/*
 * (C) Copyright 2006
 * Sven Kolberg, Maxina GmbH & Co KG, s.kolberg@maxina.de
 * Henry Margies, Maxina GmbH & Co KG, h.margies@maxina.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include "codec.h"
#include "channel.h"
#include "protocol.h"
#include "console.h"
#include "profile.h"

static struct channel channel_list[10];


/** 
 * Assigns a codec to a channel.
 * This function is called from interrupt context while
 * handling the global management channel.
 * \param idx specifies the voice channel
 * \param fmt specifies the codec to use
 * \returns zero on success
 */
int assign_channel(unsigned char idx, unsigned char fmt)
{
	struct codec* 	codec;
	struct channel*	channel;

	/* find codec in codec list */
	codec = find_codec(fmt);
	if (!codec)
		return -1;

	channel = &channel_list[idx];

	/* assign codec to channel */
	channel->codec = codec;

	/* set state to configure.. channel will be configured 
	 * later */
	channel->state = STATE_CONFIGURE;

	return 0;
}


/**
 * configure a channel
 * This function is called from process_channel. It is used
 * to initialise codecs or other stuff which should not be done
 * in interrupt context
 */
void configure_channel(struct channel* channel)
{
	PDEBUG("configure channel\n");
	channel->codec->open(channel);
	channel->state = STATE_IN_USE;
}

/**
 * Marks a channel as to be freed
 * This function is called also from interrupt context while
 * handling the management channel. It will only mark this channel
 * as to be freed.
 * \param idx specifies which channel to use
 * \returns zero on success
 */
int free_channel(unsigned char idx)
{
	struct channel*	channel;
	
	/* get channel */
	channel = &channel_list[idx];
	
	/* return an error if not in use */
	if (channel->state == STATE_UNUSED)
		return -1;

	/* mark as to be freed */
	channel->state = STATE_FREEME;
	return 0;
}


/**
 * deletes a channel which was marked as to be freed
 * Removes and deinitializes a channel. It is called from
 * process_channel.
 */
void delete_channel(struct channel* channel)
{
	PDEBUG("delete channel\n");
	channel->codec->close(channel);
	memset(channel, 0, sizeof(struct channel));
}


/**
 * Returns true if channel is ready to accept data 
 */
int channel_ready(int idx)
{
	return channel_list[idx].state == STATE_IN_USE;
}


/**
 * Reads data from sport bus to buffers
 */
void recv_channel(int idx, unsigned char* rx)
{
	struct channel* channel;
	struct ring_buf* buf;
	
	/* get channel */
	channel = &channel_list[idx];

	/* get current in-buffer for writing */
	buf = &channel->in_buf.buf[channel->in_buf.write_idx];

	/* if index is 0 we are waiting for LCT_START flag */
	if (buf->idx == 0) {
		if (*(rx+2) != LCT_START)
			return;

		if (buf->full) {
			PDEBUG("Buffer full!!!\n");
			/* ignore this frame */
			return;
		}
	}
	
	/* check for end of frame */
	if (buf->idx > 8 && *(rx+2) != LCT_CONTINUE) {
		buf->idx  = 0;
		buf->full = 1;
		channel->in_buf.write_idx++;
		channel->in_buf.write_idx%=RING_BUFFERS;
		channel->frames_received++;
		channel->frames_in_inbuf++;
		/* reread this data */
		recv_channel(idx, rx);
		return;
	}
	
	/* copy bytes */
	memcpy(&buf->data[buf->idx], rx, 2);
	buf->idx+=2;

	if (buf->idx > 8)
		return;

	/* read frame information */
	switch (buf->idx) {
		case 4: /* length */
			buf->size = *(rx+2);
			break;
		case 6: /* length */
			buf->size |= *(rx+2)<<8;
			break;
		case  8: /* samples */
			buf->samples = *(rx+2);
			break;
		case 10: /* samples */
			buf->samples |= *(rx+2)<<8;
			break;

	}


}


/**
 * Copies data from buffers to sport bus 
 */
void send_channel(int idx, unsigned char* tx)
{
	struct channel* channel;
	struct ring_buf* buf;
	
	/* get channel */
	channel = &channel_list[idx];

	/* get current in-buffer for writing */
	buf = &channel->out_buf.buf[channel->out_buf.read_idx];

	/* if size is zero, there is no data */
	if (!buf->full) {
		*(tx+2) = LCT_IDLE;
		return;
	}


	switch (buf->idx) {
		case 0:	/* LCT_START flag */
			*(tx+2) = LCT_START;
			break;

		case 2: /* length (hi) */
			*(tx+2) = buf->size >> 8;
			break;

		case 4: /* length (lo) */
			*(tx+2) = buf->size & 0xff;
			break;

		case 6: /* samples (hi) */
			*(tx+2) = buf->samples >> 8;
			break;

		case 8: /* samples (lo) */
			*(tx+2) = buf->samples & 0xff;
			break;
		default:
			*(tx+2) = LCT_CONTINUE;;
			break;
	}
			
	/* copy bytes */
	memcpy(tx, &buf->data[buf->idx], 2);
	buf->idx+=2;

	/* still data to send? */
	if (buf->idx < buf->size)
		return;

	/* end of frame */
	buf->size = 0;
	buf->idx  = 0;
	buf->full = 0;
	channel->out_buf.read_idx++;
	channel->out_buf.read_idx%=RING_BUFFERS;
	channel->frames_sent++;
	channel->frames_in_outbuf--;

}


/**
 * process one channel
 */
void process_channel(struct channel* channel)
{
	struct ring_buf* inbuf;
	struct ring_buf* outbuf;
	
	/* get current in-buffer for reading */
	inbuf = &channel->in_buf.buf[channel->in_buf.read_idx];

	if (!inbuf->full)
		return;

	/* get current out-buffer for writing */
	outbuf = &channel->out_buf.buf[channel->out_buf.write_idx];

	/* let codec calculate the data */
	channel->codec->proceed(channel, outbuf, inbuf);

	outbuf->idx  = 0;
	inbuf->size  = 0;
	inbuf->idx   = 0;
	inbuf->full  = 0;
	outbuf->full = 1;

	channel->in_buf.read_idx++;
	channel->in_buf.read_idx%=RING_BUFFERS;

	channel->out_buf.write_idx++;
	channel->out_buf.write_idx%=RING_BUFFERS;
	channel->frames_in_outbuf++;
	channel->frames_in_inbuf--;
	channel->frames_proceeded++;
}


/**
 * Go through all channels and do codec stuff
 */
void process_channels(void)
{
	int i;


	for (i=0; i<10; i++) {

		switch( channel_list[i].state ) {

			case STATE_UNUSED:
				break;

			case STATE_IN_USE:
				process_channel(&channel_list[i]);
				break;

			case STATE_CONFIGURE:
				configure_channel(&channel_list[i]);
				break;

			case STATE_FREEME:
				delete_channel(&channel_list[i]);
				print_all_profiles();
				break;
		}
	}
}


/**
 * Print channel statistic 
 */
void show_channels(void)
{
	int i;

	for (i=0; i<10; i++) {

		circ_printf("channel %d: ", i);

		switch( channel_list[i].state ) {

			case STATE_UNUSED:
				circ_printf("not in use");
				break;

			case STATE_IN_USE:
				circ_printf("in use (recv: %d proc: %d, sent: %d) frames in inbuf: %d outbuf %d srcfmt %d dstfmt %d",
					channel_list[i].frames_received,
					channel_list[i].frames_proceeded,
					channel_list[i].frames_sent,
					channel_list[i].frames_in_inbuf,
					channel_list[i].frames_in_outbuf,
					channel_list[i].codec->srcfmt,
					channel_list[i].codec->dstfmt);
				break;

			case STATE_CONFIGURE:
				circ_printf("to be configured");
				configure_channel(&channel_list[i]);
				break;

			case STATE_FREEME:
				circ_printf("to be freed");
				break;
		}

		circ_printf("\n");
	}
	circ_printf("\n");
}
