/*
 * (C) Copyright 2006
 * Sven Kolberg, Maxina GmbH & Co KG, s.kolberg@maxina.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have progress a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <malloc.h>

#include "codec.h"
#include "channel.h"
#include "console.h"
#include "protocol.h"
#include "gsm/inc/gsm.h"


static void swap16(unsigned short* b, int size)
{
	int i;
	for (i=0; i<size/2; i++,b++) {
		*b = (*b<<8) | (*b>>8);
	}
}

/**
 * GSM codec - gsm to linear
 * converts a gsm frame into a linear frame
 */
static void codec_gsm2lin_proceed(struct channel* channel, 
			struct ring_buf* dst, struct ring_buf* src)
{
	/* Gain Access to the channel data */
	gsm gsm = channel->codec_data;

	if (src->size != 33)
		circ_printf("error, frame size is %d and not 33\n", src->size);

	if (src->samples != 160)
		circ_printf("error, samples is %d\n", src->samples);

	/* Encode a frame of data */
	gsm_decode(gsm, src->data, (unsigned short*)dst->data);

	/* convert to BE */
	swap16((unsigned short*)dst->data, 320);

	dst->size    = 320;
	dst->samples = 160;
}

/**
 * GSM codec - linear to gsm
 * converts a linear frame into a gsm frame
 */
static void codec_lin2gsm_proceed(struct channel* channel, 
			struct ring_buf* dst, struct ring_buf* src)
{
	/* Gain Access to the channel data */
	gsm gsm = channel->codec_data;

	if (src->size != 320)
		circ_printf("error, frame size is %d and not 320\n", src->size);

	if (src->samples != 160)
		circ_printf("error, samples is %d\n", src->samples);

	/* convert to LE */
	swap16((unsigned short*)src->data, 320);

	/* Encode a frame of data */
	gsm_encode(gsm, (unsigned short*)src->data, dst->data);
	dst->size    = 33;
	dst->samples = 33;
}


/**
 * Allocate and initialize memory for a channel
 * \returns pointer to the memory allocated
 */
static void codec_gsm_open(struct channel* channel)
{
	channel->codec_data = gsm_create();
}


/**
 * Free the memory allocated using codec_gsm_open
 */
static void codec_gsm_close(struct channel* channel)
{
	gsm_destroy(channel->codec_data);
}


/*
 * Config-Structure 
 */
static struct codec lin2gsm_codec = {

	.name = "linear to gsm codec",

	.srcfmt = CODEC_SLINEAR,
	.dstfmt = CODEC_GSM,

	.open    = codec_gsm_open,
	.close   = codec_gsm_close,
	.proceed = codec_lin2gsm_proceed,

};

static struct codec gsm2lin_codec = {

	.name = "gsm to linear codec",

	.srcfmt = CODEC_GSM,
	.dstfmt = CODEC_SLINEAR,

	.open    = codec_gsm_open,
	.close   = codec_gsm_close,
	.proceed = codec_gsm2lin_proceed,

};

/**
 * register the GSM codec
 */
void register_codec_gsm(void)
{
	/* register the GSM-Codec, bothways */
	register_codec(&lin2gsm_codec);
	register_codec(&gsm2lin_codec);
}
