/*
 * (C) Copyright 2006
 * Sven Kolberg, Maxina GmbH & Co KG, s.kolberg@maxina.de
 * Henry Margies, Maxina GmbH & Co KG, h.margies@maxina.de
 * 
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>

/**
 * Purpose:
 * 	The Output through the console in U-Boot is a crapy piece of ...
 * 	In other words: it is blocking. it will wait until the last character has been send through 
 * 	the Serial Port, wich takes hours..., this is critical in Interrupt routines
 *
 * Solution:
 * 	Use a circular buffer wich will be emptied by an Interrupt handler or by a "user-space" routine
 */

/* The buffersize */
#define CIRCULAR_PRINTF_BUFFER_SIZE 2048

/* The maxinmum size of a printf */
#define CIRCULAR_PRINTF_SIZE 256

/* The circular console Buffer */
static unsigned char circular_console_buffer [CIRCULAR_PRINTF_BUFFER_SIZE];

/* The position of the read address, must always be smaller/equal then the write adress, but the difference must not be smaller then the CIRCULAR_BUFFER_SIZE */
static unsigned long circular_read_address = 0;

/* The position of the write address */
static unsigned long circular_write_address = 0;


/**
 * The console printf function
 * \params fmt - The Format string
 * \params ... - The Variables
 */
int circ_printf(const char *fmt, ...)
{
	va_list args;
	int i,j;
	size_t len;
	unsigned long diff;

	char buffer[CIRCULAR_PRINTF_SIZE];

	/* format the string into a buffer */
	va_start(args, fmt);
	i = vsprintf(buffer, fmt, args);
	va_end(args);

	/* check if the buffer fits into the circular buffer */
	len = strlen(buffer);
	if ((CIRCULAR_PRINTF_BUFFER_SIZE - (circular_write_address - circular_read_address) - len) < 0)
	{
		/* The String is to long, we now flush all data */
		circular_read_address = circular_write_address;

		strcpy(buffer, "WARNING: CONSOLE BUFFER FLUSHED\n");
		len = strlen(buffer);
	}

	/* insert the buffer into the circular buffer */
	for (j=0;j<len;j++)
	{
		/* copy byte by byte, implement wrap around using modular */
		circular_console_buffer[circular_write_address % CIRCULAR_PRINTF_BUFFER_SIZE] = buffer [j];
		
		/* advance the circular buffer */
		circular_write_address++;
	}

	/* adjust read and write adress */
	diff = (circular_read_address/CIRCULAR_PRINTF_BUFFER_SIZE) * CIRCULAR_PRINTF_BUFFER_SIZE;
	/* The following two instructions should be atomic FIXME */
	circular_read_address -= diff;
	circular_write_address -= diff;

	return i;
}

/**
 * The console output function which should be called
 * \returns nothing
 */
void circ_output(void)
{
	/* While characters in the buffer.. */
	while (circular_read_address < circular_write_address) {
		/* put the character to the serial port */
		serial_putc(circular_console_buffer[circular_read_address % CIRCULAR_PRINTF_BUFFER_SIZE]);
		circular_read_address++;
	}
}


void gdb_halt(void)
{
	printf("GDB_HALT\n");
}

