/*
 * (C) Copyright 2006
 * Sven Kolberg, Maxina GmbH & Co KG, s.kolberg@maxina.de
 * Henry Margies, Maxina GmbH & Co KG, h.margies@maxina.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <exports.h>

#include "dma_buffer.h"
#include "sport.h"
#include "cache.h"
#include "exceptions.h"
#include "console.h"
#include "protocol.h"
#include "channel.h"
#include "profile.h"

/* includes for the codecs, remember to call 
 * the register_codec_*() function */
#include "codec_loopback.h"
#include "codec_gsm.h"

/* include for the echoccancellation, remember to 
 * call the register_echoc_*() function */
#ifdef USE_ECHO_AEC
#include "echocan_aec.h"
#endif

#ifdef USE_ECHO_SPEEX
#include "echocan_speex.h"
#endif

static unsigned long watchdog_timer = 0;

/**
 * Is called in interrupt context after 8 bytes were received
 * This function will get the current buffer and calls the
 * real data handling function.
 */
static void transfer(void)
{
	unsigned long tx, rx;
	
	/* The pointer to the current active buffer */
	unsigned char* writeptr;
	unsigned char *readptr;

	/* The pointer to last active buffer, for sanity check  */
	static unsigned char *lastwriteptr;
	static unsigned char *lastreadptr;

	profile_pause();

#ifdef USE_DCACHE
	/* invalidate dma pointer */
	blackfin_dcache_invalidate_range(
		(unsigned char*)pDMA2_CURR_ADDR, 
		(unsigned char*)(pDMA2_CURR_ADDR+1));
#endif

	/* select which ping-pong buffer is the one we can actualy write to */
	tx = (unsigned char*)(*pDMA2_CURR_ADDR) - dma_tx_buf;

	/* create Pointer to where we can write to */
	if( tx >= (DMA_BUFFER_SIZE)) {
		writeptr = (unsigned char *) dma_tx_buf;
	} else {
		writeptr = (unsigned char *) dma_tx_buf + DMA_BUFFER_SIZE;
	}

	/* sanity check, writeptr should alternate between both possible values */
	if(writeptr == lastwriteptr) 
		circ_printf("Transmitsport didn't swap (TX)\n");
	
	/* remember the last ptr */
	lastwriteptr = writeptr;


#ifdef USE_DCACHE
	/* invalidate dma pointer */
	blackfin_dcache_invalidate_range(
		(unsigned char*)pDMA1_CURR_ADDR, 
		(unsigned char*)(pDMA1_CURR_ADDR+1));
#endif

	/* select which ping-pong buffer is the one we actually read from */
	rx = (unsigned char*)(*pDMA1_CURR_ADDR) - dma_rx_buf;

	/* create pointer where we can read from */
	if( rx >= (DMA_BUFFER_SIZE)) {
		readptr = (unsigned char *) dma_rx_buf;
	} else {
		readptr = (unsigned char *) dma_rx_buf + DMA_BUFFER_SIZE;
	}

	/* sanity check, readptr should alternate between both possible values */
	if (readptr == lastreadptr)
		circ_printf("Transmitsport didn't swap (RX)\n");
	
	/* remeber the last ptr */
	lastreadptr = readptr;
	
#ifdef USE_DCACHE
	/* invalidate receive buffer */
	blackfin_dcache_invalidate_range(readptr, readptr+DMA_BUFFER_SIZE);
#endif

	/* Handle the Communication stuff  */
	handle_data(readptr,writeptr, 8);
	watchdog_timer++;
	
#ifdef USE_DCACHE
	/* flush transmit buffer */
	blackfin_dcache_flush_range(writeptr, writeptr+DMA_BUFFER_SIZE);
#endif 

	profile_resume();
}



/**
 * The entry routine, called by the u-boot
 */
int main(int argc, char *argv[])
{
	printf ("\n\n**********************\n");
	printf (    "* Horstbox Codec DSP *\n");
	printf (    "**********************\n\n");

	/* Print the ABI version */
	app_startup(argv);
	printf ("Expected ABI version      : %d\n", XF_VERSION);
	printf ("Actual U-Boot ABI version : %d\n", (int)get_version());

	/* Check if the Version of the U-Boot is bigger/equal to us */
	if (XF_VERSION > (int)get_version()) {
		printf("Version of U-Boot is to low...\n"); 
		return 1;
	}

#ifdef USE_DEBUG_EXCEPTION_HANDLER
	/* register exception handler */
	printf("Register exception handler\n");
	if (register_exception_handler())
		printf("Failed...");

//	test_exception_handler();
#endif
	


	/* Register the Codecs */
	register_codec_loopback();
	register_codec_gsm();

#ifdef USE_ECHO_AEC
	register_echoc_aec();
#endif

#ifdef USE_ECHO_SPEEX
	register_echoc_speex();
#endif

#ifdef USE_DCACHE
	if (!dcache_status()) {
		printf("enable dcache\n");
		dcache_enable();
	}

#endif

	/* configure sport interface */
	printf("init sport\n");
	init_sport0();

	/* configure dma (do not start yet) */
	printf("init dma\n");
	init_dma_wc();

	/* initialize interrupts */
	printf("init interrupt\n");
	if (init_sport_interrupts(transfer)) {
		printf ("failed\n");
		return -1;
	}
	
	/* start dma transfer */
	printf("enable dma\n");
	enable_dma_sport0();

	/* process the codec in an endless loop */
	for(;;) {
		process_channels();

		if (watchdog_timer > 2000) {
			watchdog_timer-=2000;
#ifdef STATISTICS
			show_channels();
#endif
			circ_output();	
		}
	}

	return (0);
}




