/******************************************************************************
  Copyright(c) 2006 Sven Kolberg Germany. 

 This file is subject to the terms and conditions of the GNU Library General
 Public License. See the file "COPYING.LIB" in the main directory of this
 archive for more details.

 ******************************************************************************
  File name       :  scalar32.asm 
 
  Purpose         :  Real vector dot product for int data types.

  Description     :  This function adds each element of
		     input vector `a[]` in 4.28 to each element 
		     of input vector `b[]` in 20.12 multiplied
		     by mul in 4.28 and stores the 
		     result in a[] in 4.28.
		     a += b*mul
		     Where: 	a is A * 2^-28
		     		b is B * 2^-12
				m is M * 2^-28
				(A, B and M is a 32Bit signed integer)

		     So, e = a+b*m is 
		     	E*2^-28 = A*2^-28 + (B*2^-12 * M*2^-28) | * 2^-28

			E = A + (B*M*2^-12)

		     B*M*2^-12 in detail:

			M * B * 2^-12

			= (M.h*2^16 + M.l) * (B.h*2^16 + B.l) * 2^-12
			=   M.h*2^16 * B.h*2^16 * 2^-12
			  + M.h*2^16 * B.l      * 2^-12
			  + M.l      * B.h*2^16 * 2^-12
			  + M.l      * B.l      * 2^-12
			=   M.h * B.l * 2^20
			  + M.h * B.l * 2^4
			  + M.l * B.h * 2^4
			  + M.l * B.l * 2^-12


		 
  Registers used  :
  Operands in  R0 & R1 
  R0 - Index to vector A
  R1 - Index to vector B
  R2 - mul in 20.12
  R3 - No of elements   (16 bits)
  R3,I0,I1,P2,A0,A1

  Notes          : Input vectors should be aligned on a 4-byte address
		   boundary and be allocated in different banks to achieve 
		   the cycle count given below.

  CYCLE COUNT    : ? + N/2 where N is the number of elements

**************************************************************/
.text
.global __vect_mul_update_32
.align 2;

__vect_mul_update_32: [--SP]=(R7:0); // oush R7-R0 on stack
	I0 = R0;		// address of input vector A
	I1 = R1;		// address of input vector B
	R3 = R2;		// mul (4.28) in R3
	P2 = [SP+12+4*8] 	// fetch the size of the vector from the stack 

	/* Start of Algorithm */
	R0 = 0x0;

	LSETUP(ST_MUL_UPDATE_32, END_MUL_UPDATE_32) LC0 = P2;
ST_MUL_UPDATE_32:	

	R1 = [I1++];	// load b (20.12) in R1
	R0 = [I0];	// load a (4.28) in R5

	A0 = R3.H * R1.H (IS), A1 = R3.H * R1.L (IS,M);
	A1 += R1.H * R3.L (IS,M);
	A0 = A0 << 20;
	A1 = A1 << 4;
	A0 += A1;
	A1 = R3.L * R1.L (FU);
	A1 = A1 >>> 12;
	A0 += A1;
	R2 = A0;
	R0 = R0 + R2;

END_MUL_UPDATE_32: [I0++] = R0;

	(R7:0)=[SP++];
	RTS;

__vect_mul_update_32.end:



#if 0
__vect_mul_update_32: [--SP]=(R7:0); // oush R7-R0 on stack
	I0 = R0;	//ADDRESS OF INPUT COMPLEX VECTOR1
	I1 = R1;	//ADDRESS OF INPUT COMPLEX VECTOR2
	R3 = [SP+12+32];   // Fetch the size of the vector from the stack 
	CC = R3 <= 0;	//Checks if number of elements is zero
	IF CC JUMP RETURN_ZERO;	// zero length...

	P2 = R3;	//SET LOOP COUNTER
	CC = R3 == 1;
	IF CC JUMP SINGLE;	//IF number of elements is 1 ...

	R1 = R0 | R1;	//check that input vectors are 32 bit alligned
	R3 = 3;
	R3 = R1 & R3;
	CC = R3 == 0;
	IF !CC JUMP NOT_ALIGNED;	// should always be aligned

NOT_ALIGNED:
SINGLE:
	/* Start of Algorithm */
	R3 = R2;			// mul (4.28) in R3

	LSETUP(ST_MUL_UPDATE_32, END_MUL_UPDATE_32) LC0 = P2;
ST_MUL_UPDATE_32:	
	R1 = [I1++];	// load b (20.12) in R1
	R5 = [I0];			// load a (4.28) in R5

	A0 = R1.H * R3.H (IS), A1 = R1.H * R3.L (M, IS);      // multiply signed x unsigned
	A0 = A0 << 16;
	A1 += R3.H * R1.L (M, IS);      // multiply signed x unsigned
	A0 += A1;
	A1 = R1.L*R3.L (FU);
	A0 = A0 << 4;
	A1 = A1 >> 12;
	A0 += A1;

//	A1 = R1.L*R3.L (FU);
//	A1 = A1 >> 16;
//	A1 += R1.H * R3.L (M, IS);      // multiply signed x unsigned
//	A1 += R3.H * R1.L (M, IS);      // multiply signed x unsigned
//	A0 = R1.H * R3.H (IS);         // multiply signed x signed  
//	A0 = A0 << 16;
//	A0 += A1;
//	A0 = A0 << 4;

	R2 = A0;
	R5 = R5+R2;
END_MUL_UPDATE_32:	[I0++] = R5;	// store to vector
RETURN_ZERO:
	(R7:0)=[SP++];
	RTS;

__vect_mul_update_32.end:

#endif
