/*
 * (C) Copyright 2006
 * Sven Kolberg, Maxina GmbH & Co KG, s.kolberg@maxina.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have progress a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#ifndef __HEADER_PROFILE_H__
#define __HEADER_PROFILE_H__

#ifdef PROFILE

typedef unsigned long long profile_cycles;

/**
 * define holding structure
 */
struct profile_struct {
	char name[80];		/* The name of the profiling point */
	profile_cycles start_cycles;
	profile_cycles cycles;
	profile_cycles max_cycles;
	profile_cycles min_cycles;
	profile_cycles total_cycles;
	unsigned long num_calls;
	unsigned long user_counter;

	/* is the profiling point currently active ? */
	int active;	

	/* is this struct initialized? */
	int init;

	/* Pointer to the next profile entry */
	struct profile_struct *next; 
};

extern struct profile_struct *profile_head;
extern profile_cycles pause_time;

/**
 * read back cpu cycles 
 */
profile_cycles _read_cycles(void);

/**
 * init the profiling point and add it to the profiling list
 */
void profile_init(struct profile_struct *p, char *name); 

/**
 * remove the profile from the registration 
 */
void profile_destroy(struct profile_struct *p);
		
/**
 * start the profiling section
 */
static inline void profile_start(struct profile_struct *p)
{
	p->start_cycles = _read_cycles();
	p->active = 1;
}

/**
 * stop the profiling section
 */
static inline void profile_stop(struct profile_struct *p)
{
	/* calculate cycles */
	p->cycles = _read_cycles() - p->start_cycles;

	/* new max value ? */
	if (p->cycles > p->max_cycles)
		p->max_cycles = p->cycles;

	/* new min value? */
	if (p->cycles < p->min_cycles)
		p->min_cycles = p->cycles;

	p->total_cycles += p->cycles;
	p->num_calls++;
	p->active = 0;
}


/** 
 * pause all active channels
 * remember the time of the entry
 */
static inline void profile_pause(void)
{
	pause_time = _read_cycles();
}

/**
 * unpause all active channels
 * read the current time an update the active channels
 */
static inline void profile_resume(void)
{
	struct profile_struct *cur_profile;
	cur_profile = profile_head;
	pause_time = _read_cycles() - pause_time;
	while (cur_profile) {
		if(cur_profile->active) {
			cur_profile->start_cycles -= pause_time;
		}
		cur_profile = cur_profile->next;
	}
}

/**
 * print profile information 
 */
void print_profile(struct profile_struct *p);

/**
 * print all profile infomations
 */
void print_all_profiles(void);


#define STATIC_PROFILE_START(name) \
static struct profile_struct profile_##name = {0}; 	\
{ \
	if (!profile_##name.init) 			\
		profile_init( &profile_##name, #name); 	\
	profile_start(&profile_##name);			\
}

#define STATIC_PROFILE_STOP(name) profile_stop(&profile_##name);
	
#else

#define profile_init(A, B)	
#define profile_destroy(A)
#define profile_start(A)
#define profile_stop(A)
#define profile_pause()
#define profile_resume()
#define print_profile(A)
#define print_all_profiles()
#define STATIC_PROFILE_START(name)
#define STATIC_PROFILE_STOP(name)

#endif /* PROFILE */

#endif /* __HEADER_PROFILE_H__ */
