/*
 * (C) Copyright 2006
 * Sven Kolberg, Maxina GmbH & Co KG, s.kolberg@maxina.de
 * Henry Margies, Maxina GmbH & Co KG, h.margies@maxina.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include "console.h"
#include "protocol.h"
#include "channel.h"

/**
 * Returns the value of a register
 * \param reg the register to read
 * \returns the content of the register
 */
static unsigned char read_register(unsigned char reg)
{
	unsigned char val;

	val = 1;

	PDEBUG("reg (%d) = %d\n", reg, val); 
	return val;
}


/**
 * Writes a value to a register
 * \param reg the register to write to
 * \param val is the value
 * \returns an error code (WRITE_OK on success)
 */
static unsigned char write_register(unsigned char reg, unsigned char val)
{
	unsigned char ret = WRITE_OK;

	switch(reg) {
		/* voice registers */
		case 0: case 1:
		case 2: case 3:
		case 4: case 5:
		case 6: case 7:
		case 8: case 9:
			/* configure channel */
			if (val & CODEC_CHANNEL_ENABLE) {
				/* assign codec to channel */
				if (assign_channel(reg, val))
					ret =  WRITE_ERR_CODEC_NOT_AVIABLE;
				break;
			}
			
			/* free channel */
			if (free_channel(reg))
				ret = WRITE_ERR_CHANNEL_NOT_ASSIGNED;

			break;

		/* unsupported register */
		default:
			ret = WRITE_ERR_UNASSIGNED_REG;

	}

	PDEBUG("reg (%d) = 0x%x; ret = 0x%x\n", reg, val, ret); 

	return ret;
}


/**
 * Write cmd and value to tx buffer 
 * \param tx is the tx buffer (2 bytes will be used)
 * \param cmd is the command (left 4 bits of first byte)
 * \param reg is the register (right 4 bits of the first byte)
 * \param val is the value to write (2. byte)
 */
static inline void write_cmd(	unsigned char* tx, unsigned char cmd, 
				unsigned char reg, unsigned char val)
{
	*tx = cmd<<4 | (reg&0xf);
	*(tx+1) = val;
}


/**
 * Handles all data on the management channel.
 *   - configure voice channels
 *   - answer requests
 * \params rx pointer to the received management data
 * \params tx points to the data which will be transmited
 */
static void handle_management_channel(unsigned char *rx, unsigned char *tx)
{
	unsigned char reg, val, cmd;

	/* extract command  */
	cmd = (*rx) >> 4;

	switch (cmd) {
		case CMGMT_READ:
			/* get register */
			reg = (*rx) & 0xf;

			/* read a register value */
			val = read_register(reg);
			write_cmd(tx, CMGMT_READ_ACK, reg, val);
			break;

		case CMGMT_WRITE:
			/* get register and value */
			reg = (*rx) & 0xf;
			val = *(rx+1);

			/* write value to register */
			val = write_register(reg, val);
			write_cmd(tx, CMGMT_WRITE_ACK, reg, val);
			break;

		case CMGMT_STATUS:
			/* answer status request */
			write_cmd(tx, CMGMT_STATUS_ACK, 0x0, 0x2); /* version 0.2 */
			break;

		default:
			/* unknown command? */
			write_cmd(tx, CMGMT_UNKNOWN_CMD, cmd, 0);
	}
}


/**
 * This function handles incoming and outgoing data of the SPORT interface
 * One frame consists of 32 timeslots (each 1 byte). The first two timeslot 
 * are the global management channel. After the global management channel
 * are ten codec channels (each 3 bytes).
 *
 * Organisation of a frame:
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| mgmt| chan 0 | chan 1 | chan 2 | chan 3 | chan 4 | chan 5 | chan 6 | chan7  | chan8  | chan9  |
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *
 *  \params rx pointer to the data received from the sport
 *  \params tx pointer to the data to be transmitted by the sport
 *  \params frames number of frames in Tx and Tx
 */
void handle_data(unsigned char *rx, unsigned char *tx, unsigned int frames)
{
	int i, c;

	/* Profiling, stop the counter during interrupt, since we only 
	 * want to investigate the runtime of a section */
//	PROFILE_PAUSE();

//#ifdef USE_WATCHDOG
//	/* Advance Counter with ms precision */
//	if (ticktack++ > 2000) {
//		/* we crashed */
//		for (i=0; i<Frames; i+=32)
//			Tx[i] = 0;
//	
//		printf("DSP CRASHED\n");
//		/* print strace here... */
//
//		return;
//	}
//#else 
//	ticktack++;
//#endif

	/* iterate through all the frames */
	for (i=0; i<frames; i++) {

		handle_management_channel(rx,tx);
		tx+=2;
		rx+=2;
		for (c=0; c<10; c++) {
			if (channel_ready(c)) {
				recv_channel(c, rx);
				send_channel(c, tx);
			} else
				*(tx+2) = LCT_IDLE;

			tx+=3;
			rx+=3;
		}
	}

//	PROFILE_RESUME();
}
 

