/*
 * (C) Copyright 2006
 * Sven Kolberg, Maxina GmbH & Co KG, s.kolberg@maxina.de
 * Henry Margies, Maxina GmbH & Co KG, h.margies@maxina.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>

#include "console.h"

/* call back for interrupt */
static void (*irq_callback)(void);

/**
 * Initialize the SPORT0
 * - clockspeed is 2.048 MHz
 * - framesync is 8kHz
 * - clock is driven externally
 * - MSB first, 8 bit word length
 * - Window size 8
 * - Window offset 0
 * - 0 bit delay between FS pulse and first data bit
 * - multichannel frame mode enabled, 
 * - multichannel tx and rx DMA packing enabled 
 */
void init_sport0(void)
{
	/* set up FSYNC and optionally SCLK using Blackfin Serial port */

	*pSPORT0_RFSDIV = 255;  /* 8 kHz FSYNC with 2.048MHz PCLK  */
	*pSPORT0_TFSDIV = 255; 

	/* external tx clk, not data dependant, MSB first */
	*pSPORT0_TCR2 = 7;      /* 8 bit word length      */
	*pSPORT0_TCR1 = 0;

	/* rx enabled, MSB first, internal frame sync     */
	*pSPORT0_RCR2 = 7;      /* 8 bit word length      */
	*pSPORT0_RCR1 = 0; 

	/* Enable MCM transmit & receive channels       */
	*pSPORT0_MTCS0 = 0xffffffff;
	*pSPORT0_MRCS0 = 0xffffffff;
	
	/* MCM window size of 8 with 0 offset             */
	*pSPORT0_MCMC1 = 0x3000;

	/* 0 bit delay between FS pulse and first data bit,
	   multichannel frame mode enabled, 
	   multichannel tx and rx DMA packing enabled */
	*pSPORT0_MCMC2 = 0x001c;
}


/**
 * The Interrupt handler routine for the SPORT
 * Basicly it will clear the STATUS BIT and call a routine to handle the data
 * DO NOT CALL THIS FUNTION DIRECTLY!
 * \params irq The number of the irq wich was triggered
 * \params dev_id
 * \params regs
 * \returns IRQ_HANDLED
 */
static int sport0_isr(int irq, void *dev_id, struct pt_regs * regs)
{
	/* confirm interrupt handling, write 1 to DMA_DONE bit */
	*pDMA1_IRQ_STATUS = 0x0001;
	__builtin_bfin_ssync(); /* note without this line ints dont
                                   occur every 1ms, but get sporadic.
                                   Why?  Is it something to do with
				   next line? How could we "lose"
                                   the write to IRQ_STATUS without
				   this ssync? Maybe it just happens
				   after the ISR has finished sometimes
				   which messes things up */

	*pFIO_FLAG_S = (1<<12);
	__builtin_bfin_ssync();
	

	irq_callback();
	
	/*FIXME*/
	*pFIO_FLAG_C = (1<<12);
	__builtin_bfin_ssync();

	return 0;
}
 
/**
 * Initialize the Interrupt of SPORT RX
 * \returns zero uppon success, nonzero otherwise
 */
int init_sport_interrupts(void(*handler)(void))
{
	irq_callback = handler;

	/* Register to IVG9 and fire wenn IRQ_SPORT0 is asserted*/
	if (request_irq(IRQ_SPORT0, sport0_isr) != 0) {
    		return -1;
	}

	/* enable IVR 9 in IMASK*/
	*(unsigned volatile long *)IMASK |= (1<<9);

	/* enable DMA1 sport0 Rx interrupt */
	*pSIC_IMASK |= 0x00000200;

	__builtin_bfin_ssync();

	return 0;
}
