/** \file main.c
 *
 * Main routines for bfloader backend
 *
 * based on VDSP compatible backends provided by
 *      (c) Analog Devices, Inc. - 2001
 *
 * based on the u-boot cfi-flash_driver provided by
 * 	(C) Copyright 2002-2004
 * 		Brad Kemp, Seranoa Networks, Brad.Kemp@seranoa.com
 * 	Copyright (C) 2003 Arabella Software Ltd.
 * 		Yuli Barcohen <yuli@arabellasw.com>
 * 		Modified to work with AMD flashes
 * 	Copyright (C) 2004
 * 	        Ed Okerson
 * 	        Modified to work with little-endian systems.
 *
 *
 * This supports the following boards:
 *
 * \li Horstbox
 * \li STAMP-BF533
 * \li STAMP-BF537
 * \li EZKIT-BF537
 *
 * Modified and bug fixed for bfloader
 * Also removed unreadable hungarian variable notation
 * 16 bit Flash chips only!
 *    Martin Strubel <hackfin@section5.ch>
 *
 * Modified and bug fixed to work with cfi compliant flash
 *
 * This code is provided 'as is' without warranty. Use on your
 * own risk.
 *
 *
 */


#include "error.h"
#include "bfloader.h"
#include "flash.h"
#include <cdefBF533.h>

#define FLASH_BASE      0x20000000

#define BUFFER_SIZE		0x2000
//#define NUM_SECTORS		128	//Intel 28F128

////////////////////////////////////////////////////////////////////////////
// GLOBALS 
//

static
unsigned short g_buffer[BUFFER_SIZE >> 1];

// DATA NEEDED BY FRONTEND:

char           *AFP_Title           = "Horst CFI Flash driver";
char           *AFP_Description     = "CFI FLASH";
ProgCmd         AFP_Command         = NO_COMMAND;
int             AFP_ManCode         = -1;
int             AFP_DevCode         = -1;
unsigned long   AFP_Offset          = 0x0;
short          *AFP_Buffer;
long            AFP_Size            = BUFFER_SIZE;
long            AFP_Count           = -1;
long            AFP_Stride          = 1;
int             AFP_NumSectors      = -1;
int             AFP_Sector          = -1;
int             AFP_Error           = 0;            // error code
int             AFP_Verify          = 0;        // verify flag
unsigned long   AFP_StartOff        = 0x0;          // sector start offset
unsigned long   AFP_EndOff          = 0x0;          // sector end offset
int             AFP_FlashWidth      = 0x10;         // width of the flash
//int            *AFP_SectorInfo;

//SectorLocation SectorInfo[NUM_SECTORS];

/**
 * Initialize the Memory interface
 * Change here to make it suit to your board
 * \returns allways NO_ERR
 */
int memory_interface_init()
{
	*pEBIU_AMGCTL  = 0xff;
	*pEBIU_AMBCTL0 = 0x7bb07bb0;
	*pEBIU_AMBCTL1 = 0x7bb07bb0;
	return NO_ERR;
}

/**
 * copys data from Flash to the buffer data
 * \params info The Flash information structure for the flash we want to read from
 * \params start The First Word we want to read from
 * \params count The number of bytes we want to read
 * \params data The pointer to a buffer we shall write to
 * \returns usually NO_ERR
 */
int flash_readchunk(flash_info_t * info, unsigned long start, long count, short *data)
{
	uchar *src;
	uchar *dst;

	dst = (uchar *) data;
	src = (uchar *) start;
	memcpy(dst, src, count);
        return NO_ERR;
}
                


////////////////////////////////////////////////////////////////////////////
// MAIN

int main()
{

	AFP_Buffer = g_buffer;

	/* initialize the memory interface */
	memory_interface_init();

	/* initialize the cfi Flash */
	flash_init();

	/* get the Device and Manufacturer code */
	AFP_ManCode         = flash_info[0].vendor;
	AFP_DevCode         = flash_info[0].flash_id;
	AFP_NumSectors      = flash_info[0].sector_count;
	AFP_Sector          = -1;
	AFP_FlashWidth      = 8*flash_info[0].chipwidth;         // width of the flash


	// main loop
	while (1) {
		// the plug-in will set a breakpoint at "AFP_BreakReady" so it knows
		// when we are ready for a new command because the DSP will halt
		//
		// the jump is used so that the label will be part of the debug
		// information in the driver image otherwise it may be left out
		// since the label is not referenced anywhere
		asm("AFP_BreakReady:");
		asm("nop;");
		if (0)
			asm("jump AFP_BreakReady;");

		switch (AFP_Command)
		{
			case GET_CODES:
				AFP_ManCode         = flash_info[0].vendor;
				AFP_DevCode         = flash_info[0].flash_id;
				AFP_NumSectors      = flash_info[0].sector_count;
				AFP_Sector          = -1;
				AFP_FlashWidth      = 8*flash_info[0].chipwidth;         // width of the flash
				AFP_Error = NO_ERR;
				break;
			
			case RESET:
				//Tested, does work
				AFP_Error = flash_reset(&(flash_info[0]));
				break;
			
			case WRITE:
				switch( write_buff (&(flash_info[0]), AFP_Buffer, AFP_Offset, AFP_Count))
				{
					case 0:
						AFP_Error = NO_ERR;
						break;
					case 1:
						AFP_Error = POLL_TIMEOUT;
						break;
					case 2:
					default:
						AFP_Error = FLASH_NOT_ERASED;
						break;

				}
				break;
			case ERASE_ALL:
				flash_erase(&(flash_info[0]), 0, flash_info[0].sector_count-1);
				break;
			case ERASE_SECT:
				switch(flash_erase(&(flash_info[0]), AFP_Sector, AFP_Sector)) 
				{
					case 0:
						AFP_Error = NO_ERR;
						break;
					default:
						AFP_Error = POLL_TIMEOUT;
						break;
				}
				break;
			case READ:
				//Tested, does work
				AFP_Error = flash_readchunk(&(flash_info[0]), AFP_Offset, AFP_Count, AFP_Buffer);
				break;
			case GET_SECTNUM:
				AFP_Sector = find_sector(&(flash_info[0]),AFP_Offset);
				AFP_Error = NO_ERR;
				break;
			case NO_COMMAND:
			default:
				AFP_Error = UNKNOWN_COMMAND;
				break;
		}
		AFP_Command = NO_COMMAND;
	}
	return 0;
}


