/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Henry Margies <h.margies@maxina.de> - Maxina GmbH, Neu-       *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/

#include "lan_common.h"


/**
 * HLST call back function
 */
void nvram_avp_destroy (void   *state,
			void    *item,
			char     *key,
			unsigned klen) 
{
	if (item)
		free (item);
}


/**
 * Create list with values from nvram
 */
HLST* create_from_nvram(NVCL* nvram) 
{
	int slot, used, n, total ;
	const char *var, *val, **Val ;
	HLST* avp;

	nvcl_putcmd(nvram, "slots");
	if (nvcl_rply(nvram, 0) < 1) {
		log_msg(LOG_ERR, "Could not transmit data to nvram");
		nvcl_close(nvram);
		return NULL;
	}

	if (nvcl_getavpair(nvram, 0, &val, 0, 0) == 0) {
		log_msg(LOG_ERR, "Did not get slot value");
		nvcl_close(nvram);
		return NULL;
	}

	used = atoi(val);

	if (nvcl_getavpair(nvram, 0, &val, 0, 0) == 0) {
		log_msg(LOG_ERR, "Did not get slot value");
		nvcl_close(nvram);
		return NULL;
	}

	total = atoi(val);

	/* create list */
	avp = create_hlst(total, nvram_avp_destroy, NULL);
	if (!avp) {
		log_msg(LOG_ERR, "Unable to create list");
		nvcl_close(nvram);
		return NULL;
	}

	for (slot = 1, n = used; slot <= total && n >= 0; slot ++) {
		
		char *p, *q, t [20];

        	sprintf(t, "%u", slot);
		nvcl_putcmd(nvram, "get");
		nvcl_putavpair(nvram, t, 0, 0);

		if (nvcl_rply(nvram, 0) < 1)
			continue ;

		while (nvcl_getavpair(nvram, &var, &val, 0, 0) != 0) {

			n--;

			/* no value -> ignore */
			if (!val)
				continue;

			/* copy value */
			q = strdup(val);

			/* copy name */
			p = strdup(var);

			/* store to our list */
			Val = (const char **)make_hlst (avp, p, 0);
			/* check if key does not exist */
			if (Val) {
				/* set the variable contents */
				*Val = q;
			} else
				free(q);
			
			/* clean up */
			free (p);
		}
	}
	
	return avp;
}

/**
 * Call cb function for all values which key starts with prefix
 */
void for_each(HLST* avp, const char* prefix, cb_t cb, void* data)
{
	char **Val, *key;
	HSRCH *lookup = open_hlst_search(avp);
	
	while ((Val = (char**)next_hlst_search(lookup))) {

		if (!Val || !*Val || !**Val)
			continue;

		key = query_key_hlst((void **)Val);

		if (strncmp(prefix, key, strlen(prefix)))
			continue;

		cb(data, *Val);
	}
	close_hlst_search(lookup);
}

#if 0
/**
 * Call cb function for all values which key starts with prefix
 */
void for_seq(HLST* avp, const char* prefix, cb_t cb, void* data)
{
	char* total;
	int i;

	total = get_indexed_value(avp, prefix, "totalElements");
	if (!total)
		return;

	for (i=1; i<=atoi(total); i++) {
		char buf[32];

		snprintf(buf, sizeof(buf), "%d", i);
		cb(data, buf);
	}

}
#endif
	
char* get_list_value(HLST* avp, const char* name)
{
	char** val;

	val = (char**)find_hlst(avp, name, 0);
	
	if (!val)
		return NULL;
	return *val;
}


/**
 * Returns a pointer to a value which key is name.index
 */
char* get_indexed_value(HLST* avp, const char* name, const char* index)
{
	char* val;
	char* full_name = (char*)malloc(strlen(name)+strlen(index)+2);

	strcpy(full_name, name);
	strcat(full_name, ".");
	strcat(full_name, index);

	val = get_list_value(avp, full_name);
	
	free(full_name);

	return val;
}

int id_is_in(const char* s, const char* id) 
{
	char* bf = (char*)malloc(strlen(s)+3);
	char* sb = (char*)malloc(strlen(id)+3);
	char* r;
	
	strcpy(bf, " ");
	strcat(bf, s);
	strcat(bf, " ");

	strcpy(sb, " ");
	strcat(sb, id);
	strcat(sb, " ");

	r = strstr(bf, sb);

	free(bf);
	free(sb);

	return (r)?1:0;
}



void update_from_nvram(NVCL* nvram, HLST* avp, const char* varname)
{
	const char *var, *val, **Val;

	delete_hlst(avp, varname, 0);	

	nvcl_putcmd(nvram, "set");
	nvcl_putavpair(nvram, varname, 0, 0);

	if (nvcl_rply(nvram, 0) < 1) {
		log_msg(LOG_ERR, "Error update from nvram");
		return;
	}

	nvcl_getavpair(nvram, &var, &val, 0, 0);

	Val = (const char**) find_hlst(avp, varname, 0);
	if (!Val)
		Val = (const char **)make_hlst(avp, varname, 0);
	
	*Val = strdup(val);	
}


void update_to_nvram(NVCL* nvram, HLST* avp, const char* var, const char* val)
{
	if (val)
		nvcl_putcmd(nvram, "set");
	else {
		nvcl_putcmd(nvram, "del");
		val = NULL;
	}

	nvcl_putavpair(nvram, var, val, 0);

	if (nvcl_rply(nvram, 0) < 1) {
		log_msg(LOG_ERR, "Error update to nvram");
		return;
	}

	update_from_nvram(nvram, avp, var);
}


int is_true(HLST* avp, const char* name)
{
	char* value;

	value = get_list_value(avp, name);

	if (!value)
		return 0;

	if (!*value)
		return 0;

	if (strcmp(value, "true"))
		return 0;

	return 1;
}

#ifdef SET_LINKS_AND_TRIGGER
void nvram_link_indexed(NVCL* nvram, const char* name, const char* id, const char* mode, const char* target)
{
	char* var = (char*)malloc(strlen(name) + strlen(id) + 2);

	strcpy(var, name);
	strcat(var, ".");
	strcat(var, id);

	nvram_link(nvram, var, mode, target);
	free(var);
	
}

void nvram_unlink(NVCL* nvram, const char* name)
{
	const char *var, *val;
	size_t vlen;
	
	nvcl_putcmd(nvram, "unlink");
	nvcl_putavpair(nvram, name, 0, 0);

	if (nvcl_rply(nvram, 0) < 1) {
		log_msg(LOG_ERR, "Error removing symlink");
		return;
	}

        while (nvcl_getavpair (nvram, &var, &val, &vlen, 0) != 0) {}
}

void nvram_link(NVCL* nvram, const char* name, const char* mode, const char* target)
{
	const char *var, *val;
	size_t vlen;

	nvcl_putcmd(nvram, "link");
	nvcl_putavpair(nvram, name, mode, 0);
	nvcl_putavpair(nvram, target, 0, 0);

//	log_msg(LOG_NOTICE, "nvram -c link <%s> <%s> <%s>", name, mode, target);

	if (nvcl_rply(nvram, 0) < 1) {
		log_msg(LOG_ERR, "Error setting symlink");
		return;
	}

        while (nvcl_getavpair (nvram, &var, &val, &vlen, 0) != 0) {}

}

void nvram_trigger(NVCL* nvram, const char* name, const char* mode, const char* target)
{
	const char *var, *val;
	size_t vlen;
	
	nvcl_putcmd(nvram, "trigger");
	nvcl_putavpair(nvram, name, mode, 0);
	nvcl_putavpair(nvram, "+", target, 0);

//	log_msg(LOG_NOTICE, "nvram -c trigger %s %s + %s", name, mode, target);

	if (nvcl_rply(nvram, 0) < 1) {
		log_msg(LOG_ERR, "Error setting trigger");
		return;
	}

        while (nvcl_getavpair (nvram, &var, &val, &vlen, 0) != 0) {}
}


#endif


/**
 * misc functions
 */
char* replace_char(const char* string, char find, char replace)
{
	char *c;

	if (!string)
		return NULL;

	while((c=strchr(string, find))) {
		*c = replace;
	}
	return string;
}

