/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#define _GNU_SOURCE
#include <getopt.h>

/* ----------------------------------------------------------------------- *
 * Private: definitions & declarations
 * ----------------------------------------------------------------------- */

static const
struct option long_options [] = {
    {"mask",  0, 0, 'm'},
    {"width", 0, 0, 'w'},
    {"any",   0, 0, 'a'},
    {0, 0, 0, 0}
};

static const
char short_options [] = "+mwa";
 
/* ----------------------------------------------------------------------- *
 * Library
 * ----------------------------------------------------------------------- */

static
const char *width2mask (unsigned wid) {
    struct in_addr in;
    in.s_addr = htonl (~((1L << (32  - wid)) - 1));
    return inet_ntoa (in);
}       

static
const char *mask2width (uint32_t msk) {
    static char wds [3] ;
    uint32_t m = 0xffffffff ;
    int n = 32, w = 32 ;

    switch (msk) {
    case          0: return  "0" ;
    case 0xff000000: return  "8" ;
    case 0xffff0000: return "16" ;
    case 0xffffff00: return "24" ;
    case 0xffffffff: return "32" ;
    }

    for (;;) {
        if ((msk & m) == 0) {
            n    -= w ;
            msk >>= w ;
        }
        if ((w >>= 1) == 0)
            break ;
        m >>= w ;
    }

    sprintf (wds, "%d", n);

    m = 0xffffffff ;
    n = w = 32 ;

    for (;;) {
        if ((msk & m) == m)
            msk >>= w ;
        if ((w >>= 1) == 0)
            break ;
        m >>= w ;
    }
        
    return msk ? 0 : wds ;
}

/* ----------------------------------------------------------------------- *
 * Public: MAIN
 * ----------------------------------------------------------------------- */

static char *pname, mode = 'a' ;

static
void usage (void) {
    fprintf (stderr,
             ("Usage: %s [options] netspec ...\n"
              "Options: --mask,-m    netspec is an octet mask\n"
              "         --width,-w   netspec is a width argument\n"
              "         --any,-a     netspec can be chosen arbitrarily\n"
              "Default netspec mode is -%c\n"), pname, mode) ;
    exit (2);
}

main (int argc, char **argv) {
    int c, option_inx, raw = 0 ;

    if ((pname = strrchr (argv [0], '/')) == 0)
        pname = argv [0] ;
    else
        pname ++ ;

    /* parse opotions */
    while ((c = getopt_long (argc, argv, short_options, 
                             long_options, &option_inx)) != -1)
        switch (c) {
        case 'w':
        case 'm':
        case 'a':
            mode = c ;
            continue ;
        case '?':
            usage ();
        }

    if (optind >= argc)
	usage ();


    do {
        struct in_addr in;
        const char *s ;
        char *p ;
        long l ;

        if ((l = strtol (argv [optind], &p, 0)) < 33 && l >= 0 &&
            p [0] == '\0' && (s = width2mask (l)) != 0) {
            if (mode == 'm') goto not_a_netwidth ;
            printf ("%s\n", s);
            continue ;
        }

        if (inet_aton (argv [optind], &in) != 0 &&
            (s = mask2width (ntohl (in.s_addr))) != 0) {
            if (mode == 'w') goto not_a_netmask ;
            printf ("%s\n", s);
            continue ;
        }

        switch (mode) {
        case 'w': goto not_a_netwidth ;
        case 'm': goto not_a_netmask ;
        default:  goto neither_mask_nor_width ;
        }
    }
    while (++ optind < argc) ;

    exit (0);

neither_mask_nor_width:
    fprintf (stderr,
             "%s: Error: argument %s is neithen a netmask nor a netwidth.\n",
             pname, argv [optind]);
    exit (2);

not_a_netwidth:
    fprintf (stderr,
             "%s: Error: argument %s is not a netwidth.\n",
             pname, argv [optind]);
    exit (2);

not_a_netmask:
    fprintf (stderr,
             "%s: Error: argument %s is not a netmask.\n",
             pname, argv [optind]);
    exit (2);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
