/*
 * BiST: A telefonie hardware testing application
 *
 * Written by  Sven Kolberg <s.kolberg@maxina.de>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <zaptel.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/fcntl.h>

//#include <zap.h>

#include "Analog_Func.h"
#include "Analog_FXS.h"

#ifdef DEBUG
	#define PDEBUG(frmt, ...) \
		{\
		fprintf(stdout, "file: %s function: %s line: %i \n",__FILE__, __PRETTY_FUNCTION__, __LINE__); \
		fprintf(stdout, frmt, ##__VA_ARGS__);\
		fprintf(stdout, "\n");\
		}
#else 
	#define PDEBUG(frmt, ...)
#endif




	
/**
 * \brief Handles the Event checking on the FXS
 * \params a pointer to the file handler
 */
static void *fxs_handle_channel(void *data)
{
	/* The File descriptor of the Zaptel device */
	Analog_chan *achan = (Analog_chan *)data;
	int res;

	PDEBUG("File descriptor %d\n", achan->fd);
	while(1) 
	{
		/* check for events */
		res = zaptel_check_event(achan->fd);
		
		
		switch(res) {
			case ZT_EVENT_ONHOOK:  /* hung up */
				PDEBUG("ZT_EVENT_ONHOOK");
				/* Callbackroutine, to notify the switching-logic*/
				if(achan->call_hangup) achan->call_hangup(achan);
				break;
			case ZT_EVENT_RINGOFFHOOK:	  /* rang in or answered */
				PDEBUG("ZT_EVENT_RINGOFFHOOK");
				{
					int i = ZT_RINGOFF;  /* turn ringing seq. off */
					  /* if error, return as such */
					if(ioctl(achan->fd,ZT_HOOK,&i) == -1) return NULL;
				}
				/* Callbackroutine, to notify the switching-logic */
				if(achan->call_answered) achan->call_answered(achan);
				break;
			case ZT_EVENT_WINKFLASH:  /* got wink */
			    	PDEBUG("ZT_EVENT_WINKFLASH");
				break;
			case ZT_EVENT_ALARM:  /* got alarm */
			    	PDEBUG("ZT_EVENT_ALARM");
				break;
			case ZT_EVENT_NOALARM: /* alarm went away */
				PDEBUG("ZT_EVENT_NOALARM");
				break;
			case ZT_EVENT_ABORT: /* got HDLC abort */
			    	PDEBUG("ZT_EVENT_ABORT");
				break;
			case ZT_EVENT_OVERRUN: /* got HDLC overrun */
			    	PDEBUG("ZT_EVENT_OVERRUN");
				break;
			case ZT_EVENT_BADFCS: /* got HDLC bad FCS */
			    	PDEBUG("ZT_EVENT_BADFCS");
				break;
			case ZT_EVENT_DIALCOMPLETE: /* dial complete */
			    	PDEBUG("ZT_EVENT_DIALCOMPLETE");
				break;
			case ZT_EVENT_RINGERON: /* ringer on */
			    	PDEBUG("ZT_EVENT_RINGERON");
				break;
			case ZT_EVENT_RINGEROFF: /* ringer off */
			    	PDEBUG("ZT_EVENT_RINGEROFF");
				break;
			case ZT_EVENT_HOOKCOMPLETE: /* hook change complete */
			    	PDEBUG("ZT_EVENT_HOOKCOMPLETE");
				break;
			case ZT_EVENT_NONE:
				break;
			default: 
				if(res & (ZT_EVENT_PULSEDIGIT | ZT_EVENT_DTMFDOWN | ZT_EVENT_DTMFUP))
				{
					/* Call the Callbackroutine to notify the switching logic that a digit was recieved */
					if(achan->number_recieved) achan->number_recieved(achan, res & 0xff);
				}

				/* DEBUGing output */
				if (res & ZT_EVENT_PULSEDIGIT) 
				{
					PDEBUG("ZT_EVENT_PULSEDIGIT %c", res & 0xff);
				} 
				else if (res & ZT_EVENT_DTMFDOWN)
				{
					PDEBUG("ZT_EVENT_DTMFDOWN %c", res & 0xff);
				}
				else if (res & ZT_EVENT_DTMFUP) 
				{
					PDEBUG("ZT_EVENT_DTMFUP %c", res & 0xff);
				}
				else 
				{
					//PDEBUG("other")
					PDEBUG("Unknown Event %d\n", res);
				}
				break;
		}
	}
	return NULL;
}
	
Analog_chan *init_fxs(int channel)
{
	Analog_chan *achan;
	ZT_PARAMS p;
	ZT_BUFFERINFO bi;
	ZT_GAINS gains;
	ZT_SPANINFO spaninfo;
	
	int state;
	int tonezone;
	pthread_t thread;

	/*Get Memory for achan*/
	achan = malloc(sizeof(Analog_chan));
			
	/* at first open the device */
	achan->fd = zaptel_open(channel);
	
	if (achan->fd==-1) {
		PDEBUG("unable to open zapteldevice %i", channel);
		return NULL;
	}


	
	/* get Params */
	if(ioctl(achan->fd, ZT_GET_PARAMS, &p)) {
		PDEBUG("unable to get Parameters");
	}

	/* manipulate Params */
	//Not needed 
	

	/*set Params */
	if(ioctl(achan->fd, ZT_SET_PARAMS, &p)) {
		PDEBUG("unable to Set Parameters");
	}


	/* get Buffinfo */
	if(ioctl(achan->fd, ZT_GET_BUFINFO, &bi)) {
		PDEBUG("unable to Get Buffinfo");
	}

	/* manipulate Buffinfo */
	bi.numbufs = 4;

	/* set Buffinfo */
	if(ioctl(achan->fd, ZT_SET_BUFINFO, &bi)) {
		PDEBUG("unable to Set Buffinfo");
	}
	
	/* Get Gain */
	gains.chan = channel;
	if(ioctl(achan->fd, ZT_GETGAINS, &gains)) {
		PDEBUG("unabel to Get Gains");
	}
	
	/* Adjust Gain */
	//not needed linear is OK

	/*set Gain*/
	gains.chan = channel;
	if(ioctl(achan->fd, ZT_SETGAINS, &gains)) {
		PDEBUG("unable to set gains");
	}

	/*Set Hook*/
	state = ZT_ONHOOK;
	if(ioctl(achan->fd, ZT_HOOK, &state)) {
		PDEBUG("unable to set the Linestatus to ONHOOK");
	}

	
	/* Set the Tonezone */
	tonezone = -1;
	if (ioctl(achan->fd, ZT_SETTONEZONE, &tonezone)) {
		PDEBUG("unable to set tonezone");
	}

	/* Get Span Stats */
	if(ioctl(achan->fd, ZT_SPANSTAT, &spaninfo)) {
		PDEBUG("unable to set spaninfo");
	}

	/* Start Thread with the Start function fxs_handle_channel an pass the File handler onwards */
	if (pthread_create(&thread, NULL, fxs_handle_channel, achan)) 
	{
		PDEBUG("Unable to create Thread for fxs_handle_channel");
		return NULL;

	} 
	else 
	{
		sprintf(achan->description, "FXS(%i)", channel);
		return achan; /* The File descriptor to the caller */
	}
}

/**
 * \brief Ring a channel for nrings times
 * \params achan the Channel where we should ring
 * \returns 0 after ringing nrings times, 1 if the line got answerded, -1 on error
 */

static int fxs_ring(Analog_chan *achan)
{
	int	i,n;

	n = 0; /* init counter */

	i = ZT_RINGOFF;  /* start fresh */
	/* this may return error if not presently in ringing state. its ok. */
	ioctl(achan->fd,ZT_HOOK,&i);
	
	i = ZT_RING;  /* ring phone once */
	/* if error, return as such */
	if(ioctl(achan->fd,ZT_HOOK,&i) == -1) ;//return(-1);
	
	return(0);

}

/**
 * \brief Start a call by ringing the line contiously only stopped by gooing off-hook
 * \params achan the channel we like to place a call
 * \returns 0 upon success, 1 if already offhook (no need to ring if offhock), other values indicate an error
 */
int fxs_start_call(Analog_chan *achan)
{
	ZT_PARAMS p;
	
	/* Ring the phone */
	p.channo = 0;
	
	if(ioctl(achan->fd, ZT_GET_PARAMS, &p)) {
		PDEBUG("Error Geting Parameters");
		return -1;
	}
	
	if(p.rxisoffhook) {
		PDEBUG("OFF_HOOK");

		/* Goto OFFHOOK State, Make sure to turn ringing off */
		{
			int i = ZT_OFFHOOK;
			ioctl(achan->fd,ZT_HOOK,&i);
			/* ignore the return value of ZT_HOOK, since it returns an error if allready OFFHOOK,
			 * so everything is ok (hope so)
			 */
		}
		return 1;
	}
	
	/* let it ring, one time */
#ifdef FXS_CLID 
	if (firsttime)
	{
		firsttime = 0;
		zap_ringclid(zap, "1234", "Prank Call");/*use callerid*/
	}
	else
#endif
		fxs_ring(achan);

	

	/*We where answered*/
	return 0;
}

/**
 * \brief Stopps a call by going on-hook
 * \params achan the channel we like to stop a call
 * \returns 0 upon success, 1 if already onhook, other values indicate an error
 */
int fxs_stop_call(Analog_chan *achan)
{
	ZT_PARAMS p;
	
	/* Ring the phone */
	p.channo = 0;
	
	if(ioctl(achan->fd, ZT_GET_PARAMS, &p)) {
		PDEBUG("Error Geting Parameters");
		return -1;
	}
	
	if(!p.rxisoffhook) {
		PDEBUG("Allready OnHook");
		return 1;
	}
	
	/* Goto OFFHOOK State, Make sure to turn ringing off */
	{
		int i = ZT_ONHOOK;
		ioctl(achan->fd,ZT_HOOK,&i);
		/* ignore the return value of ZT_HOOK, since it returns an error if allready OFFHOOK,
		 * so everything is ok (hope so)
		 */
	}

	return 0;
}

