/*
 * BiST: A telefonie hardware testing application
 *
 * Written by  Sven Kolberg <s.kolberg@maxina.de>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#if defined(__linux__)
#include <linux/zaptel.h>
#elif defined(__FreeBSD__)
#include <zaptel.h>
#endif
#include <pthread.h>
#include <sys/select.h>

#include <libpri.h>
#include <pri_internal.h>

#include "ISDN_Func.h"
#include "ISDN_Network.h"
#include "ISDN_CPE.h"
#include "Analog_FXS.h"
#include "Analog_FXO.h"


/* Pointer to the structure of a device which should accept the call 
 * Will be setung using <to>
 */
void *call_acceptor = NULL;

/* Pointer to the originator of a call
 */
void *call_originator = NULL;

/* How long should it ring (e.g. how many cadences)
 */
int  take_rings = 1;

/* Who did answer the line?
 */
void *line_answered = NULL;

static void help(void)
{
	fprintf(stderr, "\nBiST [-t timeout] [-r <ringer>] <from> <to>\n"
			" <ringer> <from> <to> can by FXO, FXS1, FXS2, "
			"ISDNNT, ISDNTE\n\n");
}

static char* extract_number(char* device) 
{
	char* c;

	if (!device)
		return NULL;

	c = strchr(device, '/');
	if (!c)
		return NULL;

	/* move pointer to character after "/" and check if it is
	 * not zero */
	if (!*(++c))
		return NULL;

	return c;
}

static int extract_dev(char* device) 
{
	int i;
	char* dev_list[] = {
		"UNKNOWN",
		"FXO",
		"FXS1",
		"FXS2",
		"ISDNNT",
		"ISDNTE",
		NULL };

	if (!device)
		return 0;

	for (i=0; dev_list[i]; i++) {
		
		if (strncasecmp(device, dev_list[i], strlen(dev_list[i])))
			continue;

		return i;
	}

	return 0;
}

//Callbackroutine, will be called when a call is answered
static int call_answered(struct sanalog_chan *achan)
{
	fprintf(stdout, "Call was answered on %s\n", achan->description);
	if (call_originator != achan) line_answered = achan;

	return 0;
}

//Callbackroutine, will be called if a call is hungup
static int call_hangup(struct sanalog_chan *achan)
{
	fprintf(stdout, "Call was hung up on %s\n", achan->description);

	return 0;
}

//Callbackroutine, if a digit was recieved
static int number_recieved(struct sanalog_chan *achan, char digit)
{
	fprintf(stdout, "Digit \"%c\" on %s recieved\n", digit, achan->description);

	return 0;
}

//Callbackroutine, will be called during ringing, num is the number or rings allready recieved
static int ringing(struct sanalog_chan *achan, int num)
{
	fprintf(stdout, "Ringing on %s the %ith time\n", achan->description, num);

	if((call_acceptor == achan) && (take_rings <= num)) 
	{
		/*We have to acept the call */
		fprintf(stdout, "Accept the call on %s\n", achan->description);
		line_answered = achan;
		return 1;
	}
	else
	{
		/*We do not handle the call */
		return 0;
	}
}

//Will be called when a call is answered
static int isdn_call_answered (struct isdn_span_info *span)
{
	fprintf(stdout, "Call on %s answered\n", span->description);
	if (call_originator != span) line_answered = span;

	return 0;
}

//Will be called when a call is terminated
static int isdn_call_hangup (struct isdn_span_info *span)
{
	fprintf(stdout, "Call on %s hungup\n", span->description);

	return 0;
}

//Will be called when it's ringing
static int isdn_ringing (struct isdn_span_info *span, int nrings)
{
	fprintf(stdout, "Ringing on %s the %ith time\n", span->description, nrings);

	if((call_acceptor == span) && (take_rings <= nrings)) 
	{
		/*We have to acept the call */
		fprintf(stdout, "Accept the call on ISDN\n");
		line_answered = span;
		return 1;
	}
	else
	{
		/*We do not handle the call */
		return 0;
	}
}

int main(int argc, char *argv[])
{

	char 	params[] = "hr:t:";
	int 	i;
	int	ret;

	int timeout  = 60;
	int dev_from = 0;
	int dev_to   = 0;
	char* number = NULL;

	int ring_dev_list[8] = {};
	int ring_dev_counter = 0;

	while(1) {
		i = getopt(argc, argv, params);
		if (i == -1) 
			break;

		switch(i) {
			
			case 'r': /* ringer device */
				if (ring_dev_counter > sizeof(ring_dev_list)/sizeof(ring_dev_list[0])) {
					fprintf(stderr, "Maximum number of ring devices reached\n");
					break;
				}

				ret = extract_dev(optarg);
				if (!ret) {
					fprintf(stderr, "Invalid device %s\n", optarg);
					break;
				}
				ring_dev_list[ring_dev_counter++] = ret;

				break;				

			case 't': /* timeout */
				timeout = atoi(optarg);
				break;

			case 'h':
				help();
				return 0;
		}
	}

	if (!argv[optind] || !argv[optind+1]) {
		fprintf(stderr, "You must specify 'from' and 'to' device\n");
		help();
		return -1;
	}

	dev_from = extract_dev(argv[optind]);
	number   = extract_number(argv[optind]);
	dev_to   = extract_dev(argv[optind+1]);

	if (!dev_from) {
		fprintf(stderr, "Error parsing %s\n", argv[optind]);
		help();
		return -2;
	}


	if (!number) {
		fprintf(stderr, "Invalid number in %s\n", argv[optind]);
		help();
		return -3;
	}

	if (!dev_to) {
		fprintf(stderr, "Error parsing %s\n", argv[optind+1]);
		help();
		return -4;
	}

#ifdef DEBUG
	printf("timeout = %d\n", timeout);
	printf("from = %d\n", dev_from);
	printf("to = %d\n", dev_to);
	printf("number = %s\n", number);
	for (i=0; i< ring_dev_counter; i++) {
		printf("ringer dev [%d] = %d\n", i, ring_dev_list[i]);
	}
#endif


	/*Init all devices*/
	ISDN_SPAN_INFO *nt_span_info, *te_span_info;

	/* Start the TE */
		if(!(te_span_info = isdn_te_init_dchannel(6,2)))
		{
			fprintf(stdout, "Unable to initialize the ISDN TE on the Testdevice\n");
			exit(-5);
		}
	
		/* Assign the callbackroutines */
		te_span_info->call_answered	= isdn_call_answered;
		te_span_info->call_hangup	= isdn_call_hangup;
		te_span_info->ringing		= isdn_ringing;

	/* Start the NT */	
		if(!(nt_span_info = isdn_nt_init_dchannel(3,1)))
		{
			fprintf(stdout, "Unable to initialize the ISDN NT on the Testdevide\n");
			exit(-6);
		}
		/* Assign the callbackroutines */
		nt_span_info->call_answered	= isdn_call_answered;
		nt_span_info->call_hangup	= isdn_call_hangup;
		nt_span_info->ringing		= isdn_ringing;

	/* start Analog */
		Analog_chan *fxs1 = init_fxs(7);
		if(!fxs1) {
			fprintf(stdout, "Unable to open FXS1\n");
			exit(-7);
		}
		Analog_chan *fxs2 = init_fxs(8);
		if(!fxs2) {
			fprintf(stdout, "Unable to open FXS2\n");
			exit(-8);
		}
		Analog_chan *fxo  = init_fxo(9);
		if(!fxo) {
			fprintf(stdout, "Unable to open FXO\n");
			exit(-9);
		}

	
		/* Set Callbackroutines */
		fxs1->call_answered	= call_answered;
		fxs1->call_hangup	= call_hangup;
		fxs1->number_recieved	= number_recieved;
		fxs1->ringing		= ringing;
		
		fxs2->call_answered	= call_answered;
		fxs2->call_hangup	= call_hangup;
		fxs2->number_recieved	= number_recieved;
		fxs2->ringing		= ringing;
		
		fxo->call_answered	= call_answered;
		fxo->call_hangup	= call_hangup;
		fxo->number_recieved	= number_recieved;
		fxo->ringing		= ringing;

	/* Set which channel has to accept the call */
	switch(dev_to) {
		case 1: /* "FXO"*/
			call_acceptor = fxo;
			break;
		case 2: /* "FXS1" */
			call_acceptor = fxs1;
			break;
		case 3: /* "FXS2" */
			call_acceptor = fxs2;
			break;
		case 4: /* "ISDNNT" */
			call_acceptor = nt_span_info;
			break;
		case 5: /* "ISDNTE" */
			call_acceptor = te_span_info;
			break;
		default:
			call_acceptor = NULL;
			break;
	}

	/* Let it ring */
	switch(dev_from) {
		case 1: /* "FXO"*/
			{
				char num[256];
				sprintf(num, "T%s", number);
				fprintf(stdout, "Dailing %s using FXO\n", num);
				call_originator = fxo;
				fxo_call_number(fxo, num);
			}
			break;
		case 2: /* "FXS1" */
			{
				fprintf(stdout, "Ringing on FXS1\n");
				call_originator = fxs1;
				fxs_start_call(fxs1);
			}
			break;
		case 3: /* "FXS2" */
			{
				fprintf(stdout, "Ringing on FXS2\n");
				call_originator = fxs2;
				fxs_start_call(fxs2);
			}
			break;
		case 4: /* "ISDNNT" */
			{
				fprintf(stdout, "Dailing %s using ISDN NT\n",number);
				call_originator = nt_span_info;
				isdn_nt_call(	nt_span_info, 
					PRI_LAYER_1_ALAW, 
					
					number, //The one we are calling 
					PRI_UNKNOWN,
					0,
					
					"12345", //Who we are
					"external Network",
					PRES_ALLOWED_USER_NUMBER_NOT_SCREENED,
					PRI_NATIONAL_ISDN );
			}
			break;
		case 5: /* "ISDNTE" */
			{
				call_originator = te_span_info;
				fprintf(stdout, "Dailing %s using ISDN TE\n", number);
				isdn_te_call(   te_span_info, 
					PRI_LAYER_1_ALAW, 
					
					number, //The one we are calling 
					PRI_UNKNOWN,
					0,
					
					"*300", //We Akt as device
					"Test test",
					PRES_ALLOWED_USER_NUMBER_NOT_SCREENED,
					PRI_NATIONAL_ISDN );
			}
			break;
		default:
			/* we should never reach this one */
			fprintf(stdout, "Error, unexpected argument in <from>\n");
			return -10;
			break;
	}




	/* Wait for a while, till the timeout expires or the line gets answered */
	i=0;
	while ((i<timeout) && (!line_answered)) {
		sleep(1);
		i++;
	}

	/* turn all channels of */
	fxo_send_onhook(fxo);

	isdn_lock_span(te_span_info);
		if(te_span_info->call) pri_hangup(te_span_info->pri, te_span_info->call, PRI_CAUSE_CALL_REJECTED, -1);
	isdn_unlock_span(te_span_info);
	isdn_lock_span(nt_span_info);
		if(nt_span_info->call) pri_hangup(nt_span_info->pri, nt_span_info->call, PRI_CAUSE_CALL_REJECTED, -1);
	isdn_unlock_span(nt_span_info);
	
	
	/* Dump where it did ring and how long */
	printf("FXO rings: %i\n", fxo->numberofrings);
	printf("ISDN TE rings: %i\n", te_span_info->numberofrings);
	printf("ISDN NT rings: %i\n", nt_span_info->numberofrings);

	/* Check if we did succeed */
	if((line_answered == fxo) && (dev_to == 1))
		printf("FXO did answer the line\n");
	else if ((line_answered == fxs1) && (dev_to == 2))
		printf("FXS1 did answer the line\n");
	else if ((line_answered == fxs2) && (dev_to == 3))
		printf("FXS2 did answer the line\n");
	else if ((line_answered == te_span_info) && (dev_to == 5))
		printf("ISDN TE did answer the line\n");
	else if ((line_answered == nt_span_info) && (dev_to == 4))
		printf("ISDN NT did answer the line\n");
	else if ((line_answered == NULL))
	{
		printf("nobody did answer the line\n");
		exit(-1);
	}
	else
	{
		fprintf(stderr, "UNEXPECTED! Someone undefined did answer the line\n");
		exit(-2);
	}

	/* Exit successfully */
	exit(0);
}

