/*
 * BiST: A telefonie hardware testing application
 *
 * Written by  Sven Kolberg <s.kolberg@maxina.de>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */


/**
 * \brief ISDN TE handling functions
 * author: Sven Kolberg s.kolberg@maxina.de
 */
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#if defined(__linux__)
#include <linux/zaptel.h>
#elif defined(__FreeBSD__)
#include <zaptel.h>
#endif
//#include <zap.h>
#include <pthread.h>
#include <sys/select.h>

#include <libpri.h>
#include <pri_internal.h>
#include <tonezone.h>

/* Local Include File */
#include "ISDN_CPE.h"
#include "Analog_Func.h"

#ifdef DEBUG
	#define PDEBUG(frmt, ...) \
		fprintf(stdout, "file: %s function: %s line: %i \n",__FILE__, __PRETTY_FUNCTION__, __LINE__); \
		fprintf(stdout, frmt, ##__VA_ARGS__);\
		fprintf(stdout, "\n");
#else 
	#define PDEBUG(frmt, ...)
#endif

static void *handle_incomming_call(void *data)
{
	ISDN_SPAN_INFO *span = (ISDN_SPAN_INFO *)data;
	int nrings = 0, loop=1;

	PDEBUG("#############################################################\nCall Recieved");
	
	isdn_lock_span(span);
		PDEBUG("Acknoledge call");
		pri_acknowledge(span->pri, span->call,1,1);

	isdn_unlock_span(span);

	if(span->ringing)
	{
		while(loop)
		{
			/* Check if we accept the call */
			switch(span->ringing(span, nrings))
			{
				case 0: /* We are doing nothing here, just wait */
					break;
				case 1: /* We accept the call */
					isdn_lock_span(span);
						PDEBUG("Answer call");
						/* Send pri_answer */
						pri_answer(span->pri,span->call, 0, 1);
						/* go out of loop */
						loop=0;
					isdn_unlock_span(span);
					break;
				case -1: /* We decline the call */
					isdn_lock_span(span);
						PDEBUG("Hangup call");
						pri_hangup(span->pri,span->call, PRI_CAUSE_CALL_REJECTED, -1);
					isdn_unlock_span(span);
					break;
				default: /* should never happen */
					PDEBUG("Error: return value of span->ringing() not understandable");
			}

			nrings++;
			span->numberofrings= nrings;
			sleep(6);
		}

		sleep(10);
		isdn_lock_span(span);
			PDEBUG("Hangup call");
			pri_hangup(span->pri,span->call, PRI_CAUSE_CALL_REJECTED, -1);
		isdn_unlock_span(span);
	}
	else
	{
		/*There is no callbackroutine defined, just hangup the line, since the switching logic will never be able to handle it*/
		isdn_lock_span(span);
			PDEBUG("Hangup call");
			pri_hangup(span->pri,span->call, PRI_CAUSE_CALL_REJECTED, -1);
		isdn_unlock_span(span);
	
	}
	return NULL;
}

/**
 * \brief This Function will be called if a RING on a D-Channel is detected
 * This will happen as soon as a phone goes off hook
 * Take apropiate measures to prepare the channel, to initiate a call
 * \params pri 	The information about the D-channel
 * \params e	The event indicating the RING
 */
static void isdn_te_pri_event_ring(ISDN_SPAN_INFO *span, pri_event *e)
{
	int chanpos;

	/* the pri of isdn_span_info, just for better handling */
	struct pri *pri;
	pri = span->pri;
	
	PDEBUG("RING on TE detected!");
        
	/*Get Params of the D-Channel*/
	ZT_PARAMS param;
	if (ioctl(span->fds, ZT_GET_PARAMS, &param))
		return ;

	
	/* is a B channel allocated? */
	if (e->ring.channel == -1)	
	{
		/*We have the free choice, which B-Channel to use*/
		chanpos = isdn_get_available_bchannel(span);
	}
	else
	{
		/*The other side has already selected a channel, so use it*/
		chanpos = e->ring.channel-1; /* -1 because channel B1 has index 0*/
	}	
	PDEBUG("########################## chanpos: %i", chanpos);
	/* if no channel specified find one empty */
	if (chanpos < 0) {
		/* Nodetype BRI_CPE_PTMP */
  		PDEBUG("Ignoring callwaiting SETUP on channel %d\n", e->ring.channel);
		/*hangup*/ 
      		pri_hangup(pri, e->ring.call, PRI_CAUSE_USER_BUSY, -1);
      		return;
      	} else {
		/* OK, there is one channel */

		/* if the callednum is not set, we have to play a dailtone */
		if(strlen(e->ring.callednum) == 0) {
#warning This section is untested since The Horstbox only sends complete strings (at least until now), this case might be handled			
#if untested_enabled
			/*Play Dailtone*/
			int law; //The LAW (A or uLaw)
			
			/* This function is definied in Zaptel, so just use it */
			tone_zone_play_tone(span->bchannel[chanpos].fds /*fd_b0*/, ZT_TONE_DIALTONE);

			/* Set The Law of the channel */
			if (ioctl (span->bchannel[chanpos].fds, ZT_AUDIOMODE, &law) == -1) 
				PDEBUG("Unable to set audio mode on channel %d", chanpos);

			if (e->ring.layer1 == PRI_LAYER_1_ALAW)
				law = ZT_LAW_ALAW;
			else
				law = ZT_LAW_MULAW;
			
			ioctl(span->bchannel[chanpos].fds, ZT_SETLAW, &law);

			/* Set Gain. in General, just skip it, because its a digital channel we do not want to mess around with*/
			if (set_actual_gain(span->bchannel[chanpos].fds, 0, 0, 0, law) < 0)
				PDEBUG("Unable to set gains on channel %d", chanpos);

			
			/* Send pri_proceeding */
			//pri_proceeding(pri, e->answer.call,1,0);
			
			/* Send pri_acknowledge */
			pri_acknowledge(pri, e->answer.call,1,1);
			/* Send pri_answer */
			pri_answer(pri, e->answer.call, 0, 1);
#endif

		} 
		else 
		{
			/* This is the Normal Condition if the Horstbox akts as NT */
			int law; //The LAW (A or uLaw)
			
			/* At first the decission, do we take the call? */
			if(1) {
				/* We take the call */
				/* Play a dailtone */
				tone_zone_play_tone(span->bchannel[chanpos].fds, ZT_TONE_DIALTONE);
				
				/* Set The Law of the channel */
				law = 1;
				if (ioctl (span->bchannel[chanpos].fds, ZT_AUDIOMODE, &law) == -1) 
					PDEBUG("Unable to set audio mode on channel %d", chanpos);
	
				if (e->ring.layer1 == PRI_LAYER_1_ALAW)
					law = ZT_LAW_ALAW;
				else
					law = ZT_LAW_MULAW;
				
				ioctl(span->bchannel[chanpos].fds, ZT_SETLAW, &law);

				/* Set Gain. in General, just skip it, because its a digital channel we do not want to mess around with*/
				if (set_actual_gain(span->bchannel[chanpos].fds, 0, 0, 0, law) < 0)
					PDEBUG("Unable to set gains on channel %d", chanpos);
				
				if (1 /*e->ring.complete*/) {
					/* announce proceding */
					pri_proceeding(pri, e->ring.call, 1, 0);

					span->call = e->ring.call;

					pthread_t thread;
					/* Start Thread with the Start function dchan and the argument pri */
					if (pthread_create(&thread, NULL, handle_incomming_call, span)) 
					{
						fprintf(stderr,"could not start thread\n");
						return;
				
					}
					
					
				} else {
					/* Number is incomplete, need more info */
					pri_need_more_info(pri, e->ring.call, 1, 1);
				}
			} else {
				/* We rejekt the call */
				pri_hangup(pri, e->ring.call, PRI_CAUSE_UNALLOCATED, -1);
			}
				
		}
	}

	//remember the call
	span->call = e->ring.call;
}

/**
 * \brief Handle the PRI_EVENT_DCHAN_UP event
 * \params span the span we are working on
 * \params e the event we are working on
 */
static void isdt_te_pri_event_dchan_up(ISDN_SPAN_INFO *span, pri_event *e)
{
	
}

/**
 * \brief Handle Dchannel events
 * \params span the span we are working on
 * \params e the event recieved for the dchannel
 */
void isdn_te_handle_dchannel_events(ISDN_SPAN_INFO *span, pri_event *e)
{
	/* Network Events */
	PDEBUG("############################ %s (%d)\n", pri_event2str(e->gen.e), e->gen.e);

	switch(e->gen.e) {
		case PRI_EVENT_RING:
			isdn_te_pri_event_ring(span, e);
			break;
		case PRI_EVENT_DCHAN_UP:
			isdt_te_pri_event_dchan_up(span, e);
			break;
		case PRI_EVENT_ANSWER:
			/* Notify the Switching logic, that the call was accepted */
			if(span->call_answered) span->call_answered(span);
			break;
		case PRI_EVENT_HANGUP_REQ:
			/* Notify the switching logic, that the call needs to be hung up */
			if(span->call_hangup) span->call_hangup(span);
			break;
		default:
			break;
	}
}

/**
 * \brief handle the dchannel, in a loop call check for events an handle them
 * \params data a pointer to a ISDN_SPAN_INFO which is proper initialized
 * \returns NULL
 * remark: usualy this function will run in an extra thread, because it 
 * will never return to the caller
 */
void *isdn_te_handle_dchannel(void *data)
{
	/* The information on the ISDN_Span */
	ISDN_SPAN_INFO *isdn_span_info;
	
	/* the pri of isdn_span_info, just for better handling */
	struct pri *pri;
	
	/* next is the Time we have the check the next event,
	 * tv is the current time
	 */
	struct timeval *next, *scheduled , tv;
	
	/* An event from the D-Channel aquired using
	 * pri_schedule_run
	 * pri_check_event
	 */
	pri_event *e;

	/* Structure for select to check if read or write to a file descriptor will be blocking */ 
	fd_set fds;
	
	pthread_attr_t attr;
	
	/* The D-channel we are working on*/
	isdn_span_info = (ISDN_SPAN_INFO *) data;
	pri = isdn_span_info->pri;

	
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

	/* Result */
	int res;

	for(;;) {
		/* Calculate the Time till the next call to pri_schedule_run or pri_check_event */
		scheduled = pri_schedule_next(pri);
		
		if( (next = scheduled) )
		{
			gettimeofday(&tv, NULL);
			
			/* Make the Difference */
			tv.tv_sec = next->tv_sec - tv.tv_sec;
			tv.tv_usec = next->tv_usec - tv.tv_usec;

			/* adjust usec and sec */
			if (tv.tv_usec < 0) {
				tv.tv_usec += 1000000;
				tv.tv_sec -= 1;
			}
			if (tv.tv_sec < 0) {
				tv.tv_sec = 0;
				tv.tv_usec = 0;
			}
		} else {
			/* Just in case pri_schedule_next returned zero, ask again in 1 second */ 	
			//poll once a sec
			tv.tv_sec = 1;
			tv.tv_usec = 0;
		}

	//	PDEBUG("Timeout: %d sec %d msec %d usec\n",tv.tv_sec, tv.tv_usec/1000, tv.tv_usec%1000);

		/* Check if a read to the d-channel will be blocking or wait till the Timeout expires
		 * e.g. this does mean:
		 * 	Wait till somebody (an ISDN-Phone for instance) has send something on the D-channel
		 * 	if nobody does, just let the Timeout expire to plan further actions
		 */
		FD_ZERO(&fds);
		FD_SET(pri_fd(pri), &fds);
		res = select(pri_fd(pri) + 1, &fds, NULL, NULL, /*next ?*/ &tv/* : NULL*/);
		
		/* This section here can just be entered by a single Thread
		 * During an incomming call it may be possible that an other thread will access the span
		 */
		isdn_lock_span(isdn_span_info);
		
		if (res < 0) {
			perror("select");
		} else if (!res) {
			/* run Scheduled events */	
			e = pri_schedule_run(pri);
		} else { 
			/* An Event has been detected (at least the D-channel became readable) */
			e = pri_check_event(pri);
		}

		/* If There was an event, handle it */
		if (e) 
		{
			isdn_te_handle_dchannel_events(isdn_span_info, e);
		}
	
		/* End of critical section */
		isdn_unlock_span(isdn_span_info);
	}
	return NULL;
}

/**
 * \brief Open a D-channel an start checking for events
 * \params channelnum, the number where Dchannel is at (usualy on Horstbox at 3(NT) and 6(TE)
 * \params span, the span we are working with
 */
ISDN_SPAN_INFO *isdn_te_init_dchannel(int channelnum, int span) 
{
	ISDN_SPAN_INFO *span_info;
	int chan;

	/* Get Memory for our span_info structure */
	span_info = isdn_create_span_info();


	/* check if opening was successfull */
	if(!span_info)
		return NULL;
	
	/* Initialize the Bchannels */
	for(chan = 0; chan<NUM_BCHANNEL;chan++) {
		if(isdn_open_bchannel(&(span_info->bchannel[chan]), channelnum-NUM_BCHANNEL+chan)) 
		{
				//Add Error Handling Code here
				PDEBUG("Unable to open Bchannel %d", chan +1+channelnum);
		}
	}

	/* open Dchannel */
	isdn_open_dchannel(span_info, channelnum, span, BRI_CPE_PTMP, PRI_SWITCH_EUROISDN_E1);

	/* Start Thread with the Start function dchan and the argument pri */
	if (pthread_create(&span_info->thread, NULL, isdn_te_handle_dchannel, span_info)) 
	{
		fprintf(stderr,"could not start thread\n");
		return NULL;

	} else 
		return span_info; /* return structure to the caller */
}

/* 
 * \brief start a call on a PRI
 * \params pri the pri channel, where we should start the call on
 * \params law the Codec we should use on the B-Channel (PRI_LAYER_1_ALAW or PRI_LAYER_1_ULAW)
 * \params called_MSN a '\0' terminated string with the number we should dial
 * \params called_plan (e.g. PRI_UNKNOWN)
 * \params numcomplete is the number called_MSN complete or are there more numbers comming? (e.g 0 for numbers are complete)
 * \params caller_MSN a '\0' terminated string with the number of the originator
 * \params caller_details a '\0' terminated string with infos which can be displayed on the ISDN phone
 * \params caller_pres what shall be displayed on the phone? (e.g. PRES_ALLOWED_USER_NUMBER_NOT_SCREENED just display the caller_details)
 * \params caller_plan (e.g. PRI_NATIONAL_ISDN)
 * \returns nothing
 */
void isdn_te_call(	ISDN_SPAN_INFO *span, 
				int law,
				
				char *called_MSN,
				int called_plan,
				int numcomplete,
				
				char *caller_MSN,
				char *caller_details,
				int caller_pres,
				int caller_plan)
{
	/* Network */
	struct pri_sr *sr;
	struct pri *pri = span->pri;
	
	
	/* At first lock the span, libpri is not thread-safe */
	isdn_lock_span(span);
	
		/* Create new Call */
		span->call = pri_new_call(pri);
		
		/* Create memory for call configuration */
		sr = pri_sr_new();

		/* Set Channel configuration use channel 1 (first Bchannel)
		 * channel 0 is for aclling without B-channel?!?
		 * channel 2 is for second B-channel
		 * */
		pri_sr_set_channel(sr, 2, 1, 1);

		
		/* Set the Bearer capabilitys, in this case ALAW */
		pri_sr_set_bearer(sr, 0, law /*PRI_LAYER_1_ALAW*/, "");

		/* Set who we are calling (destination of call) */
		pri_sr_set_called(	sr,
					called_MSN/*"300"*/, //number MSN
					called_plan/*PRI_UNKNOWN*/, //calledplan
					numcomplete/*0*/); //Nummcomplete, -> complete
		
		/* Set up who we are */	
		pri_sr_set_caller(	sr, 
					caller_MSN /*"*21"*/, //our nummber
					caller_details /*"Testtest"*/, //The string to appear on the screen of the phone
					caller_plan /*PRI_NATIONAL_ISDN*/, //The callerplan
					caller_pres /*PRES_ALLOWED_USER_NUMBER_NOT_SCREENED*/); //Callerpres

		/* Set up the call, make it happen :) */
		pri_setup(pri, span->call, sr);

		/* Do cleanup */
		pri_sr_free(sr);
	
	/*unlock the span, so the normal things can proceed*/
	isdn_unlock_span(span);
}
