/*
 * BiST: A telefonie hardware testing application
 *
 * Written by  Sven Kolberg <s.kolberg@maxina.de>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */


/**
 * \brief General ISDN handling Functions
 * author: Sven Kolberg s.kolberg@maxina.de
 */
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#if defined(__linux__)
#include <linux/zaptel.h>
#elif defined(__FreeBSD__)
#include <zaptel.h>
#endif
//#include <zap.h>
#include <pthread.h>
#include <sys/select.h>

#include <libpri.h>
#include <pri_internal.h>

/* Include local defines */
#include "ISDN_Func.h"
#include "Analog_Func.h"

#ifdef DEBUG
	#define PDEBUG(frmt, ...) \
		fprintf(stdout, "file: %s function: %s line: %i \n",__FILE__, __PRETTY_FUNCTION__, __LINE__); \
		fprintf(stdout, frmt, ##__VA_ARGS__);\
		fprintf(stdout, "\n");
#else 
	#define PDEBUG(frmt, ...)
#endif

/**
 * \brief initialize a mutex for using it
 * \params pmutex a pointer to the mutex we want to initialize
 * \returns 0 uppon success, non-zero otherwise
 */
int mutex_init(pthread_mutex_t *pmutex)
{
	pthread_mutexattr_t attr;
	pthread_mutexattr_init(&attr);
	pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
	return pthread_mutex_init(pmutex, &attr);
}

/**
 * \brief destroys a mutex
 * \params pmutex pointer to the mutex to be destroyed
 */
int mutex_destroy(pthread_mutex_t *pmutex)
{
	return (pthread_mutex_destroy(pmutex));
}

/**
 * \brief locks a SPAN for exclusive use
 * \params span the SPAN to be locked
 */
int isdn_lock_span(ISDN_SPAN_INFO *span)
{
	return(pthread_mutex_lock(&span->mutex));
}

/**
 * \brief unlocks a SPAN after locking it
 * \params span the SPAN to be unlocked
 */
int isdn_unlock_span(ISDN_SPAN_INFO *span)
{
	return( pthread_mutex_unlock(&span->mutex));
}


/**
 * \brief initialize the structure isdn_span_info
 * \returns a pointer to a structure of isdn_span_info, user isdn_free_span_info() to free the associated memory
 */
ISDN_SPAN_INFO *isdn_create_span_info(void)
{
	ISDN_SPAN_INFO *p;
	
	/* allocate memory */
	if(!(p = malloc(sizeof(ISDN_SPAN_INFO)))) 
		return NULL; /* just in case that no more memory is aviable */

	/* zero all information */
	bzero(p, sizeof(ISDN_SPAN_INFO));
	
	/* initialize properly */
	p->fds = -1;
	p->pri = NULL;
	p->thread = -1;
	mutex_init(&p->mutex);

	/* return Pointer */
	return p;
}

/**
 * \brief free the Memory allocated with isdn_create_span_info
 * \params p pointer to the structure to be freed
 */
void isdn_free_span_info(ISDN_SPAN_INFO *p)
{
	if (p) free (p);
}

/**
 * \brief Searches for an unused B-channel withhin a Span and returns the number
 * \params p The Span we investigate
 * \returns the Number of the free B-channel or -1 if no channel is aviable
 */
int isdn_get_available_bchannel(ISDN_SPAN_INFO *p)
{
	int i;
	for(i=0;i<NUM_BCHANNEL;i++)	
	{
		/* if a channel is not in use, return it's index*/
		if(!p->bchannel[i].inuse) return i;
	}
	return -1;
}

/**
 * \brief checks if a caracter is a digit
 * \params c the character to be checked
 * \returns 1 if it is a digit and 0 if not
 */
int isdigit(char c)
{
	switch (c)
	{
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			return 1;
		default:
			return 0;
	}
}

/**
 * \brief Open a Zaptel channel
 * \params fn the Channel to open ("1" for channel 1, "2" for channel 2..)
 * \returns the file descriptor on success or -1 on failure
 */
int zt_open(char *fn)
{
	int fd;
	int isnum;
	int chan = 0;
	int bs;
	int x;
	isnum = 1;
	for (x=0;x<strlen(fn);x++) {
		if (!isdigit(fn[x])) {
			isnum = 0;
			break;
		}
	}
	if (isnum) {
		chan = atoi(fn);
		if (chan < 1) {
			PDEBUG("Invalid channel number '%s'\n", fn);
			return -1;
		}
		fn = "/dev/zap/channel";
	}

	fd = open(fn, O_RDWR);
	
	if (fd < 0) {
		PDEBUG("Unable to open '%s': %s\n", fn, strerror(errno));
		return -1;
	}
	if (chan) {
		if (ioctl(fd, ZT_SPECIFY, &chan)) {
			x = errno;
			zt_close(fd);
			errno = x;
			PDEBUG("Unable to specify channel %d: %s\n", chan, strerror(errno));
			return -1;
		}
	}
	bs = READ_SIZE;
	if (ioctl(fd, ZT_SET_BLOCKSIZE, &bs) == -1) return -1;
	return fd;
}

/**
 * \brief sets a translation lookup table for recieve and transmit
 * \params fd the file descriptor of the zaptel device
 * \params chan the channel on which it should be set
 * \params rxgain the gain of the recieve channel
 * \params txgain the gain of the transmit channel
 * \params law the law (A_LAW/U_LAW) of the channel
 * \returns 0 on success an non zero on failure
 * remark: right now a n:n translation is applied
 */
int set_actual_gain(int fd, int chan, float rxgain, float txgain, int law)
{
	struct	zt_gains g;
	int j;
	
	g.chan = chan;
	for (j=0;j<256;j++) {
			g.rxgain[j] = j;
			g.txgain[j] = j;
		}
		
	 /* set 'em */
	return(ioctl(fd,ZT_SETGAINS,&g));
}

/**
 * \brief this function will be called by the libpri
 * \params s the message string provided by libpri
 * remark: do not call this function directly
 */
void isdn_message(char *s)
{
	char *c;
	static int keeplast = 0;
	do {
		c = strchr(s, '\n');
		if (c) {
			*c = '\0';
			c++;
		}
		if (keeplast)
			printf("%s", s);
		else
			printf("-2 %s", s);
		if (c)
			printf("\n");
		s = c;
	} while(c && *c);
	if (!c)
		keeplast = 1;
	else
		keeplast = 0;
}

/**
 * \brief this function will be called by the libpri to display error messages
 * \params s the error string provieded by libpri
 * remark: do not call this function directly
 */
void isdn_err(char *s)
{
	char *c;
	static int keeplast = 0;
	do {
		c = strchr(s, '\n');
		if (c) {
			*c = '\0';
			c++;
		}
		if (keeplast)
			printf("%s", s);
		else
			printf("=2 %s", s);
		if (c)
			printf("\n");
		s = c;
	} while(c && *c);
	if (!c)
		keeplast = 1;
	else
		keeplast = 0;
}

/**
 * \brief dump all of the content of a pri_event_ring to stdout
 * \params e the event to be dumped
 */
void dump_pri_event_ring(pri_event_ring *e)
{
	fprintf(stdout, "e: %d\n", e->e);
	fprintf(stdout, "channel: %d\n",e->channel);
	fprintf(stdout, "callingpres: %d\n",e->callingpres);
	fprintf(stdout, "callingpresuser: %d\n", e->callingpresuser);
	fprintf(stdout, "callingplan: %d\n", e->callingplan);
	fprintf(stdout, "callingplanuser: %d\n", e->callingplanuser);
	fprintf(stdout, "callingnum: %s\n",e->callingnum);
	fprintf(stdout, "callingani: %s\n",e->callingani);
	fprintf(stdout, "callingname: %s\n",e->callingname);
	fprintf(stdout, "calledplan: %d\n", e->calledplan);
	fprintf(stdout, "ANI II: %d\n", e->ani2);
	fprintf(stdout, "callednum: %s\n",e->callednum);
	fprintf(stdout, "redirectingnum: %s\n", e->redirectingnum);
	fprintf(stdout, "useruserinfo: %s\n", e->useruserinfo);
	fprintf(stdout, "flexibl:e %i\n", e->flexible);
	fprintf(stdout, "cref: %i\n", e->cref);
	fprintf(stdout, "ctype: %i\n", e->ctype);
	fprintf(stdout, "layer1: %i\n", e->layer1);
	fprintf(stdout, "complete: %i\n", e->complete);
	fprintf(stdout, "call: %08x\n", (unsigned int)e->call);
	fprintf(stdout, "tei: %i\n", e->tei);
	fprintf(stdout, "callingsubaddr: %s\n\n", e->callingsubaddr);
}

/**
 * \brief Open a B-Channel
 * \params bchannel pointer where the associated information should be stored
 * \params channelnumber The channelnumber
 * \returns 0 on success otherwise non-zero
 */
int isdn_open_bchannel(ISDN_CHANNEL_INFO *bchannel, int channelnumber)
{
	char channel[128];
	int offset = 0;
	int tonezone;

	/* Open the Zapterldevice */
	PDEBUG("Open Bchannel /dev/zap/%d\n", channelnumber);

	sprintf(channel,"%d", channelnumber);
	bchannel->fds = zt_open(channel);

	if(bchannel->fds < 0) /* unable to open file */
	{
		PDEBUG("Error opening %d", channelnumber);
		return 1;
	}

	memset(&bchannel->zt_params, 0, sizeof(ZT_PARAMS));

        if(ioctl(bchannel->fds, ZT_GET_PARAMS, &bchannel->zt_params) < 0){
		PDEBUG("Unable to get parameters");
		zt_close(bchannel->fds)
		return 2;
	}
/*      if (p.sigtype != (signalling & 0x3ffff)) {
		ast_log(LOG_ERROR, "Signalling requested is %s but line is in %s signalling\n", sig2str(signalling), sig2str(p.sigtype));
		destroy_zt_pvt(&tmp);
		return tmp;
	}*/

	if(ioctl(bchannel->fds, ZT_AUDIOMODE, &offset)) {
		PDEBUG("Unable to set clear mode on clear channel %d", channelnumber);
		zt_close(bchannel->fds);
		return 3;
	}
	
	if(ioctl(bchannel->fds,ZT_SPANSTAT,&bchannel->zt_spaninfo)) {
		PDEBUG("Unable to get span status");
		zt_close(bchannel->fds);
		return 4;
	}

	if(ioctl(bchannel->fds, ZT_SET_PARAMS, &bchannel->zt_params) < 0){
		PDEBUG("Unable to set parameters");
		zt_close(bchannel->fds)
		return 5;
	}

	if(ioctl(bchannel->fds, ZT_GET_BUFINFO, &bchannel->zt_bufferinfo) < 0){
		PDEBUG("Unable to get bufferinfo");
		zt_close(bchannel->fds)
		return 6;
	}

	bchannel->zt_bufferinfo.numbufs=4;
	
	if(ioctl(bchannel->fds, ZT_SET_BUFINFO, &bchannel->zt_bufferinfo) < 0){
		PDEBUG("Unable to set bufferinfo");
		zt_close(bchannel->fds)
		return 7;
	}
#ifdef GainOnBChannel
	/* Get Gain */
	bchannel->zt_gain.chan = channel;
	if(ioctl(bchannel->fds, ZT_GETGAINS, &bchannel->zt_gain)) {
		PDEBUG("unabel to Get Gains");
		return 8;
	}
	
	/* Adjust Gain */
	//not needed linear is OK

	/*set Gain*/
	bchannel->zt_gain.chan = channel;
	if(ioctl(bchannel->fds, ZT_SETGAINS, &bchannel->zt_gain)) {
		PDEBUG("unable to set gains");
		return 9;
	}
#endif
	/* Set the Tonezone */
	tonezone = -1;
	if (ioctl(bchannel->fds, ZT_SETTONEZONE, &tonezone)) {
		PDEBUG("unable to set tonezone");
		return 10;
	}

	if(ioctl(bchannel->fds,ZT_SPANSTAT,&bchannel->zt_spaninfo)) {
		PDEBUG("Unable to get span status");
		zt_close(bchannel->fds);
		return 11;
	}

	
	/* evrything went ok*/
	return 0;
}

/**
 * \brief open a D-Channel and initialize it
 * \params dchannel a pointer to the structure where the associatet data should be stored, the memory must be allocated with isdn_create_dchannel()
 * \params channelnumber the Number of the Zaptel-Device
 * \params span the span we are working on
 * \params node either BRI_NETWORK_PTMP or BRI_CPE_PTMP
 * \params switchtype usualy PRI_SWITCH_EUROISDN_E1
 * \returns NULL on failure or a pointer to a ISDN_SPAN_INFO on success 
 */
ISDN_SPAN_INFO *isdn_open_dchannel(ISDN_SPAN_INFO *dchannel, int channelnum, int span, int node, int switchtype)
{
	
	int res;

	/* Open the Zaptel device */
	dchannel->fds = open("/dev/zap/channel", O_RDWR); //, 0600);
	
	/* configure the zaptel device to use the channelnum */
	if ((dchannel->fds < 0) || (ioctl(dchannel->fds,ZT_SPECIFY,&channelnum) == -1)) {
		fprintf(stderr, "Unable to open D-channel %d (%s)\n", channelnum, strerror(errno));
		return NULL;
	}
	
	/* do some sanity checks, check if it is realy a D-channel */
	res = ioctl(dchannel->fds, ZT_GET_PARAMS, &(dchannel->zt_params));
	if (res) {
		zt_close(dchannel->fds);
		fprintf(stderr, "Unable to get parameters for D-channel %d (%s)\n", channelnum, strerror(errno));
		isdn_free_span_info(dchannel);
		return NULL;
	}
	if (dchannel->zt_params.sigtype != ZT_SIG_HDLCFCS) {
		zt_close(dchannel->fds);
		fprintf(stderr, "D-channel %d is not in HDLC/FCS mode.  See /etc/zaptel.conf\n", channelnum);
		isdn_free_span_info(dchannel);
		return NULL;
	}
	
	/* check if it is possible to get the SPANSTAT */
	memset(&(dchannel->zt_spaninfo), 0, sizeof(dchannel->zt_spaninfo));
	res = ioctl(dchannel->fds, ZT_SPANSTAT, &(dchannel->zt_spaninfo));
	if (res) {
		zt_close(dchannel->fds);
		fprintf(stderr, "Unable to get span state for D-channel %d (%s)\n", channelnum, strerror(errno));
		isdn_free_span_info(dchannel);
		return NULL;
	}

	/* Set the Bufferinfo */
	dchannel->zt_bufferinfo.txbufpolicy = ZT_POLICY_IMMEDIATE;
	dchannel->zt_bufferinfo.rxbufpolicy = ZT_POLICY_IMMEDIATE;
	dchannel->zt_bufferinfo.numbufs = 32;
	dchannel->zt_bufferinfo.bufsize = 1024;
	
	if (ioctl(dchannel->fds, ZT_SET_BUFINFO, &(dchannel->zt_bufferinfo))) {
		fprintf(stderr, "Unable to set appropriate buffering on channel %d\n", channelnum);
		zt_close(dchannel->fds);
		return NULL;
	}
	
	/* Open the Dchannel itself */
	if (!(dchannel->pri= pri_new(dchannel->fds, node, switchtype,span))) {
		fprintf(stderr,"unable to init the D-channel\n");
		return NULL;
	}

	/* Set Overlapdial */
	pri_set_overlapdial(dchannel->pri,0);

	/* Set DebugLevel to Highest */
	pri_set_debug(dchannel->pri, PRI_DEBUG_Q931_ANOMALY /*PRI_DEBUG_ALL(PRI_DEBUG_Q931_DUMP | PRI_DEBUG_Q921_DUMP | PRI_DEBUG_Q921_RAW | PRI_DEBUG_Q921_STATE)*/);
	
	/* Set Network specific facility */
	pri_set_nsf(dchannel->pri, PRI_NSF_NONE);

	/* Set the desctiption */
	sprintf(dchannel->description, "ISDN-%s(%i)", (BRI_NETWORK_PTMP == node)?"NT":"TE",channelnum);
	
	/* return our pointer to the caller */
	return dchannel;
}
