package horstbox;
#
# Package horstbox
#
# This package provides functions to communicate with 
# the horstbox.
# 
use strict;
use warnings;

use Socket ;
use LWP::UserAgent;
use HTTP::Daemon;
use HTTP::Status;
use Crypt::SSLeay;
use threads::shared;

##
# Returns filename of DSL-Firmware
#
#
sub get_dsl_firmware
{
	
	my @firmware = glob("image/DLink*-C_DSL-G684T_singleimage_kernel_fs_*");

	if ( $#firmware < 0 ) {
		print "No DSL firmware found, skipping DSL upgrade...\n";
		return "";
	}

	if ( $#firmware > 0 ) {
		print "More then one DSL firmware found, skipping DSL upgrade...\n";
		return "";
	}

	return $firmware[0];
}

##
# Get DSL Version from file name
#
#
sub get_dsl_version
{
	my $version = shift;

	$version =~ s/image\/DLink.*-C_DSL-G684T_singleimage_kernel_fs_//; 

	return $version;
}

##
# Sets a nvram value.
#  
# Returns 0 on success.
sub nvram_set 
{
	my $host  = shift;
	my $user  = shift;
	my $pass  = shift;
	my $name  = shift;
	my $value = shift;
	
	my $url  = "https://${host}/nvram";

	my $ua  = new LWP::UserAgent( -timeout => 5);
	my $req = new HTTP::Request POST => $url;

	$req->authorization_basic($user, $pass);
	$req->content_type('application/x-www-form-urlencoded');
	$req->content("name=$name&value=$value");
	
	my $res = $ua->request($req);

	if ( !($res->is_success)) {
		# 1 -> Error
		# print "Error for $name, $value\n";
		return 1;	
	}

	if ( $res->as_string !~ /OK[ \t\n]*$/ ) {
		return 2;
	}

	return 0;
}

##
# Sets all nvram values which are needed for updating.
#
# Returns 0 on success.
sub update_location 
{
	my $host = shift;
	my $user = shift;
	my $pass = shift;
	my $ip   = shift;
	my $port = shift;
	my $res  = 0;
	$res = &nvram_set( $host, $user, $pass, 
		"UPDATE.CONFIG_URL", "http://$ip:$port/config.txt");
	if ($res) {
		return $res;
	}
	
	$res += &nvram_set( $host, $user, $pass, 
		"UPDATE.ZIMAGE_URL", "http://$ip:$port/zImage");
	
	$res += &nvram_set( $host, $user, $pass, 
		"UPDATE.MAIN_FS_URL", "http://$ip:$port/main-fs");
	
	$res += &nvram_set( $host, $user, $pass, 
		"UPDATE.UPDATE_FS_URL", "http://$ip:$port/update-fs");

	$res += &nvram_set( $host, $user, $pass, 
		"UPDATE.OPT_FS_URL", "http://$ip:$port/opt-fs");

	$res += &nvram_set( $host, $user, $pass, 
		"UPDATE.FINISH_URL", "http://$ip:$port/finish");

	my $dsl  = &get_dsl_firmware();
	my $ver  = &get_dsl_version($dsl);
	
	# only update, if file is there
	if ($dsl) {
		$res += &nvram_set( $host, $user, $pass, 
			"UPDATE.DSL_FS_URL", "http://$ip:$port/dsl-image");
		$res += &nvram_set( $host, $user, $pass, 
			"UPDATE.DSL_VERSION", $ver);
	}

	return $res;
}


##
# Triggers the update procedure at the horstbox.
#
#  Returns 0 on success.
sub trigger_update
{
	my $host = shift;
	my $user = shift;
	my $pass = shift;
	my $ip   = shift;
	my $port = shift;
	my $res  = 0;
	
	$res = &update_location( $host, $user, $pass, $ip, $port);
	if ($res) {
		return $res;
	}

	$res = &nvram_set( $host, $user, $pass, "sys_firmware.update", time());
	return $res;
}


sub trigger_update2
{
	my $host = shift;
	my $user = shift;
	my $pass = shift;
	my $port = shift;
	my $res  = 0;
	my $url  = "https://${host}/update";

	my $dsl  = &get_dsl_firmware();
	my $ver  = &get_dsl_version($dsl);

	my $ua  = new LWP::UserAgent( -timeout => 5);
	my $req = new HTTP::Request POST => $url;

	$req->authorization_basic($user, $pass);
	$req->content_type('application/x-www-form-urlencoded');

	# only update, if file is there
	if ($dsl) {
		$req->content("port=$port&dsl_version=$ver");
	} else {
		$req->content("port=$port");
	}
	
	$res = $ua->request($req);

	if ( !($res->is_success)) {
		# 1 -> Error
		# print "Error for $name, $value\n";
		return 1;	
	}

	if ( $res->as_string !~ /OK[ \t\n]*$/ ) {
		return 2;
	}

	return 0;
}


sub check_http 
{
	my $host = shift;
	my $user = shift;
	my $pass = shift;
	my $res  = 0;
	
	$res = &nvram_set( $host, $user, $pass, "test", "value");
	return $res;
}

sub same_network
{
	my @addr1 = split(/\./, $_[0]);
	my @addr2 = split(/\./, $_[1]);
	my @mask =  split(/\./, $_[2]);

	for (my $i=0; $i<4; $i++) {
		my $b1 = $addr1[$i] & $mask[$i];
		my $b2 = $addr2[$i] & $mask[$i];
		
		if ($b1 != $b2) {
			return 0;
		}
	}
	return 1;
}

sub get_ip_address_ip
{
	my $h = unpack 'N', inet_aton shift ;
	for (`/sbin/ip addr show`) {
		next unless /inet\s+(\d+\.\d+\.\d+\.\d+)\/(\d+)/ ;
               	my $a = unpack 'N', inet_aton $1 ;
               	my $m = $2 > 31 ? 0xffffffff : ~((1 << (32 - $2))-1);
		return $1 if (($a ^ $h) & $m) == 0 ;
	}
}

sub get_ip_address_ifconfig
{
	my $host = shift ;
	my @in = `LC_ALL=C /sbin/ifconfig | grep "inet addr"`;

	my ($addr, $mask) ;
	while ($in[0]) {
	
		$addr = $mask = shift @in;

		$addr  =~ s/^.*inet addr:([0-9\.]*).*\n/$1/; 
		$mask =~ s/^.*Mask:([0-9\.]*).*\n/$1/; 
		
		my $res = &same_network( $addr, $host, $mask);
		if ($res) {
			return $addr;
		}
	}

	return "";
}

sub get_ip_address_ipconfig
{
	my $host = shift ;
	my $addr = "";
	my $mask = "255.255.255.0";
	
	for (`ipconfig`) {
		next unless /[^0-9]*IP[^0-9]*(\d+\.\d+\.\d+\.\d+)/ ;

		$addr = $1;
#print "$addr $host\n";
		my $res = &same_network( $addr, $host, $mask);
		if ($res) {
			return $addr;
		}
	}
	return "";
}

sub get_ip_address_default
{
	return "192.168.0.2";
}

sub get_ip_address
{
	my $host = shift ;

	# use ip if available
	if (-x "/sbin/ip") {
		return &get_ip_address_ip($host);
	}

	# use ifconfig if available
	if (-x "/sbin/ifconfig") {
		return &get_ip_address_ifconfig($host);
	}

	# no ifconfig? maybe windows?
	return &get_ip_address_ipconfig($host);
	
	# for static mapping ...
#	return &get_ip_address_default($host);
}

# global variables
my $status : shared = 0; # the variable will be changed be the http daemon
			 # 9 - files where not found
			 # 0 - not running 
			 # 1 - just startet
			 # 2 - download in progress
			 # 3 - main-fs downloaded
			 # 4 - last file downloaded

sub get_status
{
	if ($status == 3) {
		$status = 2;
		return 3;
	}
		
	return $status;
}

sub check_files
{
	if (!-r "image/main-fs") {
		return 1;
	}
	if (!-r "image/opt-fs") {
		return 1;
	}
	if (!-r "image/update-fs") {
		return 1;
	}
	if (!-r "image/zImage") {
		return 1;
	}
	if (!-r "config.txt") {
		return 1;
	}

	return 0;
}

sub init_http_daemon
{
	my $port = shift;

	$status = 1;

	if (&check_files) {
		$status = 9;
		return;
	}
	
	print "Starting Update Daemon\n";
	&start_http_daemon($port);
	print "Update Daemon died\n";
	
	$status = 0;
}

sub start_http_daemon
{
	my $port = shift;
	my $d = HTTP::Daemon->new(LocalPort => $port, Reuse => 1) || return 1;

	my $last_time = time();

	while (my $c = $d->accept) {
		my $r = $c->get_request;

		if (!$r) {
			next;
		}
			
		if ($r->method ne 'GET') {
			$c->send_error(RC_FORBIDDEN);
			next;
		}

		print STDERR "Providing ", $r->url->path, " (", time() - $last_time ,")\n";
		$last_time = time();

		if ( $r->url->path eq "/config.txt") {
			$status=2;
			$c->send_file_response("config.txt");
		} elsif ( $r->url->path eq "/zImage") {
			$status=2;
			$c->send_file_response("image/zImage");
		} elsif ( $r->url->path eq "/main-fs") {
			$status=3;
			$c->send_file_response("image/main-fs");
		} elsif ( $r->url->path eq "/update-fs") {
			$status=2;
			$c->send_file_response("image/update-fs");
		} elsif ( $r->url->path eq "/opt-fs") {
			$status=2;
			$c->send_file_response("image/opt-fs");
		} elsif ( $r->url->path eq "/dsl-image") {
			my $dsl_file = &get_dsl_firmware();
			$status=2;
			$c->send_file_response("$dsl_file");
		} elsif ( $r->url->path eq "/finish") {
			$status=4;
			$c->send_error(RC_FORBIDDEN);
		} else {
			$c->send_error(RC_FORBIDDEN);
		}

		$c->close;
		undef($c);
	}

}


return 1;
