/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <linux/init.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <linux/io.h>

#include "ixp425-ledman.h"

#define LEDMAN_MAJOR	126
#define LEDMAN_NAME	"ledman"
#define LEDMAN_INTERVAL	(HZ/10)

static struct timer_list ledman_timer;
static volatile unsigned char *ledman_cs2;
static unsigned char ledman_settings_hi[2] = {0xff,0xff};
static unsigned char ledman_settings_lo[2] = {0xff,0xff};
static unsigned int ledman_counter = 0;

static void ledman_update_leds(void)
{
	int i = ((ledman_counter++)>>2)%2;

	*ledman_cs2 = ledman_settings_lo[i];
	*(ledman_cs2+2) = ledman_settings_hi[i];
}


static void ledman_timer_cb(unsigned long arg)
{
	/* trigger watchdog */
	*IXP4XX_GPIO_GPOUTR ^= 0x400;

	/* set leds */
	ledman_update_leds();

	/* Re-arm timer interrupt. */
	ledman_timer.expires = jiffies + LEDMAN_INTERVAL;
	add_timer(&ledman_timer);
}

static int ledman_init_leds(void)
{
	/* Configure CS2 for operation, 8bit and writable will do */
	*IXP4XX_EXP_CS2 = 0xbfff0003;

	/* Map address space */
	ledman_cs2 = (volatile unsigned char *) ioremap(IXP4XX_EXP_BUS_BASE(2), 512);

	*ledman_cs2 = ledman_settings_lo[0];
	*(ledman_cs2+2) = ledman_settings_hi[0];

	/* Configure GPIO10 as an output to kick watchdog */
	gpio_line_config(10, IXP4XX_GPIO_OUT);

	/* Configure GPIO1 as an output to reset the blackfin */
	gpio_line_config(1, IXP4XX_GPIO_OUT);
	
	/* setup timer for watchdog and led flashing */
	init_timer(&ledman_timer);
	ledman_timer.function = ledman_timer_cb;
	ledman_timer.data = 0;
	ledman_timer.expires = jiffies + LEDMAN_INTERVAL;
	add_timer(&ledman_timer);

	return 0;
}


static void ledman_stop_leds(void)
{
	del_timer(&ledman_timer);
	
	/* Unmap address space */
	iounmap((void*)ledman_cs2);
}

/**
 * device and ioctl functions 
 **/

static void ledman_configure_led_map(int led, int status, int map)
{
	map 	%= 2;

	if (led == 0) {
		/* turn all leds off or on */

		ledman_settings_lo[map] = (status)? 0 : 0xff;
		ledman_settings_hi[map] = (status)? 0 : 0xff;
		return;
	}

	led--;

	if (status) /* turn led on ( zeroing bit ) */
		ledman_settings_lo[map] &= ~(1 << led);
	else
		ledman_settings_lo[map] |= 1 << led;

	ledman_settings_hi[map] = ledman_settings_lo[map] >> 4;

}

int ledman_cmd(int cmd, unsigned long led)
{
	if (led > 8)
		return -EFAULT;

	switch(cmd) {
		case LEDMAN_CMD_ON:
			ledman_configure_led_map(led, 1, 0);
			ledman_configure_led_map(led, 1, 1);
			break;
		case LEDMAN_CMD_OFF:
			ledman_configure_led_map(led, 0, 0);
			ledman_configure_led_map(led, 0, 1);
			break;
		case LEDMAN_CMD_FLASH:
			ledman_configure_led_map(led, 0, 0);
			ledman_configure_led_map(led, 1, 1);
			break;
		case LEDMAN_CMD_KILLTIMER:
			del_timer(&ledman_timer);
			break;
		case LEDMAN_CMD_STARTTIMER:
			del_timer(&ledman_timer);
			add_timer(&ledman_timer);
			break;
		case LEDMAN_CMD_BLACKFIN_RST_ON:
			gpio_line_set(1, 0);
			break;
		case LEDMAN_CMD_BLACKFIN_RST_OFF:
			gpio_line_set(1, 1);
			break;
		default:
//			printk("unsupported (%x)\n", _IOC_NR(cmd));
			return -ENOTTY;
	}

	return 0;
}

EXPORT_SYMBOL(ledman_cmd);

static int ledman_ioctl(struct inode *inode, struct file *filp,
			unsigned int cmd, unsigned long arg)
{
	/* check for correct ioctl numbers */
	if (_IOC_TYPE(cmd) != LEDMAN_IOC_MAGIC) 
		return -ENOTTY;

	return ledman_cmd(cmd, arg);
}


struct file_operations ledman_fops = {
	ioctl:  ledman_ioctl,
};


static int ledman_init_dev(void)
{
	int ret;
	
	ret = register_chrdev(LEDMAN_MAJOR, LEDMAN_NAME, &ledman_fops);
	if (ret < 0) {
		printk("Error in registering device to kernel\n");
		return ret;
	}
	return 0;
}


static void ledman_remove_dev(void)
{
	int ret = unregister_chrdev(LEDMAN_MAJOR, LEDMAN_NAME);
	if (ret < 0)
		printk("Error in removing device from kernel\n");

}


static int ledman_init(void)
{
	int ret;

	ret = ledman_init_leds();
	if (ret)
		return -ENODEV;

	ret = ledman_init_dev();
	if (ret)
		return -EIO;

	return 0;
}


static void ledman_exit(void)
{
	ledman_stop_leds();
	ledman_remove_dev();
}


MODULE_DESCRIPTION("Led manager for Intel IXP425 board\n");
MODULE_AUTHOR("Henry Margies (h.margies@maxina.de");
MODULE_LICENSE("GPL");

module_init(ledman_init);
module_exit(ledman_exit);

