/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "util.h"
#include "ctrl.h"
#include "mime-types.h"
#include "buffer.h"
#include "settings.h"
#include "nvalloc.h"
#include "logger.h"

#define __COMHANDLER_PRIVATE
#include "comhandler.h"

/* ----------------------------------------------------------------------- *
 * Public: Connectionless communication handlers
 * ----------------------------------------------------------------------- */

COMM *comm_init (LIST *l, HTTP *h) {

    COMM *o = ZMALLOC (COMM);   

    if ((o->list = l) == 0) {
        logger (LOG_ERR, "comhandler: LIST subsystem not initialised!");
        return 0;
    }

    if ((o->http = h) == 0) {
        logger (LOG_ERR, "comhandler: HTTP subsystem not initialised!");
        return 0;
    }

    return o ;
}


void comm_close (COMM *o) {
    
    if (o) {
        list_close (o->list);
        free (o);
    }
}


COMM *comm_update (COMM *o, CTRL *c) {
    IOBUF b ;
    o->started = time (0);
    o->ctrl    = c ;

    XZERO (b);
    data_printf (&b, ".  verbosity-bits: 0x%02x -- not suppressing "
                 "NVRAM variable retrieving messages\n", PARSER_GETVAR_LOG);
    data_printf (&b, ".                  0x%02x -- not suppressing "
                 "NVRAM variable write back messages\n", PARSER_SETVAR_LOG);
    data_printf (&b, ".                  0x%02x -- not suppressing "
                 "NVRAM variable listing messages\n", PARSER_LISTVAR_LOG);

    ctrl_verbosity_msg (c, b.data);
    kill_space (&b);

    return o;
}


void comm_list (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];

    const char *file ;
    FILE *fp ;
    int fd ;

    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 3) goto usage;
    
    file = argc > 1 ? argv [1] : COMM_LIST_FILE_DEFAULT ;
    if ((fd = open (file, O_CREAT|O_EXCL|O_WRONLY, 0400)) < 0) {
        ctrl_reply (o->ctrl, "-ERR: cannot create %s for writing: %s",
                    file, strerror (errno));
        return;
    }
    if ((fp = fdopen (fd, "w")) == 0) {
        ctrl_reply (o->ctrl, "-ERR: cannot open %s for writing: %s",
                    file, strerror (errno));
        return;
    }
    list_print (o->list, (void(*)(void*,const char*,...))fprintf, fp);
    fprintf (fp, ".\n");
    fclose (fp);
    ctrl_reply (o->ctrl, "-OK:\n%s", file);
    return ;
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl,
                "%s [<file-name>] -- print http action table\n", me);
}


void comm_set (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    MIMETYPE mt;
    PAGE_ACT pa ;
    const char **av, *so, *s, *ena ;
    int ac, n ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc < 2) goto usage;

    if (argc > 2) {
        
        /* provide the uri as first argument */
        so       = argv [2] ;
        argv [2] = argv [1] ;

        if (argc > 3) {
            mt = mimetype_code (argv [2]);
            ac = 3 ;
        }
        else {
            mt = text_html ;
            ac = 2 ;
        }
        
        list_set_argv (o->list, paPresent, 0, so, mt, argc - ac, argv + ac);

        argv [2] = so ;
        ctrl_reply (o->ctrl, "+OK\n", argv [1], so);
        return ;
    }

    if (argc == 2) {
        if (list_get_argv
            (o->list, argv [1], &pa, &ena, &so, &mt, &ac, &av) != 0) {
            if ((s = mimetype_name (mt)) == 0) s = "" ;
            ctrl_reply (o->ctrl, "+OK:\n%s: %s(%s,%d<%s>) =>",
                        argv [1], so, s, pa, ena ? ena : "");
            for (n = 0; n < ac; n ++)
                ctrl_reply (o->ctrl, " %s", av [n] ? av [n] : "(null)");
            return ;
        }

        ctrl_reply (o->ctrl, "+OK\n", argv [1], so);
        return;
    }
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl,
                "%s <uri> [<shared-obj> [<mime-type> [<args> ... ]]]"
                " -- set/print http action\n", me);
}

    
void comm_del (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc != 2) goto usage;

    if (list_del_argv (o->list, argv [1]))
        ctrl_reply (o->ctrl, "+OK");
    else
        ctrl_reply (o->ctrl, "-ERR: %s", strerror (errno));
    return ;
        
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s <uri> -- delete http action\n", me);
}


void comm_restart (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc != 1) goto usage;

    if (http_reopen (o->http))
        ctrl_reply (o->ctrl, "+OK");
    else
        ctrl_reply (o->ctrl, "-ERR: restart failed");
    return ;

usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s -- restart the micro http server\n", me);
}


void comm_info (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    HTSTATS t;
    IOBUF b;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc != 1) goto usage;

    http_status (o->http, &t) ;
    
    ctrl_reply (o->ctrl, "+OK\n") ;

    ctrl_reply (o->ctrl, "uptime:");
    if (t.uptime > 60) {
        int min = t.uptime / 60 ;
        int h   =      min / 60 ;
        if (h > 24) {
            ctrl_reply (o->ctrl, "% 23dd %d", h / 24, h);
            h %= 24 ;
        }
        else
            ctrl_reply (o->ctrl, "% 26d", h);

        ctrl_reply (o->ctrl, ":%02d:%02d\n", min % 60, t.uptime % 60);
    }
    else
        ctrl_reply (o->ctrl, "% 33ds\n", t.uptime);

    ctrl_reply (o->ctrl, ("requests:      % 25d\n"
                          "bytes sent:    % 25d\n"
                          "bytes received:% 25d\n"),
                t.ht.request.accepted,
                t.ht.reply.bytes_sent,
                t.ht.request.bytes_read);
    
    ctrl_reply (o->ctrl, ("channels:       % 24d\n"
                          "  curently used:% 24d\n"
                          "  channel overruns:% 21d\n"),
                t.channel.available,
                t.channel.active,
                t.ht.request.overrun);

    ctrl_reply (o->ctrl, ("replies:   % 29d\n"
                          "  failed:  % 29d\n"
                          "  code 200:% 29d\n"
                          "  code 400:% 29d\n"
                          "  code 401:% 29d\n"
                          "  code 404:% 29d\n"
                          "  code 503:% 29d\n"),
                t.ht.reply.total,
                t.ht.reply.failed,
                t.ht.reply.code200,
                t.ht.reply.code400,
                t.ht.reply.code401,
                t.ht.reply.code404,
                t.ht.reply.code503);

    XZERO (b);
    data_printf (&b, "[%s]", t.ht.dll.cache);
    ctrl_reply (o->ctrl, ("Items in the dyn module cache:% 10d\n"
                          "  max entries in the cache:   % 10d\n"
                          "  objects by name:% 22s\n"),
                t.ht.dll.used_slots, t.dll.depth, b.data);
    kill_space (&b);

#   ifdef HAVE_OPENSSL
    ctrl_reply (o->ctrl, ("Items in the SSL session cache:% 9d\n"
                          "  max entries in the cache:    % 9d\n"
                          "  session cache hits:          % 9d\n"
                          "  session cache misses:        % 9d\n"
                          "  session cache timeouts:      % 9d\n"),
                t.ht.ssl.number,
                t.ht.ssl.depth,
                t.ht.ssl.hits,
                t.ht.ssl.misses,
                t.ht.ssl.timeouts) ;

    ctrl_reply (o->ctrl, "Last SSL session accept loop retries:% 3d\n",
                t.ht.ssl.accept_loop);

    ctrl_reply (o->ctrl, ("SSL server connections:          % 7d\n"
                          "  connections that have finished:% 7d\n"
                          "  renegotiations requested:      % 7d\n"),
                t.ht.ssl.accept_renegotiate,
                t.ht.ssl.accept,
                t.ht.ssl.accept_good);
#   endif

    return ;
        
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s -- print statistics\n", me);
}


// void comm_dllflush (void *state, int argc, const char **argv) {
//     COMM *o = state;
//     const char *me = argv [0];
//     
//     if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
//     if (argc != 1) goto usage;
// 
//     http_dllflush (o->http);
//     ctrl_reply (o->ctrl, "+OK");
//     return ;
//         
// usage:
//     ctrl_reply (o->ctrl, "-ERR: usage: ");
// help:
//     ctrl_reply (o->ctrl, "%s -- flush shared objects cache\n", me);
// }

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
