/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#else
#define USE_DLFCN
#endif

#ifdef USE_LOCAL_LTDL
# include "libltdl/ltdl.h"
#else
# ifdef USE_DLFCN
#  include <dlfcn.h>
#  define lt_dlopen(path)         dlopen ((path), RTLD_NOW|RTLD_LOCAL)
#  define lt_dlsym(handle, fname) dlsym  ((handle), (fname))
#  define lt_dlclose(handle)      dlclose (handle)
#  define lt_dlerror()            dlerror ()
#  define lt_dlexit()             /* empty */
#  define lt_dlinit()             /* empty */
#  define lt_dlhandle		     void*
# else
#  include <ltdl.h>
# endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <ctype.h>

#define __HTTP_DL_PRIVATE
#include "http-dl.h"

#include "util.h"
#include "list.h"
#include "hlst.h"
#include "nvalloc.h"
#include "logger.h"

/* ----------------------------------------------------------------------- *
 * Public: shared object handler
 * ----------------------------------------------------------------------- */

DLL_CACHE *call_handler (DLL_CACHE       *dll,
                         size_t       dll_dim,
                         const char       *so,
                         const char  *so_path,
                         const char   *ht_dir,
                         NVCL             *nv) {

    const char *meth ;
    time_t touched ;
    int n, inx, iny ;
    lt_dlhandle so_handle ;
    void (*fn_htrun) (void*,struct _rply_rec*,int,const char**) ;
    void *(*htinit) (const char*, NVCL*) ;

    /* check whether the dll is in the cache, already */
    for (n = 0, touched = 0, inx = iny = -1; n < dll_dim; n ++) {

        if (dll [n].so.lookup == 0) {
            if (iny < 0) iny = n ;
            continue ;
        }

        if (strcmp (dll [n].so.lookup, so) == 0)
            /* found that shared object in the cache, already */
            return dll + n ;

        if (touched == 0 || dll [n].so.touched < touched) {
            touched = dll [n].so.touched ;
            inx = n ;
        }
    }

    /* inx will always be set to either a free entry or a used entry */
    dll += (iny < 0 ? inx : iny) ;

    /* clean up old data in the cache */
    if (dll->so.lookup != 0) {
        if (dll->fn.htcleanup)
            (*dll->fn.htcleanup) (dll->fn.handle) ;
        free ((char*)dll->so.lookup);
        lt_dlclose (dll->so.handle);
        XZERO (*dll);
    }

    /* clear any existing error */
    lt_dlerror();
    /* open shared object */
    if ((so_handle = lt_dlopen (so_path)) == 0)
        goto dlopen_error ;

    /* load new data into the cache */
    if ((htinit            = lt_dlsym (so_handle, meth =    "htinit")) == 0)
        goto method_missing ;
    if ((fn_htrun          = lt_dlsym (so_handle, meth =     "htrun")) == 0)
        goto method_missing ;
    if ((dll->fn.htcleanup = lt_dlsym (so_handle, meth = "htcleanup")) == 0)
        goto method_missing ;

    /* initialise */
    dll->so.handle    = so_handle ;
    dll->so.lookup    = zstrdup (so) ;
    dll->so.touched   = time (0);
    dll->fn.htrun     = fn_htrun ;
    dll->fn.htcleanup = lt_dlsym (so_handle, "htcleanup") ;
    dll->fn.handle    = (*htinit) (ht_dir, nv);
        
    return dll ;

dlopen_error:
    logger (LOG_ERR, "Cannot load & bind method handler %s: %s",
            so_path, lt_dlerror ());
    return 0;
    
method_missing:
    logger (LOG_ERR, "No method \"%s\" handler available at %s: %s",
            meth, so_path, lt_dlerror ());
    lt_dlclose (so_handle);
    return 0;
}



void dllflush_handler (DLL_CACHE *dll, size_t dll_dim) {

    int n ;
    
    /* check whether the dll is in the cache, already */
    for (n = 0; n < dll_dim; n ++) {

        if (dll [n].so.lookup != 0) {

            if (dll [n].fn.htcleanup)
                (*dll [n].fn.htcleanup) (dll->fn.handle) ;
            
            free ((char*)dll [n].so.lookup);
            lt_dlclose (dll [n].so.handle);

            XZERO (dll [n]);
        }
    }
}

unsigned dllopen_handler (void) {
#   ifndef USE_DLFCN
    if (lt_dlinit () != 0) {
        logger (LOG_ERR, "Cannot initialise dynamic loader") ;
        return 0;
    }
#   endif
    return 1;
}

void dllclose_handler (void) {
    lt_dlexit () ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
