/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <ctype.h>

#include "util.h"
#include "list.h"
#include "hlst.h"

#define __HTTP_PRIVATE
#define __HTTP_DL_PRIVATE
#include "http-plugin.h"

/* ----------------------------------------------------------------------- *
 * Public: plugin interface
 * ----------------------------------------------------------------------- */

void header_printf (RPLY_REC *r, const char *format, ...) {
    va_list ap;
    va_start (ap, format);
    data_vprintf (& r->header, format, ap);
    va_end (ap);
}

void body_printf (RPLY_REC *r, const char *format, ...) {
    va_list ap;
    va_start (ap, format);
    data_vprintf (& r->body, format, ap);
    va_end (ap);
}

void body_append (RPLY_REC *r, const char *data, size_t len) {
    data_append (& r->body, data, len);
}

void body_app_file (RPLY_REC *r, int fd, size_t len) {
    file_append (& r->body, fd, len);
}

void rplyerror (RPLY_REC *r, const char *e) {
    r->error = e ;
}

const char *rplycode (RPLY_REC *r, unsigned n) {
    const char *s = http_codestr (r->c->base, n) ;

    if (n >= 400) {
        r->error = s ;
        r->c->req.http.should_close = 1 ;
        data_printf (& r->header, "Connection: close\r\n");
    }
    else
        data_append (& r->header, s, 0);

    return s;
}

const char *urlmethod (RPLY_REC *r) {
    return r->c->req.http.state == http_get_ok ? "GET" : "POST" ;
}

ip4addr_t http_client (RPLY_REC *r) {
    return r->c->client;
}

uint16_t http_clport (RPLY_REC *r) {
    return r->c->clport;
}

const char *http_request (RPLY_REC *r) {
    return r->c->req.http.htreq ;
}

unsigned http_fallback (RPLY_REC *r) {
    return r->c->req.http.fallback != 0 ;
}

HLST *http_args (RPLY_REC *r) {
    return r->c->req.args.htargs ;
}

const char *dlpool_p (RPLY_REC *r, const char *path) {
    return make_path (r->c->base->dll_dir, &r->path, path);
}

const char *htdocs_p (RPLY_REC *r, const char *path) {
    return make_path (r->c->base->doc_dir, &r->path, path);
}


static
void _set_value (const char  *val,
                 const char **Val,
                 size_t      *Len) {
    if (val) {
        if (Val) *Val =         val ;
        if (Len) *Len = strlen (val);
    }
    else {
        if (Val) *Val = "" ;
        if (Len) *Len =  0 ;
    }
}

unsigned nextavpair (RPLY_REC      *r,
                     const char **Var,
                     const char **Val,
                     size_t      *Len) {
    const char **VAL ;
    
    if (r->args == 0)
        r->args = open_hlst_search (r->c->req.args.htargs) ;

    if ((VAL = (const char **)next_hlst_search (r->args)) == 0) {
        close_hlst_search (r->args);
        r->args = 0 ;
        return 0 ;
    }

    *Var = query_key_hlst ((void **)VAL);

    _set_value (*VAL, Val, Len);

    return 1;
}


unsigned getavalue (RPLY_REC      *r,
                    const char  *var,
                    const char **Val,
                    size_t      *Len) {
    const char **VAL ;

    if ((VAL = (const char**)
         find_hlst (r->c->req.args.htargs, var, 0)) == 0)
        return 0;

    _set_value (*VAL, Val, Len);
    
    return 1;
}


MIMETYPE mimetype (RPLY_REC *r, MIMETYPE mt) {
    MIMETYPE old_mt ;

    if ((old_mt = r->mt) == mime_none)
        old_mt = r->c->req.args.mt ;

    if (mt)
        r->mt = mt ;

    return old_mt;
}


/* prefix the path argument by dir unless absoulte */
const char *make_path (const char *dir, IOBUF *b, const char *path) {

    if (path [0] == '/')
        return path ;

    if (path [0] == '.' && path [1] == '/')
        path += 2 ;

    kill_space (b);
    data_printf (b, "%s/%s", dir, (char*)path);
    unlink_space (b);
    return b->data ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
