/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany                                                     *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include "util.h"
#include "list.h"
#include "hlst.h"
#include "varstr.h"
#include "settings.h"
#include "ctrl.h"
#include "parser.h"
#include "nvalloc.h"
#include "tvcalc.h"
#include "logger.h"

#include <regex.h>
#include <pcre.h>

#undef WITH_DMALLOC

#define __HTTP_PRIVATE
#define __HTTP_DL_PRIVATE
#include "http.h"
#include "http-plugin.h"
#include "plugin-util.h"

#ifdef WITH_DMALLOC
typedef struct {
    const char *file ;
    int         line ;
} __avdump_rec ;
static unsigned long __mark;

static
int __avdump_print (void *state, void *val, char *key, unsigned keylen) {
    __avdump_rec *a = state ;
    dmalloc_verify (val);
    logger (LOG_DEBUG, "AVPRINT at %s(%d): %s=%s.\n",
            a->file, a->line, key, val ? val : "(null)");
    return 0;
}

static
void __avdump (HCHAN *c, const char *file, int line) {
    // __avdump_rec a = {file, line};
    dmalloc_verify (c->req.args.htargs);
    // for_hlst_do (c->req.args.htargs, __avdump_print, &a); 
    dmalloc_log_stats ();
}

#define DMALLOC_MARK() { __mark = dmalloc_mark (); }

#define DMALLOC_HUNT() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }

#define avdump(c) __avdump (c,__FILE__, __LINE__)

#define DMALLOC_VERIFY(c) {                             \
                 dmalloc_verify (c);                    \
                 dmalloc_log_stats ();                  }

#define DMALLOC_MESSAGE(g,fmt, ...) {                   \
                 dmalloc_message (fmt, ##__VA_ARGS__) ; \
                 if (g) DMALLOC_VERIFY(g) ;             }
#else
#define DMALLOC_MARK()            /* n/a */
#define DMALLOC_HUNT()            /* n/a */
#define avdump(c)                 /* n/a */
#define DMALLOC_VERIFY(g)         /* n/a */
#define DMALLOC_MESSAGE(fmt, ...) /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * Private: text/data manipulation
 * ----------------------------------------------------------------------- */

#define __str(x)   # x
#define to_str(x)  __str (x)

/* change all non-printable characters to '.' */
#define safestring(s) {char *_ ; for (_=(s); *(_); _++) \
    if (isprint ((int)*(_)) == 0 && isspace ((int)*(_)) == 0) *_='.'; }

static /* HLST call back function */
void htargs_destroy (void   *state,
                     void    *item,
                     char     *key,
                     unsigned klen) {
    if (item)
        free (item);
}
    

/* ----------------------------------------------------------------------- *
 * Private: create/destroy session management
 * ----------------------------------------------------------------------- */

static
void free_http_request (HCHAN *c) {

    if (c->req.buf.data) {
        kill_space (&c->req.buf);
        if (c->req.args.htargs) {
            avdump (c);
            destroy_hlst (c->req.args.htargs);
        }
        if (c->req.http.htreq)
            free ((char*)c->req.http.htreq);
        if (c->req.args.so)
            free ((char*)c->req.args.so) ;
        if (c->req.args.argv) {
            int n ;
            for (n = 0; n < c->req.args.argc; n++)
                if (c->req.args.argv [n])
                    free ((char*)c->req.args.argv [n]);
            free (c->req.args.argv);
        }
        XZERO (c->req);
    }
}


static
void close_http_channel (HCHAN *c) {

    // DMALLOC_HUNT ();

    if (c->fd >= 0) {
#       ifdef HAVE_OPENSSL
        if (c->ssl)
            ssl_finish (c->ssl);
#       endif
        shutdown (c->fd, 2) ;
        close (c->fd);
        undo_action (c->base->action, c->fd);
        c->fd     = -1 ;
        c->client =  0 ;
    }

    // DMALLOC_HUNT ();

    free_http_request (c);

    DMALLOC_HUNT ();
}
             
/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
const char *ht_unescape (const char *s, size_t len) {

    char *buf = (char*)xmalloc (len + 1);
    char   *t = buf ;
    int a, b ;

    while (len > 0)
        switch (*s) {
        case '+':
            * t ++ = ' ' ;
            s   ++ ;
            len -- ;
            continue ;
        case '%':
            if ((a = parse_hex_dgt (s [1])) >= 0) {
                if ((b = parse_hex_dgt (s [2])) >= 0) {
                    * t ++ = (a << 4) | b ;
                    s   += 3 ;
                    len -= 3 ;
                }
                else {
                    s   += 2 ;
                    len -= 2 ;
                }
            }
            else {
                s   ++ ;
                len -- ;
            }
            continue ;
        default:
            * t ++ = * s ++ ;
            len -- ;
        }

    * t = '\0' ;
    return buf ;
}


static
const char *parse_http_args (HCHAN *c, const char *args) {

    /* the arg string looks like: var=val&var=val ... */

    for (;;) {
        const char *end, *eq, *var, **R ;

        /* find the end of the assignment */
        for (end = args; *end && *end != '&' && !isspace (*end); end ++)
            ;

        /* find the '=' char (if any ) that separates var=val */
        for (eq = args; eq < end && *eq != '='; eq ++)
            ;
        
        /* assign/expand the variable name */
        var = ht_unescape (args, eq - args) ;

        /* store the value */
        R = (const char **) find_hlst (c->req.args.htargs, var, 0);
        if (R == 0)
            R = (const char **) make_hlst (c->req.args.htargs, var, 0) ;
        else if (*R)
            free ((char*)*R) ;

        if (R != 0)
            if (eq < end) /* assign/expand the value */
                *R = ht_unescape (eq + 1, end - eq - 1) ;

        /* clean up */
        free ((char*)var);

        /* done? */
        if (*end == '\0' || isspace (*end)) {
            avdump (c);
            return end ;
        }

        /* next assignment */
        args = end + 1;
    }

    /* NOTREACHED */
}

static
int parse_multipart_part (HCHAN *c, const char *arg) {
  pcre *re;
  const char *error;
  int erroroffset;
  int capturevector[30]; /* index should be multiple of 3 */
  char *variable, *value;
  int rc;
  FILE *uploadfile;

  re = pcre_compile("^Content-Disposition: form-data; name=\"([^\"]+)\"(?:; filename=\"([^\"]+)\"\\R+Content-Type: \\S+)*\\R+(.+)--$", PCRE_CASELESS|PCRE_MULTILINE|PCRE_DOTALL, &error, &erroroffset, NULL);

  if (re == NULL) {
    logger(LOG_ERR, "pcre_compile() failed at offset %d: %s", erroroffset, error);
    return 0;
  }

  rc = pcre_exec(re, NULL, arg, (int)strlen(arg), 0, 0, capturevector, 30);

  if (rc < 0){
    switch(rc)
      {
      case PCRE_ERROR_NOMATCH: 
	logger(LOG_ERR, "Cannot parse multipart POST data");
	break;

      default: 
	logger(LOG_ERR, "Parse error %d", rc);
	break;
      }

    pcre_free(re);
    return 0;
  }
   
  variable = zmalloc(strlen(arg)*sizeof(char *));
  pcre_copy_substring(arg, capturevector, rc, 1, variable, strlen(arg));
  variable = stripstr(variable);

  value = zmalloc(strlen(arg)*sizeof(char *));
  pcre_copy_substring(arg, capturevector, rc, 3, value, strlen(arg));

  if (strcmp(variable, "filename") == 0) {

    uploadfile = fopen("/var/tmp/mhttp-upload", "w");
    if (!uploadfile) {
      logger(LOG_ERR, "Cannot create temporary upload file");
      fclose(uploadfile);
      return 0;
    }

    fputs(value, uploadfile);
    fclose(uploadfile);
    return 1;

  } else {

    addToList(c->req.args.htargs, variable, value);

  } 

  free(variable);
  free(value);
  avdump (c);
  pcre_free(re);

  return 1; /* success */

}

/* parse multipart/form-data HTTP request */
static
int parse_multipart_args (HCHAN *c, char *args) {

  char *boundary, *p, *temp;
  int res;

  boundary = c->req.http.boundary;

  p = strstr(args, boundary);
  if (p != NULL) {
    args = p + strlen(boundary);
  }

  for (p = args; (p = strstr(args, boundary)); args = p + strlen(boundary)) {
    temp = strstr(args, boundary);
    if (temp != NULL) {
      res = parse_multipart_part(c, stripstr(substr(args, 0, (temp-args))));
      if (res == 0)
	logger(LOG_WARNING, "unparseable POST data");
    }
  }

  return 1;

}

void _parse_http_method_cpy_arglist (HCHAN        *c,
                                     int          ac,
                                     const char **av) {
    c->req.args.so   = zstrdup (c->req.args.so) ;
    c->req.args.argv = zmalloc (c->req.args.argc * sizeof (char*));
    while (-- ac >= 0)
        if (av [ac])
            c->req.args.argv [ac] = zstrdup (av [ac]) ;
    c->req.args.htargs = create_hlst (0, htargs_destroy, 0);
}


static
unsigned parse_http_method (HCHAN *c) {

    char *p ;
    int ct ;
    const char **av, *cmd ;

    /* check whether we got a complete line */
    if (strchr (c->req.buf.data, '\n') == 0)
        /* do not retry for ever */
        return c->req.access_count ++ < 100 ;
    
    /* parse the http method */
    if (strncmp (c->req.buf.data, "GET ", 4) == 0) {
        c->req.http.state = http_get_header ;
        c->req.http.url_args = 1;
        p = c->req.buf.data + 4 ;
    }
    else if (strncmp (c->req.buf.data, "POST ", 5) == 0) {
        c->req.http.state = http_post_header ;
        p = c->req.buf.data + 5 ;
    }
    else
        return 0 ;

    /* parse the request: [http[s]://<request>[:ppp]]/.. */
    while (*p == ' ') p ++ ;
    if (strncasecmp (p, "http", 4) == 0) {
        p += 4;
        if (* p == 's' || * p == 'S')
            p ++ ;
        if (* ++ p != ':' || * ++ p != '/' || * ++ p != '/')
            return 0;
        /* find the beginning of the resource looking like "/<requ>..." */
        while (*p != '/' && isgraph (*p))
            /* isgraph (): any printable character except space. */
            p ++ ;
    }
    if (* p != '/')
        return 0;

    /* extract the command token, terminate command token */
    cmd = p ++ ;
    while (* p != ' ' && * p != '?')
        if (* p ++ == '\0')
            return 0;
    
    ct = * p ; /* save/evaluate the termination character */
    * p = '\0' ;

    /* store the request for later use */
    c->req.http.htreq = zstrdup (cmd);
    
    /* find an application method: "/<request>..." */
    if (list_get_argv (c->base->ht_lst, cmd, 0, 0,
                       &c->req.args.so, &c->req.args.mt,
                       &c->req.args.argc, &av) != 0) {
        _parse_http_method_cpy_arglist (c, c->req.args.argc, av);
    }
    else if (list_get_argv (c->base->ht_lst, LIST_DEFAULT_URI, 0, 0,
                            &c->req.args.so, &c->req.args.mt,
                            &c->req.args.argc, &av) != 0) {
        c->req.http.fallback = 1;
        _parse_http_method_cpy_arglist (c, c->req.args.argc, av);
    }
    else {
        c->req.args.argc     = 1 ;
        c->req.args.argv     = zmalloc (sizeof (char*));
        c->req.args.argv [0] = zstrdup (cmd);
    }
    * p ++ = ct ; /* restore */
    
    /* read the get request and verify the request syntax */
    if (ct == '?')
        p = (char*)parse_http_args (c, p);
    else if (!isspace (ct))
        return 0 ;

    /* verify the http header syntax: "http/1.0" or similar */
    while (isspace (*p)) p ++ ;
  
    if (strncasecmp (p, "http/", 5) != 0)
        return 0 ;
    p += 5 ;

    if (!('0' <= p [0] && p [0] <= '9') ||
        !                (p [1] == '.') ||
        !('0' <= p [2] && p [2] <= '9')) 
        return 0 ;
    p += 3 ;

    if (p [0] == '\n') {
        p ++ ;
    }
    else if (p [0] == '\r' || p [1] == '\n') {
        p += 2 ;
    }
    else
        return 0 ;

    c->req.http.seen = p - c->req.buf.data ;

    return 1 ;
}


static
unsigned compare_nvram (NVCL        *nv,
                        const char *var,    /* might be NULL */
                        const char *val) {  /* not NULL */

    if (var != 0) {
        const char *v, *w ;
        int ok ;

#       ifndef DISCARD_EXTENSIVE_LOGGING
        int noisy ;
        if ((noisy = nvretr_verbosity (nv)) == -1)
            return 0 ;
#       endif

        nvcl_putcmd (nv, "set");
        nvcl_putavpair (nv, var, 0, 0);

        ok = (nvcl_rply (nv, HTTP_NVRPLY_TIMEOUT) > 0 &&
              nvcl_getavpair (nv, &v, &w, 0, 0) != 0)
            ? strcmp (val, w) == 0
            : 0
            ;

#       ifndef DISCARD_EXTENSIVE_LOGGING
        nvrestore_verbosity (nv, noisy);
#       endif

        return ok ;
    }

    return 0 ;
}


static
unsigned parse_http_header (HCHAN *c) {
    char *r, *end ;
    char *var = c->req.buf.data + c->req.http.seen ;
    char *val = strchr (var, ':');
    unsigned u ;

    if ((r = val ++) == 0)
        return 0;

    /* extract header variable name: <header-variable> [space] ':' */
    while (r [-1] == ' ') r -- ;
    r [0] = '\0';
  
    /* separate header value */
    while (* val == ' ') val ++ ;
    if ((end = strchr (val, '\n')) == 0)
        return 0 ;
    r = end ++ ;
    while (r [-1] <= ' ')
        r -- ;
    r [0] = '\0' ;

    /* evaluate particular headers */
    if (strcasecmp (var, "Content-Length") == 0) {
        if (parse_unsigned (&u, val, 0) == 0 || u < 0) return 0;
        c->req.http.clen = u ;
    }
    else if (strcasecmp (var, "Accept") == 0) {
        /* val: comma separated list */
        if (c->req.http.acc_ok == 0)
            for (;;) {
                if ((mimetype_code (val) == c->req.args.mt) ||
                    (mimetype_code (val) == mime_any)) {
                    c->req.http.acc_ok = 1 ;
                    break ;
                }
                if ((val = strchr (val, ',')) == 0) {
                    c->req.http.acc_ena = 1 ;
                    break ;
                }
                do { val ++ ; } while (isspace (*val));
            }
    }
    else if (strcasecmp (var, "Connection") == 0) {
        if (strcasecmp (val, "Keep-Alive") == 0) {
            if (c->req.http.keep_alive == 0)
                c->req.http.keep_alive = time (0) + HTTP_KEEP_ALIVE ;
        }
        else if (strcasecmp (val, "Close") == 0) {
            c->req.http.keep_alive = 0 ;
            c->req.http.conn_close = 1 ;
        }
    }
    else if (strcasecmp (var, "Content-Type") == 0) {
        if (strncasecmp (val, "multipart/form-data", 19) == 0) {
	  c->req.http.boundary = zstrdup(stripstr(substr(val, 30, strlen(val)-30)));
	  c->req.http.multipart = 1;
	}
        else if (strcasecmp (val, "application/x-www-form-urlencoded") == 0) {
	  c->req.http.multipart = 0;
	}
    }
    else if (strcasecmp (var, "Keep-Alive") == 0) {
        if (parse_unsigned (&u, val, 0) == 0 || u < 0) return 0;
        if (u && c->req.http.conn_close == 0)
            c->req.http.keep_alive = time (0) + u ;
    }
    else if (strcasecmp (var, "Authorization") == 0) {
        if (strncasecmp (val, "Basic ", 6) == 0) {
            const char *e ;
            char *q, *p = val + 6 ;
            while (isspace (*p)) p ++ ;
            /* decode base 64 string */
            if (nvcl_b64decode (&q, &u, &e, p)) {
                /* username:passwd */
                if ((p = strchr (q, ':')) != 0) {
                    HTTP *h = c->base ;
                    * p ++ = '\0' ;
                    if (compare_nvram (h->nvram, h->htpasswd, p) &&
                        compare_nvram (h->nvram, h->htadmin,  q))
                        c->req.http.basic_auth = 1 ;
#                   ifdef                HTTP_AUTH_LOG
                    else if (verbosity & HTTP_AUTH_LOG) {
                        // p="<hidden>" ;
                        logger (LOG_DEBUG, "Basic authentication failed, "
                                "credentials=\"%s:%s\"", q, p);
                    }
#                   endif
                }
                free (q) ;
            }
        }
    }

    return c->req.http.seen = end - c->req.buf.data ;
}


static
unsigned send_http_reply (HCHAN *c) {

    IOBUF rsp;
    const char *code, *mtype ;
    int n ;
    RPLY_REC io ;
    HTTP *h = c->base ;

    XZERO  (io) ;
    XZERO (rsp) ;

    io.c = c ;

    // DMALLOC_HUNT ();

    /* check for authentication */
    if (h->htadmin && !c->req.http.basic_auth) {
        io.error = http_codestr (h, 401) ;
        data_append (&io.header, "WWW-Authenticate: Basic realm=\"", 0);
        if (h->htrealm) {
#           ifndef DISCARD_EXTENSIVE_LOGGING
            int noisy ;
            if ((noisy = nvretr_verbosity (h->nvram)) != -1) {
#           endif
                const char *var, *val ;
                nvcl_putcmd (h->nvram, "set");
                nvcl_putavpair (h->nvram, h->htrealm, 0, 0);
                if (nvcl_rply (h->nvram, HTTP_NVRPLY_TIMEOUT) > 0 &&
                    nvcl_getavpair (h->nvram, &var, &val, 0, 0) != 0)
                    data_append (&io.header, val, 0);
#           ifndef DISCARD_EXTENSIVE_LOGGING
                nvrestore_verbosity (h->nvram, noisy);
            }
#           endif
        }
        data_append (&io.header, "\"\r\n", 0);
    }
    else {
        DLL_CACHE *dll, *hth = 0 ;

        /* get shared object */
        if (c->req.args.so && (dll = c->base->dll_cache) != 0) {
            size_t     dim = c->base->dll_slots ;
            const char *so = c->req.args.so ;
            const char *sp = dlpool_p (&io, so) ;
            if ((hth = call_handler
                 (dll, dim, so, sp, h->doc_dir, h->nvram)) != 0)
                (*hth->fn.htrun) (hth->fn.handle,
                                  &io,
                                  c->req.args.argc,
                                  c->req.args.argv);

            /* not needed, anymore */
            kill_space (&io.path);
        }
        
        /* error: no shared object */
        if (hth == 0) {
            rplycode (&io, 404);
            if (verbosity & HTTP_LOG_CLOSE)
                logger (LOG_ERR, "Error 404: suggesting to close the http "
                        "socket, fd=%d", c->fd);
            data_printf (&io.body, "No method handler available for %s\r\n",
                         c->req.http.htreq );
        }
    }

    // DMALLOC_HUNT ();

    /* assemble http response */
    if ((code = io.error) == 0)
        code = http_codestr (h, 200);
    if ((io.mt && ((mtype = mimetype_name (io.mt))          == 0)) ||
        (          (mtype = mimetype_name (c->req.args.mt)) == 0))
        mtype = "text/plain";

    data_printf (&rsp,
		 "HTTP/1.1 %s\r\n"
		 "Server: Maxina/mHttp-server $Revision: 3877 $\r\n"
		 "Content-Type: %s\r\n",
		 code, mtype);

    /* hh TODO
    disposition = (char **)find_hlst (c->req.args.htargs, "Content-Disposition", 0);
    if (disposition != 0)
    {
            logger(LOG_INFO, "serving as: %s\n", disposition);
	    data_printf(&rsp, "Content-Disposition: %s\r\n", disposition);
    }
    */
    /* TODO remove this hack which serves the box configuration as a downloadable file */
    if (find_hlst(c->req.args.htargs, "configuration", 0))
    {
	    data_printf(&rsp, "Content-Disposition: attachment; filename=HorstBox.cfg;\r\n");
    }

    /* append additional headers */
    if (io.header.data) {
        data_printf (&rsp, "%s", io.header.data);
        kill_space (&io.header);
    }

    if (io.body.data) {
        /* make sure that the client understands the mime type */
        if (!io.mt && c->req.http.acc_ena && !c->req.http.acc_ok) {
            kill_space (&io.body);
            kill_space (&rsp);
            logger (LOG_ERR, "Cannot send on the http socket fd=%d: "
                    "%s is not understood by the client", mtype);
            return 0;
        }
        data_printf (&rsp, "Content-Length: %u\r\n\r\n", io.body.top);
        data_append (&rsp, io.body.data, io.body.top); /* 8 bit clean */
        kill_space (&io.body);
    }
    else {
        data_printf (&rsp, "Content-Length: 0\r\n\r\n");
    }

    /* statistics */
    h->stats.ht.reply.total ++ ;

    /* send answer packet */
#   ifdef HAVE_OPENSSL
    if (c->ssl)
        n = ssl_send (c->ssl, rsp.data, rsp.top) ;
    else
#   endif
    {
        n = send (c->fd, rsp.data, rsp.top, 0) ;
        if (verbosity & HTTP_REPLY_LOG) {
            safestring (rsp.data);
            logger (LOG_DEBUG, "HTTP reply: \"%s\"", rsp.data);
        }
    }

    /* clean up */
    kill_space (&rsp);

    if (n < 0) {
        if (errno)
            logger (LOG_ERR,
                    "Send on the http socket fd=%d returned error: %s",
                    c->fd, strerror (errno));
        else
            logger (LOG_ERR, "Send error on the http socket, fd=%d", c->fd);

        /* statistics */
        h->stats.ht.reply.failed ++ ;
        return 0;
    }

    /* statistics */
    h->stats.ht.reply.bytes_sent += n ;

    return 1;
}

/* ----------------------------------------------------------------------- *
 * Private: http data callback handler
 * ----------------------------------------------------------------------- */

static
void http_data (void *data) {

    int n ;
    HCHAN *c = data ;

#   ifdef HAVE_OPENSSL
read_more:
#   endif

    /* extend read buffer */
    make_space (&c->req.buf, HTTP_MIN_REQBUF);

#   ifdef HAVE_OPENSSL
    if (c->ssl) {
        if ((n = ssl_recv (c->ssl,
                           c->req.buf.data + c->req.buf.top,
                           c->req.buf.size - c->req.buf.top - 1)) < 0) {
            if (errno != EAGAIN) { 
                logger (LOG_ERR,
                        "Closing ssl socket fd=%d on read error: %s",
                        c->fd, ssl_strerror (c->ssl));
                close_http_channel (c);
            }
            return ;
        }
    }
    else
#   endif
        if ((n = recv (c->fd,
                       c->req.buf.data + c->req.buf.top,
                       c->req.buf.size - c->req.buf.top - 1,
                       MSG_DONTWAIT)) < 0) {
            if (errno != EAGAIN) { 
                logger (LOG_ERR,
                        "Closing http socket fd=%d on read error: %s",
                        c->fd, strerror (errno));
                close_http_channel (c);
            }
            return ;
        }

    if (n == 0) {
        /* orderly shut down */
        if (verbosity & HTTP_LOG_CLOSE)
            logger (LOG_ERR, "No more data: closing http socket, fd=%d",
                    c->fd);
        goto http_close_n_leave ;
    }


    /* prepare for text parsing */
    c->req.buf.data [c->req.buf.top + n] = '\0' ;

    /* statistics */
    c->base->stats.ht.request.bytes_read += n ;

#   ifdef HTTP_REQUEST_LOG
    if (verbosity & HTTP_REQUEST_LOG)
        logger (LOG_DEBUG, (c->req.buf.top
                            ? "HTTP request (contd.): %d=\"...%s\""
                            : "HTTP request: %d=\"%s\""),
                n, c->req.buf.data + c->req.buf.top);
#   endif

    /* set to end of text */
    c->req.buf.top += n ;

    /* indicate that we have just visited that data chanel */
    c->visited = time (0);

    for (;;) {
        // DMALLOC_HUNT ();

        switch (c->req.http.state) {
        case none:
            if (parse_http_method (c) == 0)
                break;
            continue ;

        case http_post_header:
        case http_get_header: {
            char *p, *q = c->req.buf.data + c->req.http.seen ;
            if ((p = strchr (q, '\n')) == 0) {
                /* need more data */
#               ifdef HAVE_OPENSSL
                if (c->ssl)
                    goto read_more;
#               endif
                return ;
            }
            if (* q == '\r' && p == q + 1) {
                c->req.http.state = c->req.http.state == http_get_header
                    ? http_get_ok
                    : http_post_body ;
                c->req.http.seen += 2 ;
                c->req.http.hlen = c->req.http.seen ;
                continue ;
            }
            if (p == q) {
                c->req.http.state = c->req.http.state == http_get_header
                    ? http_get_ok
                    : http_post_body ;
                c->req.http.seen ++ ;
                continue ;
            }
            if ((n = parse_http_header (c)) == 0)
                break ;
            c->req.http.seen = n ;
            continue ;
        }

        case http_get_ok:
            if (send_http_reply (c) == 0)
                break ;
            c->req.http.state = http_sent_ok ;
            continue;

        case http_post_body:
            /* need to check that the body is complete */
            if (c->req.buf.top < c->req.http.hlen + c->req.http.clen) {
                /* need more data */
#               ifdef HAVE_OPENSSL
                if (c->ssl)
                    goto read_more;
#               endif
                return ;
            }
            if (c->req.http.seen < c->req.buf.top) {
	        if (c->req.http.multipart) {
		  if (parse_multipart_args (c, c->req.buf.data + c->req.http.seen) == 0)
		    logger (LOG_ERR, "Illegal POST data/variables");
		} else {
		  const char *p = (char*) parse_http_args (c, c->req.buf.data + c->req.http.seen);
		  if (*p && !isspace (*p))
		    logger (LOG_ERR, "Illegal POST data/variables");
		}
            }
            c->req.http.state = http_post_ok ;
            continue ;

        case http_post_ok:
            if (send_http_reply (c) == 0)
                break ;
            c->req.http.state = http_sent_ok ;
            continue;
    
        case http_sent_ok:
            /* remove that data channel unless keep-alive is on */
            if (c->req.http.keep_alive   == 0 ||
                c->req.http.conn_close   != 0 ||
                c->req.http.should_close != 0 ||
                c->req.http.keep_alive < time (0))
                break ;
            free_http_request (c);
#           ifdef HAVE_OPENSSL
            if (c->ssl && (ssl_flush (c->ssl), ssl_pending (c->ssl)))
                /* the ssl cache must be flushed */
                break ;
#           endif
            
            // DMALLOC_HUNT ();
            return ;

        default:
            logger (LOG_ERR, "State machine error (internal) - %u",
                    c->req.http.state);
            /* break; */
        }

        if (verbosity & HTTP_LOG_CLOSE)
            logger
                (LOG_ERR, "State machine closes http socket, fd=%d", c->fd);
        goto http_close_n_leave ;
    }
    /* NOTREACHED */

http_close_n_leave:

    /* remove that data channel */
    close_http_channel (c);
}

/* ----------------------------------------------------------------------- *
 * Private: event loop call back
 * ----------------------------------------------------------------------- */

static
void http_accept (void *data) {

    HTTP *h = data ;
    const char *reason ;
    struct sockaddr_in in ;
    socklen_t alen = sizeof (in);
    int n, fd, old_opt;

    // DMALLOC_MARK ();
    
    /* Set nonblocking mode so that accept() will never choke */
    if ((old_opt = fcntl (h->listen_fd, F_GETFL, 0)) >= 0)
        fcntl (h->listen_fd, F_SETFL, old_opt|O_NONBLOCK);
    if ((fd = accept (h->listen_fd, (struct sockaddr*)&in, &alen)) < 0) {
        logger (LOG_ERR, "Error accepting http: %s", strerror (errno));
        return ;
    }
    if (old_opt >= 0)
        fcntl (h->listen_fd, F_SETFL, old_opt);
    
    /* find a free socket descriptor */
    for (n = 0; n < h->channels; n ++)
        if (h->channel [n].fd < 0) {
            h->channel [n].client = ntohl (in.sin_addr.s_addr) ;
            h->channel [n].clport = ntohs (in.sin_port) ;
#           ifdef HAVE_OPENSSL
            if (h->cty && /* enabled ?? */ (h->channel [n].ssl = ssl_accept
                                            (h->cty, fd,
                                             h->channel [n].client,
                                             h->channel [n].clport)) == 0) {
                reason = "Error accepting ssl" ;
                goto stop ;
            }
#           endif
#           ifdef FD_CLOEXEC
            fcntl (fd, F_SETFD, FD_CLOEXEC);  /* close socket in exec */
#           endif
            h->channel [n].fd      =      fd ;
            h->channel [n].visited = time (0);
            do_action  (h->action, fd, http_data, h->channel + n);
            h->stats.ht.request.accepted ++ ;
            // DMALLOC_HUNT();
            return ;
        }

    reason = "No more slots for http" ;
    h->stats.ht.request.overrun ++ ;

#   ifdef HAVE_OPENSSL
stop:
#   endif
    /* no socket descriptor left - close up */
    logger (LOG_ERR, "%s - rejecting %s", reason, ipaddr_pp (&in.sin_addr));
    shutdown (fd, 2);
    close    (fd);
    // DMALLOC_HUNT();
}

/* ----------------------------------------------------------------------- *
 * Private: constructor
 * ----------------------------------------------------------------------- */

static
unsigned http_listen_setup (HTTP *h) {
    int n;

    /* open http socket */
    if ((h->listen_fd = socket (PF_INET, SOCK_STREAM, 0)) == 0) {
        logger (LOG_ERR, "Error creating http socket: %s", strerror (errno));
        return 0 ;
    } 
    if (setsockopt
        (h->listen_fd, SOL_SOCKET, SO_REUSEADDR, (n=1, &n), sizeof n) < 0) {
        logger (LOG_ERR,
                "Error customizing http socket: %s", strerror (errno));
        return 0 ;
    }
    if (bind
        (h->listen_fd,(struct sockaddr*)&h->listen,sizeof (h->listen)) < 0) {
        logger (LOG_ERR, "Error binding http socket: %s", strerror (errno));
        return 0 ;
    }
    if (listen (h->listen_fd,  h->channels) < 0) {
        logger (LOG_ERR,
                "Error listening on http socket: %s", strerror (errno));
        return 0 ;
    }
#   ifdef FD_CLOEXEC
    fcntl (h->listen_fd, F_SETFD, FD_CLOEXEC); /* socket close on exec */
#   endif

    /* set up the http server */
    do_action (h->action, h->listen_fd, http_accept, h);

    return 1;
}

static
void http_channel_setup (HTTP *h) {
    int n ;
    
    /* (re-)initialise channels */
    for (n = 0; n < h->channels; n ++) {
        h->channel [n].fd   = -1 ;
        h->channel [n].base = h ;
    }
}

/* ----------------------------------------------------------------------- *
 * Public: constructor/destructor
 * ----------------------------------------------------------------------- */

char *http_codestr (HTTP *h, int errnum) {
    
#   define _UNSUPPORTED_FMT "%u Unsupported code"
#   define _CASE_STRING(num,text) \
    case num: { h->stats.ht.reply.code ## num ++; \
                return to_str (num) " " text;     }

    static char buf [sizeof (_UNSUPPORTED_FMT) + 5]; /* 0xff => max 5 dgts */
    switch (errnum) {
        _CASE_STRING (200, "OK");

        _CASE_STRING (400, "Bad Request");
        _CASE_STRING (401, "Unauthorized");
        _CASE_STRING (404, "Not Found");
        
        _CASE_STRING (503, "Service Unavailable");
    }

    sprintf (buf, "%03u Unsupported code", errnum & 0xff);
    return buf;
    
#   undef _CASE_STRING
}


void http_status (HTTP *h, HTSTATS *t) {

    if (t) {
        int n, a ;
        char *p ;
        
        *t = h->stats ;

        t->uptime = time (0) - h->started ;

        for (n = 0, t->channel.active = 0; n < h->channels; n ++)
            if (h->channel [n].fd >= 0)
                t->channel.active ++ ;

        if ((p = (char*)h->stats.ht.dll.cache) == 0)
            p = zmalloc (1) ;
            
        for (n = a = t->ht.dll.used_slots = 0; n < h->dll_slots; n ++) {
            const char *s = h->dll_cache [n].so.lookup;

            if (s) {
                int b = strlen (s);
                if (memcmp (s + b - 3, ".so", 3) == 0) /* strip ",so" */
                    b -= 3 ;
                p = (char*) zrealloc (p, a + b + 3);
                if (a) {
                    memcpy (p + a, ", ", 2);
                    a += 2;
                }
                memcpy (p + a, s, b);
                p [a += b] = '\0' ;
                t->ht.dll.used_slots ++ ;
            }
        }

        t->ht.dll.cache = h->stats.ht.dll.cache = p ;

#       ifdef HAVE_OPENSSL
        t->ht.ssl = * ssl_info (h->cty);
#       endif
    }
}


void http_statflush (HTTP *h) {
    if (h->stats.ht.dll.cache)
        free ((char*)h->stats.ht.dll.cache);
    XZERO (h->stats.ht) ;
}


void http_idlecheck (HTTP *h) {
    int n ;
    time_t ttd = time (0) - h->channel_idle ;

    for (n = 0; n < h->channels; n ++)
        if (h->channel [n].fd >= 0 && h->channel [n].visited <= ttd) {
            if (verbosity & HTTP_PURGE_LOG)
                logger (LOG_INFO, "Connection to %s on slot %d(fd=%d) timed "
                        "out (last vistited at %s): closing",
                        ip4addr_pp (h->channel [n].client), n+1,
                        h->channel [n].fd,
                        time_print (h->channel [n].visited));
            DMALLOC_MESSAGE (0, "Closing socket now ...");
            close_http_channel (h->channel + n);
        }
}


HTTP *http_init (DO           *action,
                 LIST           *list, 
                 unsigned  chan_slots,
                 unsigned cache_slots,
                 ip4addr_t     ipaddr,
                 uint16_t        port,
                 const char   *dlpool,
                 const char   *htdocs,
                 const char  *htadmin,
                 const char *htpasswd,
                 const char  *htrealm,
                 NVCL          *nvram
#                ifdef HAVE_OPENSSL
                 ,CTY            *cty
#                endif
    ) {

    HTTP *h ;

    DMALLOC_MARK () ;

    if (dllopen_handler () == 0)
        return 0;

    if (chan_slots == 0 || chan_slots == -1)
        chan_slots = HTTP_DEFAULT_CHANNELS ;

    h = (HTTP*) zmalloc (sizeof (HTTP) * (chan_slots - 1)) ;

    /* set up the number of channels */
    h->channels = chan_slots ;

    if ((h->nvram = nvram) == 0) {
        logger (LOG_ERR, "Initializaion error: NVRAM connector missing");
        goto error_ok ;
    }
    if ((h->action = action) == 0) {
        logger (LOG_ERR,
                "Initializaion error: call back handler object missing");
        goto error_ok ;
    }
    if ((h->ht_lst = list) == 0) {
        logger (LOG_ERR, "Initializaion error: "
                "assignment list handler object missing");
        goto error_ok ;
    }

    /* set up the ip-address and the port */
    if (port == 0 || port == (uint16_t)-1)
        port = HTTP_DEFAULT_SERVER_PORT ;
    if (ipaddr == -1)
        ipaddr = HTTP_DEFAULT_IP_ADDRESS ;
    XZERO (h->listen);
    h->listen.sin_family = AF_INET ;
    h->listen.sin_port   = htons (port) ;
    if (ipaddr)
        h->listen.sin_addr.s_addr = htonl (ipaddr) ;

    /* initialise the listener */
    if (http_listen_setup (h) == 0)
        goto error ;

    http_channel_setup (h);
        
    /* initialise the dll chache */
    if ((h->dll_slots = cache_slots) == 0 || cache_slots == -1)
        h->dll_slots = HTTP_DEFAULT_CACHE_DIM ;
    h->dll_cache = zmalloc (h->dll_slots * sizeof (DLL_CACHE));

    /* set up strings */
    h->dll_dir = dlpool ? zstrdup (dlpool) : zstrdup (".");
    h->doc_dir = htdocs ? zstrdup (htdocs) : zstrdup (".");
    if (htadmin)
        h->htadmin  = zstrdup (htadmin) ;
    if (htpasswd)
        h->htpasswd = zstrdup (htpasswd) ;
    if (htrealm)
        h->htrealm  = zstrdup (htrealm) ;

    h->channel_idle = HTTP_DEFAULT_CHANNEL_IDLE ;   

#   ifdef HAVE_OPENSSL
    h->cty = cty ;

    /* statistics */
    h->stats.sslok             = (cty != 0);
#   endif
        
    /* statistics */
    h->started                 = time (0);
    h->stats.channel.available = h->channels ;
    h->stats.dll.depth         = h->dll_slots ;

    if (verbosity & HTTP_INTRO_LOG) {
        IOBUF  url ;
        XZERO (url);
#       ifdef HAVE_OPENSSL
        data_append (&url, cty ? "https" : "http", 0);
        data_printf (&url, "://%s", ip4addr_pp (ipaddr));
        if ((cty != 0 && port != 443) || (cty == 0 && port != 80))
            data_printf (&url, ":%u", port);
#       else
        data_printf (&url, "http://%s", ip4addr_pp (ipaddr));
        if (port != 80)
            data_printf (&url, ":%u", port);
#       endif
        logger (LOG_INFO, "Listener on %s has been established", url.data);
        kill_space (&url);
    }

    return h ;

error:
    if (h->listen_fd >= 0)
        close (h->listen_fd);

error_ok:
    free (h);
    return 0;
}


unsigned http_reopen (HTTP *h) {
    int n;
    
    /* flush active data channels */
    for (n = 0; n < h->channels; n ++) 
        close_http_channel (h->channel + n);
    dllflush_handler (h->dll_cache, h->dll_slots);

#   ifdef HAVE_OPENSSL
    /* reopen ssl */
    if (h->cty != 0 && ssl_reopen (h->cty) == 0) {
        logger (LOG_WARNING, "Cannot reinitialise for the SSL sessions");
        return 0;
    }
#   endif

    return 1;
}


void http_close (HTTP *h) {
    if (h) {
        int n;
        for (n = 0; n < h->channels; n ++) 
            close_http_channel (h->channel + n);
        if (h->listen_fd >= 0)
            close (h->listen_fd);
        if (h->dll_dir)
            free ((char*)h->dll_dir);
        if (h->htadmin)
            free ((char*)h->htadmin) ;
        if (h->htpasswd)
            free ((char*)h->htpasswd) ;
        if (h->htrealm)
            free ((char*)h->htrealm) ;
        if (h->stats.ht.dll.cache)
            free ((char*)h->stats.ht.dll.cache);

        free (h);

        dllclose_handler ();
    }
}


// void http_dllflush (HTTP *h) {
//     dllflush_handler (h->dll_cache, h->dll_slots);
// }

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
