/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: n/n (D-Link)			                          *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>

#include "hlst.h"
#include "util.h"
#include "http-plugin.h"
#include "plugin-util.h"
#include "parser.h"
#include "nvalloc.h"
#include "logger.h"


/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

#define INVALID_START_IP      	     	CUSTOM_START_ERROR + 10
#define INVALID_END_IP      	     	CUSTOM_START_ERROR + 11
#define INVALID_ADDRESS_RANGE  	     	CUSTOM_START_ERROR + 12
#define INVALID_NETWORK  	     	CUSTOM_START_ERROR + 13

#define INVALID_PRIMARY_DNS    	     	CUSTOM_START_ERROR + 20
#define INVALID_SECONDARY_DNS       	CUSTOM_START_ERROR + 21

#define INVALID_LEASE_TIME       	CUSTOM_START_ERROR + 30

#define GROUP_LABEL			"lan_dhcp"

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

int isDhcpEnabled(HLST *avp)
{
	char **val = (char**)find_hlst(avp, "lan_dhcp.isDhcp", 0);

	if(val && !strcmp(*val, "true"))
	{
		return 1;
	}

	return 0;
}

int isDnsModeAuto(HLST *avp)
{
	char **val = (char**)find_hlst(avp, "lan_dhcp.dnsMode", 0);

	if(val && !strcmp(*val, "auto"))
	{
		return 1;
	}

	return 0;
}

int checkLeaseTime(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	int error = NO_ERROR;

	if(isDhcpEnabled(avp))
	{
		int leaseTime;

		leaseTime = getVarAsLong(avp, varName);
		if((leaseTime < 1) || (leaseTime > 999999)) /* Alte D-LINK Box lies max. 6 Zeichen zu */
		{
			error = INVALID_LEASE_TIME;
		}
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}


int checkStartIp(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	int error = NO_ERROR;

	if(isDhcpEnabled(avp))
	{
		char **val = (char**)find_hlst(avp,varName,0); 
		if(!val || !validIpAddr(val))
		{
			error = INVALID_START_IP;
		}
	}

	if(error != NO_ERROR)
		handle_varError(avp,error );

	return error;
}

int checkEndIp(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	int error = NO_ERROR;

	if(isDhcpEnabled(avp))
	{
		char **val = (char**)find_hlst(avp,varName,0);
		if(!val || !validIpAddr(val))
		{
			error = INVALID_END_IP;
		}
	}

	if(error != NO_ERROR)
		handle_varError(avp,error );


        return error;
}



int checkPrimDns(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	int error = NO_ERROR;
	

	if(isDhcpEnabled(avp) && 
	   !isDnsModeAuto(avp))
	{
		char **val = (char**)find_hlst(avp,varName,0);
		if(!val || !validIpAddr(val))
		{
			error = INVALID_PRIMARY_DNS;
		}
	}

	if(error != NO_ERROR)
		handle_varError(avp,error );

        return error;
}

int checkSecDns(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	int error = NO_ERROR;
        
	if(isDhcpEnabled(avp) && 
	   !isDnsModeAuto(avp))
	{
		char **val = (char**)find_hlst(avp,varName,0);

		//The value for the alternate DNS could be empty for the case that a user only knows one dns server ip
		if(val && !strlen(removeBlanks(*val)))
		{
			//delete variable from avp list
			//delete_hlst(avp,varName,0);
			//return
			return NO_ERROR;
		}


		if(!validIpAddr(val))
		{
			error = INVALID_SECONDARY_DNS;
		}
	}

	if(error != NO_ERROR)
		handle_varError(avp,error );

        return error;
}

int checkIpRange(HLST *avp, HLST *nvram_avp, const char *varName)
{
	/* test whether filter is for any Source ip or not*/

	unsigned long ret = 0;
	struct in_addr start, end, ip, mask;
	int error = NO_ERROR;
       
	char **destFrom = (char**)find_hlst(avp,"lan_dhcp.startIp",0);
	char **destTo   = (char**)find_hlst(avp,"lan_dhcp.endIp",0);
	char **ip_addr  = (char**)find_hlst(nvram_avp,"lan_ip.ipAddress",0);
	char **ip_mask  = (char**)find_hlst(nvram_avp,"lan_ip.subnetMask",0);

	ensureCheckState(avp,nvram_avp,varName);

	if (!isDhcpEnabled(avp))
		return NO_ERROR;

	if (!destFrom || !validIpAddr(destFrom))
		return NO_ERROR; /* will be handled by checkStartIp */

	if (!destTo || !validIpAddr(destTo))
		return NO_ERROR; /* will be handled by checkEndIp */

	inet_aton(*destFrom, &start);
	inet_aton(*destTo, &end);

	if (ntohl(start.s_addr) > ntohl(end.s_addr)) {
		error = INVALID_ADDRESS_RANGE;
		handle_varError(avp, error);
	}

	if (!ip_addr || !*ip_addr || !ip_mask || !*ip_mask)
		return error; /* strange?! */

       	inet_aton(*ip_addr, &ip);
       	inet_aton(*ip_mask, &mask);

	if ( (ip.s_addr & mask.s_addr) != (start.s_addr & mask.s_addr) ||
       	     (ip.s_addr & mask.s_addr) != (end.s_addr & mask.s_addr)) {
       		error = INVALID_NETWORK;
		handle_varError(avp, error);
	}


	return error;
}

int checkDnsMode(HLST *avp, HLST *nvram_avp, const char *varName)
{
	//if(ensureRadioChecked(avp,nvram_avp,varName) != NO_ERROR)
		//addToList(avp, varName, "auto");

	return NO_ERROR;
}


HTMLOBJECT *init_htmlVars(){

	int max_size = 8;
	int i=0;
	int max_index = max_size - 1;

	HTMLOBJECT *p = (HTMLOBJECT *)xmalloc(sizeof(HTMLOBJECT) * max_size);

	p[i].revise 	= checkStartIp;
	p[i++].name	= "lan_dhcp.startIp";

	p[i].revise 	= checkEndIp;
	p[i++].name	= "lan_dhcp.endIp";

	p[i].revise 	= checkIpRange;
	p[i++].name	= "lan_dhcp.isDhcp";

	p[i].revise 	= checkLeaseTime;
	p[i++].name	= "lan_dhcp.leaseTime";

	p[i].revise 	= checkDnsMode;
	p[i++].name	= "lan_dhcp.dnsMode";

	p[i].revise 	= checkPrimDns;
	p[i++].name	= "lan_dhcp.primaryDns";

	p[i].revise 	= checkSecDns;
	p[i++].name	= "lan_dhcp.secondaryDns";
       

	//sentinel
	p[i>max_index ? max_index : i].revise 	= NULL;
	p[i>max_index ? max_index : i].name	= NULL;

	return p;
}

void free_htmlVars(HTMLOBJECT *p){
	free(p);
	p = NULL;
}



void handle_varError(HLST *avp, int ecode){

 
	switch(ecode){

        case INVALID_START_IP:		addToList(avp, "lan_dhcp.startIp.error", "1"); 
					break;
								
	case INVALID_END_IP:		addToList(avp, "lan_dhcp.endIp.error", "1"); 
					break;

	case INVALID_ADDRESS_RANGE:	addToList(avp, "lan_dhcp.addressRange.error", "1"); 
					break;

	case INVALID_NETWORK:		addToList(avp, "lan_dhcp.network.error", "1"); 
					break;

	case INVALID_PRIMARY_DNS:	addToList(avp, "lan_dhcp.primaryDns.error", "1"); 
					break;

	case INVALID_SECONDARY_DNS:	addToList(avp, "lan_dhcp.secondaryDns.error", "1"); 
					break;

	case INVALID_LEASE_TIME:	addToList(avp, "lan_dhcp.leaseTime.error", "1"); 
					break;

	default	:      			//UNKNOWN error code, nothing to do
					logger (LOG_WARNING, "UNKNOWN error code: %d: ", ecode);
					return;
	}


	addToList(avp, "error", "1");
    
	return;

}




void *htinit (const char *doc_path, NVCL *nv) {
	STATE *s = ZMALLOC (STATE) ;

	s->htmlvars.objects = init_htmlVars();
	s->htmlvars.load_on_success 		= NULL;
	s->htmlvars.group_label			= GROUP_LABEL;
	s->htmlvars.total_elements_label	= NULL;

	s->nvram  	= nv ;
	s->htpath 	= doc_path ;

	return s ;
}

void htcleanup (void *desc) {
	STATE *s = desc ;

	free_htmlVars(s->htmlvars.objects);
	free(s);
}

void htrun (void *desc, RPLY_REC *r, int argc, const char **argv) {
    
	STATE  *s = desc ;
	PARSER   *parser ;
	HLST        *avp ;
	const char *data ;
	char *html_file = (char*)argv[1];

	if (argc < 2)
		return ;

	avp = http_args (r);

	CALL_HTTPVARS_ACTION(s,avp);
        
	handle_global_settings(avp, s->nvram);

	parser = open_parser (avp, s->nvram, s->htpath) ;

	if (file_parser (parser,  htdocs_p (r, html_file)) == 0)
		return ;

	/* parser */
	run_parser (parser) ;

	avdump (avp); /* debugging */

	/* print result */
	if ((data = data_collect (parser)) != 0)
		body_append (r, data, 0);

	/* clean up */
	file_parser_end (parser);
	close_parser (parser) ;
}



/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

