/*
 * nvram lib - flash data base client library
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: nvram-lib.c 3175 2006-09-23 19:17:11Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#ifdef HAVE_LIBGEN_H
# include <libgen.h>
#endif

#ifdef HAVE_STRING_H
# include <string.h>
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#define __NVRAMLIB_CLIENT_PRIVATE
#include "nvram-lib.h"

#include "logger.h"
#include "nvalloc.h"
#include "buffer.h"
#include "base64.h"
#include "varstr.h"
#include "tvcalc.h"

#undef WITH_DMALLOC

const char nvcl_string  [] = PACKAGE_STRING ;
const char nvcl_version [] = PACKAGE_VERSION ;

/* ----------------------------------------------------------------------- *
 * private variables
 * ----------------------------------------------------------------------- */

#define mesg static const char
mesg        _errstr [] = "*ERROR" ;
mesg nvcl_timeout_s [] = "Timeout waiting for NVRAM response" ;
mesg    nvcl_busy_s [] = "No answer from NVRAM yet - call in later";
mesg    nvcl_idle_s [] = "No NVRM response job outstanding";
#undef mesg

/* ----------------------------------------------------------------------- *
 * helpers for debugging
 * ----------------------------------------------------------------------- */

#ifdef WITH_DMALLOC
#define DMALLOC_VERIFY(d) {\
     dmalloc_message ("*** %s(%d): dmalloc NVCL check point.\n", \
                      __FILE__, __LINE__);                       \
                      dmalloc_verify (d);                        \
                      dmalloc_verify ((d)->recv.resp.error);     \
                      dmalloc_verify ((d)->recv.resp.avp);       \
                      dmalloc_log_stats ();                      \
        }
#else
#define DMALLOC_VERIFY(g) /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * private prototypes
 * ----------------------------------------------------------------------- */

static void error_response (NVCL*, const char*) ;
static int select_response (NVCL*, int) ;

/* ----------------------------------------------------------------------- *
 * static: text and argument handling
 * ----------------------------------------------------------------------- */

static
void put_arg (NVCL          *d,
              unsigned   delim,    /* prepend delimiter */
              unsigned   quote,    /* use with multiple words */
              const char *text,    /* not NULL */
              size_t       len) {  /* not zero => base64 data */

    const char *p, *b64 = 0 ;
    static char intro [2] = {NVCL_BASE64_INTRO, '\0'};

    /* already initialised ? */
    if (d->send.append_offs == 0)
        return ;

    if (len != 0 || text [0] == NVCL_BASE64_INTRO) { /* binary marker */
        if (len == 0) len = strlen (text) ;
#       ifdef DISCARD_HTTP
        text = x64alloc (intro, text, len, 0);
#       else
        text = x64alloc (intro, text, len, d->ip != 0);
#       endif
        len  = strlen (text) ;
        b64  = text ;
    }
    else if ((len = strlen (text)) == 0)
        len = 2, text = "\"\"" ; /* set to empty text */

    /* append space as argument delimiter */
    if (d->send.req_max < d->send.append_offs + 1 + len)
        goto end_up ;
    if (delim)
        d->send.request [d->send.append_offs ++] = delim ;
    
    /* no spaces - simple mode */
    if (quote == 0 || strchr (text, ' ') == 0) {
        memcpy (d->send.request + d->send.append_offs, text, len);
        d->send.append_offs += len ;
        goto end_up ;
    }

    /* need at least that room to store delimiters */
    if (d->send.req_max < d->send.append_offs + 2 + len)
        goto end_up ;
 
    /* store delimiter */
    d->send.request [d->send.append_offs ++] = quote ;

    /* filter the data: map: " => \" */
    while ((p = strchr (text, quote)) != 0) {
        int n =  p - text ;
        if (d->send.req_max < d->send.append_offs + 2 + len)
            goto end_up ;
        /* store text */
        if (n > 0) {
            memcpy (d->send.request + d->send.append_offs, text, n);
            d->send.append_offs += n ;
            len                 -= n ;
        }
        /* store escape and quotation mark */
        d->send.request [d->send.append_offs ++] =  '\\' ;
        d->send.request [d->send.append_offs ++] = quote ;
        len -- ;
        text = p + 1;
    }

    if (d->send.req_max < d->send.append_offs + 1 + len)
        goto end_up ;

    /* store rightmost text */
    if (len > 0) {
        memcpy (d->send.request + d->send.append_offs, text, len);
        d->send.append_offs += len ;
    }
    /* store quotation mark */
    d->send.request [d->send.append_offs ++] = quote ;

end_up:
    if (b64) free ((char*)b64);
}


/* ----------------------------------------------------------------------- *
 * static: assignment value pairs
 * ----------------------------------------------------------------------- */

static
AVP *new_avp (NVCL *d) {
    AVP *a ;
    
    DMALLOC_VERIFY(d) ;
    
    if (d->recv.resp.avp != 0) {
        int n = d->recv.resp.entries ++ ;
        d->recv.resp.avp = (AVP*) zrealloc
            (d->recv.resp.avp, d->recv.resp.entries * sizeof (AVP));
        
        return memset (d->recv.resp.avp + n, 0, sizeof (AVP));
    }

    d->recv.resp.entries = 1 ;
    a = d->recv.resp.avp = (AVP*) zmalloc (sizeof (AVP));

    DMALLOC_VERIFY(d) ;

    return a ;
}


static
AVP *top_avp (NVCL *d) {
    return d->recv.resp.avp != 0
        ? d->recv.resp.avp + d->recv.resp.entries - 1
        : 0
        ;
}


static
void kill_avp (NVCL *d) {

    if (d->recv.resp.avp != 0) {
        int n ;
        for (n = 0; n < d->recv.resp.entries; n ++) {
            if (d->recv.resp.avp [n].var &&
                d->recv.resp.avp [n].var != _errstr)
                free ((char*)d->recv.resp.avp [n].var);
            if (d->recv.resp.avp [n].val)
                free ((char*)d->recv.resp.avp [n].val);
        }
        free (d->recv.resp.avp);
        d->recv.resp.avp = 0;
    }
}


static
void strip_avp_val (AVP *avp) {

    size_t len = avp->len ;
    
    while (len > 0 && isspace (avp->val [len - 1]))
        len -- ;

    if (len < avp->len) {
        if ((avp->len = len) > 0) {
            avp->val = (char*) zrealloc ((char*)avp->val, len + 1);
            ((char*)avp->val) [len] = '\0' ;
        }
        else {
            free ((char*)avp->val);
            avp->val = 0 ;
        }
    }
}




static /* assuming: top_avp () exists, already */
unsigned parse_b64_field (NVCL *d, const char *p) {
    const char *q ;
    size_t len ;
    const char *error = x46alloc (&q, &len, p);

    if (error == 0) {
        AVP *avp = top_avp (d) ; /* assuming: always extists */
        avp->flg.binary =   1 ;
        avp->val        =   q ;
        avp->len        = len ;
        return 1;
    }
    
    error_response (d, error);
    return 0;
}


static
unsigned avp_parse (NVCL *d, char *data, size_t data_len) {

    char *q ;
    int count = 0 ;

    /* cycle through all the assignment value pairs */
    for ( /* n/a */; data_len > 0; data = q, count ++) {

        char *val, *eol ;
        AVP *avp ;
        size_t len ;

        if ((q = strchr (data, '\n')) != 0) {
            len       = q - data ;
            eol       = q ++     ; /* q: increments data */
            data_len -= ++ len   ; /* done with this record */
        }
        else {
            len       = strlen (data) ;
            eol       = data + len ;
            data_len -= len ;
        }
        
        /* strip trailing spaces */
        while (len > 0 && isspace (eol [-1]))
            len --, eol -- ;
        eol [0] = '\0' ;

        /* find the start of the variable name */
        while (isspace (* data))
            data ++ ;

        /* handle empty records */
        if (* data == '\0')
            continue ;

        /* begin new assignment value record */
        avp = new_avp (d) ;

        /* store the variable name */
        len = (val = strchr (data, '=')) != 0
            ? val - data
            : strlen (data)
            ;
        avp->var = memcpy (xmalloc (len + 1), data, len) ;
        ((char*)avp->var) [len] = '\0' ;

        /* parse value */
        if (val ++ == 0)
            continue ;
        
        /* check for binary data */
        if (* val == NVCL_BASE64_INTRO) {

            ++ val ;
            
            /* should we decode binary data ? */
            if (d->dont_b64 == 0) {
                parse_b64_field (d, val) ;
                continue ;
            }
            
            avp->flg.base64 = 1 ;
        }

        avp->len = eol - val;
        avp->val = memcpy (xmalloc (avp->len + 1), val, avp->len + 1);
        strip_avp_val (avp) ;
    }

    return count ;
}

/* ----------------------------------------------------------------------- *
 * static: connectionless/framed request handling
 * ----------------------------------------------------------------------- */

static
void update_framed_reply_path (NVCL *d) {
    struct stat st ;
    char *s ;

    /* check the reply socket path */
    s = strrchr (d->recv.un.sun_path, '/');
    * s = '\0' ; /* dirname */

    /* check whether there is a directory */
    if (stat (d->recv.un.sun_path, &st) == 0 && S_ISDIR (st.st_mode)) {

        /* is it accessible */
        if (access (d->recv.un.sun_path, W_OK|X_OK) != 0) {
            logger (LOG_ERR,
                    "Cannot access reply socket directory %s (%s) - abort",
                    d->recv.un.sun_path, strerror (errno));
            exit (2);
        }
        /* clean up (just to make sure .. ) */
        * s = '/' ;
        unlink (d->recv.un.sun_path);
        return ;
    }

    /* create the missing last segment of the reply subdirectory path */
    if (mkdir (d->recv.un.sun_path, 700) != 0) {
        logger (LOG_ERR,
                "Cannot create reply socket directory %s (%s) - abort",
                d->recv.un.sun_path, strerror (errno));
        exit (2);
    }

    d->recv.dir_made = 1 ; /* remember for clean up */
    * s = '/' ; /* undo dirname */
}


static /* returns reply fd or -1 */
int send_framed_request (NVCL *d) {

    /* finish up the request string */
    if (d->send.append_offs == 0)
        d->send.append_offs = d->send.req_start;

    /* cancel pending request */
    if (d->sock_fd >= 0) {
        close (d->sock_fd) ;
        unlink (d->recv.un.sun_path);
    }

    /* create reply socket */
    if ((d->sock_fd = socket (PF_UNIX, SOCK_DGRAM, 0)) < 0) {
        logger (LOG_ERR,
                "Cannot create NVRAM client datagram socket: %s",
                strerror (errno));
        return 0;
    }

    /* bind to the reply socket */
    if (bind (d->sock_fd, (struct sockaddr*)&d->recv.un,
              d->recv.un_size) < 0) {
        /* clean up and retry */
        update_framed_reply_path (d) ;
        if (bind (d->sock_fd, (struct sockaddr*)&d->recv.un,
                  d->recv.un_size) < 0) {
            logger (LOG_ERR,
                    "Cannot bind NVRAM client reply socket to %s: %s",
                    d->recv.un.sun_path, strerror (errno));
            goto close_on_error;
        }
    }

    /* send the request */
    if (sendto (d->sock_fd, d->send.request, d->send.append_offs, 0,
                (struct sockaddr*)&d->send.un, d->send.un_size) < 0) {
        logger (LOG_ERR, "Cannot send NVRAM command via %s: %s",
                d->send.un.sun_path, strerror (errno));
        goto close_on_error;
    }

    return 1 ;

close_on_error:
    close (d->sock_fd) ;
    d->sock_fd = -1 ;
    unlink (d->recv.un.sun_path);
    return 0;
}

/* ----------------------------------------------------------------------- *
 * static: data stream request handling
 * ----------------------------------------------------------------------- */

#ifndef DISCARD_HTTP

static
unsigned connect_server (NVCL *d) {

    struct sockaddr_in in ;

    /* open http socket */
    if ((d->sock_fd = socket (PF_INET, SOCK_STREAM, 0)) == -1) {
        logger (LOG_ERR, "Cannot create NVRAM http socket: %s",
                strerror (errno));
        return 0;
    }

    XZERO (in);
    in.sin_family = AF_INET ;
    in.sin_port   = htons (d->port) ;
    if (d->ip) in.sin_addr.s_addr = htonl (d->ip) ;

    if (connect (d->sock_fd, (struct sockaddr*)&in, sizeof in) < 0) {
        logger (LOG_ERR, "Cannot connect to %s:%u: %s",
                ip4addr_pp (d->ip), (unsigned)d->port, strerror (errno));
        close (d->sock_fd);
        d->sock_fd = -1 ;
        return 0;
    }

    return 1;
}


static /* returns reply fd or -1 */
int send_http_request (NVCL *d) {

    int n ;

    IOBUF  buf ;
    XZERO (buf);

    data_printf (&buf,
                 "POST %s HTTP/1.1\r\n"
                 "Host: %s\r\n"
                 "Accept: text/plain\r\n"
                 "User-Agent: Maxina/NVRAM-client $Revision: 3175 $\r\n",
                 d->send.request, ip4addr_pp (d->ip));

    if (d->keep_alive)
        data_printf (&buf,
                     "Connection: Keep-Alive\r\n"
                     "Keep-Alive: %d\r\n", (int)d->keep_alive);

    if ((n = d->send.append_offs - d->send.args_start) > 0) {
        char *p = d->send.request + d->send.args_start ;
	
	/* one spare byte left */
        p [d->send.append_offs - d->send.args_start] = '\0' ; 
        data_printf (&buf, "Content-Length: %d\r\n\r\n%s", n + 2, p) ;
    }

    data_printf (&buf, "\r\n");

    /* send request packet */
    if ((n = send (d->sock_fd, buf.data, buf.top, 0)) < 0) {

        if (errno == EBADF) {
            int err = errno ;

            if (d->sock_fd >= 0 && close (d->sock_fd < 0))
                logger (LOG_ERR, "Error closing NVRAM http socket: %s",
                        strerror (errno));

            if (connect_server (d) != 0)
                n = send (d->sock_fd, buf.data, buf.top, 0) ;
            else
                errno = err ;
        }
    }
    
    kill_space (&buf);              

    if (n < 0) {
        logger (LOG_ERR, "Cannot send on the NVRAM socket: %s %u",
                strerror (errno), errno);
        return 0;
    }
 
    return 1 ;
}

#endif /* DISCARD_HTTP */


/* ----------------------------------------------------------------------- *
 * static: request handling
 * ----------------------------------------------------------------------- */

static /* returns reply fd or -1 */
int send_request (NVCL *d) {
    int n ;
    
#   ifndef DISCARD_HTTP
    if (d->ip != 0)
        n = send_http_request (d) ;
    else
#   endif /* DISCARD_HTTP */
        n = send_framed_request (d) ;

    if (n) {
        /* the date of the waiting time deadline */
        gettimeofday (&d->recv.over, 0);
        tval_add_ms  (&d->recv.over, d->recv.timeout_ms);
    }

    return n ;
}

/* ----------------------------------------------------------------------- *
 * static: general response handling, helpers
 * ----------------------------------------------------------------------- */

static
void error_response (NVCL *d, const char *p) {
    size_t n = strlen (p) + 1;

    if (d->recv.resp.error != 0)
        free (d->recv.resp.error);

    d->recv.resp.error = memcpy (xmalloc (n), p, n);

    /* strip trailing new-lines */
    while (n > 1 && isspace (d->recv.resp.error [n - 2]))
        d->recv.resp.error [n -- - 2] = '\0' ;
}



/* ----------------------------------------------------------------------- *
 * static: framed response handling
 * ----------------------------------------------------------------------- */

static
unsigned parse_framed_response (NVCL *d, char *q) {

    char *p = q ;

    /* a good answer looks like: "+OK ..." */
    if (strncmp (p, "+OK", 3) != 0) {

        if (strncmp (p, "-ERR:", 5) == 0) {
            p += 5 ; 
            while (* p && * p == ' ')
                p ++ ;
        }
        else if (* p == '\0') {
            p = "No data received from the NVRAM server." ;
        }

        error_response (d, p) ;
        return 0;
    }

    /* find start of data */
    p += 3 ; 
    while (* p && * p != '\n')
        p ++ ;

    if (* p ++ == 0) {
        if (q [3] == ':') {
            p = q + 4;
            while (* p && * ++ p == ' ')
                ;
        }
        else
            p = "" ;
    }

    avp_parse (d, p, strlen (p)) ;
    return 1;
}


/* ----------------------------------------------------------------------- *
 * static: http response handling
 * ----------------------------------------------------------------------- */

#ifndef DISCARD_HTTP

static
unsigned more_http_data (NVCL *d, IOBUF *b, char **dp) {

    int n ;
    
    /* initialize unless done yet */
    if (b->data == 0)
        data_printf (b, "%s", *dp);
    *dp = b->data ;

    make_space (b, 2048) ;

    if (select_response (d, 20) <= 0)
        return 0 ;

    if ((n = recv (d->sock_fd, b->data + b->top, b->size - b->top - 1,
                   MSG_DONTWAIT)) < 0) {
        logger (LOG_ERR,
                "Error reading from the NVRAM http socket: %s",
                strerror (errno));
        return 0;
    }

    if (n == 0) /* ench of data transmission */
        return 0 ;

    /* prepare for text parsing */
    b->data [b->top += n] = '\0' ;

    return n;
}


static
unsigned parse_http_response (NVCL *d, IOBUF *buf, char *p) {

    char *q, *r;

    size_t len, content_length  = 0 ;
    int content_type_ok = 0 ;
    int content_length_ok = 0 ;

    /* check whether we got a complete first line */
    while ((q = strchr (p, '\n')) == 0)
        if (more_http_data (d, buf, &p) == 0)
            return 0 ;
    if ((r = q ++) [-1] == '\r')
        r -- ;
    * r = '\0' ;

    /* verify the http header syntax: "http/1.0" or similar */
    if (strncasecmp (p, "http/", 5) != 0)
        return 0 ;
    p += 5 ;
    if (!('0' <= p [0] && p [0] <= '9') ||
        !                (p [1] == '.') ||
        !('0' <= p [2] && p [2] <= '9')) 
        return 0 ;
    p += 3 ;

    /* parse the return code */
    while (*p == ' ') p ++ ;
    if (strncmp (p, "200 ", 4) != 0) {
        int n ;
        if (p [0] < '1' || '9' < p [0] ||
            p [1] < '0' || '9' < p [1] ||
            p [2] < '0' || '9' < p [2] || p [3] != ' ')
            return 0 ;
        n = strlen (p) + 1;
        if (d->recv.resp.error != 0)
            free (d->recv.resp.error);
        d->recv.resp.error = memcpy (xmalloc (n), p, n);
    }
    p = q ;
    
    /* parse other header lines */
    for (;;) {
        char *var ;
        unsigned u ;

        /* check for the end-of-header */
        if (p [0] == '\r' && p [1] == '\n') {
            p += 2 ;
            break ;
        }

        /* terminate this header line */
        while ((q = strchr (p, '\n')) == 0)
            if (more_http_data (d, buf, &p) == 0)
                return 0 ;
        if ((r = q ++) [-1] == '\r')
            r -- ;
        * r = '\0' ;

        /* start of the variable */
        var = p ;

        /* advance p to the next line */
        p = q ;

        /* isolate variable */
        if ((q = strchr (var, ':')) == 0)
            return 0 ;
        r = q ++ ;
        while (r [-1] == ' ')
            r -- ;
        * r = '\0' ;

        /* get value */
        while (* q == ' ') q ++ ;
        
        if (strcasecmp (var, "Content-Type") == 0) {
            if (strcasecmp (q, "text/plain") != 0)
                return 0 ;
            content_type_ok = 1;
        }
        else if (strcasecmp (var, "Content-Length") == 0) {
            if (parse_unsigned (&u, q, 0) == 0 || u < 0)
                return 0 ;
            content_length    = u ;
            content_length_ok = 1 ;
        }
    }
    
    if (content_length_ok) {
        /* make sure that the body is complete */
        while ((len = strlen (p)) < content_length)
            /* strlen: quite inefficient but rarely used, hopefully */
            if (more_http_data (d, buf, &p) == 0)
                return 0 ;
    }
    else {
        /* take the rest as is */
        content_length = strlen (p) ;
    }

    /* check for an error explanation */
    if (d->recv.resp.error && strchr (p, '\n') != 0) {
        AVP *avp = new_avp (d) ;
        avp->var = _errstr ;
        avp->len = content_length ;
        avp->val = memcpy (xmalloc (content_length + 1), p, content_length);
        strip_avp_val (avp) ;
        return 1;
    }

    avp_parse (d, p, content_length) ;
    return 1;
}

#endif /* DISCARD_HTTP */


/* ----------------------------------------------------------------------- *
 * static: response handling
 * ----------------------------------------------------------------------- */

static
int select_response (NVCL *d, int wait_ms)  {

    fd_set set;
    struct timeval tmo;

    /* check whether we got a reponse, already */
    FD_ZERO (&set) ;
    FD_SET (d->sock_fd, &set) ;

    if (wait_ms) {

        /* wait for while reading the response ... */
        struct timeval now;
        gettimeofday (&now, 0);

        /* waiting time is over ? */
        if (tval_le_tv (&d->recv.over, &now) != 0)
            return -7 ;

        /* assemble maximum timeout */
        tmo = d->recv.over ;
        tval_sub_tv (&tmo, &now); /* tmo := over - now */

        /* use the smaller alternative */
        if (tval_lt_ms (&tmo, wait_ms)) {
            tmo.tv_sec = wait_ms / 1000 ;
            tmo.tv_usec = (wait_ms % 1000) * 1000 ;
        }
    }
    else {

        /* asynchroneous call, snoop only */
        XZERO (tmo);
    }

    /* check for data, returns -1, 0, or +1 */
    return select (d->sock_fd + 1, &set, 0, 0, &tmo) ;
}


static
void flush_response (NVCL *d) {
    if (d->recv.resp.error)
        free (d->recv.resp.error);
    kill_avp (d);
    XZERO (d->recv.resp);
}


static
unsigned parse_response (NVCL *d, char *p) {
    int n ;

    flush_response (d);
    
#   ifndef DISCARD_HTTP
    if (d->ip != 0) {
        IOBUF  buf ;
        XZERO (buf);
        n = parse_http_response (d, &buf, p) ;
        kill_space (&buf) ;
    }
    else
#   endif /* DISCARD_HTTP */

        n = parse_framed_response (d, p) ;

    return n ;
}


static
void process_errno (NVCL *d, const char *fmt, int err) {

    const char *t = strerror (err) ;
    char       *s = (char*) xmalloc (strlen (fmt) + strlen (t) + 1);
    
    sprintf (s, fmt, t);
    logger (LOG_ERR, s);
    error_response (d, s);
    free (s);
}


/* ----------------------------------------------------------------------- *
 * public
 * ----------------------------------------------------------------------- */

NVCL *nvcl_open (const char *chrt,
#                ifndef DISCARD_HTTP
                 ip4addr_t ip, uint16_t port,
#                endif /* DISCARD_HTTP */
                 unsigned tmo_ms) {
    NVCL   d ;
    XZERO (d);

#   ifndef DISCARD_HTTP
    if (ip || port) {
        if ((d.ip = ip) == 0)
            d.ip = 0x7f000001 ; /* 127.0.0.1 in host notation */
        if ((d.port = port) == 0)
            d.port = NVCL_HTTP_CONFIG_PORT ;

        if (connect_server (&d) == 0) {
            logger (LOG_INFO, "NVRAM client: cannot connect to %s:%u: %s",
                    ip4addr_pp (d.ip), (unsigned)d.port, strerror (errno));
            return 0 ;
        }
#       ifdef NVCL_OPEN_LOG
        if (verbosity & NVCL_OPEN_LOG)
            logger (LOG_INFO, "NVRAM client: connection to %s:%u ready",
                    ip4addr_pp (d.ip), (unsigned)d.port);
#       endif
    }
    else {
#   endif /* DISCARD_HTTP */
       
        /* assemble the request socket address */
        d.send.un.sun_family = AF_UNIX ;
        strncpy (d.send.un.sun_path, NVCL_REQUEST_SOCKET, UNIX_PATH_MAX);
        d.send.un.sun_path [UNIX_PATH_MAX-1] = '\0' ;
        d.send.un_size = (sizeof (d.send.un.sun_family) +
                          strlen (d.send.un.sun_path));

        /* assemble the reply prefix and store it in the request buffer */
        if (chrt == 0) chrt = NVCL_DEFAULT_CHROOT ;
        sprintf (d.send.request, "%s" NVCL_REPLY_PFX "%x", chrt, getpid ());

        /* assemble the reply socket address */
        d.recv.un.sun_family = AF_UNIX ;
        strncpy (d.recv.un.sun_path, d.send.request, UNIX_PATH_MAX);
        d.recv.un.sun_path [UNIX_PATH_MAX-1] = '\0' ;
        d.recv.un_size = (sizeof (d.recv.un.sun_family) +
                          strlen (d.recv.un.sun_path));

        /* set the append offset for the request buffer */
        d.send.req_start = strlen (d.send.request) ;
    
        /* no outstanding request, yet */
        d.sock_fd = -1 ;

        update_framed_reply_path (&d);

#       ifdef NVCL_OPEN_LOG
        if (verbosity & NVCL_OPEN_LOG)
            logger (LOG_INFO, "NVRAM client reply socket path is %s",
                    d.recv.un.sun_path);
#       endif

#   ifndef DISCARD_HTTP
    }
#   endif /* DISCARD_HTTP */

    /* initialise data send buffer */
    d.send.req_max = sizeof (d.send.request) - d.send.req_start - 1;

    /* msecs maximum waiting time */
    if ((d.recv.timeout_ms = tmo_ms) == 0) /* default timeout, msecs */
        d.recv.timeout_ms = NVCL_DEFAULT_REPLY_TMO ;

    return zmemdup (&d);
}


void nvcl_close (NVCL *d) {

    if (d->is_cloned == 0
#   ifndef DISCARD_HTTP
        && d->ip == 0
#   endif /* DISCARD_HTTP */
        ) {
        unlink (d->recv.un.sun_path) ;
        if (d->recv.dir_made) {
            * strrchr (d->recv.un.sun_path, '/') = '\0' ; /* dirname */
            rmdir (d->recv.un.sun_path) ;
        }
    }

    flush_response (d);
    if (d->is_cloned == 0 && d->sock_fd >= 0) {
#       ifndef DISCARD_HTTP
        if (d->ip != 0)
            shutdown (d->sock_fd, SHUT_RDWR);
#       endif /* DISCARD_HTTP */
        close (d->sock_fd);
    }
    free (d);
}


NVCL *nvcl_clone (NVCL *d) {
    NVCL *newd = zmemdup (d);

    newd->is_cloned = 1;
    
    /* copy AVP */
    if (d->recv.resp.entries) {
        int n ;
        newd->recv.resp.avp = zmalloc
            ((1 + d->recv.resp.entries) * sizeof (*d->recv.resp.avp));
        for (n = 0; n < d->recv.resp.entries; n ++) {
            newd->recv.resp.avp [n] = d->recv.resp.avp [n] ;
            if (d->recv.resp.avp [n].var)
                newd->recv.resp.avp [n].var = zstrdup
                    (d->recv.resp.avp [n].var);
            if (d->recv.resp.avp [n].val)
                newd->recv.resp.avp [n].val = zstrdup
                    (d->recv.resp.avp [n].val);
        }
    }

    if (d->recv.resp.error)
        newd->recv.resp.error = zstrdup (newd->recv.resp.error) ;

    return newd ;
}


void nvcl_putcmd (NVCL *d, const char *cmd) {

    size_t len = strlen (cmd);

    if (d->send.req_start + len + 2 < d->send.req_max) {

#       ifndef DISCARD_HTTP
        if (d->ip != 0) {
            if (3 + len < d->send.req_max) {
                /* remote http access */
                d->send.request [0] = '/' ;
                memcpy (d->send.request + 1, cmd, len);
                d->send.request [len + 1] = '\0' ;
                d->send.args_start =
                    d->send.append_offs = len + 2;
            }
            return ;
        }
#       endif /* DISCARD_HTTP */

        /* unix domain socket frame */
        d->send.append_offs = d->send.req_start ;
        if (d->send.append_offs + 1 + len < d->send.req_max) {
            d->send.request [d->send.append_offs ++] = ' ' ;
            memcpy (d->send.request + d->send.append_offs, cmd, len) ;
            d->send.request [d->send.append_offs += len] = '\0' ;
        }
    }
}


void nvcl_putavpair (NVCL *d,
                     const char *var,
                     const char *val, size_t len) {

#   ifndef DISCARD_HTTP
    if (d->ip != 0) {
        put_arg (d, d->send.request [d->send.append_offs - 1] ? '&' : 0, 0,
                 var, 0);
        if (val)
            put_arg (d, '=', 0, val, len);
        return ;
    }
#   endif /* DISCARD_HTTP */

    put_arg (d, ' ', 0, var, 0);
    if (val)
        put_arg (d, ' ', '"', val, len);
}


/* special handling depending on msecs:
 * 
 *    msecs < 0:      snoop only, don't return an error when the
 *                    result data have not yet been received 
 *
 *    msecs == 0:     wait until the result data have been received
 *                    (or return an error)
 *
 *    msecs > 0:      wait at most msecs milli seconds for the result
 *                    data have, return an error if this time is over
 *
 *
 * notes on the return code:
 * 
 *    ok:                1
 *    
 *    failure:          -1
 *    
 *    waiting for data:  0
 */   
  
int nvcl_rply (NVCL *d, int msecs) {

    int wait_for_data_ms ;

    if (d->send.append_offs) {
        /* aways start a new request when there is an argument set up */
        if (send_request (d) == 0)
            return 0 ;
    }
    else if (d->sock_fd < 0) {
        logger (LOG_ERR, nvcl_idle_s);
        return 0;
    }

    wait_for_data_ms = msecs == 0
        ? d->recv.timeout_ms      /* max waiting time for select() */
        : (msecs > 1 ? msecs : 0) /* msecs == -1 or 1: snoop only */
        ;

    /* efficient handling of msecs == 1:
     *
     * wait_for_data_ms == 0 => snoop only
     * msecs > 0             => error if data are not available
     */

    switch (select_response (d, wait_for_data_ms)) {
    case 0: {
        if (msecs < 0) {
#           ifdef NVCL_TIMEOUT_LOG
            if (verbosity & NVCL_TIMEOUT_LOG) logger (LOG_INFO, nvcl_busy_s);
#           endif
            /* async mode: not cleaning up, yet */
            return 0 ;
        }
        /* otherwise timeout */
#       ifdef NVCL_TIMEOUT_LOG
        if (verbosity & NVCL_TIMEOUT_LOG) logger (LOG_INFO, nvcl_timeout_s);
#       endif

        error_response (d, nvcl_timeout_s);
        return -1;
    }
        
    case 1: { /* data are available, now -- so get them */

        char buf [NVCL_IOBUF_SIZE];
        int n = recv (d->sock_fd, buf, sizeof (buf) - 1, 0);

        if (n < 0) {
            process_errno
                (d, "Error retrieving data from NVRAM (%s)", errno);
            return -1 ;
        }

        buf [n] = '\0';
        n = parse_response (d, buf);

        if (d->keep_alive == 0) {

#           ifndef DISCARD_HTTP
            if (d->ip != 0)
                shutdown (d->sock_fd, SHUT_RDWR);
            else
#           endif
                unlink (d->recv.un.sun_path);

            close (d->sock_fd) ;
            d->sock_fd = -1 ;
        }
        return n ;
    }

    case -7:
#       ifdef NVCL_TIMEOUT_LOG
        if (verbosity & NVCL_TIMEOUT_LOG)
            logger (LOG_INFO, "Too late, not waiting for NVRAM response");
#       endif

        error_response (d, "Too late, not waiting for NVRAM response");
        return 0;
    }
    
    process_errno (d, "Error waiting for a NVRAM response (%s)", errno);
    return -1 ;
}

unsigned nvcl_getavpair (NVCL *d,
                         const char **Var,
                         const char **Val, size_t *Len, AVFLGS *Flg) {
    AVP *avp ;
    
    /* check whether we got an avp to retrieve, at all */
    if (d->recv.resp.avp == 0 ||
        d->recv.resp.entries <= d->recv.resp.avp_seen)
        return 0;

    /* get record and increment to the next position */
    avp = d->recv.resp.avp + d->recv.resp.avp_seen ++ ;

    if (Var && (*Var = avp->var) == 0) *Var = "" ;
    if (Val && (*Val = avp->val) == 0) *Val = "" ;
    if (Len)    *Len = avp->len ;
    if (Flg)    *Flg = avp->flg ;
    
    return 1;
}


unsigned nvcl_base64_ok (NVCL *d, unsigned mode) {
    int n = d->dont_b64 ;
    d->dont_b64 = !mode ;
    return n ;
}


#ifndef DISCARD_HTTP
unsigned nvcl_http_ok (NVCL *d) {
    return d->ip != 0;
}
#endif


time_t nvcl_keep_alive (NVCL *d, time_t tmo) {
    time_t ka = d->keep_alive ;
    d->keep_alive = tmo ;
    return ka ;
}

const char *nvcl_rply_error (NVCL *d) {
    return d->recv.resp.error ? d->recv.resp.error : "" ;
}


unsigned nvcl_b64decode (char       **Q,
                         size_t      *L,
                         const char **E,
                         const char *in) {
    const char *err = x46alloc ((const char **)Q, L, in) ;
    if (err == 0)
        return 1 ;
    if (E)
        *E = err ;
    return 0 ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
