/*
 * Copyright (C) 2003-2006 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: varstr.h 3211 2006-10-04 11:47:04Z jordan $
 */

#ifndef __SHAT_VARSTR_H
#define __SHAT_VARSTR_H

#include "buffer.h"

typedef
enum _scalar_type {
    t_no_type = 0,
    t_bool,                  /* bit field */
    t_signed,                /* integer, displayed as signed */
    t_unsigned,              /* integer, displayed as unsigned */
    t_hex,                   /* integer, displayed as hex */
    t_string,                /* fixed length character string */
    t_data,                  /* unspecified data */
    t_enum,                  /* integer, extension possible */
} SCALAR_TYPE ;

typedef
struct _params {
    const char *name ;       /* parameter name used for access */
    struct {
        SCALAR_TYPE type ;   /* data type */
        size_t    offset ;   /* offset within the data structure */
        size_t      size ;   /* sizeof () this data type */
        unsigned     ext ;   /* bit field or enum_t type extension */
    } v ;
    short flags ; /* whatever you like, eg: 1 == edit ok */
} PARAMS ;

#ifdef __VARSTR_HELPERS

/* get the byte offset of a field */
#define __STRUCT_OFFSET(type, field) ((char*)(&((type*)0)->field)-(char*)0)
#define __STRUCT_SIZEOF(type, field) sizeof (((type*)0)->field)

/*
 * allocate a non-boolean PARAMS scalar definition:
 *
 * sctype: SCALAR_TYPE
 * type:   structure type
 * field:  name of the structure field
 * ext:    extension field "ext"
 */

#define __VAR_SCALR(sctype, type, field, ext) \
              # field, {sctype, __STRUCT_OFFSET (type, field),    \
                                __STRUCT_SIZEOF (type, field), ext}

/*
 * allocate a boolean PARAMS bit field definition:
 *
 * type:   structure type
 * field:  base name of the structure field
 * leaf:   leaf name of the structure field
 * bnum:   bit fieled number (ranging 1 .. 7)
 *
 * a typical bit field structure looks like
 *
 * struct {
 *   ...
 *   struct {
 *      leaf : 1;
 *       ...
 *   } field ;
 *   ...
 * };
 *
 * Restrictions: due to the gcc compiler method of allocating
 *               bit fields, there are at most 7 fileds supported.
 */

#define __VAR_BOOL(type, field, leaf, bnum) \
      # field "." # leaf, {t_bool, __STRUCT_OFFSET (type, field),     \
                                                    sizeof (int), bnum}
#endif

/* error handling */

typedef
enum _varstr_error_code {
    EVARSTR_no_error = 0,

    /* errno/system errors */

    EVARSTR_error_bottom = 0x100,

    EVARSTR_data_type_undefined,

    EVARSTR_signed_integer_string_termination,
    EVARSTR_signed_integer_string_out_of_range,
    EVARSTR_signed_integer_size_undefined,

    EVARSTR_unsigned_integer_string_termination,
    EVARSTR_unsigned_integer_string_out_of_range,
    EVARSTR_unsigned_integer_size_undefined,

    EVARSTR_bool_integer_size_undefined,
    EVARSTR_bool_integer_string_out_of_range,
    EVARSTR_bool_integer_bit_not_defined,

    EVARSTR_enum_integer_size_undefined,
    EVARSTR_enum_integer_string_out_of_range,
    EVARSTR_enum_integer_value_out_of_range,
    EVARSTR_enum_parser_not_available,
    EVARSTR_enum_parser_returns_error,

    EVARSTR_string_text_too_long,
    EVARSTR_string_parser_hex_dgt_error,
    EVARSTR_string_parser_oct_dgt_error,

    EVARSTR_prefixed_entry_is_not_unique,
    EVARSTR_prefixed_entry_does_not_exist,

    EVARSTR_error_top,
} varstr_error_code ;

extern varstr_error_code varstr_errno ;
extern const char *varstr_strerr (varstr_error_code) ;

/*
 * Find the index of the "testee" string in the argument "list" and return
 *   the non-negaitve index, or
 *   -1 if there is no such entry, or
 *   -2 if there are more than one entries of the argiment list that
 *      begin with the same prefix (in short: "testee" is not unique)
 *
 * If the 2nd argument "dim" is NULL, the list is assumed to terminate
 * with a NULL string pointer (and has no more than 100 entries).
 *
 * Example1:
 *
 *   char * list1 [10] = { "text1", "text2", .... } ;
 *   varstr_list_index (list1, 0, 10, sizeof (char*), "testee", 0)
 *
 * Example2:
 *
 *   #define DIM(x) (sizeof (x) / sizeof (x [0]))
 *   PARAMS params [] = { ... } ;
 *   varstr_list_index (params, 0,
                        DIM (params), sizeof (PARAMS), "testee", 0);
 *
 * Example3:
 *
 *   #define DIM(x) (sizeof (x) / sizeof (x [0]))
 *   struct { name [20], ... } list2 [] = {{"text2", ..}, {"text2", ..} ..};
 *   varstr_list_index (list2, sizeof (list2 [0].name),
 *                      DIM (list2), sizeof (list2), "testee", 0);
 *
 */

extern int varstr_list_index (const void*list, size_t len,
                              unsigned dim, size_t width,
                              const char*testee, size_t tlen);

/*
 * Use the data struct "data" and extract the value of the target
 * field as  described by "entry".  Append a human readable version of
 * that fieled name and value to the "out" data stream.
 *
 * In the target field is an enum_t type and the function pointer "extfn"
 * is non-NULL, call a function like
 *
 *    const char *extfn (unsigned value, unsigned ext)
 *
 * and pass it the actual value and extension number (see the definition
 * of PARAMS). If extfn() returns some string, use it as human readable
 * representation of value, otherwise act as there were an eror.
 *
 * varstr_param_append () returns 0 if there was a parse error, and 1
 * otherwise.
 *
 * Example:
 *
 *    IOBUF *out ;
 *    PARAMS *entry ;
 *    struct { ... } data ;
 *    // print all entries of the structure data
 *    for (n = 0; n < DIM (params); n ++)
 *       varstr_param_append (out, entry, data, extfn, 0) ;
 */

extern unsigned varstr_param_append (IOBUF *out, const PARAMS *entry,
                                     const void *data,
                                     const char *(*extfn)(unsigned,unsigned),
                                     const char *quick_delim);

/*
 * Reverse the varstr_param_append() function, parse the argument "value"
 * and store it into argument "data" (the proper position described by
 * the argument "entry").
 *
 * In case of an extension data type, if available, the argument function
 *
 *    const char *extfn (unsigned value, unsigned ext)
 *
 * is used to check an integer argument value, otherwise the argument
 * function (if available)
 *
 *    unsigned loadfn (const char *value, unsigned ext)
 *
 * is used to parse an non numeric value.
 */

extern unsigned varstr_param_load (const PARAMS *entry, void *data,
                                   const char *value, size_t len,
                                   const char *(*extfn)(unsigned,unsigned),
                                   unsigned (*loadfn)(const char*,unsigned));

extern unsigned parse_unsigned (unsigned*,const char*,const char*);
extern unsigned parse_signed   (int*,const char*,const char*);
extern int      parse_oct_dgt  (int);
extern int      parse_hex_dgt  (int);

#endif /* __SHAT_VARSTR_H */
