/*
 * Copyright (C) 2006 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: ctrl.c 3211 2006-10-04 11:47:04Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef DISCARD_CMD_SOCKET

#include <stdio.h>
#include <libgen.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_SYS_UN_H
# include <sys/un.h>
#else
# ifdef HAVE_LINUX_UN_H
# include <linux/un.h>
# endif
#endif

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif

#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NET_IF_H
#include <net/if.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_IF_ETHER_H
#include <netinet/if_ether.h>
#endif

#include "varstr.h"
#include "util.h"
#include "nvalloc.h"
#include "logger.h"

#define __CTRL_PRIVATE
#include "ctrl.h"

#undef WITH_DMALLOC

#ifdef WITH_DMALLOC
#define DMALLOC_VERIFY(c) {\
     dmalloc_message       \
       ("*** %s(%d): dmalloc check point.\n", __FILE__, __LINE__); \
     dmalloc_verify (c);   \
     dmalloc_log_stats (); }
#else
#define DMALLOC_VERIFY(c) /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * variables and definitions
 * ----------------------------------------------------------------------- */

#ifndef  O_NDELAY
# define O_NDELAY O_NONBLOCK
#endif

#ifndef   S_ISSOCK
# ifdef   S_IFSOCK
#  define S_ISSOCK(m) (((m) & S_IFMT) == S_IFSOCK)
# endif
#endif

#define DIM(a) (sizeof (a) / sizeof (a [0]))

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

/* example for a communication handler */
void usage_handler (void *ctrl, int argc, const char **argv) {
    CTRL *ct = ctrl;
    const char **p;
    int n ;

    /* prepend by a first argument "?" */
    p = (const char **)xmalloc ((argc + 1) * sizeof (*argv)) ;
    if (argc)
        memcpy (p + 1, argv, argc * sizeof (*argv)) ;
    p [0] = "?";
        
    for (n = 0; n < ct->dim; n++) {
        if (ct->service [n].fn != ctrl_usage) { /* dont call yourself */
            struct _service *svc = ct->service + n ;
            p [1] = svc->name ;
            (*svc->fn) (svc->use_ctrl ? ct : svc->fn_state,
                        argc + 1,
                        (const char**)p);
        }
    }

    free (p);
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

#ifndef DISCARD_EXTENSIVE_LOGGING
static
void delayed_debug_logger (CTRL *ct) {
    if (ct->on_leave.request_log.top) {
        if (verbosity & CTRL_REQUEST_LOG)
            logger (LOG_DEBUG, ct->on_leave.request_log.data);
        kill_space (&ct->on_leave.request_log);
    }
}
#endif

/* execute request */
void ctrl_handler (CTRL *ct) {

    int argc, argi, n;
    char buf [CTRL_MAX_REQUEST_SIZE], *argv [20], *p;
    struct sockaddr_un sun ;
    socklen_t slen ;

#   ifndef DONT_USE_RECVMSG
    struct {
        struct iovec    iov [1];
        struct msghdr   msg [1];
        struct cmsghdr cmsg [1];
        struct ucred sender [1]; /* cms data following the cmsg */
    } m ;

    XZERO (m) ;
    m.msg [0].msg_iov        = m.iov;
    m.msg [0].msg_iovlen     = 1;
    m.msg [0].msg_control    = m.cmsg;
    m.msg [0].msg_controllen = sizeof (m.cmsg) + sizeof (m.sender);

    /* read request into buffer - leave one byte before input starts */
    m.iov [0].iov_base = buf + 1;
    m.iov [0].iov_len  = sizeof (buf) - 2;
    if ((n = recvmsg (ct->cmd_fd, m.msg, MSG_TRUNC)) < 0) {
        logger (LOG_ERR, "command: Error reading from %s: ",
                ct->un.sun_path, strerror (errno));
        return ;
    }

    if ((m.msg [0].msg_flags & MSG_CTRUNC) == 0 &&
        m.cmsg [0].cmsg_len   >= sizeof (m.cmsg) + sizeof (m.sender) &&
        m.cmsg [0].cmsg_level == 1 &&
        m.cmsg [0].cmsg_type  == SCM_CREDENTIALS) {
        ct->uid = m.sender [0].uid ;
        ct->gid = m.sender [0].gid ;
        ct->pid = m.sender [0].pid ;
    }
    else {
        ct->uid = -1 ;
        ct->gid = -1 ;
        ct->pid = -1 ;
    }

#   else   
    slen = sizeof (sun) ;

    /* read request into buffer - leave one byte before input starts */
    if ((n = recvfrom (ct->cmd_fd, buf + 1, sizeof (buf) - 2, MSG_TRUNC,
                       (struct sockaddr*)&sun, &slen)) < 0) {
        logger (LOG_ERR, "command: Error reading from %s: ",
                ct->un.sun_path, strerror (errno));
        return ;
    }
#   endif

    DMALLOC_VERIFY (ct) ;

    if (n == 0) return ;
    buf [n+1] = '\0'; /* starting at buf+1, zero terminated - easy parsing */

#ifndef DISCARD_EXTENSIVE_LOGGING
    if (verbosity & CTRL_REQUEST_LOG)
        data_printf (&ct->on_leave.request_log, "command request"
#   ifndef DONT_USE_RECVMSG
                " from uid=%d, gid=%d, pid=%u: \"%s\"",
                ct->uid, ct->gid, ct->pid,
#   else
                ": \"%s\"",
#   endif
                buf+1);
#   endif /* CTRL_REQUEST_LOG */

    DMALLOC_VERIFY (ct) ;

    /* parse buffer, split into space separated arguments */
    for (argc = 0, p = buf + 1; p && *p && argc < DIM (argv); argc ++) {
        int c = ' ';
        while (* p == ' ')
            p ++ ;
        argv [argc] = p ;
        if (* p == '"' || * p == '\'') {
            c = * p ++ ;
            argv [argc] ++ ;
        }
        for (;;) {
            char *del = strchr (p,   c ) ;
            char *esc = strchr (p, '\\') ;
            /* is there an escape within the current argument? */
            if (esc == 0 || (del != 0 && del < esc)) {
                p = del ; /* no escape */
                break ;
            }
            /* this escape character: remove it */
            memmove (esc, p = esc + 1, strlen (esc)); /* incl: '\0' */
        }
        if (p)
            * p ++ = '\0' ; /* set string terminator */
    }
    if (argc == 0) return ;

    DMALLOC_VERIFY (ct) ;

    /* check whether we got a reply socket */
    if (argv [0] [0] == '/') {
        ct->reply_socket = argv [0];  /* publish the reply socket path */
        if (-- argc == 0) {           /* nothing we can do, here */
            argi = 0 ;
            argv [0] = 0 ;
            goto help_command;
        }
        argi = 1 ;                    /* first request argument */
    }
    else {
        ct->reply_socket = 0 ;
        argi = 0 ;                    /* first request argument */
    }

    /* setup for output */
    ct->ioinx          = 0 ;
    ct->iobuf_reply_ok = 0 ;

    DMALLOC_VERIFY (ct) ;

    /* find handler by its registered name and run it */
    if ((n = varstr_list_index (ct->service, sizeof (ct->service [0].name),
                                ct->dim, sizeof (ct->service [0]),
                                argv [argi], 0)) >= 0) {
        
        struct _service *svc = ct->service + n ;

#           ifndef DISCARD_EXTENSIVE_LOGGING
        /* handle debuglog */
        if (svc->fn != ctrl_verbosity)
            delayed_debug_logger (ct);
#           endif

        /* call hander */
        (*svc->fn) (svc->use_ctrl ? ct : svc->fn_state, argc,
                    (const char**)argv + argi);

#       ifndef DISCARD_EXTENSIVE_LOGGING
        /* handle debuglog */
        delayed_debug_logger (ct);
#       endif

        /* send back collected output */
        goto reply_to_sender;
    }

    DMALLOC_VERIFY (ct) ;

    /* no handler: invoke the usage handler if we can send back data */
    if (ct->reply_socket == 0) return ;

    ctrl_reply (ct, (n == -1
                     ? "-ERR: unknown command \"%s\"\n"
                     : "-ERR: command prefix \"%s\" is not unique\n"),
                argv [argi]);
    
 help_command:

#   ifndef DISCARD_EXTENSIVE_LOGGING
    /* handle debuglog */
    delayed_debug_logger (ct);
#   endif

    usage_handler (ct, argc, (const char**)argv + argi);

 reply_to_sender:
    if (ct->iobuf_reply_ok == 0 || ct->reply_socket == 0)
        goto leave ; /* no reply expected */

    DMALLOC_VERIFY (ct) ;

    /* send the command buf to server */
    sun.sun_family = AF_UNIX ;
    strncpy (sun.sun_path, ct->reply_socket, UNIX_PATH_MAX);
    sun.sun_path [UNIX_PATH_MAX-1] = '\0' ;
    slen = (offsetof (struct sockaddr_un, sun_path) +
            strlen (sun.sun_path) + 1);

    if ((n = sendto (ct->cmd_fd, ct->iobuf, ct->ioinx, 0,
                     (struct sockaddr*)&sun, slen)) < 0) {
        ct->iobuf [ct->ioinx - 1] = '\0';
        logger (LOG_ERR, "command: Cannot reply to %s: %s",
                ct->reply_socket, strerror (errno));
    }

#   ifndef DISCARD_EXTENSIVE_LOGGING
    if (verbosity & CTRL_REPLY_LOG)
        logger (LOG_DEBUG, "command reply(%d): \"%s\"", n, ct->iobuf);
#   endif

leave:
    if (ct->on_leave.set_noisy) {
        *ct->custom.noisebits = ct->on_leave.noisy;
        ct->on_leave.set_noisy = 0;
    }

    DMALLOC_VERIFY (ct) ;
}


/* collect response data - returns the number chars left in the buffer */
int ctrl_reply (CTRL *ct, const char *format, ...) {

    int left = ct->iobuf_size  - ct->ioinx;
    struct stat buf ;

    DMALLOC_VERIFY (ct) ;

    if (ct->reply_socket == 0) {        /* no reply expected */
        errno = EPIPE ;                 /* Broken pipe */
        return -1 ;
    }
    if (stat (ct->reply_socket, &buf) != 0) {
        ct->reply_socket = 0 ;
        return -1 ;
    }
#   ifdef S_ISSOCK
    if (S_ISSOCK (buf.st_mode) == 0) {
        ct->reply_socket = 0 ;
        errno = ENOTSOCK ;              /* Socket operation on non-socket */
        return -1 ;
    }
#   endif
     
    if (format) {

        va_list ap;
        va_start (ap, format);

        if (left > 0) {
        
            int n = vsnprintf (ct->iobuf + ct->ioinx, /* append data here */
                               left,       /* not more than this many */
                               format,     /* format */
                               ap);        /* these arguments */

            if (n < 0)                     /* error occured */
                left = -1 ;
            else if (n > left) {           /* output was truncated */
                left      = -1 ;
                ct->ioinx = ct->iobuf_size ;
                errno     = ENOSPC ;       /* No space left on device */
            }
            else {
                left      -= n ;           /* bookkeeping ... */
                ct->ioinx += n ;
            }
        }
        else {                             /* flag overflow */
            left   = -1 ;
            errno  = E2BIG ;               /*  Argument list too long */
        }

        va_end (ap);

        ct->iobuf_reply_ok = 1 ;           /* remember that we got in, here */
    }

    DMALLOC_VERIFY (ct) ;
    return left ;
}


/* query uid/gid of the sender for the last request (if any) */
#ifndef DONT_USE_RECVMSG
uid_t ctrl_uid (CTRL *ct) {
    return ct->uid ;
}

gid_t ctrl_gid (CTRL *ct) {
    return ct->gid ;
}

pid_t ctrl_pid (CTRL *ct) {
    return ct->pid ;
}
#endif

/* example for some easy communication handlers */
void ctrl_usage (void *ctrl, int argc, const char **argv) {
    CTRL *ct = ctrl;

    if (ctrl_reply (ct, "+OK\n") >= 0)
        usage_handler (ct, argc, argv);
    if (ct->custom.msg.usage)
        ctrl_reply (ct, "%s", ct->custom.msg.usage);
}

void ctrl_usage_msg (CTRL *ct, const char *s) {
    if (ct->custom.msg.usage)
        free ((char*)ct->custom.msg.usage);
    ct->custom.msg.usage = (s && *s) ? zstrdup (s) : 0 ;
}


void ctrl_verbosity (void *ctrl, int argc, const char **argv) {
    CTRL *ct = ctrl;
    const char *me = argv [0], *level;
    unsigned old, noise, op ;

    /* check for a help line request */
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }

    if (ct->custom.noisebits == 0) {
        ctrl_reply (ct, "+ERR: not configured for verbosity management");
        return;
    }

    /* always keep the MSBit reset */
    old = *ct->custom.noisebits & ((unsigned)-1 >> 1);

    switch (argc) {
    case 1:
        goto show_it ;
    case 2:
        op = 0 ;
        level = argv [1] ;
        break ;
    case 3:
        op = argv [1] [0] ;
        if ((op == '+' || op == '-' || op == '=') && argv [1] [1] == '\0') {
            level = argv [2] ;
            break ;
        }
        /* goto usage; */
    default:
        goto usage;
    }
    
    if (level [0] == '+' || level [0] == '-' || level [0] == '=') {
        if (op) goto usage ;
        op = * level ++ ;
    }

    if (parse_unsigned (&noise, level, 0) == 0)
        goto usage ;

    switch (op) {
    case '+':
        *ct->custom.noisebits  |=  noise;
        break ;
    case '-':
        *ct->custom.noisebits  &= ~noise;
        break ;
    case '=':
        ct->on_leave.    noisy  =  noise;
        ct->on_leave.set_noisy  =  1    ;
        break ;
    default:
        *ct->custom.noisebits   =  noise;
    }

show_it:
    ctrl_reply (ct, "+OK: %#x", old) ;
    return ;

usage:
    ctrl_reply (ct, "-ERR: usage: ");
help:
    ctrl_reply (ct, "%s [[+-=]<verbosity>] -- "
                "shows/modifies/sets the noise level\n", me);
    if (ct->custom.msg.noisy)
        ctrl_reply (ct, "%s", ct->custom.msg.noisy);
}

void ctrl_verbosity_msg (CTRL *ct, const char *s) {
    if (ct->custom.msg.noisy)
        free ((char*)ct->custom.msg.noisy);
    ct->custom.msg.noisy = (s && *s) ? zstrdup (s) : 0 ;
}

void ctrl_resource (void *ctrl, int argc, const char **argv) {
    CTRL *ct = ctrl;
    const char *me = argv [0];

#   ifdef HAVE_GETRUSAGE
    struct rusage r ;
    int rusage_ok ;
#   endif

    FILE *statm_ok;
    int level, n, pagek, size, resident, share, code, data, lib, dt, none;
    
    /* check for a help line request */
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }

    if (argc > 2) goto usage ;

    if (argc == 2) {
	static const char *tok [] = {"most", "all"};
	switch (varstr_list_index
		(tok, 0, DIM (tok), sizeof (tok [0]), argv [1], 0)) {
	case 0:
            level = 5 ;
	    break ;
	case 1:
            level = 9 ;
	    break ;
	default:
            goto usage ;
	}
    }
    else
        level = 0;
        
    ctrl_reply (ct, "+OK\n", 0) ;

    pagek = sysconf (_SC_PAGESIZE) / 1024 ;

#   ifdef HAVE_GETRUSAGE
    if (getrusage (RUSAGE_SELF, &r) < 0) {
        ctrl_reply (ct, "getrusage() returned error: %s", strerror (errno));
        rusage_ok = 0 ;
    }
    else
        rusage_ok = 1 ;
#   endif

    /* /proc/self/statm: size resident share code data lib dt */
#   define PROC_STATM "/proc/self/statm"
    if ((statm_ok = fopen (PROC_STATM, "r")) == 0) {
        ctrl_reply (ct, "Cannot open " PROC_STATM " for reading: %s",
                    strerror (errno));
    }
    else if ((n = fscanf (statm_ok,
                          "%d %d %d %d %d %d %d %d", &size, &resident,
                          &share, &code, &data, &lib, &dt, &none)) != 7) {
        char *info = n < 0
            ? strerror (errno)
            : (n > 7
               ? "more than 7 fields"
               : "less than 7 fields")
            ;
        ctrl_reply (ct, "Error parsing " PROC_STATM ": %s", info);
        fclose (statm_ok);
        statm_ok = 0 ;
    }
    else {
        fclose (statm_ok);

        size     *= pagek ;
        resident *= pagek ;
        share    *= pagek ;
        code     *= pagek ;
        data     *= pagek ;
        lib      *= pagek ;
        dt       *= pagek ;
    }
    
#   undef PROC_STATM

#   ifdef HAVE_GETRUSAGE
    if (rusage_ok) {
        ctrl_reply     (ct, "user time used:              % 9u.%02us\n",
                    r.ru_utime.tv_sec,
                        (r.ru_utime.tv_usec + 5000) / 10000);
        
        ctrl_reply     (ct, "system time used:            % 9u.%02us\n",
                        r.ru_stime.tv_sec,
                        (r.ru_stime.tv_usec + 5000) / 10000);
    }
#   endif

    if (statm_ok) {
        ctrl_reply     (ct, "currently used memory size:  % 9ukb\n", size);
        if (level > 4) {
            int ushare = size - share;
            ctrl_reply (ct, "shared memory:               % 9ukb\n", share);
            ctrl_reply (ct, "unshared memory:             % 9ukb\n", ushare);
            ctrl_reply (ct, "resident memory:             % 9ukb\n",
                        resident);
        }

        if (level > 8) {
            ctrl_reply (ct, "code size:                   % 9ukb\n", code);
            ctrl_reply (ct, "data/stack size:             % 9ukb\n", data);
            ctrl_reply (ct, "library data:                % 9ukb\n", lib);
            ctrl_reply (ct, "dirty memory:                % 9ukb\n", dt);
        }
    }
    if (level > 4) {
        ctrl_reply     (ct, "page size:                   % 9ukb\n", pagek);
        ctrl_reply     (ct, "process id:                  % 9u\n",getpid ());
    }

#   ifdef HAVE_GETRUSAGE
    if (rusage_ok) {
        if (level > 4) {
            ctrl_reply (ct, "page reclaims:               % 9u\n",
                        r.ru_minflt);
            ctrl_reply (ct, "page faults:                 % 9u\n",
                        r.ru_majflt);
            ctrl_reply (ct, "signals received:            % 9u\n",
                        r.ru_nsignals);
        }
        if (level > 8) {
            ctrl_reply (ct, "voluntary context switches:  % 9u\n",
                        r.ru_nvcsw);
            ctrl_reply (ct, "involuntary context switches:% 9u\n",
                        r.ru_nivcsw);
        }
    }
#   endif

    return ;
    
usage:
    ctrl_reply (ct, "-ERR: usage: ");
help:
    ctrl_reply (ct, "%s [most|all] -- "
                "shows system resource consumption\n", me);
}

/* ----------------------------------------------------------------------- *
 * Constructor/Destructor
 * ----------------------------------------------------------------------- */

CTRL *ctrl_init (gid_t            gid,
                 size_t    iobuf_size,
                 const char *sockpath,
                 unsigned  *Verbosity) {
    CTRL ct ;
    char *dup, *dir ;

#   ifndef DONT_USE_RECVMSG
    int one = 1 ;
#   endif

    XZERO (ct);

    if (sockpath == 0) {
        sockpath = CTRL_DEFAULT_REQUEST_PATH ;
        logger (LOG_WARNING, "No UNIX socket path given, using \"%s\""
                " instead", sockpath);
    }
    dup = zstrdup (sockpath) ;
    dir = dirname (dup) ; /* dirname modifies its argument */
    
    if ((ct.iobuf_size = iobuf_size) == 0)
        ct.iobuf_size = CTRL_DEFAULT_REPLY_SIZE;
    ct.iobuf = (char*) xmalloc (ct.iobuf_size);
    
    ct.dim = 0;
    mkdir (dir, 0700); /* may fail when existing, already */
    if (gid) {
        chown (dir, -1, gid);
        chmod (dir, 0750);
    }
    free (dup);
    
    if ((ct.cmd_fd = socket (PF_UNIX, SOCK_DGRAM, 0)) < 0) {
        logger (LOG_ERR, "Cannot create command socket: %s",
                strerror (errno));
        return 0;
    }

#   ifndef DONT_USE_RECVMSG
    /* want to receive sender credentials */
    if (setsockopt
        (ct.cmd_fd, SOL_SOCKET, SO_PASSCRED, &one, sizeof one) < 0) {
        logger (LOG_ERR, "Cannot receive credentials with this socket: %s",
                strerror (errno));
        return 0;
    }
#   endif

    ct.un.sun_family = AF_UNIX ;
    strncpy (ct.un.sun_path, sockpath, UNIX_PATH_MAX);
    ct.un.sun_path [UNIX_PATH_MAX-1] = '\0' ;
    unlink (ct.un.sun_path);

    /* The length of the address is the offset of the start of the filename,
       plus its length, plus one for the terminating null byte. */
    ct.length_un = (offsetof (struct sockaddr_un, sun_path) +
                    strlen (ct.un.sun_path) + 1);
    if (bind (ct.cmd_fd, (struct sockaddr*)&ct.un, ct.length_un) < 0) {
        logger (LOG_ERR, "Cannot bind command socket to %s: ", 
                ct.un.sun_path, strerror (errno));
        return 0;
    }

    if (gid && gid != -1) {
        chown (ct.un.sun_path, -1, gid);
        chmod (ct.un.sun_path, 0770);
    }

    if (Verbosity)
        ct.custom.noisebits = Verbosity ;

    return zmemdup (&ct);
}


void ctrl_shutdown (CTRL *ct) {
    unlink (ct->un.sun_path);
}


void ctrl_destroy (CTRL *ct) {
    close (ct->cmd_fd);
    unlink (ct->un.sun_path);
    free (ct->iobuf);
    if (ct->custom.msg.noisy)
        free ((char*)ct->custom.msg.noisy);
    if (ct->custom.msg.usage)
        free ((char*)ct->custom.msg.usage);
    free (ct);
}


int ctrl_fd (CTRL *ct) {
    return ct->cmd_fd ;
}


CTRL *ctrl_add (CTRL                                *ct,
                const char                        *name,
                void (*handler)(void*,int,const char**),
                void                     *handler_state) {

    if (ct && name && handler) {
        struct _service *svc;
        char *s ;
        int use_ctrl = (ct == handler_state) ;
        
        /* get space for an additional entry */
        ct = (CTRL*) zrealloc
            (ct, sizeof (CTRL) + ct->dim * sizeof (struct _service)) ;

        DMALLOC_VERIFY (ct) ;

        /* populate new entry */
        svc = ct->service + (ct->dim ++);
        
        strncpy (s = (char*)svc->name, name, sizeof (svc->name));
        s [sizeof (svc->name) - 1] = 0;
        svc->fn       = handler ;

        /* there is no fixup procedure needed when the CTRL ptr has
           changed due to a memory reallocation by zrealloc() */

        if ((svc->use_ctrl = use_ctrl) == 0)
            svc->fn_state = handler_state ;
    }

    return ct ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
#endif /* DISCARD_CMD_SOCKET */
