/*
 * nvtools lib - flash data base support tools
 *
 * Copyright (C) 2003-2006 Maxina GmbH - Jordan Hrycaj <jordan@maxina.de>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Ackn: Sources reused from SHAT - Source Hardware Address Translator
 *       Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 * $Id: prefix.c 2933 2006-07-28 07:26:07Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#include "hlst.h"
#include "util.h"
#include "nvalloc.h"
#include "logger.h"

#define __NVRAM_PREFIX_PRIVATE
#include <prefix.h>

#undef USE_DEBUG

#if defined (USE_DEBUG) && !defined (_DEBUG_BISECT)
#define _DEBUG_BISECT 1
#endif

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

PLST *create_plst (unsigned use8bit) {
    PLST *pfx = ZMALLOC (PLST);
    pfx->h = create_hlst (0, 0, 0);
    pfx->use8bit = use8bit ;
    return pfx ;
}

void destroy_plst (PLST *pfx) {
    destroy_hlst (pfx->h);
    free (pfx);
}

void flush_plst (PLST *pfx) {
    flush_hlst (pfx->h, 0, 0);
}


void **update_plst (PLST *pfx, const char *var, size_t len) {
    void **val = find_hlst (pfx->h, var, len) ;
    if (val == 0)
        val = make_hlst (pfx->h, var, len);
    return val ;
}

void remove_plst (PLST *pfx, const char *var, size_t len) {
    delete_hlst (pfx->h, var, len);
}

void **find_plst (PLST *pfx, const char *key, size_t key_len) {

    int min, inx, max ;
    void **val ;

    if (key == 0 || (key [0] == 0 && key_len == 0))
        return 0 ;
    
    if (key_len == 0)
        key_len = strlen (key);
    else if (pfx->use8bit == 0 && key [key_len - 1] == '\0')
        key_len -- ;
            
    /* sort the list */
    sort_hlst (pfx->h) ;

    /* initalise bisection method ... */
    min =                            0 ; /* interval start */
    max = query_hlst_size (pfx->h) - 1 ; /* interval end */
    inx =                           -1 ; /* index of last seen prefix */
    val =                            0 ; /* value ptr of last seen prefix */

#   ifdef _DEBUG_BISECT
    logger (LOG_DEBUG, "Find a prefix for %s from:", key);
    { int n; for (n = 0; n <= max; n ++) {
        void **v = inx_hlst (pfx->h, n) ;
        logger (LOG_DEBUG, "%5d - %s", n, query_key_hlst (v));
    }}
#   endif
    
    /* find the greatest entry index with an entry smaller equal "key" */
    for (;;) {
        int CMP;

        /* split into intervals */
        int INX = (min + max) >> 1 ;

        /* read least value from the interval with the greater entries */
        void **VAL = inx_hlst (pfx->h, INX) ;
        char  *VAR = query_key_hlst    (VAL);
        int    LEN = query_keylen_hlst (VAL);

        /* use '\0' terminated strings unless otherwise specified */
        if (pfx->use8bit == 0 && LEN && VAR [LEN - 1] == '\0') LEN -- ;

#       ifdef _DEBUG_BISECT
        logger (LOG_DEBUG, "Bisect: [%d, %d[%s], %d]", min, INX, VAR, max);
#       endif

        if (key_len < LEN) {
            CMP = memcmp (VAR, key, key_len) ;
        } else if ((CMP = memcmp (VAR, key, LEN)) == 0) {
#           ifdef _DEBUG_BISECT
            logger (LOG_DEBUG, "Bisect: [%s] prefixes \"%s\" "
                    "=> hold the prefix", VAR, key);
#           endif
            inx = INX ; /* this one has been seen, for sure */
            val = VAL ;
            goto use_right_interval ;
        }
                 
        /* var < key */
        if (CMP < 0) {
        use_right_interval:
            /* use the interval with the greater entries */
            if (max == INX) { /* min == max */
#               ifdef _DEBUG_BISECT
                logger (LOG_DEBUG, "Bisect: [%s] < \"%s\" and "
                        "max = inx = %d = min => stop", VAR, key, INX);
#               endif
                break ;
            }
            /* min < inx < max && min+2 == max => inx == min + 1 */
            min = (min == INX || min + 2 == max) ? max : INX ;
            
            if (LEN && VAR [0] == key [0]) {
                inx = INX ; /* remember the last var(inx) < key */
#               ifdef _DEBUG_BISECT
                logger (LOG_DEBUG,"Bisect: [%s] < \"%s\" => min := %d, "
                        "hold inx = %d", VAR, key, min, inx);
#               endif
            }
#           ifdef _DEBUG_BISECT
            else
                logger (LOG_DEBUG, "Bisect: [%s] < \"%s\" => min := %d",
                        VAR, key, min);
#           endif
        }

        /* key < var */
        else {
            /* use the interval with the smaller entries */
            if (min == INX) { /* min == max or min == max + 1 */
#               ifdef _DEBUG_BISECT
                logger (LOG_DEBUG, "Bisect: \"%s\" < [%s] and min = inx "
                        "= %d = (max or max + 1) => stop", key, VAR, INX);
#               endif
                break ;
            }
            max = (max == INX) ? min : INX ;
#           ifdef _DEBUG_BISECT
            logger (LOG_DEBUG,
                    "Bisect: \"%s\" < [%s] => max := %d", key, VAR, max);
#           endif
        }
    }

    /* check whether we have seen a prefix, already */
    if (val) {
#       ifdef _DEBUG_BISECT
        logger (LOG_DEBUG, "Bisect returns [%s] prefixes \"%s\"",
                query_key_hlst (val), key);
#       endif
        return val ;
    }

    /* search upwards for the longest prefix, this means we accept the
       first variable that is a prefix */
    while (inx >= 0) {
        void **val = inx_hlst (pfx->h, inx) ;
        char  *var = query_key_hlst    (val);
        int    len = query_keylen_hlst (val);

#       ifdef _DEBUG_BISECT
        logger (LOG_DEBUG, "Linear search compares [%s] ~ \"%s\","
                " inx = %d", var, key, inx);
#       endif

        /* stop if there are no possible prefixes, anymore */
        if (len == 0 || var [0] != key [0]) break ;
        
        /* use '\0' terminated strings unless otherwise specified */
        if (pfx->use8bit == 0 && len && var [len - 1] == '\0') len -- ;

        /* check for a prefix */
        if (key_len >= len  && memcmp (var, key, len) == 0) {
#           ifdef _DEBUG_BISECT
            logger (LOG_DEBUG, "Linear search returns [%s] prefixes \"%s\"",
                    var, key);
#           endif
            return val ;
        }

        inx -- ;
    }
        
#   ifdef _DEBUG_BISECT
    logger (LOG_DEBUG, "No prefix found for \"%s\"", key);
#   endif
    return 0 ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
