/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "util.h"
#include "nvalloc.h"
#include "settings.h"
#include "logger.h"
#include "parser.h"
#include "buffer.h"

#define __MHTTPD_LIST_PRIVATE
#include "list.h"

/* ----------------------------------------------------------------------- *
 * private
 * ----------------------------------------------------------------------- */

struct _list_print_desc {
    void (*prt) (void*, const char*, ...);
    void  *out ;
} ;


static
int list_print_cb (void   *state,
                   void    *data,
                   char     *key,
                   unsigned klen) {
    int n ;
    struct _list_print_desc *desc = state ;
    ARGV                      *av = data ;
    const char                *mt ;
    const char               *ena ;

    if ((mt = mimetype_name (av->mt)) == 0)
        mt = "" ;
    if ((ena = av->nvena) == 0)
	ena = "" ;
 
    (*desc->prt) (desc->out, "%s: %s(%s,%d<%s>) =>",
		  key, av->so, mt, av->pa, ena);
    for (n = 0; n < av->argc; n ++)
        (*desc->prt)
            (desc->out, " %s", av->argv [n] ? av->argv [n] : "(null)");
    (*desc->prt) (desc->out, "\n");

    return 0;
}


static
void argv_destroy (void   *state,
                   void    *item,
                   char     *key,
                   unsigned klen) {
    int n ;
    ARGV *av = item ;

    if (av == 0)
        return ;
    
    if (av->so)
        free ((char*)av->so);

    for (n = 0; n < av->argc; n++)
        if (av->argv [n])
            free ((char*)av->argv [n]);

    free ((char*)av);
}


static
int nvram_ok (NVCL *nv, const char *s) {

    if (s != 0) {
        const char *v, *w ;
        int ok ;
        IOBUF b;
#       ifndef DISCARD_EXTENSIVE_LOGGING
        int noisy ;
#       endif

        XZERO (b);
        data_append (&b, LIST_ACCESS_BASE, 0);
        data_append (&b, s, 0);
                     
#       ifndef DISCARD_EXTENSIVE_LOGGING
        if ((noisy = nvpacl_verbosity (nv)) == -1)
            return 0 ;
#       endif

        nvcl_putcmd (nv, "set");
        nvcl_putavpair (nv, b.data, 0, 0);

        ok = (nvcl_rply (nv, LIST_NVRAM_TIMEOUT) > 0 &&
              nvcl_getavpair (nv, &v, &w, 0, 0) != 0) ;

        kill_space (&b);

#       ifndef DISCARD_EXTENSIVE_LOGGING
        nvrestore_verbosity (nv, noisy);
#       endif

        if (ok != 0)
            switch (tolower (w [0])) {
            case 't': /* true */
            case 'y': /* yes */
            case 'e': /* enabled */
            case '1':
                return 1;
            case 'f': /* false */
            case 'n': /* no */
            case 'd': /* disabled */
            case '0':
                return 0;
            }
    }

    return -1 ;
}


/* ----------------------------------------------------------------------- *
 * Public: simple in-memory list manager
 * ----------------------------------------------------------------------- */

LIST *list_open (NVCL *nvram) {
    LIST *l ;

    if (nvram == 0) {
        logger (LOG_ERR, "Initializaion error: NVRAM connector missing");
        return 0 ;
    }

    l = ZMALLOC (LIST);
    l->table = create_hlst (100, argv_destroy, 0);
    l->nvram = nvram ;
    return l;
}



void list_flush (LIST *l) {
    flush_hlst (l->table, 0, 0);
}



void list_close (LIST *l) {
    destroy_hlst (l->table);
    free (l);
}


const char *list_get_so (LIST *l, const char *uri) {
    ARGV **Av = (ARGV**) find_hlst (l->table, uri, 0);
    return Av ? (*Av)->so : 0 ;
}


unsigned list_get_argv (LIST            *l,
                        const char    *uri,
                        PAGE_ACT       *Pa,
                        const char   **Ena,
                        const char    **So,
                        MIMETYPE       *Mt,
                        int          *Argc,
                        const char ***Argv) {
    int check_access = 1;
    ARGV **Av, *av ;

    if ((Av = (ARGV**) find_hlst (l->table, uri, 0)) == 0)
        return 0;

    av = *Av ;

    if (Pa)  { *Pa   = av->pa    ; check_access = 0; }
    if (Ena) { *Ena  = av->nvena ; check_access = 0; }

    if (check_access)
        switch (av->pa) {
        case paPresent:
            break ;
        case paShowNotByDefault:
            if (nvram_ok (l->nvram, av->nvena) != 0) 
                return 0 ;
            break ;
        case paShowByDefault:
            if (nvram_ok (l->nvram, av->nvena) == 0) 
                return 0 ;
            break ;
        case paHideNotByDefault:
            if (nvram_ok (l->nvram, av->nvena) != 1)
                break ;
            return 0 ;
        case paHideByDefault:
            if (nvram_ok (l->nvram, av->nvena) == 1)
                break ;
            /* return 0 ; */
        default:
            return 0;
        }
       
    if (So)    *So   = av->so    ;
    if (Mt)    *Mt   = av->mt    ;
    if (Argc)  *Argc = av->argc  ;
    if (Argv)  *Argv = av->argv  ;

    return 1;
}



void list_set_argv (LIST           *l,
                    PAGE_ACT       pa,
                    const char *nvena,
                    const char    *so,
                    MIMETYPE       mt,
                    int          argc,
                    const char **argv) {
    int n ;
    ARGV **Av ;

    ARGV *av = zmalloc (sizeof (ARGV) + argc * sizeof (char *));

    if (so)
        av->so = zstrdup (so);

    av->argc  =  argc;
    av->mt    =    mt;
    av->pa    =    pa;
    av->nvena = nvena;
    
    for (n = 0; n < argc; n ++)
        if (argv [n])
            av->argv [n] = zstrdup (argv [n]);
    
    /* add to hash table */
    if ((Av = (ARGV**) find_hlst (l->table, argv [0], 0)) == 0)
        Av = (ARGV**) make_hlst (l->table, argv [0], 0);
    else if (*Av)
        free (*Av);

    *Av = av ;
}



unsigned list_del_argv (LIST         *l,
                    const char *uri) {
    return delete_hlst (l->table, uri, 0) != -1 ;
}



void list_print (LIST *l, void(*prt)(void*,const char*,...), void *out) {

    struct _list_print_desc desc ;

    desc.prt = prt ;
    desc.out = out ;
    
    for_hlst_do (l->table, list_print_cb, &desc);
}

unsigned list_used (LIST *l) {
    return query_hlst_size (l->table);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
