/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include <stdarg.h>
#include <string.h>

#include "http-plugin.h" /* make_path () */
#include "settings.h"
#include "ctrl.h" /* used in macro from settings.h */
#include "varstr.h"
#include "nvalloc.h"
#include "logger.h"

#undef WITH_DMALLOC

#define __PARSER_PRIVATE
#include "parser.h"

/*
 * Syntax:
 *
 *
 * Variable substitution (value nesting supported):
 *
 *  general syntax:
 *
 *     $<varname>
 *     ${<varname>}
 *     ${<varname><op><value>}
 *
 *  in particular:
 *
 *     $<varname>            -- substitute variable
 *     ${<varname>}          -- substitute variable
 *     ${<varname>::<value>} -- set value (do not substitute)
 *     ${<varname>:=<value>} -- set and substitute value if variable is empty
 *     ${<varname>:*<value>} -- set and substitute value if var is not empty
 *     ${<varname>:-<value>} -- substitute value if variable is empty
 *     ${<varname>:+<value>} -- substitute value if variable is not empty
 *
 *  where value might be a general expression containing variables and
 *  directives itself
 *
 *
 * Directives (argument nesting supported):
 *
 *  general syntax:
 *
 *     $[<directive>]
 *     $[<directive> <argument>]
 *
 *  in particular:
 *
 *     $[include <text ...>]
 *     $[if <text ...>] ... $[fi]
 *     $[if <text ...>] ... $[else] ... $[fi]
 *     $[eval <text ...>]             -- re-evaluate argument
 *     $[# ...]                       -- comment
 *     $[off]                         -- no output
 *     $[on]                          -- enable output
 *     $[emptyok]                     -- keep empty variables
 *     $[noempty]                     -- remove variables when becoming empty
 *     $[load [<prefix>]]             -- load nvram vars and prefix them
 *     $[store [<pefix>]]             -- select by prefix and store back
 *     $[sort [<prefix>]] ... $[tros] -- do for all prefixed variables
 *     $[index <varname>]             -- assign loop variable
 *     $[value <varname>]             -- assign value of the loop variable
 *     $[count <varname>]             -- assign iteration count
 */

/* ----------------------------------------------------------------------- *
 * Private variables and definitions
 * ----------------------------------------------------------------------- */

#ifdef WITH_DMALLOC
typedef struct {
    PARSER        *g ;
    const char *file ;
    int         line ;
} __avdump_rec ;

static unsigned long __mark, __mark2;

static
int __avdump_print (void *state, void *val, char *key, unsigned keylen) {
    __avdump_rec *a = state ;
    dmalloc_verify (val);
    logger (LOG_DEBUG, "AVPRINT at %s(%d): %s=%s.",
            a->file, a->line, key, val ? val : "(null)");
    return 0;
}
static
void __avdump (PARSER *g, const char *file, int line) {
    __avdump_rec a = {g, file, line};
    dmalloc_verify (g->avp);
    for_hlst_do (g->avp, __avdump_print, &a); 
    dmalloc_log_stats ();
}
#define avdump(g) __avdump (g, __FILE__, __LINE__)
#define DMALLOC_VERIFY(g) {\
            P_STATE *q = g->state ;                        \
            dmalloc_message ("*** %s(%d): dmalloc LIST check point.\n", \
                           __FILE__, __LINE__);            \
            dmalloc_verify (g);                            \
            while (q) { dmalloc_verify (q); q = q->next; } \
            dmalloc_log_stats ();                          \
        }
#define DMALLOC_MESSAGE(g,fmt, ...) {\
                 dbglog (g, fmt, ##__VA_ARGS__);                \
                 dmalloc_message ("Message[" __FILE__ "="         \
                                  to_str (__LINE__) "] "        \
                                  fmt, ##__VA_ARGS__) ;         \
                 DMALLOC_VERIFY(g) ;                            \
        }
#define DMALLOC_MARK2() { __mark2 = dmalloc_mark (); }
#define DMALLOC_MARK() { __mark = dmalloc_mark (); }

#define DMALLOC_HUNT2() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark2,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }
#define DMALLOC_HUNT() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }
#else
#define avdump(g)                 /* n/a */
#define DMALLOC_MARK()            /* n/a */
#define DMALLOC_HUNT()            /* n/a */
#define DMALLOC_MARK2()            /* n/a */
#define DMALLOC_HUNT2()            /* n/a */
#define DMALLOC_VERIFY(g)         /* n/a */
#define DMALLOC_MESSAGE(fmt, ...) /* n/a */
#endif

/* stringify number */
#define __str(x)              # x
#define to_str(x)             __str (x)

/* dimension of static array */
#define DIM(x)                (sizeof (x) / sizeof ((x) [0]))

static const char clause_missing []= "%s clause found but %s is missing"    ;
static const char weird_elif     []= "directive(internal): Weird elif state";
static const char no_input_type  []= "%s(internal): unknown input type %d"  ;
static const char no_output_type []= "%s(internal): unknown output type %d" ;


#define command_maxlen   sizeof ("include")
#define command_list(x)  \
    x (INCLUDE, include) \
    x      (IF, if)      \
    x    (ELIF, elif)    \
    x    (ELSE, else)    \
    x      (FI, fi)      \
    x    (EVAL, eval)    \
    x (COMMENT, comment) \
    x     (OFF, off)     \
    x      (ON, on)      \
    x (NOEMPTY, noempty) \
    x (EMPTYOK, emptyok) \
    x    (LOAD, load)    \
    x   (STORE, store)   \
    x    (SORT, sort)    \
    x    (TROS, tros)    \
    x   (INDEX, index)   \
    x   (VALUE, value)   \
    x   (COUNT, count)

#define                                   mk_cmd_token(x,y) , CMD_ ## x
typedef enum { CMD_NONE = 0 command_list (mk_cmd_token) } cmd_tok ;
#undef                                    mk_cmd_token

#define                                           mk_cmd_tab(x,y) , # y
static const char *cmd_tab [] = { 0 command_list (mk_cmd_tab) };
#undef                                            mk_cmd_tab

/* ----------------------------------------------------------------------- *
 * Private: logging
 * ----------------------------------------------------------------------- */

static
void vparserlog (PARSER         *g,
                 int         level,
                 const char   *pfx,
                 const char *fmtin,
                 va_list        ap) {

    const char *name ;
    int lineno = 0 ;
    char pfx2 [] = "parser[%d] %s(%d): %s: ";
    char pfx3 [] = "parser[]: %s: ";
    char *fmtout ;
    size_t fmt_len = (sizeof (pfx2)  +
                      strlen (pfx)   +
                      strlen (fmtin) + 20) ;
                      
    if (g->state) {
        switch (g->state->in.input_type) {
        case instream: 
        case instream_clone:
            if ((name = g->state->in.reader.file->filename) == 0)
                name = "<input>" ;
            lineno = g->state->in.reader.file->lineno ;
            break ;
        case instring:
        case instring_clone:
            name = "STREAM" ;
            break ;
        default:
            name = "null" ;
        }
        fmt_len += strlen (name);
    }
    
    fmtout = (char*) xmalloc (fmt_len) ;

    if (g->state)
        sprintf (fmtout, pfx2, g->state->nesting_level, name, lineno, pfx);
    else
        sprintf (fmtout, pfx3, pfx);

    vlogger (level, strcat (fmtout, fmtin), ap);
    free (fmtout);
}


#ifdef _DEBUG
void _errlog (PARSER *g, const char *err, const char *fmt, ...) {
    va_list ap;
    va_start (ap, fmt);
    vparserlog (g, LOG_ERR, err, fmt, ap);
    va_end (ap);
}
#define errlog(g, fmt, ...) _errlog \
      (g, "Error[" __FILE__ "=" to_str (__LINE__) "]", fmt, ##__VA_ARGS__)

void _dbglog (PARSER *g, const char *err, const char *fmt, ...) {
    va_list ap;
    va_start (ap, fmt);
    vparserlog (g, LOG_DEBUG, err, fmt, ap);
    va_end (ap);
}
#define dbglog(g, fmt, ...) _errlog \
      (g, "Debug[" __FILE__ "=" to_str (__LINE__) "]", fmt, ##__VA_ARGS__)
#else
static
void errlog (PARSER *g, const char *fmt, ...) {
    va_list ap;
    va_start (ap, fmt);
    vparserlog (g, LOG_ERR, "Error", fmt, ap);
    va_end (ap);
}
#define dbglog(g, fmt, ...) /* n/a */
#endif


#if 0
static
void warnlog (PARSER *g, const char *fmt, ...) {
    va_list ap;
    va_start (ap, fmt);
    vparserlog (g, LOG_WARNING, "Warning", fmt, ap);
    va_end (ap);
}
#endif

#ifndef DISCARD_EXTENSIVE_LOGGING
/* get rid of those never ending nvram variable listings */
static
int set_nvram_verbosity (NVCL *nvram, char mode, unsigned bits) {
    unsigned n ;
    const char *v ;
    char buf [5 + sizeof (verbosity) * 2] ;

    sprintf (buf, "%c %#x", mode, bits);

    nvcl_putcmd (nvram, "noisy");
    nvcl_putavpair (nvram, buf, 0, 0);

    if (nvcl_rply (nvram, 0) < 1) {
        logger (LOG_ERR, "NVRAM service communication error") ;
        return -1;
    }

    if (nvcl_getavpair (nvram, &v, 0, 0, 0) == 0) {
        logger (LOG_WARNING, "No \"noisy\" response from NVRAM - cannot "
                 "retrieve current state");
        return -2 ;
    }
    
    if (parse_unsigned (&n, v, 0) == 0) {
        logger (LOG_WARNING, "Weird \"noisy\" response: \"%s\" - cannot "
                "retrieve current state", v);
        return -2 ;
    }

    return n ;
}

static
int nvram_verbosity (NVCL     *nvram,
                     unsigned enable,
                     char       mode,
                     unsigned   bits) {
    return ((verbosity & enable) != 0 || nvcl_http_ok (nvram) != 0)
        ? -3
        : set_nvram_verbosity (nvram, mode, bits)
        ;
}
#endif

/* ----------------------------------------------------------------------- *
 * Private: io stack management
 * ----------------------------------------------------------------------- */

static
void pop_state (PARSER *g) {

    P_STATE *pold ;

    if ((pold = g->state) != 0) {

        g->state = pold->next ;

        switch (pold->in.input_type) {
        case instream:
            free ((char*)pold->in.reader.file->filename) ;
            fclose      (pold->in.reader.file->fp) ;
            free        (pold->in.reader.file);
            break ;
        case instring:
            kill_space (pold->in.reader.iobuf->buf);
            free       (pold->in.reader.iobuf->buf);
            free       (pold->in.reader.iobuf);
            break ;
        case instream_clone:
        case instring_clone:
            break ;
        default:
            errlog (g, no_input_type, "pop_state", pold->in.input_type);
        }
        switch (pold->out.output_type) {
        case outstream:
            fclose     (pold->out.writer.file->fp);
            free       (pold->out.writer.file);
            break ;
        case outstring:
            kill_space (pold->out.writer.iobuf->buf);
            free       (pold->out.writer.iobuf->buf);
            free       (pold->out.writer.iobuf);
            break ;
        case outstring_clone:
        case outstream_clone:
        case nullout:
            break ;
        default:
            errlog (g, no_output_type, "pop_state", pold->out.output_type);
        }
        free (pold);
    }

    // DMALLOC_VERIFY (g);
}


static
void pop_state_level (PARSER *g) {

    const char nesting_error [] = "%s: nesting error" ;
    
    if (g->state->block.if_true_pending  != 0 ||
        g->state->block.if_false_pending != 0 ||
        g->state->block.else_pending     != 0) {
        errlog (g, nesting_error, "if/else");
        return;
    }

    if (g->state->block.sort_block_level > 0) {
        errlog (g, nesting_error, "sort");
        return;
    }

    if (g->state->next == 0)
        return ;

    pop_state (g);
}

static
void push_state_dup (PARSER *g) {

    P_STATE *pnew = ZMALLOC (P_STATE);

    if (g->state) {

        if (g->state->nesting_level > 10) {
            errlog (g, "PANIC: context stack overflow at nesting level: %d",
                    g->state->nesting_level);
            while (g->state->nesting_level > 1)
                pop_state (g) ;
            return ;
        }

        *pnew      = *g->state ;
        pnew->next = g->state ;

        switch (pnew->in.input_type) {
        case instream:
            pnew->in.input_type = instream_clone;
            break ;
        case instring:
            pnew->in.input_type = instring_clone ;
            break ;
        case instream_clone:
        case instring_clone:
            break ;
        default:
            errlog (g, no_input_type,
                    "push_state_dup", pnew->in.input_type);
        }
        switch (pnew->out.output_type) {
        case outstream:
            pnew->out.output_type = outstream_clone ;
            break ;
        case outstring:
            pnew->out.output_type = outstring_clone ;
            break ;
        case outstream_clone:
        case outstring_clone:
        case nullout:
            break ;
        default:
            errlog (g, no_output_type,
                    "push_state_dup", pnew->out.output_type) ;
        }
        XZERO (pnew->block);
    }

    g->state = pnew ;
    g->state->nesting_level ++ ;

    // DMALLOC_VERIFY (g);
}


static
unsigned push_file_input (PARSER        *g,
                          const char *file,
                          const char *info) {
    FILE *fp;
    IOBUF iobuf ;
    const char *s ;

    if (info == 0)
        info = "reading" ;

    XZERO (iobuf);
    if ((fp = fopen (s = make_path (g->docf, &iobuf, file), "r")) == 0) {
        errlog
            (g, "Cannot open \"%s\" for %s: %s", s, info, strerror (errno));
        kill_space (&iobuf);
        return 0;
    }
    kill_space (&iobuf);

    push_state_dup (g) ;
    g->state->in.input_type            = instream ;
    g->state->in.reader.file           = ZMALLOC (XIOFILE);
    g->state->in.reader.file->fp       = fp ;
    g->state->in.reader.file->lineno   = 1 ;
    g->state->in.reader.file->filename = (s = strrchr (file, '/'))
        ? zstrdup (s + 1)
        : zstrdup (file)
        ;
        
    // DMALLOC_VERIFY (g);
    return 1;
}


/* ----------------------------------------------------------------------- *
 * Private: read/input methods
 * ----------------------------------------------------------------------- */

static
int next_char (PARSER *g, IOBUF *iob) {
    int c ;
    P_STATE *s = g->state ;

    switch (s->in.input_type) {
    case instream:
    case instream_clone:
        if (s->in.stop || (c = getc (s->in.reader.file->fp)) == EOF)
            break ;
        if (c == '\n')
            g->state->in.reader.file->lineno ++ ;
        if (iob)
            char_append (iob, c);
        return c ;

    case instring:
    case instring_clone:
        if (s->in.stop ||
            s->in.reader.iobuf->pos >= s->in.reader.iobuf->buf->top)
            break ;
        c = s->in.reader.iobuf->buf->data [s->in.reader.iobuf->pos ++];
        if (iob)
            char_append (iob, c);
        return c ;

    default:
        errlog (g, no_input_type, "next_char", s->in.input_type);
    }

    s->in.stop = 1 ;
    return EOF ;
}


static
int peep_char (PARSER *g) {
    int c ;
    P_STATE *s = g->state ;

    switch (s->in.input_type) {
    case instream:
    case instream_clone:
        if (s->in.stop || (c = getc (s->in.reader.file->fp)) == EOF)
            break ;
        ungetc (c, s->in.reader.file->fp) ;
        return c ;

    case instring:
    case instring_clone:
        if (s->in.stop ||
            s->in.reader.iobuf->pos >= s->in.reader.iobuf->buf->top)
            break ;
        return s->in.reader.iobuf->buf->data [s->in.reader.iobuf->pos];

    default:
        errlog (g, no_input_type, "peep_char", s->in.input_type);
    }

    // DMALLOC_VERIFY (g);

    return EOF ;
}


static
void stop_input (PARSER *g) {
    g->state->in.stop = 1 ;
}

/* ----------------------------------------------------------------------- *
 * Private: write/output methods
 * ----------------------------------------------------------------------- */

static
void put_char (PARSER *g, int c) {

    P_STATE *s = g->state ;

    if (s->out.suspend_output == 0)

        switch (s->out.output_type) {
        case outstream:
        case outstream_clone:
            putc (c, s->out.writer.file->fp);
            break ;

        case outstring:
        case outstring_clone:
            char_append (s->out.writer.iobuf->buf, c) ;
            /* break ; */

        case nullout:
            break ;

        default:
            errlog (g, no_output_type, "put_char", s->out.output_type);
        }
}


static
void put_string (PARSER *g, const char *txt) {
    
    P_STATE *s = g->state ;

    if (s->out.suspend_output == 0)

        switch (s->out.output_type) {
        case outstream:
        case outstream_clone:
            fputs (txt, s->out.writer.file->fp);
            break ;
        case outstring:
        case outstring_clone:
            data_append (s->out.writer.iobuf->buf, txt, 0);
            /* break ; */
        case nullout:
            break ;
        default:
            errlog (g, no_output_type, "put_char", s->out.output_type);
        }
}

#if 0
static
void put_printf (PARSER *g, const char *format, ...) {

    P_STATE *s = g->state ;

    if (s->out.suspend_output == 0) {
        va_list   ap;
        va_start (ap, format);
    
        switch (s->out.output_type) {
        case outstream:
        case outstream_clone:
            vfprintf (s->out.writer.file->fp, format, ap);
            break ;
        case outstring:
        case outstring_clone:
            data_vprintf (s->out.writer.iobuf->buf, format, ap);
            /* break ; */
        case nullout:
            break ;
        default:
            errlog (g, "put_printf(internal): unknown output type %d",
                    s->out.output_type) ;
        }
        va_end (ap);
    }
}
#endif

/* ----------------------------------------------------------------------- *
 * Private: variables
 * ----------------------------------------------------------------------- */

static
unsigned load_from_nvram (PARSER *g, const char *pfx) {

    int slot, used, n, total ;
    const char *var, *val, **Val ;
    size_t len = zstrlen (pfx);

#   ifndef DISCARD_EXTENSIVE_LOGGING
    int noisy ;
    if ((noisy = nvload_verbosity (g->nvram)) == -1)
        return 0 ;
#   endif
    
    nvcl_putcmd (g->nvram, "slots");
    if (nvcl_rply (g->nvram, 0) < 1)
        goto com_error;

    if (nvcl_getavpair (g->nvram, 0, &val, 0, 0) == 0) {
        errlog (g, "No NVRAM slots used") ;
        return 0 ;
    }
    used = atoi (val);

    if (nvcl_getavpair (g->nvram, 0, &val, 0, 0) == 0) {
        errlog (g, "No NVRAM slots available") ;
        return 0 ;
    }
    total = atoi (val);

    if (pfx == 0)
        pfx = "" ;
    
    for (slot = 1, n = used; slot <= total && n >= 0; slot ++) {
        char *p, *q, t [20];

        // DMALLOC_VERIFY (g);
        
        sprintf (t, "%u", slot);
        nvcl_putcmd    (g->nvram, "get");
        nvcl_putavpair (g->nvram, t, 0, 0);

        if (nvcl_rply (g->nvram, 0) < 1)
            continue ;

        while (nvcl_getavpair (g->nvram, &var, &val, 0, 0) != 0) {

            n -- ; /* one less */

            if (val)
                q = zstrdup (val) ;
            else if (g->state->delete_on_empty)
                continue ;
            else
                q = zmalloc (1) ; /* strdup ("") */

            /* prepend by the prefix */
            p = strcat (strcpy ((char*) xmalloc
                                (strlen (var) + len + 1), pfx), var);
            
            /* get the variable contents */
            if ((Val = (const char **) find_hlst (g->avp, p, 0)) == 0)
                Val = (const char **) make_hlst (g->avp, p, 0) ;
            else if (*Val)
                free ((char*)*Val) ;

            /* set the variable contents */
            *Val = q ;

            /* clean up */
            free (p);
        }
    }

#   ifndef DISCARD_EXTENSIVE_LOGGING
    nvrestore_verbosity (g->nvram, noisy);
#   endif

    avdump (g) ;

    // DMALLOC_VERIFY (g);

    return used ;

    /*NOTREACED*/

com_error:
    errlog (g, "NVRAM service communication error") ;
    return 0 ;
}


static
unsigned store_into_nvram (PARSER *g, const char *pfx) {

    HSRCH *lookup = open_hlst_search (g->avp) ;
    size_t    len = zstrlen (pfx);
    int    errors = 0 ;
    int     total = 0 ;

    const char **Val ;

#   ifndef DISCARD_EXTENSIVE_LOGGING
    int noisy ;
    if ((noisy = nvstore_verbosity (g->nvram)) == -1)
        return 0 ;
#   endif

    while ((Val = (const char**) next_hlst_search (lookup)) != 0) {

        const char *var = query_key_hlst ((void **)Val) ;

        // DMALLOC_VERIFY (g);

        if (len && (len > strlen (var) || strncmp (pfx, var, len)) != 0)
            /* prefix does not match */
            continue ;

        /* strip off the prefix */
        //var += len ;

        /* update nvram */
        if (*Val == 0 || **Val == '\0') {
            nvcl_putcmd    (g->nvram, "del");
            nvcl_putavpair (g->nvram, var, 0, 0);
        }
        else {
            nvcl_putcmd    (g->nvram, "set");
            nvcl_putavpair (g->nvram, var, *Val, 0);
        }

        if (nvcl_rply (g->nvram, 0) <= 0) {
            errlog (g, "Cannot set/update NVRAM(%s=%s): %s",
                    var, *Val ? *Val : "(null)", nvcl_rply_error (g->nvram));
            errors ++ ;
        }
        else
            total ++ ;
    }

    close_hlst_search (lookup) ;

#   ifndef DISCARD_EXTENSIVE_LOGGING
    nvrestore_verbosity (g->nvram, noisy);
#   endif

    // DMALLOC_VERIFY (g);

    return errors ? - total - 1 : total ;
}

/* ----------------------------------------------------------------------- *
 * Private: parser entities
 * ----------------------------------------------------------------------- */

static
int run_parser_stop (PARSER *g, int eot);

static
const char **prefixed_entries (PARSER *g, const char *pfx) {
    
    const char **lst ;
    int n, inx = 0, len = pfx ? strlen (pfx) : 0;
        
    /* sort the list */
    sort_hlst (g->avp) ;

    if (len) {

        int min =                            0 ;
        int max = query_hlst_size (g->avp) - 1 ;
    
        /* find the least entry starting with the prefix */
        for (;;) {
            const char **Val ;
        
            if ((Val = (const char **)
                 inx_hlst (g->avp, inx = (max + min) >> 1)) == 0)
                
                n = -1 ;

            else if ((n = strncmp /* prefix should match => success */
                      (query_key_hlst ((void **)Val), pfx, len)) == 0 &&
                     len < query_keylen_hlst ((void **)Val)) {
        
                /* check whether this is the first matching entry */
                if (inx == 0)
                    break ;

                if ((Val = (const char **) inx_hlst (g->avp, inx - 1)) == 0)

                    n = -1 ;

                else if (strncmp /* should not match => success */
                         (query_key_hlst ((void **)Val), pfx, len) != 0)
                    break ;
            }

            if (n >= 0 || min + 1 == max) /* && inx == min */
                max = inx ;
            else
                min = inx ;

            if (min >= max) {
                if (n == 0)
                    break ;
                return 0 ;
            }
        }
        
    } /* pfx != 0 */

    /* make a var list */
    lst = (const char **) xmalloc (sizeof (char**));

    for (n = 0;;n ++) {
        const char *var, **Val = (const char **) inx_hlst (g->avp, inx ++) ;

        if (Val == 0)
            break ;

        var = query_key_hlst ((void **)Val) ;
        if (len && strncmp (var, pfx, len) != 0)
            break ;

        lst = (const char **) zrealloc (lst, (n + 2) * sizeof (char**));
        lst [n] = zstrdup (var) ;
    }
    
    lst [n] = 0 ;
    // DMALLOC_VERIFY (g);
    return lst ;
}


static
int arg_expression (PARSER *g, const char *expr) {

    const char *p1, *p2;

    if ((p1 = expr) == 0)
        return 0 ;

    /* compare strings */
    if((p2 = strrchr (p1, '=')) != 0){

		while (* ++ p2 == ' ')
			;

		while (* p1 == ' ')
			p1 ++;
        
		while (p1 [0] == p2 [0] && p1 [0] != '=') {
			p2 ++;
			p1 ++;
		}

		while (p2 [0] == ' ')
			p2 ++;

		while (p1 [0] == ' ')
			p1 ++;

		return p1 [0] == '=' && p2 [0] == '\0' ;
    }

    /* test for non-zero number */
    if (isdigit (p1 [0]) || p1 [0] == '+' || p1 [0] == '-') {
        unsigned long n = strtol (p1, (char**)&p2, 0) ;

        switch (n) {
        case LONG_MAX:
        case LONG_MIN: /* either a string or a non-zero number */
            return 1 ;
        case 0:
            if (errno == EINVAL) /* is a string  */
                return 1;
        }

        while (isspace (*p2))
            p2 ++ ;

        if (p2 [0] == '\0')
            return n != 0 ;

        /* was an expression rather than a number */
    }

    /* test for a non-empty string */
    return p1 [0] != '\0';

//    else
//	{
//		isTrue = expr && expr [0] && (expr [0] != '0' || expr [1] != '\0')
//            ? 1
//            : 0
//            ;
//	}
//
//    return isTrue;
}


static
cmd_tok command (PARSER *g, IOBUF *iob) {

    char buf [command_maxlen] ; /* space for the longest command */
    int n, c ;

    /* read the command into the buffer (terminates with a space) */
    for (n = 0;;) {
        if ((c = peep_char (g)) != EOF) {
            if (isspace (c) || c == ']')
                break ;
            buf [n] = next_char (g, iob);
            if (++ n < DIM (buf))
                continue ;
        }
        buf [n] = '\0' ;
        goto unknown_cmd ;
    }
    buf [n] = '\0' ;
    
    /* find the command number in the list (starting at index 1) */
    for (n = 1;;) {
        if (strcmp (cmd_tab [n], buf) == 0)
            return n ;
        if (DIM (cmd_tab) <= ++ n)
            goto unknown_cmd ;
    }

    /* NOTREACHED */

unknown_cmd:
    errlog (g, "Unknown directive: %s", buf) ;
    return 0 ;
}


static
const char *varname_alloc (PARSER *g, IOBUF *iob) {

    int c = peep_char (g) ;

    IOBUF  varname  ;
    XZERO (varname) ;

    /* first character should be plain alpha in strict mode */
    if (g->state->strict_varsyntax == 0 || isalpha (c) || c == '_')

        for (;;) {
            /* add to the variable name */
            char_append (&varname, c = next_char (g, iob));

            /* check next */
            c = peep_char (g) ;

            /* check for end-of-varname */
            if (c == '}' || c == ':' || c == EOF)
                break ;

            /* must probably be some sort of an alpha character */
            if (g->state->strict_varsyntax && isalnum (c) == 0 && c != '_')
                break ;
        }

    // DMALLOC_VERIFY (g);

    unlink_space (&varname);
    return varname.data ;
}


static
unsigned argument_alloc (PARSER        *g,
                         const char **arg,
                         IOBUF       *iob,
                         int         stop) {
    int c ;
    XIOBUF xiobuf ;
    IOBUF  argbuf ;
    
    /* will contain the result */
    XZERO (argbuf);
    XZERO (xiobuf);

    /* recursively parse the argument */
    push_state_dup (g) ;
    g->state->out.suspend_output    = 0 ;
    g->state->out.output_type       = outstring_clone ;
    g->state->out.writer.iobuf      = &xiobuf ;
    g->state->out.writer.iobuf->buf = &argbuf ;

    c = run_parser_stop (g, stop) ;
    if (argbuf.data)
        data_append (iob, argbuf.data, 0);
    char_append (iob, c);

    // DMALLOC_VERIFY (g);

    if (c == stop) {
        if ((*arg = (const char*) argbuf.data) == 0)
            *arg = zmalloc (1) ; /* strdup ("") */

        unlink_space (&argbuf);
        return 1;
    }
    
    /* should not happen - but anyway */
    if (c == EOF)
        errlog (g, "\"%c\" was expected (got end-of-file)", stop);
    else
        errlog (g, "\"%c\" was expected (got only %c)", stop, c);

    kill_space (&argbuf);

    // DMALLOC_VERIFY (g);

    return 0;
}


static
void simple_variable (PARSER *g) {

    const char *var, **Val ;
    IOBUF iobuf ;

    int flag = g->state->strict_varsyntax ;

    /* always keep track of what we have read */
    XZERO        (iobuf);
    char_append (&iobuf, '$');

    g->state->strict_varsyntax = 1 ;
    var = varname_alloc (g, &iobuf) ;
    g->state->strict_varsyntax = flag ;
    
    if (var == 0 || g->state->dont_expand)
        goto dump_and_terminate ;

    /* retrieve/substitute */
    if ((Val = (const char **) find_hlst (g->avp, var, 0)) != 0)
        put_string (g, *Val);
    goto terminate ;
    
    /* NOTREACHED */

dump_and_terminate:
    put_string (g, iobuf.data);

terminate:
    kill_space (&iobuf) ;
    if (var)
        free ((char*)var);

    // DMALLOC_VERIFY (g);
}


static
void complex_variable (PARSER *g) {

    int c, op ;
    const char *var = 0, **Val, *arg = 0;

    // DMALLOC_MARK();

    /* always keep track of what we have read */
    IOBUF         iobuf ;
    XZERO        (iobuf);

    P_STATE *s = g->state ;

    char_append (&iobuf, '$');

    if ((c = next_char (g, &iobuf)) != '{')
        goto dump_and_terminate ;

    if (s->out.output_type == nullout) {
        for(;;)
            switch(next_char (g, &iobuf)) {
            case EOF:
                errlog (g, "unexpected end of file");
            case '}':
	    		goto terminate;
            }
    }
    
    if ((var = varname_alloc (g, &iobuf)) == 0) 
        goto dump_and_terminate ;
  
    /* get variable contents */
    Val = (const char **) find_hlst (g->avp, var, 0) ;

    /* check for simple variable substitution */
    switch (c = next_char (g, &iobuf)) {
    case '}':
        if (g->state->dont_expand)
            goto dump_and_terminate ;
        if (Val) /* retrieve/substitute */
            put_string (g, *Val);
        goto terminate ;

    case ':':
        /* up to now we got: "${<varname>:", check for actions */
        switch (op = next_char (g, &iobuf)) {
        case ':': /* set value, no substitution */
        case '=': /* set and substitute value if variable is empty */
        case '*': /* set and substitute value if var is not empty */
        case '-': /* substitute value if variable is empty */
        case '+': /* substitute value if variable is not empty */

            // DMALLOC_MESSAGE (g, "argument_alloc ...");
            if (argument_alloc (g, &arg, &iobuf, '}') == 0 ||
                g->state->dont_expand)
                goto dump_and_terminate ;

            switch (op) {
            case '-': /* substitute value if variable is empty */
                if (Val && *Val)
                    put_string (g, *Val) ;
                else if (arg)
                    put_string (g, arg) ;
                goto terminate ;

            case '+': /* substitute value if variable is not empty */
                if (Val && *Val && arg)
                    put_string (g, arg) ;
                goto terminate ;

            case ':': /* set value */
                op = 0 ; /* no output */
                break ;

            case '*': /* set and substitute value if var is not empty */
                if (Val && *Val)
                    break ;
                goto terminate ;
        
            case '=': /* set and substitute value if variable is empty */
                if (Val && *Val && **Val)
                    goto terminate ;
            }

            /* assign value */
            
            if (Val == 0)                  /* prepare Val */
                Val = (const char **) make_hlst (g->avp, var, 0) ;
            else if (*Val)
                free ((char*)*Val) ;

            if (arg != 0) {
                *Val = arg ;               /* set */
                if (op)                    /* substitute when flagged ok */
                    put_string (g, arg);
                arg = 0 ;                  /* do not recycle "arg" */
                goto terminate;
            }

            if (g->state->delete_on_empty)
                delete_hlst (g->avp, var, 0) ;
            else
                *Val = zmalloc (1) ;       /* zstrdup ("") */
            goto terminate ;

        default:
            errlog (g, "unknown assignment operator \":%c\"", op);
            /* proceed */
        }
        
    default:
        ;
    }
    /* proceed */
   
dump_and_terminate:
    put_string (g, iobuf.data);
    
terminate:
    if (arg) free ((char*)arg);
    if (var) free ((char*)var);
    kill_space (&iobuf);

    // DMALLOC_VERIFY (g);
    // DMALLOC_HUNT ();
}


static
void directive (PARSER *g) {

    int c ;
    cmd_tok cmd;
    const char *arg = 0;

    DMALLOC_MARK();

    /* always keep track of what we have read */
    IOBUF         iobuf  ;
    XZERO        (iobuf) ;
    char_append (&iobuf, '$');

    if ((c = next_char (g, &iobuf)) != '[')
        goto dump_and_terminate ;

    /* parse the command name */
    cmd = command (g, &iobuf);
    if (cmd == 0)
        goto dump_and_terminate ;

    /* parse argument */
    for (;;)
        switch (c = peep_char (g)) {
        case EOF:
            goto dump_and_terminate;

        case ']':
            switch (cmd) {
            case CMD_IF:
            case CMD_ELIF:
            case CMD_INDEX:
            case CMD_VALUE:
            case CMD_COUNT:
                /* argument required */
                errlog (g, "argument required with $[%s]", cmd_tab [cmd]);
                goto next_dump_and_terminate;

            case CMD_ELSE:
                // DMALLOC_MESSAGE (g, "$[else]");

                if (g->state->dont_expand)
                    goto next_dump_and_terminate ;

                /* evaluate argument */
                if (g->state->block.if_true_pending) {
                    g->state->out.output_type       = nullout ;
                    g->state->block.if_true_pending =       0 ;
                    if (g->state->dont_exec_lvl == 0)
                        g->state->dont_exec_lvl = g->state->nesting_level ;
                }
                else if (g->state->block.if_false_pending) {
                    g->state->block.if_false_pending = 0;
                    if (g->state->dont_exec_lvl == g->state->nesting_level) 
                        g->state->dont_exec_lvl = 0 ;
                    switch (g->state->next->out.output_type) {
                    case outstream:
                    case outstream_clone:
                        g->state->out.output_type = outstream_clone ;
                        break;
                    case outstring:
                    case outstring_clone:
                        g->state->out.output_type = outstring_clone ;
                        break ;
                    case nullout:
                        break ;
                    default:
                        errlog (g, no_output_type, "directive", 
                                g->state->next->out.output_type);
                    }
                }
                else {
                    errlog (g, clause_missing, "$[else]", "$[if]");
                    goto next_dump_and_terminate ;
                }

                g->state->block.else_pending = 1;
                goto swallow_and_terminate ;

            case CMD_FI:
                // DMALLOC_MESSAGE (g, "$[fi]");

                if (g->state->dont_expand)
                    goto next_dump_and_terminate ;

                if (g->state->block.if_true_pending  == 0 &&
                    g->state->block.if_false_pending == 0 &&
                    g->state->block.else_pending     == 0) {
                    errlog (g, clause_missing, "$[fi]", "$[if]/$[else]");
                    goto next_dump_and_terminate;
                }
                XZERO (g->state->block) ;
                next_char (g, 0) ; /* swallow */
                pop_state_level (g);
                goto terminate ;

            case CMD_TROS:
                // DMALLOC_MESSAGE (g, "$[tros]");

                if (g->state->block.sort_block_level > 0) {
                    // DMALLOC_VERIFY (g);
                    if (-- g->state->block.sort_block_level == 0) {
                        next_char (g, 0) ;  /* swallow */
                        stop_input (g)   ; /* done with this block */
                        goto terminate ;
                    }
                    
                }
                else
                    errlog (g, clause_missing, "$[tros]", "$[sort]");
                goto next_dump_and_terminate;

            case CMD_COMMENT:
                goto swallow_and_terminate ;
                
            case CMD_OFF:
                // DMALLOC_MESSAGE (g, "$[off]");

                if (g->state->dont_expand)
                    goto next_dump_and_terminate ;
                g->state->out.suspend_output = 1;
                goto swallow_and_terminate ;

            case CMD_ON:
                // DMALLOC_MESSAGE (g, "$[on]");

                if (g->state->dont_expand)
                    goto next_dump_and_terminate ;
                g->state->out.suspend_output = 0;
                goto swallow_and_terminate ;

            case CMD_NOEMPTY:
                if (g->state->dont_expand)
                    goto next_dump_and_terminate ;
                g->state->delete_on_empty = 1 ;
                goto swallow_and_terminate ;
                
            case CMD_EMPTYOK:
                if (g->state->dont_expand)
                    goto next_dump_and_terminate ;
                g->state->delete_on_empty = 0 ;
                goto swallow_and_terminate ;

            case CMD_LOAD:
                // DMALLOC_MESSAGE (g, "$[load]");

                if (g->state->dont_expand || g->state->dont_exec_lvl)
                    goto next_dump_and_terminate ;
                if (load_from_nvram (g, 0))
                    goto swallow_and_terminate ;

                errlog (g, "directive: "
                        "$[load] did not find any variables");
                goto next_dump_and_terminate;

            case CMD_STORE:
                // DMALLOC_MESSAGE (g, "$[store]");

                if (g->state->dont_expand || g->state->dont_exec_lvl)
                    goto next_dump_and_terminate ;
                if (store_into_nvram (g, 0))
                    goto terminate ;

                errlog (g, "directive: "
                        "$[store] did not find any variables");
                goto next_dump_and_terminate;

            case CMD_SORT:
                // DMALLOC_MESSAGE (g, "$[sort]");

                errlog (g, "directive: Command not implementd, yet");
                goto next_dump_and_terminate;

            default:
                /* argument wanted */
                char_append (&iobuf, c);
                goto swallow_and_terminate ;
            }

        case '\t':
        case ' ':
            next_char (g, &iobuf);
            continue ;

        default:
            // DMALLOC_MESSAGE (g, "argument_alloc ...");
            if (argument_alloc (g, &arg, &iobuf, ']') == 0 ||
                g->state->dont_expand && cmd != CMD_SORT)
                goto next_dump_and_terminate ;
                        
            switch (cmd) {
            case CMD_ELSE:
            case CMD_FI:
            case CMD_TROS:
            case CMD_OFF:
            case CMD_ON:
            case CMD_NOEMPTY:
            case CMD_EMPTYOK:
                /* no argument allowed, here */
                errlog (g, "no argument allowed with $[%s]", cmd_tab [cmd]);
                goto next_dump_and_terminate ;

            case CMD_ELIF:

                // DMALLOC_MESSAGE (g, "$[elif ...]");

                if (g->state->block.if_true_pending  == 0 &&
                    g->state->block.if_false_pending == 0) {
                    /* no preceeding if clause */
                    errlog (g, clause_missing, "$[elif]", "$[if]");
                    goto next_dump_and_terminate ;
                }
                /* evaluate argument */
                if (g->state->block.if_true_pending ||
                    arg_expression (g, arg) == 0)   {
                    /* argument evaluates false after a true condition */
                    g->state->out.output_type = nullout ;
                    if (g->state->dont_exec_lvl == 0)
                        g->state->dont_exec_lvl = g->state->nesting_level ;
                }
                else {
                    /* argument evaluates true for the first time */
                    g->state->block.if_true_pending = 1 ;
                    if (g->state->dont_exec_lvl == g->state->nesting_level)
                        g->state->dont_exec_lvl = 0 ;
                    switch (g->state->next->out.output_type) {
                    case outstream:
                    case outstream_clone:
                        g->state->out.output_type = outstream_clone ;
                        break;
                    case outstring:
                    case outstring_clone:
                        g->state->out.output_type = outstring_clone ;
                        break ;
                    case nullout:
                        break ;
                    default:
                        errlog (g, no_output_type, "directive", 
                                g->state->next->out.output_type);
                    }
                }
                goto swallow_and_terminate ;
                    
            case CMD_IF:

                // DMALLOC_MESSAGE (g, "$[if ...]");

                push_state_dup (g) ;
                XZERO (g->state->block);
               
                /* evaluate argument */
                if (arg_expression (g, arg) == 0) {
                    /* argument evaluates false */
                    g->state->out.output_type        = nullout ;
                    g->state->block.if_false_pending =       1 ;
                    if (g->state->dont_exec_lvl == 0)
                        g->state->dont_exec_lvl = g->state->nesting_level ;
                }
                else {
                    /* argument evaluates true */
                    g->state->block.if_true_pending = 1 ;
                }
		
                goto terminate ;

            case CMD_INCLUDE:

                // DMALLOC_MESSAGE (g, "$[include ...]");
                
                if (g->state->dont_exec_lvl)
                    goto swallow_and_terminate ;
                if (push_file_input (g, arg, "include") == 0)
                    goto next_dump_and_terminate;

                run_parser (g);

                /* FIXME: do we really need a 2nd parser instance, here ?? */
				run_parser (g);
                /* goto swallow_and_terminate ; */

            case CMD_COMMENT:
                goto swallow_and_terminate ;

            case CMD_EVAL:
                push_state_dup (g) ;
                g->state->in.input_type        = instring ;
                g->state->in.reader.iobuf      = ZMALLOC (XIOBUF);
                g->state->in.reader.iobuf->buf = ZMALLOC (IOBUF);
                data_append (g->state->in.reader.iobuf->buf, arg, 0);
                run_parser (g);
                goto terminate ;

            case CMD_LOAD:

                // DMALLOC_MESSAGE (g, "$[load ...]");

                if (g->state->dont_exec_lvl)
                    goto swallow_and_terminate ;
                if (load_from_nvram (g, arg))
                    goto terminate ;
                errlog (g, "directive(internal): "
                        "$[load %s] did not find any variables", arg);
                goto next_dump_and_terminate;

			case CMD_STORE:

                // DMALLOC_MESSAGE (g, "$[store ...]");

                if (g->state->dont_exec_lvl)
                    goto swallow_and_terminate ;
                if (store_into_nvram (g, arg))
                    goto swallow_and_terminate ;

                errlog (g, "directive: "
                        "$[store %s] did not find any variables", arg);
                goto next_dump_and_terminate;

            case CMD_SORT: {

                int inx ;
                IOBUF outbuf ;
                XIOBUF xiobuf ;
                const char *pfx, **lst ;

                // DMALLOC_MESSAGE (g, "$[sort ...]");

                if (g->state->block.sort_block_level > 0) {
                    g->state->block.sort_block_level ++ ;
                    goto next_dump_and_terminate ;
                }
                
                pfx = arg ;
                lst = g->state->dont_exec_lvl
                    ? 0
                    : prefixed_entries (g, pfx)
                    ;

                XZERO (outbuf) ;
                XZERO (xiobuf);

                next_char (g, 0) ; /* swallow */

                /* pre-evaluate the block */
                push_state_dup (g) ;
                g->state->out.output_type       = outstring_clone ;
                g->state->out.writer.iobuf      = &xiobuf ;
                g->state->out.writer.iobuf->buf = &outbuf;
                g->state->dont_expand           = 1 ;
                g->state->block.sort_block_level ++ ;
                run_parser_stop (g, 0) ;
                
                if (lst == 0)
                    goto terminate ;

                if (outbuf.data) /* unless output is suspended */
                    for (inx = 0 ; lst [inx] != 0; inx ++) {
                        /* eval the block for all instances of the loop var */
                        push_state_dup (g) ;
                        g->state->block.loop.var       = lst [inx] ;
                        g->state->block.loop.count     = inx + 1 ;
                        g->state->in.input_type        = instring ;
                        g->state->in.reader.iobuf      = ZMALLOC (XIOBUF);
                        g->state->in.reader.iobuf->buf = ZMALLOC (IOBUF);
                        data_append  /* use the result from above */
                            (g->state->in.reader.iobuf->buf, outbuf.data, 0);
                        run_parser_stop (g, 0) ;
                        free ((char*)lst [inx]) ;
                    }

                free ((char*)lst);
                kill_space (&outbuf);
                goto terminate ;
            }

            case CMD_INDEX:
            case CMD_VALUE:
            case CMD_COUNT: {

                /* evaluate data */
                const char **Val, *lpv ;

                // DMALLOC_MESSAGE (g, "$[index/value/count ...]");

                if (g->state->dont_exec_lvl)
                    goto swallow_and_terminate ;

                /* get current loop variable name */
                if ((lpv = g->state->block.loop.var) == 0) {
                    errlog (g, "directive: $[%s %s] could not assign data",
                            cmd_tab [cmd], arg);
                    break ;
                }

                switch (cmd) {
                case CMD_COUNT:
                    /* get the current loop count */
                    sprintf ((char *) (lpv = (const char *) xmalloc (20)),
                             "%d", g->state->block.loop.count);
                    break;
                case CMD_VALUE:
                    /* need the value rather than the variable name */
                    if ((Val = (const char **)
                         find_hlst (g->avp, lpv, 0)) != 0)
                        lpv = *Val ;
                    else
                        lpv = 0 ;
                    /* proceed */
                default:
                    lpv = lpv ? zstrdup (lpv) : zmalloc (1) ;
                }

                /* get or create custom variable from the argument */
                if ((Val = (const char**) find_hlst (g->avp, arg, 0)) != 0) {
                    if (*Val)
                        free ((char*)*Val);
                }
                else
                    Val = (const char **) make_hlst (g->avp, arg, 0) ;

                /* assign current variable/value */
                *Val = lpv ;
                goto terminate ;
            }

            default:
                errlog (g, "directive(internal): "
                        "$[%s %s] is unknown and not implemented",
                        cmd_tab [cmd], arg);
                goto next_dump_and_terminate;
            } /* default: ... switch (cmd) */

            /* NOTREACHED */

        } /* for (;;) */
            
    /* NOTREACHED */

swallow_and_terminate:
    next_char (g, 0);
    goto terminate ;

next_dump_and_terminate:
    next_char (g, &iobuf);
    /* proceed */

dump_and_terminate:
    put_string (g, iobuf.data);
    /* proceed */

terminate:
    kill_space (&iobuf) ;
    if (arg)
        free ((char*)arg);

    // DMALLOC_VERIFY (g);
    // DMALLOC_HUNT ();
}


static
int run_parser_stop (PARSER *g, int eot) {
    int c ;

    for (;;)

        switch (c = next_char (g, 0)) {
        case EOF:
            goto eot_terminate ;

        case '\\':
            if (g->state->dont_expand)
                put_char (g, c);

            switch (c = next_char (g, 0)) {
            case EOF:
                goto eot_terminate ;
            case '\r':
                if (peep_char (g) == '\n')
                    next_char (g, 0) ;
                /* proceed */
            case '\n':
                continue ;
            default:
                put_char (g, c);
            }
            continue ;

        case '$':
            // avdump (g);
            switch (c = peep_char (g)) {
            case '{':
                // DMALLOC_MESSAGE
                //    (g, "run_parser_stop(%#x): complex_variable()", eot);
                complex_variable (g);
                break ;
            case '[':
                // DMALLOC_MESSAGE
                //    (g, "run_parser_stop(%#x): directive()", eot);
                directive (g);
                break ;
            default:
                // DMALLOC_MESSAGE
                //    (g, "run_parser_stop(%#x): simple_variable()", eot);
                simple_variable  (g);
            }
            // DMALLOC_MESSAGE(g, ".");
            // avdump (g);
            continue;

        case '}':
        case ']':
            if (c == eot)
                goto eot_terminate ;
            /* proceed */
          
        default:
            put_char (g, c);
        }

    /* NOTREACHED */

eot_terminate:
    pop_state_level (g);
    // avdump          (g);
    // DMALLOC_VERIFY  (g);
    return c ;
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

PARSER *open_parser (HLST *avp, NVCL *nv, const char *s) {

    PARSER *g ;
    
    if (avp == 0) {
        errlog (g, "AVP subsystem not initialised!");
        return 0;
    }
    if (s == 0)
        s = "%s" ;

    DMALLOC_MARK2 () ;

    g = ZMALLOC (PARSER);
    g->avp   = avp ;
    g->docf  =   s ;
    g->nvram =  nv ;

    return g;
}


void run_parser (PARSER *g) {
    run_parser_stop (g, 0);
}


void close_parser (PARSER *g) {
    while (g->state)
        pop_state (g);
    free (g);

    DMALLOC_HUNT2 ();
}



unsigned file_parser (PARSER *g, const char *file) {

    if (file == 0 || push_file_input (g, file, 0) == 0)
        return 0 ;
    
    g->state->out.output_type       = outstring ;
    g->state->out.writer.iobuf      = ZMALLOC (XIOBUF);
    g->state->out.writer.iobuf->buf = ZMALLOC (IOBUF);
    return 1;
}


const char *data_collect (PARSER *g) {
    if (g->state == 0 || g->state->nesting_level > 1) {
        if (g->state->nesting_level)
            errlog (g, "Parser returned error, nesting level is %d",
                    g->state->nesting_level) ;
        return 0 ;
    }
    switch (g->state->out.output_type) {
    case outstring:
    case outstring_clone:
        return g->state->out.writer.iobuf->buf->data ;
    default:
        return 0;
    }
}


void file_parser_end (PARSER *g) {

    if (g->state->in.input_type == instream &&
        g->state->nesting_level == 1)
        pop_state (g);
}

#ifndef DISCARD_EXTENSIVE_LOGGING
int nvload_verbosity (NVCL *nvram) {
    return nvram_verbosity
        (nvram, PARSER_LISTVAR_LOG, '-', PARSER_LISTVAR_NVOFF);
}

int nvretr_verbosity (NVCL *nvram) {
    return nvram_verbosity
        (nvram, PARSER_GETVAR_LOG, '-', PARSER_GETVAR_NVOFF) ;
}

int nvpacl_verbosity (NVCL *nvram) {
    return nvram_verbosity
        (nvram, PARSER_PAGE_ACL_LOG, '-', PARSER_PAGE_ACL_NVOFF) ;
}

int nvstore_verbosity (NVCL *nvram) {
    return nvram_verbosity
        (nvram, PARSER_SETVAR_LOG, '-', PARSER_SETVAR_NVOFF) ;
}

int nvrestore_verbosity (NVCL *nvram, int noisy) {
    return (noisy < 0)
        ? -3
        : set_nvram_verbosity (nvram, '=', noisy)
        ;
}
#endif

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
