/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Roman Schmidt <rs6@gmx.de>	                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>

#include "hlst.h"
#include "util.h"
#include "http-plugin.h"
#include "parser.h"
#include "nvalloc.h"
#include "plugin-util.h"
#include "logger.h"


/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

#define ACCOUNTNAME_EMPTY			CUSTOM_START_ERROR + 10
#define ACCOUNTNAME_NOT_EXISTS			CUSTOM_START_ERROR + 11
#define FAILOVERNAME_EMPTY			CUSTOM_START_ERROR + 12
#define FAILOVERNAME_NOT_EXISTS			CUSTOM_START_ERROR + 13


static int checkStandardAccount(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char **val = (char**)find_hlst(avp, varName, 0);
	

	if(!val || **val=='\0')
	{
		error = ACCOUNTNAME_EMPTY;
	}
	else if(!getExistNvramVar(nvram_avp, "phone_accounts.id", *val))
	{
		error = ACCOUNTNAME_NOT_EXISTS;
	}
       
	if(error != NO_ERROR)
		handle_varError(avp, error);


	return error;
}

static int checkFailOverAccount(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char **val = (char**)find_hlst(avp, varName, 0);
	

	if(!val || **val=='\0')
	{
		error = FAILOVERNAME_EMPTY;
	}
	else if(!getExistNvramVar(nvram_avp, "phone_accounts.id", *val))
	{
		error = FAILOVERNAME_NOT_EXISTS;
	}
       
	if(error != NO_ERROR)
		handle_varError(avp, error);


	return error;
}


HTMLOBJECT *init_htmlVars(){
	int i=0;
	int max_size 	= 3;
	int max_index 	= max_size - 1;

	HTMLOBJECT *p = (HTMLOBJECT *)xmalloc(sizeof(HTMLOBJECT) * max_size);

	p[i].revise 	= checkStandardAccount;
	p[i++].name	= "phone_dials.standardaccount";
	
	p[i].revise 	= checkFailOverAccount;
	p[i++].name	= "phone_dials.sipfailover";
	
	//sentinel
	p[i>max_index ? max_index : i].revise 	= NULL;
	p[i>max_index ? max_index : i].name		= NULL;

	return p;
}

void free_htmlVars(HTMLOBJECT *p){
	free(p);
	p = NULL;
}


void handle_varError(HLST *avp, int ecode){

 
	switch(ecode){

	case ACCOUNTNAME_EMPTY:		addToList(avp, "phone_dials.accountEmpty.error", "1");
					break;	

	case ACCOUNTNAME_NOT_EXISTS:   	addToList(avp, "phone_dials.accountNotExists.error", "1");
					break;
        
	case FAILOVERNAME_EMPTY:	addToList(avp, "phone_dials.failoverEmpty.error", "1");
					break;	

	case FAILOVERNAME_NOT_EXISTS:  	addToList(avp, "phone_dials.failoverNotExists.error", "1");
					break;
        
	default	:			//UNKNOWN error code, nothing to do
					logger (LOG_WARNING, "UNKNOWN error code: %d: ", ecode);
					return;
	}

	addToList(avp, "error", "1");
    
	return;

}


void *htinit (const char *doc_path, NVCL *nv) {
	STATE *s = ZMALLOC (STATE) ;

	s->htmlvars.objects = init_htmlVars();
	s->htmlvars.load_on_success 		= NULL;
	s->htmlvars.total_elements_label 	= NULL;
	s->htmlvars.group_label 		= NULL;

	s->nvram  	= nv ;
	s->htpath 	= doc_path ;
	return s ;
}

void htcleanup (void *desc) {
	STATE *s = desc ;

	free_htmlVars(s->htmlvars.objects);
	free(s);
}

void htrun (void *desc, RPLY_REC *r, int argc, const char **argv) {
	STATE  *s = desc ;
	PARSER   *parser ;
	HLST        *avp ;
	const char *data ;

	if (argc < 2)
		return ;

	avp = http_args (r);


	CALL_HTTPVARS_ACTION(s,avp);
        
	handle_global_settings(avp, s->nvram);

	parser = open_parser (avp, s->nvram, s->htpath) ;

	if (file_parser (parser, htdocs_p (r, argv [1])) == 0)
		return ;

	/* parser */
	run_parser (parser) ;

	avdump (avp); /* debugging */

	/* print result */
	if ((data = data_collect (parser)) != 0)
		body_append (r, data, 0);

	/* clean up */
	file_parser_end (parser);
	close_parser (parser) ;
}


/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

