/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Roman Schmidt <rs6@gmx.de>	                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>

#include "hlst.h"
#include "http-plugin.h"
#include "parser.h"
#include "nvalloc.h"
#include "plugin-util.h"
#include "logger.h"

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

#define PREFIX_INVALID				CUSTOM_START_ERROR + 10
#define PREFIX_EXISTS				CUSTOM_START_ERROR + 11
#define TIMERANGE_INVALID			CUSTOM_START_ERROR + 20
#define DAYNOTCHECKED				CUSTOM_START_ERROR + 30
#define MAPREFIX_INVALID			CUSTOM_START_ERROR + 40
#define MAMODIFIER_INVALID			CUSTOM_START_ERROR + 50
#define ACCOUNTNAME_EMPTY			CUSTOM_START_ERROR + 60
#define ACCOUNTNAME_NOT_EXISTS			CUSTOM_START_ERROR + 61

#define INVALID_TOTAL_ELEMENTS			CUSTOM_START_ERROR + 70

#define TOTAL_ELEMENTS_LABEL			"phone_dials_config.totalElements"
#define LIST_LABEL				"phone_dials_config.list"

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

int checkAccountExists(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char **val = (char**)find_hlst(avp, varName, 0);
	
	/* first check if account is needed */
	val = (char**)find_hlst(avp, "phone_dials_config.rule", 0);
	if(!val || !*val || **val=='\0') {
		logger(LOG_ERR, "Error, rule not set in phone_dial");
		return NO_ERROR;
	}

	/* if rule is block we do not need an account */
	if (!strcmp("block", *val))
		return NO_ERROR;

	val = (char**)find_hlst(avp, varName, 0);
	if(!val || !*val || **val=='\0')
	{
		error = ACCOUNTNAME_EMPTY;
	}
	else if(!getExistNvramVar(nvram_avp, "phone_accounts.id", *val))
	{
		error = ACCOUNTNAME_NOT_EXISTS;
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);


	return error;
}


int isManipulationActivated(HLST *avp)
{
	char **val, **val2;

	val 	= (char**)find_hlst(avp, "phone_dials_config.rule", 0);
	val2 	= (char**)find_hlst(avp, "phone_dials_config.manipulation", 0);

	if(val && !strcmp(*val, "connect") &&
	   val2 && !strcmp(*val2, "true"))
	{
		return 1;
	}

	return 0;
}


int validateTimeRange(HLST *avp)
{
    long startHour, startMin, endHour, endMin;
	int error = NO_ERROR;



	startHour 	= getVarAsLong(avp, "phone_dials_config.fromhours");
	startMin 	= getVarAsLong(avp, "phone_dials_config.fromminutes");
	endHour 	= getVarAsLong(avp, "phone_dials_config.tohours");
	endMin 		= getVarAsLong(avp, "phone_dials_config.tominutes");


	if(startHour < 0 || startMin < 0 ||
	   endHour	 < 0 || endMin	 < 0 ||
	   startHour > endHour ||
	   (startHour == endHour && startMin > endMin))
	{
		error = TIMERANGE_INVALID;
	}

    if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;

}


int checkPrefix(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char *existVar, **val;

	val = (char**)find_hlst(avp, varName, 0);
	if(!val || **val=='\0' || !validPrefixNumber(val))
	{
		error = PREFIX_INVALID;
	}
	else if(existVar = getExistNvramVar(nvram_avp, varName, *val))	
	{
		/* check if we found the variable that we have to edit */
		if(!isEditVariable(avp,existVar,varName))
			error = PREFIX_EXISTS;
	}


	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}



int checkPeriod(HLST *avp, HLST *nvram_avp, const char *varName)
{
	char **val;
	char *day[7];
	int i;
	int error = NO_ERROR;


	val = (char**)find_hlst(avp, varName, 0);

	if(val && !strcmp(*val, "userdefined"))
	{
		validateTimeRange(avp);
	   
		day[0] = "phone_dials_config.mon";
		day[1] = "phone_dials_config.tue";
		day[2] = "phone_dials_config.wed";
		day[3] = "phone_dials_config.thu";
		day[4] = "phone_dials_config.fri";
		day[5] = "phone_dials_config.sat";
		day[6] = "phone_dials_config.sun";

		for(i=0; i<7; i++)
		{
			val = (char**)find_hlst(avp, day[i], 0);
			if(val && !strcmp(*val, "true"))
				break;
		}

		if(i == 7)
			error = DAYNOTCHECKED;
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}

int checkMaPrefix(HLST *avp, HLST *nvram_avp, const char *varName)
{
	char **val,**val2;
	int error = NO_ERROR;


	if(isManipulationActivated(avp))
	{
		val = (char**)find_hlst(avp, varName, 0);
		val2 = (char**)find_hlst(avp, "phone_dials_config.prefix", 0);

		if(!val || !val2 || strncmp(*val, *val2, strlen(*val)))
		   error = MAPREFIX_INVALID;
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}

int checkMaModifier(HLST *avp, HLST *nvram_avp, const char *varName)
{
	char **val;
	int error = NO_ERROR;


	if(isManipulationActivated(avp))
	{
		val = (char**)find_hlst(avp, varName, 0);
	   
		if(!val || **val=='\0' || !validPhoneNumber(val))
		   error = MAMODIFIER_INVALID;
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}


int setId(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	int error = NO_ERROR;
	char *nextIndex;
	char **val;
	char **ind;

	ind		= (char**)find_hlst(avp, "index", 0);
	//val 		= (char**)find_hlst(avp, varName, 0);
	nextIndex	= getNextIndex(nvram_avp, LIST_LABEL, TOTAL_ELEMENTS_LABEL);
   
	if(!ind || **ind == '\0')
	{
		if(nextIndex)
		{
			forceAddToList(avp,varName,nextIndex);
		}
		else
		{
			error = INVALID_TOTAL_ELEMENTS;
		}
	}
	else
	{
		addToList(avp,varName,*ind);
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return NO_ERROR;
}



HTMLOBJECT *init_htmlVars(){
	int max_size = 20;
	int i=0;
	int max_index = max_size - 1;

	HTMLOBJECT *p = (HTMLOBJECT *)xmalloc(sizeof(HTMLOBJECT) * max_size);

	p[i].revise 	= setId;
	p[i++].name	= "phone_dials_config.id";

	p[i].revise 	= checkPrefix;
	p[i++].name	= "phone_dials_config.prefix";

	p[i].revise 	= checkPeriod;
	p[i++].name	= "phone_dials_config.period";

	p[i].revise 	= noError;
	p[i++].name	= "phone_dials_config.fromhours";

	p[i].revise 	= noError;
	p[i++].name	= "phone_dials_config.fromminutes";

	p[i].revise 	= noError;
	p[i++].name	= "phone_dials_config.tohours";

	p[i].revise 	= noError;
	p[i++].name	= "phone_dials_config.tominutes";

	p[i].revise 	= ensureCheckState;
	p[i++].name 	= "phone_dials_config.mon";

	p[i].revise 	= ensureCheckState;
	p[i++].name 	= "phone_dials_config.tue";

	p[i].revise 	= ensureCheckState;
	p[i++].name 	= "phone_dials_config.wed";

	p[i].revise 	= ensureCheckState;
	p[i++].name	= "phone_dials_config.thu";

	p[i].revise 	= ensureCheckState;
	p[i++].name	= "phone_dials_config.fri";

	p[i].revise 	= ensureCheckState;
	p[i++].name	= "phone_dials_config.sat";

	p[i].revise 	= ensureCheckState;
	p[i++].name	= "phone_dials_config.sun";

	p[i].revise 	= noError;
	p[i++].name	= "phone_dials_config.rule";

	p[i].revise 	= checkAccountExists;
	p[i++].name 	= "phone_dials_config.account";

	p[i].revise 	= ensureCheckState;
	p[i++].name 	= "phone_dials_config.manipulation";

	p[i].revise 	= checkMaPrefix;
	p[i++].name 	= "phone_dials_config.manipulationPrefix";

	p[i].revise 	= checkMaModifier;
	p[i++].name	= "phone_dials_config.manipulationModifier";

	
	//sentinel
	p[i>max_index ? max_index : i].revise 	= NULL;
	p[i>max_index ? max_index : i].name    	= NULL;

	return p;
}



void free_htmlVars(HTMLOBJECT *p){
	free(p);
	p = NULL;
}



void handle_varError(HLST *avp, int ecode){

 
	switch(ecode){

	case PREFIX_INVALID:			addToList(avp, "phone_dials_config.prefix.error", "1");
						break;	

	case PREFIX_EXISTS:			addToList(avp, "phone_dials_config.prefixExists.error", "1");
						break;

	case TIMERANGE_INVALID: 		addToList(avp, "phone_dials_config.timeRange.error", "1");
						break;

	case DAYNOTCHECKED:			addToList(avp, "phone_dials_config.dayNotChecked.error", "1");
						break;

	case MAPREFIX_INVALID:			addToList(avp, "phone_dials_config.manipulationPrefix.error", "1");
						break;

	case MAMODIFIER_INVALID:		addToList(avp, "phone_dials_config.manipulationModifier.error", "1");
						break;

	case ACCOUNTNAME_NOT_EXISTS:		addToList(avp, "phone_dials_config.account.notExists.error", "1");
						break;

	case ACCOUNTNAME_EMPTY:			addToList(avp, "phone_dials_config.account.error", "1");
						break;
									
	default	:				//UNKNOWN error code, nothing to do
						logger (LOG_WARNING, "UNKNOWN error code: %d: ", ecode);
						return;
	}


	addToList(avp, "error", "1");
    
	return;

}


void *htinit (const char *doc_path, NVCL *nv) {
	STATE *s = ZMALLOC (STATE) ;

	s->htmlvars.objects = init_htmlVars();
	s->htmlvars.load_on_success 		= "phone/phone_dials.html";
	s->htmlvars.total_elements_label 	= "phone_dials_config.totalElements";
	s->htmlvars.group_label 		= "phone_dials_config";

	s->nvram  	= nv ;
	s->htpath 	= doc_path ;
	return s ;
}

void htcleanup (void *desc) {
	STATE *s = desc ;

	free_htmlVars(s->htmlvars.objects);
	free(s);
}

void htrun (void *desc, RPLY_REC *r, int argc, const char **argv) {
	STATE  *s = desc ;
	PARSER   *parser ;
	HLST        *avp ;
	const char *data ;
	char **retVal;
	char *html_file;

	if (argc < 2)
		return ;

	html_file = (char*)argv[1];

	avp = http_args (r);

	if((char**)find_hlst(avp,"save", 0))
	{
		 //store
		if(save_indexHttpVars(s,avp))
			html_file = (char*)s->htmlvars.load_on_success;

	}
	else if((char**)find_hlst(avp,"delete", 0))
	{
		 //reorder, delete
		if(delete_indexHttpVars(s,avp))
			html_file = (char*)s->htmlvars.load_on_success;
	}
	else if((char**)find_hlst(avp,"add", 0))
	{
		 //do nothing for add
	}
	else
	{
		  load_indexHttpVars(s, avp);
	}
        
	handle_global_settings(avp, s->nvram);
	
	parser = open_parser (avp, s->nvram, s->htpath) ;

	if (file_parser (parser, htdocs_p (r, html_file)) == 0)
		return ;

	/* parser */
	run_parser (parser) ;

	avdump (avp); /* debugging */

	/* print result */
	if ((data = data_collect (parser)) != 0)
		body_append (r, data, 0);

	/* clean up */
	file_parser_end (parser);
	close_parser (parser) ;
}



/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

