/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Roman Schmidt <r.schmidt@maxina.de> - Maxina GmbH, Neu-       *
*   Isenburg, Germany                                                     *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>

#include "hlst.h"
#include "util.h"
#include "http-plugin.h"
#include "plugin-util.h"
#include "parser.h"
#include "nvalloc.h"
#include "logger.h"


/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

#define INVALID_USERNAME_EMPTY         	CUSTOM_START_ERROR + 10
#define INVALID_PASSWORD_EMPTY	       	CUSTOM_START_ERROR + 40
#define INVALID_PASSWORD_NONEQUAL      	CUSTOM_START_ERROR + 41

#define GROUP_LABEL			"phone_tapi"

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */


int checkUsername(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char **username;
	int len1 = 0;

	username = (char**)find_hlst(avp, varName, 0);
   
	if(!username || !(len1=strlen(*username)))
		error = INVALID_USERNAME_EMPTY;

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}



int confirmPasswd(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char **passwd1;
	char **passwd2;
	int len1 = 0;
	int len2 = 0;

	passwd1 = (char**)find_hlst(avp, varName, 0);
   
	if(!passwd1 || !(len1=strlen(*passwd1)))
		error = INVALID_PASSWORD_EMPTY;

	passwd2 = (char**)find_hlst(avp, "phone_tapi.passwd2", 0);
	
	if(!passwd2 || *passwd2 && !(len2=strlen(*passwd2)))
		error = INVALID_PASSWORD_EMPTY;

	if((error == NO_ERROR) && (len1 != len2 || strcmp(*passwd1,*passwd2)))
		error = INVALID_PASSWORD_NONEQUAL;


	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}

int checkTapiAct(HLST *avp, HLST *nvram_avp, const char *varName)
{
	char **retVal;
	retVal = (char**)find_hlst(avp, varName, 0);

	if(!retVal || !strcmp(*retVal,"enabled"))
		addToList(avp, varName, "false"); 

	return NO_ERROR;
}




HTMLOBJECT *init_htmlVars(){

	int max_size = 5;
	int i=0;
	int max_index = max_size - 1;

	HTMLOBJECT *p = (HTMLOBJECT *)xmalloc(sizeof(HTMLOBJECT) * max_size);

	p[i].revise 	= checkUsername;
	p[i++].name	= "phone_tapi.username";

	p[i].revise 	= confirmPasswd;
	p[i++].name	= "phone_tapi.passwd";

	p[i].revise 	= checkTapiAct;
	p[i++].name	= "phone_tapi.isAct";

	p[i].revise 	= noError;
	p[i++].name	= "phone_tapi.index";
	


	//sentinel
	p[i>max_index ? max_index : i].revise 	= NULL;
	p[i>max_index ? max_index : i].name	= NULL;

	return p;
}

void free_htmlVars(HTMLOBJECT *p){
	free(p);
	p = NULL;
}



void handle_varError(HLST *avp, int ecode){

 
	switch(ecode){

	case INVALID_USERNAME_EMPTY:	addToList(avp, "phone_tapi.username.empty.error", "1"); 
					break;
	case INVALID_PASSWORD_EMPTY:	addToList(avp, "phone_tapi.passwd.empty.error", "1"); 
					break;
	case INVALID_PASSWORD_NONEQUAL:	addToList(avp, "phone_tapi.passwd.nonequal.error", "1"); 
					break;

	default	:      			//UNKNOWN error code, nothing to do
					logger (LOG_WARNING, "UNKNOWN error code: %d: ", ecode);
					return;
	}


	addToList(avp, "error", "1");
    
	return;

}




void *htinit (const char *doc_path, NVCL *nv) {
	STATE *s = ZMALLOC (STATE) ;

	s->htmlvars.objects = init_htmlVars();
	s->htmlvars.load_on_success 		= NULL;
	s->htmlvars.group_label			= GROUP_LABEL;
	s->htmlvars.total_elements_label	= NULL;

	s->nvram  	= nv ;
	s->htpath 	= doc_path ;

	return s ;
}

void htcleanup (void *desc) {
	STATE *s = desc ;

	free_htmlVars(s->htmlvars.objects);
	free(s);
}

void htrun (void *desc, RPLY_REC *r, int argc, const char **argv) {
    
	STATE  *s = desc ;
	PARSER   *parser ;
	HLST        *avp ;
	const char *data ;
	char *html_file = (char*)argv[1];

	if (argc < 2)
		return ;

	avp = http_args (r);

	CALL_HTTPVARS_ACTION(s,avp);
        
	handle_global_settings(avp, s->nvram);

	parser = open_parser (avp, s->nvram, s->htpath) ;

	if (file_parser (parser,  htdocs_p (r, html_file)) == 0)
		return ;

	/* parser */
	run_parser (parser) ;

	avdump (avp); /* debugging */

	/* print result */
	if ((data = data_collect (parser)) != 0)
		body_append (r, data, 0);

	/* clean up */
	file_parser_end (parser);
	close_parser (parser) ;
}



/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

