/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Susanne Goldammer <s.goldammer@maxina.de> - Maxina GmbH,      *
*   Neu-Isenburg, Germany						  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <regex.h>

#include "plugin-util.h"
#include "util.h"
#include "nvram-lib.h"
#include "parser.h"
#include "settings.h"
#include "ctrl.h"
#include "nvalloc.h"

#include "logger.h"

int validateRegExp(char *val, char *pattern)
{
	int  status;
    regex_t   re;

	if(regcomp(&re, pattern, REG_EXTENDED|REG_NOSUB) != 0) {
		return 0;      /* report error */
    }

	status = regexec(&re, val, (size_t) 0, NULL, 0);

	regfree(&re);

    if (status != 0) {
		return 0;      /* report error */
    }

	return 1; 
}


int validPhoneNumber(char **val)
{
	char *pattern = "^([0-9]*)$";

	return validateRegExp(*val, pattern);
}

static int walkInBracket(char* c)
{
	int counter = 0;
	
	while (*c) {
		switch (*c) {
			case ']':
				return counter;

			case '-':
				if (counter != 1)
					return 0;
				if (*(c+1) < '0')
					return 0;
				if (*(c+1) > '9')
					return 0;
				if (*(c+2) != ']')
					return 0;
				break;
		
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				break;

			default:
				/* error */
				return 0;
		}
		c++;
		counter++;
	}
	return 0;
}

int validPrefixNumber(char **val)
{
	char *c = *val;
	int  res;

	while (*c) {
		switch (*c) {
			case '[':
				res = walkInBracket(++c);
				if (!res)
					return 0;
				c+=res;
				break;
		
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				break;

			default:
				/* error */
				return 0;
		}
		c++;
	}
	return 1;
}

char *removeBlanks(char *val)
{
	char *p, *start;;

	start = val;
	p = val;

	while(*val)
	{
		if(!isspace(*val))
		{
			*p = *val;
			p++;

		}
		val++;
	}

            

	*p = *val;

	return start;
}



int validIpAddr(char **val)
{
	int res = 0;
	struct in_addr n_addr;

	//char *pattern = "^(([01]?[0-9]{1,2}|2[0-4][0-9]|25[0-5])\\.){3}([1-9][0-9]?|1[0-9]{1,2}|2[0-4][0-9]|25[0-5])$";

	char *pattern = "^(([0-9]{1,3})\\.){3}([0-9]{1,3})$";


	/* inet_aton returns zero if the address format is invalid */
	if(validateRegExp(*val, pattern) && inet_aton(*val,&n_addr))
	{
		uint32_t	h_addr 	= ntohl(n_addr.s_addr);
		
		uint32_t	ip1	= (h_addr & 0xFF000000)>>24;
		uint32_t	ip2	= (h_addr & 0x00FF0000)>>16;
		uint32_t	ip3	= (h_addr & 0x0000FF00)>>8;
		uint32_t	ip4	= (h_addr & 0x000000FF);

		if(ip1 == 0 && ip2 == 0 && ip3 == 0 && ip4 == 0)
			return 0;

		if(ip1 < 1 || ip1 > 223 || ip1 == 127)
			return 0;	/* invalid ip */

		if(ip2 < 0 || ip2 > 254 )
			return 0;	/* invalid ip */

		if(ip3 < 0 || ip3 > 254 )
			return 0;	/* invalid ip */

		if(ip4 < 1 || ip4 > 254 )
			return 0;	/* invalid ip */


		/* valid ip is present */
		res = 1;
	}


	return res;
}

int validNetworkAddr(char **val)
{
	int res = 0;
	struct in_addr n_addr;

	//char *pattern = "^(([01]?[0-9]{1,2}|2[0-4][0-9]|25[0-5])\\.){3}([1-9][0-9]?|1[0-9]{1,2}|2[0-4][0-9]|25[0-5])$";

	char *pattern = "^(([0-9]{1,3})\\.){3}([0-9]{1,3})$";


	/* inet_aton returns zero if the address format is invalid */
	if(validateRegExp(*val, pattern) && inet_aton(*val,&n_addr))
	{
		uint32_t	h_addr 	= ntohl(n_addr.s_addr);
		
		uint32_t	ip1	= (h_addr & 0xFF000000)>>24;
		uint32_t	ip2	= (h_addr & 0x00FF0000)>>16;
		uint32_t	ip3	= (h_addr & 0x0000FF00)>>8;
		uint32_t	ip4	= (h_addr & 0x000000FF);


		if(ip1 == 0 && (ip2 != 0 || ip3 != 0 || ip4 != 0))
			return 0;

		if(ip1 < 0 || ip1 > 223 || ip1 == 127)
			return 0;	/* invalid ip */

		if(ip2 < 0 || ip2 > 254 )
			return 0;	/* invalid ip */

		if(ip3 < 0 || ip3 > 254 )
			return 0;	/* invalid ip */

		if(ip4 < 0 || ip4 > 254 )
			return 0;	/* invalid ip */


		/* valid ip is present */
		res = 1;
	}


	return res;
}

int validSubnetMask(char **val)
{
        int res = 0;
	struct in_addr n_addr;

	/* inet_aton returns zero if the address is invalid */
	if(inet_aton(*val,&n_addr))
	{
		uint32_t	h_addr 	= ntohl(n_addr.s_addr);

		while(h_addr & 0x80000000)
			h_addr<<=1;

		/*if not all figures in h_addr are zero, we have an incorrect subnet mask */
		if(h_addr)
			res = 0;
		else
			res = 1;
	}
   
	return res;
}


int validMacAddr(char **val)
{
	char *pattern = "^(([0-9A-Fa-f]){2}\\:){5}([0-9A-Fa-f]){2}$";

	return validateRegExp(*val, pattern);
}

int validHexOnly(char *val)
{
    char *pattern = "^([0-9A-Fa-f])+$";   
	return validateRegExp(val, pattern);
}

int validEnc64(char *val)
{
	char *pattern = "^([0-9A-Fa-f]){10}$";

	return validateRegExp(val, pattern);
}

int validEnc128(char *val)
{
	char *pattern = "^([0-9A-Fa-f]){26}$";

	return validateRegExp(val, pattern);
}

int validEnc256(char *val)
{
	char *pattern = "^([0-9A-Fa-f]){58}$";

	return validateRegExp(val, pattern);
}

//make sure that we have checked one of radio buttons
int ensureRadioChecked(HLST *avp, HLST *nvram_avp, const char *varName)
{
	char **val = (char**)find_hlst(avp, varName, 0);

	if(!val)
	{
	   return INVALID_RADIO_OPTION;
	}

	return NO_ERROR;
}


char *getExistNvramVar(HLST *nvram_avp, const char *varName, const char *newVal)
{
	
	const char 	**Val;
	char		*var = NULL;
	int 		ret=0;
	HSRCH 		*lookup = open_hlst_search (nvram_avp);

	while ((Val = (const char**) next_hlst_search (lookup)) != 0) 
	{
		if(strcasecmp(*Val,newVal)!=0)
		   continue;


		var = query_key_hlst((void **)Val) ;

		if(strncmp(varName, var, strlen(varName)) == 0)
		{
			/* we found the same (part of) value in the nvram */
			break;
 		}

		/* that was not what we're looking for */
		var = NULL;
	}

	close_hlst_search (lookup) ;

	return var;
}

long getVarAsLong(HLST *avp, const char *var)
{
	long tmp=-1;
	char *p, **retVal;

	retVal = (char**)find_hlst(avp, var, 0);

	if(retVal && **retVal!='\0')
	{
		tmp = strtol(*retVal,&p,10);

		if(*p != '\0')
			tmp = -1;
	}

	//in error case return -1 
	return tmp;
}


long getVarAsPort(HLST *avp, const char *varName)
{
	long ret = getVarAsLong(avp, varName);

	return (ret<1 || ret>65535) ? -1 : ret;
}


int isEditVariable(HLST *avp, const char *existVar, const char *varName)
{
	int isEdit = 0;
	char **ind = (char**)find_hlst(avp, "index", 0);

	/* check if we are in edit mode */
	if(ind && (**ind != '\0'))	
	{
		/* check if we found the variable we are going to edit */
		if((strlen(existVar) > strlen(varName)) &&
		   (strcmp(existVar + strlen(varName) + 1,*ind) == 0)
		  )
		{
			isEdit = 1;
		}
	}

    return isEdit;
}


void addToList(HLST *avp, const char *varName, const char *varValue)
{
	char **Val;

	Val = (char **) make_hlst(avp, varName, 0) ;

	if(Val)
	{
		/* set the value of the new variable */
		*Val = (char *)xmalloc(strlen(varValue)+1);
		strcpy(*Val, varValue);
	}
}

void forceAddToList(HLST *avp, const char *varName, const char *varValue)
{
	char **Val;

	Val = (char**)find_hlst(avp, varName, 0);


	if(Val)
	{
		/* replace the value if the variable exists */
		*Val = ZREALLOC(*Val, char, strlen(varValue)+1);
		strcpy(*Val, varValue);
	}
	else
	{
		addToList(avp, varName, varValue);
	}
}


unsigned load_from_nvram (NVCL *nvram, HLST *avp, const char *pfx ) {

    int slot, used, n, total ;
    const char *var, *val, **Val ;
    size_t len = zstrlen (pfx);

#   ifndef DISCARD_EXTENSIVE_LOGGING
    int noisy ;
    if ((noisy = nvload_verbosity (nvram)) == -1)
        return 0 ;
#   endif

    nvcl_putcmd (nvram, "slots");
    if (nvcl_rply (nvram, 0) < 1) {
        //error: "NVRAM service communication error"
        return 0 ;
    }

    if (nvcl_getavpair (nvram, 0, &val, 0, 0) == 0) {
		//error: "No NVRAM slots used"
        return 0 ;
    }
    used = atoi (val);

    if (nvcl_getavpair (nvram, 0, &val, 0, 0) == 0) {
        //error: "No NVRAM slots available"
        return 0 ;
    }
    total = atoi (val);

    if (pfx == 0)
        pfx = "" ;
    
    for (slot = 1, n = used; slot <= total && n >= 0; slot ++) {
        char *p, *q, t [20];

        DMALLOC_VERIFY (avp);
        
        sprintf (t, "%u", slot);
        nvcl_putcmd    (nvram, "get");
        nvcl_putavpair (nvram, t, 0, 0);

        if (nvcl_rply (nvram, 0) < 1)
            continue ;

        while (nvcl_getavpair (nvram, &var, &val, 0, 0) != 0) {

            n -- ; /* one less */

            if (val)
                q = zstrdup (val) ;
            else
                q = zmalloc (1) ; /* strdup ("") */

            /* prepend by the prefix */
            p = strcat (strcpy ((char*) xmalloc
                                (strlen (var) + len + 1), pfx), var);

            avdump(avp) ;

            /* get the variable contents */
            if ((Val = (const char **) find_hlst (avp, p, 0)) == 0)
                Val = (const char **) make_hlst (avp, p, 0) ;
            else if (*Val)
                free ((char*)*Val) ;

            /* set the variable contents */
            *Val = q ;

            /* clean up */
            free (p);
        }
    }

#   ifndef DISCARD_EXTENSIVE_LOGGING
    nvrestore_verbosity (nvram, noisy);
#   endif

    DMALLOC_VERIFY (avp);

    return used ;
}


int store_into_nvram (NVCL *nvram ,  HLST *avp) {

    HSRCH *lookup = open_hlst_search (avp) ;
    int    errors = 0 ;
    int     total = 0 ;

    const char **Val ;

#   ifndef DISCARD_EXTENSIVE_LOGGING
    int noisy ;
    if ((noisy = nvstore_verbosity (nvram)) == -1)
        return 0 ;
#   endif

    while ((Val = (const char**) next_hlst_search (lookup)) != 0) {

        const char *var = query_key_hlst ((void **)Val) ;

        DMALLOC_VERIFY (avp);


        /* update nvram */
        if (*Val == 0 || *Val == 0 || **Val == '\0') {
            nvcl_putcmd    (nvram, "del");
            nvcl_putavpair (nvram, var, 0, 0);
        }
        else {
            nvcl_putcmd    (nvram, "set");
            nvcl_putavpair (nvram, var, *Val, 0);
        }

        if (nvcl_rply (nvram, 0) <= 0) {
            //errlog (g, "Cannot set/update NVRAM(%s=%s): %s",
            //    var, *Val ? *Val : "(null)", nvcl_rply_error (g->nvram));
            errors ++ ;
        }
        else
            total ++ ;
    }

    close_hlst_search (lookup) ;

#   ifndef DISCARD_EXTENSIVE_LOGGING
    nvrestore_verbosity (nvram, noisy);
#   endif

    DMALLOC_VERIFY (avp);

    return errors ? -errors : total ;
}

 /* HLST call back function */
void nvram_avp_destroy (void   *state,
						void    *item,
						char     *key,
						unsigned klen) {
    if (item)
        free (item);
}


int ensureCheckState(HLST *avp, HLST *nvram_avp, const char *varName)
{
	char **val = (char**)find_hlst(avp, varName, 0);

	if(!val || strcmp(*val,"true"))
	{
		forceAddToList(avp, varName, "false");
	}


	return NO_ERROR;
}



void load_httpvars(STATE  *s,  HLST *avp)
{
	
	int c;
	char **retVal;
	HLST *nvram_avp;
	HTMLOBJECT *varlist = s->htmlvars.objects;

	nvram_avp = create_hlst(300, nvram_avp_destroy, NULL);

   
	load_from_nvram(s->nvram , nvram_avp, NULL);


	c = -1;

	while(varlist[++c].name)
	{
		retVal = (char**)find_hlst(nvram_avp,varlist[c].name, 0);

		if(retVal)
		{
			addToList(avp, varlist[c].name, *retVal); 
		}
		else
		{
			addToList(avp, varlist[c].name, ""); 
		}
		
	}

	destroy_hlst(nvram_avp);

	return;
}


void save_httpvars(STATE  *s,  HLST *avp)
{
	
	int 	c, error, success=1;
	char 	**retVal;
	HLST 	*nvram_avp;

    HTMLOBJECT *varlist = s->htmlvars.objects;

	nvram_avp = create_hlst(300, nvram_avp_destroy, NULL);
   
	load_from_nvram(s->nvram , nvram_avp, NULL);


	c = -1;

	while(varlist[++c].name)
	{
      
		error = varlist[c].revise(avp, nvram_avp, varlist[c].name);
		
		if(error != NO_ERROR)
			success = 0;
	}

	


	if(success)
	{
		HSRCH 		*lookup = open_hlst_search (avp);
		const char 	**Val;
		int 		ret;

	   

		//clean the avp list before storing
		while ((Val = (const char**) next_hlst_search (lookup)) != 0) 
		{
			const char *var = query_key_hlst ((void **)Val) ;

			//reset the list index
			c = -1;

			while(varlist[++c].name)
			{
				if(strcmp(varlist[c].name, var) == 0)
					break;
			}

			/* delete the variable if it is not found in the vars list */
			if(!varlist[c].name)
			{
				delete_hlst(avp,var,0);
			}
			else if(*Val == 0 || **Val == '\0') 
			{
				/* empty Val indicates that this variable will be deleted from nvram */
				/* so check if the variable is actually in the nvram */
				if(!find_hlst(nvram_avp,varlist[c].name, 0))
					delete_hlst(avp,var,0);		/* not found in nvram, so remove the variable */
			}
	
		}

		close_hlst_search (lookup) ;


		ret = store_into_nvram(s->nvram,  avp);

		if(ret < 0)
		{
		   //NVRAM_ERROR
		}
		else
		{
			addToList(avp, "success", "1"); 
		}
	}

	destroy_hlst(nvram_avp);

	return;
}


void load_indexHttpVars(STATE  *s,  HLST *avp)
{
	
	int c;
	char **retVal;
	char **index;
	HLST *nvram_avp;
	HTMLOBJECT *varlist = s->htmlvars.objects;
   
	nvram_avp = create_hlst(300, nvram_avp_destroy, NULL);

   
	load_from_nvram(s->nvram , nvram_avp, NULL);

	index = (char**)find_hlst(avp,"index", 0);
	if(!index)
	{
		destroy_hlst(nvram_avp);
		return;
	}

	c = -1;

	while(varlist[++c].name)
	{
		char *p = (char*)xmalloc(strlen(varlist[c].name)+strlen(*index)+2);	//2 for '.' and '\0'
		p = strcpy(p,varlist[c].name);
		p = strcat(p,".");
		p = strcat(p,*index);
		
       retVal = (char**)find_hlst(nvram_avp,p, 0);

		if(retVal)
		{
			addToList(avp, varlist[c].name, *retVal); 
		}

		free(p);
		
	}
   

	destroy_hlst(nvram_avp);

	return;
}




int intlen(int c)
{
	int i=1;

	if(c<0)
		i++;

	while(c = (int)(c/10))
		i++;

	return i;
}


char *getNextIndex(HLST *nvram_avp, const char *listLabel, const char *total_elements_label)
{
    char 	**totalElem, **indexList, *_dup, *p;
	int 	*_element, i, result = 1;

	_element 	= NULL;
	_dup 		= NULL;

	indexList	= (char**)find_hlst(nvram_avp, listLabel, 0);
	totalElem 	= (char**)find_hlst(nvram_avp,total_elements_label, 0);
	

	if(indexList && (_dup = zstrdup(*indexList)) && totalElem)
	{
		int stop;
		int total = atoi(*totalElem);

		//integer array for the elements of the indexList
		_element = (int*)xmalloc(sizeof(int) * total);

		p = strtok(_dup, " ");

		//convert ascii to int
		for(i=0; (i<total) && p; i++)
		{
			_element[i] = atoi(p);
			
			p = strtok(NULL, " ");
		}



		stop = 0;
		result = 0;
		while(!stop)
		{
			stop = 1;
			result++;
			for(i=0; i<total; i++)
			{
				if(result == _element[i])
				{
					stop = 0;
					break;
				}
			}
		}
	}

	free(_element);
	free(_dup);
  

	p = (char*)xmalloc(intlen(result) + 1);

	snprintf(p,intlen(result) + 1,"%d",result);
   
	return p;
}


char *increasedTotalElem(HLST *nvram_avp, const char* total_elements_label)
{
	char **p;
	char *newTotalElem = NULL;
	int total=1;

	if(p = (char**)find_hlst(nvram_avp, total_elements_label, 0))
	{
		total = atoi(*p);
		total++;
	}

	newTotalElem = (char*)xmalloc(intlen(total)+1);
	snprintf(newTotalElem, intlen(total)+1, "%d", total);

	return newTotalElem;
}

char *decreasedTotalElem(HLST *nvram_avp, const char* total_elements_label)
{
	char **p;
	char *newTotalElem = NULL;

	if(p = (char**)find_hlst(nvram_avp, total_elements_label, 0))
	{
		int total = atoi(*p);

		if(total>0)
		{
			total--;
			
			newTotalElem = (char*)xmalloc(intlen(total)+1);
			snprintf(newTotalElem, intlen(total)+1, "%d", total);
		}
	}

	return newTotalElem;
}

char *addNewIndex(HLST *nvram_avp, char *listLabel, const char *index)
{
    char **indexList;
	char *newIndexList = NULL;


	indexList	= (char**)find_hlst(nvram_avp, listLabel, 0);

	if(indexList)
	{
		newIndexList = zstrdup(*indexList);  	
		newIndexList = (char *)zrealloc(newIndexList, strlen(newIndexList) + strlen(index) + 2);
		if(*newIndexList != '\0')
			strcat(newIndexList, " ");
		strcat(newIndexList, index);
	}
	else
	{
		newIndexList = (char*)xmalloc(1);
		*newIndexList = '\0';
		newIndexList = (char*)zrealloc(newIndexList,strlen(index) + 1);
		strcat(newIndexList, index);
	}
        
	return newIndexList;
}

char *removeIndex(HLST *nvram_avp, char *listLabel, const char *index)
{
    char **indexList;
	char *newIndexList = NULL;


	indexList	= (char**)find_hlst(nvram_avp, listLabel, 0);

	if(indexList && (**indexList != '\0'))
	{
		char *_tmp, *p;

		_tmp = zstrdup(*indexList);

		newIndexList = (char *)xmalloc(strlen(_tmp) + 1);
		*newIndexList = '\0';

		p = strtok(_tmp, " ");

		while(p)
		{
			if(strcmp(p, index) != 0)
			{
				if(*newIndexList != '\0')
					strcat(newIndexList, " ");

				strcat(newIndexList, p);
			}
			p = strtok(NULL, " ");
		}

		free(_tmp);
	}

	return newIndexList;
}

char *joinStrings(const char *str1, const char *str2)
{
	char *joined;

	joined = (char *)xmalloc(strlen(str1) + strlen(str2) + 1);
	

	strcpy(joined, str1);
	strcat(joined, str2);

	return joined;
}


int save_indexHttpVars(STATE  *s,  HLST *avp)
{
	
	int 	c, error, success=1;
	char 	**retVal;
	HLST 	*nvram_avp = NULL;

	HTMLOBJECT *varlist = s->htmlvars.objects;

	nvram_avp = create_hlst(300, nvram_avp_destroy, NULL);
   
	load_from_nvram(s->nvram , nvram_avp, NULL);

	c = -1;

	while(varlist[++c].name)
	{
      
		error = (*varlist[c].revise)(avp, nvram_avp, varlist[c].name);

		
		if(error != NO_ERROR)
		{
			success = 0;
			//handle_varError(avp, error);
		}
	}
     

	if(success)
	{
		HSRCH 		*lookup = open_hlst_search (avp);
		const char 	**Val;
		int 		ret;
		char 		**p;
		char		*_listlabel, *_index, *_newTotalElem, *_newIndexList;
  
		//now get the index before cleanup
		if((p = (char**)find_hlst(avp,"index", 0)) && (**p != '\0')) 
		{
			_index = zstrdup(*p);
			_newTotalElem 	= NULL;
			_newIndexList 	= NULL;
			_listlabel	  	= NULL;
		}
		else
		{

			_listlabel = joinStrings(s->htmlvars.group_label, ".list");

			_index = getNextIndex(nvram_avp, _listlabel, s->htmlvars.total_elements_label);

			_newTotalElem = increasedTotalElem(nvram_avp, s->htmlvars.total_elements_label);

			_newIndexList = addNewIndex(nvram_avp, _listlabel, _index);
		}



		//clean the avp list before storing
		while ((Val = (const char**) next_hlst_search (lookup)) != 0) 
		{
			const char *var = query_key_hlst ((void **)Val) ;

			//reset the list index
			c = -1;

			while(varlist[++c].name)
			{
				if(strncmp(varlist[c].name, var, strlen(varlist[c].name)) == 0)
					break;
			}

			//delete the variable if it is not found in the vars list
			if(!varlist[c].name)
			{
				delete_hlst(avp,var,0);
			}
		}



		close_hlst_search (lookup) ;

    	c = -1;
		while(varlist[++c].name)
		{
			char **newVal,*p; 
				
			newVal = (char**)find_hlst(avp,varlist[c].name, 0);

			if(!newVal)
				continue;
               
			p = (char*)xmalloc(strlen(varlist[c].name)+strlen(_index)+2);	//2 for '.' and '\0'
			p = strcpy(p,varlist[c].name);
			p = strcat(p,".");
			p = strcat(p,_index);

			if(**newVal != '\0' || find_hlst(nvram_avp, p, 0))
			{
				addToList(avp, p, *newVal);
			}
			delete_hlst(avp, varlist[c].name, 0);

			free(p);
			
		}

		if(_newTotalElem)
		{
			addToList(avp,  s->htmlvars.total_elements_label, _newTotalElem);
		}
			  
		if(_newIndexList && _listlabel)
		{
			addToList(avp,  _listlabel, _newIndexList);
		}


		ret = store_into_nvram(s->nvram,  avp);

		if(ret < 0)
		{
			success = 0;
			//error log:
			//handle_varError(avp, NVRAM_ERROR); 
		}
		else
		{
			addToList(avp, "success", _index); 
		}

 
		free( _listlabel );
		free( _newTotalElem );
		free( _newIndexList );
		free( _index );
	}
	

	destroy_hlst(nvram_avp);

	return success;
}


int delete_indexHttpVars(STATE  *s,  HLST *avp)
{
	int 	c;
	char 	**tmp;
	char	*_total, *_listlabel, *_newIndexList, *_index;
	char 	buffer[200];
	HLST 	*nvram_avp;
	HTMLOBJECT *varlist = s->htmlvars.objects;

	if(!(tmp = (char**)find_hlst(avp,"index", 0)) || (**index=='\0') )
	{
		return 0;
	}

	_index = zstrdup(*tmp);
  
	nvram_avp = create_hlst(300, nvram_avp_destroy, NULL);
   
	load_from_nvram(s->nvram , nvram_avp, NULL);


	if(!(_total = decreasedTotalElem(nvram_avp,  s->htmlvars.total_elements_label)))
	{
		destroy_hlst(nvram_avp);
		free(_index);
		return 0;
	}

	_listlabel = joinStrings(s->htmlvars.group_label, ".list");

	if(!(_newIndexList = removeIndex(nvram_avp, _listlabel, _index)))
	{
		destroy_hlst(nvram_avp);
		free(_index);
		free(_total);
		free(_listlabel);
		return 0;
	}

	flush_hlst(avp, 0, 0);


	/* prepare the avp list with all elements to delete */
	c = -1;
	while(varlist[++c].name)
	{
		char **deleteVal; 

		if(strncmp(varlist[c].name, s->htmlvars.group_label, strlen(s->htmlvars.group_label)))
			continue;

		sprintf(buffer,"%s.%s",varlist[c].name, _index);
				
		deleteVal = (char**)find_hlst(nvram_avp, buffer, 0);
		if(!deleteVal)
			continue;

		addToList(avp, buffer, "");
	}
	
	/* set the decreased Total Elements value */
    addToList(avp,  s->htmlvars.total_elements_label, _total);

	/* set the new Index List (without the deleted index) */
	addToList(avp,  _listlabel, _newIndexList);

  
	if(store_into_nvram(s->nvram,  avp) < 0)
	{
		//error log:
		//handle_varError(avp, NVRAM_ERROR); 
	}
	else
	{
		flush_hlst(avp, 0, 0);
		addToList(avp, "success", _index ); 
	}
	
	
	destroy_hlst(nvram_avp);
	free(_index);
	free(_total);
	free(_listlabel);
	free(_newIndexList);
	  
	return 1;
}



/**
 * This functions counts all type values from nvram beginning
 * with base_name and sets variables 
 * <base_name.account|isdn|sip|external = number 
 * to avp.
 * Example: base_name = phone_accounts 
 * avp values: 	phone_accounts.analog=
 * 		phone_accounts.isdn=
 * 		phone_accounts.sip=
 * 		phone_accounts.external
 */
void count_types(STATE* s, HLST* avp, char* base_name)
{
	HLST* nvram_avp;


	HSRCH 		*lookup;
	const char	**value, *name;

	int counter_analog = 0;
	int counter_isdn   = 0;
	int counter_sip    = 0;
	int counter_ext    = 0;

	char* type_lbl 	= joinStrings(base_name, ".type");
	char* analog_lbl= joinStrings(base_name, ".analog");
	char* isdn_lbl  = joinStrings(base_name, ".isdn");
	char* sip_lbl 	= joinStrings(base_name, ".sip");
	char* ext_lbl 	= joinStrings(base_name, ".external");

	char buffer[32];
	
	nvram_avp = create_hlst(300, nvram_avp_destroy, NULL);
	load_from_nvram(s->nvram , nvram_avp, NULL);

	lookup = open_hlst_search (nvram_avp);
	while ((value = (const char**) next_hlst_search(lookup)) != 0) {

		int tmp_analog = 0;
		int tmp_isdn   = 0;
		int tmp_sip    = 0;
		int tmp_ext    = 0;

		/* first check for type value */
		if(!strcasecmp(*value, "analog")) {
			
			tmp_analog = 1;

		} else if (!strcasecmp(*value, "isdn")) {

			tmp_isdn = 1;
			
		} else if (!strcasecmp(*value, "sip")) {

			tmp_sip = 1;

		} else if (!strcasecmp(*value, "external")) {

			tmp_ext = 1;

		} else
			continue;
		
		/* check if name is correct */
		name = query_key_hlst((void **)value);

		if (strncmp(name, type_lbl, strlen(type_lbl)))
			continue;
		counter_analog += tmp_analog;
		counter_isdn   += tmp_isdn;
		counter_sip    += tmp_sip;
		counter_ext    += tmp_ext;
	}
	close_hlst_search(lookup);
	destroy_hlst(nvram_avp);

	snprintf(buffer, sizeof(buffer), "%d", counter_analog);
	addToList(avp, analog_lbl, buffer);

	snprintf(buffer, sizeof(buffer), "%d", counter_isdn);
	addToList(avp, isdn_lbl, buffer);

	snprintf(buffer, sizeof(buffer), "%d", counter_sip);
	addToList(avp, sip_lbl, buffer);
	
	snprintf(buffer, sizeof(buffer), "%d", counter_ext);
	addToList(avp, ext_lbl, buffer);

	free(type_lbl);
	free(analog_lbl);
	free(isdn_lbl);
	free(sip_lbl);
	free(ext_lbl);
}

int noError(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	return NO_ERROR;
}


/**
 * Handles language and expert mode for plugins.
 */
void handle_global_settings(HLST* avp, NVCL* nvram)
{
	char* lang;
	char* mode;
	
	/* handle language */
	lang = get_from_list(avp, "lang");
	if (lang && *lang) {
		/* store into nvram */
		add_to_nvram(nvram, "global.lang", lang);
	} else {
		/* get from nvram */
		add_to_list(avp, "lang", get_from_nvram(nvram, "global.lang"));
	}

	/* handle mode */
	mode = get_from_list(avp, "mode");
	if (mode && *mode) {
		/* store into nvram */
		add_to_nvram(nvram, "global.mode", mode);
	} else {
		/* get from nvram */
		add_to_list(avp, "mode", get_from_nvram(nvram, "global.mode"));
	}

	
	/* handle first time visit */
	mode = get_from_list(avp, "firsttime");
	if (mode && *mode) {
		/* store into nvram */
		add_to_nvram(nvram, "global.firsttime", mode);
	} else {
		/* get from nvram */
		add_to_list(avp, "firsttime", get_from_nvram(nvram, "global.firsttime"));
	}

	
}


/**
 * Adds a key value pair to the list. If the key does not exist, it will
 * be added
 */
void add_to_list(HLST* avp, const char* name, const char* value)
{
	char** val;

	if (!name || !value)
		return;

	val = (char**)find_hlst(avp, name, 0);

	if (!val)
		val = (char**)make_hlst(avp, name, 0);

	if (!val)
		return; /* error */

	*val = ZREALLOC(*val, char, strlen(value)+1);
	strcpy(*val, value);
}

/**
 * Returns a value specified by a key. If value is not found
 * null is returned
 */
char* get_from_list(HLST* avp, const char* name) 
{
	char** var;

	if (!name)
		return;

	var = (char**)find_hlst(avp, name, 0);

	if (var && *var)
		return *var;

	return 0;
}

/**
 * Adds a key value pair to nvram. If the key does not yet exist
 * it will be created.
 */
void add_to_nvram(NVCL* nvram, const char* name, const char* value)
{
	if (value)
		nvcl_putcmd(nvram, "set");
	else
		nvcl_putcmd(nvram, "del");

	nvcl_putavpair(nvram, name, value, 0);

	if (nvcl_rply(nvram, 0) < 1)
		return;
	return;
}

/**
 * Sends a command to nvram. If the key does not yet exist
 * it will be created.
 */
void cmd_to_nvram(NVCL* nvram,const char* cmd, const char* name, const char* value)
{
	nvcl_putcmd(nvram, cmd);

	nvcl_putavpair(nvram, name, value, 0);

	if (nvcl_rply(nvram, 0) < 1)
		return;
	return;
}

/**
 * Returns a value specified by a key from nvram. If value is not 
 * found null is returned
 */
char* get_from_nvram(NVCL* nvram, const char* name)
{
#   ifndef DISCARD_EXTENSIVE_LOGGING
    NVCL *nvclone;
    int noisy ;
#   endif

	const char *var, *val = 0;

	if (!name)
		return;

#   ifndef DISCARD_EXTENSIVE_LOGGING
    if ((noisy = nvretr_verbosity (nvram)) == -1)
        return 0 ;
#   endif

	nvcl_putcmd(nvram, "set");
	nvcl_putavpair(nvram, name, 0, 0);

	if (nvcl_rply(nvram, 0) > 0)
        nvcl_getavpair(nvram, &var, &val, 0, 0);

#   ifndef DISCARD_EXTENSIVE_LOGGING
    nvclone = nvcl_clone (nvram); /* save nvcl state */
    nvrestore_verbosity (nvclone, noisy);
    nvcl_close (nvclone);
#   endif

	return (char*)val;
}


/**
 * This function sends a html page to the client. Before this is done
 * the html page is parsed.
 */
void send_html_page(HLST* avp, STATE* s, RPLY_REC* r, const char* templ_name)
{
	PARSER*		parser;
	const char*	data;

	parser = open_parser(avp, s->nvram, s->htpath);

	if ( file_parser(parser, templ_name) == 0 ) {
		close_parser(parser);
		return;
	}

	/* parser */
	run_parser(parser) ;

	/* print result */
	if ((data = data_collect(parser)) != 0)
		body_append(r, data, 0);

	/* clean up */
	file_parser_end(parser);
	close_parser(parser) ;
}


/*
 * Just Check a checkbox for true and false. If value is not set 
 * it is false
 */
int checkIdBlocking(HLST *avp, HLST *nvram_avp, const char *varName) 
{
	char **val = (char**)find_hlst(avp, varName, 0);

	if(!val || !*val || !**val)
		add_to_list(avp, varName, "false");

	return NO_ERROR;
}


/*
 * Returns true if nvram value is "true"
 */
int is_enabled(NVCL* nvram, const char* name)
{
	char* value;

	value = get_from_nvram(nvram, name);

	if (!value || !*value)
		return 0;

	return !strcmp(value, "true");
}

/* 
 * check if all necessary params are set for sip account 
 * currently not used
 */
int valid_sip_parms( HLST* avp, const char* id, char* base_name)
{
	char* acc_name		= (char*)joinStrings(base_name, joinStrings(".name."		, id));
	char* acc_type		= (char*)joinStrings(base_name, joinStrings(".type."		, id));
	char* acc_number	= (char*)joinStrings(base_name, joinStrings(".number."		, id));
	char* acc_server	= (char*)joinStrings(base_name, joinStrings(".server."		, id));
	char* acc_user		= (char*)joinStrings(base_name, joinStrings(".username."	, id));
	char* acc_passwd	= (char*)joinStrings(base_name, joinStrings(".password."	, id));

	char **type, **name, **number, **server, **user, **passwd;

	type = (char**)find_hlst(avp, acc_type, 0);
	 
	if(!type || !*type || strcmp(*type,"sip"))
		return 0;

	name 	= (char**)find_hlst(avp, acc_name, 		0);
	number 	= (char**)find_hlst(avp, acc_number, 	0);
	server 	= (char**)find_hlst(avp, acc_server, 	0);
	user 	= (char**)find_hlst(avp, acc_user, 		0);
	passwd 	= (char**)find_hlst(avp, acc_passwd, 	0);

	if(	!name 	|| !*name 	|| !**name 		||
		!number || !*number || !**number	||
		!server	|| !*server	|| !**server	||
		!user	|| !*user	|| !**user		||
		!passwd	|| !*passwd	|| !**passwd 
		)
		return 0;

	return 1;	
}

/* 
 * count sip accounts with valid params
 * currently not used
 */
int count_valid_sip_accounts(STATE* s, HLST* avp, char* base_name)
{
	HLST* nvram_avp;

	HSRCH 		*lookup;
	const char	**value, *name;

	int counter_sip    = 0;

	char* phone_accounts_id_lbl 	= (char*)joinStrings(base_name, ".id");
	char* sip_lbl 					= (char*)joinStrings(base_name, ".sip");

	char buffer[32];
	
	nvram_avp = create_hlst(300, nvram_avp_destroy, NULL);
	load_from_nvram(s->nvram , nvram_avp, NULL);

	lookup = open_hlst_search (nvram_avp);
	while ((value = (const char**) next_hlst_search(lookup)) != 0) {

		if( !value || !*value || !**value)
			continue; 

		/* check if name is correct */
		name = query_key_hlst((void **)value);

		if (strncmp(name, phone_accounts_id_lbl, strlen(phone_accounts_id_lbl)))
			continue;

		if(valid_sip_parms(nvram_avp, *value, base_name))
			counter_sip    += 1;
	}
	close_hlst_search(lookup);
	destroy_hlst(nvram_avp);

	snprintf(buffer, sizeof(buffer), "%d", counter_sip);
	forceAddToList(avp, sip_lbl, buffer);

	free(phone_accounts_id_lbl);
	free(sip_lbl);
}

char *substr(char *string, int start, size_t count) {
  static char str[65536];
  str[0] = '\0'; /* The NUL string error return */

  if (string != NULL) {
    size_t len = strlen(string);
    if (len < 65536) {
      if (start < 0)
	start = len + start;
      if (start >= 0 && start < len) {
	if (count - 1 > len - start)
	  count = len - start;
	strncpy(str, string + start, count);
	str[count] = 0;
      }
    }
  }

  return str;

} 

char *stripstr(char *string) {
  char *s, *d;

  for(s = string; isspace(*s); s++)
    ;

  d = string;
  while(*d++ = *s++)
    ;

  d--;

  while(d > string) {
    d--;
    if (!isspace(*d)) break;
    *d = 0;
  }

  return string;

}
