/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef HAVE_OPENSSL

#include <arpa/inet.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#ifdef HAVE_STRING_H
# include <string.h>
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#define SYSLOG_NAMES
#include <syslog.h>

#include <openssl/ssl.h>
#include <openssl/rand.h>
#include <openssl/x509.h>
#include <openssl/err.h>

#include "util.h"
#include "nvalloc.h"
#include "settings.h"
#include "buffer.h"
#include "logger.h"

#define __MHTTPD_SSL_PRIVATE
#include "ssl.h"

#ifdef WITH_DMALLOC
static unsigned long __mark;

#define DMALLOC_MARK() { __mark = dmalloc_mark (); }

#define DMALLOC_HUNT() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }

#define DMALLOC_VERIFY(c) {                             \
                 if (c) dmalloc_verify (c);             \
                 dmalloc_log_stats ();                  }

#define DMALLOC_MESSAGE(g,fmt, ...) {                   \
                 dmalloc_message (fmt, ##__VA_ARGS__) ; \
                 if (g) DMALLOC_VERIFY(g) ;             }

#else
#define DMALLOC_MARK()            /* n/a */
#define DMALLOC_HUNT()            /* n/a */
#define DMALLOC_VERIFY(g)         /* n/a */
#define DMALLOC_MESSAGE(fmt, ...) /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * Private: variables and definitions
 * ----------------------------------------------------------------------- */

#if SSLEAY_VERSION_NUMBER < 0x00904000L
# error SSLeay version must be at least 0.9.4
#endif

/* change all non-printable characters to '.' */
#define safestring(s) {char *_ ; for (_=(s); *(_); _++) \
    if (isprint ((int)*(_)) == 0 && isspace ((int)*(_)) == 0) *_='.'; }

/* change all unsafe characters to '.' */
#define safename(s) \
    {char *_; for (_ = (s); *_; _++) if (isalnum ((int)*_) == 0) *_='.'; }

#define KEY_CACHE_LENGTH 2049
#define KEY_CACHE_TIME   3600

#define DIM(a) (sizeof (a) / sizeof (a [0]))

#define __str(x)              # x
#define str(x)                __str (x)

static int verify_depth    ;
static int verify_mode     ;
static int accept_count    ;
static int loop_count      ;

/* ----------------------------------------------------------------------- *
 * Private: diagnostics
 * ----------------------------------------------------------------------- */

static
void info_log (const char *f, ...) {
    char format [200];
    va_list ap;
    va_start (ap, f);
    sprintf (format, "SSL info(%d): %s", accept_count, f);
    vlogger (LOG_DEBUG, format, ap);
    va_end (ap);
}

static
void load_session_info (SESS_INFO *s, SSL_CTX *ctx) {
    s->accept_loop         = loop_count ;
    s->number              = SSL_CTX_sess_number (ctx);
    s->connect             = SSL_CTX_sess_connect (ctx);
    s->connect_good        = SSL_CTX_sess_connect_good (ctx);
    s->connect_renegotiate = SSL_CTX_sess_connect_renegotiate (ctx);
    s->accept              = SSL_CTX_sess_accept (ctx);
    s->accept_good         = SSL_CTX_sess_accept_good (ctx);
    s->accept_renegotiate  = SSL_CTX_sess_accept_renegotiate (ctx);
    s->hits                = SSL_CTX_sess_hits (ctx);
    s->misses              = SSL_CTX_sess_misses (ctx);
    s->timeouts            = SSL_CTX_sess_timeouts (ctx);
}

static
void info_callback (const SSL *s, int where, int ret) {
    SESS_INFO i;
    
    if (verbosity & XXL_STATS_LOG)
        switch (where) {
        case SSL_CB_ACCEPT_LOOP:
            loop_count ++ ;
            break ;
        case SSL_CB_ACCEPT_EXIT:
            info_log ("exit accept state");
            break ;
        case SSL_CB_READ_ALERT:
            info_log ("read alert: %s: %s",
                      SSL_alert_type_string_long (ret),
                      SSL_alert_desc_string_long (ret));
            break ;
        case SSL_CB_WRITE_ALERT:
            info_log ("write alert: %s: %s",
                      SSL_alert_type_string_long (ret),
                      SSL_alert_desc_string_long (ret));
            break ;
        case SSL_CB_HANDSHAKE_START:
            info_log ("starting handshake ...");
            loop_count = 0 ;
            break ;
        case SSL_CB_HANDSHAKE_DONE:
            info_log ("handshake done, statistics:");
            load_session_info (&i, SSL_get_SSL_CTX ((SSL*)s));
#           define I(fmt,val) {if (i.val != 0) info_log ("%3d " fmt,i.val);}
            I ("times accept state visited",        accept_loop);
            I ("items in the session cache",        number);
            I ("client connects (SSL_connect())",   connect);
            I ("client connects that have finished",connect_good);
            I ("client renegotiatations requested", connect_renegotiate);
            I ("server connects (SSL_accept())",    accept);
            I ("server connects that have finished",accept_good);
            I ("server renegotiatiations requested",accept_renegotiate);
            I ("session cache hits",                hits);
            I ("session cache misses",              misses);
            I ("session cache timeouts",            timeouts);
#           undef I
            break ;
        default:
            info_log ("unknown state: $#x: %s",
                      where, SSL_state_string_long (s));
        }
}


static
const char *ssl_reason_text (int n) {
#   define __case(n) case n: return # n
#   define __fmt "unknown code: %#x"

    switch (n) {
        static char buf [sizeof (__fmt) + sizeof (n) * 2] ;
    __case (SSL_ERROR_NONE);
    __case (SSL_ERROR_ZERO_RETURN);
    __case (SSL_ERROR_WANT_X509_LOOKUP);
    __case (SSL_ERROR_WANT_READ);
    __case (SSL_ERROR_WANT_WRITE);
    __case (SSL_ERROR_SYSCALL);
    __case (SSL_ERROR_SSL);
    default:
        sprintf (buf, "unknown code: %#x", n);
        return buf;
    }

#   undef __case
#   undef __fmt
}
    

static
void sslerror (char *txt) { /* SSL Error handler */
    int level= LOG_ERR, save = errno ;
    long err ;

    if ((err = ERR_get_error ()) == 0)
        logger (level, "%s: Peer suddenly disconnected", txt);
    else
        do {
            char str [120];
            ERR_error_string (err, str);
            logger (level, "%s: %s", txt, str);
            level = LOG_INFO ;
        }
        while ((verbosity & XXL_STATS_LOG) && (err = ERR_get_error ()));
    
    errno = save ;
}

/* ----------------------------------------------------------------------- *
 * Private: key and certificate management
 * ----------------------------------------------------------------------- */

static
unsigned get_nv_value (NVCL          *nv,
                       const char **Data,
                       size_t       *Len,
                       const char  *ntry) {
    if (ntry == 0)
        return 0;
    
    nvcl_putcmd    (nv, "set");
    nvcl_putavpair (nv, ntry, 0, 0);
    return
        nvcl_rply (nv, 0) > 0 && nvcl_getavpair (nv, 0, Data, Len, 0);
}


#ifndef NO_DH
static
unsigned load_dhparams (NVCL         *nv,
                        SSL_CTX     *ctx,
                        const char *ntry) {
    DH *dh = 0;
    const char *data ;
    const unsigned char *udata ;
    size_t len ;

    if (ntry == 0)
        ntry = XXL_DHPARAM_NAME ;

    if (get_nv_value (nv, &data, &len, ntry) == 0) {
        logger (LOG_ERR, "Error retrieving the dh params entry: %s", ntry);
        return 0 ;
    }

    udata = (const unsigned char *)data ; /* avoid warnings */
    if (d2i_DHparams (&dh, &udata, len) == 0) {
        logger (LOG_ERR, "Error parsing the dh params entry: %s", ntry);
        sslerror ("d2i_DHparams");
        return 0 ;
    }
    
    SSL_CTX_set_tmp_dh (ctx, dh);

    if (verbosity & XXL_SETUP_LOG)
        logger (LOG_INFO, "Diffie-Hellman initialized with %d bit key",
                8 * DH_size (dh));
    DH_free (dh);

    return 1; /* OK */
}
#endif /* NO_DH */


#ifndef NO_RSA
static
RSA *make_temp_key (int keylen) {
    RSA *result;

    if (verbosity & XXL_SETUP_LOG)
        logger (LOG_INFO, "Generating %d bit temporary RSA key...", keylen);

    result = RSA_generate_key (keylen, RSA_F4, 0, 0);

    if (verbosity & XXL_SETUP_LOG)
        logger (LOG_INFO, "Temporary RSA key created");

#   if SSLEAY_VERSION_NUMBER <= 0x0090701fL
    /* Set blinding iff it's not built into our OpenSSL version */
    RSA_blinding_on (result, 0);
#   endif

    return result;
}


static
RSA *tmp_rsa_cb (SSL *ss, int export, int keylen) {
    static int initialized = 0;
    static RSA    *longkey = 0;
    static int     longlen = 0;
    static time_t longtime = 0;
    static struct keytabstruct {
        RSA *key;
        time_t timeout;
    } keytable [KEY_CACHE_LENGTH];

    RSA *oldkey;
    time_t now;

    if (initialized == 0) {
        int i;
        for (i = 0; i < DIM (keytable); i++) {
            keytable [i].key     = 0;
            keytable [i].timeout = 0;
        }
        initialized = 1;
    }

    time (&now);

    if (keylen < DIM (keytable)) {
        if (keytable [keylen].timeout < now) {
            oldkey                    = keytable [keylen].key;
            keytable [keylen].key     = make_temp_key (keylen);
            keytable [keylen].timeout = now + KEY_CACHE_TIME;
            if (oldkey)
                RSA_free (oldkey);
        }

        return keytable [keylen].key;
    }

    /* Temp key > 2048 bits.  Is it possible? */
    if (longtime < now || longlen != keylen) {
        oldkey   = longkey;
        longkey  = make_temp_key (keylen);
        longtime = now + KEY_CACHE_TIME;
        longlen  = keylen;
        if (oldkey)
            RSA_free (oldkey);
    }

    return longkey;
}
#endif /* NO_RSA */


static
unsigned load_server_cert (NVCL         *nv,
                           SSL_CTX     *ctx,
                           const char *ntry) {
    const char *data ;
    size_t len ;

    if (ntry == 0)
        ntry = XXL_CERT_NAME ;

    if (get_nv_value (nv, &data, &len, ntry) == 0) {
        logger (LOG_ERR, "Error retrieving the server certificate entry: %s",
                ntry);
        return 0 ;
    }

    if (SSL_CTX_use_certificate_ASN1 (ctx, len, (unsigned char*)data) == 0) {
        logger (LOG_ERR, "Error using the certificate entry: %s", ntry);
        sslerror ("SSL_CTX_use_certificate_ASN1");
        return 0 ;
    }
    
    if (verbosity & XXL_SETUP_LOG)
        logger (LOG_DEBUG, "Certificate entry: %s", ntry);

    return 1;
}


static
unsigned load_server_key (NVCL         *nv,
                          SSL_CTX     *ctx,
                          const char *ntry) {
    const char *data ;
    size_t len ;

    if (ntry == 0)
        ntry = XXL_KEY_NAME ;

    if (get_nv_value (nv, &data, &len, ntry) == 0) {
        logger (LOG_ERR, "Error retrieving the server key entry: %s", ntry);
        return 0 ;
    }

#   ifdef NO_RSA
    if (SSL_CTX_use_PrivateKey_ASN1
        (y.ctx, (unsigned char*)data, len) == 0) {
        logger (LOG_ERR, "Error using the private key entry: %s", ntry);
        sslerror ("SSL_CTX_use_PrivateKey_ASN1");
        return 0 ;
    }
#   else /* NO_RSA */
    if (SSL_CTX_use_RSAPrivateKey_ASN1
        (ctx, (unsigned char*)data, len) == 0) {
        logger (LOG_ERR, "Error using the private key entry: %s", ntry);
        sslerror ("SSL_CTX_use_RSAPrivateKey_ASN1");
        return 0 ;
    }
#   endif /* NO_RSA */

    return 1;
}


static
unsigned load_ca_certs (SSL_CTX        *ctx,
                        const char *ca_file,
                        const char  *ca_dir) {
    if (ca_file != 0) {
        if (SSL_CTX_load_verify_locations (ctx, ca_file, 0) == 0) {
            logger (LOG_ERR, "Error loading verify certificates from %s",
                    ca_file);
            sslerror ("SSL_CTX_load_verify_locations");
            return 0;
        }
        SSL_CTX_set_client_CA_list (ctx, SSL_load_client_CA_file (ca_file));
        logger (LOG_DEBUG, "Loaded verify certificates from %s", ca_file);
    }

    if (ca_dir != 0) {
        if (SSL_CTX_load_verify_locations (ctx, 0, ca_dir) == 0) {
            logger (LOG_ERR, "Error setting verify directory to %s", ca_dir);
            sslerror ("SSL_CTX_load_verify_locations");
            return 0;
        }
        logger (LOG_DEBUG, "Set verify directory to %s", ca_dir);
    }

    return 1;
}


static
unsigned load_cipher_list (NVCL         *nv,
                           SSL_CTX     *ctx,
                           const char *ntry) {
    const char *data ;

    if (ntry == 0)
        return 1;

    if (get_nv_value (nv, &data, 0, ntry) == 0) {
        logger (LOG_ERR, "Error retrieving the cipher list entry: %s", ntry);
        return 0 ;
    }

    if (SSL_CTX_set_cipher_list (ctx, data) == 0) {
        sslerror ("SSL_CTX_set_cipher_list");
        return 0;
    }

    return 1;
}


/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
void cert_specs (X509 *cert,
                 char *subject, size_t slen,
                 char *issuer,  size_t ilen) {
    if (subject) {
        X509_NAME_oneline (X509_get_subject_name (cert), subject, slen);
        safestring (subject);
    }
    if (issuer) {
        X509_NAME_oneline (X509_get_issuer_name (cert), issuer, ilen);
        safestring (issuer);
    }
}


static /* our verify callback function */
int verify_callback (int preverify_ok, X509_STORE_CTX *ctx) {
    X509_OBJECT ret;
    char s [1024], t [1024];

    if ((verify_mode & ~XXL_VERIFY_USE_ONLY_MY) == SSL_VERIFY_NONE) {
        if (verbosity & XXL_VERIFY_LOG) {
            cert_specs (ctx->current_cert, s, sizeof (s), t, sizeof (t));
            logger (LOG_INFO, "CERT VERIFY IGNORE: depth=%d, %s by %s",
                    ctx->error_depth, s, t);
        }
        return 1; /* Accept connection */
    }

    if (preverify_ok == 0) {
        /* Remote site specified a certificate, but it's not correct */
        if (verbosity & XXL_VERIFY_LOG) {
            cert_specs (ctx->current_cert, s, sizeof (s), t, sizeof (t));
            logger (LOG_WARNING,
                    "CERT VERIFY ERROR: depth=%d, error=%s: %s by %s",
                    ctx->error_depth,
                    X509_verify_cert_error_string (ctx->error), s, t);
        }
        return 0; /* Reject connection */
    }

    if ((verify_mode & XXL_VERIFY_USE_ONLY_MY) &&
        ctx->error_depth == 0 &&
        X509_STORE_get_by_subject (ctx, X509_LU_X509,
                                   X509_get_subject_name (ctx->current_cert),
                                   &ret) != 1) {
        if (verbosity & XXL_VERIFY_LOG) {
            cert_specs (ctx->current_cert, s, sizeof (s), t, sizeof (t));
            logger (LOG_WARNING,
                    "CERT VERIFY ERROR ONLY MY: no cert for %s by %s", s, t);
        }
        return 0; /* Reject connection */
    }

    if (verify_depth > 0 && ctx->error_depth > verify_depth) {
        if (verbosity & XXL_VERIFY_LOG) {
            cert_specs (ctx->current_cert, s, sizeof (s), t, sizeof (t));
            logger (LOG_WARNING,  "CERT VERIFY ERROR: depth=%d>%d, %s by %s",
                    ctx->error_depth, verify_depth, s, t);
        }
        return 0; /* Reject connection */
    }
    
    if (verbosity & XXL_VERIFY_LOG) {
        cert_specs (ctx->current_cert, s, sizeof (s), t, sizeof (t));
        logger (LOG_INFO, "CERT VERIFY OK: depth=%d, %s by %s",
                ctx->error_depth, s, t);
    }

    return 1; /* Accept connection */
}


static
void ssl_accept_ok (XXL *ss) {
    X509 *peer ;
    char s [1024], t [1024];

    if (verbosity & XXL_TRAFFIC_LOG)
        logger (LOG_INFO, "SSL_accept: done on fd=%d after %d tries",
                ss->fd, accept_count);

    if (verbosity & XXL_SETUP_LOG) {
        SSL_CIPHER *cipher ;
        if ((cipher = SSL_get_current_cipher (ss->ssl)) != 0) {
            size_t len ;
            SSL_CIPHER_description (cipher, s, sizeof (s));
            if ((len = strlen (s)) > 0) s [len-1] = '\0';
            logger (LOG_INFO, "Negotiated ciphers: %s", s);
        }
    }
    
    if ((peer = SSL_get_peer_certificate (ss->ssl)) != 0) {
        cert_specs (peer, s, sizeof (s), t, sizeof (t));
        if (verbosity & XXL_SETUP_LOG)
            logger (LOG_INFO, "Peer cert %s by %s", s, t);
        ss->dn = zstrdup (s) ;
    }
}


static
int ssl_accept_continue (XXL *ss, void *buf, int len) {
    int n ;
    
    switch (ss->ssl_pending) {
    case SSL_ERROR_WANT_READ:
        if (verbosity & XXL_SETUP_LOG)
            logger (LOG_INFO, "SSL_accept continues ...");

        /* ok, mark the connection opened, note that ssl_recv()
           needs ss->ssl_pending == 0 - otherwise recursion! */
        ss->ssl_pending = 0 ;
        
        switch (n = ssl_recv (ss, buf, len)) {
        case -1:
            if (errno == EAGAIN) {
                ss->ssl_pending = SSL_ERROR_WANT_READ ;
                return -1;
            }
            /* break; */

        case 0:
            break ;

        default:
    	    if (verbosity & XXL_SETUP_LOG)
                logger (LOG_INFO, "SSL_accept on fd=%d: len=%d", ss->fd, n);
            ssl_accept_ok (ss);
            return n ;
        }
    }

    SSL_shutdown (ss->ssl); /* Send close_notify back */
    return -1;
}


/* ----------------------------------------------------------------------- *
 * Private: ssl object constructor/destructor
 * ----------------------------------------------------------------------- */

unsigned _ssl_cty_setup (CTY *y) {

    static int global_init = 0;
    const char *data;

    if (global_init == 0) {
        global_init = 1 ;
        SSLeay_add_ssl_algorithms ();
        SSL_load_error_strings ();
    }
    
    if (RAND_status () == 0) {
        logger (LOG_ERR, "Entropy missing - Problems initialising the PRNG");
        return 0;
    }

    if ((y->ctx = SSL_CTX_new (SSLv23_server_method ())) == 0) {
        sslerror ("SSL_CTX_new");
        return 0;
    }

#   ifndef NO_RSA
    SSL_CTX_set_tmp_rsa_callback (y->ctx, tmp_rsa_cb);
#   endif /* NO_RSA */

#   ifndef NO_DH
    if (load_dhparams (y->nv, y->ctx, y->xxl.dh_name) == 0)
        logger (LOG_WARNING, "Diffie-Hellman initialization failed");
#   endif /* NO_DH */
    
#   if SSLEAY_VERSION_NUMBER >= 0x00906000L
    SSL_CTX_set_mode (y->ctx, (SSL_MODE_ENABLE_PARTIAL_WRITE      |
                              SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER));
#   endif /* OpenSSL-0.9.6 */

    SSL_CTX_set_session_cache_mode (y->ctx, SSL_SESS_CACHE_SERVER);
    SSL_CTX_sess_set_cache_size (y->ctx, y->xxl.session_cache);

    if (y->xxl.session_timeout > 0) /* openssl default: 300 secs */
        SSL_CTX_set_timeout (y->ctx, y->xxl.session_timeout);

    if (load_server_cert (y->nv, y->ctx, y->xxl.cert_name) == 0)
        return 0;
    if (load_server_key (y->nv, y->ctx, y->xxl.key_name) == 0)
        return 0;
    if (SSL_CTX_check_private_key (y->ctx) == 0) {
        sslerror ("Private server key does not match the certificate");
        return 0;
    }
    
    // /* certificate verification */
    // if (verify_paths) {
    //     if (verbosity & XXL_SETUP_LOG)
    //         logger (LOG_INFO, "Initializing SSL library verify paths");
    //     if (SSL_CTX_set_default_verify_paths (y->ctx) == 0) {
    //         sslerror ("X509_set_default_verify_paths");
    //         return 0;
    //     }
    // }

    if (load_ca_certs (y->ctx, y->xxl.ca_file, y->xxl.ca_dir) == 0)
        return 0;
   
    SSL_CTX_set_verify
        (y->ctx, verify_mode & ~XXL_VERIFY_USE_ONLY_MY, verify_callback);

    SSL_CTX_set_info_callback (y->ctx, info_callback);

    if (load_cipher_list (y->nv, y->ctx, y->xxl.cipher_name) == 0)
        return 0;
        
    if (get_nv_value (y->nv, &data, 0, y->xxl.sid_name [y->sid_inx]) == 0)
        data = y->xxl.sid_ctx [y->sid_inx] ;
    y->sid_ctx = zstrdup (data);
    y->sid_len = strlen  (data);

    return 1 ;
}


static
void _cty_cleanup (CTY *y) {
    RAND_cleanup () ;
    if (y->ctx) SSL_CTX_free        (y->ctx);
    if (y->sid_ctx)     free ((char*)y->sid_ctx);
}


/* ----------------------------------------------------------------------- *
 * Public: ssl object constructor/destructor
 * ----------------------------------------------------------------------- */

const char *ssl_version (void) {
    return SSLeay_version (SSLEAY_VERSION);
}


void ssl_destroy (CTY *y) {
    if (y) {
        _cty_cleanup (y);
        free (y);
    }
}


// void ssl_check_cache (CTY *y) {
//    SSL_CTX_flush_sessions (y->ctx, time (0));
// }


CTY *ssl_create (NVCL *nv, XXL_INIT *q) {

    CTY y ;  /* global SSL context */

    DMALLOC_MARK () ;

    XZERO (y);

    y.xxl = *q ;
    y.nv  = nv ;
    
    if (y.xxl.session_cache == 0)
        y.xxl.session_cache = XXL_SESSION_CACHE ;
    if (y.xxl.sid_name [0] == 0)
        y.xxl.sid_name [0] = XXL_CTX0SID_NAME ;
    if (y.xxl.sid_name [0] == 0)
        y.xxl.sid_name [0] = XXL_CTX1SID_NAME ;

    y.xxl.sid_ctx [0] = XXL_SID0_CTX ;
    y.xxl.sid_ctx [1] = XXL_SID1_CTX ;

    if (_ssl_cty_setup (&y) == 0) {
        _cty_cleanup (&y) ;
        return 0;
    }
    
    if (y.xxl.    dh_name) y.xxl.    dh_name = zstrdup (y.xxl.    dh_name);
    if (y.xxl.  cert_name) y.xxl.  cert_name = zstrdup (y.xxl.  cert_name);
    if (y.xxl.   key_name) y.xxl.   key_name = zstrdup (y.xxl.   key_name);
    if (y.xxl.cipher_name) y.xxl.cipher_name = zstrdup (y.xxl.cipher_name);
    if (y.xxl.    ca_file) y.xxl.    ca_file = zstrdup (y.xxl.    ca_file);
    if (y.xxl.    ca_dir)  y.xxl.    ca_dir  = zstrdup (y.xxl.    ca_dir);

    y.xxl.sid_name [0] = zstrdup (y.xxl.sid_name [0]);
    y.xxl.sid_name [1] = zstrdup (y.xxl.sid_name [1]);
    
    return zmemdup (&y);
}


unsigned ssl_reopen (CTY *y) {

    CTY Y ;

    DMALLOC_HUNT ();

    /* make a copy */
    Y = *y ;
    Y.ctx = 0 ;
    Y.sid_inx = (Y.sid_inx == 0) ; /* toggle */

    /* open a new session */
    if (_ssl_cty_setup (&Y) == 0) {
        _cty_cleanup (&Y) ;
        return 0;
    }

    /* close the current ssl sessions */
    _cty_cleanup (y);

    /* save the new descriptor into the old one */
    *y = Y ;

    DMALLOC_HUNT ();

    return 1;
}


/* ----------------------------------------------------------------------- *
 * Public: ssl session management
 * ----------------------------------------------------------------------- */

void ssl_finish (XXL *ss) {
    struct timeval tv ;
    fd_set readfds, writefds;
    
    ssl_flush (ss);

    switch (SSL_get_shutdown (ss->ssl)) {
    case SSL_SENT_SHUTDOWN:
    case SSL_RECEIVED_SHUTDOWN:
    case SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN:
        break ;
    default:
        /* make sure that we can send data */
        XZERO (tv);
        FD_ZERO (&readfds);
        FD_SET (ss->fd, &readfds);
        FD_ZERO (&writefds);
        FD_SET (ss->fd, &writefds);
        if (select (ss->fd + 1, &readfds, &writefds, 0, &tv) > 0 &&
            FD_ISSET (ss->fd, &readfds)                          &&
            FD_ISSET (ss->fd, &writefds)) {
            SSL_shutdown (ss->ssl);
        }
        else if (errno)
            logger (LOG_WARNING, "Error: SSL connection on fd=%d "
                    "cannot be shut down: %s", ss->fd, strerror (errno));
    }

    SSL_free (ss->ssl);
    ERR_remove_state (0);
    free (ss);

    // DMALLOC_HUNT ();
}


XXL *ssl_accept (CTY         *y,
                 int         fd,
                 ip4addr_t ipcl,
                 uint16_t  port) {
    XXL ss, *xx;
    int old_opt, n;

    XZERO (ss) ;
    ss.y  =  y ;
    ss.fd = fd ;

    // DMALLOC_MARK () ;
    
    if (verbosity & XXL_SETUP_LOG)
        logger (LOG_INFO, "connected from %s:%d - starting SSL",
                ip4addr_pp (ipcl), port);
    
    /* use non-blocking mode with ssl */
    if ((old_opt = fcntl (fd, F_GETFL, 0)) < 0 ||
        fcntl (ss.fd, F_SETFL, old_opt|O_NONBLOCK)) {
        logger (LOG_INFO,
                "Error setting non-blocking mode for SSL connection "
                "from %s:%d:", ip4addr_pp (ipcl), port, strerror (errno));
        return 0;
    }
    
    /* set global session start/call back parameters */
    verify_depth = y->xxl.verify_depth ;
    verify_mode  = y->xxl.verify_mode  ;

    if ((ss.ssl = SSL_new (y->ctx)) == 0) {
        sslerror ("SSL_new");
        return 0;
    }

    // DMALLOC_HUNT ();

    /* Set blinding iff it's not built into our OpenSSL version */
#   if SSLEAY_VERSION_NUMBER < 0x0090702fL
    set_rsa_blinding (ss.ssl);
#   else
    if (verbosity & XXL_SETUP_LOG)
        logger (LOG_DEBUG, "Relying on OpenSSL RSA Blinding.");
#   endif

#   if SSLEAY_VERSION_NUMBER >= 0x0922
    SSL_set_session_id_context
        (ss.ssl, (const unsigned char*)y->sid_ctx, y->sid_len);
#   endif

    // DMALLOC_HUNT ();

    SSL_set_fd (ss.ssl, ss.fd);
    SSL_set_accept_state (ss.ssl);

    // DMALLOC_HUNT ();

    accept_count = 0 ; /* global, accessible by call back function */
    while ((n = SSL_accept (ss.ssl)) <= 0) {
        
        if (n < 0 && accept_count ++ < XXL_REPEAT_ACCEPT)
            switch (ss.ssl_pending = SSL_get_error (ss.ssl, n)) {
            case SSL_ERROR_WANT_READ:
                if (verbosity & XXL_SETUP_LOG)
                    logger (LOG_INFO, "SSL_accept on fd=%d needs read - "
                            "delaying", ss.fd);
                return zmemdup (&ss) ;
                
            case SSL_ERROR_WANT_WRITE:
                if (verbosity & XXL_SETUP_LOG)
                    logger (LOG_INFO, "SSL_accept on fd=%d needs write - "
                            "delaying", ss.fd);
                SSL_write (ss.ssl, "", 0);
                continue ;
            }

        sslerror ("SSL_accept");
        SSL_set_shutdown (ss.ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
        SSL_free (ss.ssl);
        ERR_remove_state (0);
        return 0;
    }

    ssl_accept_ok (xx = zmemdup (&ss));

    return xx ;
}


const char *ssl_client_dn (XXL *ss) {
    return ss->dn ? ss->dn : "<unknown>" ;
}


const char *ssl_strerror (XXL *unused) {
    static char str [120];
    long err = ERR_get_error ();
   
    if (err) {
        ERR_error_string (err, str);
        return str;
    }
    return strerror (errno);
}


int ssl_pending (XXL *ss) {
    return ss->d_len;
}


void ssl_flush (XXL *ss) {
    while (ss->delayed && (ssl_send (ss, 0, 0) < 0 && errno == EAGAIN))
        ;
}


SESS_INFO *ssl_info (CTY *y) {
    load_session_info (&y->info, y->ctx);
    y->info.depth = y->xxl.session_cache ;
    return &y->info ;
}

/* ----------------------------------------------------------------------- *
 * Public: ssl send/receive replacements
 * ----------------------------------------------------------------------- */

int ssl_recv (XXL *ss, void *buf, int len) {
    int n, r ;

    if (ss->ssl_pending)
        return ssl_accept_continue (ss, buf, len);

    n = SSL_read (ss->ssl, buf, len);
    switch (r = SSL_get_error (ss->ssl, n)) {
    case SSL_ERROR_NONE:
        if (verbosity & XXL_TRAFFIC_LOG)
            logger (LOG_INFO, "SSL_read: %d bytes from fd=%d", n, ss->fd);
        return n ;

    case SSL_ERROR_ZERO_RETURN:
        if (verbosity & XXL_TRAFFIC_LOG)
            logger (LOG_INFO, "SSL_read: shutdown on fd=%d", ss->fd);
        SSL_shutdown (ss->ssl); /* Send close_notify back */
        return 0;

    case SSL_ERROR_WANT_X509_LOOKUP:
    case SSL_ERROR_WANT_READ:
    case SSL_ERROR_WANT_WRITE:
        if (verbosity & XXL_TRAFFIC_LOG)
            logger (LOG_INFO, "SSL_read on fd=%d postponed, reason: %s",
                    ss->fd, ssl_reason_text (r));
        errno = EAGAIN ;
        return -1;

    case SSL_ERROR_SYSCALL:
        if (n < 0) { /* not EOF */
            if (errno == EINTR) {
                logger (LOG_WARNING,
                        "SSL read interrupted by a signal - retrying");
                errno = EAGAIN ;
                return -1;
            }
            logger (LOG_WARNING, "Error: SSL_read: %s", strerror (errno));
            SSL_set_shutdown
                /* do not call shut down when the conection has died */
                (ss->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
            return -1;
        }
        if (verbosity & XXL_TRAFFIC_LOG)
            logger (LOG_INFO, "SSL socket closed on SSL_read");
        SSL_set_shutdown
            /* do not call shut down when the conection has died */
            (ss->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
        return 0;
        
    case SSL_ERROR_SSL:
        sslerror ("SSL_read");
        /* return -1; */
    }
    
    return -1;
}


int ssl_send (XXL *ss, void *buf, int buflen) {
    int n, cleanup = 0, len = buflen, repeat = 0 ;
    
    if (ss->delayed) {
        if (len) {
            /* append to send buffer */
            ss->delayed = (char*) zrealloc (ss->delayed, ss->d_len + len);
            memcpy (ss->delayed + ss->d_len, buf, len) ;
        }
        buf  = ss->delayed;
        len += ss->d_len;
        ss->delayed = 0 ;
        ss->d_len   = 0 ;
        cleanup     = 1 ;
    }

    if (len == 0) {
        errno = EAGAIN ;
        return -1 ;
    }

    for (;;) {
        if ((n = SSL_write (ss->ssl, buf, len)) > 0) {
            if (n < len) {
                /* spool and send it later on */
                ss->d_len   = len - n;
                ss->delayed = cleanup
                    ? zrealloc (memmove (buf, buf + n, ss->d_len), ss->d_len)
                    : memcpy (xmalloc (ss->d_len), buf + n, ss->d_len)
                    ;
            }
	    /*
            if (verbosity & XXL_TRAFFIC_LOG) {
                char *p = memcpy (xmalloc (n + 1), buf, n);
                p [n] = '\0' ;
                safestring (p) ;
            	if (verbosity & XXL_TRAFFIC_LOG)
                	logger (LOG_DEBUG, 
				"SSL_write: sent %d bytes: \"%s\"", n, p);
                free (p);
            }
	    */
            if (buflen == 0) {
                errno = EAGAIN ;
                return -1 ;
            }
            return buflen ;
        }

        switch (SSL_get_error (ss->ssl, n)) {
        case SSL_ERROR_ZERO_RETURN: /* close_notify received */
            if (verbosity & XXL_TRAFFIC_LOG)
                logger (LOG_INFO, "SSL_write: shutdown on fd=%d", ss->fd);
            if (cleanup) /* clean up, we have allocated this space */
                free (buf);
            return 0 ;
       
        case SSL_ERROR_WANT_X509_LOOKUP: /* busy - send it later on */
        case SSL_ERROR_WANT_READ:
        case SSL_ERROR_WANT_WRITE:
            if (++ repeat < 500) {
                /* suspend some time when flushing the cache */
                struct timeval tv = {0, XXL_WAIT_REWRITE * 1000};
                if (verbosity & XXL_TRAFFIC_LOG)
                    logger (LOG_INFO, "SSL_write delayed on fd=%d (%d)",
                            ss->fd, repeat);
                select (0, 0, 0, 0, &tv);
                continue ;
            }
            else if (verbosity & XXL_TRAFFIC_LOG)
                logger (LOG_INFO, "SSL_write on fd=%d postponed", ss->fd);
            ss->d_len   = len;
            ss->delayed = cleanup
                ? buf
                : memcpy (xmalloc (len), buf, len)
                ;
            if (buflen == 0) {
                /* suspend some time when flushing the cache */
                struct timeval tv = {0, XXL_WAIT_REWRITE * 1000};
                select (0, 0, 0, 0, &tv);
                errno = EAGAIN ;
                return -1 ;
            }
            return buflen ;
        case SSL_ERROR_SYSCALL:
            if (errno == EINTR) {
                logger (LOG_WARNING,
                        "SSL_write interrupted by a signal - retrying");
                errno = EAGAIN ;
            }
            else
                logger (LOG_WARNING, "Error: SSL_write: %s", strerror (n));
            break ;
        
        case SSL_ERROR_SSL:
            sslerror ("SSL_write");
            /* break; */
        }

        if (cleanup) /* clean up, we have allocated this space */
            free (buf);
        return -1;
    }

    /* NOTREACHED */
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
#endif /* HAVE_OPENSSL */
