/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Henry Margies <h.margies@maxina.de> - Maxina GmbH, Neu-       *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "hlst.h"
#include "util.h"
#include "http-plugin.h"
#include "parser.h"
#include "nvalloc.h"
#include "plugin-util.h"

static void get_time(HLST* avp, NVCL* nvram)
{
	time_t t = time(NULL);

	add_to_list(avp, "status.time_cur", ctime(&t));

	add_to_list(avp, "status.time_type",
			get_from_nvram(nvram, "sys_time.time"));
}

static void get_version(HLST* avp)
{
	char version[1024];
	FILE* f = fopen("/etc/version", "r");

	if (f) {
		fgets(version, sizeof(version), f);
		fclose(f);
	} else {
		strncpy(version, "Error", sizeof(version));
	}

	add_to_list(avp, "status.version", version);
}


static void get_phone(HLST* avp, NVCL* nvram)
{
	char name[128], number[128], type[128];
	char* id = get_from_nvram(nvram, "phone_dials.standardaccount");

	snprintf(name  , sizeof(name),   "phone_accounts.name.%s", id);
	snprintf(number, sizeof(number), "phone_accounts.number.%s", id);
	snprintf(type, sizeof(type), "phone_accounts.type.%s", id);

	add_to_list(avp, "status.phone_name",  get_from_nvram(nvram, name));
	add_to_list(avp, "status.phone_number",get_from_nvram(nvram, number));
	add_to_list(avp, "status.phone_type",  get_from_nvram(nvram, type));
}


static void get_internet(HLST* avp, NVCL* nvram)
{
	char* status = get_from_nvram(nvram, "wan_dsl.carrier_state");

	if (!strncmp("Con", status, 3))
		add_to_list(avp, "status.net_stat", "connected");
	else if (!strncmp("\"Con", status, 4))
		add_to_list(avp, "status.net_stat", "connected");
	else if (!strncmp("Not", status, 3))
		add_to_list(avp, "status.net_stat", "not_connected");
	else if (!strncmp("\"Not", status, 4))
		add_to_list(avp, "status.net_stat", "not_connected");
	else 
		add_to_list(avp, "status.net_stat", "waiting");

	add_to_list(avp, "status.net_ip",
			get_from_nvram(nvram, "wan_dsl.ip_address.1"));
		
	add_to_list(avp, "status.net_local_ip",
			get_from_nvram(nvram, "lan_ip.ipAddress"));
		
}


static void get_wlan(HLST* avp, NVCL* nvram)
{
	add_to_list(avp, "status.ap_stat", 
			get_from_nvram(nvram, "lan_wlan.isAccesspoint"));
	
	add_to_list(avp, "status.ap_ssid", 
			get_from_nvram(nvram, "lan_wlan.ssid"));
	
	add_to_list(avp, "status.ap_security", 
			get_from_nvram(nvram, "lan_wlan.security"));
}


/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

void *htinit (const char *doc_path, NVCL *nv) 
{
	STATE *s = ZMALLOC (STATE);

	s->nvram  = nv;
	s->htpath = doc_path;

	return s;
}


void htcleanup (void *desc) 
{
	STATE *s = desc ;
	free(s);
}

void htrun (void *desc, RPLY_REC *r, int argc, const char **argv) 
{
	STATE	*s = desc;
	PARSER	*parser;
	HLST	*avp;
	FILE*	pipe;
	const char *data;
	const char* html = argv[1];

	if (argc < 2)
		return;

	avp = http_args(r);

	handle_global_settings(avp, s->nvram);
    
	parser = open_parser(avp, s->nvram, s->htpath);

	if (!strcmp(argv[0], "/")) {
		
		char* firsttime = get_from_nvram(s->nvram, "global.firsttime");

		if (!strcmp(firsttime, "true"))
			html = "welcome.html";

	}
	
	if (file_parser(parser, html) == 0)
		return;

	get_version(avp);
	get_time(avp, s->nvram);
	get_phone(avp, s->nvram);
	get_internet(avp, s->nvram);
	get_wlan(avp, s->nvram);

	count_types(s, avp, "phone_accounts");
	count_types(s, avp, "phone_devices");

	/* parser */
	run_parser(parser);

	/* print result */
	if ((data = data_collect(parser)) != 0)
		body_append(r, data, 0);

	/* clean up */
	file_parser_end(parser);
	close_parser(parser);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
