/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Susanne Goldammer <s.goldammer@maxina.de> - Maxina GmbH,      *
*   Neu-Isenburg, Germany						  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>

#include "hlst.h"
#include "util.h"
#include "http-plugin.h"
#include "parser.h"
#include "nvalloc.h"
#include "plugin-util.h"
#include "logger.h"


/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

#define INVALID_SOURCE_START_IP				CUSTOM_START_ERROR + 10
#define INVALID_SOURCE_END_IP				CUSTOM_START_ERROR + 11
#define INVALID_SOURCE_ADDRESS_RANGE		CUSTOM_START_ERROR + 12

#define INVALID_DESTINATION_START_IP		CUSTOM_START_ERROR + 20
#define INVALID_DESTINATION_END_IP			CUSTOM_START_ERROR + 21
#define INVALID_DESTINATION_ADDRESS_RANGE	CUSTOM_START_ERROR + 22

#define INVALID_SOURCE_START_PORT			CUSTOM_START_ERROR + 30
#define INVALID_SOURCE_END_PORT				CUSTOM_START_ERROR + 31
#define INVALID_SOURCE_PORT_RANGE			CUSTOM_START_ERROR + 32

#define INVALID_DESTINATION_START_PORT		CUSTOM_START_ERROR + 40
#define INVALID_DESTINATION_END_PORT		CUSTOM_START_ERROR + 41
#define INVALID_DESTINATION_PORT_RANGE		CUSTOM_START_ERROR + 42


/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */


int checkSourceIp(HLST *avp, HLST *nvram_avp, const char *varName)
{
	/* test whether filter is for any Source ip or not*/

	unsigned long ret = 0;
	struct in_addr start, end;
	int error = NO_ERROR;
	char **val = (char**)find_hlst(avp, varName, 0);

	if(strncmp(*val,"specified",strlen("specified"))==0)
	{
		/* first check starting ip */
		char **sourceFrom;
		/* second check ending ip */
		char **sourceTo;

		sourceFrom = (char**)find_hlst(avp,"wan_outbound.sourceFrom",0);
		if(sourceFrom)
		{
			ret = validIpAddr(sourceFrom);
			if(!ret)
				error = INVALID_SOURCE_START_IP;
		}
		else
			error = INVALID_SOURCE_START_IP;

		

		sourceTo = (char**)find_hlst(avp,"wan_outbound.sourceTo",0);
		if(sourceTo)
		{
			ret = validIpAddr(sourceTo);
			if(!ret)
				error = INVALID_SOURCE_END_IP;
		}
		else
			error = INVALID_SOURCE_END_IP;

		//at last check if the start ip is smaller than the end ip
		inet_aton(*sourceFrom, &start );
		inet_aton(*sourceTo, &end);
		if(start.s_addr > end.s_addr)
			error = INVALID_SOURCE_ADDRESS_RANGE;
	}
	else
	{
		delete_hlst(avp,"wan_outbound.sourceFrom",0);
		delete_hlst(avp,"wan_outbound.sourceTo",0);
	}


	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}

int checkDestIp(HLST *avp, HLST *nvram_avp, const char *varName)
{
	unsigned long ret = 0;
	struct in_addr start, end;
	int error = NO_ERROR;
	char **val = (char**)find_hlst(avp, varName, 0);

	/* test whether filter is for any Source ip or not*/

	if(strncmp(*val,"specified",strlen("specified"))==0)
	{
		/* first check starting ip */
		char **destFrom;

		destFrom = (char**)find_hlst(avp,"wan_outbound.destinationFrom",0);
		if(destFrom)
		{
			ret = validIpAddr(destFrom);
			if(!ret)
				error = INVALID_DESTINATION_START_IP;
		}
		else
			error = INVALID_DESTINATION_START_IP;

		/* second check ending ip */
		char **destTo;

		destTo = (char**)find_hlst(avp,"wan_outbound.destinationTo",0);
		if(destTo)
		{
			ret = validIpAddr(destTo);
			if(!ret)
				error = INVALID_DESTINATION_END_IP;
		}
		else
			error = INVALID_DESTINATION_END_IP;

		/* check if startIp is smaller or equal to endIp */
		ret = inet_aton(*destFrom, &start);
		ret = inet_aton(*destTo, &end);
		if(start.s_addr > end.s_addr)
			error = INVALID_DESTINATION_ADDRESS_RANGE;
	}
	else
	{
		delete_hlst(avp,"wan_outbound.destinationFrom",0);
		delete_hlst(avp,"wan_outbound.destinationTo",0);
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}

int checkSourcePort(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char **val = (char**)find_hlst(avp, varName, 0);

	if(strncmp(*val,"specified",strlen("specified"))==0)
	{
		long start,end;					//numeric values for start and end port
		char **startPort, **endPort ;
		char *s;

		//check for startPort
		startPort = (char**)find_hlst(avp,"wan_outbound.sourcePortFrom",0);
		if(startPort)
		{
			start = strtol(*startPort,&s,10);

			if(strcmp(*startPort,"") == 0 || (*s != '\0') || start<0 || start>65535 )
				error = INVALID_SOURCE_START_PORT;
		}


		//check for endPort
		endPort = (char**)find_hlst(avp,"wan_outbound.sourcePortTo",0);
		if(endPort)
		{
			end = strtol(*endPort,&s,10);

			if(strcmp(*endPort,"") == 0 || (*s != '\0') || end<0 || end>65535 )
				error = INVALID_SOURCE_END_PORT;
		}

		if(strcmp(*startPort,"") && strcmp(*endPort,"") && start > end)	//startPort should be smaller or equal to endPort
			error = INVALID_SOURCE_PORT_RANGE;
	}
	else
	{
		delete_hlst(avp,"wan_outbound.sourcePortFrom",0);
		delete_hlst(avp,"wan_outbound.sourcePortTo",0);
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}

int checkDestPort(HLST *avp, HLST *nvram_avp, const char *varName)
{
	int error = NO_ERROR;
	char **val = (char**)find_hlst(avp, varName, 0);

	if(strncmp(*val,"specified",strlen("specified"))==0)
	{
		long start,end;					//numeric values for start and end port
		char **startPort, **endPort ;
		char *s;

		//check for startPort
		startPort = (char**)find_hlst(avp,"wan_outbound.destinationPortFrom",0);
		if(startPort)
		{
			start = strtol(*startPort,&s,10);

			if(strcmp(*startPort,"") == 0 || (*s != '\0') || start<0 || start>65535 )
				error = INVALID_DESTINATION_START_PORT;
		}


		//check for endPort
		endPort = (char**)find_hlst(avp,"wan_outbound.destinationPortTo",0);
		if(endPort)
		{
			end = strtol(*endPort,&s,10);

			if(strcmp(*endPort,"") == 0 || (*s != '\0') || end<0 || end>65535 )
				error = INVALID_DESTINATION_END_PORT;
		}

		if(start > end)	//startPort should be smaller or equal to endPort
			error = INVALID_DESTINATION_PORT_RANGE;
	}
	else
	{
		delete_hlst(avp,"wan_outbound.destinationPortFrom",0);
		delete_hlst(avp,"wan_outbound.destinationPortTo",0);
	}

	if(error != NO_ERROR)
		handle_varError(avp, error);

	return error;
}

HTMLOBJECT *init_htmlVars(){
	int max_size = 1;
	int i=0;
	int max_index = max_size - 1;

	HTMLOBJECT *p = (HTMLOBJECT *)xmalloc(sizeof(HTMLOBJECT) * max_size);
	
	//sentinel
	p[i>max_index ? max_index : i].revise 	= NULL;
	p[i>max_index ? max_index : i].name		= NULL;

	return p;
}

void free_htmlVars(HTMLOBJECT *p){
	free(p);
	p = NULL;
}


void handle_varError(HLST *avp, int ecode){

 
	switch(ecode){

	default	:		    	       	//UNKNOWN error code, nothing to do
						logger (LOG_WARNING, "UNKNOWN error code: %d: ", ecode);
						return;
	}

	addToList(avp, "error", "1");
    
	return;
}


void *htinit (const char *doc_path, NVCL *nv) {
    STATE *s = ZMALLOC (STATE) ;

	s->htmlvars.objects = init_htmlVars();
	s->htmlvars.load_on_success 		= NULL;
	s->htmlvars.total_elements_label 	= NULL;
	s->htmlvars.group_label 		= NULL;

	s->nvram  	= nv ;
	s->htpath 	= doc_path ;
	return s ;
}

void htcleanup (void *desc) {
	STATE *s = desc ;

	free_htmlVars(s->htmlvars.objects);
	free(s);
}

void htrun (void *desc, RPLY_REC *r, int argc, const char **argv) {
	STATE  *s = desc ;
	PARSER   *parser ;
	HLST        *avp ;
	const char *data ;

	if (argc < 2)
		return ;

	avp = http_args (r);

	if(query_hlst_size(avp) == 0)
	{
		load_httpvars(s, avp);
	}
	
        
	handle_global_settings(avp, s->nvram);

	parser = open_parser (avp, s->nvram, s->htpath) ;

	if (file_parser (parser, htdocs_p (r, argv [1])) == 0)
		return ;

	/* parser */
	run_parser (parser) ;

	avdump (avp); /* debugging */

	/* print result */
	if ((data = data_collect (parser)) != 0)
		body_append (r, data, 0);

	/* clean up */
	file_parser_end (parser);
	close_parser (parser) ;
}


/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

