/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: flash-file.c 2658 2006-06-23 07:02:48Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "util.h"
#include "buffer.h"
#include "logger.h"
#include "flash.h"


#define FLASH_FILE "/tmp/.flash-file"

/* ----------------------------------------------------------------------- *
 * Private varables and definitions
 * ----------------------------------------------------------------------- */

/* need automatically growing read buffer */
static IOBUF iobuf ;

#define MAGIC "NVRM"

typedef struct {
	char magic [sizeof (MAGIC) - 1];
	unsigned long size;
} __attribute__((__packed__)) header_t;

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */
size_t flash_write (const char *s, size_t len) {

    int fd, n, ret;
    header_t header;

    header.size = htonl (len);
    memcpy (header.magic, MAGIC, sizeof (header.magic));
    len += sizeof(header);
    
    /* open device */
    fd = open(FLASH_FILE, O_WRONLY|O_CREAT|O_TRUNC);

    if (fd < 0) {
        logger(LOG_ERR, "could not open flash file "
               FLASH_FILE " for writing: %s", strerror(errno));
        return 0;
    }

    /* write */
    ret = write (fd, &header, sizeof (header));
    if (ret != sizeof (header)) {
        logger(LOG_ERR, "error writing to flash: %s", strerror (errno));
        close (fd);
        return 0;
    }
    len -= sizeof (header);

    n = len;
    while (n) {

        ret = write (fd, s, n);

        if (ret <= 0) {
            logger(LOG_ERR, "error writing to flash: %s", strerror (errno));
            len = 0;
            break;
        }

        n -= ret;
        s += ret;
    }

    close (fd);
    return len ;
}
 

void flash_bufclear (void) {
    kill_space (&iobuf);
}


const char *flash_read (void) {
    int fd, ret;
    unsigned size ;
    header_t header;

    fd = open (FLASH_FILE, O_RDONLY);

    if (fd< 0) {
        logger (LOG_ERR, "cannot open flash file: %s",
                strerror (errno));
        return 0;
    }

    ret = read (fd, &header, sizeof (header));
    if (ret != sizeof (header)) {
        logger (LOG_ERR, "could not read size from file: %s",
                strerror (errno));
        close (fd);
        return 0;
    }
    
    if (memcmp (header.magic, MAGIC, sizeof (header.magic))) {
        logger (LOG_ERR, "flash not valid, maybe reading it "
                "for the first time?");
        close(fd);
        return 0;
    }
    size = ntohl (header.size);

    kill_space (&iobuf);
    file_append (&iobuf, fd, size);
    
    if (iobuf.top < size) {
        logger (LOG_ERR, "error reading flash file: %s",
                errno ? strerror (errno) : "short read");
        kill_space (&iobuf);
    }

    /* add double \0\0 terminator */
    char_append (&iobuf, '\0');
    char_append (&iobuf, '\0');

    close (fd);
    return iobuf.data;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
