/*
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *   There is no permission to make any digital or printed copies of part or all
 *   of this work without the permission of D-Link.
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: flash-horst.c 2658 2006-06-23 07:02:48Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "util.h"
#include "buffer.h"
#include "logger.h"
#include "flash.h"

#include <sys/ioctl.h>
#include <mtd/mtd-user.h>

/* ----------------------------------------------------------------------- *
 * Private varables and definitions
 * ----------------------------------------------------------------------- */

/* need automatically growing read buffer */
static IOBUF iobuf ;

#define MAGIC "NVRM"

typedef struct {
	char magic [sizeof (MAGIC) - 1];
	unsigned long size;
} __attribute__((__packed__)) header_t;

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

static int open_mtd_by_name(const char* name, int flags) {

    int ret, nr, size, erase_size;
	char string[1024];
	FILE* mtd;
	int fd = -1;

	mtd = fopen("/proc/mtd", "r");
	if (!mtd)
		return -1;
	
	/* skip first line */
	fgets(string, sizeof(string), mtd);
	
	while(1) {
		
		ret = fscanf(mtd, "mtd%d: %x %x \"%[^\"]\"\n", &nr, &size, 
                     &erase_size, string);

		if (ret != 4)
			break;
		
		if (strcmp(string, name))
			continue;


		snprintf(string, sizeof(string), "/dev/mtd/%d", nr);
		fd = open(string, flags);
		if (fd != -1)
			break;
			
		snprintf(string, sizeof(string), "/dev/mtd%d", nr);
		fd = open(string, flags);
		break;
	}

	fclose(mtd);

	return fd;
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */
size_t flash_write (const char *s, size_t len) {

    struct mtd_info_user mtd;
    struct erase_info_user erase;
    int fd, n, ret;
    header_t header;

    header.size = htonl (len);
    memcpy (header.magic, MAGIC, sizeof (header.magic));
    len += sizeof(header);
    
    /* open device */
    fd = open_mtd_by_name("nvram", O_WRONLY | O_SYNC);

    if (fd < 0) {
        logger(LOG_ERR, "could not open flash file: %s", strerror(errno));
        return 0;
    }

    /* get device info */
    ret = ioctl(fd, MEMGETINFO, &mtd);
    if (-1 == ret) {
        logger(LOG_ERR,
               "could not get mtd info, is this a mtd device?: %s",
               strerror (errno));
        close(fd);
        return 0;
    }
    // printf("mtd.size = %d, mtd.erasesize = %d\n",
    //        mtd.size, mtd.erasesize);
    
    /* check whether there is enough space */
    if (mtd.size < len) {
        logger(LOG_ERR,
               "mtd device too small, want to write %d bytes, size is %d",
               len, mtd.size);
        close(fd);
        return 0;
    }

    /* calculate blocks to erase */
    erase.start = 0;
    erase.length= len & ~(mtd.erasesize - 1);
    if (len % mtd.erasesize) 
        erase.length += mtd.erasesize;
    
    //printf("erase from %d, %d bytes\n", erase.start, erase.length);

    /* erase */
    ret = ioctl(fd, MEMERASE, &erase);
    if (ret < 0) {
        logger(LOG_ERR, "could not erase flash device: %s",
               strerror (errno));
        close(fd);
        return 0;
    }
    
    /* write */
    ret = write (fd, &header, sizeof (header));
    if (ret != sizeof (header)) {
        logger(LOG_ERR, "error writing to flash: %s", strerror (errno));
        close (fd);
        return 0;
    }
    len -= sizeof (header);

    n = len;
    while (n) {

        ret = write (fd, s, n);

        if (ret <= 0) {
            logger(LOG_ERR, "error writing to flash: %s", strerror (errno));
            len = 0;
            break;
        }

        n -= ret;
        s += ret;
    }

    close (fd);
    return len ;
}
 

void flash_bufclear (void) {
    kill_space (&iobuf);
}


const char *flash_read (void) {
    int fd, ret;
    unsigned size ;
    header_t header;

    fd = open_mtd_by_name ("nvram", O_RDONLY);

    if (fd< 0) {
        logger (LOG_ERR, "cannot open flash file: %s",
                strerror (errno));
        return 0;
    }

    ret = read (fd, &header, sizeof (header));
    if (ret != sizeof (header)) {
        logger (LOG_ERR, "could not read size from file: %s",
                strerror (errno));
        close (fd);
        return 0;
    }
    
    if (memcmp (header.magic, MAGIC, sizeof (header.magic))) {
        logger (LOG_ERR, "flash not valid, maybe reading it "
                "for the first time?");
        close(fd);
        return 0;
    }
    size = ntohl (header.size);

    kill_space (&iobuf);
    file_append (&iobuf, fd, size);
    
    if (iobuf.top < size) {
        logger (LOG_ERR, "error reading flash file: %s",
                errno ? strerror (errno) : "short read");
        kill_space (&iobuf);
    }

    /* add double \0\0 terminator */
    char_append (&iobuf, '\0');
    char_append (&iobuf, '\0');

    close (fd);
    return iobuf.data;
}


/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
