/*
 * nvramd - active data base server and flash memory cache
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: comhandler.c 3837 2007-03-13 10:44:25Z hholtschneider $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "util.h"
#include "ctrl.h"
#include "varstr.h"
#include "preload.h"
#include "settings.h"
#include "nvalloc.h"
#include "logger.h"

#define __COMHANDLER_PRIVATE
#include "comhandler.h"

/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
unsigned load_defaults (COMM *o, int log, int io_mode) {

    if (preload_defaults (o->list, -1, io_mode)) {
        if ((verbosity & COMM_LOG_DEFLOAD) && log)
            logger (LOG_INFO, io_mode
                    ? "The data from the flash has been loaded."
                    : "Default data has been initialised, only.");
        return 1;
    }

    if ((verbosity & COMM_LOG_DEFLOAD) && log)
        logger (LOG_ERR, "Error reading defaults from the flash");
    return 0;
}


static
unsigned privileged (COMM *o) {
#   ifndef DONT_USE_RECVMSG
    return
        (ctrl_uid (o->ctrl) == 0 && ctrl_gid (o->ctrl) == 0) ||
        (ctrl_gid (o->ctrl) == o->admin) ;
#   else
    return 1;
#   endif
}

/* ----------------------------------------------------------------------- *
 * Public: Connectionless communication handlers
 * ----------------------------------------------------------------------- */

COMM *comm_init (LIST *l, gid_t admin) {

    COMM *o = ZMALLOC (COMM);   

    if ((o->list = l) == 0) {
#       ifdef DISCARD_INFO
        logger (LOG_ERR, "comhandler: AVP subsytem not initialised!");
#       endif
        return 0;
    }

    o->admin = admin ;

    return o ;
}


void comm_close (COMM *o) {
    
    if (o) {
        list_close (o->list);
        free (o);
    }
}


COMM *comm_update (COMM *o, CTRL *c, int defaults) {
    o->started = time (0);
    o->ctrl    = c ;
    preload_functions (o->list) ;
    if (defaults)
        load_defaults (o, 1, COMM_DEFAULT_LOAD_MODE);
    return o;
}


void comm_store (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc != 1) goto usage;

    ctrl_reply (o->ctrl, list_store (o->list)
                ? "+OK\n"
                : "+ERR: problem writing flash\n");
    return;
 
 usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, "%s -- write data to flash\n", me);
}


void comm_load (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    int mode ;

    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 2) goto usage;

    if (argc == 2) {
        unsigned u ;
        if (parse_unsigned (&u, argv [1], 0) == 0) goto usage;
        mode = u ;
    }
    else
        mode = COMM_DEFAULT_LOAD_MODE;

    if (load_defaults (o, 0, mode))
        ctrl_reply (o->ctrl, "+OK\n");
    else
        ctrl_reply (o->ctrl, "+ERR: problem reading flash\n");
        
    return;
 
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, ("%s [mode] -- load data from flash"
#                         ifndef DISCARD_INFO
                          " (default = 0x%02x)"
#                         endif /* DISCARD_INFO */
                          "\n"), me
#               ifndef DISCARD_INFO
                , COMM_DEFAULT_LOAD_MODE
#               endif /* DISCARD_INFO */
        );
#   ifndef DISCARD_INFO
#   define LEFT_SPACE ".               "
    ctrl_reply
        (o->ctrl,
         ("load mode-bits: 0x%02x -- nvram/flash data only (no defaults)\n"
          LEFT_SPACE      "0x%02x -- overwrite default flags\n"
          LEFT_SPACE      "0x%02x -- overwrite default values\n"
          LEFT_SPACE      "0x%02x -- overwrite default trigger\n"
          LEFT_SPACE      "0x%02x -- overwrite volatile variables\n"
          LEFT_SPACE      "0x%02x -- apply templates to nvram/flash vars\n"),
         import_clear_ok,    import_flags_ok,
         import_value_ok,    import_trigger_ok,
         import_volatile_ok, import_template_ok);
#   undef LEFT_SPACE
#   endif /* DISCARD_INFO */
}


void comm_slots (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];

    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 2) goto usage;
    
    ctrl_reply (o->ctrl, "+OK:\nused-slots=%u\nnum-slots=%u\n",
                list_used_slots (o->list),
                list_num_slots (o->list));
    return;
 
 usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, "%s -- print number of variable slots\n", me);
}


void comm_get (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *c, *d, *f, *i, *me = argv [0];
    unsigned u ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 2:
        f = COMM_DEF_LIST_INX_FLAGS ;
        i = argv [1] ;
        break ;
    case 3:
        f = argv [1] ;
        i = argv [2] ;
        break ;
    default:
        goto usage;
    }
    
    if (parse_unsigned (&u, i, 0) == 0) goto usage;

    if ((c = list_get_var (o->list, privileged (o), f, u)) == (char*)-1) {
        ctrl_reply (o->ctrl, "-ERR: slot filtered: %s\n", i);
    }
    else if (c && (d = list_get_val (o->list, privileged (o), c, -1)) != 0)
        ctrl_reply (o->ctrl, "+OK:\n%s=%s\n", c, d);
    else
        ctrl_reply (o->ctrl, "-ERR: no such slot: %s\n", i);

    return ;
        
    ctrl_reply (o->ctrl, "+OK\n");
    return ;
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl,
                "%s [[~]filter-flags] slot -- print variable=value\n", me);
#   ifndef DISCARD_INFO
    ctrl_reply (o->ctrl,
                ".   default filter-flags: " COMM_DEF_LIST_INX_FLAGS "\n");
#   endif /* DISCARD_INFO */
}



void comm_set (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *var, *val, *me = argv [0];
    pid_t pid = -1;
    unsigned u ;
    int n, avinx = 1;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 2:
        goto display ;

    case 3:
        if (argv [1] [0] != '#')
            goto avpair ;
        if (parse_unsigned (&u, argv [1] + 1, 0) == 0)
            goto usage ;

        pid   = u ;
        avinx = 2 ;

    display:
        var = argv [avinx] ;
        if ((val = list_get_val (o->list, privileged (o), var, pid)) != 0)
            ctrl_reply (o->ctrl, "+OK:\n%s=%s\n", var, val);
        else
            ctrl_reply (o->ctrl, "-ERR: no such variable: %s\n", var);
        return ;

    case 4:
        if (argv [1] [0] != '#' || parse_unsigned (&u, argv [1] + 1, 0) == 0)
            goto usage ;

        pid   = u ;
        avinx = 2 ;

    avpair:
        var = argv [avinx] ;
        if ((n = list_avp_replace (o->list, privileged (o),
                                   var, argv [avinx + 1], pid, 0)) != 0)
            ctrl_reply (o->ctrl, "+OK: %u\n", n);
        else
            ctrl_reply (o->ctrl, "-ERR error writing variable %s\n", var);
        return ;
    }
        
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply
        (o->ctrl, "%s [#pid] variable [value] -- set/print value\n", me);
}

    
void comm_del (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *var, *me = argv [0];
    pid_t pid = -1;
    unsigned u ;
    int n, avinx = 1;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 2:
        goto delete ;

    case 3:
        if (argv [1] [0] == '#') {
            if (parse_unsigned (&u, argv [1] + 1, 0) == 0)
                goto usage ;
            pid   = u ;
            avinx = 2 ;
        }
       
    delete:
        var = argv [avinx] ;
        if ((n = list_avp_replace (o->list, privileged (o),
                                   var, 0, pid, 0)) != 0)
            ctrl_reply (o->ctrl, "+OK: %u\n", n);
        else
            ctrl_reply (o->ctrl, "-ERR error deleting variable %s\n", var);
        return ;
    }
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s [#pid] variable -- delete variable\n", me);
}


void comm_symlink (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    const char **A, *var, *flags;
    
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }
    
    switch (argc) {
    case 2:
        if ((A = list_get_trigger (o->list, argv [1], symlink_trigger)) == 0)
            ctrl_reply (o->ctrl, "-ERR no such symlink: %s\n", argv [1]);
        else {
            ctrl_reply (o->ctrl, "+OK:\n%s=", argv [1]);
            if (A [1] && A [1] [0])
                ctrl_reply (o->ctrl, "%s ", A [1]);
            ctrl_reply (o->ctrl, "%s\n", A [2]);
        }
        return ;
    case 3:
        flags =        0 ;
        var   = argv [2] ;
        break ;
    case 4:
        flags = argv [2] ;
        var   = argv [3] ;
        break ;
    default:
        goto usage ;
    }

    ctrl_reply (o->ctrl, "+OK: %u\n", list_lnk_replace
                (o->list, privileged (o), argv [1], flags, var, 0));
    return ;

usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s variable [[flags] variable]"
                " -- symlink to trigger\n", me);
}


void comm_rm_symlink (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    int n ;
    
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }
    if (argc != 2) goto usage;

    if ((n = list_lnk_replace
         (o->list, privileged (o), argv [1], 0, 0, 0)) != 0)
        ctrl_reply (o->ctrl, "+OK: %u\n", n);
    else
        ctrl_reply (o->ctrl, "-ERR error removing trigger symlink\n");

    return ;
        
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s variable -- remove trigger symlink\n", me);
}


void comm_trigger (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    const char *flgs = 0, *uids = 0, *gids = 0, *chrt = 0;
    unsigned u = -1, g = -1 ;
    int n ;
         
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }
    if (argc < 2) goto usage;

    if (argc == 2) {
        const char **A =
            list_get_trigger (o->list, argv [1], execute_trigger);

        if (A == 0) {
            ctrl_reply (o->ctrl, "-ERR no such trigger: %s\n", argv [1]);
            return ;
        }

        ctrl_reply (o->ctrl, "+OK:\n%s=", argv [1]);

        n = 0 ;
        if (A [1] [0])        n |= 1 ;
        if (A [2] [0] != '-') n |= 2 ;
        if (A [3] [0] != '-') n |= 4 ;
        if (A [4] [0])        n |= 8 ;
        
        if (n) {
            ctrl_reply (o->ctrl, "%s ", A [1]);
            
            if (n &   (2|4|8))
                ctrl_reply (o->ctrl, "%s ", A [2]);

            if (n &     (4|8))
                ctrl_reply (o->ctrl, "%s ", A [3]);
            
            if (n &       (8))
                ctrl_reply (o->ctrl, "%s ", A [4]);

            ctrl_reply (o->ctrl, "+ ");
        }
        
        if (A [5]) {
            ctrl_reply (o->ctrl, "%s", A [5]);
            for (n = 6; A [n]; n++)
                ctrl_reply (o->ctrl, " %s", A [n]);
        }

        ctrl_reply (o->ctrl, "\n");
        return ;
    }

    /* find delimiter */
    if (argc <= (n = 7))
        n = argc ;
    while (-- n > 2 && (argv [n] [0] != '+' || argv [n] [1]))
        ;

    switch (n) {
    case 3:
        flgs = argv [2] ;
        break ;
    case 4:
        flgs = argv [2] ;
        uids = argv [3] ;
        break ;
    case 5:
        flgs = argv [2] ;
        uids = argv [3] ;
        gids = argv [4] ;
        break ;
    case 6:
        flgs = argv [2] ;
        uids = argv [3] ;
        gids = argv [4] ;
        chrt = argv [5] ;
        break ;
    default:
        n = 1 ;
    }

    if (uids && parse_unsigned (&u, uids, 0) == 0) goto usage ;
    if (gids && parse_unsigned (&g, gids, 0) == 0) goto usage ;

    n ++ ;

    ctrl_reply (o->ctrl, "+OK: %u\n", list_exe_replace
                (o->list, privileged (o),
                 argv [1], flgs, u, g, chrt, argc - n, argv + n, 0));
    return ;
        
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s variable"
                " [[flags [uid [gid [chroot]]] +] cmd [args ...]]"
                " -- set up trigger function\n", me);
}


void comm_rm_trigger (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    int n ;
    
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }
    if (argc != 2) goto usage;

    if ((n = list_exe_replace
         (o->list, privileged (o), argv [1], 0,0,0,0,0,0,0)) != 0)
        ctrl_reply (o->ctrl, "+OK: %u\n", n);
    else
        ctrl_reply (o->ctrl, "-ERR error removing trigger function\n");

    return ;
        
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s variable -- remove trigger function\n", me);
}



void comm_access (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    const char **A ;
    
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }

    switch (argc) {
    case 2:
        if ((A = list_get_trigger (o->list, argv [1], access_trigger)) == 0)
            ctrl_reply (o->ctrl, "-ERR no such access controlled "
                        "variable: %s\n", argv [1]);
        else
            ctrl_reply (o->ctrl, "+OK:\n%s=%s\n", argv [1], A [1]);
        return ;

    case 3:
        ctrl_reply (o->ctrl, "+OK: %u\n", list_access_replace
                    (o->list, privileged (o), argv [1], argv [2], 0));
        return ;
    }

/* usage: */
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s variable [permission-flags] "
                "-- variable access control\n", me);
}


void comm_rm_access (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    int n ;
    
    if (me [0] == '?' && !me [1]) { me = argv [1]; goto help; }
    if (argc != 2) goto usage;

    if ((n = list_access_replace
         (o->list, privileged (o), argv [1], 0, 0)) != 0)
        ctrl_reply (o->ctrl, "+OK: %u\n", n);
    else
        ctrl_reply (o->ctrl,
                    "-ERR error removing variable access control\n");

    return ;
        
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl,
                "%s variable -- remove variable access control\n", me);
}

/* ----------------------------------------------------------------------- *
 * Public: HTTP communication handlers
 * ----------------------------------------------------------------------- */

#ifndef DISCARD_HTTP

void store_cmd (void *state, RPLY_REC *r) {

    if (list_store (http_rplylist (r)) == 0) {
        char *s ;
        http_rplydata  (r, s = "Error writing data to the flash");
        http_rplyerror (r, e503) ;
        logger (LOG_ERR, s);
    }
}


void load_cmd (void *state, RPLY_REC *r) {

    char *s;
    
    LIST  *l = http_rplylist (r) ;

    if (preload_defaults (l, -1, COMM_DEFAULT_LOAD_MODE) != 0)
        return ;
    
    http_rplydata  (r, s = "Error reading defaults from the flash");
    http_rplyerror (r, e503) ;
        
#   ifndef USE_ORDINARY_FILE
    if (verbosity & HTTP_LOG_DEFLOAD)
#   endif
        logger (LOG_ERR, s);
}


void slots_cmd (void *state, RPLY_REC *r) {
    
    LIST *l = http_rplylist (r) ;

    char buf [100] ;

    sprintf (buf, "used-slots=%u\r\nnum-slots=%u\r\n",
             list_used_slots (l),
             list_num_slots  (l)) ;

    data_printf (http_rplybody (r), "%s", buf);
}


void list_cmd (void *state, RPLY_REC *r) {
    
    IOBUF *b = http_rplybody (r) ;
    LIST  *l = http_rplylist (r) ;

    const char *var ;

    if (http_haveargs (r)) {
        for (;;) {
            const char *val ;
            /* get the next avpair */
            if ((var = http_shiftargs (r)) == 0)
                return ;
            data_printf (b, "%s", var);
            if ((val = list_get_val (l, -1, var, -1)) != 0)
                data_printf (b, "=%s", val);
            data_printf (b, "\r\n");
        }
    }
    else {
        int used = list_used_slots (l) ;
        int    n = list_num_slots  (l) ;

        if (n > 0)
            do {
                const char *val ;
                var = list_get_var (l, -1, COMM_DEF_LIST_INX_FLAGS, n) ;
                if (var == 0 || var == (char*)-1)
                    continue ;
                data_printf (b, "%s", var);
                if ((val = list_get_val (l, -1, var, -1)) == 0)
                    data_printf (b, "=%s", var, val);
                data_printf (b, "\r\n");
                if (-- used == 0)
                    return ;
            }
            while (-- n > 0);
    }
}


void get_cmd (void *state, RPLY_REC *r) {

    IOBUF *b = http_rplybody (r) ;
    LIST  *l = http_rplylist (r) ;

    unsigned u ;
    const char *s, *var ;

    /* get the next index */
    while ((s = http_shiftargs (r)) != 0)
        /* get the next index */
        if (parse_unsigned (&u, s, 0)                               != 0 &&
            (var = list_get_var (l,-1, COMM_DEF_LIST_INX_FLAGS, u)) != 0 &&
            (var) != (char*)-1) {
            const char *val = list_get_val (l, -1, var, -1) ;
            if (val == 0) val = "" ;
            data_printf (b, "%s=%s\r\n", var, val);
        }
}


void set_cmd (void *state, RPLY_REC *r) {
    
    IOBUF *b = http_rplybody (r) ;
    LIST  *l = http_rplylist (r) ;

    for (;;) {
        /* get the next avpair */
        const char *var = http_shiftargs (r) ;
        char *val ;
  
        if (var == 0)
            return ;

        if ((val = strchr (var, '=')) == 0) {
            /* list this value only */
            const char *cval = list_get_val (l, -1, var, -1);
            if (cval == 0) cval = "";
            data_printf (b, "%s=%s\r\n", var, cval);
            continue ;
        }

        * val ++ = '\0' ;
     
        if (list_avp_replace (l, -1, var, val, -1, 0) == 0)
            /* record the variables that could not be added */
            data_printf (b, "%s\r\n", var);
    }
}


void del_cmd (void *state, RPLY_REC *r) {
    
    IOBUF *b = http_rplybody (r) ;
    LIST  *l = http_rplylist (r) ;

    for (;;) {
        /* get the next avpair */
        const char *var = http_shiftargs (r) ;

        if (var == 0)
            return ;

        if (list_avp_replace (l, -1, var, 0, -1, 0) == 0)
            /* record the variables that could not be deleted */
            data_printf (b, "%s\r\n", var);
    }
}


void error_cmd (void *state, RPLY_REC *r) {

    const char *s;
    IOBUF *b = http_rplybody (r) ;
    
    http_rplyerror (r, e404) ;

    if ((s = http_urlmethod (r)) != 0)
        data_printf
            (b, "Unknown http command request: %s", s);
    else
        data_printf
            (b, "Error decoding internal command: <%s>", http_cmdname (r));
}

#endif /* DISCARD_HTTP */

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
