/*
 * nvramd - active data base server and flash memory cache
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: dsafunctions.c 3175 2006-09-23 19:17:11Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#ifdef HAVE_OPENSSL
#define __NVRAMD_DSA_FUNCTIONS_PRIVATE
#endif

#include "dsafunctions.h"

#ifdef HAVE_OPENSSL
#include "tvcalc.h"
#include "varstr.h"
#include "base64.h"
#include "logger.h"
#include "buffer.h"
#include "nvalloc.h"

#include <openssl/err.h>
#include <openssl/ssl.h>

#if SSLEAY_VERSION_NUMBER < 0x00904000L
# error SSLeay version must be at least 0.9.4
#endif

#define __DSA_PRIVATE_KEY
#define __DSA_PUBLIC_KEY
#undef  __DSA_IMPORT_KEY
#include "crypt/dsa0key.h"

#undef  __DSA_PRIVATE_KEY
#undef  __DSA_PUBLIC_KEY
#define __DSA_IMPORT_KEY
#include "crypt/dsa1key.h"

#ifdef USE_SHA1_MESSAGE_DIGEST
# include <openssl/sha.h>
# define MD_PACK(s,n)     SHA1 (s, n, 0)
# define MD_DIGEST_LENGTH SHA_DIGEST_LENGTH
#else
# include <openssl/md5.h>
# define MD_PACK(s,n)     MD5 (s, n, 0)
# define MD_DIGEST_LENGTH MD5_DIGEST_LENGTH
#endif

/* ----------------------------------------------------------------------- *
 * Private: constants, private data, macros
 * ----------------------------------------------------------------------- */

typedef
struct _sigdata {
    const char *data ;    /* storage for computation results */
    time_t       ttd ;    /* time when this record expires */
    unsigned    slot ;    /* nvram variable name slot */
    KEY         *key ;    /* referes to key material */
    DSAOP       *ref ;    /* (recursive) pointer to substructure */
} SIGDATA ;

/* descriptor allocation */
#define DSAOP_alloc() ZMALLOC (DSAOP)

/* ----------------------------------------------------------------------- *
 * Private, call back handlers
 * ----------------------------------------------------------------------- */

static
void clean_dsadata (LIST *l, SIGDATA *sd) {

    const char *var = sd->slot ? list_get_var (l, 1, 0, sd->slot) : 0 ;

    if (sd->data)
        free ((char*)sd->data);

    if (sd->ref != 0)
        dsafree (sd->ref);
        
    XZERO (*sd);

    if (var)
        list_get_val (l, 1, var, 0); /* read provokes cleanup */
}

/* ----------------------------------------------------------------------- *
 * Private helpers
 * ----------------------------------------------------------------------- */

static
void nofree (void *desc) { }


static
DSA *private_key (KEY *k) {
    const unsigned char *p, *q ;

    switch (k->type) {
    case 1:
        /* decrypt/decode key material */
        switch (k->inst) {
        case 1:
            p = dsa1decode ((unsigned char*)k->u.b64.data, k->u.b64.len) ;
            break ;
        default:
            p = 0 ;
        }
        
        if (p != 0) {
            DSA *dsa ;

            /* read length prefix (little endian 7 bit, MSB => continue) */
            int n = 0, len = 0;
            do
                len |= ((unsigned) (p [n] & 0x7f)) << (7 * n);
            while (n < 4 && p [n ++] & 0x80);

            /* skip n length bytes, import key */
            dsa = n < k->u.b64.len
                ? d2i_DSAPrivateKey (0, (q = p + n, &q), len)
                : 0
                ;

            free ((char*)p);
            return dsa ;
        }
        return 0;

    case 3:
        switch (k->inst) {
#if __not_used_at_the_moment__
        case 1:
            return dsa_get_priv_key1 ();
#endif
        case 0:
            return dsa_get_priv_key0 ();
        }
    }

    return 0;
}


static
const char *new_sigslot (LIST          *l,
                         DSAOP       *sid,
                         const char  *pfx,
                         unsigned rvarlen,
                         unsigned   slots,
                         unsigned     ttd,
                         const char* (*f) (LIST*,void*,
                                           const char*,int,
                                           const char*),
                         int        flags) {
    IOBUF b ;
    int top, inx, n ;
    time_t now = time (0) ;
    const char *val ;

    /* initialise table structure */
    if (sid->sd == 0) {

        sid->dim = slots ;
        sid->sd  = zmalloc (sid->dim  * sizeof (SIGDATA));
        sid->l   = l ;

        SSL_load_error_strings ();
    }

    /* find a free table entry */
    for (n = 0, inx = -1; n < sid->dim; n ++) {
        if (sid->sd [n].ttd == 0) {
            /* got a free slot */
            inx = n ;
            break ;
        }
        if (sid->sd [n].ttd <= now) {
            /* check expired candidates - get the oldest */
            if (inx < 0 || sid->sd [n].ttd < sid->sd [inx].ttd)
                inx = n ;
        }
    }

    if (inx < 0) {
        logger (LOG_WARNING, "Rejecting signid request: table busy "
                "(%d slots)", sid->dim);
        return 0 ;
    }

    clean_dsadata (l, sid->sd + inx) ;

    /* register new unique identifier */
    XZERO (b);
    data_printf (&b, "%s%s.", pfx, val = rndalloc (rvarlen));
    top = b.top ;
    n   = 0 ;

    for (;;) {
        data_printf (&b, "%u", n) ;
        if (list_avp_replace (l, 1, b.data, val, 0, 1) != 0)
            break ;
        b.top = top ;
        n ++ ;
    }

    /* set up function */
    list_fregister (l, val, sid->sd + inx, f, nofree, b.data, flags);
    free ((char*)val);

    sid->sd [inx].ttd      = now + ttd;
    sid->sd [inx].key      = &sid->key ;
    sid->sd [inx].slot     = list_get_slot (l, b.data);
    kill_space (&b);

    /* return persistent memory == b.data */
    return list_get_var (l, 1, 0, sid->sd [inx].slot) ;
}


/* ----------------------------------------------------------------------- *
 * Private, temporary data functions, call back handlers
 * ----------------------------------------------------------------------- */

static
const char *dsasignx (LIST         *l,
                      void      *desc,
                      const char *var,
                      int       flags,
                      const char *val) {

    SIGDATA *sd = desc ;
    unsigned n, ok ;
    unsigned char *p, *s ;
    DSA *key ;

    if (val == 0) {
        if (sd->data != 0 ||   /* signature response */
            sd->slot == 0) {   /* kill yourself request */
            sd->ttd = 0 ;                             /* mark idle */
            list_fregister (l, val, 0, 0, 0, var, 0); /* delete */
        }
        return sd->data ;
    }

    /* load the private key */
    if ((key = private_key (sd->key)) == 0) {
        logger (LOG_ERR, "Private key not available: %s",
                ERR_error_string (ERR_get_error (), 0));
        return 0 ;
    }

    /* digest */
    p = MD_PACK ((const unsigned char *)val, strlen (val));

    /* sign */
    n  = DSA_size (key) ;
    s  = (unsigned char *)xmalloc (n) ;
    ok = DSA_sign (0, p, MD_DIGEST_LENGTH, s, &n, key) ;

    /* no private key needed, anymore */
    DSA_free (key) ;

    if (ok == 1)
        /* encode digest */
        sd->data = b64alloc ((char*)s, n);
    else
        logger (LOG_ERR, "Signing error: %s",
                ERR_error_string (ERR_get_error (), 0));

    free ((char*)s) ;
    return ok == 1 ? "" : (const char *) -1 ;
}


static
const char *dsavfyx (LIST         *l,
                     void      *desc,
                     const char *var,
                     int       flags,
                     const char *val) {
    SIGDATA *sd = desc ;
    unsigned n, ok;
    const char *s, *result = (const char *) -1 ;

    if (sd->data == 0) {

        if (val != 0) {
            /* decode signature */
            if ((ok = b46alloc (&s, &n, val)) != 0) {

                /* digest */
                unsigned char *p = MD_PACK
                    ((const unsigned char *)var, strlen (var));

                /* verify */
                ok = DSA_verify (0, p, MD_DIGEST_LENGTH,
                                 (const unsigned char*)s,
                                 n, sd->key->u.public);

                free ((char*)s);

                switch (ok) {
                case 0:
                    sd->data = zstrdup (DSAFUNC_VFY_ISFAIL);
                    return "";
                case 1:
                    sd->data = zstrdup (DSAFUNC_VFY_ISOK);
                    return "";
                }

                logger (LOG_ERR, "Signature verification error: %s",
                        ERR_error_string (ERR_get_error (), 0));
            }
        } /* val != 0 */

    } /* data == 0 */

    else if (val == 0) /* && data != 0 */
        result = sd->data ;

    sd->ttd = 0 ;                              /* mark idle */
    list_fregister  (l, val, 0, 0, 0, var, 0); /* delete */
    return result;
}


static
const char *dsakeyx (LIST         *l,
                     void      *desc,
                     const char *var,
                     int       flags,
                     const char *val) {
    SIGDATA *sd = desc ;

    if (val != 0) {
        KEY *key ;

        if (sd->ref == 0)
            sd->ref = DSAOP_alloc () ;
        else if (sd->ref->key.u.b64.data != 0) {
            free ((char*)sd->ref->key.u.b64.data) ;
            XZERO (sd->ref->key) ;
        }

        key = &sd->ref->key ; /* shortcut */

        if (b46alloc (&key->u.b64.data, &key->u.b64.len, val)) {
            const char *e ;

            if ((key->u.b64.len & 7) == 0) {
                DSA *dsa ;
                
                /* test whether we can decode this key */
                key->type = 1 ;
                key->inst = sd->key->inst ;
                if ((dsa = private_key (key)) != 0) {
                    DSA_free (dsa);
                    return "" ;
                }

                e = ERR_error_string (ERR_get_error (), 0);
            }
            else {
                e = "wrong input data format" ;
            }
            
            logger (LOG_ERR, "Error importing the private key: %s", e);

            free ((char*)key->u.b64.data) ;
            free ((char*)sd->ref);
            sd->ref = 0 ;
        }
    }
    else if (sd->ref != 0)

        return new_sigslot
            (l, sd->ref, DSAFUNC_SIGN_VARPFX, DSAFUNC_SIGN_RVARLEN,
             DSAFUNC_SIGN_SLOTS, DSAFUNC_SIGN_LIFETIME / 1000,
             dsasignx, flags);

    return (const char *) -1 ;
}
#endif /* no HAVE_OPENSSL */

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

const char *dsa1key (LIST         *l,
                     void      *desc,
                     const char *var,
                     int       flags,
                     const char *val) {
#   ifdef HAVE_OPENSSL
    if (val == 0) {
        DSAOP *sid = desc;

        const char *s = new_sigslot
            (l, sid, DSAFUNC_KEY_VARPFX, DSAFUNC_KEY_RVARLEN,
             DSAFUNC_KEY_SLOTS, DSAFUNC_KEY_LIFETIME / 1000, dsakeyx, flags);

        if (s != 0) {
            sid->key.inst = 1 ;
            // sid->key.type = 0 ; /* no key, at all */
            return s ;
        }
    }
#   endif /* no HAVE_OPENSSL */

    return (const char *) -1 ;
}


const char *dsa0sign (LIST         *l,
                      void      *desc,
                      const char *var,
                      int       flags,
                      const char *val) {
#   ifdef HAVE_OPENSSL
    if (val == 0) {
        DSAOP *sid = desc;

        const char *s = new_sigslot
            (l, sid, DSAFUNC_SIGN_VARPFX, DSAFUNC_SIGN_RVARLEN,
             DSAFUNC_SIGN_SLOTS, DSAFUNC_SIGN_LIFETIME / 1000,
             dsasignx, flags);

        if (s != 0) {
            // sid->key.inst = 0 ;
            sid->key.type = 3 ; /* no permanent (but private) key */
            return s ;
        }
    }
#   endif /* no HAVE_OPENSSL */

    /* read only function */
    return (const char *) -1 ;
}


const char *dsa0vfy (LIST         *l,
                     void      *desc,
                     const char *var,
                     int       flags,
                     const char *val) {
#   ifdef HAVE_OPENSSL
    if (val == 0) {
        DSAOP *sid = desc;

        const char *s = new_sigslot
            (l, sid, DSAFUNC_VFY_VARPFX, DSAFUNC_VFY_RVARLEN,
             DSAFUNC_VFY_SLOTS, DSAFUNC_VFY_LIFETIME / 1000,
             dsavfyx, flags) ;

        if (s != 0) {

            if (sid->key.u.public == 0) {
                if ((sid->key.u.public = dsa_get_pub_key0 ()) == 0) {
                    logger (LOG_ERR, "Public key 0 not available: %s",
                            ERR_error_string (ERR_get_error (), 0));
                    return 0 ;
                }

                sid->key.inst = 0 ; /* first key instance */
                sid->key.type = 2 ; /* type: public key */
            }

            return s ;
        }
    }
#   endif /* no HAVE_OPENSSL */

    /* read only function */
    return (const char *) -1 ;
}


void dsafree (void *desc) {
#   ifdef HAVE_OPENSSL
    DSAOP *sid = desc;
    int n ;

    switch (sid->key.type) {
    case 1:
        free ((char*)sid->key.u.b64.data) ;
        break;
    case 2:
        DSA_free (sid->key.u.public);
    }

    for (n = 0; n < sid->dim; n ++)
        clean_dsadata (sid->l, sid->sd + n);

    free (sid->sd);
    free (sid);
#   endif /* no HAVE_OPENSSL */
}

/* ----------------------------------------------------------------------- *
 * Done with ssl stuff
 * ----------------------------------------------------------------------- */

#ifndef HAVE_OPENSSL
#define DSAOP_alloc() 0
#define dsafree       0
#define dsa0sign      0
#define dsa1key       0
#endif /* no HAVE_OPENSSL */


void mkdsafunctions (LIST *l) {
    list_fregister (l, "dsa0sign", DSAOP_alloc (), dsa0sign, dsafree, 0,0);
    list_fregister (l, "dsa0vfy",  DSAOP_alloc (),  dsa0vfy, dsafree, 0,0);
    list_fregister (l, "dsa1key",  DSAOP_alloc (),  dsa1key, dsafree, 0,0);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

