/*
 * nvramd - active data base server and flash memory cache
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: http-server.c 3175 2006-09-23 19:17:11Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include "util.h"
#include "list.h"
#include "nvalloc.h"
#include "settings.h"
#include "logger.h"

#ifndef DISCARD_HTTP

#define __HTTP_SERVER_PRIVATE
#include "http-server.h"
#include "varstr.h"

char e400 [] = "400 Bad Request" ;
char e503 [] = "503 Service Unavailable" ;
char e404 [] = "404 Not Found" ;

/* ----------------------------------------------------------------------- *
 * Private: text/data manipulation
 * ----------------------------------------------------------------------- */

static /* append http response data */
void appnd_http_response (RPLY_REC *r) {

    const char *code ;
    
    if ((code = r->error) == 0)
        code = "200 OK";
  
    data_printf (&r->response,
                 "HTTP/1.1 %s\r\n"
                 "Content-Type: text/plain\r\n"
                 "Server: Maxina/NVRAM-driver $Revision: 3175 $\r\n",
                 code);

    if (r->body.data && r->body.top)
        data_printf (&r->response,
                     "Content-Length: %u\r\n\r\n%s",
                     r->body.top, r->body.data);
    else
        data_printf (&r->response,
                     "Content-Length: 0\r\n\r\n");
}



/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
void close_http_channel (HCHAN *c) {

    if (c->fd >= 0) {
        shutdown (c->fd, 2) ;
        close (c->fd);
        undo_action (c->base->action, c->fd);
        c->fd     = -1 ;
        c->client =  0 ;
    }

    if (c->req.buf.data) {
        free (c->req.buf.data);
        XZERO (c->req);
    }
}


static
unsigned parse_http_header (HCHAN *c) {

    char *p ;
    int n, ct;
    CMD_MAP *cmd;
    
    /* check whether we got a complete line */
    if (strchr (c->req.buf.data, '\n') == 0) {
        /* FIXME: deadlock protection */
        return 0;
    }

    /* parse the http method */
    if (strncmp (c->req.buf.data, "GET ", 4) == 0) {
        c->req.http.state = http_get_header ;
        c->req.http.url_args = 1;
        p = c->req.buf.data + 4 ;
    }
    else if (strncmp (c->req.buf.data, "POST ", 5) == 0) {
        c->req.http.state = http_post_header ;
        p = c->req.buf.data + 5 ;
    }
    else
        return 0 ;

    /* parse the request: [http[s]://nnnn[:ppp]]/.. */
    while (*p == ' ') p ++ ;
    if (strncasecmp (p, "http", 4) == 0) {
        p += 4;
        if (* p == 's' || * p == 'S')
            p ++ ;
        if (* ++ p != ':' || * ++ p != '/' || * ++ p != '/')
            return 0;
        /* find the beginning of the ressource looking like "/<cmd>..." */
        while (*p != '/' && isgraph (*p))
            /* isgraph (): any printable character except space. */
            p ++ ;
    }
    if (* p ++ != '/')
        return 0;

    /* remember the command token */
    c->req.http.cmd_req = p - 1 - c->req.buf.data ;

    /* parse the application method: "/<command>..." */
    for (n = 0, cmd = c->base->handler.cmds;
         n < c->base->handler.num_cmds; n ++) {
        if (strncmp (p, cmd [n].name, cmd [n].len) != 0)
            continue ;
        /* check for proper termination of the command */
        if (strchr (" ?/", p [cmd [n].len]) == 0)
            continue ;
        c->req.http.cmd         = cmd [n].cmd ;
        c->req.http.cmd_args_ok = cmd [n].args_ok ;
        p                      += cmd [n].len ;
        break ;
    }

    /* use default (NULL) => error return */
    if (c->req.http.cmd == 0) {
        while (isalnum (* p) || * p == '-')
            p ++ ;
        if (strchr (" ?/", * p) == 0)
            return 0;
    }

    /* terminate command token but save ct for syntax check */
    ct = * p ;
    * p ++ = '\0' ;
    
    /* read the get request and verify the request syntax */
    if (ct != ' ') {
        if (c->req.http.state != http_get_header || ct != '?')
            return 0;
        c->req.http.args = p - c->req.buf.data ;
        while (*p && *p != ' ')   /* advance to the next space */
            p ++ ;
        * p ++ = '\0' ;           /* term. the request string */
    }
    
    /* verify the http header syntax: "http/1.0" or similar */
    while (*p == ' ') p ++ ;
  
    if (strncasecmp (p, "http/", 5) != 0)
        return 0 ;
    p += 5 ;

    if (!('0' <= p [0] && p [0] <= '9') ||
        !                (p [1] == '.') ||
        !('0' <= p [2] && p [2] <= '9')) 
        return 0 ;
    p += 3 ;

    if (p [0] == '\n') {
        p ++ ;
    }
    else if (p [0] == '\r' || p [1] == '\n') {
        p += 2 ;
    }
    else
        return 0 ;

    return p - c->req.buf.data ;
}


static
unsigned parse_http_line (HCHAN *c) {
    char *r, *end ;
    char *var = c->req.buf.data + c->req.http.seen ;
    char *val = strchr (var, ':');
    unsigned u ;

    if ((r = val ++) == 0)
        return 0;

    /* extract header variable name: <header-variable> [space] ':' */
    while (r [-1] == ' ') r -- ;
    r [0] = '\0';
  
    /* separate header value */
    while (* val == ' ') val ++ ;
    if ((end = strchr (val, '\n')) == 0)
        return 0 ;
    r = end ++ ;
    while (r [-1] <= ' ')
        r -- ;
    r [0] = '\0' ;

    /* evaluate particular headers */
    if (strcasecmp (var, "Content-Length") == 0) {
        if (parse_unsigned (&u, val, 0) == 0 || u < 0) return 0;
        c->req.http.clen = u ;
    }
    else if (strcasecmp (var, "Accept") == 0) {
        c->req.http.acc_ena = 1 ;
        if (strstr (val, "text/plain") != 0)
            c->req.http.acc_text = 1 ;
    }
    else if (strcasecmp (var, "Connection") == 0) {
        if (strcasecmp (val, "Keep-Alive") == 0) {
            if (c->req.http.keep_alive == 0)
                c->req.http.keep_alive = time (0) + HTTP_KEEP_ALIVE ;
        }
        else if (strcasecmp (val, "Close") == 0) {
            c->req.http.keep_alive = 0 ;
            c->req.http.conn_close = 1 ;
        }
    }
    else if (strcasecmp (var, "Keep-Alive") == 0) {
        if (parse_unsigned (&u, val, 0) == 0 || u < 0) return 0;
        if (u && c->req.http.conn_close == 0)
            c->req.http.keep_alive = time (0) + u ;
    }
  
    return c->req.http.seen = end - c->req.buf.data ;
}


static
unsigned send_http_reply (HCHAN *c) {

    int n ;
    RPLY_REC io ;
    XZERO   (io) ;

    io.c = c ;
        
    if (c->req.http.cmd != 0) {
        /* check arguments */
        if (c->req.http.cmd_args_ok == 0) {
            /* no arguments wanted */
            if (c->req.http.args) {
                io.error = e400 ;
                data_printf (&io.body, "No parameters accepted with %s",
                             c->req.buf.data + c->req.http.cmd_req);
            }
        }
        else if (c->req.http.cmd_args_ok == 1) {
            /* at least one argument wanted */
            if (c->req.http.args == 0) {
                io.error = e400 ;
                data_printf (&io.body, "Parameters expected with %s",
                             c->req.buf.data + c->req.http.cmd_req);
            }
        }
        if (io.error == 0)
            /* assemble response by the argument list */
            (*c->req.http.cmd) (c, &io);
    }
    else if (c->base->handler.def_cmd.cmd) {
        (*c->base->handler.def_cmd.cmd) (c, &io);
    }
    else {
        io.error = e404 ;
        data_printf (&io.body,  "No method handler available");
    }
 
    /* assemble http response */
    appnd_http_response (&io);
    if (kill_space (&io.body)) {
        /* make sure that the client understands text plain */
        if (c->req.http.acc_ena && !c->req.http.acc_text) {
            kill_space (&io.response);
            logger (LOG_ERR,
                    "Cannot send on the http socket fd=%d: "
                    "text/plain is not understood by the client");
            return 0;
        }
    }
    
    /* send answer packet */
    n = send (c->fd, io.response.data, io.response.top, 0); 

#   ifdef HTTP_REPLY_LOG
    if (verbosity & HTTP_REPLY_LOG)
        logger (LOG_DEBUG, "HTTP reply: \"%s\"", io.response.data);
#   endif

    /* clean up */
    kill_space (&io.response);

    if (n < 0) {
        logger (LOG_ERR,
                "Send on the http socket fd=%d returned error: %s",
                c->fd, strerror (errno));
        return 0;
    }

    return 1;
}


/* ----------------------------------------------------------------------- *
 * Private: http data callback handler
 * ----------------------------------------------------------------------- */

static
void http_data (void *data) {

    int n ;
    HCHAN *c = data ;

    /* extend read buffer */
    make_space (&c->req.buf, HTTP_MIN_REQBUF) ;

    if ((n = recv (c->fd,
                   c->req.buf.data + c->req.buf.top,
                   c->req.buf.size - c->req.buf.top - 1,
                   MSG_DONTWAIT)) < 0) {
        if (errno != EAGAIN) { 
            logger (LOG_ERR,
                    "Closing http socket fd=%d on read error: %s",
                    c->fd, strerror (errno));
            close_http_channel (c);
        }
        return ;
    }

    if (n == 0) {
        /* orderly shut down */
        if (verbosity & HTTP_LOG_CLOSE)
            logger (LOG_ERR, "Closing http socket, fd=%d", c->fd);
        close_http_channel (c) ;
        return ;
    }

    /* prepare for text parsing */
    c->req.buf.data [c->req.buf.top + n] = '\0' ;

#   ifdef HTTP_REQUEST_LOG
    if (verbosity & HTTP_REQUEST_LOG)
        logger (LOG_DEBUG, (c->req.buf.top
	                    ? "HTTP request (contd.): %d=\"...%s\""
	                    : "HTTP request: %d=\"%s\""),
	        n, c->req.buf.data + c->req.buf.top);
#   endif

    /* set to end of text */
    c->req.buf.top += n ;

    for (;;) {
        switch (c->req.http.state) {

        case none:
            if ((c->req.http.seen = parse_http_header (c)) == 0)
                break;
            continue ;

        case http_post_header:
        case http_get_header: {
            char *p, *q = c->req.buf.data + c->req.http.seen ;
            if ((p = strchr (q, '\n')) == 0)
                /* need more data */
                return ;
            if (* q == '\r' && p == q + 1) {
                c->req.http.state = c->req.http.state == http_get_header
                    ? http_get_ok
                    : http_post_body ;
                c->req.http.seen += 2 ;
                continue ;
            }
            if (p == q) {
                c->req.http.state = c->req.http.state == http_get_header
                    ? http_get_ok
                    : http_post_body ;
                c->req.http.seen ++ ;
                continue ;
            }
            if ((n = parse_http_line (c)) == 0)
                break ;
            c->req.http.seen = n ;
            continue ;
        }
        case http_get_ok:
            if (send_http_reply (c) == 0)
                break ;
            c->req.http.state = http_sent_ok ;
            continue;

        case http_post_body:
            /* need to check that the body is complete */
            if (c->req.buf.size - c->req.buf.top < c->req.http.clen)
                continue ;
            if (c->req.buf.data [c->req.buf.top - 1] != '\n')
                break ;
            if (c->req.http.seen < c->req.buf.top)
                c->req.http.args = c->req.http.seen;
            c->req.http.state = http_post_ok ;
            continue ;

        case http_post_ok:
            if (send_http_reply (c) == 0)
                break ;
            c->req.http.state = http_sent_ok ;
            continue;
    
        case http_sent_ok:
            /* remove that data channel unless keep-alive is on */
            if (c->req.http.keep_alive == 0 ||
                c->req.http.conn_close != 0 ||
                c->req.http.keep_alive < time (0))
                break ;
            if (c->req.buf.data) {
                free (c->req.buf.data);
                XZERO (c->req);
            }
            return ;

        default:
            logger (LOG_ERR, "State machine error (internal) - %u",
                    c->req.http.state);
            break;
        }

        /* remove that data channel */
        close_http_channel (c);
        return ;
    }

    /* NOTREACHED */
}


HTTP *http_cmdadd (HTTP                          *h,
                   const char                 *name,
                   int                    have_args,
                   void (*handler)(void*,RPLY_REC*),
                   void              *handler_state) {

    if (h && handler) {
        
        CMD_MAP *cmd ;
            
        if (name) {
            int n = h->handler.num_cmds ++ ;

            /* get space for an additional entry */
            h->handler.cmds = (CMD_MAP*) zrealloc
                (h->handler.cmds, h->handler.num_cmds * sizeof (CMD_MAP)) ;

            cmd = h->handler.cmds + n ;
        }
        else {
            cmd = & h->handler.def_cmd ;
        }

        XZERO (*cmd) ;
        if (name) {
            cmd->name =         name  ;
            cmd->len  = strlen (name) ;
        }
        cmd->cmd      =      handler  ;
        if (have_args)
            cmd->args_ok = have_args > 0 ? 1 : 3 ;
    }

    return h ;
}

/* ----------------------------------------------------------------------- *
 * Public: call bach function helpers
 * ----------------------------------------------------------------------- */

/* terminate the avpair as string, shift data pointer */
const char *http_shiftargs (RPLY_REC *r) {

    HCHAN *c = r->c ;

    char *p, *var ;
    
    if (c->req.http.args == 0 ||
        *(var = c->req.buf.data + c->req.http.args) == '\0')
        return 0;

    if ((p = strchr (var, '&')) != 0) {
        * p ++ = '\0' ;
        c->req.http.args = p - c->req.buf.data ;
    }
    else {
        int n = strlen (var) ;
        /* chop trailing spaces like \r and \n */
        while (n > 0 && isspace (var [n - 1]))
            n -- ;
        var [n] = '\0' ;
        c->req.http.args = var + n - c->req.buf.data ;
    }

    /* undo escapes */
    if ((p = strchr (var, '%')) != 0 ||
        (p = strchr (var, '+')) != 0) {
        /* (p = strchr (var, '\\')) != 0) { // did not solve anything -- jh */
        char *q = p ;
        do {
            int a, b ;
            if (p [0] == '+') {
                * p = ' ' ;
            }
            else if (p [0] == '%' && ((a = parse_hex_dgt (p [1])) >= 0 &&
                                      (b = parse_hex_dgt (p [2])) >= 0)) {
                * q = (a << 4) | b ;
                p += 2 ;
            }
            else if (q < p) {
/**** did not solve anything - jh :(
 *           } else if (p [0] == '\\') {
 *
 *               p++;
 *               *q = *q;
 *		    
 *           } else if (q < p) {
 */
                * q = * p ;
            }
        }
        while (++ q, * ++ p) ;
        * q = '\0' ;
    }

    return var ;
}

IOBUF *http_rplybody (RPLY_REC *r) {
    return & r->body ;
}

LIST *http_rplylist (RPLY_REC *r) {
    return r->c->base->lst ;
}

void http_rplyerror (RPLY_REC *r, const char *e) {
    r->error = e ;
}

void http_rplydata (RPLY_REC *r, char *s) {
    r->body.data = s ;
}

unsigned http_haveargs (RPLY_REC *r) {
    return r->c->req.http.args!= 0 ;
}

const char *http_urlmethod (RPLY_REC *r) {
    HCHAN *c = r->c ;
    return c->req.http.cmd_req ? c->req.buf.data + c->req.http.cmd_req : 0 ;
}

const char *http_cmdname (RPLY_REC *r) {
    HTTP  *h                   = r->c->base ;
    void (*c)(void*,RPLY_REC*) = r->c->req.http.cmd ;

    int n ;
    CMD_MAP *cmd ;

    for (n = 0, cmd = h->handler.cmds; n < h->handler.num_cmds; c ++)
        if (cmd [n].cmd == c) 
            return cmd [n].name ;

    return "*unknown*" ;
}


void http_verbosity (void *state, RPLY_REC *r) {

    unsigned old = verbosity ;
    IOBUF *b = http_rplybody (r) ;
    const char *var, *val ;

    if ((var = http_shiftargs (r)) != 0) {
        unsigned noise ;
        if ((val = strchr (var, '=')) == 0) {
            val = var ;
        }
        else if (strncmp (var, "level=", 6) == 0) {
            val ++ ;
        }
        else {
            data_printf
                (b, "Syntax error ([level=]<num> expected): %s", var);
            http_rplyerror (r, e404) ;
            return ;
        }
        if (parse_unsigned (&noise, val, 0) == 0) {
            data_printf (b, "Unknown integer noise level format: %s", val);
            http_rplyerror (r, e404) ;
            return ;
        }
        verbosity = noise ;
    }

    data_printf (b, "old-level=%#x\r\n", old);
}

/* ----------------------------------------------------------------------- *
 * Public: constructor/destructor
 * ----------------------------------------------------------------------- */

HTTP *http_init (DO       *action,
                 LIST       *list, 
                 unsigned   slots,
                 ip4addr_t ipaddr,
                 uint16_t    port) {

    struct sockaddr_in in ;
    int n;
    HTTP *h ;

    if (slots == 0) slots = 1;
    h = (HTTP*) zmalloc (sizeof (HTTP) * (slots - 1)) ;

    if ((h->action = action) == 0) {
        logger (LOG_ERR,
                "Initializaion error: call back handler object missing");
        goto error_ok ;
    }
    if ((h->lst = list) == 0) {
        logger (LOG_ERR, "Initializaion error: "
                "assignment list handler object missing");
        goto error_ok ;
    }

    /* open http socket */
    if ((h->listen_fd = socket (PF_INET, SOCK_STREAM, 0)) == 0) {
        logger (LOG_ERR, "Error creating http socket: %s", strerror (errno));
        goto error ;
    }
  
    XZERO (in);
    in.sin_family = AF_INET ;
    in.sin_port   = htons (port) ;
    if (ipaddr) in.sin_addr.s_addr = htonl (ipaddr) ;
   
    if (setsockopt
        (h->listen_fd, SOL_SOCKET, SO_REUSEADDR, (n=1, &n), sizeof n) < 0) {
        logger (LOG_ERR,
                "Error customizing http socket: %s", strerror (errno));
        goto error ;
    }
    if (bind (h->listen_fd, (struct sockaddr*)&in, sizeof in) < 0) {
        logger (LOG_ERR, "Error binding http socket: %s", strerror (errno));
        goto error ;
    }
    if (listen (h->listen_fd, slots) < 0) {
        logger (LOG_ERR,
                "Error listening on http socket: %s", strerror (errno));
        goto error ;
    }
#   ifdef FD_CLOEXEC
    fcntl (h->listen_fd, F_SETFD, FD_CLOEXEC);  /* close socket in exec */
#   endif

    /* initialise slots */
    h->channels = slots ;
    for (n = 0; n < slots; n ++) {
        h->channel [n].fd   = -1 ;
        h->channel [n].base = h ;
    }
  
    return h ;

error:
    if (h->listen_fd >= 0)
        close (h->listen_fd);

error_ok:
    free (h);
    return 0;
}


int http_fd (HTTP *h) {
    return h ? h->listen_fd : -1;
}


void http_close (HTTP *h) {
  if (h) {
      int n;
      for (n = 0; n < h->channels; n ++) 
          close_http_channel (h->channel + n);
      if (h->listen_fd >= 0)
          close (h->listen_fd);
      free (h);
  }
}

/* ----------------------------------------------------------------------- *
 * Public: event loop call back
 * ----------------------------------------------------------------------- */

void http_accept (void *data) {

    HTTP *h = data ;

    struct sockaddr_in in ;
    socklen_t alen = sizeof (in) ;
    int n, fd;
  
    if ((fd = accept (h->listen_fd, (struct sockaddr*)&in, &alen)) < 0) {
        logger (LOG_ERR, "Error accepting http: %s", strerror (errno));
        return ;
    }

    /* find a free socket descriptor */
    for (n = 0; n < h->channels; n ++)
        if (h->channel [n].fd < 0) {
            h->channel [n].fd     = fd ;
            h->channel [n].client = ntohl (in.sin_addr.s_addr) ;
            do_action  (h->action, fd, http_data, h->channel + n);
#           ifdef FD_CLOEXEC /* close socket in exec */
            fcntl (h->channel [n].fd, F_SETFD, FD_CLOEXEC);
#           endif
            return ;
        }

    /* no socket descriptor left - close up */
    logger (LOG_ERR, "No more slots for http - rejecting %s",
            ipaddr_pp (&in.sin_addr));

    shutdown (fd, 2);
    close    (fd);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

#endif /* DISCARD_HTTP */
