/*
 * nvram lib - flash data base client library
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: buffer.c 3113 2006-08-20 19:48:27Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "logger.h"
#include "util.h"
#include "nvalloc.h"
#include "buffer.h"

// #undef TRACK_BUFFER_ID_ENABLE

/* ----------------------------------------------------------------------- *
 * Private variables and definitions
 * ----------------------------------------------------------------------- */

#ifdef TRACK_BUFFER_ID_ENABLE

static int buffer_id = 0;

# define TRACK_BUFFER_ID_REGISTER
# define BUFFER_ID_INTRO_TAG "track-buffers"

# define trackid_static static

# undef make_space
# undef data_printf
# undef data_vprintf
# undef data_append
# undef char_append
# undef file_append
# undef kill_space

# ifdef WITH_DMALLOC
#  define trackid_printf(fmt, ...) dmalloc_message (fmt, ##__VA_ARGS__)
# else
   static int __a;
#  define trackid_printf(fmt, ...) \
    {if (__a==0) {__a=1; logger (LOG_DEBUG, fmt, ##__VA_ARGS__); __a=0;}}
# endif

#else
# define trackid_static           /* n/a */
# define trackid_printf(fmt, ...) /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

#ifdef TRACK_BUFFER_ID_REGISTER

static void data_printf (IOBUF*, const char*, ...);
static int id_list_dim, *id_list ;

void id_register (int id) {
    int n ;

    for (n = 0; n < id_list_dim; n++)
        if (id_list [n] == 0) {
            id_list [n] = id ;
            return ;
        }

    id_list = (int*) zrealloc (id_list, (++ id_list_dim) * sizeof (int)) ;
    id_list [n] = id ;
}


void id_clear (int id, const char *s) {
    IOBUF b ;
    int n, k ;

    for (n = 0; n < id_list_dim; n++)
        if (id_list [n] == id) {
            id_list [n] = 0 ;
            goto end_up ;
        }

    trackid_printf (BUFFER_ID_INTRO_TAG " on %s: id_clear() "
                    "could not de-register id=%d.", s, id);

end_up:
    XZERO (b);
    b.data = (char*) xmalloc (b.size = 100) ;

    for (n = k = 0; n < id_list_dim; n++)
        if (id_list [n] != 0)
            data_printf (&b, (k ++ ? ", %d" : "%d"), id_list [n]);

    if (k)
        trackid_printf (BUFFER_ID_INTRO_TAG " on %s: id_clear() knows "
                        "about %d unfreed buffers: %s", s, k, b.data);

    free ((char*)b.data);

}

#endif /* TRACK_BUFFER_ID_REGISTER */

/* ----------------------------------------------------------------------- *
 * Public (unless TRACK_BUFFER_ID_ENABLE)
 * ----------------------------------------------------------------------- */

/* provide enough space for appending data */
trackid_static
void make_space (IOBUF *b, size_t space) {

    /* create buffer */
    if (b->data == 0) {
        b->size = 0 ;
        b->top  = 0 ;
    }
  
  if (b->size - b->top > space)
    return ;

  /* otherwise extend the buffer */
  b->size = b->top + space ;
  b->data = (char*) zrealloc (b->data, b->size) ; /* b->data == 0 is ok */
}


/* simply append formatted data */
trackid_static
void data_vprintf (IOBUF *b, const char *format, va_list ap) {
    size_t len, avail ;
    
    make_space (b, 100);
    avail = b->size - b->top ;
    len = vsnprintf (b->data + b->top, avail, format, ap);

    if (len >= avail) {
        make_space (b, len + 1);
        len = vsnprintf (b->data + b->top, b->size - b->top, format, ap);
    }
    b->top += len ;
}

trackid_static
void data_printf (IOBUF *b, const char *format, ...) {
    va_list ap;
    va_start (ap, format);
    data_vprintf (b, format, ap);
    va_end (ap);
}

trackid_static
void data_append (IOBUF *b, const char *data, size_t len) {
    if (len == 0) 
        len = strlen (data);
    make_space (b, len + 1);
    memcpy (b->data + b->top, data, len);
    b->top += len ;
    b->data [b->top] = '\0' ;
}   
    
trackid_static
void char_append (IOBUF *b, int c) {
    if (c != EOF) {
        make_space (b, 2); 
        if ((b->data [b->top ++] = c) != '\0')
            b->data [b->top] = '\0' ;
    }
}

trackid_static
void file_append (IOBUF *b, int fd, size_t len) {

    make_space (b, len + 1);

    while (len > 0) {
        int n = read (fd, b->data + b->top, len);
        if (n == 0)
            break ;
        if (n < 0) {
            int r = errno ;
            logger (LOG_ERR, "%s(file_append): read error: %s",
                    progname, strerror (errno));
            errno = r ;
            break ;
        }
        len    -= n ;
        b->top += n ;
    }

    b->data [b->top] = '\0' ;
}


/* clean up */
trackid_static
size_t kill_space (IOBUF *b) {
    if (b->data != 0) {
        unsigned n = b->top ;
        free (b->data) ;
        XZERO (*b);
        return n;
    }
    return 0;
}

/* ----------------------------------------------------------------------- *
 * Mem leak debugging ..
 * ----------------------------------------------------------------------- */

#ifdef TRACK_BUFFER_ID_ENABLE

#define BUFFER_ID_REGISTER(s) {                                           \
      trackid_printf (BUFFER_ID_INTRO_TAG " at %s(%d): " s "() creating " \
                      "data buffer at %#x(%#x), id=%d.", b->file = f,     \
                      b->line = l, b, b->data, b->id = ++ buffer_id);     \
      id_register (b->id);                                                }


#define BUFFER_ID_CLEAR(s,t) if (b->data) {                               \
        trackid_printf (BUFFER_ID_INTRO_TAG " at %s(%d): " s "() "        \
                        t " data buffer at %#x, id=%d.", f, l, b, b->id); \
        id_clear (b->id, s "()");                                         }

void
_make_space (const char *f, int l, IOBUF *b, size_t n) {
    int flag = (b->data == 0) ;
    make_space (b, n);
    if (flag) BUFFER_ID_REGISTER ("make_space");
}

void
_data_printf (const char *f, int l, IOBUF *b, const char*fmt, ...) {
    va_list ap;
    int flag = (b->data == 0) ;
    va_start (ap, fmt);
    data_vprintf (b, fmt, ap);
    va_end (ap);
    if (flag) BUFFER_ID_REGISTER ("data_printf");
}

void
_data_vprintf (const char *f, int l, IOBUF *b, const char*fmt, va_list ap) {
    int flag = (b->data == 0) ;
    data_vprintf (b, fmt, ap);
    if (flag) BUFFER_ID_REGISTER ("data_vprintf");
}

void
_data_append (const char *f, int l, IOBUF *b, const char*s, size_t n) {
    int flag = (b->data == 0) ;
    data_append (b, s, n);
    if (flag) BUFFER_ID_REGISTER ("data_append");
}

void
_char_append (const char *f, int l, IOBUF *b, int c) {
    int flag = (b->data == 0) ;
    char_append (b, c);
    if (flag) BUFFER_ID_REGISTER ("char_append");
}

void
_file_append (const char *f, int l, IOBUF *b, int fd, size_t n) {
    int flag = (b->data == 0) ;
    file_append (b, fd, n);
    if (flag) BUFFER_ID_REGISTER ("file_append");
}

void
_unlink_space (const char *f, int l, IOBUF *b) {
    BUFFER_ID_CLEAR ("unlink_space", "unregistering");
}

size_t
_kill_space (const char *f, int l, IOBUF *b) {
    BUFFER_ID_CLEAR ("kill_space", "freeing/recycling");
    return kill_space (b);
}

#undef BUFFER_ID_REGISTER
#undef BUFFER_ID_CLEAR
#endif /* TRACK_BUFFER_ID_ENABLE */

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
