/*
 * Copyright (C) 2003-2006 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: varstr.c 3447 2006-10-27 09:24:45Z henry $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_LIMITS_H
# include "limits.h"
/* Some gcc versions don't define LLONG_MAX, because it wasn't defined in
   the earlier C standards.  */
#endif

#ifndef  LLONG_MAX
# define LLONG_MAX   9223372036854775807LL
# warning: LLONG_MAX was undefined (known gcc/glibc problem)
#endif
#ifndef  LLONG_MIN
# define LLONG_MIN (-LLONG_MAX - 1LL)
# warning: LLONG_MIN was undefined (known gcc/glibc problem)
#endif
#ifndef  ULLONG_MAX
# define ULLONG_MAX 18446744073709551615ULL
# warning: ULLONG_MAX was undefined (known gcc/glibc problem)
#endif

#include "varstr.h"

/* ----------------------------------------------------------------------- *
 * Variables and definitions
 * ----------------------------------------------------------------------- */

varstr_error_code varstr_errno ;

#ifndef   SIZEOF_INT
# if UINT_MAX == 65535
#  define  SIZEOF_INT 2
#  warning SIZEOF_INT undefined - using 2
# elif UINT_MAX == 4294967295U
#  define  SIZEOF_INT 4
#  warning SIZEOF_INT undefined - using 4
# elif UINT_MAX == 18446744073709551615UL
#  define  SIZEOF_INT 8
#  warning SIZEOF_INT undefined - using 8
# endif
#endif

#ifndef   SIZEOF_UNSIGNED
# define  SIZEOF_UNSIGNED SIZEOF_INT
# warning SIZEOF_UNSIGNED undefined - using SIZEOF_INT
#endif

#if SIZEOF_UNSIGNED != 2 && \
    SIZEOF_UNSIGNED != 4 && \
    SIZEOF_UNSIGNED != 8
# error: sizeof (unsigned) has no proper value
#endif

#if SIZEOF_INT != 2 && \
    SIZEOF_INT != 4 && \
    SIZEOF_INT != 8
# error: sizeof (int) has no proper value
#endif

#ifndef  INT8_MAX
# define INT8_MIN  (-INT8_MAX - 1)
# define INT8_MAX   127
#endif

#ifndef  INT16_MAX
# define INT16_MIN (-INT16_MAX - 1)
# define INT16_MAX  32767
#endif

#ifndef  INT32_MAX
# define INT32_MIN (-INT32_MAX - 1)
# define INT32_MAX  2147483647
#endif

#ifndef  INT64_MAX
# define INT64_MIN (-INT64_MAX - 1)
# define INT64_MAX  9223372036854775807LL
#endif

typedef
union _var {
    int8_t    s8 ;
    uint8_t   u8 ;
    int16_t  s16 ;
    uint16_t u16 ;
    int32_t  s32 ;
    uint32_t u32 ;
    int64_t  s64 ;
    uint64_t u64 ;
    int   bool:1 ;
    struct {
        /* seven bitfield items / uint32_t is ok - compiler dependent */
        int x1:1,  x2:2, x3:3, x4:4, x5:5, x6:6, x7:7 ;
    } bit ;
} VAR ;


typedef
struct _formats {
    const char *undef_no_ext_parser_for_d ;
    const char *undef_size_d ;
    const char *undef_type_d ;
    const char *undef_bit_d ;
    const char *true ;
    const char *false ;
    const char *undef_ext_d ;
    const char *lu_enum_s ;
} FORMATS ;

static
const FORMATS std = {
    "<undef> /* no extension parser for %d */",
    "<undef> /* undefined size %d */",
    "<undef> /* undefined type %d */",
    "<undef> /* bit %d not available */",
    "1 /* true */",
    "0 /* false */",
    "<undef> /* undefined extension %d */",
    "%lu /* %s */"
} ;


static
const FORMATS quick = {
    "*",
    "*",
    "*",
    "*",
    "1",
    "0",
    "*",
    "%lu"
} ;

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

static
varstr_error_code integer_range_error (SCALAR_TYPE type) {
    switch (type) {
    case t_bool:
        return EVARSTR_bool_integer_string_out_of_range ;
    case t_enum:
        return EVARSTR_enum_integer_string_out_of_range ;
    case t_signed:
        return EVARSTR_signed_integer_string_out_of_range;
    default:
        return EVARSTR_unsigned_integer_size_undefined ;
    }
}


static
varstr_error_code integer_size_error (SCALAR_TYPE type) {
    switch (type) {
    case t_bool:
        return EVARSTR_bool_integer_size_undefined ;
    case t_enum:
        return EVARSTR_enum_integer_size_undefined ;
    case t_signed:
        return EVARSTR_signed_integer_size_undefined;
    default:
        return EVARSTR_unsigned_integer_size_undefined ;
    }
}


static
size_t parse_numeric (VAR           *v,
                      SCALAR_TYPE    t,
                      size_t        sz,
                      const char    *s,
                      size_t       len,
                      const char *term) {
    char *p = 0 ;

    if (len != 0) {
        s = memcpy (alloca (len + 1), s, len);
        ((char*)s) [len] = '\0' ;
    }

    if (t == t_signed) {

        int64_t d = strtoll (s, &p, 0);

        if (p == 0 || (*p != '\0' && (term == 0 || strchr (term, *p) == 0)))
            goto integer_signed_eterm ;

        if ((d == LLONG_MIN || d == LLONG_MAX) && errno == ERANGE)
            goto integer_erange ;
            
        switch (sz) {
        case sizeof (int8_t):
            if (d < INT8_MIN || INT8_MAX < d)
                goto integer_erange ;
            v->s8 = d ;
            break;
                
        case sizeof (int16_t):
            if (d < INT16_MIN || INT16_MAX < d)
                goto integer_erange ;
            v->s16 = d ;
            break;

        case sizeof (int32_t):
            if (d < INT32_MIN || INT32_MAX < d)
                goto integer_erange ;
            v->s32 = d ;
            break;

        case sizeof (int64_t):
#           if INT64_MAX < LLONG_MAX
            if (d < INT64_MIN || INT64_MAX < d)
                goto integer_erange ;
#           endif
            v->s64 = d ;
            break;
                
        default:
            goto integer_esize ;
        }
    }

    else {
            
        uint64_t u = strtoull (s, &p, 0);

        if (p == 0 || (*p != '\0' && (term == 0 || strchr (term, *p) == 0)))
            goto integer_unsigned_eterm ;

        if (u == ULLONG_MAX && errno == ERANGE)
            goto integer_erange ;

        switch (t) {
        case t_bool:
#           if SIZEOF_UNSIGNED == 2
            if ((u & ~ 0xffff) != 0)
                goto integer_erange ;
#           elif SIZEOF_UNSIGNED == 4
            if ((u & ~ 0xffffffff) != 0)
                goto integer_erange ;
#           endif
            v->bool = (u != 0) ;
            break;
            
        case t_enum:
#           if SIZEOF_UNSIGNED == 2
            if ((u & ~ 0xffff) != 0)
                goto integer_erange ;
#           elif SIZEOF_UNSIGNED == 4
            if ((u & ~ 0xffffffff) != 0)
                goto integer_erange ;
#           endif
            /* proceed with "unsigned" */
            
        default:
            switch (sz) {
            case sizeof (uint8_t):
                if ((u & ~ 0xff) != 0)
                    goto integer_erange ;
                v->u8 = u ;
                break;
                
            case sizeof (uint16_t):
                if ((u & ~ 0xffff) != 0)
                    goto integer_erange ;
                v->u16 = u ;
                break;
                
            case sizeof (uint32_t):
                if ((u & ~ 0xffffffff) != 0)
                    goto integer_erange ;
                v->u32 = u ;
                break;
            
            case sizeof (uint64_t):
                v->u64 = u ;
                break ;
                
            default:
                goto integer_esize ;
            }
        }
    }

    return sz ;

integer_erange:
    varstr_errno = integer_range_error (t);
    return 0;

integer_esize:
    varstr_errno = integer_size_error (t);
    return 0;

integer_signed_eterm:
    varstr_errno = EVARSTR_signed_integer_string_termination;
    return 0;

integer_unsigned_eterm:
    varstr_errno = EVARSTR_unsigned_integer_string_termination;
    return 0;
}

    
static
size_t parse_string (char *trg, size_t tsz, const char *s, size_t len) {

    size_t got = 0 ;
    
    if (len == 0)
        len = strlen (s);

    while (len > 0) {
        int c ;

        if ((c = s [0]) != '\\') {
            * trg ++ = * s ++ ;
            len -- ;
            continue ;
        }

        if (len < 2)
            return 0 ;
            
        switch (c = s [1]) {
        case 'X':
            c = 'x' ;
        case 'x':
        case '0':
        case '1':
        case '2':
        case '3': {
            int d ;
            
            if (len < 4)
                return 0 ;
            
            if (c == 'x') {
                if ((d = parse_hex_dgt (s [2])) == -1) {
                    varstr_errno = EVARSTR_string_parser_hex_dgt_error ;
                    return 0 ;
                }
                c    = d << 4 ;
                if ((d = parse_hex_dgt (s [3])) == -1) {
                    varstr_errno = EVARSTR_string_parser_hex_dgt_error ;
                    return 0 ;
                }
            }
            else {
                if ((d = parse_oct_dgt (s [1])) == -1) {
                    varstr_errno = EVARSTR_string_parser_oct_dgt_error ;
                    return 0 ;
                }
                c = d << 6;
                if ((d = parse_oct_dgt (s [2])) == -1) {
                    varstr_errno = EVARSTR_string_parser_oct_dgt_error ;
                    return 0 ;
                }
                c |= d << 3 ;
                if ((d = parse_hex_dgt (s [3])) == -1) {
                    varstr_errno = EVARSTR_string_parser_oct_dgt_error ;
                    return 0 ;
                }
            }
            
            * trg ++ = c | d ;
            s       += 4 ;
            len     -= 4 ;
            continue ;
        }
        case 'a': c = '\a' ; break ;
        case 'b': c = '\b' ; break ;
        case 'f': c = '\f' ; break ;
        case 'n': c = '\n' ; break ;
        case 'r': c = '\r' ; break ;
        case 't': c = '\t' ; break ;
        case 'v': c = '\v' ; /* break ; */
        }

        * trg ++ = c ;
        s       += 2 ;
        len     -= 2 ;
    }

    if (tsz > 0)
        * trg = '\0' ;

    return got ;
}


static
void byte_expand (IOBUF *b, unsigned char c, int always_quoted) {

    switch (c) {
    case '\'':
        data_printf (b, "'\\''");
        return ;
    case '\\':
        data_printf (b, "'\\\\'");
        return ;
    }

    if (0x20 <= c && c <= 0x80) {
        data_printf (b, "'%c'", c);
        return ;
    }

    data_printf (b, always_quoted ? "'\\x%02x'" : "0x%02x", c);
}


static
void char_comment (IOBUF *b, char c) {
    data_append (b, " /* ", 0);
    byte_expand (b, c, 1);
    data_append (b, " */", 0);
}


static
unsigned load_uint_by_size (VAR *v, const PARAMS *p, const void *data) {

    switch (p->v.size) {
    case sizeof (uint8_t):
        v->u8 = *(uint8_t*)(((uint8_t*)data) + p->v.offset) ;
        break ;
    case sizeof (uint16_t):
        v->u16 = *(uint16_t*)(((uint8_t*)data) + p->v.offset) ;
        break ;
    case sizeof (uint32_t):
        v->u32 = *(uint32_t*)(((uint8_t*)data) + p->v.offset) ;
        break ;
    case sizeof (uint64_t):
        v->u64 = *(uint64_t*)(((uint8_t*)data) + p->v.offset) ;
        break ;
    default:
        return 0;
    }

    return 1;
}


static
const uint8_t* expand_enum_type (VAR                             *v,
                                 const PARAMS                    *p,
                                 const char *(*e)(unsigned,unsigned)) {
    unsigned u ;
    const uint8_t *s ;

    switch (p->v.size) {
    case sizeof (uint8_t):
        u = v->u8 ;
        break ;
    case sizeof (uint16_t):
        u = v->u16 ;
        break ;
    case sizeof (uint32_t):
        u = v->u32 ;
        break ;
    default:
        u = v->u64 ;
    }

    varstr_errno = EVARSTR_no_error ;
    
    if ((s = (uint8_t*) (*e) (u, p->v.ext)) != 0)
        return s;

    if (varstr_errno == EVARSTR_no_error)
        varstr_errno = EVARSTR_enum_integer_value_out_of_range ;

    return 0;
}

/* ----------------------------------------------------------------------- *
 * Public: Command name and string lookup
 * ----------------------------------------------------------------------- */

int varstr_list_index (const void   *list,
                       size_t      length,
                       unsigned  dim_list,
                       size_t  width_list,
                       const char *testee,
                       size_t        tlen) {

    int inx = -1, match = -1 ;
    int null_stop = (dim_list == 0);

    if (tlen == 0)
        tlen = strlen (testee) ;
 
    if (length > 0 && tlen > length)
        return -1 ;
        
    if (null_stop)
        /* need a limit, anyway */
        dim_list = 1000 ;
    
    while (++ inx < dim_list) {

        const char *s = length > 0
            ? (const char *)list
            : *(const char **)list
            ;

        if (s != 0 && s [0] != '\0') {

            if (strncasecmp (s, testee, tlen) == 0) {
                    
                if (strlen (s) == tlen)
                    return inx ;
                    
                if (match >= 0) {
                    varstr_errno = EVARSTR_prefixed_entry_is_not_unique ;
                    return -2 ;
                }
            
                match = inx ;
            }
        }
        else if (null_stop && (length > 0 || s == 0))
            break ;
        
        list = (char*)list + width_list ;
    }

    if (match < 0)
        varstr_errno = EVARSTR_prefixed_entry_does_not_exist ;
        
    return match ;
}


unsigned varstr_param_append (IOBUF                            *b,
                              const PARAMS                     *p,
                              const void                    *data,
                              const char *(*e)(unsigned,unsigned),
                              const char *            quick_delim) {
    int n ;
    const uint8_t *e_str = 0;
    VAR var ;
    const FORMATS *f;

    if (quick_delim) {
        f = &quick ;
    }
    else {
        f =  &std ;
        quick_delim = " = ";
    }

    memset (&var, 0, sizeof (var));

    data_append (b, p->name, 0);
    data_append (b, quick_delim, quick_delim [0] ? 0 : 1);

    switch (p->v.type) {
    case t_enum:
        if (e == 0) {
            data_printf (b, f->undef_no_ext_parser_for_d, p->v.ext);
            return 0 ;
        }
        /* otherwise read scalar */
        
    case t_bool:
    case t_signed:
    case t_unsigned:
    case t_hex:

        if (load_uint_by_size (&var, p, data) == 0) {
            data_printf (b, f->undef_size_d, p->v.size);
            varstr_errno = integer_size_error (p->v.type) ;
            return 0;
        }
        break ;

    case t_string:
    case t_data:
        e_str = (((uint8_t*)data) + p->v.offset) ;
        break ;

    default:
        data_printf (b, f->undef_type_d, p->v.type);
        varstr_errno = EVARSTR_data_type_undefined ;
        return 0;
    }
  
    switch (p->v.type) {
    case t_bool:
        switch (p->v.ext) {
        case  1: n = var.bit.x1  ; break ;
        case  2: n = var.bit.x2  ; break ;
        case  3: n = var.bit.x3  ; break ;
        case  4: n = var.bit.x4  ; break ;
        case  5: n = var.bit.x5  ; break ;
        case  6: n = var.bit.x6  ; break ;
        case  7: n = var.bit.x7  ; break ;
        default:
            data_printf (b, f->undef_bit_d, p->v.ext);
            varstr_errno = EVARSTR_bool_integer_bit_not_defined;
            return 0;
        }
        data_printf (b, "%s", n ? f->true : f->false);
        break ;
        
    case t_signed:
        switch (p->v.size) {
        case sizeof (uint8_t):
            data_printf (b, "%d", (int)var.s8);
            char_comment (b, var.s8) ;
            break ;
        case sizeof (uint16_t):
            data_printf (b, "%d", (int)var.s16);
            break ;
        case sizeof (uint32_t):
            data_printf (b, "%ld", (long)var.s32);
            break ;
        case sizeof (uint64_t):
            data_printf (b, "%lld", (long long)var.s64);
        }
        break ;

    case t_unsigned:
        switch (p->v.size) {
        case sizeof (uint8_t):
            data_printf (b, "%u", (unsigned int)var.u8);
            char_comment (b, var.u8) ;
            break ;
        case sizeof (uint16_t):
            data_printf (b, "%u", (unsigned int)var.u16);
            break ;
        case sizeof (uint32_t):
            data_printf (b, "%lu", (unsigned long)var.u32);
            break ;
        case sizeof (uint64_t):
            data_printf (b, "%llu", (unsigned long long)var.u64);
        }
        break ;

    case t_hex:
        switch (p->v.size) {
        case sizeof (uint8_t):
            data_printf (b, "0x%02x", (unsigned int)var.u8);
            break ;
        case sizeof (uint16_t):
            data_printf (b, "0x%04x", (unsigned int)var.u16);
            break ;
        case sizeof (uint32_t):
            data_printf (b, "0x%08x", (unsigned long)var.u32);
            break ;
        case sizeof (uint64_t):
            data_printf (b, "0x%016x", (unsigned long long)var.u64);
        }
        break ;

    case t_string:
        n = 0;
        while (n ++ < p->v.size) {
            char *s ;
            unsigned char c = * e_str ++ ;
            switch (c) {
            case   '\0': return 1;
            case   '\a': s = "\\a"  ; break ;
            case   '\b': s = "\\b"  ; break ;
            case   '\f': s = "\\f"  ; break ;
            case   '\n': s = "\\n"  ; break ;
            case   '\r': s = "\\r"  ; break ;
            case   '\t': s = "\\t"  ; break ;
            case   '\v': s = "\\v"  ; break ;
            case   '\\': s = "\\\\" ; break ;
            default:
                if (0x20 <= c && c < 0x7f)
                    char_append (b, c);
                else
                    data_printf (b, "\\x%02x", c);
                continue ;
            }
            data_append (b, s, 0);
        }
        break ;

    case t_data:
        char_append (b, '{') ;
        if (p->v.size > 0) {
            byte_expand (b, e_str [0], 0) ;
            for (n = 1; n < p->v.size; n ++) {
                data_append (b, ", ", 0);
                byte_expand (b, e_str [n], 0) ;
            }
        }
        char_append (b, '}') ;
        break ;

    case t_enum:
        if ((e_str = expand_enum_type (&var, p, e)) == 0) {
            data_printf (b, f->undef_ext_d, p->v.ext);
            return 0;
        }
        data_printf (b, f->lu_enum_s, (long)var.u32, e_str) ;
        /* break; */

    default:
        ;
    }

    return 1 ;
}


unsigned varstr_param_load (const PARAMS                     *p,
                            void                          *data,
                            const char                   *value,
                            size_t                          len,
                            const char *(*e)(unsigned,unsigned),
                            unsigned (*f)(const char*,unsigned)) {
    int n ;
    unsigned u ;
    VAR var, v2 ;

    switch (p->v.type) {
    case t_bool:
        if (parse_numeric (&v2, t_bool, p->v.size,
                           value, len, '\0') == 0)
            return 0 ;

        load_uint_by_size (&var, p, data);

        switch (p->v.ext) {
        case  1: var.bit.x1 = v2.bool ; break ;
        case  2: var.bit.x2 = v2.bool ; break ;
        case  3: var.bit.x3 = v2.bool ; break ;
        case  4: var.bit.x4 = v2.bool ; break ;
        case  5: var.bit.x5 = v2.bool ; break ;
        case  6: var.bit.x6 = v2.bool ; break ;
        case  7: var.bit.x7 = v2.bool ; break ;
        default:
            varstr_errno = EVARSTR_bool_integer_bit_not_defined ; 
            return 0;
        }

        goto store_unsigned;
        
    case t_signed:
        if (parse_numeric (&var, p->v.type, p->v.size,
                           value, len, '\0') == 0)
            return 0;
        switch (p->v.size) {
        case sizeof (int8_t):
            *(int8_t*)(((char*)data) + p->v.offset) = var.s8 ;
            break ;
        case sizeof (int16_t):
            *(int16_t*)(((char*)data) + p->v.offset) = var.s16 ;
            break ;
        case sizeof (int32_t):
            *(int32_t*)(((char*)data) + p->v.offset) = var.s32 ;
            break ;
        case sizeof (int64_t):
            *(int64_t*)(((char*)data) + p->v.offset) = var.s64 ;
        }
        break ;

    case t_enum:
        if (parse_numeric (&var, p->v.type, p->v.size,
                           value, len, '\0') == 0) {

            if (f == 0) {
                varstr_errno = EVARSTR_enum_parser_not_available ;
                return 0 ;
            }

            /* obvously not a proper integer, so have it parsed */
            varstr_errno = EVARSTR_no_error ;
            if ((u = (*f) (value, p->v.ext)) == -1) {
                if (varstr_errno == EVARSTR_no_error)
                    varstr_errno = EVARSTR_enum_parser_returns_error ;
                return 0;
            }

            switch (p->v.size) {
            case sizeof (uint8_t):
                if ((u & ~ 0xff) == 0) {
                enum_overflow:
                    varstr_errno = EVARSTR_enum_integer_string_out_of_range ;
                    return 0;
                }
                *(uint8_t*)(((char*)data) + p->v.offset) = u ;
                break ;
            case sizeof (uint16_t):
#               if SIZEOF_UNSIGNED == 8 || \
                   SIZEOF_UNSIGNED == 4
                if ((u & ~ 0xffff) == 0)
                    goto enum_overflow ;
#               endif
                *(uint16_t*)(((char*)data) + p->v.offset) = u ;
                break ;
            case sizeof (uint32_t):
#               if SIZEOF_UNSIGNED == 8
                if ((u & ~ 0xffffffff) == 0)
                    goto enum_overflow ;
#               endif
                *(uint32_t*)(((char*)data) + p->v.offset) = u ;
                break ;
            case sizeof (uint64_t):
                *(uint64_t*)(((char*)data) + p->v.offset) = u ;
            }
            break ;
        }

        /* check enum range */
        if (e != 0 && expand_enum_type (&var, p, e) == 0)
            return 0;

        goto store_unsigned;

    case t_unsigned:
    case t_hex:
        if (parse_numeric (&var, p->v.type, p->v.size,
                           value, len, '\0') == 0)
            return 0 ;  

    store_unsigned:
        switch (p->v.size) {
        case sizeof (uint8_t):
            *(uint8_t*)(((char*)data) + p->v.offset) = var.u8 ;
            break ;
        case sizeof (uint16_t):
            *(uint16_t*)(((char*)data) + p->v.offset) = var.u16 ;
            break ;
        case sizeof (uint32_t):
            *(uint32_t*)(((char*)data) + p->v.offset) = var.u32 ;
            break ;
        case sizeof (uint64_t):
            *(uint64_t*)(((char*)data) + p->v.offset) = var.u64 ;
        }
        break ;
        
    case t_string:
        if ((n = parse_string (data, len, value, 0)) == 0)
            return 0 ;
        if (n < strlen (value)) {
            varstr_errno = EVARSTR_string_text_too_long ;
            return 0;
        }
        break;

    case t_data:
        if (len > 0)
            memcpy (data, value, len);
        break;

    default:
        varstr_errno = EVARSTR_data_type_undefined ;
        return 0;
    }

    return 1 ;
}


unsigned parse_integer (int *i, const char *s, const char *term) {
    VAR var ;

    if (parse_numeric (&var, t_signed, sizeof (int), s, 0, term) == 0)
        return 0 ;

#   if SIZEOF_INT == 2
    *i = var.s16 ;
#   elif SIZEOF_INT == 4
    *i = var.s32 ;
#   elif SIZEOF_INT == 8
    *i = var.s64 ;
#   endif

    return 1;
}


unsigned parse_unsigned (unsigned *u, const char *s, const char *term) {
    VAR var ;

    if (parse_numeric (&var, t_unsigned, sizeof (unsigned), s, 0, term) == 0)
        return 0 ;

#   if SIZEOF_UNSIGNED == 2
    *u = var.u16 ;
#   elif SIZEOF_UNSIGNED == 4
    *u = var.u32 ;
#   elif SIZEOF_UNSIGNED == 8
    *u = var.u64 ;
#   endif

    return 1;
}


int parse_oct_dgt (int c) {
    if ('0' <= c && c <= '7') return c - '0';
    return -1 ;
}

 

int parse_hex_dgt (int c) {
    if ('0' <= c && c <= '9') return c - '0';
    if ('a' <= c && c <= 'f') return c - 'a' + 10;
    if ('A' <= c && c <= 'F') return c - 'A' + 10;
    return -1 ;
}

const char *varstr_strerr (varstr_error_code e) {

    if (EVARSTR_error_bottom < e)

        switch (e) {
        case EVARSTR_data_type_undefined:
            return  "data type undefined" ;

        case EVARSTR_signed_integer_string_termination:
            return  "signed integer string termination" ;
        case EVARSTR_signed_integer_string_out_of_range:
            return  "signed integer string out of range" ;
        case EVARSTR_signed_integer_size_undefined:
            return  "signed integer size undefined" ;

        case EVARSTR_unsigned_integer_string_termination:
            return  "unsigned integer string termination" ;
        case EVARSTR_unsigned_integer_string_out_of_range:
            return  "unsigned integer string out of range" ;
        case EVARSTR_unsigned_integer_size_undefined:
            return  "unsigned integer size undefined" ;

        case EVARSTR_bool_integer_size_undefined:
            return  "bool integer size undefined" ;
        case EVARSTR_bool_integer_string_out_of_range:
            return  "bool integer string out of range" ;
        case EVARSTR_bool_integer_bit_not_defined:
            return  "bool integer bit not defined" ;

        case EVARSTR_enum_integer_size_undefined:
            return  "enum integer size undefined" ;
        case EVARSTR_enum_integer_string_out_of_range:
            return  "enum integer string out of range" ;
        case EVARSTR_enum_integer_value_out_of_range:
            return  "enum integer value out of range" ;
        case EVARSTR_enum_parser_not_available:
            return  "enum parser not available" ;
        case EVARSTR_enum_parser_returns_error:
            return  "enum parser returns error" ;

        case EVARSTR_string_text_too_long:
            return  "string text too long" ;
        case EVARSTR_string_parser_hex_dgt_error:
            return  "string parser hex dgt error" ;
        case EVARSTR_string_parser_oct_dgt_error:
            return  "string parser oct dgt error" ;

        case EVARSTR_prefixed_entry_is_not_unique:
            return  "prefixed entry is not_ nique" ;
        case EVARSTR_prefixed_entry_does_not_exist:
            return  "prefixed entry does not exist" ;

            /* avoid compiler warnings */
        case EVARSTR_no_error:
        case EVARSTR_error_bottom:
        case EVARSTR_error_top:
            ;
        }

    return strerror (e);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
