/*
 * nvramd - active data base server and flash memory cache
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: list.h 3496 2006-10-31 19:08:13Z jordan $
 */

#ifndef __NVRAMD_LIST_H
#define __NVRAMD_LIST_H

typedef enum _action_type {
    no_trigger      =   0,
    access_trigger  = ')',
    execute_trigger = '!',
    symlink_trigger = '?',
    dont_touch_trg  = '+'
} ACTION_TYPE ;

/* import flags usage (applies to list load):

   clear_ok:      flush list before reloading
   flags_ok:      redefine flags when available
   value_ok:      redefine value when available
   trigger_ok:    redefine link/trigger/access data when available
   volatile_ok:   apply other flags to volatile variables, as well
   template_ok:   apply templates after loading
   
 */

#define IMPORT_FLAGS(x) \
    x (clear_ok)        \
    x (flags_ok)        \
    x (value_ok)        \
    x (trigger_ok)      \
    x (volatile_ok)     \
    x (template_ok)

typedef enum _import_bits {
    import_no_bit = 0
#   define        __bits(a) , import_bit_ ## a
    IMPORT_FLAGS (__bits)
#   undef         __bits
} IMPORT_BITS ;

typedef enum _import_flags {
#   define        __flags(a) import_ ## a = (1 << (import_bit_ ## a - 1)),
    IMPORT_FLAGS (__flags)
#   undef         __flags
} IMPORT_FLAGS ;

#include "util.h"
#include "do.h"

#ifdef __NVRAM_LIST_PRIVATE

/* action flags usage:

   miscellaneous flags (applies to variables):
   -------------------------------------------
   volatile_data:    does not invoke any auto save mechanism when writing
   template_prefix:  apply data set before creating a variable
   function_ref:     value is a function name that handles read/write

   access flags (applies to variables, triggers or links):
   -------------------------------------------------------
   private_ok:       access is restricted to root, only
   public_read:      public variable read access, otherwise restricted
   public_write:     public variable write access, otherwise restricted
   public_modify:    public link/trigger write or variable access
                     modification, otherwise restricted

   selection/filter flags (applies to trigger or link):
   ----------------------------------------------------
   enable_write:     is activated when the variable changes
   enable_read:      is activated when the variable is read
   
   activation mode flags (applies to trigger or link):
   ---------------------------------------------------
   link_exec_ok:     trigger is an executable link target
   exec_ok:          trigger or link is executable

   activation time flags (applies to trigger or link):
   ---------------------------------------------------
   force_immediate:  execute immediately (overwrites delayed_run)
   delayed_run:      execute later on
*/


#define ACTION_FLAGS(x)   \
    x (private_ok)        \
    x (public_read)       \
    x (public_write)      \
    x (public_modify)     \
    x (volatile_data)     \
    x (exec_ok)           \
    x (link_exec_ok)      \
    x (enable_write)      \
    x (enable_read)       \
    x (force_immediate)   \
    x (delayed_run)       \
    x (template_prefix)   \
    x (function_ref)

typedef enum _action_bits {
    action_no_bit = 0
#   define        __bits(a) , action_bit_ ## a
    ACTION_FLAGS (__bits)
#   undef         __bits
} ACTION_BITS ;

typedef enum _action_flags {
#   define        __flags(a) action_ ## a = (1 << (action_bit_ ## a - 1)),
    ACTION_FLAGS (__flags)
#   undef         __flags
} ACTION_FLAGS ;


#define NO_FLAGS_SYMBOL no_flags


typedef enum _action_constants {
    action_field_delim =  ':',
    action_flag_delim  =  ',',
    escape_character   = '\\',
} ACTION_CONSTANTS ;


typedef struct _child {
    unsigned   slot_inx ;
    pid_t         child ; /* same as AVPAIR->t.execute.child */
    struct _child *next ;
} CHILD ;


typedef struct _avpair {

    struct _avcommon {
        int          flags ; /* access and trigger flags */
        unsigned       inx ; /* index into the slots table */
        const char    *val ; /* assignment value */
        char  trigger_type ; /* type of the variant, below */
        int just_created:1 ; /* temporary flag */
    } a ;

    union _trigger {    /* trigger when data are rewritten */

        struct _execute {
            uid_t               uid ;
            gid_t               gid ;
            int                argc ;
            pid_t             child ; /* only one child at a time */
            unsigned       link_inx ; /* last time triggered by this link */
            time_t         lifetime ; /* seconds of life */
            time_t         time2die ; /* time to die */
            struct timeval time2run ; /* time to run */
            const char       **argv ;
            const char   chroot [1] ; /* variable string length */
        } execute ;

        struct _symlink {
            const char      ref [1] ; /* variable string length */
        } symlink ;
        
    } t ;

    /* variable size */

} AVPAIR ;
#endif


typedef struct _list
#ifdef __NVRAM_LIST_PRIVATE
{
    IOBUF                   data ; /* data unload buffer */
    IOBUF                    var ; /* variables name buffer */
    struct _list_entries {
        unsigned         entries ;
        AVPAIR           ***slot ;
    } list ;
    CHILD              *children ;
    HLST                 *lookup ;
    HLST               *function ;
    PLST               *template ;
    DO                    *sched ;
    struct timeval   time2commit ; /* next time to save data base */
    struct timeval    time2check ; /* next time to do something */
    struct {
        unsigned wait_after_kill ; /* ms */
        unsigned    nvram_commit ; /* ms */
        unsigned trigger_wake_up ; /* ms */
    } lazy ;
}
#endif
LIST;


#ifdef __NVRAM_LIST_PRIVATE
typedef struct _valfun {
    void *fstate ;
    const char* (*set) (LIST*,void*,const char*,int,const char*);
    void (*free)(void*);
} VALFUN ;
#endif

extern LIST        *list_open           (DO*,unsigned,unsigned,unsigned);
extern void         list_close          (LIST*);
extern void         list_flush          (LIST*);

extern unsigned     list_fregister      (LIST*,const char*,void*,
                                         const char*(*)(LIST*,void*,
                                                        const char*,int,
                                                        const char*),
                                         void(*)(void*),
                                         const char*,unsigned);

extern unsigned     list_store          (LIST*);
extern unsigned     list_load           (LIST*,const char*,int mode);

extern unsigned     list_get_slot       (LIST*,const char*);
extern const char  *list_get_var        (LIST*,int,const char*,unsigned);
extern const char  *list_get_val        (LIST*,int,const char*,pid_t);
extern const char **list_get_trigger    (LIST*,const char*,ACTION_TYPE);

extern unsigned     list_avp_replace    (LIST*,int,const char*,const char*,
                                         pid_t,int);
extern unsigned     list_exe_replace    (LIST*,int,const char*,const char*,
                                         uid_t,gid_t,const char*,unsigned,
                                         const char**,unsigned);
extern unsigned     list_lnk_replace    (LIST*,int,const char*,const char*,
                                         const char*,unsigned);
extern unsigned     list_access_replace (LIST*,int,const char*,const char*,
                                         unsigned);

extern unsigned     list_used_slots     (LIST*);
extern unsigned     list_num_slots      (LIST*);
extern void         list_do_events      (LIST*);

#endif /* __NVRAMD_LIST_H */
