/*
 * nvramd - active data base server and flash memory cache
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: nvram.c 3175 2006-09-23 19:17:11Z jordan $
 *
 *
 * simple stupid example of a command line client
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <libgen.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#define _GNU_SOURCE
#include <getopt.h>

#include "varstr.h"
#include "nvalloc.h"
#include "nvram-lib.h"
#include "settings.h"
#include "logger.h"

/* ----------------------------------------------------------------------- *
 * variables and definitions
 * ----------------------------------------------------------------------- */

static NVCL *nvcl;

static int listmost  =  0 ;
static int listall   =  0 ;
static int reload    =  0 ;
static int store     =  0 ;
static char *del     =  0 ;
static char *cmd     =  0 ;
static char *filearg =  0 ;

#define __str(x)              # x
#define str(x)                __str (x)

/* ----------------------------------------------------------------------- *
 * variable, definitions
 * ----------------------------------------------------------------------- */

static const
struct option long_options [] = {
    {"help",      0, 0, 'h'},
    {"list-most", 0, 0, 'l'},
    {"list-all",  0, 0, 'L'},
    {"delete",    1, 0, 'd'},
    {"file",      1, 0, 'f'},
    {"command",   1, 0, 'c'},
    {"reload",    0, 0, 'r'},
    {"store",     0, 0, 's'},
    {"raw-b64",   0, 0, 'b'},
#   ifndef DISCARD_HTTP
    {"server",    1, 0, 'S'},
    {"port",      1, 0, 'P'},
#   endif
    {"verbose",   0, 0, 'v'},
    {"pid",       0, 0, 'p'},
    {"timeout",   0, 0, 't'},
    {0, 0, 0, 0}
};

static const
char short_options [] = ("+brslLhvc:d:f:"
#    ifndef DISCARD_HTTP
                         "S:P:"
#    endif
                         "p:t:");

typedef
enum {
    o_none     = 0,
    o_command  = 1,
    o_delete   = 2,
    o_store    = 3,
    o_reload   = 4,
    o_listmost = 5,
    o_listall  = 6,
    o_filearg  = 7,
} OPT_TYPE ;

static
char *opt_blurb [] = {
    0,              /* 0 == o_none     */
    "command/-c",   /* 1 == o_command  */
    "delete/-d",    /* 2 == o_delete   */
    "store/-s",     /* 3 == o_store    */
    "reload/-r",    /* 4 == o_reload   */
    "list-most/-l", /* 5 == o_listmost */
    "list-all/-L",  /* 6 == o_listall  */
    "file/-f",      /* 7 == o_filearg  */
};


/* ----------------------------------------------------------------------- *
 * helpers
 * ----------------------------------------------------------------------- */

static
void signal_exit (int unused) { 
    nvcl_close (nvcl);
    exit (2);
}


static
void usage (const char *s) {

    if (s != 0)
        fprintf (stderr, "%s\n\n", s);

    fprintf
        (stderr,
         "Usage: %s [options] [--] [<variable> [<assign-value>]]\n\n"
         "Options: -l,--list-most  list nvram entries\n"
         "         -L,--list-all   list all nvram entries, incl. templates\n"
         "         -c,--command    send raw argument list as command\n"
         "         -s,--store      commit changes, write to nvram\n"
         "         -f,--file=FILE  load <assign-value> from file\n"
         "         -r,--reload     undo changes, reload from nvram\n"
         "         -d,--delete     delete variable\n"
         "         -b,--raw-b64    do not decode base 64 to binary data\n"
         "         -p,--pid=NUM    send this number as pid with the "
         ""                        "assignment or delete command\n"
#       ifndef DISCARD_HTTP
         "         -P,--port       nvram server HTTP port\n"
         "         -S,--server     nvram server HTTP address\n"
#       endif
         "         -v,--verbose    print debugging information\n"
         "         -t,--timeout=MS wait at most MS milli seconds "
         ""                        "(default: " str (NVCL_REPLY_TMO) ")\n"
         "         -h,--help       print this message\n",
         progname);

    exit (2);
}

static
void no_args_with (const char *s) {
#   define __blurb "No arguments allowed with --%s"
    char *t = (char*) xmalloc (strlen (s) + sizeof (__blurb));
    sprintf (t, __blurb, s);
    usage (t); /* no return */
#   undef __blurb
}

static
void set_signals (void) {
    signal (SIGINT,  signal_exit);
    signal (SIGHUP,  signal_exit);
    signal (SIGTERM, signal_exit);
    signal (SIGUSR1, signal_exit);
    signal (SIGUSR2, signal_exit);
}


OPT_TYPE collision_type (OPT_TYPE opt) {
    if (cmd      != 0 && opt != o_command)  return o_command ;
    if (del      != 0 && opt != o_delete)   return o_delete  ;
    if (store    != 0 && opt != o_store)    return o_store   ;
    if (reload   != 0 && opt != o_reload)   return o_reload  ;
    if (listmost != 0 && opt != o_listmost) return o_listmost ;
    if (listall  != 0 && opt != o_listall)  return o_listall ;
    if (filearg  != 0 && opt != o_filearg)  return o_filearg ;
    return o_none ;
}

static
void check_collision (OPT_TYPE opt) {
    char *s ;

    if ((s = opt_blurb [collision_type (opt)]) != 0) {
#       define ERRTXT "The --%s option is not allowed with --%s"
        char *u = opt_blurb [opt];
        char *t = (char*) xmalloc (sizeof (ERRTXT)+strlen (u)+strlen (s));
        sprintf ((char*)t, ERRTXT, u, s);
#       undef ERRTXT
        usage (t);
    }
}

/* ----------------------------------------------------------------------- *
 * private methods
 * ----------------------------------------------------------------------- */

static
int list_entries (int all) {

    const char *var, *val;
    int slot, used, total ;

    nvcl_putcmd (nvcl, "slots");
    if (nvcl_rply (nvcl, 0) < 1)
        return 1;

    if (nvcl_getavpair (nvcl, 0, &val, 0, 0) == 0)
        return 1;
    used = atoi (val);

    if (nvcl_getavpair (nvcl, 0, &val, 0, 0) == 0)
        return 1;
    total = atoi (val);
    
    for (slot = 1; slot <= total && used >= 0; slot ++) {
        char t [20];

        sprintf (t, "%u", slot);
        nvcl_putcmd    (nvcl, all ? "get !no_flags" : "get");
        nvcl_putavpair (nvcl, t, 0, 0);

        if (nvcl_rply (nvcl, 0) < 1)
            continue ;

        while (nvcl_getavpair (nvcl, &var, &val, 0, 0) != 0) {
            printf ("%s=%s\n", var, val);
            used -- ;
        }
    }
    return 0;
}


static
char *load_file (size_t *Len, const char *file) {

    char *s ;
    int fd, n, not_stdin = strcmp (file, "-") ;

    if (not_stdin) {
        struct stat st ;

        if ((fd = open (file, O_RDONLY)) < 0) {
            logger (LOG_ERR, "cannot open file %s: %s",
                    file, strerror (errno));
            exit (2);
        }

        if (fstat (fd, &st) < 0) {
            logger (LOG_ERR, "cannot stat file %s: %s",
                    file, strerror (errno));
            exit (2);
        }

        if (st.st_size > NVCL_MAX_VALUE_SIZE) {
            logger (LOG_ERR, "File %s is too large (max %u bytes)",
                    file, NVCL_MAX_VALUE_SIZE);
            exit (2);
        }

        /* add \0 terminator */
        s = (char*)xmalloc (st.st_size + 1);
        
        if ((n = read (fd, s, st.st_size)) < st.st_size) {
            logger (LOG_ERR, "error reading file %s: %s",
                    file, n < 0 ? strerror (errno) : "short read");
            exit (2);
        }

        close (fd);
    }
    else {
        int left = NVCL_MAX_VALUE_SIZE + 1;
        
        if ((fd = fileno (stdin)) < 0) {
            logger (LOG_ERR, "cannot read from stdin - no file number: %s",
                    strerror (errno));
            exit (2);
        }

        s = (char*)xmalloc (left);

        n = 0 ;
        do {
            int m = read (fd, s + n, left) ;
            if (m <= 0) {
                if (m == 0)
                    break ;
                logger (LOG_ERR, "error reading from stdin: %s",
                        strerror (errno));
                exit (2);
            }
            n    += m ;
            left -= m ;
        }
        while (left > 0);

        if (n > NVCL_MAX_VALUE_SIZE) {
            logger (LOG_ERR, "Too much data from stdin (max %u bytes)",
                    NVCL_MAX_VALUE_SIZE);
            exit (2);
        }
    }
    
    s [n] = 0;

    if (Len != 0)
        *Len = n ;

    return s;
}

/* ----------------------------------------------------------------------- *
 * MAIN
 * ----------------------------------------------------------------------- */

int main (int argc, char *argv[]) {

    int n, c, option_inx ;
    int raw = 0;
    int use_pid = -1 ;
    unsigned u, data_len = 0;
    char *root  = "" ;
    unsigned  tmo_ms = NVCL_REPLY_TMO ;
#   ifndef DISCARD_HTTP
    struct in_addr ip ;
    ip4addr_t server = 0 ;
    uint16_t    port = 0 ;
#   endif

#   ifdef WITH_DMALLOC
    dmalloc_debug_setup ("debug=0x4f47d03,log=/tmp/dmalloc.log");
#   endif

    syslog_init (0, 0);

    /* progname ::= only the last path segment */
    if ((progname = strrchr (argv [0], '/')) == 0)
        progname = argv [0] ;
    else
        ++ progname ;

    while ((c = getopt_long
            (argc, argv, short_options, long_options, &option_inx)) != -1)
        switch (c) {
        case 'b':
            raw = 1 ;
            continue ;
        case 'c':
            check_collision (o_command);
            cmd = optarg ;
            continue ;
        case 'd':
            check_collision (o_delete);
            del = optarg ;
            continue ;
        case 'v':
            verbosity = -1;
            continue ;
        case 'l':
            check_collision (o_listmost);
            listmost = 1;
            continue ;
        case 'L':
            check_collision (o_listall);
            listall = 1;
            continue ;
        case 'f':
            check_collision (o_filearg);
            filearg = load_file (&data_len, optarg) ;
            continue ;
        case 'r':
            check_collision (o_reload);
            reload = 1;
            continue ;
        case 's':
            check_collision (o_store);
            store = 1;
            continue ;
        case 't':
            if (parse_unsigned (&u, optarg, 0) == 0 || (u & ~0xffff) != 0)
                usage ("Timeout parser error with --pid/-p") ;
            tmo_ms = u ;
            continue ;
#       ifndef DISCARD_HTTP
        case 'p':
            if (parse_unsigned (&u, optarg, 0) == 0 || (u & ~0xffff) != 0)
                usage ("Pid parser error with --pid/-p") ;
            use_pid = u ;
            continue ;
        case 'P':
            if (parse_unsigned (&u, optarg, 0) == 0 || (u & ~0xffff) != 0)
                usage ("HTTP port parser error with --port/-P") ;
            port = u ;
            continue ;
        case 'S':
            if (inet_aton (optarg, &ip) < 0)
                usage ("HTTP address parser error with --address/-S") ;
            server = ntohl (ip.s_addr) ;
            continue ;
#       endif    
        case 'h':
        case '?':
            usage (0);
        }

    if ((nvcl = nvcl_open (root,
#                          ifndef DISCARD_HTTP
                           server, port,
#                          endif
                           tmo_ms)) == 0)
        exit (2);

    set_signals ();

    if (raw)
        nvcl_base64_ok (nvcl, 0);

    /* store arguments */
    if (reload != 0) {
        if (optind < argc) no_args_with ("reload/-r!");
        nvcl_putcmd (nvcl, "load");
    }
    else if (store != 0) {
        if (optind < argc) no_args_with ("store/-s!");
        nvcl_putcmd (nvcl, "store");
    }
    else if (del) {
        if (optind < argc) no_args_with ("delete/-d!");
        nvcl_putcmd (nvcl, "del");
        nvcl_putavpair (nvcl, del, 0, 0);
    }
    else if (listmost) {
        if (optind < argc) no_args_with ("list-most/-L!");
        nvcl_base64_ok (nvcl, 0);
        exit (list_entries (0));
    }
    else if (listall) {
        if (optind < argc) no_args_with ("list-all/-l!");
        nvcl_base64_ok (nvcl, 0);
        exit (list_entries (1));
    }
    else {
        if (cmd != 0) {
            nvcl_base64_ok (nvcl, 0);
            nvcl_putcmd (nvcl, cmd);
        }
        else {
            if (argc <= optind)
                usage ("Arguments missing!");
            nvcl_putcmd (nvcl, "set");
        }
        if (use_pid >= 0) {
            char buf [3 * sizeof (use_pid) + 1] ;
            sprintf (buf, "#%u", use_pid);
            nvcl_putavpair (nvcl, buf, 0, 0);
        }
        if (filearg != 0) {
            if (argc != optind + 1)
                usage ("No <assign-values> with --file/-f!");
            nvcl_putavpair (nvcl, argv [optind], filearg, data_len);
        }
        else {
            for (n = optind; n < argc; n += 2)
                nvcl_putavpair
                    (nvcl, argv [n], (n+1 < argc) ? argv [n+1] : 0, 0);
        }
    }
    
    if ((n = nvcl_rply (nvcl, 0)) > 0) {
        size_t vlen ;
        const char *var, *val;
        while (nvcl_getavpair (nvcl, &var, &val, &vlen, 0) != 0)
            if (cmd == 0)
                fwrite (val, vlen, 1, stdout);
            else if (*val)
                printf ("%s=%s\n", var, val);
            else
                printf ("%s\n", var);
        exit (0);
    }
    
    fprintf (stderr, "%s: %s\n", progname, nvcl_rply_error (nvcl));
    exit (1);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
