#include "printserv.h"

#define TCP_PORT		9100
#define DEFAULT_PRINTER		"/dev/usb/lp0"
#define LOCK_FILE		"/var/lock/printserv"
#define MAX_NAME		100

volatile sig_atomic_t child_count;      /* number of kids forked */
char _print_interface[MAX_NAME];
int  _tcp_port;

void usage(){
	
	printf	("\nusage: printserv [OPTION]\n"
		 "Options:\n"
		 "-d printer    Specifies the printer device  [%s]\n"
		 "-p port       Specifies the listen tcp port [%d]\n" 
		 "-h            This infos\n", 
		 DEFAULT_PRINTER, TCP_PORT);

}

int set_args(int argc, char  **argv){

 	extern char *optarg;
        extern int optind;
	int  option;
	int  ret = 1;

	/* set defaults */
	strcpy(_print_interface, DEFAULT_PRINTER); 
	_tcp_port = TCP_PORT;	

	do {
		option = getopt(argc, argv, "hd:p:");
		switch(option){
		case 'd':	strncpy(_print_interface,optarg,MAX_NAME);
				break;

		case 'p':	_tcp_port = atoi(optarg);
				break;
		
		case -1:	break;
		
		default:	usage();
				ret = 0;
				break;
		}
	}while(option > 0);

	return ret;
}


void limit_childs(int child_max)
{
	int sleeptime = 10;

 	while (child_max < child_count) {
        	syslog(LOG_WARNING,
			"too many children, sleeping for %d seconds",
		                            sleeptime);
		sleep(sleeptime);
	}
}

static void reapchild(int signo)
{
        int save_errno = errno;
	int status;

	while (waitpid((pid_t)-1, &status, WNOHANG) > 0)
		child_count--;
	errno = save_errno;
}

static void finalize(int signo)
{
	kill(0,SIGTERM);
	exit(0);
}

int set_lock(int fd, int kdo, int lock_type, 
		off_t offset, int whe, off_t length)
{
	struct flock	lock;

	lock.l_type	= lock_type;
	lock.l_start	= offset;
	lock.l_whence	= whe;
	lock.l_len	= length;

	return( fcntl(fd, kdo, &lock) );
}


int daemoninit()
{
	pid_t	pid;

#if 0
	if( (pid = fork()) < 0 )
		return -1;
	else if (pid != 0)
		exit(0);
#endif
	setsid();

	chdir("/");

	umask(0);

//	openlog("printserv", LOG_PID, LOG_DAEMON);

	return(0);
}
		


ssize_t readn(int fd, void *vptr, size_t n)
{
	size_t		nleft;
	ssize_t		nread;
	char		*ptr;

	ptr = vptr;
	nleft = n;

	while (nleft > 0){
		if ( (nread = read(fd, ptr, nleft)) < 0) {
			if (errno == EINTR)
				nread = 0;	/* and call read() again */
			else
				return (-1);
		}else if (nread == 0)
			break;			/* EOF */

		nleft -= nread;
		ptr += nread;
	}

	return (n - nleft);			/* return >= 0 */
}


void doit(int fd)
{
	char buff[MAXLINE];
	int read;
	int lfd, pfd;
	int sum = 0;
	

	
	if( (lfd = open(LOCK_FILE, O_RDWR | O_CREAT, 660)) < 0 ){
		syslog(LOG_ERR,"can't open lockfile: %m");
		exit(1);
	}
	
	/* block until we get our lock or exit if it failed */
	if (set_writelock(lfd)<0){
		syslog(LOG_ERR,"can't lock the lockfile: %m");
		exit(1);
	}

	/* open the printer device */
	if((pfd = open(_print_interface, O_WRONLY)) < 0 ){
		syslog(LOG_ERR,"can't open  printer interface: %m");
		exit(1);
	}
#if 0
	if((pfd = open("/dev/null", O_WRONLY)) < 0 ){
		syslog(LOG_ERR,"can't open  printer interface: %m");
		exit(1);
	}
#endif	
	
	while((read = readn(fd,buff, MAXLINE)) > 0){
	
		if (write(pfd, buff, read) != read){
			syslog(LOG_ERR,"error while writing to printer:%m");
			break;
		}

		sum += read;
	}

	syslog(LOG_INFO, "%d bytes received and printed", sum);
	close(pfd);
	
	if (release_lock(lfd)<0){
		syslog(LOG_ERR,"failed to unlock the lock file: %m");
	}
	close(lfd);
}



int main(int argc, char **argv)
{
	int serversock, connfd;
	socklen_t len;
	int child_max = 32;
	struct sockaddr_in servaddr, cliaddr;
	pid_t	pid;

	
	if( !set_args(argc, argv) )
		exit(1);
	

	syslog(LOG_INFO,"started with printer: %s and listen port: %d...",
			_print_interface, _tcp_port);

	if( daemoninit() != 0){
		syslog(LOG_ERR, "error while daemoninit(): %m");
		exit(1);
	}
	
	signal(SIGCHLD, reapchild);
	signal(SIGTERM, finalize);
	child_count = 0;

 	/* Create the TCP socket */
        if ((serversock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		syslog(LOG_ERR,"Failed to create socket: %m");
		exit(1);
	}


	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(_tcp_port);

	/* Bind the server socket */
	if (bind(serversock, (SA *) &servaddr,
			sizeof(servaddr)) < 0) 
	{
		syslog(LOG_ERR,"Failed to bind the server socket: %m");
		exit(1);
	}

	/* Listen on the server socket */
	if (listen(serversock, LISTENQ) < 0) {
		syslog(LOG_ERR,"Failed to listen on server socket: %m");
		exit(1);
	}

	len = sizeof (struct sockaddr_in);
	memset (&cliaddr, 0, len); 
	
	while( 1 ) {
		
		limit_childs(child_max);
		
		if(((connfd = accept(serversock, (SA *) &cliaddr, &len)) < 0)) {
			syslog(LOG_ERR,"Failed to accept client connection: %m");
			exit(1);
		}
		
		syslog(LOG_INFO, "Client from %s accepted",  inet_ntoa(cliaddr.sin_addr)); 

		if( (pid = fork()) == 0){
			close(serversock);
			doit(connfd);
			exit(0);
		}
		
		close(connfd);
		child_count++;
	}

	return 0;
}
