/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include "util.h"
#include "ctrl.h"
#include "varstr.h"
#include "buffer.h"
#include "nvalloc.h"
#include "settings.h"
#include "logger.h"
#include "tvcalc.h"

#define __COMHANDLER_PRIVATE
#include "comhandler.h"

/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
void startstop (void       *state,
                int          argc,
                const char **argv,
                const char*(*setup)(FRUN_SPEC*,const char*),
                int      force_nl,
                const char  *expr,
                const char  *info,
                int        append){

    COMM *o = state;
    const char *me = argv [0], *s;
    FRUN_SPEC *desc ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 1:
        goto usage ;
    case 2:
        desc = filter_find (o->freg, argv [1]);
        break ;
    default:
        desc = filter_spec (o->freg, argv [1]);
    }

    if (desc == 0) {
        ctrl_reply (o->ctrl, "-ERR: undefined handle \"%s\"\n", argv [1]);
        return ;
    }
    
    if (argc > 2) {
        int n ;
        IOBUF  pat ;
        XZERO (pat) ;

        if (append) {
            if ((s = (*setup) (desc, 0)) == 0)
                goto filter_error ;
            data_append (&pat, s, 0);
        }

        data_append (&pat, argv [2], 0);
        for (n = 3; n < argc; n ++) 
            data_printf (&pat, " %s", argv [n]);

        if (force_nl && pat.data [pat.top - 1] != '\n')
            char_append (&pat, '\n');

        s = (*setup) (desc, pat.data);
        kill_space (&pat);

        if (s)
            ctrl_reply (o->ctrl, "+OK");
        else
            goto filter_error ;
    }
    else if ((s = (*setup) (desc, 0)) == 0)
        ctrl_reply (o->ctrl, "-ERR: no %s for \"%s\"", info, argv [1]);
    else
        ctrl_reply (o->ctrl, "+OK\n%s", s);

    return ;

filter_error:
    ctrl_reply (o->ctrl, "-ERR: %s", filter_error (desc));
    return ;
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply
        (o->ctrl, "%s <key> [<%s> ... ] -- %s/show %s\n",
         me, expr, append ? "append" : "set", info);
}


static
unsigned msnap (COMM *o, int ms) {

    if (ctrl_reply (o->ctrl, 0) != -1) {
        struct timeval nap;

        if (ms == 0)
            return 1 ;

        nap.tv_sec  =  ms / 1000 ;
        nap.tv_usec = (ms % 1000) * 1000 ;
        select (0, 0, 0, 0, &nap) ;

        if (ctrl_reply (o->ctrl, 0) != -1)
            return 1 ;
    }

    logger
        (LOG_ERR, "comhandler: broken client connection", strerror (errno));
    return 0;
}


static
FILE *open_file (COMM *o, const char *name) {
    int fd ;
    FILE *fp ;
    
    if ((fd = open (name, O_CREAT|O_EXCL|O_WRONLY, 0400)) < 0) {
        ctrl_reply (o->ctrl, "-ERR: cannot create %s for writing: %s",
                    name, strerror (errno));
        return 0;
    }
    if ((fp = fdopen (fd, "w")) == 0) {
        ctrl_reply (o->ctrl, "-ERR: cannot open %s for writing: %s",
                    name, strerror (errno));
        return 0;
    }

    return fp ;
}

const char *ftmpl_load (FRUN_SPEC *q, const char *fmt) {
    return filter_tmpl (q, fmt, 0);
}

/* ----------------------------------------------------------------------- *
 * Public: Connectionless communication handlers
 * ----------------------------------------------------------------------- */

COMM *comm_init (SHELL *s, FILTER *f) {

    COMM *o = ZMALLOC (COMM);   

    if ((o->shell = s) == 0) {
        logger (LOG_ERR, "comhandler: SHELL subsytem not initialised!");
        return 0;
    }
    if ((o->freg = f) == 0) {
        logger (LOG_ERR, "comhandler: FILTER subsytem not initialised!");
        return 0;
    }

    return o ;
}


void comm_close (COMM *o) {
    
    if (o) {
        free (o);
    }
}


COMM *comm_update (COMM *o, CTRL *c) {
    o->started = time (0);
    o->ctrl    = c ;
    return o;
}


void comm_status (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 1) goto usage;

    {
        time_t uptime;
        size_t in, out;
        const char *rc, *rl;
        unsigned req;
        unsigned kpa;

        shell_status (o->shell, &uptime, &in, &out, &rl, &rc, &req, &kpa);

        ctrl_reply (o->ctrl, "uptime:        ");
        if (uptime > 60) {
            int min = uptime / 60 ;
            int h   =    min / 60 ;
            if (h > 24) {
                ctrl_reply (o->ctrl, "%dd ", h / 24);
                h %= 24 ;
            }
            ctrl_reply (o->ctrl, "%d:%02d:%02d\n", h, min % 60, uptime % 60);
        }
        else
            ctrl_reply (o->ctrl, "% 6ds\n", uptime);

        if (rl)
            ctrl_reply (o->ctrl, "local init script: %s\n", (int)rl);
        if (rc)
            ctrl_reply (o->ctrl, "rc script sent:    %s\n", (int)rc);

        ctrl_reply (o->ctrl, "requests:       % 6d\n", req);
        ctrl_reply (o->ctrl, "bytes sent:     % 6d\n", (int)out);
        ctrl_reply (o->ctrl, "bytes received: % 6d\n", (int)in);
        ctrl_reply (o->ctrl, "keep-alive:        %s\n", kpa ? " on" : "off");

        return ;
    }
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, "%s -- print status\n", me);
}


void comm_file (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0], *s;
    unsigned timeout = COMM_WAIT_REPLY_DEFAULT, inx ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 2:
        inx = 1 ;
        break ;
    case 3:
        if (parse_unsigned (&timeout, argv [1], 0)) {
            inx = 2 ;
            break ;
        }
    default:
        goto usage;
    }

    if ((s = shell_file (o->shell, argv [inx], 0, 0, timeout)) == 0)
        ctrl_reply (o->ctrl, "-ERR: %s", shell_error (o->shell));
    else
        ctrl_reply (o->ctrl, "+OK\n%s", s);

    return;
 
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl,
                "%s [<timeout>] <file> -- run command(s) from file\n", me);
}


void comm_string (void *state, int argc, const char **argv) {

    COMM *o = state;
    IOBUF b ;
    const char *me = argv [0], *s;
    unsigned timeout = COMM_WAIT_REPLY_DEFAULT, inx;
    int n ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc < 2) goto usage;

    inx = (argc > 2 &&
	   isdigit (argv [1][0]) &&
	   parse_unsigned (&timeout, argv [1], 0))
        ? 2
        : 1
        ;

    /* concatenate ... */
    XZERO (b);
    data_append (&b, argv [n = inx], 0);
    while (++ n < argc) {
        char_append (&b, ' ');
        data_append (&b, argv [n], 0);
    }
    char_append (&b, '\n');
    
    /* run command string */
    s = shell_data (o->shell, b.data, b.top, 0, 0, timeout) ;
    free (b.data);
    
    if (s == 0)
        ctrl_reply (o->ctrl, "-ERR: %s", shell_error (o->shell));
    else
        ctrl_reply (o->ctrl, "+OK\n%s", s);

    return;
 
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, "%s [<timeout>] <cmd> ... -- run command\n", me);
}


void comm_flush (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 1) goto usage;

    shell_flush (o->shell);
    ctrl_reply (o->ctrl, "+OK");
    return;
 
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s -- clean up pending output\n", me);
}


void comm_reopen (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 1) goto usage;

    if (shell_restart (o->shell))
        ctrl_reply (o->ctrl, "+OK");
    else
        ctrl_reply (o->ctrl, "-ERR: %s", shell_error (o->shell));

    return;
 
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s -- reconnect ssh/shell\n", me);
}


void comm_ping (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0] ;
    unsigned retry, tmo = COMM_WAIT_EREQ_DEFAULT ;
    struct timeval endup;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 1:
        retry = 0 ;
        break ;
    case 2:
        if (isdigit (argv [1][0]) != 0 &&
	    parse_unsigned (&tmo, argv [1], 0) != 0) {
            retry = 1 ;
            tval_nxt_ms (&endup, tmo);
            break ;
        }
    default:
        goto usage;
    }

    for (;;)
        /* run command string */
        switch (shell_contest (o->shell, tmo, 1)) {
        case 0:
            if (retry && (tmo = ms_tmo_tv (&endup)) > 0)
                continue ;
            ctrl_reply (o->ctrl, "-ERR: %s\n", shell_error (o->shell));
            return;
        case 1:
            ctrl_reply (o->ctrl, "+OK");
            return;
        default:
            if (retry) {
                if ((tmo = ms_tmo_tv (&endup)) > 0)
                    continue ;
                ctrl_reply (o->ctrl, "-ERR: no connection, yet\n");
            }
            else
                ctrl_reply (o->ctrl, "+OK: reconnecting ...");
            return ;
        }
 
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl,
                "%s [<timeout>] -- verify ssh/shell connection\n", me);
}


void comm_kpalve (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    unsigned keep_alive ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 1:
        keep_alive = -1 ;
        break ;
    case 2:
        if (parse_unsigned (&keep_alive, argv [1], 0) != 0) {
            keep_alive = (keep_alive != 0) ; /* in case it becomes -1 */
            break ;
        }
    default:
        goto usage;
    }

    ctrl_reply (o->ctrl, "+OK\n%#x", shell_kpalve (o->shell, keep_alive));
    return ;

usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, "%s [0|1] -- set/show keep alive mode\n", me);
}


void comm_atpl (void *state, int argc, const char **argv) {
    startstop (state, argc, argv, ftmpl_load, 1,
               "command-template", "command template", 1);
}

void comm_tmpl (void *state, int argc, const char **argv) {
    startstop (state, argc, argv, ftmpl_load, 1,
               "command-template", "command template", 0);
}


void comm_stop (void *state, int argc, const char **argv) {
    startstop (state, argc, argv, filter_stop, 0,
               "regular-expression", "filter stop reading expression", 0);
}


void comm_redo (void *state, int argc, const char **argv) {
    startstop (state, argc, argv, filter_redo, 0,
               "regular-expression", "filter restart command template", 0);
}


void comm_start (void *state, int argc, const char **argv) {
    startstop (state, argc, argv, filter_start, 0,
               "regular-expression", "filter replace enable expression", 0);
}


void comm_replace (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0], *s;
    FRUN_SPEC *desc ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }

    switch (argc) {
    case 1:
        goto usage ;
    case 2:
        desc = filter_find (o->freg, argv [1]);
        break ;
    default:
        desc = filter_spec (o->freg, argv [1]);
    }

    if (desc == 0) {
        ctrl_reply (o->ctrl, "-ERR: undefined handle \"%s\"\n", argv [1]);
        return ;
    }

    if (argc > 2) {
        int n ;
        IOBUF  pat ;
        XZERO (pat) ;
        data_append (&pat, argv [2], 0);
        for (n = 3; n < argc; n ++) 
            data_printf (&pat, " %s", argv [n]);
        s = filter_replace (desc, pat.data);
        kill_space (&pat);
        if (s)
            ctrl_reply (o->ctrl, "+OK");
        else
            ctrl_reply (o->ctrl, "-ERR: %s", filter_error (desc));
    }
    else if ((s = filter_replace (desc, 0)) == 0)
        ctrl_reply
            (o->ctrl, "-ERR: no replace expression for \"%s\"", argv [1]);
    else
        ctrl_reply (o->ctrl, "+OK\n%s", s);

    return ;
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s <key> [<regular-expression> ... ]"
                "-- set/show filter replacement expression\n", me);
}


void comm_run (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0], *cmd, *run_cmd, *s;
    unsigned timeout =  COMM_WAIT_RUN_DEFAULT ;
    unsigned timemax = (COMM_WAIT_RERUN_MAX + 500) / 1000;
    time_t gameover ;
    FRUN_SPEC *desc ;
    int n, inx;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc < 2) goto usage;

    inx = (argc > 2 &&
	   isdigit (argv [1][0]) &&
	   parse_unsigned (&timeout, argv [1], 0))
        ? 2
        : 1
        ;

    if ((desc = filter_find (o->freg, argv [inx])) == 0) {
        ctrl_reply (o->ctrl, "-ERR: undefined handle \"%s\"", argv [inx]);
        return ;
    }

    /* load argument stack */
    filter_push (desc, 0);
    for (n = inx + 1; n < argc; n ++)
        filter_push (desc, argv [n]);
    
    if ((cmd = filter_tmpl (desc, 0, 1)) == 0) {
        ctrl_reply (o->ctrl, "-ERR: no template for \"%s\"", argv [inx]);
        return ;
    }

    gameover = time (0) + timemax ;
    run_cmd  = cmd ;
    for (;;) {
        int stopped, redo ;
 
#       ifdef USE_DEBUG
        if (run_cmd && (verbosity & COMM_LOG_RERUN) != 0)
            logger (LOG_WARNING,
                    "comhandler: sending command \"%s\"", run_cmd);
#       endif

        // if (strcmp (argv [inx], "dsl-show") == 0)
        //    logger (LOG_DEBUG, "Got it!");

        if ((s = shell_data (o->shell, run_cmd, 0,
                             (int(*)(void*,IOBUF*)) filter_run, (void*)desc,
                             timeout)) == 0) {
            ctrl_reply (o->ctrl, "-ERR: \"%s\": %s",
                        argv [inx], shell_error (o->shell));
            return ;
        }
        
        if ((n = filter_matches (desc, 0, &stopped, &redo, 0, 0)) >= 0) {

            if (stopped) {
                ctrl_reply (o->ctrl, "-ERR: \"%s\": intercepted by the "
                            "stop pattern - giving up", argv [inx]);
                return ;
            }

            if (redo) {
                if (time (0) < gameover) {
                    /* allow some sleep */
                    if (msnap (o, COMM_WAIT_RERUN_DELAY) == 0)
                        return ;
                    if (verbosity & COMM_LOG_RERUN)
                        logger (LOG_WARNING, "comhandler: \"%s\": "
                                "intercepted by the stop pattern - "
                                "retrying", argv [inx]);
                    run_cmd = cmd ;
                    continue ;
                }
                ctrl_reply (o->ctrl, "-ERR: \"%s\": intercepted by the "
                            "stop pattern - timeout, giving up", argv [inx]);
                return ;
            }

            ctrl_reply (o->ctrl, "+OK: %d replacements\n%s", n, s);
            return ;
        }
           
        if (gameover < time (0)) {
            if (verbosity & COMM_LOG_RERUN)
                logger (LOG_WARNING, "comhandler: timeout - no response "
                        "from the ssh/shell running \"%s\"", argv [inx]);
            ctrl_reply
                (o->ctrl, "-ERR: timeout - no response from the ssh/shell");
            return ;
        }

        if (msnap (o, COMM_WAIT_RERUN_DELAY) == 0)
            return ;

        if (verbosity & COMM_LOG_RERUN)
            logger (LOG_WARNING, "comhandler: \"%s\": no proper response "
                    "from the ssh/shell - parser continues", argv [inx]);
        run_cmd = 0 ; /* don't repeat the command */
    }

    /* NOTREACHED */

usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s [<timeout>] <key> [<args> ...] "
                "-- run command by template\n", me);
}


void comm_del (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    FRUN_SPEC *desc ;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc < 2) goto usage;

    if ((desc = filter_find (o->freg, argv [1])) == 0)
        ctrl_reply (o->ctrl, "-ERR: undefined handle \"%s\"", argv [1]);
    else {
        /* clear argument stack */
        filter_delete (desc);
        ctrl_reply (o->ctrl, "+OK");
    }
    
    return ;
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s <key> -- delete command template\n", me);
}


void comm_list (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0], *file;
    FILE *fp ;

    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 3) goto usage;

    file = argc > 1 ? argv [1] : COMM_LIST_FILE_DEFAULT ;
    if ((fp = open_file (o, file)) != 0) {
        filter_print (o->freg, (void(*)(void*,const char*,...))fprintf, fp);
        fprintf (fp, ".\n");
        fclose (fp);
        ctrl_reply (o->ctrl, "+OK:\n%s", file);
    }

    return ;
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s [<file-name>] -- dump filter table\n", me);
}


void comm_iodump (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0], *file;
    FILE *fp ;

    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 3) goto usage;

    if ((fp = shell_iodump (o->shell, 0)) != 0) {
        fclose (fp) ;
        ctrl_reply (o->ctrl, "+OK: io dump closed");
        return ;
    }

    file = argc > 1 ? argv [1] : COMM_IODUMP_FILE_DEFAULT ;
    if ((fp = open_file (o, file)) != 0) {
        shell_iodump (o->shell, fp);
        ctrl_reply (o->ctrl, "+OK:\n%s", file);
    }

    return ;
    
usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply (o->ctrl, "%s [<file-name>] -- dump io to file\n", me);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */

