/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "util.h"
#include "settings.h"
#include "nvalloc.h"
#include "logger.h"

#define __FILTER_PRIVATE
#include "filter.h"

#ifdef WITH_DMALLOC
static unsigned long __mark;

#define DMALLOC_MARK() { __mark = dmalloc_mark (); }

#define DMALLOC_HUNT() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }

#define DMALLOC_VERIFY(c) {                             \
                 if (c) dmalloc_verify (c);             \
                 dmalloc_log_stats ();                  }

#define DMALLOC_MESSAGE(g,fmt, ...) {                   \
                 dmalloc_message (fmt, ##__VA_ARGS__) ; \
                 DMALLOC_VERIFY(g) ;                    }

#else
#define DMALLOC_MARK()            /* n/a */
#define DMALLOC_HUNT()            /* n/a */
#define DMALLOC_VERIFY(g)         /* n/a */
#define DMALLOC_MESSAGE(fmt, ...) /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * Private: definitions & declarations
 * ----------------------------------------------------------------------- */

#define DIM(x) (sizeof (x) / sizeof ((x) [0]))

struct _filter_print_desc {
    void (*prt) (void*, const char*, ...);
    void  *out ;
} ;

/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
int filter_print_cb (void   *state,
                     void    *data,
                     char     *key,
                     unsigned klen) {
    struct _filter_print_desc *desc = state ;
    FRUN_SPEC *q                    = data  ;

    (*desc->prt) (desc->out, "\n%s:\n", key);
    if (q->send.format) {
        int n = strlen (q->send.format);
        (*desc->prt) (desc->out,
                      "  send request:     %s", q->send.format);
        if (n > 0 && q->send.format [n - 1] != '\n')
            (*desc->prt) (desc->out, "\n");
    }
    if (q->stop.pattern)
        (*desc->prt) (desc->out,
                      "  stop reading:     %s\n", q->stop.pattern);
    if (q->redo.pattern)
        (*desc->prt) (desc->out,
                      "  restart command:  %s\n", q->redo.pattern);
    if (q->find.pattern)
        (*desc->prt) (desc->out,
                      "  enable replace:   %s\n", q->find.pattern);
    if (q->replace.match.pattern)
        (*desc->prt) (desc->out,
                      "  replace response: %s\n", q->replace.match.pattern);

    return 0;
}


static
void free_args (const char **argv) {
    const char **p ;

    if ((p = argv) != 0) {
        while (*p)
            free ((char*) * p ++) ;
        free (argv) ;
    }
}


static /* destructor call back for the hlst */
void fspec_free (void *state, void* item, char* key, unsigned keylen) {
    if (item) {
        FRUN_SPEC *q = item ;
        free_args (q->send.argv);
        if (q->send.format)
            free ((char*)q->send.format);
        if (q->replace.format)
            free ((char*)q->replace.format);
        if (q->error)
            free ((char*)q->error);
        if (q->find.pattern) {
            free ((char*)q->find.pattern);
            regfree (&q->find.regex);
        }
        if (q->stop.pattern) {
            free ((char*)q->stop.pattern);
            regfree (&q->stop.regex);
        }
        if (q->replace.match.pattern) {
            free ((char*)q->replace.match.pattern);
            regfree (&q->replace.match.regex);
        }
        kill_space (&q->handy);
        free (q) ;
    }
}

/* ----------------------------------------------------------------------- *
 * Private: template parser and helpers
 * ----------------------------------------------------------------------- */

static
const char *get_submatch_text (size_t     *Len,
                               const char *buf,
                               void      *argv,
                               int         inx) {
    int offs = ((regmatch_t*)argv) [inx].rm_so ;
    if (offs != -1) {
        *Len = ((regmatch_t*)argv) [inx].rm_eo - offs ;
        return buf + offs ;
    }
    return 0;
}


static
const char *get_arg_text (size_t     *Len,
                          const char *buf,
                          void      *argv,
                          int         inx) {
    const char *s = ((const char**)argv) [inx] ;
    if (s) {
        *Len = strlen (s) ;
        return s ;
    }
    return 0;
}


static
const char *strchr_unescaped (const char *s, int c) {

    const char *t ;

    while ((t = strchr (s, c)) != 0) {
        int n = 0 ;

        /* check for an odd numer of backslash escapes */
        while (s + n < t && t [- n - 1] == '\\')
            n ++;
        if ((n & 1) == 0)
            return t;

        /* n & 1: odd number => this char is escaped */
        s = t + 1;
    }

    return 0 ;
}


static
unsigned parse_1hex (int *c, const char *s) {

    if ('0' <= *s && *s <= '9') {
        *c = *s - '0' ;
        return 1;
    }
    if ('a' <= *s && *s <= 'f') {
        *c = *s - 'a' + 10;
        return 1;
    }
    if ('A' <= *s && *s <= 'F') {
        *c = *s - 'A' + 10;
        return 1;
    }

    return 0 ;
}


static
unsigned parse_2hex (int *c, const char *s) {
    int low, high ;

    if (parse_1hex (&high, s) && parse_1hex (&low, s+1)) {
        *c = (high << 4) | low ;
    }

    return 0;
}


static
unsigned parse_3octal (int *c, const char *s) {

    if ('0' <= *s && *s <= '7') {
        int accu = (* s ++ - '0') ;

        if ('0' <= *s && *s <= '7') {
            accu <<= 3 ;
            accu  += (* s ++ - '0') ;

            if ('0' <= *s && *s <= '7') {
                accu <<= 3 ;
                accu  += (* s ++ - '0') ;

                *c = accu ;
                return 1 ;
            }
        }
    }

    return 0 ;
}


static
void append_escaped (IOBUF          *out,
                     const char       *s,
                     size_t          len,
                     const char *inhibit) {
    int c ;
    const char *t, *z ;

    if (len == 0 && (len = strlen (s)) == 0) {
        data_append (out, s, 0);
        return ;
    }
    z = s + len ;
    
    /* check for shortcut characters */
    while ((t = strchr (s, '\\')) != 0 && t < z) {

        if (s < t)
            data_append (out, s, t - s);

        if (inhibit && strchr (inhibit, t [1]))
            goto dont_replace ;

        switch (t [1]) {
        case  'a': c =   '\a' ; break ;
        case  'b': c =   '\b' ; break ;
        case  'e': c = '\x1b' ; break ;
        case  'f': c =   '\f' ; break ;
        case  'n': c =   '\n' ; break ;
        case  'r': c =   '\r' ; break ;
        case  's': c =    ' ' ; break ;
        case  't': c =   '\t' ; break ;
        case  'v': c =   '\v' ; break ;
        case '\\': c =   '\\' ; break ;
        case  '0':
        case  '1':
        case  '2':
            if (t + 2 < z && parse_3octal (&c, t + 1)) {
                t += 2 ;
                break ;
            }
            goto dont_replace ;
        case 'x':
            if (t + 2 < z && parse_2hex (&c, t + 2)) {
                t += 2 ;
                break ;
            }
            /* goto dont_replace ; */
        default:
        dont_replace:
            char_append (out, '\\');
            c = t [1] ;
        }

        char_append (out, c);
        s = t + 2 ;
    }

    if (s < z)
        data_append (out, s, z - s);
}


static
const char *xpand (IOBUF                                      *out,
                   const char                                   *s,
                   const char*(*fn)(size_t*,const char*,void*,int),
                   const char                                 *buf,
                   int                                        argc,
                   void                                      *argv) {
    
    const char *t ;

    if (s == 0)
        return 0 ;

    while ((t = strchr_unescaped (s, '$')) != 0) {
        int inx, var_name, repl_len, default_ok, replace_ok ;
        size_t len ;
        const char *u, *repl ;

        /* append prefix */
        if (s < t)
            append_escaped (out, s, t - s, 0);

        if (t [1] == '\0') {
            char_append (out, '$') ;
            return out->data ;
        }

        /* check whether there is some default assignment */
        if (t [1] == '{'&& (u = strchr_unescaped (t + 1, '}')) != 0) {
            
            /* loop continues ... */
            s = u + 1 ;

            if (t + 4 < u && t [3] == ':') { /* if unset/empty or null */
                repl = t + 5 ;
                switch (t [4]) {
                case '-':
                    default_ok = 1 ;
                    replace_ok = 0 ;
                    break;
                case '+':
                    default_ok = 0 ;
                    replace_ok = 1 ;
                    break;
                default:
                    goto not_a_variable ;
                }
            }
            else if (t + 3 < u) { /* if unset/empty, only */
                repl       = t + 4 ;
                default_ok =     0 ;
                switch (t [3]) {
                case '-':
                    replace_ok = 0 ;
                    break;
                case '+':
                    replace_ok = 1 ;
                    break;
                default:
                    goto not_a_variable ;
                }
            }
            else
                goto not_a_variable ;

            repl_len   = u - repl ;
            var_name   = t [2] ;
        }
        else {
            /* loop continues ... */
            s          = t + 2 ;

            repl       = t ;
            repl_len   = 2 ;
            var_name   = t [1] ;
            default_ok = 0 ;
            replace_ok = 0 ;
        }
        
        /* get the index of the next argument */
        if ('0' <=  var_name &&  var_name <= '9')
            inx =  var_name - '0' ;
        else if ('a' <=  var_name &&  var_name <= 'z')
            inx =  var_name - 'a' + 10 ;
        else if ('A' <= t [1] &&  var_name <= 'Z')
            inx =  var_name - 'A' + 10 ;
        else {
        not_a_variable:
            /* not a proper variable definition */
            data_append (out, t, s - t) ;
            continue ;
        }

        /* append argument */
        if (inx < argc) {
            const char *a = (*fn) (&len, buf, argv, inx);
            if (a && len) { /* if unset/empty */
                if (replace_ok) {
                    if (repl_len)
                        data_append (out, repl, repl_len);
                }
                else
                    data_append (out, a, len);
                default_ok = 0 ;
            }
        }
        
        if (default_ok && repl_len)
            data_append (out, repl, repl_len) ;
    }

    /* append tail */
    append_escaped (out, s, 0, 0);

    return out->data ;
}


/* ----------------------------------------------------------------------- *
 * Private: regex parser and helpers
 * ----------------------------------------------------------------------- */

static
unsigned compile (FILTER_FIND    *rex,
                  const char  **Error,
                  const char *pattern, /* pattern to compile */
                  const char   *store, /* stored pattern */
                  int           flags) {
    int n;

    if (rex->pattern) {
        DMALLOC_VERIFY (rex->pattern);
        free ((char*)rex->pattern);
        rex->pattern = 0;
        regfree (&rex->regex);
    }

    /* compile and check for errors */
    if ((n = regcomp (&rex->regex, pattern, flags)) != 0) {
        size_t m = regerror (n, &rex->regex, 0, 0);
        *Error = (const char*) zrealloc ((void*)*Error, m);
        regerror (n, &rex->regex, (char*)*Error, m);
        return 0 ;
    }

    if (*Error) {
        free ((char*)*Error);
        *Error = 0;
    }
    rex->pattern = zstrdup (store);
    DMALLOC_VERIFY (rex->pattern);

    return 1;
}


static
const char *parse_delim (FRUN_SPEC  *q,
                         int    *Flags,
                         int *Globally,
                         int  *Matched,
                         const char *s,
                         int     delim) {
    int n, flags = REG_EXTENDED|REG_NEWLINE, globally = 0, matched = 0;
    const char *t ;

    if ((t = strchr_unescaped (s, delim)) == 0) {
#       define delim_error "terminating delimiter missing"
        q->error = (const char*)
            zrealloc ((char*)q->error, sizeof (delim_error));
        memcpy ((char*)q->error, delim_error, sizeof (delim_error));
#       undef delim_error
        return 0;
    }

    for (n = 1; t [n] != 0; n ++)
        switch (t [n]) {
        case 'i':
            flags |= REG_ICASE ;
            continue ;
        case 's':
            flags &= ~REG_NEWLINE ;
            continue ;
        case 'x':
            if (Matched) {
                matched = 1 ;
                continue ;
            }
            goto flag_unknown;
        case 'g':
            if (Globally) {
                globally = 1 ;
                continue ;
            }
            /* goto flag_unknown; */
        default:
        flag_unknown:
            if (Globally || Matched) {
#               define flags_error \
                     "unsupported regex flag \"%c\" (g, v, s or i, only)"
                s = flags_error ;
                n = sizeof (flags_error) ;
#               undef flags_error
            }
            else {
#               define flags_error \
                     "unsupported regex flag \"%c\" (s or i, only)"
                s = flags_error ;
                n = sizeof (flags_error) ;
#               undef flags_error
            }
            {
                IOBUF  buf ;
                XZERO (buf);
                data_printf (&buf, s, t [1]);
                q->error = buf.data ;
                unlink_space (&buf) ;
            }
            return 0;
        }

    if (Flags)
        *Flags = flags;
    if (Globally)
        *Globally = globally;
    if (Matched)
        *Matched = matched;
    return t;
}


static
const char *filter_pattern (FRUN_SPEC    *q,
                            FILTER_FIND  *t,
                            const char *pat) {
    if (pat) {
        int n, delim = * pat ++, flags ;
        const char *s ;
        IOBUF out;

        if ((s = parse_delim (q, &flags, 0, 0, pat, delim)) == 0)
            return 0;

        XZERO (out);
        
        /* expand the pattern */
        append_escaped (&out, pat, s - pat, 0);

        /* compile the pattern */
        n = compile (t, &q->error, out.data, pat - 1, flags);

        kill_space (&out);

        if (n == 0)
            return 0;
    }

    return t->pattern ;
}

/* ----------------------------------------------------------------------- *
 * Public: Cnstructor/destructor
 * ----------------------------------------------------------------------- */

FILTER *filter_open (unsigned subdim) {
    FILTER f ;

    XZERO (f);
    f.lookup       = create_hlst (0, fspec_free, 0);
    if ((f.submatch_dim = subdim) == 0)
        f.submatch_dim = FILTER_SUBMATCH_DIM ;
    return zmemdup (&f);
}


void filter_close (FILTER *f) {
    if (f) {
        destroy_hlst (f->lookup);
        free (f);
    }
}


FRUN_SPEC *filter_spec (FILTER *f, const char *var) {
    FRUN_SPEC **F ;
    if ((F = (FRUN_SPEC **) find_hlst (f->lookup, var, 0)) == 0) {
        F = (FRUN_SPEC **) make_hlst (f->lookup, var, 0) ;
        *F = (FRUN_SPEC*) zmalloc
            (sizeof (FRUN_SPEC) + (f->submatch_dim-1) * sizeof (regmatch_t));
        (*F)->base = f ;
        (*F)->key  = query_key_hlst ((void**)F);
    }
    return *F;
}


FRUN_SPEC *filter_find (FILTER *f, const char *var) {
    FRUN_SPEC **F = (FRUN_SPEC **) find_hlst (f->lookup, var, 0);
    return F ? *F : 0 ;
}


void filter_delete (FRUN_SPEC *q) {
    delete_hlst (q->base->lookup, q->key, 0);
}

/* ----------------------------------------------------------------------- *
 * Public: string and pattern handling
 * ----------------------------------------------------------------------- */

void filter_push (FRUN_SPEC *q, const char *arg) {
    if (arg) {
        q->send.argv = (const char**) /* q->send_argv == NULL is ok */
            zrealloc (q->send.argv, (q->send.argc + 2) * sizeof (arg));
        q->send.argv [q->send.argc ++ ] = zstrdup (arg) ;
        q->send.argv [q->send.argc]     =            0  ;
    }
    else {
        /* reset arguments */
        if (q->send.argv) {
            free_args (q->send.argv) ;
            q->send.argv = 0 ;
            q->send.argc = 0 ;
        }
        /* reset registers */
        filter_matches (q, 0, 0, 0, 0, 0);
    }
}


const char *filter_tmpl (FRUN_SPEC *q, const char *fmt, int expand) {

    if (fmt) {
        if (q->send.format)
            free ((char*)q->send.format);
        q->send.format = zstrdup (fmt) ;
    }

    if (expand) {
        kill_space (&q->handy);
        return xpand (&q->handy, q->send.format,
                      get_arg_text, 0, q->send.argc, q->send.argv) ;
    }
    
    return q->send.format ;
}


const char *filter_start (FRUN_SPEC *q, const char *pattern) {
    return filter_pattern (q, &q->find, pattern);
}


const char *filter_stop (FRUN_SPEC *q, const char *pattern) {
    return filter_pattern (q, &q->stop, pattern);
}


const char *filter_redo (FRUN_SPEC *q, const char *pattern) {
    return filter_pattern (q, &q->redo, pattern);
}


const char *filter_replace (FRUN_SPEC *q, const char *replace) {

    if (replace) {
        int n, m, delim = * replace ++, flags ;
        const char *s, *t ;
        IOBUF out;

        if ((s = strchr_unescaped (replace, delim)) == 0) {
#           define delim_error "in-beween delimiter missing"
            q->error = (const char*)
                zrealloc ((char*)q->error, sizeof (delim_error));
            memcpy ((char*)q->error, delim_error, sizeof (delim_error));
#           undef delim_error
            return 0;
        }

        if ((t = parse_delim (q, &flags, &n, &m, s + 1, delim)) == 0)
            return 0;
        q->replace.globally     = n ; /* assign bit field */
        q->replace.matched_only = m ; /* assign bit field */

        XZERO (out);

        /* expand the replace pattern */
        append_escaped (&out, replace, s - replace, 0);

        /* compile the pattern */
        n = compile
            (&q->replace.match, &q->error, out.data, replace-1, flags);
        
        kill_space (&out);

        if (n == 0)
            return 0;
        
        /* set up the replacement string, is at least "" */
        n = t - s ;
        q->replace.format = (const char*)
            memcpy (zrealloc ((char*)q->replace.format, n), s + 1, n);
        ((char*)q->replace.format) [n - 1] = '\0' ;
    }

    return q->replace.match.pattern ;
}


const char *filter_error (FRUN_SPEC *q) {
    return q->error ;
}


int filter_matches (FRUN_SPEC *q,
                    int  *Visits,
                    int   *Stops,
                    int   *Redo,
                    int   *Found,
                    int   *Edits) {

    int n = (q->stop.match_hits ||
             q->find.match_hits ||
             q->redo.match_hits ||
             (q->visited && (q->stop.pattern == 0 &&
                             q->find.pattern == 0 &&
                             q->redo.pattern == 0)))
        ? q->replace.match.match_hits
        : -1
        ;

    if ((verbosity & FILTER_LOG_COUNTERS) != 0 &&
        (Visits || Stops || Found || Edits))
        logger (LOG_DEBUG, "filter_matches: filter visits: %d, stop: %d%s, "
                "redo: %d%s, replace enabled: %d%s, replacements: %d%s",
                q->visited,
                q->stop.match_hits,
                q->stop.pattern          ? "" : "(disabled)",
                q->redo.match_hits,
                q->redo.pattern          ? "" : "(disabled)",
                q->find.match_hits,
                q->find.pattern          ? "" : "(disabled)",
                q->replace.match.match_hits,
                q->replace.match.pattern ? "" : "(disabled)");
    
    if (Visits) *Visits = q->visited ;
    if (Stops)  *Stops  = q->stop.match_hits ;
    if (Redo)   *Redo   = q->redo.match_hits ;
    if (Found)  *Found  = q->find.match_hits ;
    if (Edits)  *Edits  = q->replace.match.match_hits ;
    
    q->visited                  = 0 ;
    q->stop.match_hits          = 0 ;
    q->redo.match_hits          = 0 ;
    q->find.match_hits          = 0 ;
    q->replace.match.match_hits = 0 ;

    return n ;
}


/* returns 0 on match */
int filter_run (FRUN_SPEC *q, IOBUF *text) {

    q->visited ++ ;
 
    if (text->data == 0)
        /* empty strings always matches */
        return 0 ;

    DMALLOC_VERIFY (q);

    if (q->stop.pattern && regexec (&q->stop.regex,
                                    text->data,
                                    q->base->submatch_dim,
                                    q->submatch,
                                    REG_NOTBOL|REG_NOTEOL) == 0) {
        q->stop.match_hits ++ ;
        return 0;
    }

    if (q->redo.pattern && regexec (&q->redo.regex,
                                    text->data,
                                    q->base->submatch_dim,
                                    q->submatch,
                                    REG_NOTBOL|REG_NOTEOL) == 0) {
        q->redo.match_hits ++ ;
        return 0;
    }

    if (q->find.pattern == 0) /* find == 0 && stop == 0 => no match */
        return q->stop.pattern == 0 ;
    
    if (regexec (&q->find.regex,
                 text->data,
                 q->base->submatch_dim,
                 q->submatch,
                 REG_NOTBOL|REG_NOTEOL) != 0)
        return 1;

    q->find.match_hits ++ ;

    /* pattern substitution */
    if (q->replace.format != 0) {
        const char *in ;
        int start_match ;

        IOBUF  out ;
        XZERO (out);

        /* make sure the character string terminates */
        char_append (text, 0);
        in = text->data ;
        
        while (in [0] != '\0' &&
               regexec (&q->replace.match.regex, in,
                        q->base->submatch_dim, q->submatch,
                        REG_NOTBOL|REG_NOTEOL) == 0 &&
               (start_match = q->submatch [0].rm_so) != -1) {

            int end_match = q->submatch [0].rm_eo ;

            q->replace.match.match_hits ++ ;

            /* copy prefix */
            if (start_match > 0 && q->replace.matched_only == 0)
                data_append (&out, in, start_match);

            /* append replacement */
            xpand (&out, q->replace.format, get_submatch_text,
                   in, q->base->submatch_dim, q->submatch);

            if (end_match == 0) break ;

            in += end_match ;

            if (q->replace.globally == 0) break ;
        }

        if (in [0] != '\0' && q->replace.matched_only == 0)
            data_append (&out, in, 0);
        
        if (verbosity & FILTER_LOG_REPLACE)
            logger (LOG_DEBUG, "filter_run: replacing \"%s\" by \"%s\"",
                    text->data, out.data);

        /* replace buffers */
        kill_space (text);
        *text = out ;

        DMALLOC_VERIFY (text->data);
    }

    return 0;
}


void filter_print (FILTER *f, void(*prt)(void*,const char*,...), void *out) {

    struct _filter_print_desc desc ;

    desc.prt = prt ;
    desc.out = out ;
    
    for_hlst_do (f->lookup, filter_print_cb, &desc);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
