/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/


/* simple stupid example of a command line client */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <libgen.h>
#include <unistd.h>
#include <signal.h>

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#define _GNU_SOURCE
#include <getopt.h>

#include "util.h"
#include "nvalloc.h"
#include "settings.h"
#include "logger.h"
#include "varstr.h"

/* ----------------------------------------------------------------------- *
 * variables and definitions
 * ----------------------------------------------------------------------- */
static char *client = 0;

#define __str(x)              # x
#define str(x)                __str (x)

#define REPLY_SOCKET_TMPL "/tmp/replyto-%s-%x"

static int no_timeout_message = 0 ;

/* ----------------------------------------------------------------------- *
 * MAIN
 * ----------------------------------------------------------------------- */

#ifdef DISCARD_CMD_SOCKET

int main () {
    fprintf (stderr,
             "Communication support was disabled by compiler option!\n");
    exit (2);
}

#else

static const
struct option long_options [] = {
    {"file",    1, 0, 'f'},
    {"help",    0, 0, 'h'},
    {"root",    1, 0, 'r'},
    {"wait",    0, 0, 'w'},
    {"strip",   0, 0, 's'},
    {"command", 0, 0, 'c'},
    {0, 0, 0, 0}
};
static const
char short_options [] = "+cf:r:hsw";
   
static char
*append (const char *p) {
    static char    *str ;
    static unsigned str_len ;

    if (str_len == 0)
        return str = strcpy
            ((char*) zrealloc (str, (str_len = strlen (p) + 1)), p);

    str = strcat ((char*) zrealloc (str, (str_len += strlen (p) + 1)), " ");
    return strcat (str, p);
}

static
void tmo_exit (int unused) {
    unlink (client);
    if (no_timeout_message == 0)
        logger (LOG_ERR, "%s: TIMEOUT", progname);
    exit (2);
}

static
void signal_exit (int unused) { 
    unlink (client);
    logger (LOG_ERR, "%s: INTERRUPT", progname);
    exit (2);
}

static
void usage (void) {
    fprintf
        (stderr,
         "Usage: %s [options] <cmd> [<args> ...]\n"
         "       %s -\n\n"
         "Options: -                   read fron stdin\n"
         "         -s, --strip         strip leading +OK/-ERR\n"
         "         -f, --file=<FILE>   read from file\n"
         "         -r, --root=<PREFIX> server runs in a chroot environment\n"
         "         -w, --wait          wait a long time for a server reply - "
         ""                           "use\n"
         "                             \"%s -w contest [timeout]\" for start "
         ""                            "up synchronisation.\n"
         "         -c, --command       send command to the rcmdd "
         ""                            "(compatibility only)\n"
         "         -h, --help          print this message\n",
         progname, progname, progname);
    exit (2);
}


static
char *load_stdin (size_t *Len) {

    char *s ;
    int fd, n, left = RCMD_MAX_VALUE_SIZE + 1;
        
    if ((fd = fileno (stdin)) < 0) {
        logger (LOG_ERR, "cannot read from stdin: %s", strerror (errno));
        exit (2);
    }

    s = (char*)xmalloc (left);
    
    strcpy (s, MAIN_CLI_CMD " ");
    n = sizeof (MAIN_CLI_CMD) ; /* == strlen (MAIN_CLI_CMD) + 1 */
    
    do {
        int m = read (fd, s + n, left) ;
        if (m <= 0) {
            if (m == 0)
                break ;
            logger (LOG_ERR, "error reading from stdin: %s",
                    strerror (errno));
            exit (2);
        }
        n    += m ;
        left -= m ;
    }
    while (left > 0);

    if (n > RCMD_MAX_VALUE_SIZE) {
        logger (LOG_ERR, "Too much data from stdin (max %d bytes)",
                RCMD_MAX_VALUE_SIZE);
        exit (2);
    }
        
    s [n] = 0;

    if (Len != 0)
        *Len = n ;

    return s;
}


int main (int argc, char *argv[]) {
    char *data = 0;
    char *file = 0;
    int fd, un_size, n, c, option_inx, strip = 0, xcode, use_command = 0 ;
    unsigned u, longtmo = 0 ;
    struct sockaddr_un un ;
    char *p, *q, buf [RCMD_CTRL_IOBUF_SIZE], *root = "";
    
    /* progname ::= only the last path segment */
    if ((progname = strrchr (argv [0], '/')) == 0)
        progname = argv [0] ;
    else
        ++ progname ;

    if (argc == 1)
        usage ();

    if (argc == 2 && argv [1] [0] == '-' && argv [1] [1] == '\0') {
        data = load_stdin (0);
    }
    else {
        while ((c = getopt_long (argc, argv, short_options, 
                                 long_options, &option_inx)) != -1)
            switch (c) {
            case 'c':
                use_command = 1;
                continue ;
            case 's':
                no_timeout_message = 1 ;
                strip ++ ;
                continue ;
            case 'w':
                longtmo = RCMD_LONGRECV_TIMEOUT ;
                continue ;
            case 'f':
                file = optarg ;
                continue ;
            case 'r':
                root = optarg ;
                continue ;
            case 'h':
            case '?':
                usage ();
            }

        if (file && optind < argc) {
            fprintf (stderr,  "%s: Cannot have the --file option and "
                     "line commands at the same time.\n", progname);
            usage () ;
        }
    }

   
    /* create reply socket */
    if ((fd = socket (PF_UNIX, SOCK_DGRAM, 0)) < 0) {
        logger (LOG_ERR, "Cannot create command socket: %s",
                strerror (errno));
        exit (2);
    }

    /* assemble the reply command socket path */
    client = (char*) xmalloc
        (sizeof (REPLY_SOCKET_TMPL) + strlen (progname) + 30);
    sprintf (client, REPLY_SOCKET_TMPL, progname, getpid ());
    if (strlen (client) > UNIX_PATH_MAX-1) client [UNIX_PATH_MAX-1] = '\0' ;
  
    /* make an argument string: /reply/socket <cmd> <args> ... */
    p = append (client) ;
    if (data)
        p = append (data);
    else if (file) {
        p = append ("by");
        p = append (file);
    }
    else {
        const char *rcmd [] = {MAIN_CLI_CMD, MAIN_FILE_CMD, MAIN_RECONN,
                               MAIN_KPALVE, MAIN_FLUSH, MAIN_REDO,
                               MAIN_TEMPL, MAIN_ATEMPL, MAIN_DELETE,
                               MAIN_RUN, MAIN_STOP, MAIN_FIND, MAIN_REPLACE,
                               MAIN_FLIST, MAIN_IODUMP,
                               "help", "rusage", "noisy", "info", 0} ;
        const char **q = rcmd ;
        if (strcmp (argv [optind], MAIN_PING) == 0) {
            if (longtmo) {
                if (optind >= argc - 1) {
                    p = append (MAIN_PING) ;
                    p = append (str (RCMD_LONGRECV_TIMEOUT)) ;
                    optind ++ ;
                }
                else if (parse_unsigned (&u, argv [optind + 1], 0) != 0)
                    longtmo = u ;
            }
        }
        else if (use_command == 0) {
            while (argc > optind && strcmp (argv [optind], *q) != 0 && * ++ q)
                ;
            if (* q == 0)
                p = append ("do");
        }
        for (n = optind; n < argc; n ++)
            if (argv [n] [0])
                p = append (argv [n]);
            else
                p = append ("\"\"");
    }
    
    /* prepend the /reply/socket path by the root prefix */
    q = (char*) xmalloc (strlen (root) + strlen (client) + 1);
    client = strcat (strcpy (q, root), client);

    /* make directory unless present */
    q = zstrdup (client); /* dirname changes the arg, so duplicate  */
    mkdir (dirname (q), 0700);
    free (q);

    /* prepare for clean up */
    signal (SIGINT,  signal_exit);
    signal (SIGHUP,  signal_exit);
    signal (SIGTERM, signal_exit);
    signal (SIGUSR1, signal_exit);
    signal (SIGUSR2, signal_exit);

    /* bind to the reply socket */
    unlink (client);
    un.sun_family = AF_UNIX ;
    strcpy (un.sun_path, client);
    un_size = sizeof (un.sun_family) + strlen (un.sun_path);
    if (bind (fd, (struct sockaddr*)&un, un_size) < 0) {
        logger (LOG_ERR, "Cannot bind command socket to %s: %s", 
                un.sun_path, strerror (errno));
        exit (2);
    }

    /* send the command buf to server */
    strncpy (un.sun_path, RCMD_CTRL_REQ_PATH, UNIX_PATH_MAX);
    un.sun_path [UNIX_PATH_MAX-1] = '\0' ;
    un_size =
        offsetof (struct sockaddr_un, sun_path) +
        strlen (un.sun_path) + 1;
    if (sendto (fd, p, strlen (p), 0, (struct sockaddr*)&un, un_size) < 0) {
        logger (LOG_ERR, "Cannot send command via %s: %s",
                un.sun_path, strerror (errno));
        unlink (client);
        exit (2);
    }

    /* receive answer */
    signal (SIGALRM, tmo_exit);
    if (longtmo == 0) longtmo = RCMD_RECV_TIMEOUT ;
    alarm ((longtmo + 500) / 1000);
    n = recvfrom (fd, buf, sizeof (buf) -1, 0, 0, 0);
    alarm (0);
    unlink (client);
    if (n < 0) {
        logger (LOG_ERR, "Error reading from %s: ",
                un.sun_path, strerror (errno));
        exit (2) ;
    }
    buf [n] = '\0'; /* zero terminated - makes parsing easier */

    p = buf ;
    xcode = (buf [0] == '+') ;

    if (strip) {
        if (strncmp (p, "+OK", 3) == 0)
            xcode = 0 ;
        else if (strncmp (p, "-ERR", 4) == 0)
            xcode = 1 ;
        else
            strip = 0 ;
        if (strip) {
            /* strip first line */
            while (*p && *p != '\n') p ++ ;
            if (*p == '\n') p ++;
        }
    }

    if ((n = strlen (p)) > 0 && p [n-1] == '\n')
        p [-- n] = '\0' ;

    if (strip == 0 || * p)
        printf ("%s\n", p);

    exit (xcode);
}

#endif /* DISCARD_CMD_SOCKET */

/* ------------------------------------------------------------------------ *
 * End
 * ------------------------------------------------------------------------ */
