/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <grp.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "varstr.h"
#include "comhandler.h"
#include "filter.h"
#include "settings.h"
#include "do.h"
#include "util.h"
#include "nvalloc.h"
#include "filter-defaults.h"
#include "logger.h"

/* ----------------------------------------------------------------------- *
 * variables and definitions
 * ----------------------------------------------------------------------- */

#ifdef _DEBUG
#define _MAIN_DEBUG
#endif

typedef
struct _desc {
    DO       *loop ;
    COMM     *comm ;
    SHELL     *cmd ;
    FILTER *filter ;
    int no_idle_check ;
} SESSION ;


extern char *optarg;
extern int optind, opterr, optopt;

static jmp_buf go_here;           /* jump buffer for long jumps */

/* ----------------------------------------------------------------------- *
 * Fiddling around with options ...
 * ----------------------------------------------------------------------- */

#define __arg1of4(a,b,c,d)    a
#define __arg2of4(a,b,c,d)    b
#define __arg3of4(a,b,c,d)    c
#define __arg4of4(a,b,c,d)    d
#define __str(x)              # x
#define str(x)                __str (x)

/* I am fed up with option name and option key management using strings
   and letters at several places - so keep them together -- jordan */

#define opt_name(x)         x (__arg1of4)
#define opt_key(x)          x (__arg2of4)
#define opt_argtype(x)      x (__arg3of4)
#define opt_info(x)         x (__arg4of4)

#define OPT_VERSION(x)      x ("version",      'V', 0,           "print version and exit")
#define OPT_VERBOSE(x)      x ("verbose",      'v', 0,           "print extra diagnostic messages, repeat for more")
#define OPT_HELP(x)         x ("help",         'h', 0,           "instead of a manual page")
#define OPT_SYSLOG(x)       x ("syslog",       's', "FACILITY",  "use syslog with the given facility")
#define OPT_NOTIMEPFX(x)    x ("no-timestamp", 'n', 0,           "do not prefix log data by a time stamp")
#define OPT_DETACH(x)       x ("detach",       'd', 0,           "daemon mode, put yourself in the background")
#define OPT_NOIACHECK(x)    x ("no-interactive-check", 'N', 0,   "ignore password requests etc. on (re)opening the ssh/shell")
#define OPT_NOIDLECHECK(x)  x ("no-idle-test", 'A', 0,           "do not keep-alive-check the connection in the idle loop")
#define OPT_FORCESTARTUP(x) x ("force-startup",'F', 0,           "always start up no matter if we can connect, or not")
#define OPT_PID_FILE(x)     x ("pid-file",     'p', "FILE",      "save the process pid")
#define OPT_CHROOT(x)       x ("chroot",       'r', "DIR",       "chroot to DIR after initialisation")
#define OPT_GROUPID(x)      x ("groupid",      'g', "GID",       "allow this group to connect and send admin requests")
#define OPT_STRIP_PRMT(x)   x ("strip",        'Z', "PROMPT",    "strip the prompt from the response (default: \"" SHELL_STRIP_PROMPT "\")")
#define OPT_POOL(x)         x ("pool",         'D', "DIR",       "files are executed from this directory (default: " SHELL_DEFAULT_POOL ")")
#define OPT_PORT(x)         x ("port",         'P', "PORT",      "pass this as port/connect option to the ssh/shell")
#define OPT_USER(x)         x ("user",         'L', "USER",      "pass this as user/login option to the ssh/shell")
#define OPT_IDFILE(x)       x ("idfile",       'I', "IDFILE",    "pass this as idfile/key option to the ssh/shell")
#define OPT_RCLOCAL(x)      x ("rclocal",      'l', "FILE",      "run local script on connection initialisation, first")
#define OPT_RCSCRIPT(x)     x ("rcscript",     'f', "FILE",      "pass the contents to the ssh/shell on connection initialisation")
#define OPT_NULL(x)         x (0,               0 , 0,            0)


#define LONG_OPTION_RECORD(x,y)  {opt_name (x), y, 0, opt_key (x)}
static const
struct option long_options [] = {
    LONG_OPTION_RECORD (OPT_VERSION,      0),
    LONG_OPTION_RECORD (OPT_VERBOSE,      0),
    LONG_OPTION_RECORD (OPT_NOTIMEPFX,    0),
    LONG_OPTION_RECORD (OPT_NOIACHECK,    0),
    LONG_OPTION_RECORD (OPT_NOIDLECHECK,  0),
    LONG_OPTION_RECORD (OPT_FORCESTARTUP, 0),
#   ifndef DISCARD_INFO
    LONG_OPTION_RECORD (OPT_HELP,         0),
#   endif /* DISCARD_INFO */
    LONG_OPTION_RECORD (OPT_RCLOCAL,      1),
    LONG_OPTION_RECORD (OPT_RCSCRIPT,     1),
    LONG_OPTION_RECORD (OPT_PORT,         1),
    LONG_OPTION_RECORD (OPT_USER,         1),
    LONG_OPTION_RECORD (OPT_IDFILE,       1),              
    LONG_OPTION_RECORD (OPT_GROUPID,      1),
    LONG_OPTION_RECORD (OPT_STRIP_PRMT,   1),
    LONG_OPTION_RECORD (OPT_POOL,         1),
    LONG_OPTION_RECORD (OPT_DETACH,       0),
    LONG_OPTION_RECORD (OPT_CHROOT,       1),
    LONG_OPTION_RECORD (OPT_PID_FILE,     1),
    LONG_OPTION_RECORD (OPT_SYSLOG,       1),
    LONG_OPTION_RECORD (OPT_NULL,         0)
};
#undef LONG_OPTION_RECORD


static const
char short_options [] = {
    '+',
    opt_key (OPT_VERSION),
    opt_key (OPT_VERBOSE),
    opt_key (OPT_NOTIMEPFX),
    opt_key (OPT_NOIACHECK),
    opt_key (OPT_NOIDLECHECK),
    opt_key (OPT_FORCESTARTUP),
#   ifndef DISCARD_INFO
    opt_key (OPT_HELP),
#   endif /* DISCARD_INFO */
    opt_key (OPT_RCLOCAL),      ':',
    opt_key (OPT_RCSCRIPT),     ':',
    opt_key (OPT_PORT),         ':',
    opt_key (OPT_USER),         ':',
    opt_key (OPT_IDFILE),       ':',      
    opt_key (OPT_GROUPID),      ':',
    opt_key (OPT_STRIP_PRMT),   ':',
    opt_key (OPT_POOL),         ':',
    opt_key (OPT_DETACH),
    opt_key (OPT_CHROOT),       ':',
    opt_key (OPT_PID_FILE),     ':',
    opt_key (OPT_SYSLOG),       ':',
    opt_key (OPT_NULL)
} ;

/* ----------------------------------------------------------------------- *
 * Timeout signal handler
 * ----------------------------------------------------------------------- */
static void s_go_here (int n) { longjmp (go_here, 1); }

/* ----------------------------------------------------------------------- *
 * call back functions
 * ----------------------------------------------------------------------- */

static
void idle (void *dsc) {
    SESSION *s = dsc ;

    if (verbosity & MAIN_LOG_IDLE_LOOPING)
        logger (LOG_INFO, "rcmd: Idle/cleanup/exec handler.");

    if (s->no_idle_check == 0 &&
        shell_contest (s->cmd, COMM_WAIT_REPLY_DEFAULT, 0) == 1)
        shell_flush (s->cmd);
}

/* ----------------------------------------------------------------------- *
 * helpers
 * ----------------------------------------------------------------------- */

static
void cpry (void) {
    logger (LOG_INFO, "RCMD - remote shell command wrapper");
    logger (LOG_INFO, "Copyright (c) 2006 - D-Link (Europe) Ltd., London, UK");
    logger (LOG_INFO, "Version " VERSION " -- Jordan Hrycaj <jordan@maxina.de>");
}

#ifndef DISCARD_INFO
static
void usage_line (const char indent,
                 const  char *name,
                 char          key,
                 const char   *arg,
                 const char  *info) {
    char longopt [100], format [30] ;
    strcpy (longopt, name) ;
    if (arg) { strcat (longopt, "="); strcat (longopt, arg); }
    sprintf (format, "%%%us-%%c, --%%-15s %%s\n", indent);
    fprintf (stderr, format, "", key, longopt, info);
}
#endif /* DISCARD_INFO */


static
void usage (const char *info, const char *opt) {

#   ifndef DISCARD_INFO
    if (info != 0) {
        fprintf (stderr, "\n%s: ERROR: %s", progname, info);
        if (opt) fprintf (stderr, " (option arg was %s)", opt);
        fprintf (stderr, "\n");
    }

#   define SHOW_OPTION(indent,opt) usage_line (indent,            \
                                               opt_name    (opt), \
                                               opt_key     (opt), \
                                               opt_argtype (opt), \
                                               opt_info    (opt)  )

    fprintf (stderr, "\nUsage: %s [options] <command-host>\n", progname);

    fputs ("\nGeneral options:", stderr);
    SHOW_OPTION ( 3, OPT_HELP);
    SHOW_OPTION (19, OPT_POOL);
    SHOW_OPTION (19, OPT_PORT);
    SHOW_OPTION (19, OPT_USER);
    SHOW_OPTION (19, OPT_IDFILE);

    fputs ("\nData flow options:", stderr);
    SHOW_OPTION ( 1, OPT_NOIACHECK);
    SHOW_OPTION (19, OPT_NOIDLECHECK);
    SHOW_OPTION (19, OPT_FORCESTARTUP);
    SHOW_OPTION (19, OPT_STRIP_PRMT);
    SHOW_OPTION (19, OPT_RCLOCAL);
    SHOW_OPTION (19, OPT_RCSCRIPT);

    fputs ("\nLogging options:", stderr);
    SHOW_OPTION ( 3, OPT_VERBOSE);
    SHOW_OPTION (19, OPT_SYSLOG);
    SHOW_OPTION (19, OPT_NOTIMEPFX);

    fputs ("\nProcess control:", stderr);
    SHOW_OPTION ( 3, OPT_DETACH);
    SHOW_OPTION (19, OPT_CHROOT);
    SHOW_OPTION (19, OPT_PID_FILE);
    SHOW_OPTION (19, OPT_GROUPID);

#   undef SHOW_OPTION
#   endif /* DISCARD_INFO */

    exit (1);
}


#ifndef DISCARD_INFO
static
void help (void) {
    fputs ("\n", stderr);
    syslog_init (0, 0); /* no timestamps with the introduction */
    cpry ();
    fputs ("\n"
           "\n ...\n"
           "",  stderr);
}
#endif /* DISCARD_INFO */



static
ip4addr_t hostip (const char *s) {
    
    extern int h_errno;
    int       e ;
    ip4addr_t ip ;

    struct hostent* h = gethostbyname (s) ;

    if (h != 0) {
        struct in_addr sin_addr;
        sin_addr.s_addr = *(in_addr_t*)h->h_addr;
        return ntohl (sin_addr.s_addr);
    }

    e = h_errno ;
    if ((ip = ntohl (inet_addr (s))) != 0 && ip != -1)
        return ip;
    errno = e ;
    return 0;
}

/* ----------------------------------------------------------------------- *
 * MAIN
 * ----------------------------------------------------------------------- */

#ifndef _MAIN_DEBUG
#undef main /* in case some debugging redirection applied */
#endif

int main (int argc, char *argv[]) {

    int  ok, n, c, option_inx ;
    char     *new_root =  0 ;
    char    *facility  =  0 ;
    char   *cmd_server =  0 ;
    char       *prompt =  0 ;
    char         *pool =  0 ;
    char     *rcscript =  0 ;
    char      *rclocal =  0 ;
    char         *user =  0 ;
    char       *idfile =  0 ;
    unsigned      port =  0 ;
    int        timepfx =  1 ;
    int         pid_fd = -1 ;
    int         detach =  0 ;
    int   no_iam_check =  0 ;
    int        verbose =  0 ;
    unsigned   set_gid =  0 ;
    int   idle_timeout =  MAIN_IDLE_TIMEOUT ;
    int  force_startup =  0 ;
    unsigned re_subdim =  0 ;
    CTRL         *ctrl =  0 ;
    
    SESSION s ;
    XZERO  (s);

#   ifdef WITH_DMALLOC
#   define  DMALLOC_LOG "/var/tmp/dmalloc.log"
    unlink (DMALLOC_LOG "~");
    rename (DMALLOC_LOG, DMALLOC_LOG "~");
    dmalloc_debug_setup ("debug=0x4f47d03,log=" DMALLOC_LOG);
    dmalloc_message ("*** %s(%d): dmalloc starts.\n", __FILE__, __LINE__);
#   undef   DMALLOC_LOG
#   endif

    /* progname ::= only the last path segment */
    if ((progname = strrchr (argv [0], '/')) == 0)
        progname = argv [0] ;
    else
        ++ progname ;
    
    while ((c = getopt_long (argc, argv, short_options, 
                             long_options, &option_inx)) != -1)
        switch (c) {
            struct group *grp;
        case opt_key (OPT_FORCESTARTUP):
            force_startup = 1 ;
            continue ;
        case opt_key (OPT_NOIDLECHECK):
            s.no_idle_check = 1 ;
            continue ;
        case opt_key (OPT_NOIACHECK):
            no_iam_check = 1 ;
            continue ;
        case opt_key (OPT_RCLOCAL):
            if (optarg [0] != '/')
                usage ("the local rc-script must be an absolute path", optarg);
            rclocal = optarg ;
            continue ;
        case opt_key (OPT_RCSCRIPT):
            if (optarg [0] != '/')
                usage ("the rc-script must be an absolute path", optarg);
            rcscript = optarg ;
            continue ;
        case opt_key (OPT_GROUPID):
            if ((parse_unsigned (& set_gid, optarg, 0) == 0 &&
                 (grp = getgrgid (set_gid))  == 0) ||
                (grp = getgrnam (optarg)) == 0)
                usage ("group name does not exist", optarg);
            set_gid = grp->gr_gid;
            continue ;
        case opt_key (OPT_STRIP_PRMT):
            if (optarg [0] == '\0')
                usage ("prompt cannot be empty", optarg);
            prompt = optarg ;
            continue ;
        case opt_key (OPT_POOL):
            if (optarg [0] != '/')
                usage
                    ("the pool directory must be an absolute path", optarg);
            pool = optarg ;
            continue ;
        case opt_key (OPT_PORT):
            if (parse_unsigned (& port, optarg, 0) == 0 || (port & ~ 0xffff))
                usage ("bogous ssh port", optarg);
            continue ;
        case opt_key (OPT_USER):
            user = optarg ;
            continue ;
        case opt_key (OPT_IDFILE):
            idfile = optarg ;
            continue ;
        case opt_key (OPT_CHROOT):
            new_root = optarg;
            continue ;
        case opt_key (OPT_NOTIMEPFX):
            timepfx = 0 ;
            continue ;
        case opt_key (OPT_DETACH):
            detach = 1 ;
            continue ;
        case opt_key (OPT_PID_FILE):
            unlink (optarg); /* don't be tricked by a symlink */
            if ((pid_fd = open (optarg, O_CREAT|O_EXCL|O_WRONLY, 0644)) < 0)
                usage ("cannot open or create pid file", optarg);
            continue ;
        case opt_key (OPT_SYSLOG):
            facility = optarg ;
            continue ;
        case opt_key (OPT_VERSION):
            cpry ();
            exit (0);
        case opt_key (OPT_VERBOSE):
            verbose ++ ;
            continue ;
#       ifndef DISCARD_INFO
        case '?':
            usage (0, "?") ; /* no return */
        case opt_key (OPT_HELP):
            help (); /* and usage */
#       endif /* DISCARD_INFO */
        default:
            usage (0,0) ; /* no return */
        }

    if (optind >= argc)
        usage ("missing <command-host>",0) ; /* no return */
    else {
        ip4addr_t ip = hostip (argv [optind++]);
        if (ip == 0)
            usage ("cannot resolve <command-host>",0) ;
        cmd_server = zstrdup (ip4addr_pp (ip));
        if (cmd_server [0] == '0')
            usage ("<command-host> address makes no sense",0) ;
    }
    
    if (optind < argc) {
        fprintf
            (stderr, "%s: Unexpected non-optional arguments: ", progname);
        while (optind < argc)
            fprintf (stderr, "%s ", argv [optind++]);
        fprintf (stderr, "\n");
        exit (2);
    }

    /* we must use syslog when running in the background */
    if (detach && facility == 0) facility = "daemon" ;

    /* initialise logging subsystem */
    if (facility) {
        int devnull ;
        /* ready for closing stdin/stdout/stderr */
        if ((devnull = open ("/dev/null", O_RDWR)) < 0) {
            usage ("Cannot open /dev/null: %s", strerror (errno));
            exit (2);
        }
        if (syslog_init (facility, timepfx) == 0)
            usage ("cannot open the syslog facility", facility);
        /* detach, run in the backgrond */
        if (detach) {
            if (fork ()) exit (0);
            setsid ();
        }
        for (n = 0; n < 3; n ++) {
            close (n) ;
            if (dup2 (devnull, n) == n)
                continue ;
            logger (LOG_ERR, "Cannot dup /dev/null to %d: %s",
                    n, strerror (errno));
            exit (2) ;
        }
    }
    else {
        syslog_init (0, timepfx);
        close (1) ;
        if (dup2 (2, 1) != 1) {
            logger (LOG_ERR, "Cannot dup stdout to stderr: %s",
                    strerror (errno));
            exit (2);
        }
    }

    /* say hello */
    cpry ();

    /* record own process id */
    if (pid_fd >= 0) {
        char buf [20];
        sprintf (buf, "%d\n", getpid ());
        write (pid_fd, buf, strlen (buf));
        close (pid_fd);
    }

    /* initialise the server control socket before chroot */
    ctrl = ctrl_init (set_gid,
		      RCMD_CTRL_IOBUF_SIZE, RCMD_CTRL_REQ_PATH,
		      &verbosity);

    /* change root environment */
    if (new_root && (chroot (new_root) || chdir ("/"))) {
        logger (LOG_ERR, "chroot failed: %s", strerror (errno));
        exit (2);
    }

    /* initialise the action loop */
    s.loop = do_init (idle, &s);

    /* set verbosity */
    switch (verbose) {
    case 2:
        verbosity = COMM_LOG_MORE ;
        break ;
    case 1:
        verbosity = COMM_LOG_DEFLOAD ;
        break ;
    case 0:
        break ;
    default:
        verbosity = -1 ;
    }

    logger (LOG_INFO, "Starting command filtering subsytem");
    if ((s.filter = filter_open (re_subdim)) == 0) {
        logger (LOG_ERR, "Cannot initialise filter subsystem");
        exit (2);
    }

    n = filter_load_defaults (s.filter);
    logger (LOG_INFO, "Default filter sets have been loaded: %d records", n);

    if (verbosity)
        logger (LOG_INFO, "Opening pipe to the remote shell at %s", cmd_server);

    ok = shell_open (&s.cmd, user, idfile, cmd_server, port,
                     prompt, pool, rclocal, rcscript, no_iam_check) != 0 ;
    if (s.cmd == 0 || (force_startup == 0 && ok == 0)) {
        logger (LOG_ERR, "Cannot initialise command shell");
        shell_close (s.cmd);
        exit (2);
    }

    /* add services ... */
    logger (LOG_INFO, "Adding control socket services ...");
    if ((s.comm = comm_init (s.cmd, s.filter)) == 0) {
        logger (LOG_ERR, "Cannot initialise local commumication system");
        exit (2);
    }
    ctrl = ctrl_add (ctrl, MAIN_FILE_CMD, comm_file,    s.comm);
    ctrl = ctrl_add (ctrl, MAIN_CLI_CMD,  comm_string,  s.comm);
    ctrl = ctrl_add (ctrl, MAIN_RECONN,   comm_reopen,  s.comm);
    ctrl = ctrl_add (ctrl, MAIN_PING,     comm_ping,    s.comm);
    ctrl = ctrl_add (ctrl, MAIN_FLUSH,    comm_flush,   s.comm);
    ctrl = ctrl_add (ctrl, MAIN_KPALVE,   comm_kpalve,  s.comm);
    ctrl = ctrl_add (ctrl, MAIN_TEMPL,    comm_tmpl,    s.comm);
    ctrl = ctrl_add (ctrl, MAIN_ATEMPL,   comm_atpl,    s.comm);
    ctrl = ctrl_add (ctrl, MAIN_STOP,     comm_stop,    s.comm);
    ctrl = ctrl_add (ctrl, MAIN_REDO,     comm_redo,    s.comm);
    ctrl = ctrl_add (ctrl, MAIN_FIND,     comm_start,   s.comm);
    ctrl = ctrl_add (ctrl, MAIN_REPLACE,  comm_replace, s.comm);
    ctrl = ctrl_add (ctrl, MAIN_DELETE,   comm_del,     s.comm);
    ctrl = ctrl_add (ctrl, MAIN_RUN,      comm_run,     s.comm);
    ctrl = ctrl_add (ctrl, MAIN_FLIST,    comm_list,    s.comm);
    ctrl = ctrl_add (ctrl, MAIN_IODUMP,   comm_iodump,  s.comm);
    ctrl = ctrl_add (ctrl, "info",        comm_status,  s.comm);
    ctrl = ctrl_add (ctrl, "rusage",ctrl_resource,  ctrl);
    ctrl = ctrl_add (ctrl, "noisy", ctrl_verbosity, ctrl);
    ctrl = ctrl_add (ctrl, "help",  ctrl_usage,     ctrl);
    s.comm = comm_update (s.comm, ctrl);

       
    if (setjmp (go_here) == 0) {
        signal (SIGINT,  s_go_here);
        signal (SIGTERM, s_go_here);

        /* accept commands */
        do_action
            (s.loop, ctrl_fd (ctrl), (void(*)(void*))ctrl_handler, ctrl);

        do_action
            (s.loop, shell_fd (s.cmd), (void(*)(void*))shell_flush, s.cmd);
        
        for (;;) {
            n = do_loop (s.loop, idle_timeout * 1000);
            if (verbose > 1) 
                logger (LOG_INFO, "%s: do_loop -> %d\n", progname, n);
        }
        /*NOTREACHED*/
    }

    shell_close (s.cmd);

#   ifdef _MAIN_DEBUG
    /* release dispatcher */
    do_destroy (s.loop);
    
    /* release the server control socket */
    ctrl_destroy (ctrl) ;

    return 0;
#   else
    exit (0);
#   endif /* _MAIN_DEBUG */
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
