/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_SIGNAL_H
# include <signal.h>
#endif
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif
#ifdef HAVE_SIGNAL_H
# include <signal.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif
#ifdef HAVE_SETJMP_H
# include <setjmp.h>
#endif

#include <stdio.h>

#include "util.h"
#include "settings.h"
#include "ptycall.h"
#include "nvalloc.h"
#include "logger.h"
#include "tvcalc.h"

#define __SHELL_PRIVATE
#include "shell.h"

#ifdef WITH_DMALLOC
static unsigned long __mark;

#define DMALLOC_MARK() { __mark = dmalloc_mark (); }

#define DMALLOC_HUNT() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }

#define DMALLOC_VERIFY(c) {                             \
                 if (c) dmalloc_verify (c);             \
                 dmalloc_log_stats ();                  }

#define DMALLOC_MESSAGE(g,fmt, ...) {                   \
                 dmalloc_message (fmt, ##__VA_ARGS__) ; \
                 DMALLOC_VERIFY(g) ;                    }

#else
#define DMALLOC_MARK()            /* n/a */
#define DMALLOC_HUNT()            /* n/a */
#define DMALLOC_VERIFY(g)         /* n/a */
#define DMALLOC_MESSAGE(fmt, ...) /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * Private: definitions & declarations
 * ----------------------------------------------------------------------- */

#define DIM(x)   (sizeof (x) / sizeof ((x) [0]))

static unsigned shell_reopen (SHELL *s);

static jmp_buf rclocal_stop ;

static
const char *shell_dialogue_phrases [] = {
    SHELL_DIALOGUE_PHRASES
};

/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
unsigned alloc_file_data (SHELL         *s,
                          IOBUF         *u,
                          const char *file,
                          int       rcfile) {
    int fd ;
    struct stat st ;

    if ((fd = open (file, O_RDONLY)) < 0) {
        s->err.num = cannot_open_file ;
        goto error ;
    }
    if (fstat (fd, &st) < 0) {
        s->err.num = cannot_stat_file ;
        goto error ;
    }
    if (st.st_size > (rcfile ?SHELL_MAX_RCFILE_SIZE :SHELL_MAX_FILE_SIZE)) {
        s->err.num = file_too_large ;
        s->err.len = st.st_size ;
        goto error ;
    }

    /* read all into memory */
    kill_space  (u);
    file_append (u, fd, st.st_size);
    close (fd);

    /* pipe through the command channel */
    return 1 ;

error:
    s->err.no  = errno ;
    logger (LOG_ERR, "File \"%s\": %s", file, shell_error (s));
    return 0;
}



static
void kill_shell (SHELL *s) {
    struct timeval nap;
    
    if (s->cmd.io.fp) {
        if (s->cmd.dump_fp != 0)
            fwrite ("\n<CLOSE>\n", 9, 1, s->cmd.dump_fp);
        if (fclose (s->cmd.io.fp) < 0)
            logger (LOG_ERR, "fclose () returned error while terminating "
                    "the shell: %s", strerror (errno));
        XZERO (s->cmd.io);
        nap.tv_sec  =  SHELL_WAIT_AFTER_CLOSE / 1000 ;
        nap.tv_usec = (SHELL_WAIT_AFTER_CLOSE % 1000) * 1000 ;
        select (0, 0, 0, 0, &nap) ;
    }

    if (s->cmd.pid == 0)
        return;

    if (waitpid (s->cmd.pid, 0, WNOHANG) > 0) {
        s->cmd.pid = 0 ;
        return ;
    }

    if (verbosity & SHELL_LOG_TIME2DIE)
        logger (LOG_INFO, "Time to die: sending signal to process %d",
                s->cmd.pid);

    if (s->cmd.dump_fp != 0)
        fwrite ("\n<KILL>\n", 8, 1, s->cmd.dump_fp);

    if (kill (s->cmd.pid, SIGTERM) < 0)
        return ;

    nap.tv_sec  =  SHELL_WAIT_AFTER_KILL / 1000 ;
    nap.tv_usec = (SHELL_WAIT_AFTER_KILL % 1000) * 1000 ;
    select (0, 0, 0, 0, &nap) ;

    if (waitpid (s->cmd.pid, 0, WNOHANG) > 0) {
        s->cmd.pid = 0 ;
        return ;
    }

    logger (LOG_INFO, "Time to die (killing hard): "
            "sending signal to process %d", s->cmd.pid);
    kill (s->cmd.pid, SIGKILL);

    waitpid (s->cmd.pid, 0, WNOHANG) ;
    s->cmd.pid = 0 ;
}


unsigned lost_session (SHELL *s) {
    if (s->cmd.io.fp == 0 || s->cmd.pid == 0)
        return 1;

    if (s->cmd.dump_fp != 0)
        fwrite ("\n<GONE>\n", 8, 1, s->cmd.dump_fp);

    if (waitpid (s->cmd.pid, 0, WNOHANG) > 0)
        goto pid_clean_up ;
    
    if (kill (s->cmd.pid, 0) == -1)
        goto pid_clean_up ;

    return 0;

pid_clean_up:
    s->cmd.pid = 0 ;
    kill_shell (s);
    return 1 ;
}


static
int receive_ok (SHELL *s, struct timeval *tv) {
    fd_set fds;
    if (s->rd_cache.top)
        return 1;
    FD_ZERO (&fds);
    FD_SET (s->cmd.io.fd, &fds);
    return select (s->cmd.io.fd+1, &fds, 0, 0, tv);
}


static
size_t receive_data (SHELL *s, char *buf, size_t len, int cache_ok) {
    int n, m = 0 ;
    
    if (cache_ok) {
        fd_set fds;
        struct timeval tv;

        if (s->rd_cache.top) {
            if ((m = len) > s->rd_cache.top)
                m = s->rd_cache.top ;
            memcpy (buf, s->rd_cache.data, m);
            if (s->rd_cache.top <= m)
                kill_space (&s->rd_cache);
            else
                memmove (s->rd_cache.data,
                         s->rd_cache.data + m, s->rd_cache.top - m);
            if ((len -= m) <= 0)
                return len ;
            buf += m ;
        }

        FD_ZERO (&fds);
        FD_SET (s->cmd.io.fd, &fds);
        XZERO (tv);
        if (select (s->cmd.io.fd+1, &fds, 0, 0, &tv) <= 0)
            return m ;
    }

    if ((n = fread (buf, 1, len, s->cmd.io.fp)) > 0) {
        s->cmd.io.inbytes += n ;
        if (s->cmd.dump_fp != 0) {
            fwrite ("\n<< ", 4, 1, s->cmd.dump_fp);
            fwrite     (buf, n, 1, s->cmd.dump_fp);
        }
        return n + m;
    }

    return m ;
}


static
int send_ok (SHELL *s, int tmo_ms, int read_ok) {

    struct timeval tv ;
    fd_set fdw, *_fdr, fdr;

    FD_ZERO (&fdw);
    FD_SET (s->cmd.io.fd, &fdw);

    if (read_ok) {
        FD_ZERO (&fdr);
        FD_SET (s->cmd.io.fd, &fdr);
        _fdr = &fdr ;
    }
    else
        _fdr = 0 ;

    tv.tv_sec  =  tmo_ms / 1000 ;
    tv.tv_usec = (tmo_ms % 1000) * 1000 ;

    /* return code: -1 - error
                     0 - timeout, cannot write nor read
                     1 - can write
                     2 - cannot write but read is possible
    */

    switch (select (s->cmd.io.fd+1, _fdr, &fdw, 0, &tv)) {
    case 1:
        if (FD_ISSET (s->cmd.io.fd, &fdw) == 0)
            return 2 ;
    case 2:
        return 1 ;
    case 0:
        return 0 ;
    }

    return -1 ;
}


static
size_t send_data (SHELL *s, const char *d, size_t len, int reopen_ok) {

    const char *save_d   =   d ;
    int         save_len = len ;

    for (;;) {
        int n, m ;

        if ((m = len) > SHELL_INTERNAL_SEND_BLOCK)
            m = SHELL_INTERNAL_SEND_BLOCK ;

        switch (send_ok (s, SHELL_WAIT_SEND_JAM, 1 /* read test ok */)) {
        case 0: /* timeout */

            // logger (LOG_ERR, "select timeout waiting for output");
            errno = EAGAIN ;
            break ;

        case 1: /* can write */
        case_1:

            // logger (LOG_ERR, "select: write ok");
            if ((n = fwrite (d, 1, m, s->cmd.io.fp)) <= 0) {
                if (reopen_ok) break ;
                s->err.no  = errno ;
                s->err.num = write_error ;
                goto shell_error_exit ;
            }

            s->cmd.io.outbytes += n ;

            if (s->cmd.dump_fp != 0) {
                fwrite ("\n>> ", 4, 1, s->cmd.dump_fp);
                fwrite       (d, n, 1, s->cmd.dump_fp);
            }

            d   += n ;
            len -= n ;

            if (len > 0)
                continue ;

            return save_len ;

        case 2: /* can read */

            // logger (LOG_ERR, "select read ok");
            make_space (&s->rd_cache, SHELL_INTERNAL_RECV_BLOCK);
                
            if ((n = receive_data (s, s->rd_cache.data + s->rd_cache.top,
                                   SHELL_INTERNAL_RECV_BLOCK, 0)) <= 0) {
                if (n == 0) {
                    // logger (LOG_ERR, "select read ok - null bytes");
                    if ((n = send_ok (s, SHELL_WAIT_SEND_JAM, 0)) > 0)
                        goto case_1 ;
                    s->err.no  = n ? errno : ENODATA ;
                    s->err.num = read_while_no_write ;
                    goto shell_error_exit ;
                }
                break ;
            }
                
            s->rd_cache.top += n ;
            // logger (LOG_ERR, "select read ok -> %d bytes", n);
            continue;
                
        default:
            logger (LOG_ERR, "select error %d waiting for the output "
                    "channel: %s", errno, strerror (errno));
        }

        if (reopen_ok == 0) {
            s->err.no  = errno ;
            s->err.num = output_jam ;
            goto shell_error_exit ;
        }

        if (shell_reopen (s) == 0)
            return 0;

        /* retry */
        reopen_ok = 0 ;
        d   = save_d ;
        len = save_len ;
    }

    /* NOTREACHED */

shell_error_exit:
    logger (LOG_ERR, shell_error (s));
    return 0;
}


static
const char *xprintf (const char *format, ...) {
    IOBUF ioenv;
    va_list ap;
    va_start (ap, format);

    XZERO (ioenv);
    data_vprintf (&ioenv, format, ap);

    va_end (ap);
    return ioenv.data ; /* must be free()-d */
}


static
char* const* xenv (char* const* av, int *N, const char *str) {
    char **xv ;
    xv = (char**) zrealloc ((void*)av, (*N + 1) * sizeof (char*)) ;
    xv [*N] = (char*) str;
    if (str) (*N) ++ ;
    return (char* const*) xv ;
}


static
const char *envstr (const char *var) {
    const char *val = getenv (var) ;
    return val ? xprintf ("%s=%s", var, val) : 0 ;
}


static
void sig_wait (int sig) {
    signal (sig, SIG_IGN);
    longjmp (rclocal_stop, 1);
}


static
unsigned init_rc (SHELL *s) {
    if (s->rclocal == 0) {
        return 1;
    }
    else {
        int status, n ;
        const char *info = s->rcl_script ? "shell script" : "program" ;

        if ((s->rclocal_pid = fork ()) == 0) {
            /* child */
            char *const argv [3] ;
            char* const* envp = 0;

            setsid ();

            n    = 0 ;
            envp = xenv (envp, &n, envstr  ("PATH"));
            envp = xenv (envp, &n, envstr  ("HOME"));
            envp = xenv (envp, &n, envstr  ("SHELL"));
            envp = xenv (envp, &n, envstr  ("PWD"));
            envp = xenv (envp, &n, envstr  ("TMPDIR"));
            envp = xenv (envp, &n, envstr  ("LD_LIBRARY_PATH"));
            envp = xenv (envp, &n, envstr  ("LD_PRELOAD")) ;
            envp = xenv (envp, &n, xprintf ("RCMD_EXE=%s", s->rclocal));
            envp = xenv (envp, &n, 0);
           
            if (s->rcl_script) {
                ((const char**)argv) [0] =  "/bin/sh" ;
                ((const char**)argv) [1] = s->rclocal ;
                ((const char**)argv) [2] =          0 ;
                execve (argv [0], argv, envp);
            }
            else {
                if ((((char**)argv) [0] = strrchr (s->rclocal, '/')) == 0)
                    ((const char**)argv) [0] = s->rclocal ;
                else
                    ((char**)argv) [0] ++ ;
                ((char**)argv) [1] = 0 ;
                execve (s->rclocal, argv, envp);
            }
            
            logger (LOG_ERR, "Executing %s %s failed: ",
                    info, s->rclocal, strerror (errno));

            exit (2);
        }

        if (s->rclocal_pid == -1) {
            logger (LOG_ERR, "Cannot fork for executing %s %s: %s",
                    info, s->rclocal, strerror (errno));
            return 0;
        }

        if (s->cmd.dump_fp != 0)
            fwrite ("\n<RCINIT>\n", 10, 1, s->cmd.dump_fp);

        if (waitpid (s->rclocal_pid, &status, WNOHANG) != s->rclocal_pid) {
            struct sigaction sa, oa;
            unsigned     save_time = alarm (0);
            pid_t         wait_pid = 0;

            if (verbosity & SHELL_LOG_RCSCRIPT_LOG)
                logger (LOG_INFO, "Waiting at most %u secs for %s %s "
                        "to terminate, pid=%d", s->rclocal_ttd, info,
                        s->rclocal, s->rclocal_pid);

            XZERO (sa);
            sa.sa_handler = sig_wait;
            sa.sa_flags   = SA_NODEFER;

            if (setjmp (rclocal_stop) == 0) {
                volatile pid_t xpid ;
                sigaction (SIGALRM, &sa, &oa);
                alarm     (s->rclocal_ttd) ;
                xpid = waitpid (s->rclocal_pid, &status, 0);
                wait_pid = xpid ;
                if (verbosity & SHELL_LOG_RCSCRIPT_LOG)
                    logger (LOG_INFO, "%s %s has terminated with status %d",
                            info, s->rclocal, WEXITSTATUS (status));
            }
            
            alarm (0) ;
            sigaction (SIGALRM, &oa, 0);
            alarm (save_time);
            
            switch (wait_pid) {
                struct timeval nap;
            case 0:
                if (verbosity & SHELL_LOG_RCSCRIPT_LOG)
                    logger (LOG_ERR, "Sending termination signal to %s %s",
                            info, s->rclocal, strerror (errno));
                if (kill (-s->rclocal_pid, SIGTERM) == 0) {
                    nap.tv_sec  =  SHELL_WAIT_AFTER_KILL / 1000 ;
                    nap.tv_usec = (SHELL_WAIT_AFTER_KILL % 1000) * 1000 ;
                    select (0, 0, 0, 0, &nap) ;
                }
                if ((wait_pid = waitpid
                     (s->rclocal_pid, &status, WNOHANG)) > 0)
                    break ;
                if (wait_pid == 0) {
                    if (kill (-s->rclocal_pid, SIGKILL) == 0) {
                        nap.tv_sec  =  SHELL_WAIT_AFTER_KILL / 1000 ;
                        nap.tv_usec = (SHELL_WAIT_AFTER_KILL % 1000)*1000;
                        select (0, 0, 0, 0, &nap) ;
                    }
                    if ((wait_pid = waitpid
                         (s->rclocal_pid, &status, WNOHANG)) > 0)
                        break ;
                }
                /* proceed: wait_pid == -1 */
                
            case -1:
                logger (LOG_ERR, "Error waiting for %s %s (pid=%d): %s",
                        info, s->rclocal, s->rclocal_pid, strerror (errno));
                s->rclocal_pid = -1;
                return 0;
            }
        }
        else {
            if (verbosity & SHELL_LOG_RCSCRIPT_LOG)
                logger (LOG_INFO, "Executed %s %s", info, s->rclocal);
        }

        s->rclocal_pid = 0 ;
        
        if ((n = WEXITSTATUS (status)) == 0) {
            s->cmd.io.rc_init = 1 ;
            return 1 ;
        }
        
        return 0;
    }
}


static
unsigned send_rc (SHELL *s) {

    if (s->rcscript == 0) {
        return 1;
    }
    else {
        IOBUF  u ;
        XZERO (u);
        
        /* read contents */
        if (alloc_file_data (s, &u, s->rcscript, 0)) {
            int n ;
            if (verbosity & SHELL_LOG_RCSCRIPT_LOG)
                logger (LOG_INFO, "Sending rc-script %s", s->rcscript);
            if (s->cmd.dump_fp != 0)
                fwrite ("\n<RCDATA>\n", 10, 1, s->cmd.dump_fp);
            n = send_data (s, u.data, u.top, 0) ;
            free (u.data);
            if (n > 0) {
                s->cmd.io.rc_sent = 1 ;
                return n ;
            }
        }
    }

    return 0;
}


static
void shell_flush_internally (SHELL       *s,
                             int    timeout,
                             int reread_tmo,
                             int  reconnect) {
    int n ;
    struct timeval tmo ;

    /* timeout for the first check, only  */
    tmo.tv_sec  =  timeout / 1000 ;
    tmo.tv_usec = (timeout % 1000) * 1000 ;

    kill_space (&s->data);
    if (reread_tmo == 0)
        /* we append to the input buffer when a reread timeout is set */
        make_space (&s->data, RCMD_CTRL_IOBUF_SIZE + 1);
    
    /* flush remainder data from the last session */
    for (;;) {
        switch (n = receive_ok (s, &tmo)) {
        case -1:
        test_reconnect:
            if (reconnect && lost_session (s)) {
                if (verbosity & SHELL_LOG_SESSION)
                    logger (LOG_INFO, "Lost session - reconnecting ...");
                shell_reopen (s);
            }
            /* break; */

        case 0:
            break ;

        default:
            if (reread_tmo)
                /* append to the input buffer when a reread timeout is set */
                make_space (&s->data, RCMD_CTRL_IOBUF_SIZE + 1);
            else
                /* be ready for we overwriting the input buffer */
                s->data.top = 0 ;
        
            /* read data - append to the input buffer */
            if ((n = receive_data (s, s->data.data + s->data.top,
                                   RCMD_CTRL_IOBUF_SIZE, 1)) <= 0)
                /* select will return 1 on eof - so check for it */
                goto test_reconnect ;

            if (reread_tmo)
                /* append to the input buffer when a reread timeout is set */
                s->data.top += n;
            else if (verbosity & SHELL_LOG_REPLY_DUMP) {
                s->data.data [s->data.top + n] = '\0';
                logger (LOG_ERR, "Flushing message from the shell: %s",
                        s->data.data + s->data.top);
            }

            tmo.tv_sec  =  reread_tmo / 1000 ;
            tmo.tv_usec = (reread_tmo % 1000) * 1000 ;
            continue;
        }

        break ;
    }

    /* keep the string space ready for overwriting */
    if (s->data.data) {
        s->data.data [s->data.top] = '\0';
        s->data.top = 0 ;
    }
}


static
unsigned intro_message_ok (SHELL *s) {
    int n ;

    if (s->data.data == 0)
        return 1;

    for (n = 0; n < DIM (shell_dialogue_phrases); n ++) {
        char *p = s->data.data, *q ;
        size_t len = strlen (shell_dialogue_phrases [n]) ;
        /* note that the last character from unwanted [] must not be
           case sensitive - otherwise two list entries are needed */
        while ((q = strchr (p, shell_dialogue_phrases [n] [len - 1])) != 0) {
            if (q ++ && q - p >= len &&
                strncasecmp (q - len, shell_dialogue_phrases [n], len) == 0){
                s->err.num = dialogue_mode_shell ;
                s->err.no  = n ;
                logger (LOG_ERR, "Shell error: %s", shell_error (s));
                return 0 ;
            }
            p = q ;
        }
    }

    return 1;
}


static
unsigned shell_reopen (SHELL *s) {
    int n, rc_checked, rc_initok ;

    kill_shell (s);

    /* initialise locally */
    rc_initok = (init_rc (s) != 0) ;

    if ((s->cmd.io.fp = ptyexecvp
         (SSH_EXEC_PATH, s->cmd.argv, &s->cmd.pid)) == 0) {
        s->err.no  = errno ;
        s->err.num = reinit_error ;
        logger (LOG_ERR, shell_error (s));
        return 0;
    }

    s->cmd.io.started = time (0);
    s->cmd.io.fd      = fileno (s->cmd.io.fp);

    if (s->cmd.dump_fp != 0)
        fwrite ("\n<OPEN>\n", 8, 1, s->cmd.dump_fp);

    /* flush introducer data .. */
    shell_flush_internally
        (s, SHELL_WAIT_HELLO_CONNECT, COMM_WAIT_REREAD_DEFAULT, 1);

    /* Check for the string "Password:" in the introducer text. */
    if (s->no_iamode_check != 0)
        rc_checked = 1 ;
    else if (s->data.data == 0)
        rc_checked = 0 ;
    else if (intro_message_ok (s) == 0)
        return 0 ;
    else
        rc_checked = 1 ;
    
    /* initialise on the server */
    if (rc_initok && send_rc (s) > 1)
        shell_flush_internally
            (s, SHELL_WAIT_HELLO_CONNECT, COMM_WAIT_REREAD_DEFAULT, 0);

    if (waitpid (s->cmd.pid, &n, WNOHANG) <= 0 || rc_checked == 0) {
        struct timeval tmo ;

        /* wait some time to make sure the child is running */
        tmo.tv_sec  =  SHELL_WAIT_AFTER_CONNECT / 1000 ;
        tmo.tv_usec = (SHELL_WAIT_AFTER_CONNECT % 1000) * 1000 ;
        select (0, 0, 0, 0, &tmo) ;

        shell_flush_internally
            (s, SHELL_WAIT_HELLO_CONNECT, COMM_WAIT_REREAD_DEFAULT, 0);

        if (waitpid (s->cmd.pid, &n, WNOHANG) <= 0) {

            if (s->data.data && s->data.data [0] &&
                (verbosity & SHELL_LOG_REPLY_DUMP))
                logger (LOG_ERR, "Shell hello message: %s", s->data.data);

            /* Check for the string "Password:" in the introducer text. */
            if (intro_message_ok (s) == 0)
                return 0 ;
            
            return 1;
        }
    }

    if (s->data.data && s->data.data [0])
        logger (LOG_ERR, "Shell termination message: %s", s->data.data);

    s->err.num = child_error ;
    logger (LOG_ERR, "Shell error: %s (status %d)",
            shell_error (s), WEXITSTATUS (n));

    return 0 ;
}


static
const char *read_data (SHELL                    *s,
                       const char               *d,
                       size_t                  len,
                       int                 timeout,
                       int (*filter)(void*,IOBUF*),
                       void           *filter_data,
                       int           keep_alive_ok) {
    int n;
    unsigned cutnschift = 0 ;
    struct timeval tmo ;

    if (keep_alive_ok && s->echo.keep_alive)
        shell_contest (s, timeout, 0);

    if (d != 0) {
        if (len == 0)
            len = strlen (d) ;
    
        shell_flush_internally (s, 0, 0, 1 /* reconnect */);

        if (waitpid (s->cmd.pid, 0, WNOHANG) > 0)
            s->cmd.pid = 0 ;

        if (s->cmd.io.rc_init == 0 && shell_reopen (s) == 0)
            return 0 ;
            
        if (s->cmd.io.rc_sent == 0)
            switch (send_rc (s)) {
            case 0:
                return 0 ;
            case 1:
                break ;
            default:
                if (verbosity & SHELL_LOG_REPLY_STATUS)
                    logger (LOG_INFO, "Flushing data after rc-script "
                            "before processing data");
                shell_flush_internally (s, 0, 0, 0);
            }

        if (verbosity & SHELL_LOG_REPLY_STATUS) {
            char *s, buf [15] ;
            if (len > 15) {
                s = memcpy (buf, d, 10) ;
                memcpy (buf + 10, "...", 4) ;
            }
            else
                s = (char*)d ;
            logger (LOG_INFO, "Sending new command: \"%s\"", s);
        }

        if (send_data (s, d, len, 1) == 0)
            return 0;
    }
    else if (verbosity & SHELL_LOG_REPLY_STATUS)
        logger (LOG_INFO, "Continue reading/parsing input ...");
        

    /* collect response - timeout for the first check, only  */
    tmo.tv_sec  =  timeout / 1000 ;
    tmo.tv_usec = (timeout % 1000) * 1000 ;

    /* get response */
    for (;;) {
        int stop_reading ;

        switch (receive_ok (s, &tmo)) {
        case -1:
        test_reconnect:
            if (lost_session (s)) {
                if (verbosity & SHELL_LOG_SESSION)
                    logger (LOG_INFO, "Lost session - reconnecting ...");
                shell_reopen (s);
            }
            /* break; */

        case 0:
            break ;

        default:
            /* read data - stop if there are none */
            make_space (&s->data, SHELL_INTERNAL_RCV_BUF_SIZE + 1);
            if ((n = receive_data (s, s->data.data + s->data.top,
                                    SHELL_INTERNAL_RCV_BUF_SIZE, 1)) <= 0)
                /* select will return 1 on eof - so check for it */
                goto test_reconnect ;
                
            s->data.top += n ;
            
            /* check data */
            if (filter && filter (filter_data, &s->data) == 0) {
                stop_reading = 1 ;
                cutnschift   = 0 ;
            }
            else
                stop_reading = 0 ;

            tmo.tv_sec  =  COMM_WAIT_REREAD_DEFAULT / 1000 ;
            tmo.tv_usec = (COMM_WAIT_REREAD_DEFAULT % 1000) * 1000 ;

            if (stop_reading == 0) {

                /* we cannot handle more than that bytes at once */
                if (s->data.top > SHELL_INTERNAL_RCV_BUF_SIZE) {
                    if (verbosity & SHELL_LOG_STRIP_REPLY)
                        logger (LOG_INFO, "Relpy data too large - %s tail "
                                "(%d bytes)",
                                (s->save_resp_headr
                                 ? "cutting tail" : "shifting"),
                                (s->save_resp_headr
                                 ? s->data.top - SHELL_INTERNAL_RCV_BUF_SIZE
                                 : SHELL_INTERNAL_RCV_BUF_SIZE));
                    if (s->save_resp_headr == 0) {
                        memmove (s->data.data,
                                 (s->data.data +
                                  s->data.top - SHELL_INTERNAL_RCV_BUF_SIZE),
                                 SHELL_INTERNAL_RCV_BUF_SIZE);
                        cutnschift = 1 ;
                    }
                    s->data.top = SHELL_INTERNAL_RCV_BUF_SIZE ;
                }
        
                continue ;
            }

            /* we cannot send back CTRL_IOBUF_SIZE bytes at once */
            if (s->data.top > RCMD_CTRL_IOBUF_SIZE) {
                if (verbosity & SHELL_LOG_STRIP_REPLY)
                    logger (LOG_INFO, "Data too large for the reply "
                            "buffer - %s tail (%d bytes)",
                            (s->save_resp_headr ?"cutting tail" :"shifting"),
                            (s->save_resp_headr
                             ? s->data.top - RCMD_CTRL_IOBUF_SIZE
                             : RCMD_CTRL_IOBUF_SIZE));
                if (s->save_resp_headr == 0) {
                    memmove (s->data.data,
                             (s->data.data +
                              s->data.top - RCMD_CTRL_IOBUF_SIZE),
                             RCMD_CTRL_IOBUF_SIZE);
                    cutnschift = 1 ;
                }
                s->data.top = RCMD_CTRL_IOBUF_SIZE ;
            }
        }

        break;
    }

    if (s->cmd.dump_fp != 0)
        fflush (s->cmd.dump_fp);

    if (s->data.data == 0)
        return "" ;

#   if    RCMD_CTRL_IOBUF_SIZE < 6
#   error RCMD_CTRL_IOBUF_SIZE must be a lot larger than 6!!!
#   endif

    if (cutnschift)
        memcpy (s->data.data, "[...]\n", 6);

    for (;;) {
        /* strip trailing new lines */
        while (s->data.top > 1 && iscntrl (s->data.data [s->data.top - 1]))
            s->data.top -- ;
        s->data.data [s->data.top] = '\0';

        /* strip trailing prompt */
        if (s->prompt.len               &&
            s->prompt.len < s->data.top &&
            strcmp (s->data.data + s->data.top-s->prompt.len,
                    s->prompt.ps) == 0) {
            if (verbosity & SHELL_LOG_STRIP_REPLY) {
#               define __blurb "Command prompt tail: stripping %d " \
                               "characters from the reply"
#               ifdef USE_DEBUG
                logger (LOG_INFO, __blurb ": %s", s->prompt.len,
                        s->data.data + s->data.top - s->prompt.len);
#               else
                logger (LOG_INFO, __blurb, s->prompt.len);
#               endif
#               undef __blurb 
            }
            s->data.data [s->data.top -= s->prompt.len] = '\0';
            continue ;
        }
        break ;
    }

    /* check whether the channel reflects the command itself and strip it */
    while (len > 1 && isspace (d [len - 1])) len -- ;
    if (strncmp (s->data.data, d, len) == 0) {
        d = s->data.data + len ;
        if (len && (verbosity & SHELL_LOG_STRIP_REPLY) != 0) {
#           define __blurb "Reflected request prefix: stripping %d " \
                           "characters from the reply"
#           ifdef USE_DEBUG
            int save = s->data.data [len];
            s->data.data [len] = '\0';
            logger (LOG_INFO, __blurb ": %s", len, s->data.data);
            s->data.data [len] = save ;
#           else
            logger (LOG_INFO, __blurb, len);
#           endif
#           undef __blurb 
        }
    }
    else
        d = s->data.data ;

    /* strip leading spaces */
    while (isspace (d [0])) d ++ ;
    return d ;
}


unsigned shell_kpalve (SHELL *s, int yes) {
    int save = s->echo.keep_alive ;
    if (yes != -1)
        s->echo.keep_alive = (yes != 0) ;
    return save ;
}


const char *shell_file (SHELL                   *s,
                        const char            *file,
                        int (*filter)(void*,IOBUF*),
                        void           *filter_data,
                        int                 timeout) {

    if (strchr (file, '/')) {
        s->err.num = illegal_file_path ;
        s->err.no  = errno ;
        logger (LOG_ERR, "File \"%s\": %s", file, shell_error (s));
    }
    else {
        /* assemble file name */
        IOBUF  u ;
        XZERO (u);
        data_printf (&u, "%s/%s", s->pool, file);

        /* read contents */
        if (alloc_file_data (s, &u, u.data, 0)) {
            /* pipe through the command channel */
            const char *p = read_data (s, u.data, u.top, timeout, 0, 0, 1);
            free (u.data);
            return p;
        }
    }

    return 0 ;
}

/* ----------------------------------------------------------------------- *
 * Public: Connectionless communication handlers
 * ----------------------------------------------------------------------- */

SHELL *shell_open (SHELL         **Desc,        
                   const char     *user,
                   const char   *idfile,
                   const char      *trg,
                   uint16_t        port,
                   const char   *prompt,
                   const char     *pool,
                   const char  *rclocal,
                   const char *rcscript,
                   int  no_iamode_check) {
    SHELL  s ;
    XZERO (s);

    /* set up the (re-)connection command */
    s.cmd.argv = (const char**)append_argv (0, SSH_PROG_NAME);
    if (user) {
        s.cmd.argv = (const char**)append_argv
            ((char**)s.cmd.argv, SHELL_SSH_USER_OPTION);
        s.cmd.argv = (const char**)append_argv
            ((char**)s.cmd.argv, user);
    }
    if (idfile) {
        if (idfile [0] != '/') {
            logger (LOG_ERR, "Not accepting identity file \"%s\" - full "
                    "path expected", idfile);
            goto just_exit;
        }
        s.cmd.argv = (const char**)append_argv
            ((char**)s.cmd.argv, SHELL_SSH_IDFILE_OPTION);
        s.cmd.argv = (const char**)append_argv
            ((char**)s.cmd.argv, idfile);
    }
    if (port) {
        char pbuf [10] ;
        s.cmd.argv = (const char**)append_argv
            ((char**)s.cmd.argv, SHELL_SSH_PORT_OPTION);
        sprintf (pbuf, "%d", port & 0xffff);
        s.cmd.argv = (const char**)append_argv
            ((char**)s.cmd.argv, pbuf);
    }
    s.cmd.argv = (const char**)append_argv
        ((char**)s.cmd.argv, SHELL_SSH_PTY_OPTION);
    s.cmd.argv = (const char**)append_argv
        ((char**)s.cmd.argv, trg);

    if ((s.prompt.ps = prompt) == 0 || *s.prompt.ps == '\0')
        s.prompt.ps = SHELL_STRIP_PROMPT ;
    s.prompt.ps = zstrdup (s.prompt.ps) ;
    s.prompt.len = strlen (s.prompt.ps) ;

    if ((s.pool = pool) == 0 || *s.pool == '\0')
        s.pool = SHELL_DEFAULT_POOL ;
    s.pool = zstrdup (s.pool) ;
    
    if (rclocal) {
        s.rclocal = zstrdup (rclocal) ;
        if (strcmp (s.rclocal + strlen (s.rclocal) - 3, ".sh") == 0)
            s.rcl_script = 1 ;
    }

    if (rcscript)
        s.rcscript = zstrdup (rcscript) ;

    if (no_iamode_check)
        s.no_iamode_check = 1 ;

    s.echo.keep_alive      =  SHELL_KEEP_ALIVE_DEFAULT ;
    s.echo.min_wait_repeat = (SHELL_WAIT_REPEAT_ECHO_REQ + 500) / 1000 ;

#   if      SHELL_RCLOCAL_TTD < 500
#   warning SHELL_RCLOCAL_TTD adjusted to 1000 ms
#   define  SHELL_RCLOCAL_TTD 1000
#   endif

    s.rclocal_ttd = (SHELL_RCLOCAL_TTD + 500) / 1000 ;

#   define e_req SHELL_ECHO_REQUEST_CMD " "
    s.echo.cmd_len = strlen (e_req);
    s.echo.req_len = s.echo.cmd_len +
        SHELL_ECHO_REQUEST_BLOCKS * (SHELL_ECHO_REQUEST_BLOCK_LEN + 1);
    s.echo.req     = memcpy
        (xmalloc (s.echo.req_len + 1), e_req, s.echo.cmd_len + 1);
    s.echo.req [s.echo.req_len] = '\0' ;
#   undef e_req

    if (shell_reopen (&s)) {
        SHELL *d = zmemdup (&s) ;
        if (verbosity & SHELL_LOG_OPEN_LOG)
            logger (LOG_INFO, "Shell/ssh channel has been opened");
        if (Desc)
            *Desc = d ;
        return d;
    }
    if (verbosity & SHELL_LOG_OPEN_LOG)
        logger (LOG_INFO, "Failed to open shell/ssh channel");

    /* no_iamode_check => (s.err.num != dialogue_mode_shell) */
    if (s.err.num != dialogue_mode_shell && Desc) {
        *Desc = zmemdup (&s) ;
        return 0;
    }
     
    free ((char*)s.pool);
    free ((char*)s.prompt.ps);
    
just_exit:
    destroy_argv ((char**)s.cmd.argv);
    return 0;
}


void shell_term (SHELL *s) {
    kill_shell (s);
}


int shell_fd (SHELL *s) {
    return s->cmd.io.fd ;
}


void shell_close (SHELL *s) {
    
    if (s) {
        if (s->rclocal_pid > 0)
            kill (-s->rclocal_pid, SIGTERM);
        if (waitpid (s->rclocal_pid, 0, WNOHANG) > 0)
            s->rclocal_pid = 0;
        kill_shell (s);
        destroy_argv ((char**)s->cmd.argv);
        kill_space (&s->data);
        free ((char*)s->pool);
        free ((char*)s->prompt.ps);
        if (s->echo.req)
            free ((char*)s->echo.req);
        if (s->rclocal)
            free ((char*)s->rclocal);
        if (s->rcscript)
            free ((char*)s->rcscript);
        if (s->cmd.dump_fp && fileno (s->cmd.dump_fp) > 2)
            fclose (s->cmd.dump_fp);
        if (s->rclocal_pid > 0)
            kill (-s->rclocal_pid, SIGKILL);
        waitpid (s->rclocal_pid, 0, WNOHANG);
        free (s);
    }
}



void shell_flush (SHELL *s) {
    shell_flush_internally (s, 0, 0, 1 /* reconnect */);
}


unsigned shell_restart (SHELL *s) {

    if (verbosity & SHELL_LOG_SESSION)
        logger (LOG_INFO, "Starting new ssh/shell session ...");

    return shell_reopen (s);
}


unsigned shell_contest (SHELL *s, int timeout, int force) {

    const char *rply, *req_data ;
    int n ;

    if (force == 0 && time (0) < s->echo.wait_repeat) {
        s->err.no  = 0 ;
        s->err.num = echo_too_early ;
        if (verbosity & SHELL_LOG_REPLY_DUMP)
            logger (LOG_INFO, shell_error (s));
        return 0 ;
    }

    /* set the time when the next keep alive check is possible */
    s->echo.wait_repeat = time (0) + s->echo.min_wait_repeat ;

    req_data = s->echo.req + s->echo.cmd_len ;
    for (n = 0 ; n < s->echo.req_len - s->echo.cmd_len; n ++) {
        if ((n % (SHELL_ECHO_REQUEST_BLOCK_LEN+1)) !=
            SHELL_ECHO_REQUEST_BLOCK_LEN) {
            char m = 10, c [1] ;
            do {
                rndata (c, 1);
                c [0] +=  '0' ;
                c [0] &= 0x7f ;
            }
            while (-- m && isalnum (c [0]) == 0) ;
            ((char*)req_data) [n] = m ? c [0] : 'a' + (n % 26) ;
        }
        else
            ((char*)req_data) [n] = ' ' ;
    }
    s->echo.req [s->echo.req_len - 1] = '\n' ;
    req_data = memcpy (xmalloc (n), req_data, n);
    ((char*)req_data) [n - 1] = '\0' ;

    shell_flush_internally (s, 0, 0, 0);
        
    if (verbosity & SHELL_LOG_REPLY_DUMP)
        logger (LOG_INFO, "Sending echo request: %s", s->echo.req);

    if ((rply = read_data
         (s, s->echo.req, s->echo.req_len, timeout, 0, 0, 0)) != 0) {

        if (rply [0] == '\0') /* not the answer expected, yet */
            rply = read_data (s, 0, 0, SHELL_WAIT_REDO_ECHO_REQ, 0, 0, 0) ;

        if (rply) {
            if (strcmp (rply, req_data) == 0) {
                free ((char*)req_data) ;
                if (verbosity & SHELL_LOG_REPLY_DUMP)
                    logger (LOG_INFO, "Accepted echo reply: %s.", rply);
                return 1 ;
            }

            if (verbosity & SHELL_LOG_REPLY_DUMP)
                logger (LOG_INFO, "No proper echo response: %s.", rply);
        }
    }

    free ((char*)req_data) ;
    return shell_restart (s) ? 2 : 0;
}


const char *shell_data (SHELL                    *s,
                        const char               *d,
                        size_t                  len,
                        int (*filter)(void*,IOBUF*),
                        void           *filter_data,
                        int                 timeout) {
    /* d != 0: do not check for keep alive when read continuation applies */
    return read_data (s, d, len, timeout, filter, filter_data, d != 0);
}


void shell_status (SHELL             *s,
                   time_t       *uptime,
                   size_t           *in,
                   size_t          *out,
                   const char **rc_init,
                   const char **rc_sent,
                   unsigned   *requests,
                   unsigned   *kp_alive) {
    if (uptime)   *uptime   = time (0) - s->cmd.io.started;
    if (in)       *in       = s->cmd.io.inbytes ;
    if (out)      *out      = s->cmd.io.outbytes ;
    if (rc_init)  *rc_init  = s->cmd.io.rc_init ? s->rclocal : 0 ;
    if (rc_sent)  *rc_sent  = s->cmd.io.rc_sent ? s->rcscript : 0 ;
    if (requests) *requests = s->cmd.io.requests ;
    if (kp_alive) *kp_alive = s->echo.keep_alive ;
}


FILE *shell_iodump (SHELL *s, FILE *fp) {
    FILE *old = s->cmd.dump_fp ;
    s->cmd.dump_fp = fp ;
    return old ;
}


const char *shell_error (SHELL *s) {
    kill_space (&s->data);

    switch (s->err.num) {
    case illegal_file_path:
        data_printf (&s->data, "path not allowed: can only use "
                     "plain file names");
        break ;
    case reinit_error:
        data_printf (&s->data, "cannot initialise new " SSH_PROG_NAME
                     " session: %s", strerror (errno));
        break ;
    case child_error:
        data_printf (&s->data, SSH_PROG_NAME " session stopped early");
        break ;
    case write_error:
        data_printf (&s->data, "error writing shell data: %s",
                     strerror (s->err.no));
        break ;
    case cannot_open_file:
        data_printf (&s->data, "cannot open: %s",
                     strerror (s->err.no));
        break ;
    case cannot_stat_file:
        data_printf (&s->data, "cannot stat: %s",
                     strerror (s->err.no));
        break ;
    case file_too_large:
        data_printf (&s->data, "file size %d is too large: max %u bytes",
                     s->err.len, SHELL_MAX_FILE_SIZE);
        break ;
    case rcfile_too_large:
        data_printf (&s->data, "rcscript size %d is too large: max %u bytes",
                     s->err.len, SHELL_MAX_RCFILE_SIZE);
        break ;
    case echo_too_early:
        data_printf (&s->data,
                     "Not sending echo request: retry time not reached "
                     "(%d seconds delay)", s->echo.min_wait_repeat);
        break ;
    case dialogue_mode_shell:
        data_printf (&s->data, "The text \"%s\" was seen - the ssh/shell "
                     "seems to run in dialogue mode",
                     shell_dialogue_phrases [s->err.no]);
        break ;
    case output_jam:
        data_printf (&s->data, "cannot write data fast enough: %s",
                     strerror (s->err.no));
        break ;
    case read_while_no_write:
        data_printf (&s->data, "failed to read while output channel has "
                     "blocked: %s", strerror (s->err.no));
        break ;
    default:
        return "no error" ;
    }

    return s->data.data ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
