/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies                                   *
 *   henry.margies@gmx.de                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "redconfig.h"


static void print_script(char* s)
{
	char* c;
	
	while( c = strchr(s, '\n') ) {
		*c = 0;
		printf(".. %s\n", s);
		s = c+1;
	}
}

/**
 * Reads a value from redconfig.
 * \param fd is the file descriptor to read from.
 * \param type is the type to read.
 * \returns zero if reading was successful
 */
static int mtd_read_value(int fd, int type)
{
	int i,b,r;
	char buffer[MAX_SIZE];
	unsigned char ip[8];
	
	switch (type) {
		case CONFIG_BOOL:
		case CONFIG_INT:
			r = mtd_get_int(fd, &b);
			printf("Value: %d\n\n", b);
			break;
		case CONFIG_STRING:
		case CONFIG_NETPORT:
			r = mtd_get_string(fd, buffer, STRING_SIZE);
			printf("Value: %s\n\n", buffer);
			break;
		case CONFIG_SCRIPT:
			r = mtd_get_string(fd, buffer, SCRIPT_SIZE);
			printf("Value:\n");
			print_script(buffer);
			printf("\n");
			break;
		case CONFIG_IP:
			r = mtd_get_ip(fd, ip);
			printf("Value: %i.%i.%i.%i\n\n", ip[0],ip[1],ip[2],ip[3]);
			break;
		case CONFIG_ESA:
			r = mtd_get_mac(fd, ip);
			printf("Value: %02x:%02x:%02x:%02x:%02x:%02x\n\n", 
					ip[0],ip[1],ip[2],ip[3],ip[4],ip[5]);
			break;
		default:
			fprintf(stderr, "Unknown type\n");
			return -1;
	}

	return r;
}

/**
 * Reads an item from redconfig.
 * \param fd is the file descriptor to read from.
 * \returns zero if reading was successul.
 */
static int mtd_read_item(int fd) 
{
	header_t header;
	int 	 ret;
	char	 buffer[1024];

	/* read header information */
	ret = read(fd, &header, sizeof(header));
	if (0 == ret)
		return 0;
	
	if (ret != sizeof(header))
		return -1;

	while (header.type == CONFIG_EMPTY) {
		header.type = header.name_length;
		header.name_length = header.sense;
		header.sense = header.dependency_length;
		ret = read(fd, &header.dependency_length, 1);
		if (ret!= 1)
			return -1;
	}

	if (	header.type == 0xde && header.name_length == 0xad &&
		header.sense== 0xde && header.dependency_length == 0xad )
	{
		int crc32;
		mtd_get_int(fd, &crc32);
		printf("\nChecksum: %d\n", crc32);
		return 1;
	}

	/* print header */
	printf("Header.Type: %d\n", header.type);
	printf("Header.NameLength: %d\n", header.name_length);
	printf("Header.Sense: %d\n", header.sense);
	printf("Header.DependencyLength: %d\n", header.dependency_length);

	/* read and print name of entry */
	if (header.name_length) {
		ret = read(fd, buffer, header.name_length);
		if (ret != header.name_length)
			return -2;

		printf("Name: %s\n", buffer);
	}

	/* read and print dependency name of available */
	if (header.dependency_length) {
		ret = read(fd, buffer, header.dependency_length);
		if (ret != header.dependency_length)
			return -3;
		printf("Dependency: %s\n", buffer);
	}
		
	/* read and print value */
	return mtd_read_value(fd, header.type);
}


static int mtd_dump(int fd)
{
	int file_length;
	int magic;
	int ret;
	
	/* read filelength */
	ret = read(fd, &file_length, sizeof(file_length));
	if (ret != sizeof(file_length)) {
		fprintf(stderr, "Could not read file length information (%d)\n", ret);
		return -1;
	}
	printf("Filelength: %d\n", file_length);
	
	/* read magic number */
	ret = read(fd, &magic, sizeof(magic));
	if (ret != sizeof(magic)) {
		fprintf(stderr, "Could not read magic number\n");
		return -1;
	}
	printf("Magic: %x\n", magic);

	fd = use_tmp_file(fd, file_length, magic);
	if (fd == -1)
		return -1;

	/* read rest of config */
	while( !mtd_read_item(fd) );

	return 0;
}

int main()
{
	int ret;
	int  fd = 0;

	fd = open_mtd_by_name("RedBoot config", O_RDONLY);
	if (-1 == fd) {
		perror("Unable to open device: ");
		return 10;
	}

	ret = mtd_dump(fd);
	close(fd);

	return ret;
}
