/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies                                   *
 *   henry.margies@gmx.de                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "redconfig.h"

#define PERROR(fmt, args...) 	fprintf(stderr,"[%s:%d] Error: " fmt,	\
							__FILE__,	\
							__LINE__, 	\
							##args)

/* 
 * Reads the 4 byte header into p_header.
 * Returns 0 on success
 */
int read_header(int fd, header_t* p_header)
{
	int ret;

	ret = read(fd, p_header, sizeof(header_t));
	if (ret == sizeof(header_t))
		return NO_ERROR;

	if (0 == ret)
		return ERROR_EOF;

	PERROR("Unable to read enough bytes (wanted %d, got %d)\n",sizeof(header_t),ret);
	return ERROR_NOT_ENOUGH;
}


/*
 * Compares the search_string with the next string from fd.
 * Returns 1, if both strings are equal, 0 if not, anything else if an error 
 * occurred
 */
int is_string(int fd, int length, char* search_string)
{
	char buffer[4096];
	int ret;

	if (0 == length)
		return ERROR_INVALID_LENGTH;
	
	if (length < 0 || length > sizeof(buffer)) {
		PERROR("Invalid string length (%d)\n", length);
		return ERROR_INVALID_LENGTH;
	}

	ret = read(fd, buffer, length);
	if (ret != length) {
		PERROR("Unable to read enough bytes(%d), just got %d\n", 
			length, ret);
		return ERROR_NOT_ENOUGH;
	}

	return (strncmp(buffer, search_string, length))?0:1;
}


/*
 * Reads and prints out a string.
 * Returns 0 on success
 */
int get_name(int fd, int show, int length)
{
	char buffer[4096];
	int ret;

	if (0 == length)
		return NO_ERROR;
	
	if (length < 0 || length > sizeof(buffer)) {
		PERROR("Invalid string length (%d)\n", length);
		return ERROR_INVALID_LENGTH;
	}

	ret = read(fd, buffer, length);
	if (ret != length) {
		PERROR("Unable to read enough bytes(%d), just got %d\n", 
			length, ret);
		return ERROR_NOT_ENOUGH;
	}

	buffer[ret-1] = '\0';

	if (show)
		printf("%s", buffer);

	return NO_ERROR;
}


/*
 * Reads and prints out a string containing the dependency.
 * Returns 0 on success
 */
int get_dependency(int fd, int show, int length)
{
	char buffer[4096];
	int ret;

	if (0 == length)
		return NO_ERROR;
	
	if (length < 0 || length > sizeof(buffer)) {
		PERROR("Invalid string length (%d)\n", length);
		return ERROR_INVALID_LENGTH;
	}

	ret = read(fd, buffer, length);
	if (ret != length) {
		PERROR("Unable to read enough bytes(%d), just got %d\n", 
			length, ret);
		return ERROR_NOT_ENOUGH;
	}
	/* nobody wants to see this! */
	if (show)
		printf(" (Depends on %s)", buffer);

	return NO_ERROR;
}


/*
 * Reads and prints out an integer value
 * Returns 0 on success
 */
int get_int(int fd, int show)
{
	int i;
	int ret;
	
	ret = read(fd, &i, sizeof(i));
	if (ret != sizeof(i)) {
		PERROR("Unable to read enough bytes(%d), just got %d\n", 
			sizeof(i), ret);
		return ERROR_NOT_ENOUGH;
	}
	
	if (show)
		printf("%d (0x%x)", i,i);
	
	return NO_ERROR;
}


/*
 * Reads and prints out a boolean value.
 * Returns 0 on success
 */
int get_bool(int fd, int show)
{
	int i;
	int ret;
	
	ret = read(fd, &i, sizeof(i));
	if (ret != sizeof(i)) {
		PERROR("Unable to read enough bytes(%d), just got %d\n", 
			sizeof(i), ret);
		return ERROR_NOT_ENOUGH;
	}
	
	if (show)
		printf("%s", (i)?"True":"False");
	
	return NO_ERROR;
}


/* 
 * Reads and prints out a script.
 * Return 0 on success
 */
int get_script(int fd, int show, int size)
{
	int ret;
	char script[MAX_SIZE];

	ret = read(fd, script, size);
	if (ret != size) {
		PERROR(	"Unable to read enough bytes(%d), "
			"just got %d\n", size, ret);
		return ERROR_NOT_ENOUGH;
	}

	script[size-1] = '\0';

	if (show)
		printf("%s", script);

	return NO_ERROR;
}


/* 
 * Reads and prints out an IP address
 * Returns 0 on success
 */
int get_ip(int fd, int show)
{
	int ret;
	unsigned char ip[4];

	ret = read(fd, ip, sizeof(ip));
	if (ret != sizeof(ip)) {
		PERROR("Unable to read enough bytes(%d), just got %d\n",
			sizeof(ip), ret);
		return ERROR_NOT_ENOUGH;
	}

	if (show)
		printf("%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);

	return NO_ERROR;
}


/*
 * Reads and printfs out a MAC address.
 * Returns 0 on success
 */
int get_mac(int fd, int show)
{
	int ret;
	unsigned char mac[8];
	/* I know, that a mac address is only 6 byte big, but redboot uses
	 * 8 bytes to store this data type
	 */

	ret = read(fd, mac, sizeof(mac));
	if (ret != sizeof(mac)) {
		PERROR("Unable to read enough bytes(%d), just got %d\n", 
			sizeof(mac), ret);
		return ERROR_NOT_ENOUGH;
	}

	if (show)
		printf("%x:%x:%x:%x:%x:%x", 	mac[0], mac[1], mac[2], 
						mac[3], mac[4], mac[5]);

	return NO_ERROR;
}


/*
 * Reads a value, according to the type.
 * Returns 0 on success
 */
int read_value(int fd, int show, int type)
{
	int ret;

	switch(type) {
		case CONFIG_BOOL:
			ret = get_bool(fd, show);
			break;
		case CONFIG_INT:
			ret = get_int(fd, show);
			break;
		case CONFIG_NETPORT:
		case CONFIG_STRING:
			ret = get_script(fd, show, STRING_SIZE);
			break;
		case CONFIG_SCRIPT:
			ret = get_script(fd, show, SCRIPT_SIZE);
			break;
		case CONFIG_IP:
			ret = get_ip(fd, show);
			break;
		case CONFIG_ESA:
			ret = get_mac(fd, show);
			break;
		case CONFIG_EMPTY:
			PERROR("Type empty not allowed here\n");
			ret = ERROR_INVALID_TYPE;
			break;
		default:
			ret = ERROR_UNKNOWN_TYPE;
			PERROR("Unsupported type %d\n", type);
	}
	return ret;
}


/* 
 * Reads the given file and prints the content :)
 * Returns 0 on success.
 */
int read_redboot_cfg(int fd, char* search_string)
{
	header_t header;
	int ret;
	int show = 1;
	int file_length;
	int magic;
	int crc32;


	/* skip first 8 characters */
	read(fd, &file_length, sizeof(file_length));
	read(fd, &magic, sizeof(magic));

	if (0xbadface != magic)
		fprintf(stderr, "WARNING: magic key is not valid\n");

	/* use temp file */
	fd = use_tmp_file(fd, file_length, magic);
	if (fd == -1)
		return -1;


	while(1) {

		/* read header information */
		ret = read_header(fd, &header);
		if (ret)
			return (ERROR_EOF==ret)?0:ret;
			
		/* If type is empty, read next byte until we read something 
		 * non zero 
		 */
		while(CONFIG_EMPTY==header.type) {
			header.type = header.name_length;
			header.name_length = header.sense;
			header.sense = header.dependency_length;

			ret = read(fd,  &header.dependency_length, 
					sizeof(header.dependency_length));

			if (ret == sizeof(header.dependency_length))
				continue;

			PERROR("Unable to read enough bytes(%d), just got %d\n", 
				sizeof(header.dependency_length), ret);
			return ERROR_NOT_ENOUGH;
		}

		if ( header.type == 0xde && header.name_length == 0xad &&
		     header.sense== 0xde && header.dependency_length == 0xad ) {
			/* CRC32 */
			read(fd, &crc32, sizeof(crc32));
			return 0;	/* exit here, end of file */
		}
		
		if (header.type > 7) {
			PERROR("Unsupported type %d\n", header.type);
			return ERROR_UNKNOWN_TYPE;
		}

		/* print name of entry */
		if (NULL == search_string) {
			ret = get_name(fd, show, header.name_length);
			if (ret)
				return ret;
			if (show)
				printf(": ");
		} else {
			show = is_string(fd, header.name_length, search_string);
		}

		/* print dependency */
		ret = get_dependency(fd, /*show*/ 0, header.dependency_length);
		if (ret)
			return ret;

		/* print value */
		ret = read_value(fd, show, header.type);
		if (ret)
			return ret;

		if (show)
			printf("\n");
	}
	
}

void print_help()
{
	printf(	"read_redconfig, Reads redboot configuration from mtd "
		"partition.\n"
		"Copyright (C) 2005 Henry Margies <h.margies@maxina.de>\n"
		"This program is free software and comes with ABSOLUTELY NO\n"
		"WARRANTY.\n"
		"Syntax: read_redconfig [name]\n"
		"\n"
		"Examples:\n"
		"	read_redconfig\n"
		"	read_redconfig npe_eth0_esa\n"
		"\n");

}

int main(int nargs, char** argv)
{
	int ret;
	int  fd = 0;
	
	if (nargs > 2) {
		print_help();
		return 1;
	}

	fd = open_mtd_by_name("RedBoot config", O_RDONLY);
	if (-1 == fd) {
		perror("Unable to open device");
		return 10;
	}
	ret = read_redboot_cfg(fd, argv[1]);
	close(fd);

	return ret;
}
