/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies                                   *
 *   henry.margies@gmx.de                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "redconfig.h"

#define PERROR(fmt, args...) 	fprintf(stderr,"[%s:%d] Error: " fmt,	\
							__FILE__,	\
							__LINE__, 	\
							##args)

/**
 * Returns a valid file descriptor to a mtd partition.
 * The name used to open a mtd partition can be found
 * in /proc/mtd.
 * \param name specifies the name of the partition.
 * \param flags specifies the flags used in open.
 * \return non zero on success.
 */
int open_mtd_by_name(const char* name, int flags)
{
	int ret, nr, size, erase_size;
	char string[1024];
	FILE* mtd;
	int fd = -1;

	mtd = fopen("/proc/mtd", "r");
	if (!mtd)
		return -1;
	
	/* skip first line */
	fgets(string, sizeof(string), mtd);
	
	while(1) {
		
		ret = fscanf(mtd, "mtd%d: %x %x \"%[^\"]\"\n", &nr, &size, 
			&erase_size, string);

		if (ret != 4)
			break;
		
		if (strcmp(string, name))
			continue;


		snprintf(string, sizeof(string), "/dev/mtd/%d", nr);
		fd = open(string, flags);
		if (fd == -1) {
			snprintf(string, sizeof(string), "/dev/mtd%d", nr);
			fd = open(string, flags);
		}
		fclose(mtd);
		return fd;
	}

	fclose(mtd);
	fprintf(stderr, "No such partition: %s\n", name);
	return -1;
}


/**
 * Reads the whole file and inserts it to a temporary one.
 * This is a workarround...
 */
int use_tmp_file(int fd, int file_length, int magic)
{
	int ret;
	int bytes_to_read;
	unsigned char* buf;

	/* read old file data */
	buf = malloc(file_length);
	if (!buf) {
		fprintf(stderr, "Could not allocate %d bytes of memory\n", 
				file_length);
		return -1;
	}
	
	memcpy(buf, &file_length, sizeof(file_length));
	buf+=sizeof(file_length);
	
	memcpy(buf, &magic, sizeof(file_length));
	buf+=sizeof(magic);
	
	bytes_to_read = file_length - sizeof(file_length) - sizeof(magic);
	
	ret = read(fd, buf, bytes_to_read);
	if (ret != bytes_to_read) {
		fprintf(stderr, "Could not read %d bytes, read %d instead\n",
				bytes_to_read, ret);
		return -1;
	}

	close(fd);

	/* write new filedata */
	buf -= sizeof(file_length) + sizeof(magic);

	fd = open("/tmp/temp_mtd_file", O_RDWR | O_CREAT);
	unlink("/tmp/temp_mtd_file");

	ret = write(fd, buf, file_length);
	if (ret != file_length) {
		fprintf(stderr, "Could not write %d bytes (%d)\n", 
				file_length, ret);
		close(fd);
		return -1;
	}


	/* rewind */
	lseek(fd, sizeof(file_length) + sizeof(magic), SEEK_SET);
	
	return fd;
}


/**
 * Reads a 32 bit integer value.
 * \param fd is the file desciptor to read from.
 * \param i is the address the integer will be stored to.
 * \return zero on success.
 */
int mtd_get_int(int fd, int* i)
{
	int r;

	r = read(fd, i, sizeof(*i));
	if (r != sizeof(*i)) {
		fprintf(stderr, "Error reading integer value\n");
		return -1;
	}

	return 0;
}


/**
 * Reads a 32 bit boolean value.
 * \param fd is the file desciptor to read from.
 * \param b is the address where the boolean will be stored to.
 * \return zero on success.
 */
int mtd_get_bool(int fd, int* b)
{
	return mtd_get_int(fd, b);
}


/**
 * Reads a string with a given size.
 * \param fd is the file desciptor to read from.
 * \param dest points to free memory to store the string.
 * \param size is the number of bytes to read.
 * \return zero on success.
 */
int mtd_get_string(int fd, char* dest, int size)
{
	int  ret;

	ret = read(fd, dest, size);
	if (ret != size) {
		fprintf(stderr, "Could not read script\n");
		return -1;
	}
	dest[ret-1] = 0;
	return 0;
}


/**
 * Reads an ip address.
 * \param fd is the file desciptor to read from.
 * \param dest points to free memory to store the ip.
 * \return zero on success.
 */
int mtd_get_ip(int fd, unsigned char* ip)
{
	int ret;

	ret = read(fd, ip, 4);
	if (ret != 4) {
		fprintf(stderr, "Could not read ip address\n");
		return -1;
	}
	return 0;
}


/**
 * Reads a mac address.
 * \param fd is the file desciptor to read from.
 * \param dest points to free memory to store the mac.
 * \return zero on success.
 */
int mtd_get_mac(int fd, unsigned char* mac)
{
	int ret;

	ret = read(fd, mac, 8);
	if (ret != 8) {
		fprintf(stderr, "Could not read mac address\n");
		return -1;
	}
	return 0;
}



