/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Henry Margies <h.margies@maxina.de> - Maxina GmbH, Neu-       *
*   Isenburg, Germany                                                     *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
**************************************************************************/



#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>


pid_t client_pid;


int syntax()
{
	fprintf(stderr, "start - Starts a service into background\n"
			"Written by Henry Margies <h.margies@maxina.de>\n\n"
			"-b program name to start\n"
			"-p pid-file\n"
			"-r restart program if it exits\n"
			"-v verbose output\n"
			"-h this text\n\n");
	return -1;
}


int daemonize(char* pid_file_name)
{
	FILE* f;
	pid_t pid = fork();

	switch(pid) {
		case 0: /* child */
			setsid();
			return 0;
		case -1:
			perror("Error while forking: ");
			return -1;
	}

	if (!pid_file_name)
		exit(0);

	f = fopen(pid_file_name, "w");
	if (!f) {
		perror("Could not open pid file: ");
		return -1;
	}

	fprintf(f, "%d", pid);
	fclose(f);
	exit(0);
}


void run(char* prog_name, char** prog_args, int restart)
{
	int status;
	int counter = 0;
	int tries   = 0;
	time_t tm   = time(NULL);

	openlog("start", 0, LOG_DAEMON);
	
	do {
		if (tries++ )
			syslog(LOG_NOTICE, "Starting %s (%d. time)",
					prog_name, tries);

		client_pid = fork();
		switch(client_pid) {
			case 0: /* child */
				status = execvp(prog_name, prog_args);
			case -1:
				syslog(	LOG_ERR, 
					"Error while starting program: %s",
					strerror(errno));
				return;
		}

		wait(&status);
		syslog(LOG_NOTICE, "%s exited with return code %d",
				prog_name, WEXITSTATUS(status));

		if (time(NULL)-tm > 60) {
			tm = time(NULL);
			counter = 0;
			continue;
		}
		if (counter++ < 30) 
			continue;

		syslog(LOG_ERR, "Process exited too often in last minute");
		syslog(LOG_ERR, "Exiting...");
		break;
		
	} while(restart);

	closelog();
}


void killed(int signal)
{
	syslog(LOG_NOTICE, "Killing client ...\n");
	kill(client_pid, SIGTERM);
	syslog(LOG_NOTICE, "Exiting...");
	exit(0);
}

void reload(int signal)
{
	syslog(LOG_NOTICE, "Sending SIGHUP to client ...\n");
	kill(client_pid, SIGHUP);
	syslog(LOG_NOTICE, "Done...");
	//exit(0);
}

int main(int n_args, char** v_args)
{
	/* configuration */
	char* pid_file_name = NULL;
	char* program_name  = NULL;
	int   restart	    = 0;
	int   verbose	    = 0;


	extern char *optarg;
	extern int optind;

	char* params = "rvp:b:";
	int c, ret;


	do {
		c = getopt(n_args, v_args, params);
		switch(c) {
			case 'p':
				pid_file_name = optarg;
				break;
			case 'r':
				restart = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'b':
				program_name = optarg;
				break;
			case -1:
				break;
			default:
				printf("default(%d)\n", c);
				break;
		}
	} while(c>0);
	v_args+=optind-1;
	*v_args = program_name;

	if (!program_name) {
		fprintf(stderr, "You must specify a program to start\n");
		return syntax();
	}

	if (verbose) {
		char** v = v_args;
		printf("prog name = %s\n", program_name);
		printf("pid file  = %s\n", pid_file_name);
		printf("prog args = ");

		while(*v)
			printf("%s ", *v++);
		printf("\n");
	}

	if (daemonize(pid_file_name))
		return -1;

	signal(SIGTERM, killed);
	signal(SIGHUP, reload);

	run(program_name, v_args, restart);

	return 0;
}
