#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "hbx_dsp.h"

/* imports for getopt */
extern char *optarg;
extern int optind, opterr, optopt;

int verbose = 0;

//efine GRANULARITY 320

/**
 * Converts a string to the Codecnumber
 */
int string_to_codecnum(char *str)
{
	if(!strncmp(str,"slin", 5))
	{
		return CODEC_LIN;
	}
	if(!strncmp(str,"gsm", 5))
	{
		return CODEC_GSM;
	}
	if(!strncmp(str,"echo_alaw", 11 ))
	{
		return ECHOC_FMT_ALAW;
	}
/*	if(!strncmp(str,"g.726", 5))
	{
		return CODEC_G726;
	}
	if(!strncmp(str,"g.729", 5))
	{
		return CODEC_G729;
	}
	if(!strncmp(str,"speex", 6))
	{
		return CODEC_SPEEX;
	}
	if(!strncmp(str,"ilbc", 5))
	{
		return CODEC_ILBC;
	}
*/
}

/**
 * print the help of the programm
 */
void print_help(void)
{
	printf( "\ndsp_test\n"
		"options:\n"
		"\t-h	print this help\n"
		"\t-s <filename>	source file\n"
		"\t-d <filename>	destination file\n"
		"\t-t	perform loopbacktest\n"
		"\t-i	perform loopbacktest and print information about fw\n"
		"\t-b	adjust blocksize (in Bytes)\n"
		"\t-c	number of blocks to transmit (only vaild with -t or -i)\n"
		"\t-f <format>	file format of source file\n"
		"\t-z <format>	file format of destination file\n"
		"\t-r <integer>	number of samples per block\n"
		"\t-q <integer>	number of blocks to write to the queue before reading\n"
		"\t <format> can be slin, gsm, echo_alaw [ g.726, g.729, speex, currently not supported ]\n\n" );

}

int get_numeric_file_content (char *filename)
{
	const int READSIZE = 30;

	/* Open the file readonly */
	FILE *fhand =  fopen(filename,"r");
	char buffer[READSIZE];

	memset(buffer, 0, READSIZE);

	if(!fhand) 
	{
		fprintf(stderr, "Error opening file: %s\n", filename);
		return -1;
	}
	
	/* read the file */
	fread(buffer, READSIZE, 1, fhand);
	
	/* close the file */
	fclose(fhand);

	/* return the value read */
	return atoi(buffer);
}

/**
 * Print the version of the Codec-FW and the DSP-FW 
 */
void print_version(void)
{
	char fw_version_file[] = "/sys/devices/platform/hbx_dsp.0/fw_version";
	char dsp_version_file[] =  "/sys/devices/platform/hbx_dsp.0/dsp_version";
	char initialized_file[] = "/sys/devices/platform/hbx_dsp.0/initialized";

	int fw_version = get_numeric_file_content(fw_version_file);
	int dsp_version = get_numeric_file_content(dsp_version_file);
	int initialized = get_numeric_file_content(initialized_file);

	if (initialized == -1) {
		printf("DSP not present\n");
		return;
	}

	printf(	"DSP Version: %i.%i\n"
		"DSP Firmware Version: %i.%i\n"
		"DSP is %s\n",
		(dsp_version>>8) & 0xff,
		dsp_version & 0xff,
		(fw_version>>8) & 0xff,
		fw_version & 0xff,
		(initialized)? "initialized":"not inizialized");

}

/**
 * Do the selftest
 * \params blocksize Size of a Block in Bytes
 * \params count number of blocks to send and recieve
 * \returns 0 uppon success, nonzero otherwise
 */
int action_selftest(int blocksize, int count)
{
	unsigned char data_rx[blocksize];
	unsigned char data_tx[blocksize];
	int i,j;
	int ret;

	/**
	 * Open the dsp-device
	 */
	int fd = open("/dev/dsp", O_RDWR);
	if (-1 == fd) {
		perror("Could not open file: /dev/dsp");
		return errno;
	}

	/**
	 * Configure the Codecchannel
	 */
	ret = ioctl(fd, HBX_DSP_SET_CODEC,SRC_CODEC(CODEC_LIN) | DST_CODEC(CODEC_LIN) );
	if (ret) {
		perror("Could not set codec: ");
		return ret;
	}

	ret = ioctl(fd, HBX_DSP_SET_SAMPLES, blocksize/2);
	if (ret) {
		perror("Could not set samples: ");
		return ret;
	}

	for(i=0; i<count;i++)
	{
		/**
		 * fill the tx-buffer with random data
		 */
		for(j=0;j<blocksize;j++)
		{
			data_tx[j] = (j*1345+i*3198) % 271;
		}
	
		/**
		 * write buffer to the dsp
		 */
		ret = write(fd, data_tx, blocksize);
		if (ret != blocksize)
			fprintf(stderr, "error: write ret = %d\n", ret);
	
		/**
		 * read the result from the DSP
		 */
		ret = read(fd, data_rx, blocksize);
		if (ret != sizeof(data_rx))
			fprintf(stderr, "error: read ret = %d\n", ret);

		/**
		 * compare 
		 */
		ret = memcmp(data_tx, data_rx, sizeof(data_rx));
		if (ret)
		{
			fprintf(stderr, "error: data diffs\nsrc:");
			for(j=0;j<blocksize;j++)
			{
				fprintf(stderr, "%02x ", data_tx[j]);
			}
			fprintf(stderr, "\ndst:");
			for(j=0;j<blocksize;j++)
			{
				fprintf(stderr, "%02x ", data_rx[j]);
			}
			fprintf(stderr, "\n");
			return 2; /* There was a failure, return nonzero */
		}

	}

	close(fd);
	return 0;	/* ALL frames verifyed ok */

}

/**
 * performs a coding action. it takes a file as source, processes it using the DSP
 * and stores the result to a file
 * \params from the source file descriptor
 * \params to the destination file descriptor
 * \params source_codec - Source Codec
 * \params dest_codec - destination codec
 * \params blocksize -  the size of the block to send
 * \params spb - samples per block
 * \params queuesize - Number of Bytes to write befor read
 * \returns 0 uppon success, nonzero otherwise
 */
int action_coding(FILE *from, FILE *to, int source_codec, int dest_codec, int blocksize, int spb, int queuesize)
{
	int MAXSIZE = 320;
	unsigned char data_rx[MAXSIZE];
	unsigned char data_tx[blocksize];
	int k =0, i;
	int ret;


	/**
	 * Open the dsp-device
	 */
	int fd = open("/dev/dsp", O_RDWR);
	if (-1 == fd) {
		perror("Could not open file: /dev/dsp");
		return errno;
	}

	/**
	 * Configure the Codecchannel
	 */
	ret = ioctl(fd, HBX_DSP_SET_CODEC,SRC_CODEC(source_codec) | DST_CODEC(dest_codec) );
	if (ret) {
		perror("Could not set codec: ");
		return ret;
	}

	ret = ioctl(fd, HBX_DSP_SET_SAMPLES, spb);
	if (ret) {
		perror("Could not set samples: ");
		return ret;
	}

	/**
	 * Fill the Queue first
	 */
	for (i=0; i< queuesize; i++)
	{
		/**
		 * read data from <from>
		 */
		ret = fread(data_tx, blocksize, 1, from);
	
		if (ret < 0)
		{
			perror("error read from file: ");
			break;
		}
		else if(ret == 0)
		{
			perror("end of file:");
			break;
		}
		else if(ret != 1)
		{
			fprintf(stderr, "size read does not match (%d)\n", ret);
		}

		/**
		 * write buffer to the dsp
		 */
		ret = write(fd, data_tx, blocksize);
		if (ret != blocksize)
			fprintf(stderr, "error: write ret = %d\n", ret);
	
	}

	while(1)
	{
		/**
		 * read the result from the DSP
		 */
		ret = read(fd, data_rx, MAXSIZE);
		if (ret < 0 )
			fprintf(stderr, "error: read ret = %d\n", ret);
		else 
		{
			k +=ret;
			fprintf(stderr, "Bytes recieved: %i\r", k);
		}

		/**
		 * write the result to <to>
		 */
		ret = fwrite( data_rx, ret, 1, to);

		/**
		 * read from <from>
		 */

		ret = fread(data_tx, blocksize, 1, from);
	
		if (ret < 0)
		{
			perror("error read from file: ");
			break;
		}
		else if(ret == 0)
		{
			perror("end of file:");
			break;
		}
		else if(ret != 1)
		{
			fprintf(stderr, "size read does not match (%d)\n", ret);
		}

		/**
		 * write buffer to the dsp
		 */
		ret = write(fd, data_tx, blocksize);
		if (ret != blocksize)
			fprintf(stderr, "error: write ret = %d\n", ret);

	}

	/**
	 * Empty the Queue
	 */
	for (i=0; i< (queuesize-1); i++)
	{
		/**
		 * read the result from the DSP
		 */
		ret = read(fd, data_rx, MAXSIZE);
		if (ret < 0 )
			fprintf(stderr, "error: read ret = %d\n", ret);
		else 
		{
			k +=ret;
			fprintf(stderr, "Bytes recieved: %i\r", k);
		}
	}

	close(fd);


}

int main(int argc, char *argv[], char *argi[])
{
	/* The allowed parameters */
	char params[] = "htis:d:b:c:f:z:r:q:";
	int i;
	
	/* File handler for the source of data an for the sink */
	FILE * from, *to;


	/* Status wich action to take */
	#define ACTION_SELFTEST 1
	#define ACTION_CODING 2
	#define ACTION_INFO 3
	int perform_action = ACTION_CODING;
	from = stdin;
	to = stdout;
	int blocksize	= 320;
	int spb	= 160;
	int blockcount	= 4;
	int queuesize = 1;
	int source_codec = CODEC_LIN;
	int dest_codec = CODEC_LIN;

	/**
	 * parse the arguments
	 */
	while (1)
	{
		i = getopt(argc, argv, params);
		if (i == -1)
			break; /* Thats it... we ar outahere */

		switch (i)
		{
			case 'h':
				/* Print help */
				print_help();
				return 0;
			case 'i':
				/* Perform Selftest and dump version infos */
				perform_action = ACTION_INFO;
				break;
			case 't':
				/* Perform Selftest */
				perform_action = ACTION_SELFTEST;
				break;
			case 's':
				/* Open File as source, read only */
				from = fopen(optarg, "r"); 
				if (!from)
				{
					printf("unable to open source file\n");
					return 1;
				}
				break;
			case 'd':
				/* Open File as destination, write only */
				to = fopen(optarg,"w");
				if(!to)
				{
					printf("unable to open destination file\n");
					return 2;
				}
				break;
			case 'b':
				/* Blocksize in Bytes */
				blocksize = atoi(optarg);
				break;
			case 'c':
				/* Number of Blocks to send and transmit */
				blockcount = atoi(optarg);
				break;
			case 'r':
				/* Samples per Block */
				spb = atoi(optarg);
				break;
			case 'v':
				verbose++;
				break;
			case 'f':	
				/* Source Codec */
				source_codec = string_to_codecnum(optarg);
				break;
			case 'z':
				/* Destination Codec */
				dest_codec = string_to_codecnum(optarg);
				break;
			case 'q':
				/* Queuesize */
				queuesize = atoi(optarg);
				break;

		}
	}


	switch(perform_action)
	{
		case ACTION_SELFTEST:
			return action_selftest(blocksize, blockcount);
			break;	
		case ACTION_CODING:
			return action_coding(from, to, source_codec, dest_codec, blocksize, spb, queuesize);
			break;
		case ACTION_INFO:
			print_version();
			break;
		default:
			perror("Unkown Command, Aborting\n");
			return 3;
	}

	return 0;
}
