/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *   			   Sven Kolberg  <s.kolberg@maxina.de>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/version.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/firmware.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/delay.h>
#include <linux/crc32.h>

#include <asm/uaccess.h>

#include "hbx_dsp.h"
#include "hbx_dsp_com.h"
#include "ixp425_hss.h"

	
/* Use macro PDEBUG for printing debug messages */
#undef PDEBUG
#ifdef HBX_DSP_DEBUG
	#define PDEBUG(fmt, args...) 	\
		printk("[%s:%d] "fmt, __FILE__,__LINE__, ##args)
#else
	#define PDEBUG(fmt, args...) {}
#endif

/* number of ring buffers in a rwbuf */
#define HBX_DSP_RBS 	8


/* size in bytes of the data buffer */
#define HBX_DSP_BSZ	320

/* number of logical channels */
#define HBX_DSP_LCH	10

/* number of errors received till reset */
#define HBX_DSP_MAX_ERR	10000

/* number of frames, between requesting heartbeat */
#define HBX_DSP_WD_REQ	16000

/* number of frames whithout receiving heartbeat needed to reset DSP */
#define HBX_DSP_WD_RES	32000

/**
 * Data Buffer (Frame buffer)
 * The data buffer contains exactly one voice frame. The voice data is stored
 * in data_buf and informational data is stored in info_buf. info_buf is used
 * to show when a frame starts (FRAME_START), how many bytes (length) and how 
 * many samples (samples) it contains.
 *
 * These data will later be copied into and from the HSS bus as follows:
 *
 * +-------------+---------------+---------------+
 * | Logical channel for one voice channel       |
 * | HSS TS n      | HSS TS n+1    | HSS TS n+2  |
 * +---------------+---------------+-------------+
 * | Voice Byte  0 | Voice Byte  1 | LCT_START   |
 * | Voice Byte  2 | Voice Byte  3 | Length[0]   |
 * | Voice Byte  4 | Voice Byte  5 | Length[1]   |
 * | Voice Byte  6 | Voice Byte  7 | Samples[0]  |
 * | Voice Byte  8 | Voice Byte  9 | Samples[1]  |
 * | Voice Byte 10 | Voice Byte 11 | LCT_CONTINUE|
 * | Voice Byte 12 | Voice Byte 13 | LCT_CONTINUE|
 * | Voice Byte 14 | Voice Byte 15 | LCT_CONTINUE|
 * | Voice Byte 16 | Voice Byte 17 | LCT_CONTINUE|
 * | Voice Byte 18 | Voice Byte 19 | LCT_CONTINUE|
 *             ...                               
 * | Voice Byte 98 | Voice Byte 99 | LCT_CONTINUE|
 * | anythig       | anything      | LCT_IDLE    |
 * +-------------+---------------+---------------+
 *
 * If there is another frame ready to send FRAME_START can be used after the 
 * last FRAME_CONT.
 *
 */
struct dbuf {
	unsigned short	data_buf[HBX_DSP_BSZ/2];
	unsigned char	info_buf[HBX_DSP_BSZ/2];

	/* how many bytes are used in buffer */
	int		size;
	/* current position in buffer */
	unsigned int	idx;
};

/**
 * Ring Buffer
 * This ring buffer is used for writing and reading.
 * For writing data to the HSS bus it is like:
 *
 *                      wbuf ---> +---------+
 *                                | dbuf[0] |
 *                                +---------+
 *                                | dbuf[1] | ---> read_idx ---> write_hss
 *                                +---------+         |        (hbx_dsp_send)
 * user_write ---> write_idx ---> | dbuf[2] |         |
 * (hbx_dsp_write)     |          +---------+        \|/
 *                     |          | dbuf[3] |
 *                    \|/         +---------+
 *                                | dbuf[4] |
 *                                +---------+
 *                                | dbuf[5] |
 *                                +---------+
 *
 * For reading data from HSS bus it is like:
 *
 *                      rbuf ---> +---------+
 *                                | dbuf[0] |
 *                                +---------+
 *                                | dbuf[1] |
 *                                +---------+
 *                                | dbuf[2] | ---> read_idx ---> user_read
 *                                +---------+         |        (hbx_dsp_read)
 * hss_read   ---> write_idx ---> | dbuf[3] |         |
 * (hbx_dsp_recv)      |          +---------+        \|/
 *                     |          | dbuf[4] |
 *                    \|/         +---------+
 *                                | dbuf[5] |
 *                                +---------+
 */
struct rwbuf {
	struct dbuf	dbuf[HBX_DSP_RBS];

	/* shows which buffer is currently used for reading */
	unsigned int	read_idx;

	/* shows which buffer is currently used to write to */
	unsigned int	write_idx;

};

/**
 * configuration for DSP */
struct hbx_dsp_config {
        
	/* state machine for fw upload */
	unsigned int state;
	unsigned int t101;

	/* history for all states */
	unsigned char  history[PAGE_SIZE];
	unsigned char* hist_start;
	unsigned char* hist_stop;

	/* character device */
	struct device*	dev;
	struct cdev 	cdev;
	dev_t		first_dev;

	/* firmware */
	const struct firmware*	fw_entry;
	unsigned char* 		fw_pointer;
	unsigned int 		fw_ready;
	unsigned int 		fw_size;
	unsigned int 		fw_crc;

	/* dsp status */
	unsigned int initialized;
	unsigned int failed;
	unsigned int dsp_version;
	unsigned int fw_version;
	unsigned int rtt;
	unsigned int pc;

	/* locks for command queue */
	struct semaphore rq_sem;

	/* command queue */
	unsigned char rq_sts;
	unsigned char rq_cmd;
	unsigned char rq_arg;

	/* usage of channels */
	unsigned char ch_used[HBX_DSP_LCH];

	/* buffer */
	struct rwbuf rbuf[HBX_DSP_LCH];
	struct rwbuf wbuf[HBX_DSP_LCH];

	/* to do things later */
	struct workqueue_struct* work_q;
	struct work_struct	 work_req_fw;
	struct work_struct	 work_rel_fw;
	struct work_struct	 work_rst_dsp;
	struct work_struct	 work_stp_hss;

	/* verbose */
	unsigned int tx,rx,err, cmd_ack_err, c1, c2;

	/* watchdog counter */
	unsigned int wd_counter;
};

static void hbx_dsp_release_fw(struct hbx_dsp_config *cfg);
static void hbx_dsp_request_fw(struct hbx_dsp_config* cfg);
static void hbx_dsp_reset(struct hbx_dsp_config* cfg);

/**
 * SysFS Vars (outsourced)
 */
#include "hbx_dsp_sysfs.c"

/**
 * File operations (outsources)
 */
#include "hbx_dsp_fops.c"

/**
 * Communiaction with DSP (outsourced)
 * (these function do all run in interrupt context)
 */
#include "hbx_dsp_com.c"

/**
 * Interface for other modules for echo canceller (outsourced)
 */
#include "hbx_dsp_ec.c"

/**
 * Resets the DSP
 */
static void hbx_dsp_reset(struct hbx_dsp_config* cfg)
{
	cfg->err	 = 0;
	cfg->initialized = 0;
	cfg->wd_counter  = 0;
	cfg->rq_sts	 = EMPTY;
	memset(cfg->ch_used, 0, sizeof(cfg->ch_used));

	PDEBUG("Activate DSP...\n");
	/* configure as output, then hold RST and release it later */
	gpio_line_config(1, IXP4XX_GPIO_OUT);
	gpio_line_set(1, 0);
	mdelay(20);
	gpio_line_set(1, 1);
}
			

/**
 * Requests the firmware from hotplug 
 */
static void hbx_dsp_request_fw(struct hbx_dsp_config* cfg)
{
	int ret;

	/* already loaded? */
	if (cfg->fw_ready)
		return;

	PDEBUG("Requesting firmware...\n");
	ret = request_firmware(&cfg->fw_entry, "hbx_dsp_fw", cfg->dev);
	if (ret) {
		printk(KERN_ERR "hbx_dsp: Firmware load failed\n");
		return;
	}
	cfg->fw_crc   = crc32_le(~0, cfg->fw_entry->data, cfg->fw_entry->size);
	cfg->fw_ready = 1;
}


/**
 * Release memory for fw
 */
static void hbx_dsp_release_fw(struct hbx_dsp_config *cfg)
{
	/* release firmware */
	if (cfg->fw_ready && cfg->fw_entry) {

		PDEBUG("Release firmware... \n");
		cfg->fw_ready = 0;
		release_firmware(cfg->fw_entry);
		cfg->fw_entry = NULL;
	}
}


/**
 * Stop HSS bus...
 */
static void hbx_dsp_stop_hss(struct hbx_dsp_config *cfg)
{
	PDEBUG("Stopping HSS bus ... \n");
	
	/* release HSS bus */
	ixp425_hss_unregister2();
}


/**
 * Create workqueue for work out of interrupt context
 */
static void hbx_dsp_create_work_q(struct hbx_dsp_config* cfg)
{
	cfg->work_q = create_singlethread_workqueue("hbx_dsp_wq");

	INIT_WORK(&cfg->work_req_fw,  (void(*)(void*))hbx_dsp_request_fw, cfg);
	INIT_WORK(&cfg->work_rel_fw,  (void(*)(void*))hbx_dsp_release_fw, cfg);
	INIT_WORK(&cfg->work_rst_dsp, (void(*)(void*))hbx_dsp_reset, cfg);
	INIT_WORK(&cfg->work_stp_hss, (void(*)(void*))hbx_dsp_stop_hss, cfg);
}


/**
 * Initialize device ... and stuff
 * This function will do:
 * 	- initalize config structure
 * 	- load firmware via fw API
 * 	- register and start hss bus
 * 	- register a character device
 * 	- create sysfs entries
 * 	- reset DSP
 * 	- return zero on success
 */ 
static int hbx_dsp_probe(struct platform_device * dev)
{
	struct hbx_dsp_config* cfg;
	int ret;

	PDEBUG("Probing for device...\n");

	/* some structure initialisations */
	cfg = kmalloc(sizeof(struct hbx_dsp_config), GFP_KERNEL);
	if (!cfg) {
		printk(KERN_ERR "Could not allocate memory\n");
		return -ENOMEM;
	}
	memset(cfg, 0, sizeof(struct hbx_dsp_config));
	cfg->hist_start = cfg->history;
	cfg->hist_stop  = cfg->history + sizeof(cfg->history) - 4;
	cfg->dev	= &dev->dev;
	dev->dev.driver_data = cfg;

//	/* get fw */
//	hbx_dsp_request_fw(cfg);
//	if (!cfg->fw_ready) {
//		kfree(cfg);
//		return -ENODEV;
//	}

	PDEBUG("Register on HSS bus...\n");
	ret = ixp425_hss_register2(hbx_dsp_transfer, &dev->dev);
	if (ret) {
		printk(KERN_ERR "hbx_dsp: Error registering on HSS bus\n");
		release_firmware(cfg->fw_entry);
		kfree(cfg);
		return -EIO;
	}
	
	PDEBUG("Create device node...\n");
	ret = alloc_chrdev_region(&cfg->first_dev, 0, 1, dev->name);
	if (ret) {
		printk(KERN_ERR "hbx_dsp: Error getting major number\n");
		ixp425_hss_unregister2();
		release_firmware(cfg->fw_entry);
		kfree(cfg);
		return ret;
	}

	cdev_init(&cfg->cdev, &hbx_dsp_fops);
	cfg->cdev.owner = THIS_MODULE;
	cfg->cdev.ops   = &hbx_dsp_fops;
	ret = cdev_add(&cfg->cdev, cfg->first_dev, 1);
	if (ret) {
		printk(KERN_ERR "hbx_dsp: Error adding character device\n");
		ixp425_hss_unregister2();
		release_firmware(cfg->fw_entry);
		kfree(cfg);
		return ret;
	}

	PDEBUG("Creating sysfs entries...\n");
	hbx_dsp_sysfs_load(&dev->dev);

	PDEBUG("Create work queue...\n");
	hbx_dsp_create_work_q(cfg);

	/* initialize request queue lock */
	init_MUTEX(&cfg->rq_sem);

	/* reset DSP */
	hbx_dsp_reset(cfg);

	/* register echo canceller routines */
	hbx_dsp_register_ec(cfg);

	return 0;
}


/**
 * This function is called when the device should be removed (e.g.: when 
 * unloading the module)
 * It will:
 * 	- remove sysfs entries
 * 	- unregister from hss bus
 * 	- release firmware data
 * 	- remove character device
 * 	- free memory
 */ 
static int hbx_dsp_remove(struct platform_device * dev)
{
	struct hbx_dsp_config* cfg = dev->dev.driver_data;

	PDEBUG("Removing device...\n");

	/* remove sysfs entries */
	hbx_dsp_sysfs_unload(&dev->dev);

	/* release HSS bus */
	ixp425_hss_unregister2();

	/* release firmware */
	hbx_dsp_release_fw(cfg);
	
	/* remove cdev */
	cdev_del(&cfg->cdev);
	unregister_chrdev_region(cfg->first_dev, 1);

	/* remove workqueue */
	if (cfg->work_q) {
		flush_workqueue(cfg->work_q);
		destroy_workqueue(cfg->work_q);
	}

	/* free memory */
	if (cfg)
		kfree(cfg);

	return 0;
}

/**
 * This function is called when module gets unloaded.
 * All deinitialization is done in hbx_dsp_remove.
 */ 
static void hbx_dsp_release(struct device * dev)
{
	PDEBUG("Releasing device...\n");
}


/* driver structure */
static struct platform_driver hbx_dsp_driver = {
	.probe          = hbx_dsp_probe,
	.remove         = hbx_dsp_remove,
	.driver		= {
		.name	= "hbx_dsp",
	}
};

/* device structure */
static struct platform_device hbx_dsp_device = {
	.name		= "hbx_dsp",
	.dev		= {
		.release= &hbx_dsp_release,
	}
};

/**
 * Called when insmoding the module. 
 */ 
static int hbx_dsp_init(void) 
{
	int ret;

	PDEBUG("Initialize module...\n");
	
	/* register driver */
	ret = platform_driver_register(&hbx_dsp_driver);
	if (ret) 
		return ret;

	/* register device */
	ret = platform_device_register(&hbx_dsp_device);
	if (!ret)
		return 0;

	/* registering device failed, unregister driver */
	platform_driver_unregister(&hbx_dsp_driver);
	return ret;
}

/**
 * Clean up all 
 */ 
static void hbx_dsp_exit(void) 
{
	PDEBUG("Release module...\n");
	platform_device_unregister(&hbx_dsp_device);
	platform_driver_unregister(&hbx_dsp_driver);
}


MODULE_DESCRIPTION("Driver for DSP used in HorstBox");
MODULE_AUTHOR("Henry Margies (h.margies@maxina.de");
MODULE_LICENSE("GPL");

module_init(hbx_dsp_init);
module_exit(hbx_dsp_exit);


