/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *   			   Sven Kolberg  <s.kolberg@maxina.de>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
 * NOTE: hbx_dsp_transfer is the entry function, which is called by
 * ixp425_hss driver as a callback.
 */


/**
 * This function is called when the state of the sate machine changes.
 * It also adds this to the history in
 * /sys/devices/platform/hbx_dsp.0/history 
 */
static void hbx_dsp_state_transition(struct hbx_dsp_config* cfg, int newstate)
{
	if (cfg->hist_start < cfg->hist_stop) {	
		*cfg->hist_start++ = '0'+newstate;
		*cfg->hist_start++ = ' ';
	}

	cfg->state = newstate;
}

/**
 * State machine for firmware upload.
 * This state machine is for communicating with the boot loader in 
 * hbx dsp. It will upload the firmware. After firmware is uploaded
 * the state machine is not used anymore.
 */
static unsigned char hbx_dsp_state_machine(	struct hbx_dsp_config* cfg, 
						unsigned char *rx_buf,
						unsigned char *tx_buf)
{
	unsigned char cmd;
	unsigned int length, crc;

	cmd = *rx_buf;
 
 	/* counts all frames */
	cfg->pc++;

	switch(cfg->state) {

		/* default state when not connected */
		case STATE_INIT_COM:
			if (cmd==FW_IDLE_ACK) {
				hbx_dsp_state_transition(cfg, STATE_CONNECTED);
				/* read version number */
				memcpy((unsigned char*)&length, rx_buf+1, 4);
				cfg->dsp_version = le32_to_cpu(length);
			}
			break;

		/* communication established */
		case STATE_CONNECTED:
			
			/* only FW_IDLE_ACK allowed */
			if (cmd!=FW_IDLE_ACK) {
				hbx_dsp_state_transition(cfg, STATE_INIT_COM);
				break;
			}

			/* firmware upload needed ? */
			if (!cfg->initialized && !cfg->failed) {
				
				if (!cfg->fw_ready) { 
					/*request firmware first */
					queue_work(cfg->work_q, &cfg->work_req_fw);
					return FW_IDLE_QUERY;
				}

				if (cfg->t101++ > 50) {
					cfg->failed = 1;
					PDEBUG("t101 expired...\n");
					printk("ERROR: Could not upload DSP firmware!\n");
					queue_work(cfg->work_q, &cfg->work_stp_hss);
					break;
				}

				cfg->rtt	= cfg->pc;
 				cfg->fw_pointer = cfg->fw_entry->data;
				cfg->fw_size    = cfg->fw_entry->size;
				hbx_dsp_state_transition(cfg, STATE_FW_UPLOAD);
				length = cpu_to_le32(cfg->fw_size);
				/* send length of firmware in TS 1-4 */
				memcpy(tx_buf+1, (unsigned char*)&length, 4);
				/* send FW_START cmd */
				return FW_START;
			}

			break;

		/* during fw upload */
		case STATE_FW_UPLOAD:

			switch (cmd) {
				case FW_START_ACK:
					cfg->rtt = cfg->pc - cfg->rtt;
				case FW_CONT_ACK:
				case FW_IDLE_ACK:
					if (cfg->fw_size <= 0) {
						/*no data to send */
						hbx_dsp_state_transition(cfg, STATE_FW_VERIFY);
						crc = cpu_to_le32(cfg->fw_crc);
						/* send crc of fw in TS 1-4 */
						memcpy(tx_buf+1, (unsigned char*)&crc, 4);
						/* send FW_STOP */
						return FW_STOP;
					}
					/* send fw package */
					length = (cfg->fw_size>31)?31:cfg->fw_size;
					/* send fw in TS 1-31 */
					memcpy(tx_buf+1, cfg->fw_pointer, length);
					cfg->fw_size    -= length;
					cfg->fw_pointer += length;
					return FW_CONT;
			}

			PDEBUG("Error: received 0x%x in state %d\n", cmd, cfg->state);
			hbx_dsp_state_transition(cfg, STATE_INIT_COM);
			break;

		/* waiting for verification of fw from dsp */
		case STATE_FW_VERIFY:

			/* fw is ok */
			if (cmd == FW_CRC_OK)
				hbx_dsp_state_transition(cfg, STATE_FW_COMPLETE);
			else if (cmd != FW_CONT_ACK) {
				PDEBUG("Error: received 0x%x in state %d\n", cmd, cfg->state);
				hbx_dsp_state_transition(cfg, STATE_INIT_COM);
				return FW_IDLE_QUERY;
			}
			return FW_LOAD_IDLE;

		/* fw upload complete */
		case STATE_FW_COMPLETE:

			if (cmd!=FW_LOAD_IDLE_ACK) {
				PDEBUG("Error: received 0x%x in state %d\n", cmd, cfg->state);
				hbx_dsp_state_transition(cfg, STATE_INIT_COM);
				return FW_IDLE_QUERY;
			}

			hbx_dsp_state_transition(cfg, STATE_FW_EXECUTE);
			/* send FW_EXEC */
			return FW_EXEC;

		/* waiting for ACK for fw execution */
		case STATE_FW_EXECUTE:

			if (cmd == FW_LOAD_IDLE_ACK) {
				/* send FW_LOAD_IDLE wait */
				return FW_LOAD_IDLE;
			}

			if (cmd == FW_EXEC_ACK) {
				/* release fw */
				queue_work(cfg->work_q, &cfg->work_rel_fw);
				/* DSP is now ready */
				cfg->initialized = 1;
			} else
				PDEBUG("Error: received 0x%x in state %d\n", cmd, cfg->state);
			hbx_dsp_state_transition(cfg, STATE_INIT_COM);
			break;

	}

	return FW_IDLE_QUERY;
}


/**
 * This function handles the global management channel. This channel is used 
 * to:
 * 	- get fw version 
 * 	- assign a codec to a logical channel
 * 	- unassign a codec from a logical channel
 *
 * The management channel uses two timeslots on HSS bus. The first is used to 
 * specify what to do, the second is used as argument.
 *
 * +-----+-----+-----+-----+-----+--------------------------
 * | TS0 | TS1 | TS2 | TS3 | TS4 | TS5 | TS6 | TS7 | TS8 ...
 * +-----+-----+-----+-----+-----+--------------------------
 * | cmd | arg | log. voice ch 1 | log. voice ch 2 |     ...
 *
 * There are three kind of commands:  read, write, status. The status command 
 * is also used as default or idle command. The last 4 bits of first byte are
 * used to specify the command,  the first  4 bits define the logical channel
 * to write to or to read from (0-9). The second byte define what to write.
 *
 *                                       +-----------------+-----------------+
 *                                       | cmd (TS0)       | arg (TS1)       |
 *                                (bits) | 8 7 6 5 4 3 2 1 | 8 7 6 5 4 3 2 1 |
 *                                       +-----------------+-----------------+
 * read command   (CMGMT_READ)           | 0 1 1 1| channel|                 |
 * write command  (CMGMT_WRITE)          | 1 0 0 0| channel| value           |
 * status command (CMGMT_STATUS)         | 1 0 0 1|        |                 |
 * 					 +--------+--------+-----------------+
 *
 * The answer will look like:            +-----------------+-----------------+
 *                                       | cmd (TS0)       | arg (TS1)       |
 *                                (bits) | 8 7 6 5 4 3 2 1 | 8 7 6 5 4 3 2 1 |
 *                                       +-----------------+-----------------+
 * answer for read   (CMGMT_READ_ACK)    | 0 1 1 1| channel| value read      |
 * answer for write  (CMGMT_WRITE_ACK)   | 1 0 0 0| channel| return code     |
 * answer for status (CMGMT_STATUS_ACK)  | 1 0 0 1| version| subversion      |
 * answer for an unknown command         | 1 1 1 1| channel| value received  |
 *                   (CMGMT_UNKNOWN_CMD) +--------+--------+-----------------+
 *
 */ 
static unsigned char hbx_dsp_management_ch(	struct hbx_dsp_config* cfg, 
						unsigned char *rx_buf,
						unsigned char *tx_buf)
{
	unsigned char cmd;

	cmd = *rx_buf;

	switch (cmd>>4) {

		/* received an answer for a read or write command */
		case CMGMT_READ_ACK:

			/* received heartbeat */
			if (cmd == ((CMGMT_READ_ACK<<4)|10))
				cfg->wd_counter = 0;

		case CMGMT_WRITE_ACK:

			if (cfg->rq_sts != WAIT) {
				/* Ack received and no command in queue */
				cfg->cmd_ack_err++;
				break;
			}

			if (cmd != cfg->rq_cmd) {
				cfg->cmd_ack_err++;
				/* Ack received for different command as in queue */
				break;
			}

			cfg->rq_arg = rx_buf[1];
			PDEBUG("command receivced %x %x\n", cmd, cfg->rq_arg);
			cfg->rq_sts = RECV;
			wake_up_interruptible(&cmd_wait_q);
			break;

		/* received status ack */
		case CMGMT_STATUS_ACK:
			cfg->fw_version = ((cmd & 0xf) << 8) | rx_buf[1];
			break;

		/* DSP does not know our command */
		case CMGMT_UNKNOWN_CMD:
		default:
			cfg->err++;
			if (cfg->err>HBX_DSP_MAX_ERR) {
				/* reset dsp */
				cfg->err = 0;
				cfg->initialized = 0;
				cfg->wd_counter = 0;
				queue_work(cfg->work_q, &cfg->work_rst_dsp);
			}

	}
	
	/* command to send? */
	if (cfg->rq_sts == SEND) {
		PDEBUG("sending command 0x%x 0x%x\n", cfg->rq_cmd,  cfg->rq_arg);
		cfg->rq_sts = WAIT;
		tx_buf[1] = cfg->rq_arg;
		return cfg->rq_cmd;

	}

	if (cfg->wd_counter++ == HBX_DSP_WD_REQ)
		return (CMGMT_READ<<4) | 10;
	
	
	/* default command */
	return CMGMT_STATUS<<4;
}


/**
 * Transfer data from data buffer to hss bus.
 * This function gets the data buffer and puts 2 bytes to the first timeslots
 * of the logical channel. Then it copies another byte from the info buffer
 * to the third timeslot.
 *
 * voice data                                 info data
 * +--------+                                 +--------+
 * | 1 byte | ----+                +--------- | 1 byte |
 * | 1 byte | ----|--------+       |          | 1 byte |
 * ////////       |        |       |          //////////
 *               \|/      \|/     \|/
 * -+--------+--------+--------+--------+--------+--------+
 *           | TS n   | TS n+1 | TS n+2 |
 *           | logical voice channel    |
 * -+--------+--------+--------+--------+--------+--------+
 *
 */
static int hbx_dsp_send(struct rwbuf *wbuf, unsigned char* tx)
{
	struct dbuf *dbuf;

	/* get write data buffer to read from
	 * wbuf     - the ring buffer for writing to hss bus
	 * read_idx - which data buffer is used to read from
	 */
	dbuf = &wbuf->dbuf[wbuf->read_idx];

	/* nothing to send ? */
	if (dbuf->size <= 0)
		return 0;

	/* copy voice data and informational data */
	memcpy(tx,   dbuf->data_buf+dbuf->idx, 2);
	memcpy(tx+2, dbuf->info_buf+dbuf->idx, 1);
	
	dbuf->idx  += 1;
	dbuf->size -= 2;

	/* if data buffer is empty, use next one and wake up 
	 * user write */
	if (dbuf->size <= 0) {
		wbuf->read_idx++;
		wbuf->read_idx%=HBX_DSP_RBS;
		wake_up_interruptible(&write_wait_q);
	}

	/* we wrote two bytes */
	return 2;
}


/**
 * Transfer data from hss bus to data buffer.
 * See hbx_dsp_send for details.
 */
static int hbx_dsp_recv(struct rwbuf *rbuf, unsigned char* rx)
{
	struct dbuf *dbuf;

	/* get read data buffer to write to
	 * rbuf      - ring buffer for data from hss bus
	 * write_idx - which data buffer is used to write to
	 */
	dbuf = &rbuf->dbuf[rbuf->write_idx];
	
	/* we are waiting for START frame if idx is zero */
	if (dbuf->idx == 0 && *(rx+2) != LCT_START)
		return 0;

	/* check for end of frame */
	if (dbuf->idx > 4 && *(rx+2) != LCT_CONTINUE) {
		dbuf->size = dbuf->idx*2;
		dbuf->idx  = 0;
		rbuf->write_idx++;
		rbuf->write_idx%=HBX_DSP_RBS;
		/* wake up user read */
		wake_up_interruptible(&read_wait_q);
		/* re-read this data */
		/* XXX remove recursion */
		return hbx_dsp_recv(rbuf, rx);
	}

	if (dbuf->idx >= HBX_DSP_BSZ/2) {
		dbuf->idx = 0;
		PDEBUG("Error, frame too long\n");
	}
	
	/* copy voice data */
	memcpy(dbuf->data_buf+dbuf->idx, rx, 2);

	/* copy info */
	*(dbuf->info_buf+dbuf->idx) = *(rx+2);

	dbuf->idx++;

	/* received 2 bytes */
	return 2;
}


/**
 * This functions fills all 32 time slots for all frames with data.
 * It will:
 * 	- run the state machine on TS0,TS1 if fw upload is still needed
 * 	  (hbx_dsp_state_machine)
 * 	- handle the global management channel on TS0 and TS1 after fw was
 * 	  uploaded (hbx_dsp_management_ch)
 * 	- send and receive data to and from logical channels if active
 * 	  (hbx_dsp_send, hbx_dsp_recv)
 *
 * 	  This is done for all 8 frames we receive.
 */ 
static void hbx_dsp_transfer(unsigned char* rx, unsigned char* tx, void* data)
{
	struct device* dev = (struct device*)data;
	struct hbx_dsp_config* cfg = dev->driver_data;
	int frame, channel;
	unsigned char *r, *t;

	/* clean transmit buffer */
	memset(tx, LCT_IDLE, 32*8);

	/* handle all 8 frames */
	for (frame=0, r=rx, t=tx; frame<8; frame++) {
		
		/* if not yet initialized run state machine for fw upload */
		if (!cfg->initialized) {

			*t = hbx_dsp_state_machine(cfg, r, t);
			/* skip other timeslots */
			t+=32;
			r+=32;
			continue;
		}

		/* handle global management channel (TS0,TS1) */
		*t = hbx_dsp_management_ch(cfg, r, t);
		t+=2;
		r+=2;

		/* handle all logical channels (which consist of 3 timeslots) */
		for (channel=0; channel<HBX_DSP_LCH; channel++,t+=3,r+=3) {

			/* if not active do nothing */
			if (!cfg->ch_used[channel])
				continue;

			/* receive data */
			cfg->rx += hbx_dsp_recv(&cfg->rbuf[channel], r);

			/* send data */
			cfg->tx += hbx_dsp_send(&cfg->wbuf[channel], t);
		}
	}

#ifdef HBX_DSP_WATCHODOG_ENABLE
	/* check wd_counter */
	if (cfg->wd_counter > HBX_DSP_WD_RES) {
		PDEBUG("Resetting DSP (no heartbeat received)...\n");
		cfg->err = 0;
		cfg->initialized = 0;
		cfg->wd_counter = 0;
		queue_work(cfg->work_q, &cfg->work_rst_dsp);
	}
#endif

}


