/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *   			   Sven Kolberg  <s.kolberg@maxina.de>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * Wait queues for read, write operations and the command queue
 */
static DECLARE_WAIT_QUEUE_HEAD(read_wait_q);
static DECLARE_WAIT_QUEUE_HEAD(write_wait_q);
static DECLARE_WAIT_QUEUE_HEAD(cmd_wait_q);

/** 
 * Command queue
 * The command queue is used to send commands to the DSP via hss bus. The
 * queue can only hold a single command and uses the following variables:
 *
 * cfg->lock
 * 	Whenever a command should be send, this lock must be used to make
 * 	sure nobody else tries to send a command.
 *
 * cfg->rq_sts
 * 	This is the status of the queue, it can be set to EMPTY, SEND, WAIT
 * 	or RECV.
 *
 * cfg->rq_cmd
 * 	This value holds  the command which is later  copied to timeslot 0
 * 	on the hss bus. It contains the kind of command and the channel if
 * 	needed.
 *
 * cfg->rq_arg
 * 	This value holds the argument for write commands. It will be copied
 * 	to  timeslot  1 on the HSS bus.  For read and write operations this 
 * 	value will also hold the answer (from timeslot 1 of hss bus).
 *
 * Sending a command will work like this:
 * 
 * 1) Enable locking so nobody else will try to send commands. Now the status
 *    (rq_sts) must be EMPTY.
 * 2) Set rq_cmd and rq_arg now (leave rq_sts to EMPTY!)
 * 3) Set rq_sts to SEND. The command will be send within the next frame.
 * 4) Sleep until rq_sts is set to RECV
 * 5) Read returned value from rq_arg.
 * 6) Set rq_sts to EMPTY
 * 7) Unlock command queue
 */

/* internal states for command queue */
#define EMPTY	0x0
#define SEND	0x11
#define WAIT	0x22
#define RECV	0x33


/**
 * This function requests a codec.
 * To do so it will:
 * 	- find out which logical channel is not used already
 * 	- send a command to DSP to assign a channel for the 
 * 	  given codec
 * 	- return the logical channel or -1 in case of an error
 */
static int hbx_dsp_request_codec( struct hbx_dsp_config* cfg,
				  unsigned int codec) 
{
	unsigned char channel = -1;
	int i;

	PDEBUG("Request channel for codec %d\n", codec);

	/* activate lock */
	if (down_interruptible(&cfg->rq_sem))
		return -ERESTARTSYS;

	/* find free logical channel */
	for (i=0;i<sizeof(cfg->ch_used);i++) 
		if (!cfg->ch_used[i]) {
			channel = i;
			break;
		}

	/* request codec */
	if (cfg->rq_sts != EMPTY) {
		printk("Internal Error! Command queue not empty\n");
		up(&cfg->rq_sem);
		return -EIO;
	}

	cfg->rq_cmd = (CMGMT_WRITE<<4) | channel;
	cfg->rq_arg = CODEC_CHANNEL_ENABLE | codec;
	cfg->rq_sts = SEND;

	/* sleep until answer arrived */
	interruptible_sleep_on_timeout(&cmd_wait_q, 100);

	/* check if command was sent */
	if (cfg->rq_sts != RECV) {
		printk("Error: Request timed out!\n");
		cfg->rq_sts = EMPTY;
		up(&cfg->rq_sem);
		return -EIO;
	}

	/* check return code from dsp */
	if (cfg->rq_arg != 1) {
		/* error reported from dsp */
		printk("Error: Command failed (%d)\n", cfg->rq_arg);
		cfg->rq_sts = EMPTY;
		up(&cfg->rq_sem);
		return -EIO;
	}

	/* make queue empty */
	cfg->rq_sts = EMPTY;

	/* mark logical channel as used */
	cfg->ch_used[channel] = 1;

	/* release lock */
	up(&cfg->rq_sem);

	/* empty buffer */
	memset(&cfg->rbuf[channel], 0, sizeof(cfg->rbuf[channel]));
	memset(&cfg->wbuf[channel], 0, sizeof(cfg->wbuf[channel]));

	return channel;
}

/**
 * This function releases a channel from a codec.
 * For this purpose it will send a command to the DSP
 * and mark the channel as unused on success.
 */
static int hbx_dsp_release_codec( struct hbx_dsp_config* cfg,
				  unsigned char channel) 
{
	PDEBUG("Release codec for channel %d\n", channel);

	/* activate lock */
	down(&cfg->rq_sem);

	/* request release */
	if (cfg->rq_sts != EMPTY) {
		printk("Internal Error! Command qeue not empty\n");
		up(&cfg->rq_sem);
		return -1;
	}

	cfg->rq_cmd = (CMGMT_WRITE<<4) | channel;
	cfg->rq_arg = 0;
	cfg->rq_sts = SEND;

	/* sleep until answer arrived */
	interruptible_sleep_on_timeout(&cmd_wait_q, 100);

	if (cfg->rq_sts != RECV) {
		printk("Error: Relase request timed out!\n");
		cfg->rq_sts = EMPTY;
		up(&cfg->rq_sem);
		return -1;
	}
	cfg->rq_sts = EMPTY;

	/* set logical channel to unused */
	cfg->ch_used[channel] = 0;

	/* release lock */
	up(&cfg->rq_sem);

	return 0;
}



/**
 * FILE interfaces
 */

/** 
 * config structure per opened file
 */
struct hbx_dsp_file_config {
	
	/* dsp config */
	struct hbx_dsp_config* config;

	/* number of samples in next write */
	unsigned short	samples;

	/* logical channel used */
	int		channel;

};

/**
 * Called when somebody opens our device.
 * This function will only setup the per file config structure.
 */
static int hbx_dsp_open(struct inode *inode, struct file *filp)
{
	struct hbx_dsp_file_config* 	cfg;
	struct hbx_dsp_config* 		config;

	PDEBUG("open...\n");

	/* get cfg structure and save it for read and write */
	config = container_of(inode->i_cdev,struct hbx_dsp_config,cdev);
	if (!config->initialized)
		return -ENODEV;
	
	/* get memory for cfg structure per file transfer */
	cfg = kmalloc(sizeof(struct hbx_dsp_file_config), GFP_KERNEL);
	if (!cfg) {
		printk(KERN_ERR "Could not allocate memory\n");
		return -ENOMEM;
	}
	memset(cfg, 0, sizeof(struct hbx_dsp_file_config));

	/* store config for later reads and writes */
	cfg->config  = config;
	cfg->channel = -1;
	cfg->samples =  0;
	filp->private_data = cfg;

	return 0;
}

/**
 * Called when file will be closed
 * It will release the logical channel if used and free
 * the per file config memory.
 */
static int hbx_dsp_close(struct inode *inode, struct file *filp)
{
	struct hbx_dsp_file_config* cfg = filp->private_data;
	PDEBUG("close...\n");

	/* was a channel configured? */
	if (cfg->channel >= 0)
		hbx_dsp_release_codec(cfg->config, cfg->channel);

	if (cfg)
		kfree(cfg);

	return 0;
}


/**
 * User wants to read from our buffer 
 */
static ssize_t hbx_dsp_read(	struct file *filp, 
				char __user *buf, 
				size_t count, loff_t *offp)
{
	struct hbx_dsp_file_config* cfg = filp->private_data;
	struct rwbuf *rbuf;
	struct dbuf *dbuf;
	unsigned short length, samples;
	
	/* if no channel was configured */
	if (cfg->channel < 0)
		return -ENODEV;

	/* get read ring buffer for channel */
	rbuf = &cfg->config->rbuf[cfg->channel];

	while (rbuf->write_idx == rbuf->read_idx) {

		/* if non blocking, return */
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;

		/* wait until date is available */
		interruptible_sleep_on(&read_wait_q);

		/* check for pending signals */
		if (signal_pending(current))
			return -ERESTARTSYS;
	}

	/* get data buffer */
	dbuf = &rbuf->dbuf[rbuf->read_idx];

	/* user is insane! he should know how much to read */
	if (count < dbuf->size)
		return -EIO;

	/* copy data to user */
	copy_to_user(buf, dbuf->data_buf, dbuf->size);
	
	/* ignore frame information? */
	memcpy(&length,  dbuf->info_buf+1, sizeof(length));
	memcpy(&samples, dbuf->info_buf+3, sizeof(samples));

	/* not needed because DSP send us BE
	length  = le16_to_cpu(length);
	samples = le16_to_cpu(samples);
	*/

	/*
	PDEBUG("Samples = %d\n", samples);
	PDEBUG("Length = %d\n", length);
	*/

	dbuf->size = 0;
	dbuf->idx  = 0;

	rbuf->read_idx++;
	rbuf->read_idx%=HBX_DSP_RBS;

	return length;
}


/**
 * User wants to write into our buffer 
 */
static ssize_t hbx_dsp_write(	struct file *filp, 
				const char __user *buf, 
				size_t count, loff_t *offp)
{
	struct hbx_dsp_file_config* cfg = filp->private_data;
	struct rwbuf *wbuf;
	struct dbuf *dbuf;
	unsigned short length, samples;	

	/* if channel is unconfigured, return an error */
	if (cfg->channel < 0)
		return -ENODEV;

	/* get write ring buffer for channel */
	wbuf = &cfg->config->wbuf[cfg->channel];
	/* get data buffer (to write to it) */
	dbuf = &wbuf->dbuf[wbuf->write_idx];

	while (dbuf->size > 0 ) {
		
		/* if non blocking, return */
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;

		/* wait until free buffer is available */
		interruptible_sleep_on(&write_wait_q);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}
	/* only send HBX_DSP_BSZ bytes */
	if (count > HBX_DSP_BSZ)
		count = HBX_DSP_BSZ;

	/* calculate length in le */
	length	= cpu_to_le16((unsigned short)count);
	/* number of samples was set by ioctl from user */
	samples = cpu_to_le16(cfg->samples);

	/* copy voice data to buffer */
	copy_from_user(dbuf->data_buf, buf, count);

	/* fill info buffer */
	memset(dbuf->info_buf, LCT_CONTINUE, sizeof(dbuf->info_buf));
	/* set START_FRAME flag */
	dbuf->info_buf[0] = LCT_START;
	/* set length and samples */
	memcpy(dbuf->info_buf+1, &length,  sizeof(length));
	memcpy(dbuf->info_buf+3, &samples, sizeof(samples));

	dbuf->size = count;
	dbuf->idx  = 0;

	wbuf->write_idx++;
	wbuf->write_idx%=HBX_DSP_RBS;

	return count;
}

/**
 * IOCTL
 * Used to:
 * 	- configure which codec to use
 * 	- set number of samples with each write
 */
static int hbx_dsp_ioctl(	struct inode *inode, 
				struct file *filp, 
				unsigned int cmd, unsigned long arg)
{
	struct hbx_dsp_file_config* cfg = filp->private_data;
	int ret;

	/* check for correct ioctl numbers */
	if (_IOC_TYPE(cmd) != HBX_DSP_IOC_MAGIC) 
		return -ENOTTY;

	switch (cmd) {
		
		case HBX_DSP_SET_CODEC:
			PDEBUG("Set codec %lu\n", arg);
			ret = hbx_dsp_request_codec(cfg->config, arg);
			if (ret < 0)
				return ret;

			cfg->channel = ret;
			PDEBUG("Aquired channel %d\n", ret);
			break;

		case HBX_DSP_SET_SAMPLES:
			cfg->samples = arg;
			return 0;

		default:
			PDEBUG("Unknown ioctl %d\n", cmd);
			return -ENOTTY;
	}

	return 0;
}

/**
 * list of supported file operations 
 */
struct file_operations hbx_dsp_fops = {
	.owner =    THIS_MODULE,
	.read =     hbx_dsp_read,
	.write =    hbx_dsp_write,
	.ioctl =    hbx_dsp_ioctl,
	.open =     hbx_dsp_open,
	.release =  hbx_dsp_close,
};


