/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>             *
 *                                                                         *
 *   IXP425 High Speed Serial Driver for Linux				   *
 * 									   *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
 * System-defined header files
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>


/*
 * Intel IXP400 Software specific header files
 */
#include <IxQMgr.h>
#include <IxNpeDl.h>
#include <IxNpeMh.h>
#include <IxOsServices.h>
#include <IxOsCacheMMU.h>

#include <IxHssAcc.h>

#include "ixp425_hss.h"
#include "ixp425_hss_port0_cfg.h"
#include "ixp425_hss_port1_cfg.h"

/* Use Macro PDEBUG for Printing Debug-Messages */
#undef PDEBUG
#ifdef IXP425_HSS_DEBUG
	#define PDEBUG(fmt, args...) 	\
		printk("[%s:%d] "fmt, __FILE__,__LINE__, ##args)
#else
	#define PDEBUG(fmt, args...)
#endif


#define __unused 	__attribute__((__unused__))
#define UINT		unsigned int
#define	CHUNKSIZE	8
#undef tv_sec

/*
 * Prototypes
 */
/* Initialize hss/pcm bus */
int ixp425_hss_init(void);

/* Releases the HSS bus.  */
void ixp425_hss_release(void);


/*
 * Generic port configuration
 */
struct si3210_hss_config {
	/* Number of channels we support. Note: this is not the number of
	 * channels we have to use. It is more some kind of maximum 
	 * possible devices we can support. */
	unsigned	channels;

	/* This is the buffer where npe stores the data to. 
	 * Size is calculated by rx_bytes_per_ts * channels */
	unsigned char*	rx_buffer;
	
	/* Number of bytes for one receive time slot (channel). */
	unsigned	rx_bytes_per_ts;

	/* This is the buffer where npe reads data from. Size is 
	 * calculated by tx_bytes_per_ch * channels and should be 
	 * equal to the size of the rx_buffer for easier handling.*/
	unsigned char*	tx_buffer;

	/* Buffer sizeof rx and tx buffers */
	unsigned	tx_buffer_size;
	unsigned	rx_buffer_size;
	
	/* This is a list of pointers. Each entry points to data with 
	 * size of tx_bytes_per_ts. There are tx_ts_per_chan pointers 
	 * for each channel. So the pointer list holds tx_ts_per_chan *
	 * channels pointers. */
	unsigned char**	tx_pointer_buffer;

	/*  Number of bytes in a transmit timeslot. For each channel there
	 *  are tx_ts_per_chan of these buffers. */
	unsigned	tx_bytes_per_ts;

	/* Number of timeslots for each transmit channel */
	unsigned	tx_ts_per_chan;

	/* This is the size of all transmit timeslots for each channel.
	 * (tx_bytes_per_ts * tx_ts_per_chan ) */
	unsigned	tx_bytes_per_ch;
	
	/* Number of bytes to receive until an interrupt will occure */
	unsigned	rx_chunk_size;

	/* These are counters to remember where we read or wrote last. */
	unsigned	rx_current;
	unsigned	tx_current;

	/* if zero, pcm bus is not running */
	int		is_pcm_running;

	struct {
		transfer_cb	func;
		void*		data;
	} transfer[8];

	void(*trans_fn)(IxHssAccHssPort, unsigned, unsigned, unsigned);
	/* for statistic */
	unsigned	drops_total;
	unsigned	loopback_success;
	unsigned	loopback_fails;

	/* packatized buffer */
	IX_OSAL_MBUF_POOL *poolIdPtr;

} hss_config[2];

/* Functions for channelized services */
#include "ixp425_hss_chn.c"

/* Functions for packetized services */
#include "ixp425_hss_pkt.c"

/* functions for proc file system */
#include "ixp425_hss_proc.c"

/**
 * Stops channelized pcm transfer on port 0.
 * \param cfg	the hss configuration.
 */
static void ixp425_hss_stop_pcm0(struct si3210_hss_config* cfg)
{
	int res;
	
	/* disconnecr */
	res = ixHssAccChanDisconnect(0);
	if (IX_SUCCESS != res) {
		/* did not work? just print out a warning */
		printk(KERN_ERR "Error: Unable to disconnect from hss bus\n");
	}

	/* free memory */
	
	/* memory for pointer list */
	ixOsalCacheDmaFree(cfg->tx_pointer_buffer);

	/* memory for receive buffer */
	ixOsalCacheDmaFree(cfg->rx_buffer);
	
	/* memroy for transmit buffer */
	ixOsalCacheDmaFree(cfg->tx_buffer);

	cfg->tx_pointer_buffer 	= NULL;
	cfg->rx_buffer		= NULL;
	cfg->tx_buffer		= NULL;
}


/**
 * Stops packetized data transfer on port 1.
 * \param cfg	the hss configuration.
 */
static void ixp425_hss_stop_pcm1(struct si3210_hss_config* cfg)
{
	int res;
	
	if (!cfg->is_pcm_running)
		return;

	res = ixHssAccPktPortDisable(1,0);
	if (IX_SUCCESS != res) {
		/* did not work? just print out a warning */
		printk(KERN_ERR "Error: Unable to disable hss bus\n");
	}

	rxFreeLowCallback(0);

	/* disconnecr */
	res = ixHssAccPktPortDisconnect(1,0);

	/* if the service is in the process of disconnnecting */
	if (res == IX_HSSACC_PKT_DISCONNECTING) {   
		int elapsedMilliseconds = 0;

		/* wait for a maximum of 250ms */
		while (elapsedMilliseconds < 250) {   
			/* if the disconnect is complete then exit with success */
			if (ixHssAccPktPortIsDisconnectComplete(1,0)) {   
				res = IX_SUCCESS;
				break;
			}

			/* wait for 50ms */
			ixOsalSleep (50);
			elapsedMilliseconds += 50;
		}
	}	

	if (IX_SUCCESS != res) {
		/* did not work? just print out a warning */
		printk(KERN_ERR "Error: Unable to disconnect from hss bus\n");
	}

	/* free memory */
	ixOsalBuffPoolUninit(cfg->poolIdPtr);
	
	/* XXX Does not really work :-( */
}


/**
 * Stops pcm transfer.
 * \param port Specifies the port to stop
 */
static void ixp425_hss_stop_pcm(int port)
{
	struct si3210_hss_config* cfg;
	
	cfg = &hss_config[port];
	
	if (!cfg->is_pcm_running)
		return;

	switch(port) {
		case 0:
			ixp425_hss_stop_pcm0(cfg);
			break;
		case 1:
			ixp425_hss_stop_pcm1(cfg);
			break;
	}

	cfg->is_pcm_running = 0;
}



/**
 * Starts channelized pcm transfer on port 0.
 * \return zero on success
 */
static int ixp425_hss_start_pcm0(void)
{
	unsigned ts, ch;
	int	 ret;
	struct si3210_hss_config* 	cfg;
	
#ifndef IXP425_HSS_ENABLE_PORT0
	return 0;
#endif

	cfg = &hss_config[0];
	memset(cfg, 0, sizeof(struct si3210_hss_config));

	cfg->channels	= hss_port0_config.numChannelised;
	cfg->trans_fn	= IXP425_HSS_PORT0_TRANS_FN;

	PDEBUG("Port %d is configured with %d channels\n", 0, cfg->channels);

	cfg->rx_bytes_per_ts	= 128;	/* important, limited to 8bit!!! */
	cfg->rx_chunk_size	=   8; /* 16 works */
	cfg->rx_current		=   0; 
	cfg->tx_current		=  16; 
	cfg->tx_bytes_per_ts	=  32;	/* Minimum 16, Maximum 48 */
	cfg->tx_ts_per_chan	=   4;	/* 32 * 4 == 128 */
	cfg->tx_bytes_per_ch 	= cfg->tx_bytes_per_ts * cfg->tx_ts_per_chan;
	cfg->tx_buffer_size	= cfg->tx_bytes_per_ch * cfg->channels;
	cfg->rx_buffer_size	= cfg->rx_bytes_per_ts * cfg->channels;
	
	memset(cfg->transfer, 0, sizeof(cfg->transfer));

	/* get memory for receive buffer */
	cfg->rx_buffer = ixOsalCacheDmaMalloc(cfg->rx_buffer_size);

	
	/* get memory for transmit buffer (should have the size of rx_buffer)*/
	cfg->tx_buffer = ixOsalCacheDmaMalloc(cfg->tx_buffer_size);

	/* get memory for pointer list */
	cfg->tx_pointer_buffer = ixOsalCacheDmaMalloc(
		sizeof(UINT*) * cfg->channels * cfg->tx_ts_per_chan);

	/* check if malloc succeeded */
	if ( 	cfg->rx_buffer == NULL || 
		cfg->tx_buffer == NULL || 
		cfg->tx_pointer_buffer == NULL) {
		
		printk("Error: Unable to allocate memory\n");
		return -1;
	}

	/* initialize with idle pattern */
	memset(cfg->rx_buffer, 0xd5, cfg->rx_bytes_per_ts * cfg->channels);
	memset(cfg->tx_buffer, 0xd5, cfg->tx_bytes_per_ch * cfg->channels);
	
	/* set up pointer list */
	for (ts=0; ts<cfg->tx_ts_per_chan; ts++) {
		for (ch=0; ch<cfg->channels; ch++) {
			cfg->tx_pointer_buffer[ts*cfg->channels+ch] =
				(unsigned char*)
				IX_OSSERV_SWAP_NPE_SHARED_LONG(
				IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(
				cfg->tx_buffer + 
				ts * cfg->tx_bytes_per_ts +
				ch * cfg->tx_bytes_per_ch));
		}
	}

	/* flush cached memory area */
	IX_ACC_DATA_CACHE_FLUSH(cfg->tx_pointer_buffer, 
			sizeof(UINT*) * cfg->channels * cfg->tx_ts_per_chan);

	IX_ACC_DATA_CACHE_FLUSH(
			cfg->tx_buffer,cfg->tx_bytes_per_ch * cfg->channels);

	IX_ACC_DATA_CACHE_FLUSH(
			cfg->rx_buffer,cfg->rx_bytes_per_ts * cfg->channels);
	
	/* last configuration steps */
	PDEBUG("ixHssAccChanPortConnect for port %d ...\n", 0);
	ret = ixHssAccChanConnect(0,
				cfg->rx_chunk_size,
				cfg->rx_buffer,
				cfg->rx_bytes_per_ts,
				(UINT32*)cfg->tx_pointer_buffer,
				cfg->tx_ts_per_chan,
				cfg->tx_bytes_per_ts,
				cfg->trans_fn);
	
	if (IX_SUCCESS != ret) {
		printk("Error: Unable to connect to port 0\n");
		/* free memory */
		ixp425_hss_stop_pcm(0);
		return -EIO;
	}
	
	/* start transfers */
	PDEBUG("ixHssAccChanPortEnable for port %d ...\n", 0);
	ret = ixHssAccChanPortEnable(0);
	if (IX_SUCCESS != ret) {
		printk("Error: Unable to enable port %d\n", 0);
		/* free memory */
		ixp425_hss_stop_pcm(0);
		return -EIO;
	}

	cfg->is_pcm_running = 1;
	return 0;
}


static int ixp425_hss_start_pcm1(void)
{
	int ret;
	struct si3210_hss_config* 	cfg;
	
	cfg = &hss_config[1];
	memset(cfg, 0, sizeof(struct si3210_hss_config));


	PDEBUG("ixHssAccPktPortConnect for port 1 ...\n");

	ret = ixHssAccPktPortConnect(
/* IxHssAccHssPort hssPortId, */			1,
/* IxHssAccHdlcPort hdlcPortId, */			0,
/* BOOL hdlcFraming, */					0,
/* IxHssAccHdlcMode hdlcMode, */			hdlcMode,
/* BOOL hdlcBitInvert, */				0,
/* unsigned blockSizeInWords, */ 			64,
/* UINT32 rawIdleBlockPattern, */			0,
/* IxHssAccPktHdlcFraming hdlcTxFraming, */		hdlcTxFraming,
/* IxHssAccPktHdlcFraming hdlcRxFraming, */		hdlcRxFraming,
/* unsigned frmFlagStart, */				1,
/* IxHssAccPktRxCallback rxCallback, */			rxCallback,
/* IxHssAccPktUserId rxUserId, */			0,
/* IxHssAccPktRxFreeLowCallback rxFreeLowCallback, */	rxFreeLowCallback,
/* IxHssAccPktUserId rxFreeLowUserId, */		0,
/* IxHssAccPktTxDoneCallback txDoneCallback, */		txDoneCallback,
/* IxHssAccPktUserId txDoneUserId) */			0);

	if (IX_SUCCESS != ret) {
		printk("Error: Unable to connect to port 1\n");
		return -EIO;
	}

	
	cfg->poolIdPtr = IX_OSAL_MBUF_POOL_INIT(128, 32*8, "HssAcc Codelet Pool");

	rxFreeLowCallback(0);

	PDEBUG("ixHssAccPktPortEnable for port 1 ...\n");
	ret = ixHssAccPktPortEnable(1, 0);
	if (IX_SUCCESS != ret) {
		printk("Error: Unable to enable port 1\n");
		return -EIO;
	}


	return 0;
}


static int ixp425_hss_start_pcm(int port) 
{
	switch (port) {
		case 0:
			return ixp425_hss_start_pcm0();
		case 1:
			return ixp425_hss_start_pcm1();
	}

	return -1;
}
int ixp425_hss_register(transfer_cb cb, void* data)
{
	struct si3210_hss_config* cfg;
	int i;
	
	cfg = &hss_config[0];

	/* start hss bus */
	if (!cfg->is_pcm_running)
		ixp425_hss_start_pcm(0);

	/* find free slot */
	for (i=0; i<8; i++) {
		if (cfg->transfer[i].func)
			continue;

		cfg->transfer[i].data = data;
		cfg->transfer[i].func = cb;
		PDEBUG("Register channel #%d for hss bus\n", i);
		return i * 16;
	}

	PDEBUG("Unable to register channel for hss bus\n");
	return -1;
}

	
void ixp425_hss_unregister(int bit)
{
	struct si3210_hss_config* cfg;
	
	cfg = &hss_config[0];

	cfg->transfer[bit/16].func = NULL;
	PDEBUG("Unregister channel #%d for hss bus\n", bit/16);

#ifdef IXP425_HSS_DYNAMIC_PORT0
	/* find free slot */
	for (bit=0; bit<8; bit++) {
		if (cfg->transfer[bit].func)
			return;

	}

	PDEBUG("Stopping hss bus\n");
	ixp425_hss_stop_pcm(0);
#endif
}


int ixp425_hss_register2(transfer_cb cb, void* data)
{
	struct si3210_hss_config* cfg;
	
	cfg = &hss_config[1];

	/* start hss bus */
	if (!cfg->is_pcm_running)
		ixp425_hss_start_pcm(1);

	if (cfg->transfer[0].func) {
		PDEBUG("Unable to register channels for hss bus 2\n");
		return -EBUSY;
	}

	cfg->transfer[0].data = data;
	cfg->transfer[0].func = cb;
	PDEBUG("Register all channels for hss bus 2\n");

	return 0;
}

	
void ixp425_hss_unregister2(void)
{
	struct si3210_hss_config* cfg;
	
	cfg = &hss_config[1];

	cfg->transfer[0].func = NULL;
	PDEBUG("Unregister all channels for hss bus 2\n");

#ifdef IXP425_HSS_DYNAMIC_PORT1
	PDEBUG("Stopping hss bus\n");
	ixp425_hss_stop_pcm(1);
#endif
}


static void ixp425_hss_error_cb(unsigned last_error, unsigned port)
{
	printk("error call back was called\n");
	printk("last_error = 0x%x, port = 0x%x\n", last_error, port);
}

/**
 * Initilizes HSS bus.
 * \returns zero on success.
 */
int ixp425_hss_init(void)
{
	int ret;
	IxQMgrDispatcherFuncPtr dispatcherFunc;
	
	/*
	 * Initialize HSS;
	 * this bus will be used for data/voice transfer
	 */
	PDEBUG("Initialize HSS ...\n");

	dispatcherFunc = NULL;

#ifdef IXP425_HSS_DEBUG
	ixOsServLogLevelSet(LOG_ALL);
#else
	ixOsServLogLevelSet(LOG_ERROR);
#endif
	
	PDEBUG("ixNpeDlNpeInitAndStart ...\n");
	ret = ixNpeDlNpeInitAndStart(IX_NPEDL_NPEIMAGE_NPEA_HSS_2_PORT);
	if (IX_SUCCESS != ret) {
		printk("Error: Could not initilize NPE DL\n");
		return -EIO;
	}

	PDEBUG("ixNpeMhInitialized ...\n");
	ret = ixNpeMhInitialize(IX_NPEMH_NPEINTERRUPTS_YES);
	if (IX_SUCCESS != ret) {
		printk("Error: Could not initilize NPE\n");
		return -EIO;
	}

	PDEBUG("ixQMgrInit ...\n");
	ret = ixQMgrInit();
	if (IX_SUCCESS != ret) {
		printk("Warning: Queue Manager already initialised\n");
	}

	ixQMgrDispatcherLoopGet(&dispatcherFunc);

	PDEBUG("ixOsServIntBind ...\n");
	ret = ixOsServIntBind(		IX_OSAL_IXP400_QM1_IRQ_LVL,
					(IxVoidFnVoidPtr)dispatcherFunc,
					(void *) IX_QMGR_QUELOW_GROUP);
	if (IX_SUCCESS != ret) {
		printk("ixOsServIntBind failed\n");
		return -EIO;
	}
	
	PDEBUG("ixHssAccInit ...\n");
	ret = ixHssAccInit();
	if (IX_SUCCESS != ret) {
		printk("Error: Could not initialize high speed serial"
				"component\n");
		return -EIO;
	}

#ifdef IXP425_HSS_ENABLE_PORT0
	hss_port0_config.txPortConfig = hss_port0_config_tx;
	hss_port0_config.rxPortConfig = hss_port0_config_rx;

	PDEBUG("ixHssAccPortInit for port 0 ...\n");
	ret = ixHssAccPortInit(	IX_HSSACC_HSS_PORT_0,
				&hss_port0_config,
				hss_tdm_config0,
				ixp425_hss_error_cb);
	if (IX_SUCCESS != ret) {
		printk("Error: Unable to setup port 0\n");
		return -EIO;
	}
#endif

#ifdef IXP425_HSS_ENABLE_PORT1
	hss_port1_config.txPortConfig = hss_port1_config_tx;
	hss_port1_config.rxPortConfig = hss_port1_config_rx;

	PDEBUG("ixHssAccPortInit for port 1 ...\n");
	ret = ixHssAccPortInit(	IX_HSSACC_HSS_PORT_1,
				&hss_port1_config,
				hss_tdm_config1,
				ixp425_hss_error_cb);
	if (IX_SUCCESS != ret) {
		printk("Error: Unable to setup port 1\n");
		return -EIO;
	}
#endif

	/* start hss bus on module load */
#ifndef IXP425_HSS_DYNAMIC_PORT0
	ixp425_hss_start_pcm(0);
#endif

#ifndef IXP425_HSS_DYNAMIC_PORT1
	ixp425_hss_start_pcm(1);
#endif
	
	ixp425_hss_proc_register();
	
	return 0;
}

/**
 * Releases the HSS bus.
 * \returns zero on success.
 */
void ixp425_hss_release(void)
{
	PDEBUG("Unregister proc files\n");
	ixp425_hss_proc_unregister();
	
	PDEBUG("Stopping hss bus\n");
	ixp425_hss_stop_pcm(0);
	ixp425_hss_stop_pcm(1);

	PDEBUG("Unload intel stuff\n");
	
	PDEBUG("Unbind routine from hardware irq\n");
	ixOsServIntUnbind(IX_OSAL_IXP400_QM1_IRQ_LVL);

	PDEBUG("Unload queue manager\n");
//	ixQMgrUnload();
	
	PDEBUG("Unload npe message handler\n");
	ixNpeMhUnload();

	PDEBUG("Unload npe download manager\n");
	ixNpeDlUnload();
}

MODULE_DESCRIPTION("High Speed Serial Driver for Intel IXDP425 Boards");
MODULE_AUTHOR("Henry Margies (h.margies@maxina.de");
MODULE_LICENSE("GPL");


module_init(ixp425_hss_init);
module_exit(ixp425_hss_release);

EXPORT_SYMBOL(ixp425_hss_register);
EXPORT_SYMBOL(ixp425_hss_unregister);
EXPORT_SYMBOL(ixp425_hss_register2);
EXPORT_SYMBOL(ixp425_hss_unregister2);
