/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>             *
 *                                                                         *
 *   IXP425 High Speed Serial Driver for Linux				   *
 * 									   *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#if (defined IXP425_HSS_PORT0_LOOPBACK_CLIENT) && (defined IXP425_HSS_PORT0_LOOPBACK_SERVER)
#error "You can not define loopback client and server at the same time for port 0"
#endif

/*
 * Buffer organisation for channelized services(for port 0)
 * ========================================================
 *
 * Receive buffer
 * --------------
 *  
 * The receive buffer is very simple organised. rx_buffer points to the 
 * beginning of this buffer. It starts with a circular buffer for the first
 * channel. The size is rx_bytes_per_ts. After this buffer is the circular 
 * buffer for the next channel and so on.
 * NOTE: maximum value of rx_bytes_per_ts is 255 (our default is 128).
 *
 * rx_buffer ------->	+-----------------------+
 * 			| Receive buffer for 	|
 * 			| channel 0 (128 bytes)	|
 * 			+-----------------------+
 * 			| Receive buffer for 	|
 * 			| channel 1 (128 bytes)	|
 * 			+-----------------------+
 * 			|	....		|
 * 			+-----------------------+
 * 			| Receive buffer for	|
 * 			| channel 7 (128 bytes) |
 * 			+-----------------------+
 *
 * rx_current goes from 0 to (exclusive) rx_bytes_per_ts. 
 *
 * Transmit buffer
 * ---------------
 *
 * The transmit buffer is more complex. The NPE uses a circular pointer 
 * list and writes to the data where the pointers point to. These
 * pointers are set by ixp425_hss in a way that it seams there is only 
 * one big transmit buffer per channel as it is for the receive buffer.
 *
 *  +-- tx_pointer_buffer
 *  |   tx_buffer ---------+
 *  \/			   |
 *  +-----------+          +--> +-----------------------+ <--+
 *  | TS:0 CH:0	| ------------>	| TS:0 CH:0 (32 bytes)	|    |	
 *  +-----------+		+-----------------------+    |
 *  | TS:0 CH:1	| -+  +------->	| TS:1 CH:0 (32 bytes)	|    +- tx_bytes_per_ch
 *  +-----------+  |  | 	+-----------------------+    |  ( 32 * 3 bytes)
 *  | TS:0 CH:2	|  |  | +----->	| TS:2 CH:0 (32 bytes)	|    |
 *  +-----------+  |  | |	+-----------------------+ <--+
 *  |	...	|  +--|-|-----> | TS:0 CH:1 (32 bytes)	|
 *  +-----------+     | |	+-----------------------+
 *  | TS:0 CH:7	|     | |  +-->	| TS:1 CH:1 (32 bytes)	|
 *  +-----------+     | |  |	+-----------------------+
 *  | TS:1 CH:0	| ----+ |  | +>	| TS:2 CH:1 (32 bytes)	|
 *  +-----------+	|  | |	+-----------------------+
 *  | TS:1 CH:1	| ------|--+ |	| TS:0 CH:2 (32 bytes)	|
 *  +-----------+	|    |	+-----------------------+
 *  |	...	|	|    |	|	...		|
 *  +-----------+	|    |	+-----------------------+
 *  | TS:1 CH:7	|  	|    |	|	...		|
 *  +-----------+	|    |	+-----------------------+
 *  | TS:2 CH:0	| ------+    |	| TS:2 CH:6 (32 bytes)	|
 *  +-----------+	     |	+-----------------------+
 *  | TS:2 CH:1	| -----------+	| TS:0 CH:7 (32 bytes)	|
 *  +-----------+		+-----------------------+
 *  |	...	|		| TS:1 CH:7 (32 bytes) 	|
 *  +-----------+		+-----------------------+
 *  | TS:2 CH:7	| ------------>	| TS:2 CH:7 (32 bytes)	|
 *  +-----------+		+-----------------------+
 *
 * For this figure: 
 * 			tx_bytes_per_ts = 32 (default 32)
 * 			tx_ts_per_chan	=  3 (default  4)
 * 			tx_bytes_per_ch = 96 (default is 128)
 * 					= tx_bytes_per_ts * tx_ts_per_chan
 * tx_current should run between 0 (inclusive) and 128 (exclusive). The 
 * start value for tx_current and rx_current are set in ixp425_hss.c. So, 
 * filling the transmit buffer should work in a same way the rx_buffer
 * is handled. There is no need to use tx_pointer_buffer (to be precise:
 * don't use it!)
 */
				
#ifdef IXP425_HSS_PORT0_LOOPBACK_CLIENT
#define IXP425_HSS_PORT0_TRANS_FN ixp425_hss_loopback_cb

/**
 * This function tranfers the data between hss bus and zaptel.
 */
static void ixp425_hss_loopback_cb(IxHssAccHssPort port, 
					unsigned rx_offset, 
					unsigned tx_offset __unused, 
					unsigned num_hss_errs)
{
	struct si3210_hss_config* 	cfg;
	int bytes_received;
	unsigned i,o;
	unsigned char* tx_ptr;
	unsigned char* rx_ptr;

	static int send_counter[2][8] = {
				{ 0, 0, 0, 0, 0, 0, 0, 0},
				{ 0, 0, 0, 0, 0, 0, 0, 0}};

	static int last_counter[2][8] = {
				{-1,-1,-1,-1,-1,-1,-1,-1},
				{-1,-1,-1,-1,-1,-1,-1,-1}};

	static int firsttime[2] = {0,0};

	cfg 	= &hss_config[port];

	/* calculate received bytes */
	bytes_received = rx_offset - cfg->rx_current;
	if (bytes_received < 0)
		bytes_received += cfg->rx_bytes_per_ts;

	if (cfg->rx_current >= cfg->rx_bytes_per_ts)
		cfg->rx_current = 0;

	if (cfg->tx_current >= cfg->tx_bytes_per_ch)
		cfg->tx_current = 0;

	rx_ptr = cfg->rx_buffer + cfg->rx_current;
	tx_ptr = cfg->tx_buffer + cfg->tx_current;

	/* read all channels */
	for (i=0; i<1/*cfg->channels*/; i++) {

		IX_ACC_DATA_CACHE_INVALIDATE((__u32)rx_ptr, 32);

		for (o=0; o<bytes_received;o++) {
			*tx_ptr = send_counter[port][i]++;

			if (last_counter[port][i] < 0) { /* test is not active */
				if (*rx_ptr == 0) { /* start of test */
					last_counter[port][i] = 0;
					cfg->loopback_success++;
					if (!firsttime[port]++)
					printk("[P%d] Loopback Test [%d]: started\n",port, i);
				}
			} else {
				if (*rx_ptr != last_counter[port][i]+1) {
					printk("[P%d] Loopback Test [%d]: failed "
						"(read %d, wanted %d)\n", port,
						i, *rx_ptr, last_counter[port][i]+1);
					last_counter[port][i] = -1;
					cfg->loopback_fails++;
				} else
					last_counter[port][i]++;
			}
	
			if (send_counter[port][i] > 255 )
				send_counter[port][i]=0;
			if (last_counter[port][i] == 255) {
				last_counter[port][i]=-1;
			}
			tx_ptr++;
			rx_ptr++;
		}
		tx_ptr -= bytes_received;
		rx_ptr -= bytes_received;
			
		IX_ACC_DATA_CACHE_FLUSH((__u32)tx_ptr, 32);

		tx_ptr += cfg->tx_bytes_per_ch;
		rx_ptr += cfg->rx_bytes_per_ts;
	}

	cfg->rx_current	+= bytes_received;
	cfg->tx_current	+= bytes_received;

		
	if (num_hss_errs != 0)
		printk("Error occurd %d\n", num_hss_errs);
}
#endif


#ifdef IXP425_HSS_PORT0_LOOPBACK_SERVER
#define IXP425_HSS_PORT0_TRANS_FN ixp425_hss_loopback2_cb

static void ixp425_hss_loopback2_cb(IxHssAccHssPort port, 
					unsigned rx_offset, 
					unsigned tx_offset __unused, 
					unsigned num_hss_errs)
{
	struct si3210_hss_config* 	cfg;
	int bytes_received;
	unsigned i;

	cfg = &hss_config[port];

	/* calculate received bytes */
	bytes_received = rx_offset - cfg->rx_current;
	if (bytes_received < 0)
		bytes_received += cfg->rx_bytes_per_ts;
				
	IX_ACC_DATA_CACHE_INVALIDATE((__u32)cfg->rx_buffer, cfg->rx_buffer_size);

	while (bytes_received >= CHUNKSIZE) {

		unsigned char* tx_ptr;
		unsigned char* rx_ptr;

		cfg->drops_total++;

		if (cfg->rx_current >= cfg->rx_bytes_per_ts)
			cfg->rx_current = 0;

		if (cfg->tx_current >= cfg->tx_bytes_per_ch)
			cfg->tx_current = 0;

		rx_ptr = cfg->rx_buffer + cfg->rx_current;
		tx_ptr = cfg->tx_buffer + cfg->tx_current;

		/* read all channels */
		for (i=0; i<cfg->channels; i++) {

			memcpy(	tx_ptr, rx_ptr, CHUNKSIZE);
			cfg->loopback_success++;

			/* also call registered functions
			 * (for timing issue) */
			if (cfg->transfer[i].func) {
				
				unsigned char fake[8];
				
				cfg->transfer[i].func(	fake, fake, 
							cfg->transfer[i].data); 
			
			}
			tx_ptr += cfg->tx_bytes_per_ch;
			rx_ptr += cfg->rx_bytes_per_ts;
		}

		cfg->rx_current	+= CHUNKSIZE;
		cfg->tx_current	+= CHUNKSIZE;
		bytes_received 	-= CHUNKSIZE;

		
	}
				
	IX_ACC_DATA_CACHE_FLUSH((__u32)cfg->tx_buffer, cfg->tx_buffer_size);

	cfg->drops_total--;

	if (num_hss_errs != 0) {
		ixHssAccLastErrorRetrievalInitiate(port);
		printk("Error occurd %d\n", num_hss_errs);
	}
}
#endif


#if (!defined IXP425_HSS_PORT0_TRANS_FN) && (defined IXP425_HSS_ENABLE_PORT0)
#define IXP425_HSS_PORT0_TRANS_FN ixp425_hss_data_transfer_cb

/**
 * This function tranfers the data between hss bus and zaptel.
 */
static void ixp425_hss_data_transfer_cb(IxHssAccHssPort port,
					unsigned rx_offset, 
					unsigned tx_offset __unused, 
					unsigned num_hss_errs)
{
	struct si3210_hss_config* 	cfg;
	int bytes_received;
	unsigned i;

#ifdef IXP425_HSS_PROFILE
	static unsigned last_ji = 0;
	static int counter = 0;

	struct timeval	start_time;
	struct timeval	end_time;

	do_gettimeofday(&start_time);
#endif
	
	cfg = &hss_config[port];

	/* calculate received bytes */
	bytes_received = rx_offset - cfg->rx_current;
	if (bytes_received < 0)
		bytes_received += cfg->rx_bytes_per_ts;
				
	IX_ACC_DATA_CACHE_INVALIDATE((__u32)cfg->rx_buffer, cfg->rx_buffer_size);

	while (bytes_received >= CHUNKSIZE) {

		unsigned char* tx_ptr;
		unsigned char* rx_ptr;

		cfg->drops_total++;

		if (cfg->rx_current >= cfg->rx_bytes_per_ts)
			cfg->rx_current = 0;

		if (cfg->tx_current >= cfg->tx_bytes_per_ch)
			cfg->tx_current = 0;

		rx_ptr = cfg->rx_buffer + cfg->rx_current;
		tx_ptr = cfg->tx_buffer + cfg->tx_current;

		/* read all channels */
		for (i=0; i<cfg->channels; i++) {

			if (cfg->transfer[i].func) {
				
				cfg->transfer[i].func(	rx_ptr, tx_ptr, 
							cfg->transfer[i].data); 
			
			}

			tx_ptr += cfg->tx_bytes_per_ch;
			rx_ptr += cfg->rx_bytes_per_ts;
		}

		cfg->rx_current	+= CHUNKSIZE;
		cfg->tx_current	+= CHUNKSIZE;
		bytes_received 	-= CHUNKSIZE;

		
	}
				
	IX_ACC_DATA_CACHE_FLUSH((__u32)cfg->tx_buffer, cfg->tx_buffer_size);

	cfg->drops_total--;

	if (num_hss_errs != 0) {
		ixHssAccLastErrorRetrievalInitiate(port);
		printk("Error occurd %d\n", num_hss_errs);
	}

#ifdef IXP425_HSS_PROFILE
	do_gettimeofday(&end_time);

	counter += (end_time.tv_usec - start_time.tv_usec) + 
		   (end_time.tv_sec - start_time.tv_sec)*1000000L;
	
	if (last_ji <= jiffies) {
		printk("%d jiffies occured. Time in interrupt per second: %d usecs \n",
			jiffies-last_ji+1000, counter/10);
		last_ji = jiffies + 1000;
		counter = 0;
	}
#endif
}
#endif

