/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
 * This is the main source code for the si3050 zaptel driver.
 * It uses:
 * 	ixp425_spi.h:		for activating the SPI bus on ixp425 boards
 * 	ixp425_hss.h:		for handling the hss(pcm) bus on ixp425 boards
 * 	si3050_driver.h:	for driver initialization, interrupt handling
 * 				and all chipset specific things
 * The main tasks for the code in this file are:
 * 	- trigger initialization of the hardware
 * 	- register on zaptel
 * 	- handle events between zaptel and si3050
 * 	- transfer voice data
 *
 * TODO:
 * 	- write si3050_exit
 * 	- write si3050_zap_close
 * 	- try to avoid memcpy
 *
 */

#include <linux/init.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/delay.h>


#include <linux/zaptel.h>

#include "ixp425_spi.h"
#include "ixp425_hss.h"
#include "hbx_dsp.h"

#include "si3050_driver.h"

/* Use macro PDEBUG for printing debug messages */
#undef PDEBUG             // undef it, just in case
#ifdef SI3050_DEBUG
	#define PDEBUG(fmt, args...) 	\
		printk("[%s:%d] "fmt, __FILE__,__LINE__, ##args)
#else
	#define PDEBUG(fmt, args...) {}
#endif

#define __unused __attribute__((__unused__))

static struct si3050_config {
	struct si3050_driver_config 	driver_config;

	struct zt_span span;
	struct zt_chan chan;
	
	struct work_struct  work_req_ec;
	struct work_struct  work_rel_ec;

} si3050_config;

static void si3050_status_changed(struct si3050_driver_config* chip_config);
static void si3050_dtmf_detected(struct si3050_driver_config* driver_config);
static void si3050_polarity_changed(struct si3050_driver_config* driver_config);

/*
 *
 * Functions for transfering voice data between hss and zaptel.
 *
 */


/**
 * This function tranfers the data between hss bus and zaptel.
 */
static void si3050_data_transfer_cb(	unsigned char* rx_ptr,
					unsigned char* tx_ptr,
					void* data)
{
	struct zt_chan* chan = (struct zt_chan*)data;;

	zt_transmit(chan->span);

	/* try to not use memcpy */
	#ifdef SI3050_ZAP_USE_MEMCPY
	memcpy(	chan->readchunk,
		rx_ptr,
		ZT_CHUNKSIZE);
	
	#else
		chan->readchunk = rx_ptr;
	#endif
			
	memcpy(	tx_ptr,
		chan->writechunk,
		ZT_CHUNKSIZE);

	
	/* echo cancel if needed */
	if (chan->ec) {
		int ec_ch = si3050_config.driver_config.ec_channel;

		if (ec_ch >= 0) {

			hbx_dsp_put_ec(ec_ch, tx_ptr, chan->readchunk);
			hbx_dsp_get_ec(ec_ch, tx_ptr, chan->readchunk);

		} else {
			zt_ec_chunk(chan, chan->readchunk,chan->writechunk);
		}
	}

	zt_receive(chan->span);

}



/*
 *
 * callback functions for zaptel
 *
 */

/**
 * Opens a channel.
 * \param chan Zaptel channel description.
 * \returns zero on success
 */
static int si3050_zap_open(struct zt_chan *chan) 
{ 
	struct si3050_driver_config*	drv_cfg;
	int ret;

	drv_cfg	= &((struct si3050_config*)chan->pvt)->driver_config;

	if (drv_cfg->in_use)
		return -EBUSY;

	PDEBUG("si3050_zap_open (%s)\n", chan->name);
	drv_cfg->in_use     =  1;
	drv_cfg->ec_channel = -1;

	ret = ixp425_hss_register(	si3050_data_transfer_cb, 
					(void*)chan);
	
	if (ret < 0){
		printk("Unable to start hss/pcm transfers\n");
		return -EIO;
	}
	si3050_start_pcm(drv_cfg->hss_start_bit=ret);
	
	si3050_config.driver_config.dtmf_detected  = si3050_dtmf_detected;
	si3050_config.driver_config.status_changed = si3050_status_changed;
	si3050_config.driver_config.polarity_changed = si3050_polarity_changed;

	return 0;
}


/**
 * Closes a channel.
 * \param chan Zaptel channel description
 * \returns zero on success.
 */
static int si3050_zap_close(struct zt_chan *chan) 
{

	struct si3050_driver_config*	drv_cfg;

	drv_cfg	= &((struct si3050_config*)chan->pvt)->driver_config;

	/* release echo canceller if in use */
	if (drv_cfg->ec_channel >= 0) 
		schedule_work(&si3050_config.work_rel_ec);
	
	/* check if chan device was really opened!! */
	if (!drv_cfg->in_use)
		return 0;

	PDEBUG("si3050_zap_close (%s)\n", chan->name);
	ixp425_hss_unregister(drv_cfg->hss_start_bit);
	drv_cfg->in_use = 0;
	drv_cfg->ec_channel = -1;
	
	return 0;
}


/**
 * Ioctl function for our zaptel channels.
 * \param chan Zaptel channel description
 * \param cmd to handle.
 * \param data for handling the command.
 * \returns zero on success.
 */
static int si3050_zap_ioctl(	struct zt_chan *chan __unused, 
				unsigned int cmd, 
				unsigned long data)
{
	unsigned int i = 0;

	PDEBUG("si3050_zap_ioctl (%s)\n", chan->name);

	switch(cmd) {
		case ZT_TONEDETECT:
#ifdef SI3050_DTMF_DETECT
			PDEBUG("ZT_TONEDETECT: \n");
			get_user(i, (int*)data);
			if (i & ZT_TONEDETECT_ON)
				PDEBUG("\tZT_TONEDETECT_ON\n");
			if (i & ZT_TONEDETECT_MUTE)
				PDEBUG("\tZT_TONEDETECT_MUTE\n");
			break;
#else
			return -ENOTTY;
#endif

#ifdef ZT_SETPOLARITY
		case ZT_SETPOLARITY:
			PDEBUG("ZT_SETPOLARITY %d\n", i);
			get_user(i, (int*)data);
//			si3050_set_polarity(chan->chanpos-1, i);
			break;
#endif

		default:
			PDEBUG("cmd = %x, data = %x\n", cmd, (unsigned)data);
			return -ENOTTY;
	}
	return 0;
}


/**
 * This function is used by zaptel to tell us to change to off hook or start
 * ringing.
 * \param chan Zaptel channel description
 * \param txsig is the new status we should change to
 * \returns 0 on success.
 */
static int si3050_zap_hooksig(struct zt_chan *chan, zt_txsig_t txsig)
{
	struct si3050_driver_config* 	drv_cfg;
	int chip;
	
	drv_cfg	= &((struct si3050_config*)chan->pvt)->driver_config;

	chip	= chan->chanpos-1;

	PDEBUG("hooksig from zaptel for chip %s: ", 
			chan->name);

	switch (txsig) {
		/* On hook */
		case ZT_TXSIG_ONHOOK:
			PDEBUG("ZT_TXSIG_ONHOOK\n");
			si3050_go_onhook();
			if (drv_cfg->ec_channel >= 0)
				schedule_work(&si3050_config.work_rel_ec);
			break;

		/* Off hook */
		case ZT_TXSIG_OFFHOOK:
			PDEBUG("ZT_TXSIG_OFFHOOK\n");
			if (drv_cfg->ec_channel < 0)
				schedule_work(&si3050_config.work_req_ec);
			si3050_go_offhook();
			break;

		/* Start / Ring */
		case ZT_TXSIG_START:
			PDEBUG("ZT_TXSIG_START\n");
//			si3050_start_ringing(chip);
			break;

		/* Drop battery if possible */
		case ZT_TXSIG_KEWL:
			PDEBUG("ZT_TXSIG_KEWL\n");
			printk("KEWL signal not yet supported\n");
			break;

		default:
			printk("unkown hooksig\n");

	}
	
	return 0;
}

/*
 *
 * Callback function for si3050_driver
 *
 */

/**
 * This function reports detected dtmf digits to zaptel.
 * It is configured as callback function and gets called by
 * si3050_driver.
 * \param chip_config is the chip configuration.
 */
static void si3050_dtmf_detected(struct si3050_driver_config* driver_config)
{
#ifdef SI3050_DTMF_DETECT
	PDEBUG("dtmf = %c\n", driver_config->dtmf);
	zt_qevent_lock(	&si3050_config.chan, 
			(ZT_EVENT_DTMFUP | driver_config->dtmf));
#endif
}

/**
 * This function reports a polarity change event.
 */
static void si3050_polarity_changed(struct si3050_driver_config* 
					driver_config __unused)
{
//	PDEBUG("polarity changed\n");
//	zt_qevent_lock(	&si3050_config.chan, ZT_EVENT_POLARITY);
}


/**
 * This function reports changed status to zaptel.
 * Changing status means from/to on hook, off hook or ringing. This
 * function gets call by si3050_driver, because it is configured as 
 * callback function.
 * \param chip_config is the chip configuration.
 */
static void si3050_status_changed(struct si3050_driver_config* chip_config)
{
	zt_rxsig_t sig;

	PDEBUG("hooksig changed for chip %d: ", 
			chip_config->chip_number);
	
	
	/* report to zaptel */
	switch(chip_config->status) {
		case ON_HOOK:
			PDEBUG("ZT_RXSIG_ONHOOK\n");
			sig = ZT_RXSIG_ONHOOK;
			break;
		case OFF_HOOK:
			PDEBUG("ZT_RXSIG_OFFHOOK\n");
			sig = ZT_RXSIG_OFFHOOK;
			break;
		case RINGING:
			PDEBUG("ZT_RXSIG_RING\n");
			sig = ZT_RXSIG_RING;
			break;
		default:
			printk("Unknown chip state!!!\n");
			return;
	}

	zt_hooksig(&si3050_config.chan, sig);
}



/* 
 *
 * Initialization functions
 *
 */


/**
 * Register on zaptel.
 * \param config ...
 * \returns zero on success.
 */
static int si3050_zap_init(struct si3050_config* config)
{
	int ret;

	/* initialize span and chan config */
	memset(&config->span, 0, sizeof(config->span));
	memset(&config->chan, 0, sizeof(config->chan));

	/* set name and description */
	snprintf(config->span.name, sizeof(config->span.name), "si3050");
	snprintf(	config->span.desc, 
			sizeof(config->span.desc), 
			"Silicon Laboratories ProSLIC 3050");

	snprintf(config->chan.name, sizeof(config->chan.name), "si3050");

	config->chan.pvt	= config;
	config->chan.chanpos	= 1;
	config->chan.sigcap	= 	
					ZT_SIG_SF |
//					ZT_SIG_CLEAR |
					ZT_SIG_FXSKS| ZT_SIG_FXSLS;

	config->span.hooksig 	= si3050_zap_hooksig;
	config->span.open 	= si3050_zap_open;
	config->span.close 	= si3050_zap_close;
	config->span.ioctl 	= si3050_zap_ioctl;

	config->span.chans	= &config->chan;
	config->span.channels	= 1;
#ifdef SI3050_USE_ULAW
	config->span.deflaw	= ZT_LAW_MULAW;
#else
	config->span.deflaw	= ZT_LAW_ALAW;
#endif
	config->span.flags	= ZT_FLAG_RBS;
	config->span.pvt 	= config;

	INIT_WORK(&config->work_req_ec,  
		(void(*)(void*))hbx_dsp_request_ec, 
		&config->driver_config.ec_channel);

	INIT_WORK(&config->work_rel_ec,  
		(void(*)(void*))hbx_dsp_release_ec, 
		&config->driver_config.ec_channel);

	init_waitqueue_head(&config->span.maintq);
	
	/* register device on zaptel */
	ret = zt_register(&config->span,0);
	if (ret) {
		printk(KERN_ERR "Unable to register with zaptel\n");
		return -1;
	}
	
	return 0;
}


/**
 * Module init.
 */
static int si3050_init(void)
{
	int ret;

	PDEBUG("init\n");

	/* basic initialization of si3050 chips */
	ret = si3050_initialize(&si3050_config.driver_config);
	if (ret)
		return ret;
	
	/* 
	 * Register on zaptel;
	 * Zaptel is the interface between the low level driver and the
	 * PBX (asterisk)
	 */
	PDEBUG("Register chips on zaptel ...\n");
	ret = si3050_zap_init(&si3050_config);
	if (ret)
		return ret;

	return 0;
}


/**
 * Module exit
 */
static void si3050_exit(void)
{
	int ret;
	
	PDEBUG("exit\n");

	/* unregister zaptel device  */
	PDEBUG("Unregister on zaptel\n");
	ret = zt_unregister(&si3050_config.span);
	if (ret)
		printk(KERN_ERR "Unable to unregister with zaptel\n");
	
	/* power down chip */
//	si3050_poweroff();
	
	PDEBUG("Release all si3050 chips\n");
	si3050_release(&si3050_config.driver_config);
}

MODULE_DESCRIPTION("Zaptel driver for Si3050 chips");
MODULE_AUTHOR("Henry Margies (h.margies@maxina.de");
MODULE_LICENSE("GPL");

module_init(si3050_init);
module_exit(si3050_exit);
