/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *   			   Sven Kolberg  <s.kolberg@maxina.de>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * This code is responsible for:
 * 	- initialization of si3210 chips
 * 	- handle interrupts
 * 	- start/stop ringing
 *
 *
 * TODO:
 * 	- check timer handling for callback functions
 * 	- think about what happens if phone is off hook on calibration
 * 	- start pcm traffic only if needed?
 * 	- write release_chips to release all hold resources
 * 	- write release_chip to power off a single chip
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <asm/arch/irqs.h>

#if (defined SI3210_USE_LEDMAN) && (defined CONFIG_LEDMAN)
#include <linux/ledman.h>
#else
#define ledman_cmd(a,b)
#endif

#include "ixp425_spi.h"
#include "si3210_regs.h"
#include "si3210_driver.h"
#include "si3210_proc.h"

/* Use macro PDEBUG for printing debug messages */
#undef PDEBUG				// undef it, just in case
#ifdef SI3210_DEBUG
	#define PDEBUG(fmt, args...) 	\
		printk("[%s:%d] "fmt, __FILE__,__LINE__, ##args)
#else
	#define PDEBUG(fmt, args...)	{}
#endif

#define __unused __attribute__((__unused__))

/* Voltage settings */
#define SI3210_VBATH	26
#define SI3210_VBATL	0x0a
#define SI3210_ONHOOK_V	0x0a

/*
 * This struct is used to have lists of default register settings.
 */
struct si3210_reg_list
{
    unsigned char  reg;
    unsigned short val;
};

/* 
 * List of default register values for indirect registers.
 */
struct si3210_reg_list si3210_indirect_reg_list[] = {
	{0,   0x55C2}, /* DTMF_ROW_0_PEAK_REG */
	{1,   0x51E6}, /* DTMF_ROW_1_PEAK_REG */
	{2,   0x4B85}, /* DTMF_ROW_2_PEAK_REG */
	{3,   0x4937}, /* DTMF_ROW_3_PEAK_REG */
	{4,   0x3333}, /* DTMF_COL_PEAK_REG */
	{5,   0x0202}, /* DTMF_FWD_TWIST_THR_REG */
	{6,   0x0202}, /* DTMF_REV_TWIST_THR_REG */
	{7,   0x0198}, /* DTMF_ROW_RATIO_THR_REG */
	{8,   0x0198}, /* DTMF_COL_RATIO_THR_REG */
	{9,   0x0611}, /* DTMF_ROW_SEC_HARMO_THR_REG */
	{10,  0x0202}, /* DTMF_COL_SEC_HARMO_THR_REG */
	{11,  0x00E5}, /* DTMF_PWR_MIN_THR_REG */
	{12,  0x0A1C}, /* DTMF_HOT_LIMIT_THR_REG */
	{13,  0x6465}, /* OSC_1_FREQ_COEF_REG */
	{14,  0x0590}, /* OSC_1_AMP_REG */
	{15,  0x0000}, /* OSC_1_PHASE_REG */
	{16,  0x7870}, /* OSC_2_FREQ_COEF_REG */
	{17,  0x007D}, /* OSC_2_AMP_REG */
	{18,  0x0000}, /* OSC_2_PHASE_REG */
	/* set offset to 0V */
	{19,  0x0000}, /* RING_OSC_DC_OFFSET_REG */
	/* 
	 * Ringing frequency should be set to 25 HZ
	 *        RCO	= coeff * 2^15
	 *        coeff	= cos( 2 * PI * freq / 1000HZ)
	 *
	 * Set freq to 25HZ
	 *        coeff	= 0.9877
	 *        RCO	= 32364.571 = 0x7e6c
	 */
	{20,  0x7e6c}, /* RING_OSC_FREQ_COEF_REG */
	/*
	 * Ringing amplitude
	 *
	 * RNGX	= 1/4 * sqrt( (1-coeff)/(1+coeff) ) * 2^15 * Vpk / 96V
	 *
	 * Vpk should be 35
	 * RNGX	= 235.06 = 0xEB
	 */
//	{21,  0x00eb}, /* RING_OSC_AMP_REG */
	{21,  0x010c}, /* RING_OSC_AMP_REG@40V */
	{22,  0x0000}, /* RING_OSC_PHASE_REG */
	{23,  0x2000}, /* PULSE_METERING_OSC_RAMP_RATE_REG */
	{24,  0x2000}, /* PULSE_METERING_OSC_AMP_REG */
	{25,  0x0000}, /* PULSE_METERING_OSC_FREQ_COEF */
	{26,  0x4000}, /* RECV_DIG_GAIN_REG */
	{27,  0x4000}, /* XMIT_DIG_GAIN_REG */
	{28,  0x1000}, /* LOOP_CLOSURE_THR_REG */
	{29,  0x3600}, /* RING_TRIP_THR_REG */
	{30,  0x1000}, /* COM_MODE_MIN_THR_REG */
	{31,  0x0080}, /* COM_MODE_MAX_THR_REG */
	{32,  0x07c0}, /* POWER_ALARM_THR_Q1_Q2_REG */
	{33,  0x376F}, /* POWER_ALARM_THR_Q3_Q4_REG */
	{34,  0x1B80}, /* POWER_ALARM_THR_Q5_Q6_REG */
	{35,  0x8000}, /* LOOP_CLOSURE_FILTER_COEF_REG */
	{36,  0x8000}, /* RING_TRIP_FILTER_COEF_REG */
	{37,  0x8000}, /* THERMAL_LP_FILTER_POLE_Q1_Q2_REG */
	{38,  0x8000}, /* THERMAL_LP_FILTER_POLE_Q3_Q4_REG */
	{39,  0x8000}, /* THERMAL_LP_FILTER_POLE_Q5_Q6_REG */
	{40,  0x0000}, /* COM_MODE_BIAS_RINGING_REG */
	{41,  0x0C00}, /* DC_DC_MIN_VOLTAGE_REG */
	{43,  0x1000}, /* LOOP_CLOSURE_THR_LB_REG */
	{99,  0x01b4}, /* FSK_AMP_COEF_SPACE_REG */
	/*
	 * OSC = cos(2*Pi*freq/24000HZ) * 2^15
	 * freq for Space is 2300HZ
	 * freq for Mark is 1300HZ
	 * NOTE: Base frequency is only 24000HZ, if FSK is active
	 */
//	{100, 0x6d23}, /* FSK_FREQ_COEF_SPACE_REG (2300HZ) */
	{100, 0x6b60}, /* FSK_FREQ_COEF_SPACE_REG (2200HZ) */
	{101, 0x00e9}, /* FSK_AMP_COEF_MARK_REG */
	/* See notes for register 100 */
//	{102, 0x78a8}, /* FSK_FREQ_COEF_MARK_REG (1300HZ) */
	{102, 0x79c0}, /* FSK_FREQ_COEF_MARK_REG (1200HZ) */
	{103, 0x1110}, /* FSK_TRANS_PARAM_0_TO_1_REG */
	{104, 0x3c00}, /* FSK_TRANS_PARAM_1_TO_0_REG */
	{ 97, 0x0000}, /* TRASMIT_FILTER */
};

/*
 * List of default values for direct registers.
 */
struct si3210_reg_list si3210_direct_reg_list[] = {
	/* write pcm mode and clock slot assignment */
	{ 1,	0x00 },	// PCM Mode Select ( 1PCLK/Data, Tri-State bit 0)
	{ 2,	0x00 },	// PCM TX Clock Slot Low Byte (1 PCLK cycle/LSB)
	{ 3,	0x00 },	// PCM TX Clock Slot High Byte
	{ 4,	0x00 },	// PCM RX Clock Slot Low Byte (1 PCLK cycle/LSB)
	{ 5,	0x00 },	// PCM RX Clock Slot High Byte
	/* initialize voice path */
//	{ 8,	0x01 },	// Loopbacks (digital loopbackt)
	{ 8,	0x00 },	// Loopbacks (no loopback)
	/* -3.5dB on receive and transmit path */
	{ 9,	0x09 },	// Transmit and receive path gain and control (-3.3db)
//	{ 9,	0x0f },	// Transmit and receive path gain and control (Mute)
//	{ 9,	0x00 },	// Transmit and receive path gain and control
	 /* 220 Ohm + 820 Ohm || 115nF */
	{ 10,	0x0E },	// Initialization Two-wire impedance
	{ 11,	0x33 },	// Transhybrid Balance/Four-wire Return Loss
	/* initialize OSC1, OSC2, ringing, pulse metering */
	{ 32,	0x00 },	// Oper. Oscillator 1 Controltone generation
	{ 33,	0x00 },	// Oper. Oscillator 2 Controltone generation
	{ 34,	0x18 },	// Initialization Ringing Oscillator Control
	{ 35,	0x00 },	// Oper. Pulse Metering Oscillator Control
	{ 36,	0x00 },	// Initialization OSC1 Active Low Byte (125 s/LSB)
	{ 37,	0x00 },	// Initialization OSC1 Active High Byte (125 s/LSB)
	{ 38,	0x00 },	// Initialization OSC1 Inactive Low Byte (125 s/LSB)
	{ 39,	0x00 },	// Initialization OSC1 Inactive High Byte (125 s/LSB)
	{ 40,	0x00 },	// Initialization OSC2 Active Low Byte (125 s/LSB)
	{ 41,	0x00 },	// Initialization OSC2 Active High Byte (125 s/LSB)
	{ 42,	0x00 },	// Initialization OSC2 Inactive Low Byte (125 s/LSB)
	{ 43,	0x00 },	// Initialization OSC2 Inactive High Byte (125 s/LSB)
	{ 44,	0x00 },	// Initialization Pulse Metering Active Low Byte
	{ 45,	0x00 },	// Initialization Pulse Metering Active High Byte
	{ 46,	0x00 },	// Initialization Pulse Metering Inactive Low Byte
	{ 47,	0x00 },	// Initialization Pulse Metering Inactive High Byte
	/* Set active timer to 1sec -> 0x1f40 * 125usec = 1sec
	{ 48,	0x40 },	// Ringing Osc. Active Timer Low Byte 
	{ 49,	0x1f },	// Ringing Osc. Active Timer High Byte 
	 * Set inactive timer to 4sec -> 0x7d00 * 125usec = 4sec
	{ 50,	0x00 },	// Ringing Osc. Inactive Timer Low Byte 
	{ 51,	0x7d },	// Ringing Osc. Inactive Timer High Byte
	 * 
	 * NOTE: Zaptel is doing the ringing timing for us. So we
	 * should ring as long as possible. Zaptel will then send
	 * us a OFF-Hook signal to stop ringing and sending the
	 * caller id (after first ringing). Then it starts ringing
	 * again.
	 */
	{ 48,	0x80 },	// Ringing Osc. Active Timer Low Byte 
	{ 49,	0x3e },	// Ringing Osc. Active Timer High Byte 
	{ 50,	0x00 },	// Ringing Osc. Inactive Timer Low Byte 
	{ 51,	0x00 },	// Ringing Osc. Inactive Timer High Byte
	/* write detect thresholds and filters */
	{ 63,	0x54 },	// Ringing Mode Loop Closure Debounce Interval
	{ 67,	0x17 },	// Automatic/Manual Control
	{ 69,	0x0C },	// Loop Closure Debounce Interval
	{ 70,	0x0A },	// Ring Trip Debounce Interval (1.25 ms/LSB)
	/* write dc-feed paramters */
	{ 65,	0x61 },	// External Bipolar Transistor Settings
	{ 66,	0x0b },	// Battery Control
	{ 71,	0x01 },	// Off-Hook Loop Current Limit (20 mA + 3 mA/LSB)
	/* ON-HOOK LINE VOLTAGE */
	{ 72,	SI3210_ONHOOK_V }, // On-Hook Voltage (open circuit voltage) = 15V
	{ 73,	0x03 },	// Common Mode VoltageVCM = 3 V(1.5 V/LSB)
	/* set rest of direct registers as needed */
	{ 52,	0x00 },	// Normal Oper. FSK Data Bit
	/* VBATH and VBATL */
	{ 74,	SI3210_VBATH },	// Initialization VBATH = 37.5 V (1.5 V/LSB)
	{ 75,	SI3210_VBATL },	// Initialization VBATL = 15 V 
	//{ 108, 0x82} /* Current Limit Increase , DC-DC Converter Squelch enabled */
	{108, 0x22} /*DC-DC Converter Speedup and Noise removal */ 
};

/*
 *
 * Functions for start/stop ringing.
 *
 */

/**
 * Generates a ring on the specified chip.
 * \param chip the chip where it should start ringing [0-7]
 */
void si3210_start_ringing(int chip)
{
	/* automatic / manual setting, only set if you want to measure
	 * ringing voltages
	 * si321x_write_cmd(chip, 67, 0x19);
	 */
	
	/* enable ringing and active timer */
	si321x_write_cmd(chip, SI3210_RINGING_OSCIL_CTRL_REG, 0x14);
	/* Set the linefeed control register to ringing */
	si321x_write_cmd(chip, SI3210_LINEFEED_CTRL_REG, 0x4);
	
	/* flash leds to show activity */
	ledman_cmd(LEDMAN_CMD_FLASH, (chip)?7:6);
}

/**
 * Changes device to off-hook.
 * \param chip the chip where it should stop ringing [0-7]
 */
void si3210_go_offhook(int chip)
{
	/* disable ringing osc */
	si321x_write_cmd(chip, SI3210_RINGING_OSCIL_CTRL_REG, 0x0);

	/* Set the linefeed control register to Forward on-hook transm */
	si321x_write_cmd(chip, SI3210_LINEFEED_CTRL_REG, 0x2);
	
	/* flash leds to show activity */
	ledman_cmd(LEDMAN_CMD_FLASH, (chip)?7:6);
}


/**
 * Changes device to on-hook.
 * \param chip the chip where it should stop ringing [0-7]
 */
void si3210_go_onhook(int chip)
{
	/* disable ringing osc */
	si321x_write_cmd(chip, SI3210_RINGING_OSCIL_CTRL_REG, 0x0);

	/* Set the linefeed control register to active forward */
	si321x_write_cmd(chip, SI3210_LINEFEED_CTRL_REG, 0x1);
			
	/* turn led on, fxs ready but not active */
	ledman_cmd(LEDMAN_CMD_OFF, (chip)?7:6);
}


/**
 * Sets the polarity of the line.
 * \param chip is the chip where to change the polarity
 * \param polarity 1 if inverse polarity
 */
void si3210_set_polarity(int chip, int polarity)
{
	unsigned char p;

	if (polarity != 0 || polarity != 1)
		return;

	/* read values */
	p = si321x_read_cmd(chip, SI3210_ONHOOK_LINE_VOLTAGE_REG);

	/* only change the polarity */
	p = (p & 0x3f)+(polarity<<6);
	si321x_write_cmd(chip, SI3210_ONHOOK_LINE_VOLTAGE_REG, p);
}


/**
 * Starts pcm transfers on the given chip.
 * \param chip chipnumber [0-7]
 */
void si3210_start_pcm(int chip, int start_bit)
{
#ifdef SI3210_DEBUG
	char* freqs[ ] = {	"8192",	"4028",	"2048",	"1024",	"512",
				"256",	"1536",	"768",	"32768"};
#endif
	unsigned char v;

	/* set pcm offset for tx and rx */
	si321x_write_cmd(chip, SI3210_PCM_TX_START_COUNT_LB_REG, start_bit);
	si321x_write_cmd(chip, SI3210_PCM_TX_START_COUNT_HB_REG, 0);

	si321x_write_cmd(chip, SI3210_PCM_RX_START_COUNT_LB_REG, start_bit);
	si321x_write_cmd(chip, SI3210_PCM_RX_START_COUNT_HB_REG, 0);

	/* 
	 * enable PCM transfers 
	 * - 1 PCLK per data bit
	 * - 8-bit transfer
	 * - A-Law
	 * - Enable PCM transfers
	 */
#ifdef SI3210_USE_ULAW
	si321x_write_cmd(chip, SI3210_PCM_MODE_SELECT_REG, 0x28);
#else
	si321x_write_cmd(chip, SI3210_PCM_MODE_SELECT_REG, 0x20);
#endif
	
	v = si321x_read_cmd(chip, 13)>>4;
	PDEBUG("PCM for chip %d is active (Freq = %skHz, Startbit = %d)\n",
			chip,freqs[v], start_bit);
}


/**
 * Stops pcm transfers on the given chip.
 * \param chip chipnumber [0-7]
 */
void si3210_stop_pcm(int chip)
{
	si321x_write_cmd(chip, SI3210_PCM_MODE_SELECT_REG, 0x00);
}


#ifdef SI3210_DEBUG_VOLTAGES
static void si3210_show_voltages(int chip)
{
	int ret;
	int v, a;

	/* Loop voltage and current */
	ret = si321x_read_cmd(chip, SI3210_LOOP_VOLTAGE_SENSE_REG);
	PDEBUG("CHIP %d: reg 78 = %x\n", chip, ret);
	v = (ret & 63) * 1500 * ((ret&64)?-1:1);

	ret = si321x_read_cmd(chip, SI3210_LOOP_CURRENT_SENSE_REG);
	PDEBUG("CHIP %d: reg 79 = %x\n", chip, ret);
	a = (ret & 63) * 1250 * ((ret&64)?-1:1);
	
	printk("CHIP %d: Loop Sense = %dmV %duA\n", chip, v, a);

	
	/* tip voltage */	
	ret = si321x_read_cmd(chip, SI3210_TIP_VOLTAGE_SENSE_REG);
	PDEBUG("CHIP %d: reg 80 = %x\n", chip, ret);
	v = ret * 376;
	printk("CHIP %d: Tip Sense = %dmV\n", chip, v);
	
	/* ring voltage */	
	ret = si321x_read_cmd(chip, SI3210_RING_VOLTAGE_SENSE_REG);
	PDEBUG("CHIP %d: reg 81 = %x\n", chip, ret);
	v = ret * 376;
	printk("CHIP %d: Ring Sense = %dmV\n", chip, v);
	
	/* battery 1 voltage */	
	ret = si321x_read_cmd(chip, SI3210_BATTERY_VOLTAGE_SENSE_1_REG);
	PDEBUG("CHIP %d: reg 82 = %x\n", chip, ret);
	v = ret * 376;
	printk("CHIP %d: Battery 1 Sense = %dmV\n", chip, v);
	
	/* battery 2 voltage */	
	ret = si321x_read_cmd(chip, SI3210_BATTERY_VOLTAGE_SENSE_2_REG);
	PDEBUG("CHIP %d: reg 83 = %x\n", chip, ret);
	v = ret * 376;
	printk("CHIP %d: Battery 2 Sense = %dmV\n", chip, v);
}
#endif

/*
 *
 * Interrupt functions
 *
 */

/*
 * This timer is used to trigger a call to a function from within an 
 * interrupt. The function will be started right after the interrupt.
 */
static struct timer_list cb_timer;

/* Prototype of callback function */
typedef void (*si3210_cb_func)(void *);

/* 
 * Use this function to trigger the call of a function.
 * \param func is the callback function you want to start
 * \param data will be the first parameter of the callback function
 */
static void si3210_run_cb(si3210_cb_func func, void* data)
{
	if (!func)	/* no callback function, nothing to do */
		return;

	/* because of some reason.. no timer is used */
	func(data);
	return;
	
	/* XXX somebody should check, if the timer is free */

	init_timer(&cb_timer);
	cb_timer.data = (unsigned long)data;
	cb_timer.expires = jiffies;
	cb_timer.function = (void(*)(unsigned long))func;
	add_timer(&cb_timer);
}


/**
 * Interrupt service routine handling interrupts on status register 1.
 * Note: This functions runs in interrupt context!
 * \param c is the chip number [0-7].
 * \param config points to the driver configuration.
 */
static void si3210_isr_1(int c, struct si3210_driver_config* config __unused)
{
	unsigned char b;

	b = si321x_read_cmd(c, SI3210_INTERRUPT_STATUS_1_REG);
	
	if (!b) /* nothing happens here */
		return;
	
	/* clear all pending interrupts */
	si321x_write_cmd(c, SI3210_INTERRUPT_STATUS_1_REG,0xff);

	if (b & SI3210_PULSE_METERING_INACTIVE_TIMER_INT_PENDING)
		PDEBUG("Pulse Metering Inactive Timer on chip %d\n", c);
	
	if (b & SI3210_PULSE_METERING_ACTIVE_TIMER_INT_PENDING)
		PDEBUG("Pulse Metering Active Timer on chip %d\n", c);
	
	if (b & SI3210_RINGING_INACTIVE_TIMER_INT_PENDING)
		PDEBUG("Inactive timer on chip %d\n", c);

	if (b & SI3210_RINGING_ACTIVE_TIMER_INT_PENDING)
		PDEBUG("Active timer on chip %d\n", c);

	if (b & SI3210_OSC_2_INACTIVE_TIMER_INT_PENDING)
		PDEBUG("Oscillator 2 Inactive Timer on chip %d\n", c);
	
	if (b & SI3210_OSC_2_ACTIVE_TIMER_INT_PENDING)
		PDEBUG("Oscillator 2 Ative Timer on chip %d\n", c);
	
	if (b & SI3210_OSC_1_INACTIVE_TIMER_INT_PENDING)
		PDEBUG("Oscillator 1 Inactive Timer on chip %d\n", c);
	
	if (b & SI3210_OSC_1_ACTIVE_TIMER_INT_PENDING)
		PDEBUG("Oscillator 1 Active Timer on chip%d\n", c);
}


/**
 * Interrupt service routine handling interrupts on status register 2.
 * Note: This functions runs in interrupt context!
 * \param c is the chip number [0-7].
 * \param config points to the driver configuration.
 */
static void si3210_isr_2(int c, struct si3210_driver_config* config)
{
	unsigned char b;

	b = si321x_read_cmd(c, SI3210_INTERRUPT_STATUS_2_REG);
	
	if (!b) 
		return;

	/* clear all pending interrupts */
	si321x_write_cmd(c, SI3210_INTERRUPT_STATUS_2_REG,0xff);

	if (b & SI3210_POWER_ALARM_Q_6_INT_PENDING)
		printk("Power Alarm Q6 on chip %d\n", c);
	
	if (b & SI3210_POWER_ALARM_Q_5_INT_PENDING)
		printk("Power Alarm Q5 on chip %d\n", c);
	
	if (b & SI3210_POWER_ALARM_Q_4_INT_PENDING) {
		printk("Power Alarm Q4 on chip %d: ", c);
		PDEBUG("Current Sense on T4 is %iuA\n",
			376*si321x_read_cmd(c, 87)/10);
	}
	
	if (b & SI3210_POWER_ALARM_Q_3_INT_PENDING) {
		printk("Power Alarm Q3 on chip %d: ", c);
		PDEBUG("Current Sense on T3 is %iuA\n",
			376*si321x_read_cmd(c, 86)/10);
	}
	
	if (b & SI3210_POWER_ALARM_Q_2_INT_PENDING)
		printk("Power Alarm Q2 on chip %d\n", c);
	
	if (b & SI3210_POWER_ALARM_Q_1_INT_PENDING)
		printk("Power Alarm Q1 on chip %d\n", c);
	
	if (b & SI3210_RING_TRIP_INT_PENDING) {
		PDEBUG("Ring Trip on chip %d\n", c);
		if (si321x_read_cmd(c, 68) & 0x2)
			config->chip[c].status = OFF_HOOK;
	
		/* trigger start of cb function */
		si3210_run_cb(	(si3210_cb_func) config->status_changed, 
				&config->chip[c]);
	}

	if (b & SI3210_LOOP_CLOSURE_TRANS_INT_PENDING) {
		PDEBUG("Loop Closure Transition on chip %d\n", c);
		
		if (si321x_read_cmd(c, 68) & 0x1) {
			config->chip[c].status = OFF_HOOK;
			ledman_cmd(LEDMAN_CMD_FLASH, (c)?7:6);
		} else {
			config->chip[c].status = ON_HOOK;
			ledman_cmd(LEDMAN_CMD_OFF, (c)?7:6);
		}
	
		/* trigger start of cb function */
		si3210_run_cb(	(si3210_cb_func) config->status_changed, 
				&config->chip[c]);
	}

	if (b & SI3210_POWER_ALARM)
		panic("Power Alarm");
	
}


/**
 * Interrupt service routine handling interrupts on status register 3.
 * Note: This functions runs in interrupt context!
 * \param c is the chip number [0-7].
 * \param config points to the driver configuration.
 */
static void si3210_isr_3(int c, struct si3210_driver_config* config)
{
	unsigned char b;

	b = si321x_read_cmd(c, SI3210_INTERRUPT_STATUS_3_REG);
	
	if (!b) return;

	/* clear all pending interrupts */
	si321x_write_cmd(c, SI3210_INTERRUPT_STATUS_3_REG,0xff);

	if (b & SI3210_COM_MODE_CALIB_ERROR_INT_PENDING)
		PDEBUG("Common Mode Calibration Error on chip %d\n", c);
	
	if (b & SI3210_INDIRECT_REG_ACCESS_SERVICED_INT_PENDING)
		PDEBUG("Indirect Register Access Serviced on chip %d\n", c);
	
	/* DTMF digit detected? */
	if (b & SI3210_DTMF_DETECTED_INT_PENDING) {
		unsigned char v;
		char dtmf[] = "D1234567890*#ABC";
		PDEBUG("DTMF Tone Detected on chip %d\n", c);
		/* read detected digit */
		v = si321x_read_cmd(c, SI3210_DECODE_STATUS_REG);
		config->chip[c].dtmf = dtmf[v&0xf];
		/* call cb function */
		si3210_run_cb(	(si3210_cb_func) config->dtmf_detected, 
				&config->chip[c]);
	}
}


/**
 * Main interrupt service routine handling all chips and all interrupts.
 * \param irq Interrupt number.
 * \param data Points to private data, in our case to driver config struct.
 * \param regs Not important.
 */
static irqreturn_t si3210_isr(	int irq __unused, 
			void *data, 
			struct pt_regs* regs __unused)
{
	int i;
	struct si3210_driver_config * config = data;

	/* Tell spi, that we are in interrupt context. This is extremely
	 * important for spi functions to know, because it changes how to 
	 * lock and unlock access to spi.
	 */
	SPI_IRQ_Start();
	
	/* Check all status register for all chips */
	for (i=0; i<config->num_of_chips; i++) {
		si3210_isr_1(i, config);
		si3210_isr_2(i, config);
		si3210_isr_3(i, config);
	}

	SPI_IRQ_Stop();

	return IRQ_HANDLED;
}


/**
 * This function disables all interrupts for a single chip.
 * After calling this no interrupt will occur from this chip.
 * \param chip is the chip number [0-7]
 */
static void si3210_disable_interrupts(int chip)
{
	/* activate interrupts */
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_1_REG, 0x00); 
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_2_REG, 0x00); 
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_3_REG, 0x00); 
}


/**
 * This function enables all interrupts of a single chip.
 * It also clears all pending interrupts before.
 * \param chip is the chip number where the action should take place [0-7]
 */
static void si3210_enable_interrupts(int chip)
{
	/* clear all pending interrupts */
	si321x_write_cmd(chip, SI3210_INTERRUPT_STATUS_1_REG, 0xff); 
	si321x_write_cmd(chip, SI3210_INTERRUPT_STATUS_2_REG, 0xff); 
	si321x_write_cmd(chip, SI3210_INTERRUPT_STATUS_3_REG, 0xff); 
	
	/* activate interrupts */
#ifdef SI3210_DEBUG
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_1_REG, 0xff); 
#else
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_1_REG, 0x10); 
#endif
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_2_REG, 0xff); 
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_3_REG, 0x05); 
}


/*
 *
 * Functions for initialization of si3210 chips
 *
 */

/**
 * Enable power on a single chip.
 * \param chip The chip to be powered up [0-7]
 * \returns 0 upon success
 */
static int si3210_power_up(int chip) 
{
	unsigned long i;
	unsigned char v;
	
	/* Set the period of the DC-DC converter */
	si321x_write_cmd(chip,SI3210_DC_DC_CONVERTER_PWM_PERIOD_REG, 0xff);
	si321x_write_cmd(chip,SI3210_DC_DC_CONVERTER_SWITCHING_DELAY_REG, 0x19);
	
	//Turn DC-DC Converter on by clearing bit 4
	si321x_write_cmd(chip, SI3210_POWER_DOWN_CTRL_1_REG, 0x00);

	i = 0;
	do {
		v = si321x_read_cmd(chip, SI3210_BATTERY_VOLTAGE_SENSE_1_REG);
		mdelay(1);
		if (i++ > 1000) {
			printk("Error: Timeout in power calibration\n");
			return -1;
		}
		
//	} while(v < 0x7D); //47V
	} while(v < 0x14); //30V

	si321x_write_cmd(chip, SI3210_DC_DC_CONVERTER_SWITCHING_DELAY_REG,0x9d);

	i = 0;
	do {
		v = si321x_read_cmd(chip, 
				SI3210_DC_DC_CONVERTER_SWITCHING_DELAY_REG);
		if (i++ > 1000) {
			printk("Error: Timeout in calibration of dc-dc conv\n");
			return -1;
		}
	} while(v&0x80);
    
	return 0;
}


#ifdef SI3210_LEAK_TEST_ENABLE
/**
 * Tests the device for power leakage, by powering down the source, waiting 
 * 2 sec and reading the remaining voltage.
 * If the drop in voltage is to big, we may have a leakage
 * \param chip The chip to be tested [0-7]
 * \returns 0 upon succes
 */
static int si3210_power_leak_test(int chip) 
{
	unsigned char v_bat;
	
	/* Set linefeed state to open */
	si321x_write_cmd(chip, SI3210_LINEFEED_CTRL_REG, 0x00);
	/* Switch off DC-DC circuitry */
	si321x_write_cmd(chip, SI3210_POWER_DOWN_CTRL_1_REG, 0x10);
	/* Wait */
	mdelay(1000);
	
	/* Read the value of VBAT */
	v_bat = si321x_read_cmd(chip, SI3210_BATTERY_VOLTAGE_SENSE_1_REG);
	if (v_bat < 0x04)
	{
	        printk("Power leak test failure VBAT = 0x%x\n", v_bat);
	        return -1;
	}
	
	return 0;
}
#endif // SI3210_LEAK_TEST_ENABLE


/**
 * Writes default values to indirect registers.
 * \param chip the number of the chip to be initialized [0-7]
 */
static void si3210_initialize_indirect_regs(int chip)
{
	int i;
	int entries;

	entries = sizeof(si3210_indirect_reg_list) 
		/ sizeof(struct si3210_reg_list);
	
	for(i=0; i<entries; i++) {
		/* write value to the corresponding indirect register */
		si321x_write_indirect(	chip, 
					si3210_indirect_reg_list[i].reg, 
					si3210_indirect_reg_list[i].val);
	}
}


/**
 * Writes default values to direct registers.
 * \param chip the number of the chip to be initialized [0-7]
 */
static void si3210_initialize_direct_regs(int chip)
{
	int i;
	int entries;

	entries = sizeof(si3210_direct_reg_list) 
		/ sizeof(struct si3210_reg_list);
	
	for(i=0; i<entries; i++) {
		/* write value to the corresponding direct register */
		si321x_write_cmd(	chip, 
					si3210_direct_reg_list[i].reg, 
					si3210_direct_reg_list[i].val);

	}
}


/**
 * Perfoms calibration of voltage etc.
 * \param chip The chip where calibration should bei performed. [0-7]
 * \returns zero on success.
 */
static int si3210_calibrate(int chip)
{
	int i;
	int v;

	/* set slic to open state */
	si321x_write_cmd(chip, SI3210_LINEFEED_CTRL_REG, 0);

	/* perform slic calibration */
	si321x_write_cmd(chip, SI3210_CALIB_CONTROL_STATUS_2_REG, 0x1e);
	si321x_write_cmd(chip, SI3210_CALIB_CONTROL_STATUS_1_REG, 0x47);

	i = 0;
	do {
		v = si321x_read_cmd(chip, SI3210_CALIB_CONTROL_STATUS_1_REG);
		mdelay(500); /* this value is ok as it is [HM] */
		if (i++ > 10 ) /* 5 sec */ {
			printk("Timeout\n");
			return -1;
		}

	} while(v);

	/* manual ring gain mismatch calibration */
	for (i=0x1f; i>0; i--) {

		si321x_write_cmd(chip, 
				SI3210_RING_GAIN_MISMATCH_CALIB_RESULT_REG, i);
		mdelay(40); /* wait 40ms */

		v = si321x_read_cmd(chip,SI3210_TRANSISTOR_5_CURRENT_SENSE_REG);
		if (v == 0)
			break;
	}

	/* manual tip gain mismatch calibration */
	for (i=0x1f; i>0; i--) {

		si321x_write_cmd(chip, 
				SI3210_TIP_GAIN_MISMATCH_CALIB_RESULT_REG, i);
		mdelay(40); /* wait 40ms */

		v = si321x_read_cmd(chip,SI3210_TRANSISTOR_6_CURRENT_SENSE_REG);
		if (v == 0)
			break;
	}

	/* turn on interrupt for calibration error */
	si321x_write_cmd(chip, SI3210_INTERRUPT_ENABLE_3_REG, 4);

	/* perform slic calibration (again) */
	/* FIXME: make sure we are onhook */
	si321x_write_cmd(chip, SI3210_CALIB_CONTROL_STATUS_2_REG, 0x01);
	si321x_write_cmd(chip, SI3210_CALIB_CONTROL_STATUS_1_REG, 0x40);
	
	i = 0;
	do {
		v = si321x_read_cmd(chip, SI3210_CALIB_CONTROL_STATUS_1_REG);
		mdelay(500); /* this value is ok!! [HM]*/
		if (i++ > 10) {
			printk("Timeout\n");
			return -1;
		}

	} while(v);

	/* flush out energy accumulators */
	for (i=88; i<=95; i++)
		si321x_write_indirect(chip, i, 0);
	
	si321x_write_indirect(chip, 97, 0);

	for (i=193; i<=211; i++)
		si321x_write_indirect(chip, i, 0);

	return 0;
}


/**
 * Initializes a single chip.
 * \param chip specifies the chip to be initialized [0-7]
 * \returns zero on success
 */
int si3210_initialize_chip(int chip)
{
	int ret;
#ifdef CONFIG_LEDMAN
	int led = (chip)?7:6;
#endif

	PDEBUG("Activating chip %d ...\n", chip);
	ledman_cmd(LEDMAN_CMD_ON, led);

	si3210_disable_interrupts(chip);

	si3210_initialize_indirect_regs(chip);

	/* clear out digital loopbacks */
	si321x_write_cmd(chip, SI3210_AUDIO_PATH_LOOPBACK_REG, 0x0);
	/* set Rev.C and above enhancements */
	si321x_write_cmd(chip, SI3210_ENHANCEMENT_ENABLE_REG, 0xeb);
	/* set high and low battery voltage */
	si321x_write_cmd(chip, SI3210_HIGH_BATTERY_VOLTAGE_REG,	SI3210_VBATH);
	si321x_write_cmd(chip, SI3210_LOW_BATTERY_VOLTAGE_REG,  SI3210_VBATL);
	/* make vbat switch not automatic */
	si321x_write_cmd(chip, SI3210_AUTO_MANUAL_CONTROL_REG, 0x17);
	/* Q7 should be set to OFF for si3210 */
	si321x_write_cmd(chip, SI3210_BATTERY_FEED_CTRL_REG, 0x1);

	ret = si3210_power_up(chip);
	if (ret) {
		ledman_cmd(LEDMAN_CMD_OFF, led);
		return ret;
	}
#ifdef SI3210_LEAK_TEST_ENABLE
	si3210_power_leak_test(chip);
	si3210_power_up(chip);
#endif // SI3210_LEAK_TEST_ENABLE

	ret = si3210_calibrate(chip);
	if (ret) {
		ledman_cmd(LEDMAN_CMD_OFF, led);
		return ret;
	}

	si3210_initialize_direct_regs(chip);

	/* writing Power Thresholds 35-39 */
	si321x_write_indirect(chip,37,0x008c);
	si321x_write_indirect(chip,38,0x008c);
	si321x_write_indirect(chip,39,0x0010);
	
	/* go active */
	si321x_write_cmd(chip, SI3210_LINEFEED_CTRL_REG, 0x1);
	
	si3210_enable_interrupts(chip);
	ledman_cmd(LEDMAN_CMD_OFF, led);

	return 0;
}


/**
 * Turns off a single chip.
 * \param chip specifies the chip to be initialized [0-7]
 */
void si3210_poweroff_chip(int chip)
{
	/* stop pcm transfer */
	si3210_stop_pcm(chip);

	/* disable interrupts */
	si3210_disable_interrupts(chip);

	/* go to open linefeed state */
	si321x_write_cmd(chip, SI3210_LINEFEED_CTRL_REG, 0x0);
	
	/* turn all off */
	si321x_write_cmd(chip, SI3210_POWER_DOWN_CTRL_1_REG, 0x1b);
	si321x_write_cmd(chip, SI3210_POWER_DOWN_CTRL_2_REG, 0x2a);
	
	ledman_cmd(LEDMAN_CMD_OFF, (chip)?7:6);
}


/**
 * The timerlist for the si3210 watchdog
 */
static struct timer_list si3210_watchdog_timer;

#define SI3210_WATCHDOG_INTERVAL (10 * HZ)

/**
 * The Watchdog interrupt
 * read some registers to make sure we have contact
 * if not -> AEIY panic 
 */
static void si3210_watchdog_timer_cb(unsigned long arg)
{
	
	static char spi_mode_select_reg[2];
	static char pcm_mode_select_reg[2];
	static int firsttime = 1;
	
	int i;
	
	if (firsttime) 
	{
		/**
		 * remember the current setting of the spi and pcm mode register
		 */
		for(i=0;i<2;i++) 
		{
			spi_mode_select_reg[i] = si321x_read_cmd(i, SI3210_SPI_MODE_SELECT_REG);
			pcm_mode_select_reg[i] =  si321x_read_cmd(i, 108);
		}

		firsttime = 0;
	}
	else
	{
		/**
		 * compare the current setting with the one remembered
		 */
		for(i=0;i<2;i++)
		{
			if (spi_mode_select_reg[i] != si321x_read_cmd(i, SI3210_SPI_MODE_SELECT_REG))
				panic("PANIC spi mode register different");
			if (pcm_mode_select_reg[i] !=  si321x_read_cmd(i, 108))
				panic("PANIC pcm mode register different");
		}
	}


	/* Re.arm timer interrupt */
	si3210_watchdog_timer.expires = jiffies + SI3210_WATCHDOG_INTERVAL;
	add_timer(&si3210_watchdog_timer);
}

/**
 * init a timer which does a read to the si3210 wich are in the config
 */
static int init_si3210_watchdog(void)
{
	PDEBUG("Arming Watchdog\n");
	/* setup timer for watchdog */
	init_timer(&si3210_watchdog_timer);
	si3210_watchdog_timer.function = si3210_watchdog_timer_cb;
	si3210_watchdog_timer.data = 0;
	si3210_watchdog_timer.expires = jiffies + SI3210_WATCHDOG_INTERVAL;
	add_timer(&si3210_watchdog_timer);
	
	return 0;
}
	
/**
 * deinit
 */
static int deinit_si3210_watchdog(void)
{
	del_timer(&si3210_watchdog_timer);	
}

/**
 * This function performs a basic setup of all si3210 chips.
 * Call ixp425_hss_init and ixp425_spi_init before. The driver
 * configuration will be initialized, so setting values before 
 * makes no sense.
 * \param config points to the driver configuration structure.
 * \returns zero on success
 */
int si3210_initialize_chips(struct si3210_driver_config* config)
{
	int i, v, ret;
	char* names[] = { "Si3210", "Si3211", "Unused", "Si3210M" };

	memset(	config, 0, sizeof(struct si3210_driver_config));

	for (i=0; i<8; i++) {
		v = si321x_read_cmd(i, SI3210_SPI_MODE_SELECT_REG);	

		if (v == 0xff)
			break;

		if (!v&0x80) {
			printk("Error: Chip is not in daisy chain mode\n");
			break;
		}

		if ( ((v>>4)&0x3) == 2 ) {
			printk("Error: Chip has invalid Part Number\n");
			break;
		}

		if (!v&0xf) {
			printk("Error, Chip has invalid Revision Number\n");
			break;
		}
		
		printk("Found chip(%d) %s Revision %c\n", i, names[(v>>4)&3], 
							'A'+(v&0xf)-1 );
		si3210_disable_interrupts(i);

		config->chip[i].status = ON_HOOK;
		config->chip[i].chip_number = i;
	}

	config->num_of_chips = i;

	//Request an Interrupt
	ret = request_irq(	IRQ_IXP4XX_GPIO11, si3210_isr,
				SA_SHIRQ | SA_TRIGGER_LOW, 
				"si3210_irq", (void*)config);

	if (ret) {
		printk("Error while requesting interrupt for si3210\n");
		return -1;
	}

	ret = 0;
	si3210_proc_register();
	for (i=0; i<config->num_of_chips; i++) {
		ret += si3210_initialize_chip(i);
		si3210_proc_chip_register(i);
	}


#ifdef SI3210_DEBUG_VOLTAGES
	mdelay(2000);
	for (i=0; i<config->num_of_chips; i++)
		si3210_show_voltages(i);
#endif
	
	/* Start the watchdog */
	init_si3210_watchdog();

	return ret;
}


/* This function __should__ power off all active chips and release all 
 * resources.
 * \param config points to the driver configuration structure.
 */
void si3210_release_chips(struct si3210_driver_config* config)
{
	int i;

	for (i=0; i<config->num_of_chips; i++)
		si3210_proc_chip_unregister(i);
	si3210_proc_unregister();

	/* remove interrupt */
	free_irq(IRQ_IXP4XX_GPIO11, (void*)config);
	deinit_si3210_watchdog();

}
