/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *   			   Sven Kolberg  <s.kolberg@maxina.de>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define __unused __attribute__((__unused__))
#include <linux/proc_fs.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>

#include "ixp425_spi.h"

/* Use macro PDEBUG for printing debug messages */
#undef PDEBUG				// undef it, just in case
#ifdef SI3210_DEBUG
	#define PDEBUG(fmt, args...) 	\
		printk("[%s:%d] "fmt, __FILE__,__LINE__, ##args)
#else
	#define PDEBUG(fmt, args...)	{}
#endif

/*  proc dir for si3210 */
static struct proc_dir_entry* si3210_proc_root_dir;
static struct proc_dir_entry* si3210_proc_chip_dir[8];
static struct proc_dir_entry* si3210_proc_regs_dir[8];

#define IS_STRANGE	0
#define IS_ON		1
#define IS_OFF		2

static int is_on(const char* buffer, unsigned long count)
{
	char buf[4];

	if (count > 4) {
		return -EFAULT;
	}

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) )
		return -EFAULT;

	if (!strncmp("on", buf,2))
		return IS_ON;
	else if (!strncmp("off", buf,3))
		return IS_OFF;

	return IS_STRANGE;
}



/*
 * mutes receive and transmit path 
 */
static int si3210_proc_read_mute(char *page, char **start __unused,
			    off_t off __unused, int count __unused,
			    int *eof __unused, void *data)
{
	int chip = (int)data;
	int val  = si321x_read_cmd(chip, 9);

	if (val & 0x30)
		return sprintf(page, "on\n");
	return sprintf(page, "off\n");
}

static int si3210_proc_write_mute(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	int chip = (int)data;

	int val  = si321x_read_cmd(chip, 9);
	int ret  = is_on(buffer, count);

	switch(ret) {
		case IS_ON:
			si321x_write_cmd(chip, 9, val | 0x30);
			break;
		case IS_OFF:
			si321x_write_cmd(chip, 9, val & (~0x30));
			break;
		case IS_STRANGE:
			/* do nothing */
		default:
			return ret;
	}
	return count;
}

/*
 * starts and stops pcm
 */
static int si3210_proc_read_pcm(char *page, char **start __unused,
			    off_t off __unused, int count __unused,
			    int *eof __unused, void *data)
{
	int chip = (int)data;
	int val  = si321x_read_cmd(chip, 1);

	if (val & 0x20)
		return sprintf(page, "on\n");
	return sprintf(page, "off\n");
}

static int si3210_proc_write_pcm(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	int chip = (int)data;
	int val  = si321x_read_cmd(chip, 1);
	int ret  = is_on(buffer, count);

	switch(ret) {
		case IS_ON:
			si321x_write_cmd(chip, 1, val | 0x20);
			break;
		case IS_OFF:
			si321x_write_cmd(chip, 1, val & (~0x20));
			break;
		case IS_STRANGE:
			/* do nothing */
		default:
			return ret;
	}
	return count;

}


/*
 * set and show registers
 */
static int si3210_proc_read_reg(char *page, char **start __unused,
			    off_t off __unused, int count __unused,
			    int *eof __unused, void *data)
{
	int chip = ((int)data) & 0xff;
	int reg  = ((int)data) >> 8;
	int val  = si321x_read_cmd(chip, reg);

	return sprintf(page, "0x%x\n", val);
}

static int si3210_proc_write_reg(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	char buf[128];
	int chip = (int)data;
	int reg  = ((int)data) >> 8;
	int val;
	
	if (count > sizeof(buf)) {
		return -EFAULT;
	}

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) )
		return -EFAULT;

	val = simple_strtol(buf, NULL, 0);

	si321x_write_cmd(chip, reg, val);

	return count;
}


/*
 * generic functions for registering and unregistering proc entries 
 */

#define si3210_create_proc_entry(name, root)			\
{								\
	struct proc_dir_entry* e;				\
	e = create_proc_entry(__stringify(name), 0444, root);	\
	if (e) {						\
		e->read_proc 	= si3210_proc_read_##name;	\
		e->write_proc	= si3210_proc_write_##name;	\
		e->data		= (void*)chip;			\
	}							\
}								


void si3210_proc_chip_register(int chip)
{
	struct proc_dir_entry *root, *card, *regs;
	char name[128];
	int i;

	root = si3210_proc_root_dir;
	snprintf(name, sizeof(name), "%d", chip);

	PDEBUG("Registering proc files for si3210/%d\n", chip);

	if (root == NULL) {
		printk("Error creating proc directory (si3210)\n");
		return;
	}

	/* register proc directory */
	card = proc_mkdir(name, root);
	if (card == NULL) {
		printk("Error creating proc directory (si3210/%s)\n", name);
		return;
	}
	si3210_proc_chip_dir[chip] = card;


	regs = proc_mkdir("register", card);
	if (regs == NULL) {
		printk("Error creating proc directory (si3210/%s/register)\n", name);
		return;
	}
	si3210_proc_regs_dir[chip] = regs;

	si3210_create_proc_entry(mute,	card);
	si3210_create_proc_entry(pcm,	card);

	for (i=0; i<108;i++) {
		struct proc_dir_entry* e;

		snprintf(name, sizeof(name), "%d", i);
	
		e = create_proc_entry(name , 0444, regs);
		if (e) {
			e->read_proc 	= si3210_proc_read_reg;
			e->write_proc	= si3210_proc_write_reg;
			e->data		= (void*)(chip | (i<<8));
		}
	}
	
}

void si3210_proc_chip_unregister(int chip)
{
	char name[128];
	int i;
	
	/* remove proc files */
	PDEBUG("Unregister proc entries for chip %d\n", chip);

	for (i=0; i<108;i++) {
		snprintf(name, sizeof(name), "%d", i);
	
		remove_proc_entry(name,	si3210_proc_regs_dir[chip]);
	}

	remove_proc_entry("pcm", 	si3210_proc_chip_dir[chip]);
	remove_proc_entry("mute", 	si3210_proc_chip_dir[chip]);

	/* remove proc dirs */
	snprintf(name, sizeof(name), "%d", chip);
	remove_proc_entry("register", 	si3210_proc_chip_dir[chip]);
	remove_proc_entry(name, 	si3210_proc_root_dir);
}


void si3210_proc_register(void)
{
	PDEBUG("Register proc directory\n");
	si3210_proc_root_dir = proc_mkdir("si3210", NULL);

}


void si3210_proc_unregister(void)
{
	PDEBUG("Unregister proc directory\n");
	remove_proc_entry("si3210", NULL);
}

