/***************************************************************************
 *   Copyright (C) 2005 by Henry Margies <h.margies@maxina.de>		   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
 * This is the main source code for the si3210 zaptel driver.
 * It uses:
 * 	ixp425_spi.h:		for activating the SPI bus on ixp425 boards
 * 	ixp425_hss.h:		for handling the hss(pcm) bus on ixp425 boards
 * 	si3210_driver.h:	for driver initialization, interrupt handling
 * 				and all chipset specific things
 * The main tasks for the code in this file are:
 * 	- trigger initialization of the hardware
 * 	- register on zaptel
 * 	- handle events between zaptel and si3210
 * 	- transfer voice data
 *
 * TODO:
 * 	- write si3210_exit
 * 	- write si3210_zap_close
 * 	- try to avoid memcpy
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/delay.h>


#include <linux/zaptel.h>

#include "ixp425_hss.h"
#include "ixp425_spi.h"
#include "si3210_driver.h"
#include "hbx_dsp.h"

/* Use macro PDEBUG for printing debug messages */
#undef PDEBUG             // undef it, just in case
#ifdef SI3210_DEBUG
	#define PDEBUG(fmt, args...) 	\
		printk("[%s:%d] "fmt, __FILE__,__LINE__, ##args)
#else
	#define PDEBUG(fmt, args...) {}
#endif

#define __unused __attribute__((__unused__))

static struct si3210_config {
	struct si3210_driver_config 	driver_config;

	struct zt_span span;
	struct zt_chan chans[SI3210_MAX_CHIPS];
	
	struct work_struct  work_req_ec[SI3210_MAX_CHIPS];
	struct work_struct  work_rel_ec[SI3210_MAX_CHIPS];

} si3210_config;

/*
 *
 * Functions for transfering voice data between hss and zaptel.
 *
 */


/**
 * This function tranfers the data between hss bus and zaptel.
 */
static void si3210_data_transfer_cb(	unsigned char* rx_ptr,
					unsigned char* tx_ptr,
					void* data)
{
	struct zt_chan* chans;
	unsigned channel = (unsigned) data;

	chans 	= si3210_config.chans;

	if (0 == channel)
		zt_transmit(chans[0].span);

	/* try to not use memcpy */
	#ifdef SI3210_ZAP_USE_MEMCPY
	memcpy(	chans[channel].readchunk,
		rx_ptr,
		ZT_CHUNKSIZE);
	
	#else
		chans[channel].readchunk = rx_ptr;
	#endif
			
	memcpy(	tx_ptr,
		chans[channel].writechunk,
		ZT_CHUNKSIZE);

//	/* loopback */
//	memcpy( tx_ptr,
//		rx_ptr,
//		ZT_CHUNKSIZE);
//			
//	memcpy(	chans[channel].readchunk,
//		chans[channel].writechunk,
//		ZT_CHUNKSIZE);
			
	/* echo cancel if needed */
	if (chans[channel].ec) {
		int ec_ch = si3210_config.driver_config.chip[channel].ec_channel;

		if (ec_ch >=0) {

			hbx_dsp_put_ec(ec_ch, tx_ptr, chans[channel].readchunk);
			hbx_dsp_get_ec(ec_ch, tx_ptr, chans[channel].readchunk);

		} else {
			zt_ec_chunk(	&chans[channel],
					chans[channel].readchunk, 
					chans[channel].writechunk);
		}
	}

	/* FIXME: 1 is the last channel */
	if (1 == channel)
		zt_receive(chans[0].span);

}



#ifdef SI3210_ZAPTEL_LOOPBACK
/**
 * This is a loop function to test:
 * fxs driver -> zaptel -> asterisk -> zaptel -> fxs
 */
static void si3210_loopback_transfer_cb(unsigned char* rx_ptr,
					unsigned char* tx_ptr,
					void* data)
{
	struct zt_chan* chans;
	unsigned channel = (unsigned) data;

	static int rx_counter[8] = {0};
	static int tx_counter[8] = {0};
	static int complete[8]   = {0};
	static int started[8]    = {0};
	static unsigned last_j   = 0;
	int i;

	chans 	= si3210_config.chans;

	if (0 == channel)
		zt_transmit(chans[0].span);

	for (i=0;i<ZT_CHUNKSIZE; i++) {
		chans[channel].readchunk[i] = tx_counter[channel];

		if (++tx_counter[channel] > 255)
			tx_counter[channel] = 0;
	}
	
	for (i=0;i<ZT_CHUNKSIZE; i++) {

		unsigned char c = chans[channel].writechunk[i];
	
		if (!started[channel]) {
			if (c == 0) {
				started[channel]=1;
				rx_counter[channel]=1;
			}
			continue;
		}

		if (c == rx_counter[channel]) {
			if (++rx_counter[channel] > 255) {
				rx_counter[channel] = 0;
				if (complete[channel]++> 100) {
					complete[channel]=0;
					printk("Complete on channel %d\n", channel);
				}
			}
			continue;
		}
		
		started[channel]=0;

		if (rx_counter[channel] < 10)
			continue;

		printk("Test failed on channel %d: received %d expected %d\n", channel, c, rx_counter[channel]);
		for (i=0;i<ZT_CHUNKSIZE; i++) {
			unsigned char c = chans[channel].writechunk[i];
			printk("%03d ", c);
		}
		printk("\n");
		
	}

	if (last_j <= jiffies) {
		last_j = jiffies + 100;
		for (i=0;i<ZT_CHUNKSIZE; i++) {
			unsigned char c = chans[channel].writechunk[i];
			printk("%03d ", c);
		}
		printk(" (%d)\n", channel);
	}
	
//	memcpy(	chans[channel].readchunk,
//		chans[channel].writechunk,
//		ZT_CHUNKSIZE);
			
	/* FIXME: 1 is the last channel */
	if (1 == channel)
		zt_receive(chans[0].span);

}
#endif



/*
 *
 * callback functions for zaptel
 *
 */

/**
 * Opens a channel.
 * \param chan Zaptel channel description.
 * \returns zero on success
 */
static int si3210_zap_open(struct zt_chan *chan) 
{ 
	struct si3210_driver_config*	drv_cfg;
	struct si3210_chip_config*	chip_cfg;
	int ret;

	drv_cfg	= &((struct si3210_config*)chan->pvt)->driver_config;
	chip_cfg= &drv_cfg->chip[chan->chanpos-1];

	if (chip_cfg->in_use)
		return -EBUSY;

	PDEBUG("si3210_zap_open (%s)\n", chan->name);
	chip_cfg->in_use 	=  1;
	chip_cfg->ec_channel 	= -1;

#ifdef SI3210_ZAPTEL_LOOPBACK
	ret = ixp425_hss_register(	si3210_loopback_transfer_cb, 
					(void*)(unsigned)chan->chanpos-1);
#else
	ret = ixp425_hss_register(	si3210_data_transfer_cb, 
					(void*)(unsigned)chan->chanpos-1);
#endif
	
	if (ret < 0){
		printk("Unable to start hss/pcm transfers\n");
		return -EIO;
	}
	chip_cfg->hss_start_bit = ret;
	si3210_start_pcm(chan->chanpos-1, ret);
	
	return 0;
}


/**
 * Closes a channel.
 * \param chan Zaptel channel description
 * \returns zero on success.
 */
static int si3210_zap_close(struct zt_chan *chan) 
{

	struct si3210_driver_config*	drv_cfg;
	struct si3210_chip_config*	chip_cfg;
	int chip;

	drv_cfg	= &((struct si3210_config*)chan->pvt)->driver_config;
	chip = chan->chanpos-1;
	chip_cfg= &drv_cfg->chip[chip];

	/* release echo canceller if in use */
	if (chip_cfg->ec_channel >= 0) 
		schedule_work(&si3210_config.work_rel_ec[chip]);
	
	flush_scheduled_work();

	/* check if chan device was really opened!! */
	if (!chip_cfg->in_use)
		return 0;

	PDEBUG("si3210_zap_close (%s)\n", chan->name);
	ixp425_hss_unregister(chip_cfg->hss_start_bit);
	chip_cfg->in_use     =  0;
	chip_cfg->ec_channel = -1;
	
	return 0;
}


/**
 * Ioctl function for our zaptel channels.
 * \param chan Zaptel channel description
 * \param cmd to handle.
 * \param data for handling the command.
 * \returns zero on success.
 */
static int si3210_zap_ioctl(	struct zt_chan *chan __unused, 
				unsigned int cmd, 
				unsigned long data)
{
	PDEBUG("si3210_zap_ioctl (%s)\n", chan->name);

	switch(cmd) {
		case ZT_TONEDETECT:
#ifdef SI3210_DTMF_DETECT
		{
			unsigned int i = 0;

			PDEBUG("ZT_TONEDETECT: \n");
			get_user(i, (int*)data);
			if (i & ZT_TONEDETECT_ON)
				PDEBUG("\tZT_TONEDETECT_ON\n");
			if (i & ZT_TONEDETECT_MUTE)
				PDEBUG("\tZT_TONEDETECT_MUTE\n");
		}
#else
			return -ENOTTY;
#endif
			break;

#ifdef ZT_SETPOLARITY
		case ZT_SETPOLARITY:
		{
			unsigned int i = 0;

			PDEBUG("ZT_SETPOLARITY %d\n", i);
			get_user(i, (int*)data);
			si3210_set_polarity(chan->chanpos-1, i);
			break;
		}
#endif

		default:
			PDEBUG("cmd = %x, data = %x\n", cmd, (unsigned)data);
			return -ENOTTY;
	}
	return 0;
}


/**
 * This function is used by zaptel to tell us to change to off hook or start
 * ringing.
 * \param chan Zaptel channel description
 * \param txsig is the new status we should change to
 * \returns 0 on success.
 */
static int si3210_zap_hooksig(struct zt_chan *chan, zt_txsig_t txsig)
{
	struct si3210_driver_config* 	drv_cfg;
	struct si3210_chip_config*	chp_cfg;
	int chip;
	
	drv_cfg	= &((struct si3210_config*)chan->pvt)->driver_config;

	chip	= chan->chanpos-1;
	chp_cfg	= &drv_cfg->chip[chip];

	PDEBUG("hooksig from zaptel for chip %s: ", 
			chan->name);

	switch (txsig) {
		/* On hook */
		case ZT_TXSIG_ONHOOK:
			PDEBUG("ZT_TXSIG_ONHOOK\n");
			si3210_go_onhook(chip);
			if (chp_cfg->ec_channel >= 0)
				schedule_work(&si3210_config.work_rel_ec[chip]);
			break;

		/* Off hook */
		case ZT_TXSIG_OFFHOOK:
			PDEBUG("ZT_TXSIG_OFFHOOK\n");
			if (chp_cfg->ec_channel < 0)
				schedule_work(&si3210_config.work_req_ec[chip]);
			si3210_go_offhook(chip);
			break;

		/* Start / Ring */
		case ZT_TXSIG_START:
			PDEBUG("ZT_TXSIG_START\n");
			si3210_start_ringing(chip);
			break;

		/* Drop battery if possible */
		case ZT_TXSIG_KEWL:
			PDEBUG("ZT_TXSIG_KEWL\n");
			printk("KEWL signal not yet supported\n");
			break;

		default:
			printk("unkown hooksig\n");

	}
	
	return 0;
}

/*
 *
 * Callback function for si3210_driver
 *
 */

/**
 * This function reports detected dtmf digits to zaptel.
 * It is configured as callback function and gets called by
 * si3210_driver.
 * \param chip_config is the chip configuration.
 */
static void si3210_dtmf_detected(struct si3210_chip_config* chip_config)
{
#ifdef SI3210_DTMF_DETECT
	PDEBUG("dtmf = %c\n", chip_config->dtmf);
	zt_qevent_lock(	&si3210_config.chans[chip_config->chip_number], 
			(ZT_EVENT_DTMFUP | chip_config->dtmf));
#endif
}


/**
 * This function reports changed status to zaptel.
 * Changing status means from/to on hook, off hook or ringing. This
 * function gets call by si3210_driver, because it is configured as 
 * callback function.
 * \param chip_config is the chip configuration.
 */
static void si3210_status_changed(struct si3210_chip_config* chip_config)
{
	zt_rxsig_t sig;
	int	   chip = chip_config->chip_number;

	PDEBUG("hooksig changed for chip %d: ", chip);
	
	
	/* report to zaptel */
	switch(chip_config->status) {
		case ON_HOOK:
			PDEBUG("ZT_RXSIG_ONHOOK\n");
			sig = ZT_RXSIG_ONHOOK;
			break;
		case OFF_HOOK:
			PDEBUG("ZT_RXSIG_OFFHOOK\n");
			sig = ZT_RXSIG_OFFHOOK;
			break;
		default:
			printk("Unknown chip state!!!\n");
			return;
	}
	
	zt_hooksig(&si3210_config.chans[chip], sig);
}



/* 
 *
 * Initialization functions
 *
 */


/**
 * Register on zaptel.
 * \param config ...
 * \returns zero on success.
 */
static int si3210_zap_init(struct si3210_config* config)
{
	int i, ret;

	/* initialize span and chan config */
	memset(&config->span, 0, sizeof(config->span));
	memset(&config->chans, 0, sizeof(config->chans));

	/* set name and description */
	snprintf(config->span.name, sizeof(config->span.name), "si3210");
	snprintf(	config->span.desc, 
			sizeof(config->span.desc), 
			"Silicon Laboratories ProSLIC 3210");

	for (i=0; i<config->driver_config.num_of_chips; i++) {
		snprintf(config->chans[i].name, 
			 sizeof(config->chans[i].name),
			 "Si3210-%d", i);

		config->chans[i].pvt	= config;
		config->chans[i].chanpos= i+1;
		config->chans[i].sigcap	= 	
						ZT_SIG_SF |
						ZT_SIG_EM_E1 |
						ZT_SIG_CLEAR |
					/*	ZT_SIG_FXOKS | 	*/
						ZT_SIG_FXOLS;

		INIT_WORK(&config->work_req_ec[i],  
			(void(*)(void*))hbx_dsp_request_ec, 
			&config->driver_config.chip[i].ec_channel);

		INIT_WORK(&config->work_rel_ec[i],  
			(void(*)(void*))hbx_dsp_release_ec, 
			&config->driver_config.chip[i].ec_channel);

	}		

	config->span.hooksig 	= si3210_zap_hooksig;
	config->span.open 	= si3210_zap_open;
	config->span.close 	= si3210_zap_close;
	config->span.ioctl 	= si3210_zap_ioctl;

	config->span.chans	= config->chans;
	config->span.channels	= config->driver_config.num_of_chips;
#ifdef SI3210_USE_ULAW
	config->span.deflaw	= ZT_LAW_MULAW;
#else
	config->span.deflaw	= ZT_LAW_ALAW;
#endif
	config->span.flags	= ZT_FLAG_RBS;
	config->span.pvt 	= config;

	init_waitqueue_head(&config->span.maintq);
	
	/* register device on zaptel */
	ret = zt_register(&config->span,0);
	if (ret) {
		printk(KERN_ERR "Unable to register with zaptel\n");
		return -1;
	}
	
	return 0;
}


/**
 * Module init.
 */
static int si3210_init(void)
{
	int ret;

	PDEBUG("init\n");

	/* basic initialization of si3210 chips */
	ret = si3210_initialize_chips(&si3210_config.driver_config);
	if (ret)
		return ret;
	
	PDEBUG("Found %d chips\n", si3210_config.driver_config.num_of_chips);

	si3210_config.driver_config.dtmf_detected  = si3210_dtmf_detected;
	si3210_config.driver_config.status_changed = si3210_status_changed;

	/* 
	 * Register on zaptel;
	 * Zaptel is the interface between the low level driver and the
	 * PBX (asterisk)
	 */
	PDEBUG("Register chips on zaptel ...\n");
	ret = si3210_zap_init(&si3210_config);
	if (ret)
		return ret;

	return 0;
}


/**
 * Module exit
 */
static void si3210_exit(void)
{
	int i, ret;
	
	PDEBUG("exit\n");

	/* unregister zaptel device  */
	PDEBUG("Unregister on zaptel\n");
	ret = zt_unregister(&si3210_config.span);
	if (ret)
		printk(KERN_ERR "Unable to unregister with zaptel\n");
	
	/* power down chip */
	for (i=0; i<si3210_config.driver_config.num_of_chips; i++)
		si3210_poweroff_chip(i);
	
	PDEBUG("Release all si3210 chips\n");
	si3210_release_chips(&si3210_config.driver_config);
}

MODULE_DESCRIPTION("Zaptel driver for Si3210 chips");
MODULE_AUTHOR("Henry Margies (h.margies@maxina.de");
MODULE_LICENSE("GPL");

module_init(si3210_init);
module_exit(si3210_exit);
