/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: comhandler.c 3181 2006-09-25 10:42:20Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <time.h>

#include "util.h"
#include "ctrl.h"
#include "stun.h"
#include "varstr.h"
#include "nvalloc.h"
#include "logger.h"

#define __COMHANDLER_PRIVATE
#include "comhandler.h"


/* ----------------------------------------------------------------------- *
 * defs, macros
 * ----------------------------------------------------------------------- */

#define __str(x)      # x
#define   str(x) __str (x)

/* ----------------------------------------------------------------------- *
 * Private: helpers
 * ----------------------------------------------------------------------- */

static
void info (COMM *o) {
    int  n = time (0) - o->started ;

    ctrl_reply (o->ctrl, "+OK:\nuptime: ");

    if (n > 60) {
        int min =   n / 60 ;
        int h   = min / 60 ;
        if (h > 24) {
            ctrl_reply (o->ctrl, "%dd ", h / 24);
            h %= 24 ;
        }
        ctrl_reply (o->ctrl, "%d:%02d:%02d\n", h, min % 60, n % 60);
    }
    else
        ctrl_reply (o->ctrl, "%ds\n", n);

    rfc3489_show
        (o->r3489, (void(*)(void*,const char*,...)) ctrl_reply, o->ctrl);
}

/* ----------------------------------------------------------------------- *
 * Public: Communication handlers
 * ----------------------------------------------------------------------- */

COMM *comm_init (R3489 *s) {

    COMM *o = ZMALLOC (COMM);

    if ((o->r3489 = s) == 0) {
        logger (LOG_ERR, "Uninitialised RFC3489 state machine");
        free (o);
        return 0 ;
    }

    return o ;
}


void comm_close (COMM *o) {
    
    if (o) {
        free (o);
    }
}


COMM *comm_update (COMM *o, CTRL *c) {
    o->started = time (0);
    o->ctrl    = c ;      
    return o;
}


void comm_info (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
 
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 1) goto usage;

    info (o) ;
    return;
 
 usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, "%s -- print setup info\n", me);
}


void comm_reload (void *state, int argc, const char **argv) {

    COMM *o = state;
    const char *me = argv [0];
    int n ;
     
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 1) goto usage;

    n = rfc3489_update (o->r3489) ;

    if (n > 0)
        ctrl_reply
            (o->ctrl, "+OK: %d nvram records loaded\n", n);
    else
        ctrl_reply
            (o->ctrl, "-ERR: query failed, no nvram records available\n");
    return;
 
 usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
 help:
    ctrl_reply (o->ctrl, "%s -- reload settings from the nvram\n", me);
}


void comm_run (void *state, int argc, const char **argv) {
    COMM *o = state;
    const char *me = argv [0];
    unsigned timeout = -1, recur = -1;
    
    if (me [0] == '?' && !me [1]) {me = argv [1]; goto help; }
    if (argc > 3) goto usage;

    if (argc == 2 && parse_unsigned (&timeout, argv [1], 0) == 0) goto usage;
    if (argc == 3 && parse_unsigned (&recur,   argv [2], 0) == 0) goto usage;

    rfc3489_tmo
        (o->r3489, timeout, recur);
    rfc3489_run
        (o->r3489, (void(*)(void*,const char*,...)) ctrl_reply, o->ctrl);
    return ;

usage:
    ctrl_reply (o->ctrl, "-ERR: usage: ");
help:
    ctrl_reply
        (o->ctrl, "%s [<timeout>] [<recur>] -- update nat layout\n", me);
}


/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
