/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: device.c 1981 2006-04-05 20:30:24Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/if_tun.h>
#include <net/route.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>

#include "util.h"
#include "settings.h"
#include "nvalloc.h"
#include "logger.h"

#define __DEVICE_PRIVATE
#include "device.h"

/* ----------------------------------------------------------------------- *
 * variables and definitions
 * ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- *
 * Fancy error message
 * ----------------------------------------------------------------------- */

static
void init_error (const char   *msg, 
                const char    *arg, 
                const char *action,
                int            xit) {
    if (arg == 0 && action == 0)
        logger (LOG_ERR, "%s: %s", progname, msg);
    else if (arg == 0 && action != 0)
        logger (LOG_ERR, "%s: %s: %s (%s)",
                 progname, msg, action, strerror (errno));
    else if (arg != 0 && action == 0)
        logger (LOG_ERR, "%s: %s %s", progname, msg, arg);
    else /*  arg != 0 && action != 0 */
        logger (LOG_ERR, "%s: %s %s: %s (%s)",
                progname, msg, arg, action, strerror (errno));
    if (xit) {
        sleep (5); /* do not exit immediately - wait for some seconds */
        exit (2);
    }
}

static
void if_init_error (const char *dev, const char *action, int xit) {
    return init_error
        ("Failed to initialise interface", dev, action, xit);
}

/* ----------------------------------------------------------------------- *
 * inbound interface query
 * ----------------------------------------------------------------------- */

ip4addr_t ifquery (const char *name, IP4_POOL *ifaddr, int xit) {
    
    int ioctl_fd ;
    static struct in_addr ipa ;
    struct ifreq ifr ;
    char ifname [IFNAMSIZ];   /* interface name */

    /* set up interface info structure */
    if (name == 0 || *name == 0)
        name = DEFAULT_STUNBOUND;

    /* create ioctl command socket */
    if ((ioctl_fd = socket (PF_PACKET, SOCK_DGRAM, 0)) < 0) {
        init_error ("Cannot create a packet socket", "SOCK_DGRAM",
                    "PF_PACKET", xit);
        return 0;
    }

    /* GET INTERFACE INFO */

    /* get interface index for the socket address and use it in
       the queries, following */
    XZERO (ifr); 
    strcpy (ifr.ifr_name, strncpy (ifname, name, IFNAMSIZ-1));
    if ((ioctl (ioctl_fd, SIOCGIFINDEX, &ifr)) < 0) {
        if_init_error (ifr.ifr_name, "SIOCGIFINDEX", xit);
        close (ioctl_fd) ;
        return 0 ;
    }

    /* check interface flags */
    if (ioctl (ioctl_fd, SIOCGIFFLAGS, &ifr) < 0) {
        init_error ("Can't get interface flags of",
                    ifname, "SIOCGIFFLAGS", xit);
        goto close_ioctl_exit ;
    }
    if ((ifr.ifr_flags & IFF_UP) == 0) {
        init_error ("Interface is down", ifr.ifr_name, 0, xit);
        goto close_ioctl_exit ;
    }
    if ((ifr.ifr_flags & IFF_LOOPBACK) != 0) {
        init_error
            ("The loopback interface will not work: ", ifname,0,xit);
        goto close_ioctl_exit ;
    }

    /* get (primary) ip address */
    XZERO (ifr.ifr_ifru);
    ifr.ifr_addr.sa_family = AF_INET;
    if (ioctl (ioctl_fd, SIOCGIFADDR, &ifr) < 0) {
        if_init_error (ifname, "SIOCGIFADDR", xit);
        goto close_ioctl_exit ;
    }
    memcpy (&ipa,
            &((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr,
            sizeof (ipa));

    if (ifaddr) {
        struct ifconf ifc ;
        int n ;
        /* collect all the device addresses */
        for (n = 4; /* empty */ ; n += 3) {
            int len = n * sizeof (struct ifreq);
            ifc.ifc_len = len ;
            ifc.ifc_req = (struct ifreq *) zmalloc (ifc.ifc_len) ;

            /* Get a list of all interface addresses */
            if (ioctl (ioctl_fd, SIOCGIFCONF, &ifc) < 0) {
                if_init_error (ifr.ifr_name, "SIOCGIFCONF", xit);
                free (ifc.ifc_req);
                goto close_ioctl_exit ;
            }

            /* Make sure that we got all. Thus we want the system to have
               less space filled than offered. */
            if (len <= ifc.ifc_len) {
                free (ifc.ifc_req);
                continue ;
            }

            /* filter out the LAN interface addresses */
            n = ifc.ifc_len / sizeof (struct ifreq);
            while (-- n >= 0)
                if (strcmp (ifname, ifc.ifc_req [n].ifr_name) == 0)
                    /* add to the address list */
                    ip4_fetch
                        (ifaddr, &((struct sockaddr_in*)
                                   &ifc.ifc_req [n].ifr_addr)->sin_addr);
            free (ifc.ifc_req);
            break ;
        }
    }
    
    // /* get mtu */
    // if (ioctl (ioctl_fd, SIOCGIFMTU, &ifr) < 0) {
    //    if_init_error (ifr.ifr_name, "SIOCGIFMTU", xit);
    //    goto close_ioctl_exit ;
    // }
    // ifmtu = ifr.ifr_mtu ;

    /* not needed anymore */
    close (ioctl_fd);

    return ntohl (ipa.s_addr) ;
    /*NOTREACHED*/

close_ioctl_exit:
    close (ioctl_fd);
    return 0 ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
