/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: ipcalc.c 1070 2006-01-17 21:27:20Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#include "util.h"
#include "ipcalc.h"

/* ----------------------------------------------------------------------- *
 * private
 * ----------------------------------------------------------------------- */

/* IP checksum */
static 
uint16_t in_cksum (const uint16_t *ptr,
                   uint32_t        sum,
                   int             len) {
    while (len > 1)  {
        sum += ((uint8_t*)ptr) [0] | (((uint8_t*)ptr) [1] << 8);
        ptr ++ ;
        len -= 2;
    }
    if (len == 1) {
        uint16_t odd = 0;
        *(uint8_t*)&odd = *(uint8_t*)ptr;
        sum += odd;                         /* adding [<last-byte>, 0] */
    }
    sum  = (sum >> 16) + (sum & 0xffff);  /* add high-16 to low-16 */
    sum += (sum >> 16);                   /* add carry */
    return ~sum ;          /* ones-complement, then truncate to 16 bits */
}


uint16_t ip_cksum (const struct ip *pkt) {
    uint32_t c_sum = in_cksum
        ((const uint16_t*)pkt, -pkt->ip_sum, pkt->ip_hl*4);
    c_sum   = c_sum + (c_sum >> 16) ;
    c_sum   = c_sum + (c_sum >> 16) ;
    return c_sum ;
}


static 
uint16_t phdr_ip_cksum (const struct ip *pkt,
                        uint16_t     old_cks,
                        unsigned short   max) {
    uint32_t c_sum ;
    long len = ntohs (pkt->ip_len) - sizeof (struct ip);
    if (len > max || len < 0) return -1 ;
    c_sum = in_cksum
        ((const uint16_t*)(((uint8_t*)(pkt+1))-8), -old_cks, len+8);
    /* add more pseudo header stuff */
    c_sum -= ntohs ((uint16_t)len) ;
    c_sum -= ntohs (pkt->ip_p);
    c_sum  = c_sum + (c_sum >> 16) ;
    c_sum  = c_sum + (c_sum >> 16) ;
    return c_sum ;
}


static
char *cks_ok (uint16_t cks_is, uint16_t cks_should) {
    static char buf [50];
    if (cks_is == cks_should)
        sprintf (buf, "sum %#x ok", ntohs (cks_is));
    else
        sprintf (buf, "sum %#x should be %#x",
                 ntohs (cks_is), ntohs (cks_should));
    return buf ;
}


static
char *ipkt_pp_appnd (char *s, int max, const struct ip *pkt, int verbose) {
    char *p;
    unsigned long n ;

    p = memchr (s, 0, max); if (p == 0) return s; else max -= (p - s);
    switch (pkt->ip_p) {
    case IPPROTO_ICMP: {
        struct icmphdr *i = (struct icmphdr *)(pkt + 1);
        sprintf (p, "%s > %s: (%u): ",
                 ipaddr_pp (&pkt->ip_src), 
                 ipaddr_pp (&pkt->ip_dst),
                 (int)i->type);
        if ((n = icmp_cksum (pkt, -1)) != i->checksum || verbose) {
            p = memchr (s, 0, max);
            if (p == 0) return s; else max -= (p - s);
            sprintf (p, "[icmp %s]", cks_ok (i->checksum, n));
        }
        else
            sprintf (p, "[icmp sum ok]");
        break ;
    }
    case IPPROTO_UDP: {
        struct udphdr *u = (struct udphdr *)(pkt + 1);
        sprintf (p, "%s.%u > %s.%u: ",
                 ipaddr_pp (&pkt->ip_src), (int)ntohs (u->source),
                 ipaddr_pp (&pkt->ip_dst), (int)ntohs (u->dest));  
        if ((n = udp_cksum (pkt, -1)) != u->check || verbose) {
            if (u->check) {
                p = memchr (s, 0, max);
                if (p == 0) return s; else max -= (p - s);
                sprintf (p, "[udp %s]", cks_ok (u->check, n));
            }
            else
                sprintf (p, "[udp no-cks]");
        }
        else
            sprintf (p, "[udp sum ok]");
        break ;
    }
    case IPPROTO_TCP: {
        struct tcphdr *t = (struct tcphdr *)(pkt + 1);
        sprintf (p, "%s.%u > %s.%u: ",
                 ipaddr_pp (&pkt->ip_src), (int)ntohs (t->source),
                 ipaddr_pp (&pkt->ip_dst), (int)ntohs (t->dest));  
        n = 0 ;
        if (t->psh) { strcat (p, "P"); n ++; }
        if (t->syn) { strcat (p, "S"); n ++; }
        if (t->fin) { strcat (p, "F"); n ++; }
        if (t->rst) { strcat (p, "R"); n ++; }
        if (t->urg) { strcat (p, "U"); n ++; }
        if (n == 0)   strcat (p, ".");

        if ((n = tcp_cksum (pkt, -1)) != t->check || verbose) {
            p = memchr (s, 0, max);
            if (p == 0) return s; else max -= (p - s);
            sprintf (p, " [tcp %s]", cks_ok (t->check, n));
        }
        else
            strcat (p, " [tcp sum ok]");
    
        p = memchr (s, 0, max); if (p == 0) return s; else max -= (p - s);
        n = ntohs (pkt->ip_len) - 4*pkt->ip_hl - 4*t->doff;
        sprintf (p, " %lu:%lu(%lu)",
                 (long)ntohl (t->seq), (long)ntohl (t->seq) + n, n);

        if (t->ack) {
            p = memchr (s, 0, max);
            if (p == 0) return s; else max -= (p - s);
            sprintf (p, " ack %u", ntohl (t->ack_seq));
        }
        break ;
    }
    default:
        sprintf (p, "%s > %s ipproto=%u",
                 ipaddr_pp (&pkt->ip_src), 
                 ipaddr_pp (&pkt->ip_dst),
                 (int)pkt->ip_p);
    }

    /* sub protocol options here */
    if (ntohs (pkt->ip_off) & IP_DF)
        strcat (s, " (DF)");
    if (ntohs (pkt->ip_off) & IP_MF)
        strcat (s, " (MF)");

    /* fragmentation */
    if (ntohs (pkt->ip_off) & IP_OFFMASK) {
        p = memchr (s, 0, max); if (p == 0) return s; else max -= (p - s);
        sprintf (p, " (offset %d,", ntohs (pkt->ip_off));
    }
    else
        strcat (s, " (");

    /* general ip specs */
    p = memchr (s, 0, max); if (p == 0) return s; else max -= (p - s);
    if ((n = ip_cksum (pkt)) != pkt->ip_sum || verbose) {
        sprintf (p, "ttl %u, id %u, len %d, %s)",
                 (int)pkt->ip_ttl, 
                 (int)ntohs (pkt->ip_id),
                 (int)ntohs (pkt->ip_len),
                 cks_ok (pkt->ip_sum, n));
    }
    else
        sprintf (p, "ttl %u, id %u, len %d)", 
                 (int)pkt->ip_ttl, 
                 (int)ntohs (pkt->ip_id),
                 (int)ntohs (pkt->ip_len));

    return s ;
}



/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

/* pretty print the given integer target */
const char *ulongp_pp (const long *p) {
    static char buf [1 + sizeof (long) * 3] ;
    if (p == 0)
        return "<no-int>";
    sprintf (buf, "%lu", *p);
    return buf ;
}


const char *ushortp_pp (const short *p) {
    static char buf [1 + sizeof (long) * 3] ;
    if (p == 0)
        return "<no-int>";
    sprintf (buf, "%hu", *p);
    return buf ;
}


/* pretty print the given integer target */
const char *xlongp_pp (const long *p) {
    static char buf [5 + sizeof (long) * 2] ;
    if (p == 0)
        return "<no-int>";
    sprintf (buf, "%lx", *p);
    return buf ;
}


/* pretty print the given quad integer target */
const char *xquadp_pp (const uint128_t *p) {

    static char buf [3 + sizeof (uint128_t) * 2] = "0x" ;
    static char hex [] = "0123456789abcdef" ;
    int i = 0 ;

    if (p == 0)
        return "<no-quad-int>";

    while (i < 32 && p->oct [i])
        i += 2 ;
    if (i == 30)
        return "0x0" ;
    if ((p->oct [i] & 0xf0) == 0)
        i ++ ;
    while (i < 32) {
        buf [i + 2] = (i & 1)
            ? hex [p->oct [i >> 1] >>  4]
            : hex [p->oct [i >> 1] & 0xf]
            ;
        i ++ ;
    }
    buf [i] = '\0' ;

    return buf ;
}

uint16_t tcp_cksum (const struct ip *pkt, unsigned short max) {
    return phdr_ip_cksum (pkt, ((struct tcphdr*)(pkt+1))->check, max);
}


uint16_t udp_cksum (const struct ip *pkt, unsigned short max) {
    return phdr_ip_cksum (pkt, ((struct udphdr*)(pkt+1))->check, max);
}

uint16_t icmp_cksum (const struct ip *pkt, unsigned short max) {
    uint16_t old_cks = ((struct icmphdr*)(pkt+1))->checksum ;
    long len = ntohs (pkt->ip_len) - sizeof (struct ip);
    if (len > max || len < 0) return -1 ;
    return in_cksum
        ((const uint16_t*)((struct icmphdr*)(pkt+1)), -old_cks, len);
}


const char *ipkt_pp (const struct ip *pkt, int verbose) {
    static char buf [1024]; /* should be big enough */
    buf [0] = '\0';
    return ipkt_pp_appnd (buf, sizeof buf, pkt, verbose);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
