/*
 * nvram lib - flash data base client library
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: base64.c 3175 2006-09-23 19:17:11Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#include "util.h"
#include "nvalloc.h"
#include "base64.h"

/* ----------------------------------------------------------------------- *
 * Local variables
 * ----------------------------------------------------------------------- */

static
const char _b64str [] = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                         "abcdefghijklmnopqrstuvwxyz"
                         "0123456789"
                         "+/=");

static
const char _alpha  [] = ("0123456789"
                         "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                         "abcdefghijklmnopqrstuvwxyz"
                         "_#") ;

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

static
int _46b (const char **p) {

    int c = * (*p) ++ ;
    
    if (c == '%') {
        if ((c  = parse_hex_dgt (* (*p)) << 8) & ~0xf0) return 0;
        (*p) ++ ;
        if ((c |= parse_hex_dgt (* (*p)))      & ~0xff) return 0;
        (*p) ++ ;
    }
    
    if ('A' <= c && c <= 'Z') return c - 'A' ;
    if ('a' <= c && c <= 'z') return c - 'a' + 'Z' - 'A' + 1;
    if ('0' <= c && c <= '9') return c - '0' + 'Z' - 'A' + 'z' - 'a' + 2;

    switch (c) {
    case '+': return 62;
    case '/': return 63;
    case '=': return -1;
    }

    (*p) -- ;
    return 0;
}


static
void _put_b64 (char      **buf,
               const char *map,
               size_t    *Offs,
               size_t     *Max,
               size_t   sextet) {

    sextet &= 0x3f ;

    /* assuming the buffer is big enough for the base 64 encoded string */
    if (Max == 0 || sextet < 62) {
        (* buf) [(* Offs) ++] = map [sextet] ;
        return ;
    }

    if (*Max <= *Offs + 3)
        * buf = (char*) zrealloc (* buf, (*Max) += 30);

    (* buf) [* Offs ++] = '%' ;

    switch (sextet) {
    case 62: /* '+' */
        (* buf) [(* Offs) ++] = '2' ;
        (* buf) [(* Offs) ++] = 'b' ;
        break;
    case 63: /* '/' */
        (* buf) [(* Offs) ++] = '2' ;
        (* buf) [(* Offs) ++] = 'f' ;
    }
}


static
void _put_b64_pad (char **buf, size_t *Offs, size_t *Max) {

    if (Max == 0) {
        (* buf) [(* Offs) ++] = '=' ;
        return ;
    }
    
    if (*Max <= *Offs + 3)
        * buf = (char*) zrealloc (* buf, (*Max) += 30);
    (* buf) [(* Offs) ++] = '%' ;
    (* buf) [(* Offs) ++] = '3' ;
    (* buf) [(* Offs) ++] = 'd' ;
}


static
const char *_2b64 (const char         *intro,
                   const char           *map,
                   const unsigned char *data,
                   size_t                len,
                   int                httpok) {
    long cluster ;
    size_t   inx ;
    size_t  *Max ;

    size_t  ilen = intro ? strlen (intro) : 0 ;
    size_t bsize = len * 4 / 3 + 5 ;
    char    *b64 = (char*) xmalloc (bsize + ilen) ;


    if (ilen) {
        memcpy (b64, intro, ilen);
        inx = ilen ;
    }
    else
        inx = 0 ;

    /* no adjustment unless http - buffer is always big enough */
    Max = (httpok == 0) ? 0 : & bsize ;

    /* need http escaped base 64 */
    while (len >= 3) {
        cluster = ((((data [0] << 8) | data [1]) << 8) | data [2]) ;
        _put_b64 (&b64, map, &inx, Max, cluster >> 18) ;
        _put_b64 (&b64, map, &inx, Max, cluster >> 12) ;
        _put_b64 (&b64, map, &inx, Max, cluster >>  6) ;
        _put_b64 (&b64, map, &inx, Max, cluster)       ;
        len  -= 3 ;
        data += 3 ;
    }

    switch (len) {
    case 2:
        cluster = ((data [0] << 8) | data [1]) ;
        _put_b64 (&b64, map, &inx, Max, cluster >> 10) ;
        _put_b64 (&b64, map, &inx, Max, cluster >>  4) ;
        _put_b64 (&b64, map, &inx, Max, cluster <<  2) ;
        _put_b64_pad (&b64, &inx, Max) ; /* '=' */
        break ;
    case 1:
        cluster = data [0] ;
        _put_b64 (&b64, map, &inx, Max, cluster >>  2) ;
        _put_b64 (&b64, map, &inx, Max, cluster <<  4) ;
        _put_b64_pad (&b64, &inx, Max) ; /* '=' */
        _put_b64_pad (&b64, &inx, Max) ; /* '=' */
        break ;
    }

    b64 [inx] = '\0' ;

    return b64;
}


/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

const char *x64alloc (const char *intro,
                      const char     *s,
                      size_t          n,
                      int        httpok) {
    return _2b64 (intro, _b64str, (const unsigned char*)s, n, httpok);
}


const char *b64alloc (const char *s, size_t n) {
    return _2b64 (0, _b64str, (const unsigned char*)s, n, 0);
}


const char *x46alloc (const char **Q, size_t *Len, const char *p) {

    const char *q = p + strlen (p);
    char *a, *r;

    /* strip trailing blanks */
    while (p < q && isspace (q [-1]))
        q -- ;
    
    a = r = (char*) xmalloc ((q - p) * 3 / 4 + 1);

    while (p + 4 <= q) {
        int c ;
        long cluster ;

        cluster   = _46b (&p);
        cluster <<= 6 ;
        cluster  |= _46b (&p);

        /* check for padding */
        if ((c = _46b (&p)) == -1) {
            if ((c = _46b (&p)) != -1) {
                free (a) ;
                return "Final padding character in base 64 string "
                    "was expected.";
            }
            cluster >>= 4 ;
            * r ++ = cluster & 0xff ;
            break ;
        }

        cluster <<= 6 ;
        cluster  |= c ;

        /* check for padding */
        if ((c = _46b (&p)) == -1) {
            cluster >>= 2 ;
            * r ++ = (cluster >> 8) & 0xff ;
            * r ++ =  cluster       & 0xff ;
            break ;
        }

        cluster <<= 6 ;
        cluster  |= c ;
 
        * r ++ = (cluster >> 16) & 0xff ;
        * r ++ = (cluster >>  8) & 0xff ;
        * r ++ =  cluster        & 0xff ;
    }

    if (p < q) {
        free (a) ;
        return "Base 64 string length must be a multiple of 4." ;
    }

    if (Q) {
        * r = '\0' ; /* place EOT but don't increment */
        * Q =   a  ;
    }

    if (Len)
        * Len = r - a ;

    return 0;
}



const char *rndalloc (size_t n) {

    size_t m = n * 3 / 4 + 2 ;
    char *t, *s = (char *) xmalloc (n + 1);

    rndata (s, m) ;
    // t = (char *)_2b64 (_alpha, (unsigned char*)s, m) ;
    t =  (char *)_2b64 (0, _alpha, (unsigned char*)s, m, 0);
    memcpy (s, t, n);
    s [n] = '\0' ;
    free (t);

    if ((t = strchr (s, '#')) != 0) {
        int max = 2 * n ;
        /* get rid of '#' characters */
        do {
            char c ;
            rndata (&c, 1) ;
            * t = _alpha [c & 0x3f] ;
        } while (-- max > 0 && (t = strchr (s, '#')) != 0) ;
        /* with some weird pobablility, all rand chars are '#' */
        while ((t = strchr (s, '#')) != 0)
            * t = '_' ;
    }

    return s;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
