/*
 * nvram lib - flash data base client library
 *
 * Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA
 *
 * Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-Isenburg, Germany
 *
 * $Id: util.c 2640 2006-06-22 10:01:06Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_LIBGEN_H
# include <libgen.h>
#endif
#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#define __NVRAMLIB_UTIL_PRIVATE
#include "util.h"

#include "logger.h"
#include "nvalloc.h"

/* ----------------------------------------------------------------------- *
 * Global variables
 * ----------------------------------------------------------------------- */

char nvcl_request_socket [] = NVCL_REQUEST_SOCKET ;

char *progname = "<unset>" ;      /* program name */
unsigned verbosity = 0;

/* ----------------------------------------------------------------------- *
 * Local variables
 * ----------------------------------------------------------------------- */

static char *random_device = 0 ;
static int   random_fd     = -1;
static char  random_buf [200] ;
static int   random_inx = sizeof (random_buf);

/* ----------------------------------------------------------------------- *
 * random stuff
 * ----------------------------------------------------------------------- */

static
void open_random (char *rd) {

    if (rd)
        random_device = rd;
    else if (random_device == 0)
        random_device = DEFAULT_RANDOM_DEV ;
 
    if ((random_fd = open (random_device, O_RDONLY)) == 0) {
        logger (LOG_ERR, "%s: open (%u) returned error (%s)!\n",
                progname, random_device, strerror (errno));
        exit (2) ;
    }
}

static
void refill_random_buf (void) {
    int errors = 0, offset = 0 ;

    while (offset < sizeof (random_buf)) {
        int n = read
            (random_fd, random_buf + offset, sizeof (random_buf) + offset);

        if (n > 0) {
            offset += n ;
            continue ;
        }
 
        /* reopen the random device and retry once */
        if (errors ++ == 0) {
            close (random_fd) ;
            open_random (0) ;
            continue ;
        }

        /* otherwise die on repeated errors */
        if (n < 0)
            logger (LOG_ERR, "%s: read (%u) returned error (%s)!\n",
                    progname, random_device, strerror (errno));
        else
            logger (LOG_ERR, "%s: read (%u) returned zero!\n",
                    progname, random_device);
        exit (2) ;
    }

    random_inx = 0 ;
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

/* random data */
void rndata (char *s, size_t n) {
    if (random_fd < 0)
        open_random (0);
    while (n > 0) {
        unsigned m ;
        if (random_inx >= sizeof (random_buf))
            refill_random_buf ();
        if ((m = sizeof (random_buf) - random_inx) > n)
            m = n ;
        memcpy (s, random_buf + random_inx, m);
        s          += m ;
        random_inx += m ;
        n          -= m ;
        // * s ++ = random_buf [random_inx ++] ;
    }
}

/* pretty print the given time stamp */
char *stamp_pp (const struct timeval *tv) {

    struct timeval _tv ;
    struct tm *tm ;
    char fm [50] ;
    int  sz ;

    static char buf [60] ;
  
    if (tv == 0) {
        if (gettimeofday (&_tv, 0) < 0)
            memset (&_tv, 0, sizeof (_tv));
        tv = &_tv ;
    }

    tm = localtime (&tv->tv_sec) ;
    sz = tm ?strftime (fm, sizeof (fm), "%Y-%d-%m %H:%M:%S.%%06u", tm) :0 ;

    if (sz > 0)
        sprintf (buf, fm, tv->tv_usec);
    else
        sprintf (buf, "%u.%06u",
                 (unsigned)tv->tv_sec, (unsigned)tv->tv_usec);

    return buf ;
}


/* pretty print the given time integer */
char *time_pp (long t) {
    struct tm *tm = localtime (&t) ;
    static char buf [50] ;

    if (tm) {
        size_t sz = strftime (buf, sizeof (buf), "%Y-%d-%m %H:%M:%S", tm);
        if (sz > 0)
            return buf ;
    }
    
    sprintf (buf, "%lu", t);
    return buf ;
}

/* pretty print an ipv4 address */
const char *ipaddr_pp (const struct in_addr *ia) { /* probably unaligned */
  
    char *p ;
    struct in_addr i; /* aligned on stack */

    /* provide for some buffers of length 16 bytes */
    static char buf [16 * 4] ;
    static int  inx ;

    if (ia == 0)
        return "<no-ip>";

    /* rotate: get the other buffer */
    if ((inx += 16) >= sizeof (buf))
        inx = 0 ;

    if (ia == 0 ||
        (p = inet_ntoa (*(struct in_addr*)memcpy (&i,ia,sizeof (i)))) == 0 ||
        strlen (p) >= 16)
        return "???.???.???.???" ;
  
    return strcpy (buf + inx, p);
}


const char *ip4addr_pp (ip4addr_t ip) {
    struct in_addr ia;
    ia.s_addr = htonl (ip) ;
    return ipaddr_pp (&ia) ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
