/*
 * Copyright (C) 2006 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: do.c 3502 2006-10-31 22:53:03Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#include "util.h"
#include "tvcalc.h"
#include "nvalloc.h"
#include "logger.h"

#ifdef _TESTING
#include "../exit23.h"
#endif

#define __DO_PRIVATE
#include "do.h"

// #if defined (USE_DEBUG) && !defined (_DEBUG_IDLE_SCHED)
#define _DEBUG_IDLE_SCHED 1
// #endif

#undef WITH_DMALLOC

#ifdef WITH_DMALLOC
static unsigned long __mark, __mark0;

#define DMALLOC_MARK0() { __mark = dmalloc_mark (); }
#define DMALLOC_MARK() { __mark = dmalloc_mark (); }

#define DMALLOC_HUNT0() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }
#define DMALLOC_HUNT() { \
                  dmalloc_message (__FILE__ "(%d): "\
                     "chasing memory leaks ...", __LINE__) ;               \
                  dmalloc_log_changed (__mark,                             \
                     1 /* log unfreed pointers */,                         \
                     0 /* do not log freed pointers */,                    \
                     1 /* log individual pointers otherwise a summary */); }
#else
#define DMALLOC_MARK0()           /* n/a */
#define DMALLOC_HUNT0()           /* n/a */
#define DMALLOC_MARK()            /* n/a */
#define DMALLOC_HUNT()            /* n/a */
#endif

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

void do_destroy (DO *dp) {
    if (dp->set != 0)
        free (dp->set);
    free (dp);
}


DO *do_init (void (*cb) (void*), void *data) {

#   ifdef _TESTING
    X23SNOOP();
#   endif

    DO *dp = (DO*) zmalloc (sizeof (DO)) ;

    dp->idle   =   cb ;
    dp->data   = data ;
    dp->max_fd =   -1 ;

    // FD_ZERO (&dp->fds);
    return dp ;
}



void do_action (DO *dp, int fd, void (*cb)(void*), void *cb_data) {

    int inx ;
    
    if (fd < 0 || cb == 0)
        return ;

    DMALLOC_MARK ();

    for (inx = 0; inx < dp->length; inx ++)
        if (dp->set [inx].rfd < 0 || dp->set [inx].rfd == fd)
            goto reuse ;

    dp->length ++ ;
    dp->set = (DO_SET*) zrealloc /* dp->set == NULL is ok */
        (dp->set, dp->length * sizeof (DO_SET)) ;

    memset (dp->set + inx, 0, sizeof (DO_SET)) ; /* length + 1 == inx */
    
reuse:
    dp->set [inx].rfd     =      fd ;
    dp->set [inx].cb      =      cb ;
    dp->set [inx].cb_data = cb_data ;
    FD_SET (fd, &dp->fds);
    
    if (fd > dp->max_fd)
        dp->max_fd = fd ;

    if (DO_LOG_ACTION_REGISTER & verbosity)
        logger (LOG_INFO, "Added action for fd=%u at slot %u", fd, inx);

    DMALLOC_HUNT();
}


int do_expire (DO             *dp,
               int             fd,
               unsigned   timeout,
               void (*xh) (void*),
               void      *xh_data) {
    int n ;

    if (fd >= 0)

        for (n = 0; n < dp->length; n ++)

            if (dp->set [n].rfd == fd) {

                if (dp->set [n].expiry) {

                    if (timeout == 0) {
                        dp->set [n].timeout = 0 ;
                        dp->set [n].expiry  = 0 ;
                        dp->with_ttd -- ; /* one less */
                        return fd ;
                    }
                }

                /* expiry is unset */
                else { 
                    if (timeout == 0)
                        return fd ;
                    dp->with_ttd ++ ;     /* one more */
                }

                dp->set [n].timeout =            timeout ;
                dp->set [n].expiry  = time (0) + timeout ;
                dp->set [n].xh      =                 xh ;
                dp->set [n].xh_data =            xh_data ;
                return fd ;
            }

    return -1 ;
}


int undo_action (DO *dp, int fd) {

    if (fd >= 0) {
        int n, inx = -1, max_fd = -1, with_ttd = 0 ;

        for (n = 0; n < dp->length; n ++) {

            if (dp->set [n].rfd == fd) {
                dp->set [n].rfd = -1 ;
                FD_CLR (fd, &dp->fds);
                inx = n ;
                continue ;
            }

            if (dp->set [n].rfd == -1)
                continue ;

            if (max_fd < dp->set [n].rfd)
                max_fd = dp->set [n].rfd ;

            if (dp->set [n].expiry != 0)
                with_ttd ++ ;
        }

        if (inx >= 0) {
            /* update statistics/registry */
            dp->max_fd   =   max_fd ;
            dp->with_ttd = with_ttd ;

            if (DO_LOG_ACTION_REGISTER & verbosity)
                logger (LOG_INFO, "Released action for fd=%u at slot %u",
                        fd, inx);

            return fd ;
        }
    }

    return -1 ;
}


void do_resched (DO *dp, int tmo_ms) {

    if (tmo_ms == 0)
        tmo_ms = dp->default_timeout ;

    if (tmo_ms  > 0)
        tval_set_ms (&dp->tv, tmo_ms);

    dp->timeout = tmo_ms ;
    tval_nxt_tv (&dp->due, &dp->tv) ; /* next run due */

    if (verbosity & DO_LOG_ACTION_NEXTIDLE) {
        struct timeval tmp ;
        char b1 [100];
        tval_nxt_ms (&tmp, 0);
        strcpy (b1, tval_print (&tmp));
        logger (LOG_INFO, "Do_resched at %s: Next select timeout at "
                "%s(%dns)", b1, tval_print (&dp->due), tmo_ms * 1000);
    }
}


time_t do_time (DO *dp) {
    return dp->now ? dp->now : time (0);
}


time_t do_time_cont (DO *dp, time_t delta) {
    return dp->time_leap + delta < time (0) ;
}


int do_loop (DO *dp, int tmo_ms) {
      
#   ifdef _TESTING
    X23CHECK();
#   endif

    dp->default_timeout = 
        dp->timeout = tmo_ms ;
 
    if (tmo_ms > 0) {
        tval_nxt_ms (&dp->tv, tmo_ms);   /* compare with due */
        tval_min_tv (&dp->due, &dp->tv); /* due: maybe from last session */
        tval_set_ms (&dp->tv, tmo_ms);
    }
    else {
        if (tmo_ms == 0)
            XZERO (dp->tv);
        XZERO (dp->due) ;
    }

    /* action loop */
    for (;;) {
        struct timeval rtv, *_tv, now;
        int io_events;

        fd_set rfds = dp->fds ;

        /* use a fixed time base */
        gettimeofday (&now, 0);

        /* reset time parameters, note that we may have been rescheduled */
        if (dp->timeout > 0) {

            /* rtv: get the time date when the next idle run is due */
            rtv = now ;
            tval_add_tv (&rtv, &dp->tv);

            /*
             * Note:
             *
             * Usually, the time dp->due will have been set, already.  Due
             * to a change of the system time, this value may be inaccurate.
             * So taking the smaller date activates the the event handlers
             * just one time too often, in the worst case.
             */
            if (tval_lt_tv (&dp->due, &rtv)) {

                /* is dp->due in the future or in the past? */
                if (tval_le_tv (&now, &dp->due)) {

                    /* future: figure out the corresponding timeout */
                    rtv = dp->due ;
                    tval_sub_tv (&rtv, &now);
                }

                /* system date has jump-changed, dp->due < current date? */
                else {
                    /* just some early non-zero timeout */
                    rtv.tv_sec = 0 ;
                    rtv.tv_usec = 50 ;
                    XZERO (dp->due) ;
                }
            }

            /* the timeout dp->tv is smaller than "dp->due - now" */
            else
                rtv = dp->tv ;

            /*
             * Always use a copy:
             * The select() call changes the contents of _tv
             */

            _tv = &rtv ;
        }

        else if (dp->timeout == 0) {
            rtv = dp->tv ;
            _tv = &rtv ;
        }
        else
            _tv = 0 ;

#       ifdef _DEBUG_IDLE_SCHED
        if ((verbosity & DO_LOG_ACTION_NEXTIDLE) != 0 && _tv != 0) {
            struct timeval tmp ;
            char b1 [100];
            tval_nxt_ms (&tmp, 0);
            strcpy (b1, tval_print (&tmp));
            tval_nxt_tv (&tmp, _tv);
            logger (LOG_DEBUG,
                    "Do_loop at %s: Select timeout at %s(%d.%03d)",
                    b1, tval_print (&tmp),
                    _tv->tv_sec, (_tv->tv_usec + 500L) / 1000L);
        }
#       endif

        /* wait for an event */
        if ((io_events = select (dp->max_fd + 1, &rfds, 0, 0, _tv)) < 0) {
            /* next idle time call in dp->def.tv seconds */
            tval_nxt_tv (&dp->due, &dp->tv) ;
            return -1;
        }

        /* set the time for the event handlers */
        dp->now = time (0) ;

        /* check whether somebody tuned the system date */
        if (dp->now < now.tv_sec ||
            (_tv != 0 && now.tv_sec + _tv->tv_sec < dp->now)) {

#           ifdef _DEBUG_IDLE_SCHED
            if (verbosity & DO_LOG_ACTION_NEXTIDLE) {
                struct timeval tmp ;
                char b1 [100];
                tval_nxt_ms (&tmp, 0);
                strcpy (b1, tval_print (&tmp));
                tval_nxt_ms (&tmp, 0);
                logger (LOG_INFO, "Do_loop at %s: System time has changed, "
                        "assuming it is %s for the moment",
                        b1, time_print (dp->now));
            }
#           endif

            dp->time_leap = dp->now ;
            dp->now       = now.tv_sec ;
        }

        if (io_events > 0 || dp->with_ttd > 0) {
            int dead_channels, n, max_fd ;

            /* loop over all the event handlers */
            for (n = max_fd = dead_channels = 0; n < dp->length; n ++) {
                if (dp->set [n].rfd < 0)
                    continue ;

                if (dp->with_ttd > 0) {
                    /* check for pending release requests */
                    if (dp->set [n].expiry &&
                        dp->set [n].expiry < dp->now) {

                        /* this handler may reset the time-to-die */
                        (*dp->set [n].xh) (dp->set [n].xh_data) ;

                        /* then recheck the time-to-die */
                        if (dp->set [n].expiry &&
                            dp->set [n].expiry < dp->now) {
                            if (DO_LOG_ACTION_EXPIRE & verbosity)
                                logger (LOG_INFO, "Removed expired "
                                        "action for fd=%u at slot %u",
                                        dp->set [n].rfd, n);
                            dp->set [n].expiry =  0 ;
                            dp->set [n].rfd    = -1 ;
                            dead_channels ++ ;
                            dp->with_ttd -- ;
                            continue ;
                        }
                    }

                    /* re-calculate the max-fd */
                    if (max_fd < dp->set [n].rfd)
                        max_fd = dp->set [n].rfd ;
                }

                if (io_events > 0) {
                    /*  action handler (at least one should apply) */
                    if (FD_ISSET (dp->set [n].rfd, &rfds))
                        (*dp->set [n].cb) (dp->set [n].cb_data) ;
                }

            } /* for (all event handlers) */

            /* update max-fd (in case of changes, only) */
            if (dead_channels)
                dp->max_fd = max_fd ;
        }

        /* return immediately if we are polling or on no timeout */
        else if (io_events == 0 && dp->timeout <= 0) {
            dp->now = 0 ;
            return 0;
        }

        /* indicate: event handler done */
        dp->now = 0 ;

        /* do not call the idle time task while it is too early */
        if (tval_wait (&dp->due))
            continue ;

#       ifdef _DEBUG_IDLE_SCHED
        if (verbosity & DO_LOG_ACTION_NEXTIDLE) {
            struct timeval tmp, tmo ;
            char b1 [100];
            char b2 [100];
            tval_nxt_ms (&tmp, 0);
            tval_nxt_tv (&tmo, &dp->tv);
            strcpy (b1, tval_print (&tmp));
            strcpy (b2, tval_print (&tmo));
            logger (LOG_INFO, "Do_loop at %s: Idle time reached (%s), "
                    "next due at %s", b1, tval_print (&dp->due), b2);
        }
#       endif

        /* call idle loop call back function */
        if (dp->idle != 0)
            (*dp->idle) (dp->data) ;

        /* next idle time call in dp->def.tv seconds */
        tval_nxt_tv (&dp->due, &dp->tv) ;
    }
    /*NOTREACHED*/
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
