/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: tvcalc.c 2542 2006-06-06 08:47:50Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#include "tvcalc.h"
#include "util.h"

/* ----------------------------------------------------------------------- *
 * helpers
 * ----------------------------------------------------------------------- */

#define tv_time(now) gettimeofday ((now), 0)

void tval_normalise_usec (struct timeval *tv) {
    if (tv->tv_usec >= 1000000L) {
        tv->tv_sec  += ( tv->tv_usec / 1000000L ) ;
        tv->tv_usec %=                 1000000L   ;
    }
    else if (tv->tv_usec < 0) {
        tv->tv_sec  -= 1 + (- tv->tv_usec / 1000000L ) ;
        tv->tv_usec %=                      1000000L   ;
        tv->tv_usec +=                      1000000L   ;
    }
}

/* ----------------------------------------------------------------------- *
 * timeval calculation
 * ----------------------------------------------------------------------- */

void tval_add_ms (struct timeval *sum, int msecs) {

    sum->tv_sec  +=  msecs / 1000 ;
    sum->tv_usec += (msecs % 1000) * 1000L ;
    tval_normalise_usec (sum);	
}


void tval_set_ms (struct timeval *sum, int msecs) {
    sum->tv_sec  =  msecs / 1000 ;
    sum->tv_usec = (msecs % 1000) * 1000L ;
    tval_normalise_usec (sum);	
}


// unsigned tval_lt_ms (struct timeval *val, int msecs) {
// 
//     int secs = msecs / 1000 ;
//     msecs %= 1000 ;
// 
//     return
//         (val->tv_sec  < secs) ||
//         (val->tv_sec == secs && val->tv_usec < (msecs * 1000L))
//         ;
// }


void tval_sub_tv (struct timeval *res, struct timeval *sub) {

    res->tv_sec  -= sub->tv_sec ;
    res->tv_usec -= sub->tv_usec ;
    tval_normalise_usec (res);	
}


unsigned tval_eq_tv (struct timeval *a, struct timeval *b) {
    return a->tv_sec == b->tv_sec && a->tv_usec == b->tv_usec ;
}


int tval_cmp_tv (struct timeval *a, struct timeval *b) {
    return
        ((a->tv_sec < b->tv_sec) ||
         (a->tv_sec == b->tv_sec && a->tv_sec < b->tv_sec))
        ? (-1)
        : (a->tv_sec != b->tv_sec || a->tv_usec != b->tv_usec)
        ;
}


void tval_add_tv (struct timeval *sum, struct timeval *add) {
    sum->tv_sec  += add->tv_sec  ;
    sum->tv_usec += add->tv_usec ;
    tval_normalise_usec (sum);	
}


int tval_min_tv (struct timeval *a, struct timeval *b) {
    if ((a->tv_sec == 0 && a->tv_usec == 0) || tval_gt_tv (a, b)) {
        *a = *b ;
        return -1 ;
    }
    return a->tv_sec != b->tv_sec || a->tv_usec != b->tv_usec ;
}


// int tval_max_tv (struct timeval *a, struct timeval *b) {
//     if (tval_lt_tv (a, b)) {
//         *a = *b ;
//         return 1 ;
//     }
//     return a->tv_sec != b->tv_sec || a->tv_usec != b->tv_usec ? -1 : 0;
// }


void tval_nxt_ms (struct timeval *now, int ms) {
    tv_time (now);
    if (ms) tval_add_ms (now, ms);
}


void tval_nxt_tv (struct timeval *now, struct timeval *offs) {
    tv_time (now);
    if (offs) tval_add_tv (now, offs);
}


void tval_tmo_tv (struct timeval *offs, struct timeval *when) {
    *offs = *when;
    if (when->tv_sec || when->tv_usec) {
        struct timeval now ;
        tv_time (&now);
        tval_sub_tv (offs, &now);
    }
}


int ms_tmo_tv (struct timeval *when) {
    struct timeval offs;
    tval_tmo_tv (&offs, when);
    if (offs.tv_sec < 0) { /* normalise for proper round up */
        offs.tv_sec ++ ;
        offs.tv_usec -= (1000000L + 1000L) ;
    }
    return offs.tv_sec + (offs.tv_usec + 500L) / 1000L ;
}


const char *time_print (time_t t) {
    struct tm *lt = localtime (&t);
    static char buf [100] ;
    /* 20 chars is enough space for tval_print (), below */
    snprintf (buf, sizeof (buf) - 20, "%04d-%02d-%02d %02d:%02d:%02d",
              lt->tm_year +1900, lt->tm_mon + 1, lt->tm_mday,
              lt->tm_hour, lt->tm_min, lt->tm_sec);
    return buf ;
}


const char *tval_print (struct timeval *tv) {
    const char *s = time_print (tv->tv_sec);
    char buf [20] ; /* see time_print (), above */
    snprintf (buf, sizeof (buf), ".%06ld", (long)tv->tv_usec & 0xffffffff);
    return strcat ((char*)s, buf) ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
