/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: rfc3489.c 3181 2006-09-25 10:42:20Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <time.h>

#include "util.h"
#include "ctrl.h"
#include "stun.h"
#include "settings.h"
#include "nvram-lib.h"
#include "nvalloc.h"
#include "logger.h"

#define __RFC3489_PRIVATE
#include "rfc3489.h"

/* ----------------------------------------------------------------------- *
 * Private: variables and definitions
 * ----------------------------------------------------------------------- */

/* stringify */
#define __str(x)              # x 
#define to_str(x)             __str (x)

/* dimension of a static array */
#define DIM(x) (sizeof (x) / sizeof (x [0]))

static const char *nat_state_name [] = {
    "no_NATSTATE"
#   define         __symbols(s) , to_str (s)
    STATE_SYMBOLS (__symbols)
#   undef          __symbols
} ;

/* ----------------------------------------------------------------------- *
 * Private: misc
 * ----------------------------------------------------------------------- */

static
const char **split_fields (const char *s) {

    const char **P = 0;
    unsigned     n = 0;

    goto start_here;

    do {
        s ++ ; /* skip delimiter */
        if (s [-2] == escape_character)
            continue ;

    start_here:
        P = (const char **) zrealloc (P, (n + 2) * sizeof (char*));
        P [n ++] = s ;
    }
    while ((s = strchr (s, stun_fields_delim)) != 0) ;

    P [n] = 0 ;
    return P;
}


static
ip4addr_t hostip (const char *s) {

    struct hostent* h = gethostbyname (s) ;
    struct in_addr sin_addr;

    if (h == 0) {
        extern int h_errno;
        int        e = h_errno ;
        ip4addr_t ip = ip2l (s) ;

        if (ip != 0 && ip != -1)
            return ip;

        errno = e ;
        return INADDR_NONE ;
    }

    sin_addr.s_addr = *(in_addr_t*)h->h_addr;
    return ntohl (sin_addr.s_addr);
}

static
long portnum (const char *s) {

    char *q = 0 ;
    long ln = strtol (s, &q, 10) ;

    if ((ln & ~0xffffL) != 0) {
        errno = ERANGE ;
        return LONG_MAX ;
    }

    if (q == 0 || *q != '\0')
        return LONG_MAX ;

    return ln;
}


/* ----------------------------------------------------------------------- *
 * Private: table/list look up
 * ----------------------------------------------------------------------- */
   
static
STUN_SERVER *stun_server_find (R3489 *s, ip4addr_t a) {
    STUN_SERVER **Q ;
    struct in_addr in ;
    in.s_addr = htonl (a);
    if ((Q = (STUN_SERVER **) ip4_lookup (s->stun.lst, &in)) != 0 && *Q)
        return *Q ;
    return 0 ;
 }


static
STUN_SERVER *stun_server_fetch (R3489 *s, ip4addr_t a) {
    STUN_SERVER **Q ;
    struct in_addr in ;
    in.s_addr = htonl (a);
    Q = (STUN_SERVER **) ip4_fetch (s->stun.lst, &in) ;
    if (*Q == 0)
        *Q = ZMALLOC (STUN_SERVER);
    return *Q ;
 }


static
void stun_server_delete (R3489 *s, ip4addr_t a) {
    struct in_addr in ;
    in.s_addr = htonl (a);
    ip4_delete (s->stun.lst, &in);
 }


/* ----------------------------------------------------------------------- *
 * Private: call back handlers
 * ----------------------------------------------------------------------- */

static
void record_state_response (STUN_SERVER  *q,
                            NATSTATE  state,
                            ADDR      *addr,
                            const char *txt) {
    if (q) {
        q->nat.state = state ;
        if (addr && (addr->ip4addr || addr->port))
            q->nat.src = *addr ;
    }

    if (verbosity & R3489_RESPONSE_LOG)
        logger (LOG_INFO, txt) ;
}


static
unsigned nv_update (NVCL         *nv,
                    const char  *var,
                    const char  *val,
                    const char *info) {

    nvcl_putcmd (nv, "set");
    nvcl_putavpair (nv, var, val, 0);

    if (nvcl_rply (nv, 10) > 0)
        return 1;
    
    logger (LOG_ERR, "Cannot update the NAT/STUN %s \"%s=%s\" in the NVRAM",
            info, var, val) ;

    return 0;
}

static
void nv_update_state (R3489        *s,
                      STUN_SERVER  *q,
                      NATSTATE  state,
                      int       trans) {
    if (q) {
        size_t len = strlen (q->label) ;

        const char *entry = (const char*) xmalloc
            (s->stun.pfx.len + len + s->stun.sipip.len + 1);

        char *e = memcpy ((char*)entry, s->stun.pfx.s, s->stun.pfx.len) ;
        memcpy (e += s->stun.pfx.len, q->label, len) ;
        memcpy (e += len, s->stun.sipip.s, s->stun.sipip.len + 1) ;

        if (nv_update
            (s->nv, entry, ip4addr_pp (q->nat.src.ip4addr), "ip")) {

            char buf [200] ;

            entry = (const char*) zrealloc
                ((char*)entry, s->stun.pfx.len + len + s->stun.type.len + 1);
            e =  (char*)entry + s->stun.pfx.len + len ;
            memcpy (e, s->stun.type.s, s->stun.type.len + 1) ;

            if (q->nat.state)
                sprintf (buf, "%u=%s", q->nat.state,
                         q->nat.state < DIM (nat_state_name)
                         ? nat_state_name [q->nat.state] : "<illegal-state>");
            else
                sprintf (buf, "0=#%x(%s)", trans, nat_state_name [0]);

            nv_update (s->nv, entry, buf, "state");
        }
        
        free ((char*) entry);
    }
}


static
void final_state (R3489        *s,
                  STUN_SERVER  *q,
                  NATSTATE  state,
                  ADDR      *addr,
                  const char *txt) {

    record_state_response (q, state, addr, txt);
    nv_update_state    (s, q, state, 0);
}


static
void state_transition (R3489          *s,
                       RSLT           *r,
                       ip4addr_t  dst_ip,
                       uint16_t dst_port,
                       int      test_num,
                       STUN_SERVER    *q,
                       NATSTATE    state,
                       ADDR        *addr,
                       const char   *txt) {
    RSLT *result ;

    record_state_response (q, state, addr, txt);
    nv_update_state    (s, q, state, test_num);

    result = stun_send (r, dst_ip, dst_port, test_num) ;

    if (verbosity & R3489_TRANSITION_LOG)
        logger (LOG_INFO,
                result ? " %#x?%s:%u=>%s:%u" : " (%#x?%s:%u=>%s:%u)",
                test_num,
                ip4addr_pp (r->src.ip4addr), r->src.port,
                ip4addr_pp (r->dst.ip4addr), r->dst.port);
}


static /* no response/timeout handler */
void noresponse (void *state, RSLT *r) {
    R3489 *s = state;

    switch (r->test_num) {
    case TEST_CHAIN_RFC3489|TEST_STATE_A|0:
        final_state (s, stun_server_find (s, r->lookup), UDP_Blocked, 0,
                     "NAT Check: Test I(timeout): UDP Blocked") ;
        return ;

    case TEST_CHAIN_RFC3489|TEST_STATE_B|3:
        final_state (s, stun_server_find (s, r->lookup), Sym_UDP_Firewall, 0,
                     "NAT Check: Test I=>II(timeout): Sym UDP Firewall");
        return;

    case TEST_CHAIN_RFC3489|TEST_STATE_C|3:
        state_transition (s, r, r->dst.ip4addr, r->dst.port,
                          TEST_CHAIN_RFC3489|TEST_STATE_D|0, 0, 0, 0,
                          "NAT Check: Test I=>II(timeout)/Different IP ...");
        return;

    case TEST_CHAIN_RFC3489|TEST_STATE_E|2:
        final_state
            (s, stun_server_find (s, r->lookup), Port_Restricted, 0,
             "NAT Check: Test I=>II=>I=>III(timeout): Port Restricted") ;
        return ;

    default:
        stun_server_delete (s, r->lookup);
    }
}


static /* response handler */
void response (void *state, RSLT *r) {
    R3489 *s = state;
         
    /* standard test automaton from RFC3489 */
    switch (r->test_num) {
       STUN_SERVER *q ;

    case TEST_CHAIN_RFC3489|TEST_STATE_A|0:
        /* create/update state entry */
        if ((q = stun_server_fetch (s, r->lookup)) != 0) {
            if (r->ok.mapped_address &&
                r->src.ip4addr == r->mapped_address.ip4addr &&
                r->src.port    == r->mapped_address.port)
                /* IP Same */
                state_transition
                    (s, r, r->dst.ip4addr, r->dst.port,
                     TEST_CHAIN_RFC3489|TEST_STATE_B|3, q, no_NATSTATE,
                     0, "NAT Check: Test I/Same IP ...") ;
            else
                /* not: IP Same */
                state_transition
                    (s, r, r->dst.ip4addr, r->dst.port,
                     TEST_CHAIN_RFC3489|TEST_STATE_C|3, q, no_NATSTATE,
                     &r->mapped_address, "NAT Check: Test I/behind NAT ...");
        }
        return;

    case TEST_CHAIN_RFC3489|TEST_STATE_B|3:
        /* arriving here implies: Open Internet => no SIP mapping */
        final_state (s, stun_server_find (s, r->lookup), Open_Internet, 0,
                     "NAT Check: Test I=>II: Open Internet") ;
        return;
        
    case TEST_CHAIN_RFC3489|TEST_STATE_C|3:
        /* arriving here implies: Full Cone => SIP mapping needed */
        final_state
            (s, stun_server_find (s, r->lookup), Full_Cone,
             &r->mapped_address, "NAT Check: Test I=>II: Full Cone") ;
        return;

    case TEST_CHAIN_RFC3489|TEST_STATE_D|0:
        if ((q = stun_server_find (s, r->lookup)) != 0) {
            /* IP Same ? */
            if (r->ok.mapped_address &&
                r->src.ip4addr == r->mapped_address.ip4addr &&
                r->src.port    == r->mapped_address.port)
            
                state_transition
                    (s, r, r->dst.ip4addr, r->dst.port,
                     TEST_CHAIN_RFC3489|TEST_STATE_E|2, q, no_NATSTATE, 0,
                     "NAT Check: Test I=>II=>I/Same IP ...");
            else
                /* not: IP Same */
                final_state (s, q, Symmetric_NAT, &r->mapped_address,
                             "NAT Check: Test I=>II=>I: Symmetric NAT");
        }
        return ;

    case TEST_CHAIN_RFC3489|TEST_STATE_E|2:
        final_state (s, stun_server_find (s, r->lookup), Restricted, 0,
                     "NAT Check: Test I=>II=>I=>III: Restricted");
        return ;

    default:
        logger (LOG_ERR, "Unknown test case/state (responded): %#x",
                r->test_num);
        stun_server_delete (s, r->lookup);
    }
}


/* ----------------------------------------------------------------------- *
 * Public: constructor/destructor
 * ----------------------------------------------------------------------- */

void rfc3489_destroy (R3489 *s) {
    if (s) {
        if (s->sp)           stun_close (s->sp);
        if (s->nv)           nvcl_close (s->nv);
        if (s->stun.lst) ip4_pool_close (s->stun.lst, free) ;
        if (s->stun.pfx.s)         free ((char*)s->stun.pfx.s) ;
        if (s->stun.query.s)       free ((char*)s->stun.query.s) ;
        if (s->stun.type.s)        free ((char*)s->stun.type.s) ;
        if (s->stun.sipip.s)       free ((char*)s->stun.sipip.s) ;
        free (s) ;
    }
}


R3489 *rfc3489_open (DO             *dp,
                     ip4addr_t  localip,
                     const char *nv_lst,
                     const char *nv_pfx,
                     const char *nv_qry,
                     const char *nv_tpe,
                     const char *nv_map) {

    R3489 *s = ZMALLOC (R3489);

    if ((s->sp = stun_open (dp, localip)) == 0) {
        logger (LOG_ERR, "Cannot initialise STUN subsytem");
        goto error;
    }
    if ((s->nv = nvcl_open (0,
#                           ifndef DISCARD_HTTP
                            0, 0,
#                           endif /* DISCARD_HTTP */
                            0)) == 0) {
        logger (LOG_ERR, "Cannot initialise NVRAM subsystem");
        goto error;
    }
    if ((s->stun.lst = ip4_pool_init ()) == 0) {
        logger (LOG_ERR, "Cannot initialise NVRAM cache");
        goto error;
    }
        
    s->timeout        = STUN_RESPONSE_TIMEOUT ;
    s->localip        = localip ;

    s->stun.list      = zstrdup (nv_lst) ;

    s->stun.pfx.s     = zstrdup (nv_pfx) ;
    s->stun.pfx.len   = strlen (s->stun.pfx.s) ;

    s->stun.query.s   = zstrdup (nv_qry) ;
    s->stun.query.len = strlen (s->stun.query.s) ;

    s->stun.type.s    = zstrdup (nv_tpe) ;
    s->stun.type.len  = strlen (s->stun.type.s) ;

    s->stun.sipip.s   = zstrdup (nv_map) ;
    s->stun.sipip.len = strlen (s->stun.sipip.s) ;

    rfc3489_update (s);
    return s ;

error:
    rfc3489_destroy (s) ;
    return 0 ;
}


/* update from the nvram */
unsigned rfc3489_update (R3489 *s) {
    const char *p, **P ;
    int count = 0 ;
    IP4_WALK *w ;

    nvcl_putcmd    (s->nv, "set");
    nvcl_putavpair (s->nv, s->stun.list, 0, 0);
    if (nvcl_rply      (s->nv, 10)           < 1 ||
        nvcl_getavpair (s->nv, 0, &p, 0, 0) == 0)
        return 0 ;

    /* mark list entries as "reset" */
    if ((w = ip4_walk_init (s->stun.lst, 0)) != 0) {
        STUN_SERVER **V = (STUN_SERVER **) ip4_this (w);
        do
            if (*V)
                (*V)->reset = 1 ;
        while ((V = (STUN_SERVER **) ip4_next (w)) != 0);
        ip4_walk_close (w);
    }
    
    /* parse a list of space separated token list */
    p = zstrdup (p) ;
    if ((P = split_fields (p)) != 0) {
        int f ;
        
        /* prepare variable look up */
        const char *entry = (const char*) xmalloc (s->stun.pfx.len);
        memcpy ((char*)entry, s->stun.pfx.s, s->stun.pfx.len);
        
        for (f = 0; P [f]; f ++) {
            STUN_SERVER **V, *v ;
            size_t len = P [f+1] ? P [f+1] - P [f] - 1 : strlen (P [f]);
            const char *q ;
            char *e;

            /* glue together the variable name */
            entry = (const char*) zrealloc
                ((char*)entry, s->stun.pfx.len +len +s->stun.query.len +1);
            memcpy (e = (char*)entry + s->stun.pfx.len , P [f], len) ;
            memcpy (e += len, s->stun.query.s, s->stun.query.len + 1) ;
        
            /* fetch the variable with the defs from the nvram */
            nvcl_putcmd (s->nv, "set");
            nvcl_putavpair (s->nv, entry, 0, 0);

            if (nvcl_rply (s->nv, 10) > 0 &&
                nvcl_getavpair (s->nv, 0, &q, 0, 0) > 0) {

                const char **Q = split_fields (q) ;

                /* syntax:

                    <stun_ip>[:<stun_port>][,[<src_ip>][:<src_port>]]
                */

                if (Q != 0) {
                    struct in_addr in ;
                    char *r, c = 0 ; /* "= 0" keeps the compiler happy */
                    ip4addr_t stun_ip, src_ip = 0 ;
                    long stun_port = 0, src_port = 0 ;

                    /* stun ip and port: <stun_ip>[:<stun_port>] */
                    int i = 0;
                    if (Q [i+1]) {
                        c = Q [i+1] [-1];
                        ((char**)Q) [i+1] [-1] = '\0';
                    }
                    if ((r = strchr (Q [i], ':')) != 0)
                        * r = '\0';
                    if ((stun_ip = hostip (Q [i])) == INADDR_NONE)
                        logger (LOG_WARNING, "Cannot resolve the stun host "
                                "name %s from the nvram: %s",
                                Q [i], strerror (errno));
                    else if (stun_ip == 0) {
                        logger (LOG_WARNING, "The stun host name %s from the "
                                "nvram must not be zero", Q [i]) ;
                        stun_ip = INADDR_NONE ;
                    }
                    if (r) {
                        if ((stun_port = portnum (r+1)) == LONG_MAX)
                            logger (LOG_WARNING, "Error parsing the stun port "
                                    "%s from the nvram: %s",
                                    r+1, strerror (errno));
                        * r = ':' ;
                    }
                    if (Q [i+1])
                        ((char**)Q) [i+1] [-1] = c ;
                    if (stun_ip == INADDR_NONE || stun_port == LONG_MAX)
                        goto endup ;
                    
                    /* src/local ip and port: [[<src_ip>][:<src_port>] */
                    if (Q [++ i]) {
                        if (Q [i+1]) {
                            c = Q [i+1] [-1];
                            ((char**)Q) [i+1] [-1] = '\0';
                        }
                        if ((r = strchr (Q [i], ':')) != 0)
                            * r = '\0';
                        if ((src_ip = hostip (Q [i])) == INADDR_NONE)
                            logger (LOG_WARNING, "Cannot resolve the stun "
                                    "source host name %s from the nvram: %s",
                                    Q [i], strerror (errno));
                        if (r) {
                            if ((src_port = portnum (r+1)) == LONG_MAX)
                                logger (LOG_WARNING, "Error parsing the "
                                        "source port %s from the nvram: %s",
                                        r+1, strerror (errno));
                            * r = ':' ;
                        }
                        if (Q [i+1])
                            ((char**)Q) [i+1] [-1] = c ;
                        if (src_ip == INADDR_NONE || src_port == LONG_MAX)
                            goto endup ;
                    } /* Q [++ i] */
                    
                    /* allocate the token record */
                    in.s_addr = htonl (stun_ip);
                    V = (STUN_SERVER**) ip4_fetch (s->stun.lst, &in) ;
                    if (*V) {
                        const char *keep ;
                        v = *V ;
                        if (v->label &&
                            strlen (v->label) == len &&
                            strncmp (v->label, P [f], len) == 0)
                            keep = v->label ;
                        else {
                            free ((char*)v->label);
                            keep = 0 ;
                        }
                        XZERO (*v);
                        v->label = keep ;
                    }
                    else {
                        v = ZMALLOC (STUN_SERVER) ;
                        *V = v ;
                    }

                    if (stun_port) v->stun_port     = stun_port ;
                    if (src_ip)    v->local.ip4addr =  src_ip   ;
                    if (src_port)  v->local.port    =  src_port ;

                    if (v->label == 0) {
                        v->label = (const char*)
                            memcpy (zmalloc (len+1), P [f], len);
                        ((char*)v->label) [len] = '\0' ;
                    }
                    count ++ ;

                endup:
                    free (Q);
                } /* if () */
            } /* nvcl_rply () */
        }  /* for (P []) */

        free ((char*)entry) ;
        free            (P) ;
        free     ((char*)p) ;
    } /* if P */

    if ((w = ip4_walk_init (s->stun.lst, 0)) != 0) {
        STUN_SERVER **V = (STUN_SERVER**) ip4_this (w);
        do
            if (*V && (*V)->reset) {
                free (*V) ;
                /* delete marked data records */
                stun_server_delete (s, ip4_inx2ip4addr (w)) ;
            }
        while ((V = (STUN_SERVER**) ip4_next (w)) != 0);
        ip4_walk_close (w);
    }

    return count;
}


/* ----------------------------------------------------------------------- *
 * Public: run the state machine
 * ----------------------------------------------------------------------- */

void rfc3489_run (R3489 *s, void(*say)(void*,const char*,...), void *d) {

    IP4_WALK *w ;

    if ((w = ip4_walk_init (s->stun.lst, 0)) != 0) {
        STUN_SERVER **V = (STUN_SERVER **) ip4_this (w);
        if (say)
            say (d, "+OK:");
        do {
            /* loop over the stun servers */
            ip4addr_t trg_ip   = ip4_inx2ip4addr (w) ;
            ip4addr_t cln_ip   = (*V)->local.ip4addr ;
            uint16_t  cln_port = (*V)->local.port ;
            uint16_t  trg_port = (*V)->stun_port ;

            if (cln_ip   == 0) cln_ip   = s->localip ;
            if (trg_port == 0) trg_port =  STUN_PORT ;
                                   
            RSLT *r = stun_session (s->sp, trg_ip,
                                    response, s,
                                    noresponse, s, s->timeout,
                                    trg_ip, trg_port,
                                    TEST_CHAIN_RFC3489|TEST_STATE_A|0,
                                    cln_ip, cln_port);
            if (say) {
                if (r)
                    say (d, " %#x?%s:%u=>%s:%u",
                         TEST_CHAIN_RFC3489|TEST_STATE_A|0,
                         ip4addr_pp (r->src.ip4addr), r->src.port,
                         ip4addr_pp (r->dst.ip4addr), r->dst.port);
                else
                    say (d, " (%#x?%s=>%s:%u)",
                         TEST_CHAIN_RFC3489|TEST_STATE_A|0,
                         ip4addr_pp (cln_ip),
                         ip4addr_pp (trg_ip), trg_port);
                continue;
            }
                
            if (r) {
                if (R3489_SEND_LOG & verbosity)
                    logger (LOG_INFO, "Sent %#x?%s:%u=>%s:%u",
                            TEST_CHAIN_RFC3489|TEST_STATE_A|0,
                            ip4addr_pp (cln_ip), r->src.port,
                            ip4addr_pp (trg_ip), trg_port);
                continue;
            }

            logger (LOG_WARNING, "Sending %#x?%s=>%s:%u failed",
                    TEST_CHAIN_RFC3489|TEST_STATE_A|0,
                    ip4addr_pp (cln_ip),
                    ip4addr_pp (trg_ip), trg_port);
        }
        while ((V = (STUN_SERVER **) ip4_next (w)) != 0);
        ip4_walk_close (w);

        if (say)
            say (d, "\n");
    }
    else {
        const char blurb [] = "no STUN entries cached from NVRAM\n";
        if (say)
            say (d, "-ERR: %s", blurb);
        else
            logger (LOG_WARNING, "rfc3489_run: %s", blurb);
        return ;
    }

    /* set the time when the next test should run */
    s->start_test = s->recur ? time (0) + s->recur : 0 ;
}


void rfc3489_auto (R3489 *s) {
    /* run it when it is time */
    if (s->start_test && s->start_test <= time (0)) {
        if (R3489_AUTOMATIC_LOG & verbosity)
            logger (LOG_INFO, "Starting the rfc3489 state machine");
        rfc3489_run (s, 0, 0);
    }
}

/* ----------------------------------------------------------------------- *
 * Public: misc/maintenance
 * ----------------------------------------------------------------------- */

void rfc3489_show (R3489 *s, void (*say) (void*,const char*,...), void *d) {
    IP4_WALK *w;

    if (s->start_test)
        say (d, "Next automated NAT test due: %s", ctime (&s->start_test));
    say (d, "Automated NAT test recurrence: %us\n", s->recur);
    say (d, "NAT test reply timeout: %us\n", s->timeout);

    if ((w = ip4_walk_init (s->stun.lst, 0)) != 0) {
        STUN_SERVER *v, **V = (STUN_SERVER**) ip4_this (w);
        int intro = 0 ;
        do
            if ((v = *V) != 0) {
                if (intro ++ == 0)
                    say (d, "STUN/NAT registry:");
                say (d, " %s=%s:%u(%#x)%s", v->label,
                     ip4addr_pp (v->nat.src.ip4addr), v->nat.src.port,
                     v->nat.state,
                     ip4addr_pp (ip4_inx2ip4addr (w)));
                if (v->stun_port)
                    say (d, ":%u", v->stun_port);
                if (v->local.ip4addr) {
                    say (d, ",%s", ip4addr_pp (v->local.ip4addr));
                    if (v->local.port)
                        say (d, ":%u", v->local.port);
                }
                else
                    if (v->local.port)
                        say (d, ",:%u", v->local.port);
            }
        while ((V = (STUN_SERVER**) ip4_next (w)) != 0);
        ip4_walk_close (w);
        if (intro) 
            say (d, "\n");
    }
}


void rfc3489_tmo (R3489 *s, unsigned timeout, unsigned recur) {
    if (timeout != -1)
        s->timeout = timeout ;
    if (recur != -1)
        s->recur = recur ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
