/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: stun.c 1981 2006-04-05 20:30:24Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netdb.h>

#include "util.h"
#include "do.h"
#include "settings.h"
#include "nvalloc.h"
#include "ipcalc.h"
#include "logger.h"

#define __STUN_PRIVATE
#include "stun.h"

/* ----------------------------------------------------------------------- *
 * Macros etc.
 * ----------------------------------------------------------------------- */

#define DIM(x) (sizeof (x) / sizeof (x [0]))

/* ----------------------------------------------------------------------- *
 * Helpers
 * ----------------------------------------------------------------------- */

static
unsigned parse_address (ADDR          *a,
                        char     *a_flag,
                        const char    *p,
                        size_t       len,
                        const char *info) {
    if (len == 8) {
        p ++; /* skip padding */

        if (* p ++ == IPV4FAMILY) {
            a->port    = ldx16b     (p);
            a->ip4addr = ip2ip4addr (p+2);
            *a_flag    = 1 ;

            if (verbosity & STUN_MSG_PARSER_LOG) {
                if (info == 0) info = "got" ;
                logger (LOG_NOTICE, "Response: %s address = %s:%u",
                        info, ip4addr_pp (a->ip4addr), (unsigned)a->port);
            }

            return 1;
        }
   
        if (* -- p == IPV6FAMILY)
            logger (LOG_WARNING, "Response: ipv6 is unsupported");
        else
            logger (LOG_WARNING, "Response: bad address family: %#x", *p);
    }
    else
        logger (LOG_WARNING,
                "Response: Wrong address header length %d (expecting 8)",
                len);

    if (info == 0) info = "given" ;
    logger (LOG_WARNING, "Response: Problem parsing the %s address", info);
    return 0;
}


static 
unsigned parse_string (TEXT          *t,
                       char     *t_flag,
                       const char    *p,
                       size_t       len,
                       const char *info) {
   
    if (len < sizeof (t->txt)-1) {
        if (len % 4 == 0) {
            memcpy (t->txt, p, t->len = len);
            t->txt [len] = '\0';
            *t_flag      =   1 ;
 
            if (verbosity & STUN_MSG_PARSER_LOG) {
                if (info == 0) info = "text" ;
                logger (LOG_NOTICE, "Response: %s = %s", info, t->txt);
            }
 
            return 1;
        }
        logger (LOG_WARNING,
                "Response: No mod4 string length header %d", len);
    }
    else
        logger (LOG_WARNING,
                "Response: header size %d of string too large (at most %d)",
                len, sizeof (t->txt));
 
    if (info == 0) info = "text" ;
    logger (LOG_WARNING, "Response: Problem parsing the %s", info);
    return 0;
}

/* ----------------------------------------------------------------------- *
 * Response handler
 * ----------------------------------------------------------------------- */

static
void response_handler (void *h) {
    struct sockaddr_in from;
    char *p, buf [MAX_MESSAGE_SIZE] ;
    size_t len ;

    int i, j, n = sizeof (from);
    PEND *u = h ;

    /* accept message frame */
    if ((len = recvfrom (u->fd, buf, sizeof (buf) - 1, 0,
                         (struct sockaddr *) &from,
                         (socklen_t       *) &n)) < 20) {

        logger (LOG_WARNING, "Error receiving STUN data on fd=%d: %s", u->fd,
                n < 0 ? strerror (errno) : "got less than 20 bytes");
        goto done ;
    }
    buf [len] = '\0';

    /* evaluate response */
    switch (u->r.type = n16toh (buf)) {
    case BIND_RESPONSE_MSG:
    case BIND_ERROR_RESPONSE_MSG:
    case SHARED_SECRET_RESPONSE_MSG:
    case SHARED_SECRET_ERROR_RESPONSE_MSG:
        break;
    default:
        logger (LOG_WARNING, "Unexpected STUN response type %#x on fd=%d",
                (int)u->r.type, u->fd);
        return ;
    }
    if ((n = n16toh (buf+2)) != (len -= 20)) {
        logger (LOG_WARNING, "Error receiving STUN data on fd=%d: bogous "
                "frame payload length %u (should be %d)", u->fd, n, len);
        return ;
    }
    if (memcmp (buf+4, &u->id, 16) != 0) {
        logger (LOG_WARNING, "Error receiving STUN data on fd=%d: %s",
                u->fd, "wrong transaction id");
        return;
    }
    
    /* collect sender infos */
    u->r.rsp.ip4addr = ntohl (from.sin_addr.s_addr);
    u->r.rsp.port    = ntohs (from.sin_port);

    for (p = buf + 20; len > 0; /* n/a */) {
        int   atype = n16toh (p) ; p += 2 ;
        size_t alen = n16toh (p) ; p += 2 ;

        if (alen + 4 > len) {
            logger (LOG_WARNING, "Attribute %#x is larger (%u) than the "
                    "remaining message size %u", atype, alen + 4, len); 
            goto done ;
        }
        len -= 4;
                 
        switch (atype) {
        case MAPPED_ADDRESS:
            if (parse_address (&u->r.mapped_address,
                               &u->r.ok.mapped_address,
                               p, alen, "mapped") == 0)
                goto done;
            break;
        case RESPONSE_ADDRESS:
            if (parse_address (&u->r.response_address,
                               &u->r.ok.response_address,
                               p, alen, "response") == 0)
                goto done;
            break;    
        case SOURCE_ADDRESS:
            if (parse_address (&u->r.source_address,
                               &u->r.ok.source_address,
                               p, alen, "source") == 0)
                goto done;
            break;    
        case CHANGED_ADDRESS:
            if (parse_address (&u->r.changed_address,
                               &u->r.ok.changed_address,
                               p, alen, "changed") == 0)
                goto done;
            break;    
        case XORMAPPED_ADDRESS:
            if (parse_address (&u->r.xor_mapped_address,
                               &u->r.ok.xor_mapped_address,
                               p, alen, "xor-mapped") == 0)
                goto done;
            break;    
        case SECONDARY_ADDRESS:
            if (parse_address (&u->r.secondary_address,
                               &u->r.ok.secondary_address,
                               p, alen, "secondary") == 0)
                goto done;
            break;    
        case REFLECTED_FROM:
            if (parse_address (&u->r.reflected_from,
                               &u->r.ok.reflected_from,
                               p, alen, "reflected-from") == 0)
                goto done;
            break;    
//        case USERNAME:
//            if (parse_string (&u->r.uname,
//                              &u->r.ok.uname,
//                              p, alen, "user name") == 0)
//                goto done;
//            break;
//        case PASSWORD:
//            if (parse_string (&u->r.passwd,
//                              &u->r.ok.passwd,
//                              p, alen, "password") == 0)
//                goto done;
//            break;
        case SERVERNAME:
            if (parse_string (&u->r.server_name,
                              &u->r.ok.server_name,
                              p, alen, "server name") == 0)
            break;
                goto done;
        case ERROR_CODE:
            if (alen < 4) {
                logger (LOG_WARNING,
                        "Response: Error code message length %d too small",
                        len);
                goto done ;
            }
            if (alen >= 4 + sizeof (u->r.error_code.txt)) {
                logger (LOG_WARNING,
                        "Response: Error message length %u too large "
                        "(at most %u)",
                        alen - 4, sizeof (u->r.error_code.txt) -1);
                goto done ;
            }
            u->r.error_code.class  = p [2] & 0x7 ;
            u->r.error_code.number = p [3] ;
            u->r.error_code.len    = alen - 4;
            memcpy (u->r.error_code.txt, p + 4, alen - 4);
            u->r.error_code.txt [alen - 4] = '\0' ;
         
            if (verbosity & STUN_MSG_PARSER_LOG)
                logger (LOG_NOTICE, "Response: Error code = %d %d %s",
                        u->r.error_code.class,
                        u->r.error_code.number,
                        u->r.error_code.txt);
            break;
        case MESSAGE_INTEGRITY:
            if (alen != 20) {
                logger (LOG_WARNING, "Response: Bogus hmac length %d", alen);
                goto done ;
            }
            break ;
        case XOR_ONLY:
            if (alen != 0) {
                logger (LOG_WARNING,
                        "Response: Bogus xor flag length %d", alen);
                goto done ;
            }
            if (verbosity & STUN_MSG_PARSER_LOG)
                logger (LOG_NOTICE, "Response: xorOnly = true");
            u->r.ok.xor_only = 1;
            break;
        case UNKNOWN_ATTRIBUTES:
            if (alen < 4 || (alen & 3) == 0) {
                if (alen < 2) {
                    logger (LOG_WARNING,
                            "Response: Unknown attributes list too short");
                    goto done ;
                }
                logger (LOG_WARNING,
                        "Response: Unknown attributes list has size %u "
                        "(mod 4 expected)", alen);
                goto done ;
            }
            if ((n = alen >> 2) > DIM (u->r.unknown_attribs.type)) {
                logger (LOG_WARNING,
                        "Response: Unknown attributes list has %u entries "
                        "(at most %u allowed)",
                        n, DIM (u->r.unknown_attribs.type));
                goto done ;
            }
            for (i = 0; i < n; i ++) {
                int type = ldx16b (p + 2*i);
                for (j = 0; j <= i; j ++)
                    if (u->r.unknown_attribs.type [j] == type)
                        goto next ;
                u->r.unknown_attribs.type
                    [u->r.unknown_attribs.num ++] = type;
            next:
                ;
            }
            break;
        default:
            if (verbosity & STUN_MSG_PARSER_LOG)
                logger (LOG_NOTICE,
                        "Response: Unknown attribute: %d", atype);
            if (atype <= 0x7FFF)
                goto done ;
        }
        p   += alen ;
        len -= alen ;
    }
    
    /* user exit - register result */
    if (u->cb) {
        u->sent_new_data = 0;
        (*u->cb) (u->cb_data, &u->r);
        if (u->sent_new_data != 0)
            return ;
    }

done:
    /* unregister and clean up */
    stun_session_delete (u->master, u->fd) ;
}



static
void timeout_handler (void *h) {
    PEND *u = h ;

    if (u->to) {
        u->sent_new_data = 0 ;
        (*u->to) (u->to_data, &u->r);
        if (u->sent_new_data != 0)
            return ;
    }
    else if (verbosity & STUN_TIMEOUT_LOG)
        logger (LOG_NOTICE, "Timeout quering %s:%u",
                ip4addr_pp (u->r.dst.ip4addr), u->r.dst.port) ;

    /* unregister and clean up */
    stun_session_delete (u->master, u->fd) ;
}


static
PEND *stun_session_data (RSLT *r) {
    return (PEND*) ((char*)r - ((char*)&(((PEND*)0L)->r) - (char*)0L));
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

SREQ *stun_open (DO *dp, ip4addr_t ip) {

    if (dp != 0) {
        SREQ   sr  ;
        XZERO (sr) ;
        sr.dp     = dp ;
        sr.dev_ip = ip ;
        return zmemdup (&sr);
    }

    return 0;
}


void stun_close (SREQ *sr) {
    if (sr == 0)
        return ;

    while (sr->pending != 0) {
        PEND *s = sr->pending->next ;
        if (sr->pending->fd >= 0) {
            close (sr->pending->fd) ;
            undo_action (sr->dp, sr->pending->fd);
        }
        free (sr->pending) ;
        sr->pending = s ;
    }
}


/* ----------------------------------------------------------------------- *
 * Session
 * ----------------------------------------------------------------------- */

int stun_session_delete (SREQ *sr,
                         int   fd) {

    PEND *p, **P = &sr->pending ;

    while ((p = *P) != 0) {
        if (p->fd == fd) {
            *P = p->next ;
            if (p->fd >= 0) {
                close (p->fd) ;
                undo_action (sr->dp, p->fd);
            }
            free (p);
            return fd;
        }
        P = &p->next ;
    }

    logger (LOG_WARNING, "No such session to delete: fd =%d", fd);
    return -1;
}


RSLT *stun_send (RSLT           *r,
                 ip4addr_t  dst_ip,
                 uint16_t dst_port,
                 int      test_num) {
    
    /* get session record in list */
    PEND *pnd = stun_session_data (r);
    SREQ  *sr = pnd->master ;
        
    char buf [MAX_MESSAGE_SIZE], *p ;
    struct sockaddr_in addr;
    int n, chg_ip_port ;

    /* prepare for the test case */
    switch ((pnd->r.test_num = test_num) & 0xf) {
    case 0: chg_ip_port =                                 0 ; break ;
    case 1: chg_ip_port = CHANGE_IP_FLAG                    ; break ;
    case 2: chg_ip_port =                  CHANGE_PORT_FLAG ; break ;
    case 3: chg_ip_port = CHANGE_IP_FLAG | CHANGE_PORT_FLAG ; break ;
    default:
        logger (LOG_WARNING, "No test valid case spcified: %#x", test_num);
        return 0;
    }

    /* create a new message id */
    rndata (((char*) & pnd->id) + sizeof (unsigned),
            sizeof    (pnd->id) - sizeof (unsigned));
    ((unsigned*) & pnd->id) [0] = ~(++ sr->msg_id); /* unique in sequ. */

    /* write message header */
    stx16b (buf, BIND_REQUEST_MSG); p = buf + 4; /* second word left free */
    p = memcpy (p, (char*)&pnd->id, sizeof (pnd->id)) + sizeof (pnd->id);

    /* change request */
    stx16b  (p, CHANGE_REQUEST); p += 2 ;
    * p ++ = 0 ; /* msb(4) */
    * p ++ = 4 ; /* lsb(4) */
    ip4addr2ip (p, chg_ip_port); p += 4 ;

    /* second word from the beginning: store length */
    n = p - buf - 20 ;
    stx16b (buf+2, n) ;

    /* send this message */
    XZERO (addr);
    addr.sin_family      = AF_INET;
    addr.sin_port        = htons (dst_port);
    addr.sin_addr.s_addr = htonl (dst_ip);

    if ((pnd->sent_new_data =
         sendto (pnd->fd, buf, n + 20, 0, /* watch out: <= 0 */
                 (struct sockaddr*) &addr, sizeof (addr))) <= 0) {
        logger (LOG_WARNING, "Could not sent %s:%u: %s",
                ip4addr_pp (dst_ip), dst_port, strerror (errno));
        return 0;
    }
    pnd->r.dst.ip4addr = dst_ip   ;
    pnd->r.dst.port    = dst_port ;

    /* set automatic deregistering on timeout */
    if (pnd->timeout)
        do_expire (sr->dp, pnd->fd, pnd->timeout, timeout_handler, pnd);
    
    if (verbosity & STUN_NEWSESSION_LOG)
        logger (LOG_INFO, "Sent %d bytes message to %s:%u",
                n, ip4addr_pp (dst_ip), dst_port);

    return r;
}


RSLT *stun_session (SREQ                 *sr,
                    ip4addr_t         lookup,
                    void (*cb) (void*,RSLT*), /* result handler */
                    void            *cb_data,
                    void (*to) (void*,RSLT*), /* timeout handler */
                    void            *to_data,
                    unsigned         timeout,
                    ip4addr_t         dst_ip,
                    uint16_t        dst_port,
                    int             test_num,
                    ip4addr_t         src_ip,
                    uint16_t        src_port) {
    PEND pnd ;
    struct sockaddr_in addr;

    if (src_ip == 0)
        src_ip = sr->dev_ip ;
    if (src_port == 0) {
        rndata ((char*)&src_port, sizeof (src_port));
        src_port |= 0x4000;
        src_port &= 0x7FFF;
    }

    /* initialise the communication reccord */
    XZERO (pnd);

    /* open socket */
    if ((pnd.fd = socket (PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
        logger (LOG_WARNING, "Could not create a UDP socket: %s",
                strerror (errno));
        return 0;
    }
    
    XZERO (addr);
    addr.sin_family      = AF_INET;
    addr.sin_addr.s_addr = htonl (INADDR_ANY);
    addr.sin_port        = htons (src_port);
    
    if (src_ip && src_ip != 0x100007f) {
        int one = 1 ;

        addr.sin_addr.s_addr = htonl (src_ip);
        if (verbosity & STUN_NEWSESSION_LOG)
            logger (LOG_INFO, "Binding to source ip address %s:%u",
                    ip4addr_pp (src_ip), src_port);

        if (setsockopt
            (pnd.fd, SOL_SOCKET, SO_REUSEADDR,
             (const char*)&one, sizeof one) < 0)
            logger (LOG_WARNING, "Could not prepare the socket to "
                    "%s:%u for bind reuse: %s",
                    ip4addr_pp (src_ip), src_port, strerror (errno));
    }
    	
    if (bind (pnd.fd, (struct sockaddr*)&addr, sizeof (addr)) != 0) {
        logger (LOG_WARNING, "Could not bind socket to %s:%u: %s",
                ip4addr_pp (src_ip), src_port, strerror (errno));
        return 0;
    }
   
    pnd.r.src.ip4addr = src_ip   ;
    pnd.r.src.port    = src_port ;

    pnd.r.lookup =  lookup ;
    pnd.timeout  = timeout ;
    pnd.cb       = cb      ; /* response handler */
    pnd.cb_data  = cb_data ;
    pnd.to       = to      ; /* timeout handler */
    pnd.to_data  = to_data ;

    /* install/push the communication record */
    pnd.next     = sr->pending ;
    pnd.master   = sr ;
    sr->pending  = (PEND*) zmemdup (&pnd);

    if (stun_send (&sr->pending->r, dst_ip, dst_port, test_num) == 0) {
        free (sr->pending);
        sr->pending = pnd.next ;
        close (pnd.fd) ;
        return 0;
    }
    
    /* register this message for answer reception */
    do_action (sr->dp, pnd.fd, response_handler, sr->pending);

    return &sr->pending->r ;
}


/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
