/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: stun.h 743 2005-12-10 16:43:51Z jordan $
 */

#ifndef __STUN_H
#define __STUN_H

#include <stdio.h>
#include <time.h>
#include "do.h"
#include "ipcalc.h"

#define MAX_MESSAGE_SIZE                    2048

#define IPV4FAMILY                       (0x0001)
#define IPV6FAMILY                       (0x0002)

/* flags */
#define CHANGE_IP_FLAG                   (0x0004)
#define CHANGE_PORT_FLAG                 (0x0002)

/* attributes */
#define MAPPED_ADDRESS                   (0x0001)
#define RESPONSE_ADDRESS                 (0x0002)
#define CHANGE_REQUEST                   (0x0003)
#define SOURCE_ADDRESS                   (0x0004)
#define CHANGED_ADDRESS                  (0x0005)
#define USERNAME                         (0x0006)
#define PASSWORD                         (0x0007)
#define MESSAGE_INTEGRITY                (0x0008)
#define ERROR_CODE                       (0x0009)
#define UNKNOWN_ATTRIBUTES               (0x000A)
#define REFLECTED_FROM                   (0x000B)
#define XORMAPPED_ADDRESS                (0x0020)
#define XOR_ONLY                         (0x0021)
#define SERVERNAME                       (0x0022)
#define SECONDARY_ADDRESS                (0x0050)

/* message types */
#define BIND_REQUEST_MSG                 (0x0001)
#define BIND_RESPONSE_MSG                (0x0101)
#define BIND_ERROR_RESPONSE_MSG          (0x0111)
#define SHARED_SECRET_REQUEST_MSG        (0x0002)
#define SHARED_SECRET_RESPONSE_MSG       (0x0102)
#define SHARED_SECRET_ERROR_RESPONSE_MSG (0x0112)


typedef struct _addr {
    ip4addr_t ip4addr ;
    uint16_t     port ;
} ADDR ;

typedef struct _text {
    short         len ;
    char    txt [256] ;      
} TEXT ;

typedef struct _creq {
    ip4addr_t ip4addr ;
} CREQ ;

typedef struct _errc {
    uint8_t     class ;
    uint8_t    number ;
    short         len ;
    char    txt [256] ;
} ERRC ;

typedef struct _aany {
    short         num ;
    uint16_t type [8] ;
} AANY ;


typedef struct _rslt {

    /* response message */
    uint16_t               type;

    /* state machine */
    ip4addr_t            lookup; /* session id */
    unsigned           test_num; /* session state */

    struct {
        char     mapped_address;
        char   response_address;
        char     source_address;
        char    changed_address;
        char xor_mapped_address;
        char  secondary_address;
        char     reflected_from;
//      char              uname;
//      char             passwd;
        char        server_name;
        char         error_code;
        char    unknown_attribs;
        char           xor_only;
    } ok ;
    
    ADDR         mapped_address;
    ADDR       response_address;
    ADDR         source_address;
    ADDR        changed_address;
    ADDR     xor_mapped_address;
    ADDR      secondary_address;
    ADDR         reflected_from;

//  TEXT                  uname;
//  TEXT                 passwd;
    TEXT            server_name;

    ERRC             error_code;
    AANY        unknown_attribs;

    /* sent and received addresses */
    ADDR                    src; /* sent from */
    ADDR                    dst; /* sent to */
    ADDR                    rsp; /* received from */
} RSLT ;


#ifdef __STUN_PRIVATE
typedef struct _pend {
    /* request descriptor, io channel */
    int                     fd;
    uint128_t               id; /* transaction id */
    int          sent_new_data;

    /* collected request and response */
    RSLT                     r;
    
    /* back reference to master descriptor */
    struct _sreq       *master;

    /* forward reference to the next item in the chain */
    struct _pend         *next;

    /* custom result handlers */
    void  (*cb) (void*, RSLT*); /* response */
    void              *cb_data;
    unsigned           timeout;
    void  (*to) (void*, RSLT*); /* timeout */
    void              *to_data;
} PEND ;
#endif


typedef struct _sreq {
#   ifdef __STUN_PRIVATE
    DO           *dp ; /* select/call back handler */
    PEND    *pending ; /* queries awaiting answers */
    ip4addr_t dev_ip ; /* default ip */
    unsigned  msg_id ; /* last message id, sent */
#   else
    char any ;
#   endif
} SREQ ;


extern SREQ *stun_open           (DO*,ip4addr_t);
extern void  stun_close          (SREQ*);
extern int   stun_session_delete (SREQ*, int);
extern RSLT *stun_send           (RSLT*,ip4addr_t,uint16_t,int);
extern RSLT *stun_session        (SREQ*,
                                  ip4addr_t,
                                  void(*)(void*,RSLT*),void*,
                                  void(*)(void*,RSLT*),void*,
                                  unsigned,ip4addr_t,uint16_t,
                                  int,ip4addr_t,uint16_t);

#endif /* __STUN_H */
