/*
 * Copyright (C) 2003-2005 Maxina GmbH - Jordan Hrycaj
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy.net.com>
 *
 * $Id: stuncd.c 3181 2006-09-25 10:42:20Z jordan $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define _GNU_SOURCE
#include <getopt.h>

// #include "pool.h"

#include "device.h"
#include "do.h"
#include "varstr.h"
#include "nvalloc.h"
#include "ctrl.h"
#include "util.h"
#include "comhandler.h"
#include "rfc3489.h"
#include "settings.h"
#include "logger.h"

/* ----------------------------------------------------------------------- *
 * variables and definitions
 * ----------------------------------------------------------------------- */

#ifdef _DEBUG
#define _MAIN_DEBUG
#endif

typedef
struct _desc {
    DO     *loop ;
    COMM   *comm ;
    R3489 *r3489 ;
    ip4addr_t ip ;
} SESSION ;


extern char *optarg;
extern int optind, opterr, optopt;

static jmp_buf go_here;           /* jump buffer for long jumps */

/* ----------------------------------------------------------------------- *
 * Fiddling around with options ...
 * ----------------------------------------------------------------------- */

#define __arg1of4(a,b,c,d)    a
#define __arg2of4(a,b,c,d)    b
#define __arg3of4(a,b,c,d)    c
#define __arg4of4(a,b,c,d)    d
#define __str(x)              # x
#define str(x)                __str (x)

/* I am fed up with option name and option key management using strings
   and letters at several places - so keep them together -- jordan */

#define opt_name(x)         x (__arg1of4)
#define opt_key(x)          x (__arg2of4)
#define opt_argtype(x)      x (__arg3of4)
#define opt_info(x)         x (__arg4of4)

#define OPT_VERSION(x)      x ("version",       'V', 0,           "print version and exit")
#define OPT_VERBOSE(x)      x ("verbose",       'v', 0,           "print extra diagnostic messages")
#define OPT_HELP(x)         x ("help",          'h', 0,           "instead of a manual page")
#define OPT_SYSLOG(x)       x ("syslog",        's', "FACILITY",  "use syslog with the given facility")
#define OPT_NOTIMEPFX(x)    x ("no-timestamp",  'n', 0,           "do not prefix log data by a time stamp")
#define OPT_DETACH(x)       x ("detach",        'd', 0,           "daemon mode, put yourself in the background")
#define OPT_PID_FILE(x)     x ("pid-file",      'p', "FILE",      "save the process pid")
#define OPT_CHROOT(x)       x ("chroot",        'r', "DIR",       "chroot to DIR after initialisation (will not restore /proc settings)")
#define OPT_GROUPID(x)      x ("groupid",       'g', "GID",       "allow this group to connect and send admin requests")
#define OPT_STUN_DEVICE(x)  x ("stun-device",   'I', "NAME",      "the STUN interface name (default=" DEFAULT_STUNBOUND ")")
#define OPT_STUN_ADDRESS(x) x ("stun-address",  'A', "IP",        "the STUN interface address (default=" STUN_DEFAULT_IP ")")
#define OPT_NULL(x)         x (0,               0 , 0,            0)


#define LONG_OPTION_RECORD(x,y)  {opt_name (x), y, 0, opt_key (x)}
static const
struct option long_options [] = {
    LONG_OPTION_RECORD (OPT_VERSION,      0),
    LONG_OPTION_RECORD (OPT_VERBOSE,      0),
    LONG_OPTION_RECORD (OPT_NOTIMEPFX,    0),
#   ifndef DISCARD_INFO
    LONG_OPTION_RECORD (OPT_HELP,         0),
#   endif /* DISCARD_INFO */
    LONG_OPTION_RECORD (OPT_GROUPID,      1),
    LONG_OPTION_RECORD (OPT_DETACH,       0),
    LONG_OPTION_RECORD (OPT_CHROOT,       1),
    LONG_OPTION_RECORD (OPT_PID_FILE,     1),
    LONG_OPTION_RECORD (OPT_STUN_DEVICE,  1),
    LONG_OPTION_RECORD (OPT_STUN_ADDRESS, 1),
    LONG_OPTION_RECORD (OPT_SYSLOG,       1),
    LONG_OPTION_RECORD (OPT_NULL,         0)
};
#undef LONG_OPTION_RECORD


static const
char short_options [] = {
    opt_key (OPT_VERSION),
    opt_key (OPT_VERBOSE),
    opt_key (OPT_NOTIMEPFX),
#   ifndef DISCARD_INFO
    opt_key (OPT_HELP),
#   endif /* DISCARD_INFO */
    opt_key (OPT_GROUPID),      ':',
    opt_key (OPT_DETACH),
    opt_key (OPT_CHROOT),       ':',
    opt_key (OPT_PID_FILE),     ':',
    opt_key (OPT_STUN_DEVICE),  ':',
    opt_key (OPT_STUN_ADDRESS), ':',
    opt_key (OPT_SYSLOG),       ':',
    opt_key (OPT_NULL)
} ;

/* ----------------------------------------------------------------------- *
 * Timeout signal handler
 * ----------------------------------------------------------------------- */
static void s_go_here (int n) { longjmp (go_here, 1); }

/* ----------------------------------------------------------------------- *
 * call back functions
 * ----------------------------------------------------------------------- */

static
void idle (void *dsc) {
    SESSION *s = dsc ;

    if (verbosity & MAIN_LOG_IDLE_LOOPING)
        logger (LOG_INFO, "Idle handler.");

    rfc3489_auto (s->r3489);
}


#if 0
static
void resched (void *dsc, int timeout) {  /* reschedule idle timer */
    SESSION *s = dsc ;

    if (verbosity & MAIN_LOG_IDLE_LOOPING)
        logger (LOG_INFO, "Rescheduling the idle timer: %d.", timeout);

    /* the scheduler wants millisecs while the arp handler knows seconds */
    do_resched (s->loop, timeout * 1000);
}
#endif /* 0 */

/* ----------------------------------------------------------------------- *
 * helpers
 * ----------------------------------------------------------------------- */

static
void cpry (void) {
    logger (LOG_INFO, "STUNCD - STUN client over NVRAM");
    logger (LOG_INFO, "Copyright (c) 2003-2005 Maxina GmbH -- GPL-2 licenced");
    logger (LOG_INFO, "Version " VERSION " -- Jordan Hrycaj <jordan@mjh.teddy-net.com>");
}

#ifndef DISCARD_INFO
static
void usage_line (const char indent,
                 const  char *name,
                 char          key,
                 const char   *arg,
                 const char  *info) {
    char longopt [100], format [30] ;
    strcpy (longopt, name) ;
    if (arg) { strcat (longopt, "="); strcat (longopt, arg); }
    sprintf (format, "%%%us-%%c, --%%-24s %%s\n", indent);
    fprintf (stderr, format, "", key, longopt, info);
}
#endif /* DISCARD_INFO */


static
void usage (const char *info, const char *opt) {

#   ifndef DISCARD_INFO
    if (info != 0) {
        fprintf (stderr, "\n%s: ERROR: %s", progname, info);
        if (opt) fprintf (stderr, " (option arg was %s)", opt);
        fprintf (stderr, "\n");
    }

#   define SHOW_OPTION(indent,opt) usage_line (indent,            \
                                               opt_name    (opt), \
                                               opt_key     (opt), \
                                               opt_argtype (opt), \
                                               opt_info    (opt)  )

    fprintf (stderr, "\nUsage: %s [options]\n", progname);

    fputs ("\nGeneral options:", stderr);
    SHOW_OPTION ( 4, OPT_VERBOSE);
    SHOW_OPTION (20, OPT_HELP);
    SHOW_OPTION (20, OPT_SYSLOG);
    SHOW_OPTION (20, OPT_CHROOT);
    SHOW_OPTION (20, OPT_NOTIMEPFX);
    SHOW_OPTION (20, OPT_DETACH);
    SHOW_OPTION (20, OPT_PID_FILE);
    SHOW_OPTION (20, OPT_GROUPID);

    fputs ("\nSTUN processing:", stderr);
    SHOW_OPTION ( 4, OPT_STUN_DEVICE);
    SHOW_OPTION (20, OPT_STUN_ADDRESS);
 
    if (info == 0 && opt == 0) fputs
        ("\n"
         " ...\n"
         "",  stderr);

#   undef SHOW_OPTION
#   endif /* DISCARD_INFO */

    exit (1);
}


#ifndef DISCARD_INFO
static
void help (void) {
    fputs ("\n", stderr);
    syslog_init (0, 0); /* no timestamps with the introduction */
    cpry ();
    fputs ("\n"
           "\n ...\n"
           "",  stderr);
}
#endif /* DISCARD_INFO */


/* ----------------------------------------------------------------------- *
 * MAIN
 * ----------------------------------------------------------------------- */

#ifndef _MAIN_DEBUG
#undef main /* in case some debugging redirection applied */
#endif

int main (int argc, char *argv[]) {

    int n, c, option_inx ;

    char        *new_root =  0 ;
    char       *facility  =  0 ;
    int           timepfx =  1 ;
    int            pid_fd = -1 ;
    int            detach =  0 ;
    int           verbose =  0 ;
    int      idle_timeout =  MAIN_IDLE_TIMEOUT ;
    char        *stun_dev =  0 ;
    in_addr_t   stun_addr =  0 ;
    unsigned      set_gid =  0 ;
    CTRL            *ctrl =  0 ;

    SESSION s ;
    XZERO  (s);

    /* progname ::= only the last path segment */
    if ((progname = strrchr (argv [0], '/')) == 0)
        progname = argv [0] ;
    else
        ++ progname ;
    
    while ((c = getopt_long (argc, argv, short_options, 
                             long_options, &option_inx)) != -1)
        switch (c) {
            struct passwd *pwd;
        case opt_key (OPT_GROUPID):
            if ((parse_unsigned (& set_gid, optarg, 0) == 0 ||
                 (pwd = getpwuid (set_gid)) == 0) && 
                (pwd = getpwnam (optarg)) == 0)
                usage ("group name dooes not exist", optarg);
            set_gid = pwd->pw_gid;
            continue ;
        case opt_key (OPT_CHROOT):
            new_root = optarg;
            continue ;
        case opt_key (OPT_NOTIMEPFX):
            timepfx = 0 ;
            continue ;
        case opt_key (OPT_DETACH):
            detach = 1 ;
            continue ;
        case opt_key (OPT_PID_FILE):
            unlink (optarg); /* don't be tricked by a symlink */
            if ((pid_fd = open (optarg, O_CREAT|O_EXCL|O_WRONLY, 0644)) < 0)
                usage ("cannot open or create pid file", optarg);
            continue ;
        case opt_key (OPT_STUN_DEVICE):
            if (*(stun_dev = optarg) == 0)
                usage ("STUN device must not be an empty string", optarg);
            continue ;
        case opt_key (OPT_STUN_ADDRESS):
            if ((stun_addr = inet_addr (optarg)) == ntohl (INADDR_NONE))
                usage ("error parsing the STUN address", optarg);
            continue ;
        case opt_key (OPT_SYSLOG):
            facility = optarg ;
            continue ;
        case opt_key (OPT_VERSION):
            cpry ();
            exit (0);
        case opt_key (OPT_VERBOSE):
            verbose = -1 ;
            continue ;
#       ifndef DISCARD_INFO
        case '?':
            usage (0, "?") ; /* no return */
        case opt_key (OPT_HELP):
            help (); /* and usage */
#       endif /* DISCARD_INFO */
        default:
            usage (0,0) ; /* no return */
        }

    if (optind < argc) {
        fprintf
            (stderr, "%s: Unexpected non-optional arguments: ", progname);
        while (optind < argc)
            fprintf (stderr, "%s ", argv [optind++]);
        fprintf (stderr, "\n");
        exit (2);
    }

    /* we must use syslog when running in the background */
    if (detach && facility == 0) facility = "daemon" ;

    /* initialise logging subsystem */
    if (facility) {
        if (syslog_init (facility, timepfx) == 0)
            usage ("cannot open the syslog facility", facility);
        /* detach, run in the backgrond */
        if (detach) {
            if (fork()) exit (0);
            setsid();
        }
        freopen ("/dev/null", "r", stdin);
        freopen ("/dev/null", "w", stdout);
        freopen ("/dev/null", "w", stderr);
    }
    else
        syslog_init (0, timepfx);
    
    cpry ();

    /* record own process id */
    if (pid_fd >= 0) {
        char buf [20];
        sprintf (buf, "%d\n", getpid ());
        write (pid_fd, buf, strlen (buf));
        close (pid_fd);
    }

    /* initialise the server control socket before chroot */
    ctrl = ctrl_init (set_gid,
		      MAIN_CTRL_IOBUF_SIZE, MAIN_CTRL_REQ_PATH,
		      &verbosity);

    /* change root environment */
    if (new_root && (chroot (new_root) || chdir ("/"))) {
        logger (LOG_ERR, "chroot failed: %s", strerror (errno));
        exit (2);
    }

    /* initialise the action loop */
    s.loop = do_init (idle, &s);

    s.ip = stun_addr
        ? stun_addr
        : ifquery (stun_dev, 0, 1 /* exit on error */)
        ;
    
    /* initialise STUN detection state machine */
    if ((s.r3489 = rfc3489_open
         (s.loop, s.ip, R3489_NVRAM_SVCLIST, R3489_NVRAM_PREFIX,
          R3489_NVRAM_QUERY, R3489_NVRAM_STATUS, R3489_NVRAM_MAPPED)) == 0) {
        logger (LOG_ERR, "Cannot initialise RFC3489 state machine");
        exit (2);
    }

    /* add services ... */
    logger (LOG_INFO, "Adding control socket services ...");
    if ((s.comm = comm_init (s.r3489)) == 0) {
        logger (LOG_ERR, "Cannot initialise commumication system");
        exit (2);
    }
    ctrl = ctrl_add (ctrl, "info",   comm_info,      s.comm);
    ctrl = ctrl_add (ctrl, "reload", comm_reload,    s.comm);
    ctrl = ctrl_add (ctrl, "stun",   comm_run,       s.comm);
    ctrl = ctrl_add (ctrl, "rusage", ctrl_resource,  ctrl);
    ctrl = ctrl_add (ctrl, "noisy",  ctrl_verbosity, ctrl);
    ctrl = ctrl_add (ctrl, "help",   ctrl_usage,     ctrl);
    s.comm = comm_update (s.comm, ctrl);

    if (verbose)
        verbosity = ~ (0) ;

    if (setjmp (go_here) == 0) {
        signal (SIGINT,  s_go_here);
        signal (SIGTERM, s_go_here);

        /* accept commands */
        do_action
            (s.loop, ctrl_fd (ctrl), (void(*)(void*))ctrl_handler, ctrl);
        
        for (;;) {
            n = do_loop (s.loop, idle_timeout * 1000);
            if (verbose) 
                logger (LOG_INFO, "s%: do_loop -> %d\n", progname, n);
        }
        /*NOTREACHED*/
    }

#   ifdef _MAIN_DEBUG
    /* release dispatcher */
    do_destroy (s.loop);
    
    /* release the server control socket */
    ctrl_destroy (ctrl) ;

    rfc3489_destroy (s.r3489) ;

    return 0;
#   else
    exit (0);
#   endif /* _MAIN_DEBUG */
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
