/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#include "util.h"
#include "logger.h"
#include "nvalloc.h"
#include "varstr.h"
#include "settings.h"
#include "helpers.h"
#include "base64.h"

#define __WPOST_HTTP_PRIVATE
#include "http.h"

/* ----------------------------------------------------------------------- *
 * Private variables and settings
 * ----------------------------------------------------------------------- */

#define __str(x)   # x
#define to_str(x)  __str (x)

#define DIM(x)     (sizeof (x) / sizeof (x [0]))

/* ----------------------------------------------------------------------- *
 * Private helpers
 * ----------------------------------------------------------------------- */

static
void to_http (IOBUF    *buf,
              const char *s) {
    int c ;
 
    while ((c = * s ++) != '\0')

        if (isalnum (c))
            char_append (buf, c) ;
        else
            switch (c) {
            case ' ':
                c = '+';
            case '.':
            case '_':
                char_append (buf, c) ;
                continue ;
            default:
                data_printf (buf, "%%%02X", c);
            }
}

static
unsigned read_append (HTDIAL *d) {

    int zero_packets = 0 ;
    unsigned tmo_ms = d->data.send.wait_first ;
        
    for (;;) {
        char *buf, *s ;
        int n ;
        fd_set fds;
        struct timeval tv;
        FD_ZERO (&fds);

        tv.tv_sec  =  tmo_ms / 1000 ;
        tv.tv_usec = (tmo_ms % 1000) * 1000 ;
        FD_SET (d->sock_fd, &fds);

        if ((n = select (d->sock_fd+1, &fds, 0, 0, &tv)) > 0) {
        
            make_space (&d->data.resp, HTTP_READ_BLOCK + 1) ;
            buf = d->data.resp.data + d->data.resp.top ;
        
            if ((n = recv (d->sock_fd, buf, HTTP_READ_BLOCK, 0)) < 0) {
                /* stupid software may behave weird upon errors */
                switch (errno) {
                case ECONNRESET:
                    d->econnreset = 1 ;
                    if (d->data.resp.top > 50)
                        goto leave_ok ;
                }
                logger (LOG_ERR, "Error reading HTTP response: %s",
                        strerror (errno));
                return 0 ;
            }

            if (n == 0) {
                if (++ zero_packets > d->data.send.max_zero)
                    break ;
            }
            else {
                buf [n] = '\0' ;

                if (verbosity & HTTP_LOG_RECV_DUMP) {
                    logger (LOG_DEBUG, "Receiving from %s:%d: %s.",
                            ip4addr_pp (d->ip), (unsigned)d->port, buf);
                }

                d->data.resp.top += n ;

                /* check for a terminating html tag */
                if ((s = strrchr (buf, '<')) != 0  &&
                    strncasecmp (s, "</html>", 7) == 0)
                    goto leave_ok ;
            }
        }
        else if (n < 0) {
            // switch (errno) {
            // }
            logger (LOG_ERR, "Error %d reading HTTP response: %s",
                    errno, strerror (errno));
            break ;
        }

        tmo_ms = d->data.send.wait_next ;
    }
    
    if (d->data.resp.top == 0) {
        logger (LOG_ERR, "Error reading HTTP response: no data yet");
        return 0;
    }

leave_ok:
    d->data.resp.data [d->data.resp.top] = '\0' ;
    return 1;
}


static
char *msgbf (char *p, size_t len, const char *s) {
    char *q = p;
    int n = len ;

    while (n -- > 0) {
        int c = (* p ++ = * s ++) ;

        switch (c) {
        case '\0':
            break ;
        case '\t':
            p [0] = 't';
            goto escape_char ;
        case '\n':
            p [0] = 'n';
            goto escape_char ;
        case '\r':
            p [0] = 'r';
            goto escape_char ;
        case '\f':
            p [0] = 'f';
            /* goto escape_char ; */
        escape_char:
            p [-1] = '\\'; 
            -- n ;
            ++ p ;
            continue ;
        default:
            if (iscntrl (c))
                p [-1] = '^' ;
            else if ((c & ~0x7f) != 0)
                p [-1] = '_' ;
            continue ;
        }

        break ;
    }
    
    p [n] = '\0' ;
    return q ;
}


static
void first_line_parse_error (HTDIAL *d) {
    char b [20] ;
    logger (LOG_ERR, "Parsing HTTP response: Error parsing first "
            "line (%s ...)",  msgbf (b, sizeof (b), d->data.resp.data));
}


static
unsigned parse_html_response (HTDIAL *d, char *p) {
    const char *s ;
    unsigned u;
    char b [30] ;

#   define parse_simple_tag(tag,error) {               \
    while (isspace (*p)) p ++ ;                        \
    if (strncasecmp (p, tag, sizeof (tag) - 1) != 0) { \
        s = tag " expected" ;                          \
        goto error ;                                   \
    }                                                  \
    p += sizeof (tag) - 1 ;                            }

    parse_simple_tag ( "<HTML>",  first_line_error) ;
    parse_simple_tag ( "<HEAD>",  html_parse_error) ;
    parse_simple_tag ( "<TITLE>", html_parse_error) ;
#   undef parse_simple_tag

    /* parse reply_code: "[...]400 " */
    while (isspace (p [0])) p ++ ;
    if (parse_unsigned (&u, p, " ") == 0) {
        logger (LOG_ERR, "Parsing HTML response: "
                "Error parsing the response code");
        return 0 ;
    }
    if (d->data.reply.code == 0)
        d->data.reply.code = u ;
    else if (u != d->data.reply.code)
        logger (LOG_WARNING, "The code in the HTML title %d differs "
                "from the http reply code %d", u,  d->data.reply.code);
    
    p = strchr (p, ' ') ;
    while (* ++ p == ' ')
        ;
    if (d->data.reply.text == 0)
        d->data.reply.text = p - d->data.resp.data ;
    if ((p = strchr (p, '<')) != 0)
        * p = '\0' ;

    /* in case there is no body, at all */
    d->data.reply.data = d->data.reply.text ;
    
    /* find the start of the reply body */
    for (;;) {
        if ((p = strchr (p + 1, '<')) == 0)
            return 1;
        if (strncasecmp (p, "<BODY", 5) != 0)
            continue ;
        if ((p = strchr (p, '>')) == 0)
            return 1;
        while (isspace (* ++ p))
            ;
        break ;
    }

    /* skip header like <h#> ... </h#> */
    if (strncasecmp (p, "<H", 2) == 0) {
        if ((p = strchr (p, '>')) == 0)
            return 1;

        /* in case there is no other body text */
        d->data.reply.data = ++ p - d->data.resp.data ;

        if ((p = strchr (p, '<')) == 0)
            return 1;
        * p ++ = '\0';
        if (strncasecmp (p, "/H", 2) != 0 || p [3] != '>')
            return 1;
        p += 4 ;
        while (isspace (*p)) p ++ ;
    }

    /* remember data text */
    d->data.reply.data = p - d->data.resp.data ;

    /* find the end of the message, terminate with \0 */
    for (;;) {
        if ((p = strchr (p, '<')) == 0)
            return 1;
        if (strncasecmp (p,    "<HR", 3) == 0 ||
            strncasecmp (p,  "<ADDR", 5) == 0 ||
            strncasecmp (p, "</BODY", 6) == 0 ||
            strncasecmp (p,   "<!--", 4) == 0)
            break ;
        p ++ ;
    }

    while (isspace (p [-1])) p -- ;
    * p = '\0' ;
    return 1 ;
    
first_line_error:
    first_line_parse_error (d);
    return 0;

html_parse_error:
    logger (LOG_ERR, "Parsing HTML response (%s ...): %s",
            msgbf (b, sizeof (b), d->data.resp.data), s);
    return 0 ;
}


static
unsigned parse_http_response (HTDIAL *d) {
    char *p, *r;

    if (strncasecmp (d->data.resp.data, "HTTP/", 5) != 0   ||
        isdigit (d->data.resp.data [5])             == 0   ||
        d->data.resp.data [6]                       != '.' ||
        isdigit (d->data.resp.data [7])            == 0    ||
        d->data.resp.data [8]                      != ' '  ) {
        first_line_parse_error (d);
        return 0;
    }

    /* skip additional spaces */
    p = d->data.resp.data + 8 ;
    while (* ++ p == ' ')
        ;

    /* parse reply_code: "[...]200 " */
    if (parse_unsigned (&d->data.reply.code, p, " ") == 0) {
        logger (LOG_ERR, "Parsing HTTP response: "
                "Error parsing the response code");
        return 0 ;
    }

    /* remember the reply text following the reply code */
    p = strchr (p, ' ') ;
    while (* ++ p == ' ')
        ;
    d->data.reply.text = p - d->data.resp.data ;
    if ((r = strchr (p, '\r')) != 0 ||
        (r = strchr (p, '\n')) != 0)
        * (p = r) = '\0' ;

    /* find the start of the reply body */
    for (;;) {
        char *q;
        if ((q = strchr (p + 1, '\r')) != 0) {
            if (strncmp (q, "\r\n\r\n", 4) == 0) {
                p = q + 4 ;           
                break ;
            }
        }
        else if ((q = strchr (p + 1, '\n')) != 0) {
            if (strncmp (q, "\n\n", 2) == 0) {
                p = q + 2 ;
                break ;
            }
        }
        else {
            logger (LOG_ERR, "Parsing HTTP response: Did not find out "
                    "where the header ends");
            return 0 ;
        }
        p = q ;
    }

    /* skip leading spaces and zeros */
    while ((* p == '\0' || isspace (*p)) &&
           p < d->data.resp.data + d->data.resp.top)
        ++ p;

    if (d->data.reply.code / 100 == 4 &&
        parse_html_response (d, p) != 0)
        return 1;

    d->data.reply.data = p - d->data.resp.data ;
    return 1;
}

static
void http_reset (HTDIAL *d) {
    NVCL *nvram = d->nvram ;
    XZERO (*d);

    d->nvram                = nvram ;
    d->sock_fd              = -1 ;
    d->data.send.max_zero   = HTTP_READ_MAX_ZERO ;
    d->data.send.wait_first = HTTP_READ_WAIT_FIRST ;
    d->data.send.wait_next  = HTTP_READ_WAIT_MORE;
}


static
void http_clear (HTDIAL *d) {
    int n ;
    
    if (d->sock_fd > 0) {
        shutdown (d->sock_fd, SHUT_RDWR);
        close (d->sock_fd);
    }

    for (n = 0; n < d->data.headers; n ++) {
        free ((char*)d->data.header [n].var) ;
        free ((char*)d->data.header [n].val) ;
    }
    free ((char*)d->data.header) ;

    kill_space (&d->data.post) ;

    http_reset (d);
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */


HTDIAL *http_open (NVCL *nvram) {
    HTDIAL d ;

    if (nvram == 0) {
        logger (LOG_ERR, "http: NVRAM subsystem not initialised!");
        return 0 ;
    }

    http_reset (&d);
    d.nvram = nvram ;
    return zmemdup (&d);
}



void http_target (HTDIAL *d, ip4addr_t ip, uint16_t port) {

    if (d->sock_fd >= 0)
        http_clear (d) ;

    d->ip   =   ip ;
    d->port = port ;
}


void http_close (HTDIAL *d) {
    http_clear (d) ;
    free ((char*)d);
}


unsigned http_connect (HTDIAL *d) {

    if (d->sock_fd < 0) {
        struct sockaddr_in in ;

        /* open http socket */
        if ((d->sock_fd = socket (PF_INET, SOCK_STREAM, 0)) == -1) {
            logger
                (LOG_ERR, "Cannot create http socket: %s", strerror (errno));
            return 0;
        }

        XZERO (in);
        in.sin_family = AF_INET ;
        in.sin_port   = htons (d->port) ;
        if (d->ip) in.sin_addr.s_addr = htonl (d->ip) ;

        if (connect (d->sock_fd, (struct sockaddr*)&in, sizeof in) < 0) {
            logger (LOG_ERR, "Cannot connect to %s:%u: %s",
                    ip4addr_pp (d->ip), (unsigned)d->port, strerror (errno));
            http_disconnect (d);
            return 0;
        }
    }
    
    return 1;
}


void http_disconnect (HTDIAL *d) {
    if (d->sock_fd >= 0) {
        shutdown (d->sock_fd, SHUT_RDWR);
        close (d->sock_fd);
    }
    d->sock_fd  = -1 ;
    http_resp_flush (d) ;
}


void http_hdr_edit (HTDIAL       *d,
                    const char *var,
                    const char *val,
                    int       force) {
    int n ;

    /* force: 0 -> default, 1 -> overwrite, else -> append */
    if (force == 0 || force == 1) {

        /* check whether we know this header already */
        for (n = 0; n < d->data.headers; n ++)

            if (strcasecmp (d->data.header [n].var, var) == 0) {
        
                if (force) {
                    int len1 = strlen (val) + 1;
                    d->data.header [n].val = (const char *) zrealloc
                        ((char*)d->data.header [n].val, len1);
                    memcpy ((char*)d->data.header [n].val, val, len1);
                }

                return ;
            }
    }
    else
        /* replace == -1: append only */
        n = d->data.headers ;

    /* n == d->data.headers */

    d->data.header = (struct _header *) zrealloc
        (d->data.header, ++ d->data.headers * sizeof (struct _header));
    d->data.header [n].var = zstrdup (var) ;
    d->data.header [n].val = zstrdup (val) ;
}


const char *http_basic_auth (HTDIAL *d, const char *u, const char *p) {
    const char *val ;
    size_t len ;
    IOBUF b ;

    if ((val = nvalue (d->nvram, u)) == 0) {
        logger (LOG_WARNING, "Baisic-Auth: No NVRAM user name %s", val);
        return 0;
    }

    XZERO (b);
    data_append (&b, val, 0);
    char_append (&b, ':');

    if ((val = nvalue (d->nvram, p)) == 0) {
        logger (LOG_WARNING, "Baisic-Auth: No NVRAM password %s", val);
        kill_space (&b);
        return 0;
    }

    data_append (&b, val, 0);
    u   = b.data ;
    len = b.top ;
    unlink_space (&b);
    
    XZERO (b) ;
    data_printf (&b, "Basic %s", p = b64alloc (u, len));
    unlink_space (&b);

    free ((char*)p);

    return b.data ;
}


void http_avp_append (HTDIAL *d, const char *var, const char *val) {
    to_http     (&d->data.post, var);
    char_append (&d->data.post, '=');
    to_http     (&d->data.post, val);
    char_append (&d->data.post, '&');
}


const char *http_val (HTDIAL *d, const char *var) {
    const char *val = nvalue (d->nvram, var) ;
    if (val == 0)
        return "" ;
    return val ;
}

unsigned http_var_append (HTDIAL       *d,
                          const char *pfx,
                          const char *var,
                          const char *(*map)(IOBUF*, const char*)) {
    const char *val ;

    if ((val = nvalue (d->nvram, var)) == 0)
        return 0 ;

    if (map) {
        IOBUF  b ;
        XZERO (b) ;
        http_avp_append (d, pfx, (*map) (&b, val)) ;
        kill_space (&b);
    }
    else
        http_avp_append (d, pfx, val);
        
    return 1 ;
}


void http_resp_flush (HTDIAL *d) {
    kill_space (&d->data.resp);
    XZERO (d->data.reply) ;
}


void http_var_flush (HTDIAL *d) {
    kill_space (&d->data.post);
}


unsigned http_send (HTDIAL *d, const char *uri) {
    int n ;
    const char *p ;
    size_t left ;
    IOBUF b ;

    XZERO (b) ;
    if ((n = d->data.post.top) != 0) n += 2 ; /* \r\n appended */
    data_printf (&b, "%u", n);
    http_hdr_replace (d, "Content-Length", b.data);
    
    b.top = 0 ;
    data_printf (&b, "%s:%u", ip4addr_pp (d->ip), (unsigned)d->port);
    http_hdr_default (d, "Host", b.data); /* Host: 192.0.2.1:8080 */

    b.top = 0 ;
    data_printf (&b, "%s %s HTTP/1.1\r\n",
                 d->data.post.top ? "POST" : "GET", uri);

    for (n = 0; n < d->data.headers; n ++)
        data_printf (&b, "%s: %s\r\n",
                     d->data.header [n].var, d->data.header [n].val);
    
    data_append (&b, "\r\n", 0);
    if (d->data.post.top) {
        data_append (&b, d->data.post.data, 0);
        data_append (&b, "\r\n", 0);
    }

    // fprintf (stderr, "%s", b.data);

    /* send data */
    p    = b.data ;
    left = b.top  ;
    do {
        if ((n = send (d->sock_fd, p, left, 0)) < 0)
            goto send_error;
        if (n == 0)
            goto zero_send;
        if (verbosity & HTTP_LOG_SEND_DUMP) {
            int save = p [n] ;
            ((char*)p) [n] = '\0' ;
            logger (LOG_DEBUG, "Sending to %s:%d: %s.",
                    ip4addr_pp (d->ip), (unsigned)d->port, p);
            ((char*)p) [n] = save ;
        }
    }
    while (p += n, left -= n) ;
    
    kill_space (&b);
    return 1;

send_error:
    logger (LOG_ERR, "Error sending HTTP request: %s", strerror (errno));
    return 0 ;

zero_send:
    logger (LOG_ERR, "Sending HTTP request: zero packet sent");
    return 0 ;
}


unsigned http_recv (HTDIAL *d) {   
    char *p;

    if (read_append (d) == 0)
        return 0 ;

    /* parse header: "HTTP/1.0 " */
    if (d->data.resp.data [0] == 'H' ||
        d->data.resp.data [0] == 'h')
        return parse_http_response (d) ;

    /* skip additional spaces */
    for (p = d->data.resp.data; isspace (p [0]); p ++)
        ;

    if (p [0] == '<')
        return parse_html_response (d, p) ;

    first_line_parse_error (d);
    return 0;
}


unsigned http_rply_code (HTDIAL *d) {
    return d->data.reply.code ;
}


const char *http_rply_text (HTDIAL *d) {
    return d->data.resp.data && d->data.reply.text
        ? (d->data.resp.data + d->data.reply.text)
        : 0
        ;
}


const char *http_rply_data (HTDIAL *d) {
    return d->data.resp.data && d->data.reply.data
        ? (d->data.resp.data + d->data.reply.data)
        : 0
        ;
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
