/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include "varstr.h"
#include "logger.h"
#include "nvalloc.h"
#include "http.h"
#include "helpers.h"

#include "set-dsl.h"

/* ----------------------------------------------------------------------- *
 * Private variables and settings
 * ----------------------------------------------------------------------- */

#define NVRAM_PREFIX "lan_dhcp."

#define DIM(x)  (sizeof (x) / sizeof (x [0]))

/* ----------------------------------------------------------------------- *
 * Private: map functions
 * ----------------------------------------------------------------------- */

static
const char *dnsmode (IOBUF *b, const char *s) {
    return strncasecmp (s, "auto", 4) == 0
        ? "0" /* auto */
        : "1" /* manual */
        ;
}

static
const char *mac_bpack2alpha (IOBUF *b, const char *s) {
    char* p;

    while ( (p=strchr(s, ':')) ) {
    	*p = '-';
    }

    return s;
}

/* ----------------------------------------------------------------------- *
 * Private: parameter and assignment lists
 * ----------------------------------------------------------------------- */

struct _intro_avp {
    const char *var ;
    const char *val ;
} ;


static
const struct _intro_avp intro_avp [] = {
    {"getpage", "../html/home/home_dhcp.htm"},
};


static
const struct _intro_avp intro_off [] = {
    {"landhcps0:settings/state",    "0"},
};


static
const struct _append_vars {
    const char *var ;
    const char *val ;
    const char *(*map)(IOBUF*, const char*);
} append_vars [] = {

    {"landhcps0:settings/state",    "isDhcp", false_is0},
    {"landhcps0:settings/start",    "startIp"},
    {"landhcps0:settings/end",      "endIp"},
    {"landhcps0:settings/lease",    "leaseTime"},
    {"landhcps0:settings/dns1mode", "dnsMode", dnsmode},
    {"landhcps0:settings/dns1",     "primaryDns"},
    {"landhcps0:settings/dns2",     "secondaryDns"},

    {"landhcpfwd0:settings/ip",     "0.0.0.0"},
    {"landhcpfwd0:settings/state",  "0"},
};

static
const struct {
    const char *var ;
    const char *val ;
    const char *(*map)(IOBUF*, const char*);
} macip_vars [] = {
    {"landhcps0:settings/staticip%d/ip",  "ipAddress.%d"},
    {"landhcps0:settings/staticip%d/mac", "macAddress.%d", mac_bpack2alpha},
};

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

static
unsigned post_vars (HTDIAL *http, const char *instance, int extra) {
    int n, slot, id ;
    unsigned hexdgts ;
    IOBUF i, j ;
    
    XZERO (i) ;
    XZERO (j) ;

    if (instance == 0)
        hexdgts = 0x54321 ;
    else if (parse_unsigned (&hexdgts, instance, 0) == 0) {
        if (extra == 0)
            logger (LOG_ERR, "dhcp_vars: illegal instance variable \"%s\"",
                    instance);
        return 0 ;
    }

    for (n = 0; n < DIM (intro_avp) ; n++ )
        http_avp_append (http, intro_avp [n].var, intro_avp [n].val);

    if (extra) {
        for (n = 0; n < DIM (intro_off) ; n++ )
            http_avp_append (http, intro_avp [n].var, intro_off [n].val);
    }
    else {
        j.top = 0 ;
        data_append (&j, NVRAM_PREFIX, 0);

        for (n = 0; n < DIM (append_vars) ; n++ ) {
            j.top = sizeof (NVRAM_PREFIX) - 1;
            data_append (&j, append_vars [n].val, 0);
            http_var_append
                (http, append_vars [n].var, j.data, append_vars [n].map);
        }

        for (slot = 1; slot <= 5; slot ++) {
            id = (hexdgts >> ((slot - 1) * 4)) & 0xf ;
            for (n = 0; n < DIM (macip_vars) ; n++ ) {
                i.top = 0 ;
                data_printf (&i, macip_vars [n].var, slot);
                if (id) {
                    j.top = sizeof (NVRAM_PREFIX) - 1;
                    data_printf (&j, macip_vars [n].val, id);
                    http_var_append (http, i.data, j.data, macip_vars[n].map);
                }
                else {
                    http_avp_append (http, i.data, "");
                }
            }
        }
    }
    
    kill_space   (&i);
    kill_space   (&j);

    return 1;
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

const char *dhcp_help (unsigned indent, unsigned ff) {
    char b [20] ;
    IOBUF i ;
    XZERO (i) ;

    sprintf (b, "%%-%ds%%-%ds%%s\n", indent, ff);

    data_printf (&i, b, "", "instance:",
                 "number");
    data_printf (&i, b, "", "with:",
                 "mac/ip-slot1 as \"[ number            ] % 16\"");
    data_printf (&i, b, "", "",
                 "mac/ip-slot2 as \"[ number / (16 ^ 1) ] % 16\"");
    data_printf (&i, b, "", "", "...");
    data_printf (&i, b, "", "",
                 "mac/ip-slot5 as \"[ number / (16 ^ 4) ] % 16\"");
    data_printf (&i, b, "", "default:",
                 "0x54321 - uses nvram slots 1 2 3 4 5");

    unlink_space (&i);
    return i.data ;
}


const char *dhcp_getp (HTDIAL *http, const char *instance) {
    return 0;
}

unsigned dhcp_extra (HTDIAL *http, const char *instance) {
    return post_vars (http, instance, 1);
}


unsigned dhcp_vars (HTDIAL *http, const char *instance) {
    return post_vars (http, instance, 0);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
