/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_CTYPE_H
#include "ctype.h"
#endif

#include "buffer.h"
#include "logger.h"
#include "nvalloc.h"
#include "http.h"
#include "helpers.h"
#include "encodec.h"

#include "set-snmp.h"

/* ----------------------------------------------------------------------- *
 * Private variables and settings
 * ----------------------------------------------------------------------- */

#define NVRAM_PREFIX      "lan_snmp."
#define NVRAM_COMM_PREFIX "lan_snmp_community."
#define NVRAM_TRAP_PREFIX "lan_snmp_traps."

#define DIM(x)  (sizeof (x) / sizeof (x [0]))

/* ----------------------------------------------------------------------- *
 * Private: map functions
 * ----------------------------------------------------------------------- */

static
const char *rwmap (IOBUF *b, const char *s) {
    if (strcasecmp (s, "readwrite") == 0)
        return "rw";
    return "ro";
}

static
const char *tvr (IOBUF *b, const char *s) {
    if (strcasecmp (s, "snmp v1") == 0)
        return "1";
    if (strcasecmp (s, "snmp v2") == 0)
        return "2";
    return "";
}

/* ----------------------------------------------------------------------- *
 * Private: parameter and assignment lists
 * ----------------------------------------------------------------------- */

struct _intro_avp {
    const char *var ;
    const char *val ;
};

struct _append_vars {
    const char *var ;
    const char *val ;
    const char *(*map)(IOBUF*, const char*);
};

/* ----------------------------------------------------------------------- *
 * SNMP
 * ----------------------------------------------------------------------- */

static
const struct _intro_avp intro_avp [] = {
    {"getpage",                       "../html/advanced/adv_snmp.htm"},
};

static /* SNMP standard */
const struct _intro_avp intro_std [] = {
    {"snmpcm:settings/system/sysoid", "1.3.6.1.4.1.294"},
};

static
const struct _append_vars append_snmp [] = {
    {"snmpcm:settings/snmpstate",          "hasSnmpAgent", ena_is1},
    {"snmpcm:settings/trapstate",          "hasSnmpTraps", ena_is1},
    {"snmpcm:settings/system/syscontact",  "contact"},
    {"snmpcm:settings/system/sysname",     "name"},
    {"snmpcm:settings/system/syslocation", "location"},
};

#define snmp_com_slots 2
static
const struct _append_vars append_snmp_com [] = {
    {"snmpcm:settings/community/entry%d/name",   "communityName.%d"},
    {"snmpcm:settings/community/entry%d/access", "accessRights.%d", rwmap},
};

#define snmp_trap_slots 3
static
const struct _append_vars append_snmp_trap [] = {
    {"snmpcm:settings/trapinfo/entry%d/trapdest",        "destinationIp.%d"},
    {"snmpcm:settings/trapinfo/entry%d/trapcommunity", "trapCommunityId.%d"},
    {"snmpcm:settings/trapinfo/entry%d/trapversion",   "trapVersion.%d", tvr}
};

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

static
unsigned post_vars (HTDIAL *http, const char *instance, int extra) {
    int n, k, id ;
    const char *list ;
    IOBUF h, i, j ;
    char *p, *q, *r ;

    for (n = 0; n <DIM (intro_avp); n ++)
        http_avp_append (http, intro_avp [n].var, intro_avp [n].val);

    if (extra)
        return 1 ;

    for (n = 0; n <DIM (intro_std); n ++)
        http_avp_append (http, intro_std [n].var, intro_std [n].val);

    XZERO (h);
    XZERO (i);
    XZERO (j);

    data_append (&j, NVRAM_PREFIX, 0);
    for (n = 0; n < DIM (append_snmp); n ++) {
        j.top = sizeof (NVRAM_PREFIX) - 1;
        data_append (&j, append_snmp [n].val, 0);
        http_var_append
            (http, append_snmp [n].var, j.data, append_snmp [n].map);
    }


    /* NVRAM_COMM_PREFIX */
    if ((list = http_val (http, NVRAM_COMM_PREFIX "list")) != 0 &&
        list [0] != '\0')
        r = p = zstrdup (list) ;
    else {
        p  = 0 ;
        id = 0 ;
    }

    j.top = 0 ;
    data_append (&j, NVRAM_COMM_PREFIX, 0);
    for (k = 0; k < snmp_com_slots; k ++) {

        if (p != 0 &&
            ((id = strtol (q = r, &r, 0)) == LONG_MIN || id == LONG_MAX))
            id = 0 ;

        for (n = 0; n < DIM (append_snmp_com); n ++) {
            i.top = 0 ;
            data_printf (&i, append_snmp_com [n].var, k);

            if (id == 0) {
                http_avp_append (http, i.data, "");
                continue ;
            }

            h.top = 0 ;
            data_printf (&h, append_snmp_com [n].val, id);
            j.top = sizeof (NVRAM_COMM_PREFIX) - 1;
            data_append (&j, h.data, 0);

            http_var_append
                (http, i.data, j.data, append_snmp_com [n].map);
        }
    }
    if (p) free (p);


    /* NVRAM_TRAP_PREFIX */
    if ((list = http_val (http, NVRAM_TRAP_PREFIX "list")) != 0
        && list [0] != '\0')
        r = p = zstrdup (list) ;
    else {
        p  = 0 ;
        id = 0 ;
    }

    j.top = 0 ;
    data_append (&j, NVRAM_TRAP_PREFIX, 0);
    for (k = 0; k < snmp_trap_slots; k ++) {

        if (p != 0 &&
            ((id = strtol (q = r, &r, 0)) == LONG_MIN || id == LONG_MAX))
            id = 0 ;

        for (n = 0; n < DIM (append_snmp_trap); n ++) {
            i.top = 0 ;
            data_printf (&i, append_snmp_trap [n].var, k);

            if (id == 0) {
                http_avp_append (http, i.data, "");
                continue ;
            }

            h.top = 0 ;
            data_printf (&h, append_snmp_trap [n].val, id);
            j.top = sizeof (NVRAM_TRAP_PREFIX) - 1;
            data_append (&j, h.data, 0);

            http_var_append
                (http, i.data, j.data, append_snmp_trap [n].map);
        }
    }
    if (p) free (p);

    
    kill_space (&h);
    kill_space (&i);
    kill_space (&j);

    return 1;
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

const char *snmp_help (unsigned indent, unsigned ff) {
    char b [20] ;
    IOBUF i ;
    XZERO (i) ;

    sprintf (b, "%%-%ds%%-%ds%%s\n", indent, ff);
    data_printf (&i, b, "", "instance:", "n/a (ignored)");
    unlink_space (&i);
    return i.data ;
}


const char *snmp_getp (HTDIAL *http, const char *instance) {
    return 0;
}

unsigned snmp_extra (HTDIAL *http, const char *instance) {
    return post_vars (http, instance, 1);
}


unsigned snmp_vars (HTDIAL *http, const char *instance) {
    return post_vars (http, instance, 0);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
