/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
# include <sys/types.h>
# else
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined (STDC_HEADERS) && defined (HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_CTYPE_H
#include "ctype.h"
#endif

#include "buffer.h"
#include "logger.h"
#include "nvalloc.h"
#include "http.h"
#include "helpers.h"
#include "encodec.h"
#include "parameters.h"

#include "set-wlan.h"

/* ----------------------------------------------------------------------- *
 * Private variables and settings
 * ----------------------------------------------------------------------- */

#define DIM(x)  (sizeof (x) / sizeof (x [0]))

/* ----------------------------------------------------------------------- *
 * Private: parameter and assignment lists
 * ----------------------------------------------------------------------- */

struct _intro_avp {
    // const char *var ;
    const PARAM_MAP *var ;
    const char *val ;
};

/* ----------------------------------------------------------------------- *
 * Common introducers
 * ----------------------------------------------------------------------- */

static/* extra and/or standard */
const struct _intro_avp intro_avp_off [] = {
    {ap_settings_ap_enabled_0,     "0"},
};

static /* extra */
const struct _intro_avp intro_avp_on0 [] = {
    {_APEnable_0,                 "on"},
    {_SSIDName_0,                 "HB"},
    {_Channel_0,                   "9"},
};

static /* standard */
const struct _intro_avp intro_avp_on1 [] = {
    {ap_settings_ap_enabled_0,     "1"},
    {var_apenabled_0,               ""},
    {ap_settings_user_isolation_0, "0"},
};

static /* standard */
const PARAM_MAP *append_vars [] = {
    ap_settings_ssid_0,
    var_ssid_0,
    ap_settings_channel_g_0,
    var_channel_0
};

/* ----------------------------------------------------------------------- *
 * NONE (neither WEP not WPA)
 * ----------------------------------------------------------------------- */

static
const struct _intro_avp intro_avp_none [] = {
    {_getpage_0, "../html/home/wireless_none.htm"},
};

static /* extra */
const struct _intro_avp intro_avp_none0 [] = {
    {_APSecurity_0, "1"},
    {_uiSecurity_0, "on"}
};

static /* standard */
const struct _intro_avp intro_avp_none1 [] = {
    {ap_settings_privacy_type_0, "0"},
};

/* ----------------------------------------------------------------------- *
 * WEP
 * ----------------------------------------------------------------------- */

static
const struct _intro_avp intro_avp_wep [] = {
    {_getpage_0, "../html/home/wireless_security_wep.htm"},
};

static /* WEP extra */
const struct _intro_avp intro_avp_wep0 [] = {
    {_APSecurity_0, "2"},
    {_uiSecurity_0, "on"}
};

static /* WEP standard */
const struct _intro_avp intro_avp_wep1 [] = {
    {ap_settings_wep_privacy_0,  "1"},

    /* FIXME: type=1 is documented but type=2 works! */
    {ap_settings_privacy_type_0, "2"}
};


static /* WEP standard */
const PARAM_MAP *append_vars_wep [] = {
    ap_settings_wep_auth_type_0,
    ap_settings_key_id_0,
    ap_settings_key_value1_0,
    ap_settings_key_len1_0,
    ap_settings_key_value2_0,
    ap_settings_key_len2_0,
    ap_settings_key_value3_0,
    ap_settings_key_len3_0,
    ap_settings_key_value4_0,
    ap_settings_key_len4_0
};

/* ----------------------------------------------------------------------- *
 * WPA
 * ----------------------------------------------------------------------- */

static
const struct _intro_avp intro_avp_wpa [] = {
    {_getpage_0, "../html/home/wireless_security_wpa.htm"},
};

static /* WPA extra */
const struct _intro_avp intro_avp_wpa0 [] = {
    {_APSecurity_0, "2"},
};

static
const PARAM_MAP *append_vars_wpa [] = {
    ap_settings_radius_ip_0,
    ap_settings_radius_port_0,
    ap_settings_radius_secret_0,
    ap_settings_interval_0
};

/* ----------------------------------------------------------------------- *
 * more WPA: 802.1x (standard)
 * ----------------------------------------------------------------------- */

static
const struct _intro_avp intro_avp_wpa_8021x [] = {
    {ap_settings_psktype_0,      "0"}, /* wpatype == 8021x */
    {ap_settings_privacy_type_0, "1"}
};

/* ----------------------------------------------------------------------- *
 * more WPA: psk hex string (standard)
 * ----------------------------------------------------------------------- */

static
const struct _intro_avp intro_avp_wpa_pskhex [] = {
    {ap_settings_psktype_0,      "1"}, /* wpatype == pskhex */
    {ap_settings_privacy_type_0, "1"}
};

static
const PARAM_MAP *append_vars_wpa_pskhex [] = {
    ap_settings_pskvalue_0
};

/* ----------------------------------------------------------------------- *
 * more WPA: psk text string (standard)
 * ----------------------------------------------------------------------- */

static
const struct _intro_avp intro_avp_wpa_pskstring [] = {
    {ap_settings_psktype_0,      "2"}, /* wpatype == pskstring */
    {ap_settings_privacy_type_0, "1"}
};

static
const PARAM_MAP *append_vars_wpa_pskstring [] = {
    ap_settings_pskvalue_1
};

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

static
unsigned post_vars (HTDIAL *http, const char *instance, int extra) {
    int n, k ;

    struct {
        unsigned inx ;
        struct {
            unsigned dim ;
            const struct _intro_avp *p ;
        } v [6] ;
    } a ;

    struct {
        unsigned iny ;
        struct {
            unsigned dum ;
            const PARAM_MAP **q  ;
        } v [5] ;
    } z ;

#   define AVP_REGISTER(sym) \
        (a.v [a.inx].dim = DIM (sym), a.v [a.inx ++].p = (sym))

#   define AVP_LOOP(u,b) \
        for ((u) = 0; (u) < a.inx; (u) ++) \
            for ((b) = 0; (b) < a.v [u].dim; (b) ++)

#   define ZVQ_REGISTER(sym) \
        (z.v [z.iny].dum = DIM (sym), z.v [z.iny ++].q = (sym))

#   define ZVQ_LOOP(u,b) \
        for ((u) = 0; (u) < z.iny; (u) ++) \
            for ((b) = 0; (b) < z.v [u].dum; (b) ++)

#   define NVRAM_VALUE(sym) \
        http_val (http, (sym)->nvram.var)


    const char *security ;

    a.inx = z.iny = 0 ;

    if (ena_is1 (0, NVRAM_VALUE (ap_settings_ap_enabled_0)) [0] == '1' &&
        (security = NVRAM_VALUE (ap_settings_privacy_type_0)) != 0) {

        if (strcasecmp (security, "none") == 0) {

            AVP_REGISTER (intro_avp_none);

            if (extra)
                AVP_REGISTER (intro_avp_none0);

            else
                AVP_REGISTER (intro_avp_none1);
        }
        else if (strcasecmp (security, "wep") == 0) {

            AVP_REGISTER (intro_avp_wep);

            if (extra)
                AVP_REGISTER (intro_avp_wep0);

            else {
                AVP_REGISTER (intro_avp_wep1);
                ZVQ_REGISTER (append_vars_wep);
            }
        }
        else if (strcasecmp (security, "wpa") == 0) {

            AVP_REGISTER (intro_avp_wpa);

            if (extra)

                AVP_REGISTER (intro_avp_wpa0);

            else {
                const char *wpat = http_val
                    (http, ap_settings_psktype_0->nvram.var);
                
                if (strcasecmp (wpat, "8021x") == 0)
                    AVP_REGISTER (intro_avp_wpa_8021x);

                else if (strcasecmp (wpat, "pskhex") == 0) {
                    AVP_REGISTER (intro_avp_wpa_pskhex);
                    ZVQ_REGISTER (append_vars_wpa_pskhex);
                }

                else if (strcasecmp (wpat, "pskstring") == 0) {
                    AVP_REGISTER (intro_avp_wpa_pskstring);
                    ZVQ_REGISTER (append_vars_wpa_pskstring);
                }

                else
                    goto ap_is_off ;

                // AVP_REGISTER (intro_avp_wpa1);
                ZVQ_REGISTER (append_vars_wpa);
            }
        }
        else
            goto ap_is_off ;

        if (extra)
            AVP_REGISTER (intro_avp_on0);
        else {
            AVP_REGISTER (intro_avp_on1);
            ZVQ_REGISTER (append_vars);
        }
    }
    else {
    ap_is_off:
        AVP_REGISTER (intro_avp_off);
        z.iny = 0 ;
    }

    AVP_LOOP (k, n)
        http_avp_append (http,
                         a.v [k].p [n].var->alpha.var,
                         a.v [k].p [n].val);

    ZVQ_LOOP (k, n)
        http_var_append (http,
                         z.v [k].q [n]->alpha.var,
                         z.v [k].q [n]->nvram.var,
                         z.v [k].q [n]->alpha.map);   
    return 1;

#   undef NV_VARIABLE
#   undef AVP_REGISTER
#   undef AVP_LOOP
#   undef ZVQ_REGISTER
#   undef ZVQ_LOOP
}

/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

const char *wlan_help (unsigned indent, unsigned ff) {
    char b [20] ;
    IOBUF i ;
    XZERO (i) ;

    sprintf (b, "%%-%ds%%-%ds%%s\n", indent, ff);
    data_printf (&i, b, "", "instance:", "n/a (ignored)");

    unlink_space (&i);
    return i.data ;
}

unsigned wlan_extra (HTDIAL *http, const char *instance) {
    return post_vars (http, instance, 1);
}


unsigned wlan_vars (HTDIAL *http, const char *instance) {
    return post_vars (http, instance, 0);
}

/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
