/**************************************************************************
*   Copyright (C) 2006 - D-Link (Europe) Ltd., London, UK                 *
*                                                                         *
*   Author: Jordan Hrycaj <jordan@maxina.de> - Maxina GmbH, Neu-          *
*   Isenburg, Germany							  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
*                                                                         *
***************************************************************************/



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "helpers.h"
#include "logger.h"
#include "settings.h"
#include "buffer.h"
#include "nvalloc.h"
#include "varstr.h"
#include "set-wpost-module.h"

/* ----------------------------------------------------------------------- *
 * Private variables and settings
 * ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- *
 * Private
 * ----------------------------------------------------------------------- */

static
const char *wpost_lazy_nvalue (NVDATA *d,  const char *var) {

    if (d->nv == 0)
        wpost_need_nvram (d);

    return nvalue (d->nv, var);
}


/* ----------------------------------------------------------------------- *
 * Public
 * ----------------------------------------------------------------------- */

void wpost_need_nvram (NVDATA *d) {

    if (d->nv != 0)
        return ;

#   ifdef WPOST_NVDEF_TIMEOUT
    if (d->nvtmo == 0)
        d->nvtmo = WPOST_NVDEF_TIMEOUT ;
#   endif

    /* initalise the nvram server */
    if ((d->nv = nvcl_open (0, d->nvserver, d->nvport, d->nvtmo)) != 0)
        return ;
    
    logger (LOG_ERR, "Cannot initialise NVRAM commumication system - exit");
    exit (2);
}

const char *wpost_normalise_uri (MODULE_URI *q, MODULE *m, const char *uri) {
    const char nvrfmt3s [] =
        "%s: cannot retrieve the %s \"%s\" from the nvram" ;

    const char *s ;
    static IOBUF b ;
    
    if (m->meth == 0) {
        /* only http supported (yet) */
        m->meth = proto_http ;

        /* server name from nvram */
        if (m->v.hostname == 0) {
            if ((s = wpost_lazy_nvalue (&m->v, WPOST_NVDEF_HOSTNAME)) == 0)
                goto nvserver_error ;
            m->v.hostname = zstrdup (s);
        }
        m->host = m->v.hostname ;

        /* ip address from the server name */
        if ((m->addr = hostip (m->host)) == -1)
            goto resolver_error ;

        /* port from nvram */
        if (m->v.port == 0) {
            if ((s = wpost_lazy_nvalue (&m->v, WPOST_NVDEF_PORT)) == 0)
                goto nvport_error ;
            m->v.port = strcmp (s, "80") == 0
                ? zstrdup ("")
                : zstrdup (s)
                ;
        }
        if (m->v.port [0] == '\0')
            m->port = 80 ;
        else if (parse_unsigned (&m->port, m->v.port, 0) == 0)
            goto port_scan_error;
    }

    XZERO (b);

    /* http:// <host> [:port] /ressource/path */
    data_printf (&b, "http://%s", m->v.hostname);
    if (m->port != 80)
        data_printf (&b, ":%d", m->port);
    data_append (&b, uri, 0);

    q->normalised.uri = b.data ;
    q->raw       .uri = uri ;

    unlink_space (&b); /* useless unless debugging */
    return 0 ;
   
resolver_error:
    kill_space (&b);
    data_printf (&b, "%s: cannot resolve the hostname \"%s\": %s",
                 progname, m->host, strerror (errno));
    return b.data ;
    
port_scan_error:
    kill_space (&b);
    data_printf (&b, "%s: error scanning the port number \"%s\": %s",
                 progname, m->v.port, varstr_strerr (varstr_errno));
    return b.data ;

 nvport_error:
    kill_space (&b);
    data_printf (&b, nvrfmt3s, progname, "port number", WPOST_NVDEF_PORT);
    return b.data ;

nvserver_error:
    kill_space (&b);
    data_printf (&b, nvrfmt3s, progname, "server name",WPOST_NVDEF_HOSTNAME);
    return b.data ;
}


unsigned wpost_http_request (HTDIAL *http, MODULE_URI *u, const char *inf) {
    
    if (verbosity & WPOST_LOG_HTTP_REQU)
        logger (LOG_INFO, "Requesting %s URI: %s", u->normalised.uri, inf);

    if (http_send (http, u->raw.uri) == 0)
        return 0 ;
        
    if (verbosity & WPOST_LOG_HTTP_REQU)
        logger (LOG_INFO, "Waiting for %s response", inf);
        
    if (http_recv (http) == 0)
        return 0 ;

    return 1;
}


int wpost_http_response (HTDIAL     *http,
                        MODULE_URI  *uri,
                         const char *name) {

    int r = 0, n, hterr = 0 ;
    const char *s = "" ;
    
    switch (n = http_rply_code (http)) {
    case 200:
        break ;

    case 400:
    case 404:
        hterr = 1 ;
        /* goto default; */

    default:
        s = "Unwanted HTTP response code: ";
        r = 3 ;
    }

    if ((verbosity & WPOST_LOG_HTTP_REPLY) != 0 || r)
        logger (r ? LOG_ERR : LOG_INFO, "%s%s(%s) -> %d %s", s,
                uri->normalised.uri, name, n, http_rply_text (http));
    
    if ((verbosity & WPOST_LOG_HTTP_RDATA) != 0 || hterr)
        logger (hterr ? LOG_ERR : LOG_DEBUG,
                "Reply data: %s", http_rply_data (http));

    return r;
}

int wpost_module_set (MODULE *m, MODULE_URI *uri, int xtra) {
    int r ;
    
    if (verbosity & WPOST_LOG_POST_URL)
        logger (LOG_INFO, "POST URL: %s", uri->normalised.uri);
   
    if (xtra && parset_post_xtra (m->set, m->inst) == 0) {
        logger (LOG_WARNING, "Error setting up extra post data");
        xtra = 0;
    }
    
    if (xtra == 0 && parset_post_body (m->set, m->inst) == 0)
        exit (2) ;
    
    if (verbosity & WPOST_LOG_HTTP_PCONN)
        logger (LOG_INFO, "Connecting to %s:%d",
                ip4addr_pp (m->addr), m->port);
    
    if (xtra) {

        /* connect to http server */
        if (http_connect (m->http)) {

            if (wpost_http_request (m->http, uri, "extra") == 0)
                exit (2) ;  

            /* prepare for reconnecting, below */
            if (parset_post_body (m->set, m->inst) == 0)
                exit (2) ;

            http_disconnect (m->http);
        }
        else
            logger (LOG_WARNING, "Error connecting for extra response");

        /* repeat on error */
        xtra = MAIN_XPOST_FAIL_REPEAT;
    }

    for (;;) {
        struct timeval nap;
 
        if (http_connect (m->http) == 0)
            goto loop_on_error ;
        
        if (wpost_http_request (m->http, uri, "POST") == 0)
            goto loop_on_error ;

        break;

    loop_on_error:
        if (xtra -- <= 0)
            exit (2);
        
        http_disconnect (m->http);

        if (verbosity & WPOST_LOG_HTTP_REQURPT)
            logger (LOG_INFO, "Retrying (%d more tries)", xtra + 1);

        nap.tv_sec  =  MAIN_WAIT_AFTER_XFAIL / 1000 ;
        nap.tv_usec = (MAIN_WAIT_AFTER_XFAIL % 1000) * 1000 ;
        select (0, 0, 0, 0, &nap) ;
    }

    r = wpost_http_response (m->http, uri, m->name);
    http_disconnect (m->http);
    return r ;
}


/* ----------------------------------------------------------------------- *
 * End
 * ----------------------------------------------------------------------- */
