/*
 * zaphfc.c - Zaptel driver for HFC-S PCI A based ISDN BRI cards
 * kernel module inspired by zaphfc driver written by Klaus-Peter
 * Junghanns.
 *
 * This program is free software and may be modified and
 * distributed under the terms of the GNU Public License.
 *
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <linux/zaptel.h>

#ifdef CONFIG_LEDMAN
# include <linux/ledman.h>
#else
# define ledman_cmd(a,b)
#endif

#ifdef ZAPHFC_HSS
# include <ixp425_hss.h>
#else
# include <khz-timer.h>
#endif

#include "zaphfc.h"

#if KERNEL_VERSION(2,6,0) > LINUX_VERSION_CODE
# include <asm/arch/gpio.h>
# define IRQ_HANDLED
# define irqreturn_t		void
# define IXP425_ISDN_PwrOn	IXP425_GPIO_PIN_9
#else
# define MOD_INC_USE_COUNT
# define MOD_DEC_USE_COUNT
# define IXP425_ISDN_PwrOn	9
# define IXP425_ISDN_PwrWarn	IRQ_IXP4XX_GPIO0	
# define IXP425_GPIO_OUT	IXP4XX_GPIO_OUT
# define IXP425_GPIO_IN		IXP4XX_GPIO_IN
# define IXP425_GPIO_TRANSITIONAL   0 /* not implemented? */
#endif /* KERNEL 2.6 */

#define read_z(x)	(__le16_to_cpu(x))
#define write_z(x)	((unsigned short)(__cpu_to_le16(x)))
#define hfc_outb(a,b,c) (writeb((u8)(c),(u32)((a)->pci_io+(b))))
#define hfc_inb(a,b) 	(readb((u32)((a)->pci_io+(b))))

#define __unused __attribute__((__unused__))

#define ZAPHFC_T3_PERIOD	800
#define ZAPHFC_T4_PERIOD	55



struct hfc_card {

	/* hardwarre */
	struct pci_dev* pci_dev;
	unsigned char*	pci_io;
    	dma_addr_t 	dma_handle;
	void*		fifomem;
	volatile void*	fifos;	/* 32k aligned mem for the fifos */
	int		irq;
	unsigned char	state; /* layer 1 state */

	/* zaptel */
	struct zt_span	span;
	struct zt_chan  chans[3];

	/* config */
	int		nt_mode;
	int		master;
	int		cardno;
	int		hss_handle;

#ifdef ZAPHFC_HSS
	/* hss */
	int		hss_start_bit[2];
#endif

	/* counter */
	int		drec_count;
	unsigned	last_msg;

	/* list */
	struct hfc_card* prev;
	struct hfc_card* next;

	/* buffers */
	unsigned char drecbuf[hfc_D_FIFO_SIZE];
	unsigned char dtransbuf[hfc_D_FIFO_SIZE];

	/* timer */
	struct timer_list t3;
	struct timer_list t4;

};

static int 		 hfc_card_count = 0;
static struct hfc_card*  hfc_cards = NULL;
static int 		 debug 	= 0;
static int 		 master = 1; /* defines which cards will be master */
static int 		 mode	= 1; /* defines which card is NT or TE, ( bit = 1 => NT, bit = 0 => TE) */
static int		 loopback = 0;
static int 		 nt_protection = 0; /* defines if the nt should be automaticaly powered off, default is no, use module parm to change */

/* prototypes */
static void zaphfc_reset(struct hfc_card* hfctmp);
#ifdef ZAPHFC_HSS
static void zaphfc_transfer_hss(unsigned char*, unsigned char*, void*);
#else
static void zaphfc_transfer(struct hfc_card* hfctmp);
#endif

/*
 * functions for proc file system 
 */

/*  proc dir for hss statistics */
static struct proc_dir_entry* zaphfc_proc_dir;

/* file and directory names */
#define ZAPHFC_PROC_DIR			"hfc"
#define ZAPHFC_PROC_FILE_STATE		"state"
#define ZAPHFC_PROC_FILE_DEBUG		"debug"
#define ZAPHFC_PROC_FILE_LOOPBACK	"loopback"
#define ZAPHFC_PROC_FILE_ALARM		"alarm"

static int zaphfc_proc_read_state(char *page, char **start __unused,
			    off_t off __unused, int count __unused,
			    int *eof __unused, void *data __unused)
{
	struct hfc_card* hfctmp;
	int bytes = 0;
	
	
	for (	hfctmp = hfc_cards; hfctmp; hfctmp = hfctmp->prev) {

		bytes += sprintf(page+bytes, "card #%d: State = %c%d\n", 
				hfctmp->cardno, (hfctmp->nt_mode)?'G':'F', 
				hfctmp->state);
		
	}

	return bytes;
			
}


static int zaphfc_proc_write_state(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	struct hfc_card* hfc;
	char buf[4];

	if (count > 4) {
		return -EFAULT;
	}

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) )
		return -EFAULT;

	if (!strncmp("on", buf,2))
		for (	hfc= hfc_cards; hfc; hfc= hfc->prev) {
			/* activation request */
			hfc_outb(hfc, hfc_STATES, hfc_STATES_DO_ACTION | hfc_STATES_ACTIVATE);

			/* start t3 timer */
			if (!hfc->nt_mode) {
				hfc->t3.expires = jiffies + ZAPHFC_T3_PERIOD;
				add_timer(&hfc->t3);
			}
		}

	return count;
}

static int zaphfc_proc_read_debug(char *page, char **start __unused,
			    off_t off __unused, int count __unused,
			    int *eof __unused, void *data __unused)
{
	if (debug)
		return sprintf(page, "on\n");
	return sprintf(page, "off\n");
}

static int zaphfc_proc_write_debug(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	char buf[4];

	if (count > 4) {
		return -EFAULT;
	}

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) )
		return -EFAULT;

	if (!strncmp("on", buf,2))
		debug = 1;
	else if (!strncmp("off", buf,3))
		debug = 0;

	return count;
}

static int zaphfc_proc_read_loopback(char *page, char **start __unused,
			    off_t off __unused, int count __unused,
			    int *eof __unused, void *data __unused)
{
	if (loopback)
		return sprintf(page, "on\n");
	return sprintf(page, "off\n");
}

static void zaphfc_set_loopback_mode(int on) 
{
	struct hfc_card* hfc;

	on = (on)?64:0;
	
	for (	hfc = hfc_cards; hfc; hfc = hfc->prev) {
		
		hfc_outb(hfc, hfc_B1_SSL, (hfc->hss_start_bit[0]/8) | 128 );
		hfc_outb(hfc, hfc_B1_RSL, (hfc->hss_start_bit[0]/8) | 128 | on );
	
		hfc_outb(hfc, hfc_B2_SSL, (hfc->hss_start_bit[1]/8) | 128 );
		hfc_outb(hfc, hfc_B2_RSL, (hfc->hss_start_bit[1]/8) | 128 | on );
	}
}

static int zaphfc_proc_write_loopback(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	char buf[4];

	if (count > 4) {
		return -EFAULT;
	}

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) )
		return -EFAULT;

	if (!strncmp("on", buf, 2))
		loopback = 1;
	else if (!strncmp("off", buf, 3))
		loopback = 0;
	
	zaphfc_set_loopback_mode(loopback);
	
	return count;
}


static int zaphfc_proc_read_alarm(char *page, char **start __unused,
			    off_t off __unused, int count __unused,
			    int *eof __unused, void *data __unused)
{
	struct hfc_card* hfctmp;
	int bytes = 0;
	
	
	for (	hfctmp = hfc_cards; hfctmp; hfctmp = hfctmp->prev) {

		bytes += sprintf(page+bytes, "card #%d: Alarm = 0x%x\n", 
				hfctmp->cardno, hfctmp->span.alarms);
		
	}

	return bytes;
			
}


static int zaphfc_proc_write_alarm(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	struct hfc_card* hfctmp;
	char buf[4];

	if (count > 4) {
		return -EFAULT;
	}

	/* write data to the buffer */
	if ( copy_from_user(buf, buffer, count) )
		return -EFAULT;

	if (!strncmp("off", buf, 3)) {
		for (	hfctmp = hfc_cards; hfctmp; hfctmp = hfctmp->prev) {
			if (hfctmp->nt_mode)
				continue;

			hfctmp->span.alarms = ZT_ALARM_NONE;
			zt_alarm_notify(&hfctmp->span);
		}
	}
	
	return count;
}


static void zaphfc_proc_register(void)
{
	struct proc_dir_entry* e;
	/* register proc directory */
	zaphfc_proc_dir = proc_mkdir(ZAPHFC_PROC_DIR, NULL);
	if(zaphfc_proc_dir == NULL) {
		printk("Error creating proc directory\n");
		return;
	}

	e = create_proc_entry(	ZAPHFC_PROC_FILE_STATE, 0444, 
				zaphfc_proc_dir);
	if (e) {
		e->read_proc = zaphfc_proc_read_state;
		e->write_proc= zaphfc_proc_write_state;
	}
	
	e = create_proc_entry(ZAPHFC_PROC_FILE_DEBUG, 0644, zaphfc_proc_dir);
	if (e) {
		e->write_proc=zaphfc_proc_write_debug;
		e->read_proc=zaphfc_proc_read_debug;
	}
	
	e = create_proc_entry(ZAPHFC_PROC_FILE_LOOPBACK, 0644, zaphfc_proc_dir);
	if (e) {
		e->read_proc = zaphfc_proc_read_loopback;
		e->write_proc = zaphfc_proc_write_loopback;
	}
	
	e = create_proc_entry(ZAPHFC_PROC_FILE_ALARM, 0644, zaphfc_proc_dir);
	if (e) {
		e->read_proc = zaphfc_proc_read_alarm;
		e->write_proc = zaphfc_proc_write_alarm;
	}
}
	

static void zaphfc_proc_unregister(void)
{
	/* remove proc file */
	remove_proc_entry(ZAPHFC_PROC_FILE_STATE, zaphfc_proc_dir);
	remove_proc_entry(ZAPHFC_PROC_FILE_DEBUG, zaphfc_proc_dir);
	remove_proc_entry(ZAPHFC_PROC_FILE_LOOPBACK, zaphfc_proc_dir);
	remove_proc_entry(ZAPHFC_PROC_FILE_ALARM, zaphfc_proc_dir);
	remove_proc_entry(ZAPHFC_PROC_DIR, NULL);
}


/**
 * Zaptel 
 */
static int zaphfc_zaptel_open(struct zt_chan *chan __unused) 
{
    MOD_INC_USE_COUNT;
    return 0;
}

static int zaphfc_zaptel_close(struct zt_chan *chan __unused) 
{
    MOD_DEC_USE_COUNT;
    return 0;
}

static int zaphfc_zaptel_ioctl(	struct zt_chan *chan __unused, 
				unsigned int cmd, 
				unsigned long data)
{
	switch(cmd) {
		case ZT_TONEDETECT:
			break;

		default:
			if (debug)
				printk(	"Unknown ioctl in zaphfc: "
					"cmd = %x, data = %x\n", cmd, 
					(unsigned)data);
			return -ENOTTY;
	}
	return 0;
}

static int zaphfc_zaptel_startup(struct zt_span *span) 
{
	struct hfc_card* hfc = span->pvt;
	int ret, i;

	if (debug)
		printk("zaptel startup on %s\n", span->name);
	
	if (!hfc) {
		printk(KERN_INFO "zaphfc: no card for span at startup\n");
		return -1;
	}
    
	if (span->flags & ZT_FLAG_RUNNING)
		return 0;

#ifdef ZAPHFC_HSS
	/* register each channel */
	for (i=0; i<2; i++) {
		ret = ixp425_hss_register(	zaphfc_transfer_hss,
						(void*)(span->chans+i));
	
		if (ret < 0){
			printk(	"Unable to start hss/pcm transfers "
				"for channel %d\n", i);
			return -1;
		}
		hfc->hss_start_bit[i] = ret;
	}
#else
	/* register on hss bus just for timing (per span)*/
	register_1khz_callback((timer_func)zaphfc_transfer, (void*)hfc);
#endif

	
	span->chans[2].flags &= ~ZT_FLAG_HDLC;
	span->chans[2].flags |= ZT_FLAG_BRIDCHAN;
	
	span->flags |= ZT_FLAG_RUNNING;

#ifdef ZAPHFC_HSS
	hfc_outb(hfc, hfc_B1_SSL, (hfc->hss_start_bit[0]/8) | 128 );
	hfc_outb(hfc, hfc_B1_RSL, (hfc->hss_start_bit[0]/8) | 128 | ((loopback)?64:0));
	
	hfc_outb(hfc, hfc_B2_SSL, (hfc->hss_start_bit[1]/8) | 128 );
	hfc_outb(hfc, hfc_B2_RSL, (hfc->hss_start_bit[1]/8) | 128 | ((loopback)?64:0));

	hfc_outb(hfc, hfc_MST_EMOD, 1 | 2 /*| 16 | 32 | 128*/ );
	hfc_outb(hfc, hfc_CONNECT, 2 | 4 | 16 | 32);

	if (debug) {
		printk("Registering %s on channel %d and %d\n",
			span->name,
			(hfc->hss_start_bit[0]/16), (hfc->hss_start_bit[1]/16));
	}
#endif

	hfc_outb(hfc, hfc_FIFO_EN,  hfc_FIFOEN_D | hfc_FIFOEN_B1 | hfc_FIFOEN_B2);

	/* activation request */
	hfc_outb(hfc, hfc_STATES, hfc_STATES_DO_ACTION | hfc_STATES_ACTIVATE);

	/* start t3 timer */
	if (!hfc->nt_mode) {
		hfc->t3.expires = jiffies + ZAPHFC_T3_PERIOD;
		add_timer(&hfc->t3);
	}

	return 0;
}


static int zaphfc_zaptel_shutdown(struct zt_span *span) 
{
	struct hfc_card* hfc = span->pvt;

	if (debug)
		printk("zaptel shutdown on %s\n", span->name);

#ifdef ZAPHFC_HSS
	ixp425_hss_unregister(hfc->hss_start_bit[0]);
	ixp425_hss_unregister(hfc->hss_start_bit[1]);
#else
	/* unregister on hss bus for timing
	 * NOTE: We are unregistering here twice (all devices)
	 * because we can not make sure that the right timer 
	 * would be removed. In that case the callback of card #0
	 * would be invoked with data of card #1 which would be
	 * not valid anymore. This has to be fixed in 1khz_timer.*/
	unregister_1khz_callback((timer_func)zaphfc_transfer);
	unregister_1khz_callback((timer_func)zaphfc_transfer);
#endif

	return 0;

}

static void zaphfc_zaptel_init(struct hfc_card* hfc)
{
	int i;

	/* reset info */
	memset(&hfc->span, 0x0, sizeof(struct zt_span));
	
	/* set name */
	sprintf(hfc->span.name, "ZAPHFC%d", hfc_card_count + 1);
	sprintf(hfc->span.desc, "HFC-S PCI A ISDN card %d [%s]", 
			hfc_card_count + 1,
			(hfc->nt_mode)?"NT":"TE");

	hfc->span.startup = zaphfc_zaptel_startup;
	hfc->span.shutdown= zaphfc_zaptel_shutdown;
	hfc->span.open	  = zaphfc_zaptel_open;
	hfc->span.close   = zaphfc_zaptel_close;
	hfc->span.ioctl   = zaphfc_zaptel_ioctl;
	
	hfc->span.chans   = hfc->chans;
	hfc->span.channels = 3;
	hfc->span.deflaw  = ZT_LAW_ALAW;
	hfc->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_CCS;
	hfc->span.offset = 0;
	init_waitqueue_head(&hfc->span.maintq);
	hfc->span.pvt = hfc;

	for (i = 0; i < hfc->span.channels; i++) {
		memset(&(hfc->chans[i]), 0x0, sizeof(struct zt_chan));
		sprintf(hfc->chans[i].name, "ZAPHFC%d/%d/%d", hfc_card_count + 1,0,i + 1);
		hfc->chans[i].pvt = hfc;
		hfc->chans[i].sigcap =  ZT_SIG_EM | ZT_SIG_CLEAR | ZT_SIG_FXSLS 
					| ZT_SIG_FXSGS | ZT_SIG_FXSKS | ZT_SIG_FXOLS 
					| ZT_SIG_FXOGS | ZT_SIG_FXOKS | ZT_SIG_CAS 
					| ZT_SIG_SF;
		hfc->chans[i].chanpos = i + 1; 
	}
	
	memset(hfc->drecbuf, 0x0, sizeof(hfc->drecbuf));
	hfc->chans[2].readchunk = hfc->drecbuf;

	memset(hfc->dtransbuf, 0x0, sizeof(hfc->dtransbuf));
	hfc->chans[2].maxbytes2transmit = hfc_D_FIFO_SIZE;
	hfc->chans[2].writechunk = hfc->dtransbuf;

	if (zt_register(&hfc->span,0))
		printk(KERN_WARNING "unable to register zaptel device!\n");
}


/**
 * Hardware initialization 
 */
static void zaphfc_debug_dchan_tx(struct hfc_card* hfctmp, int count)
{
	int x;
	
	if (!debug)
		return;

	printk(KERN_CRIT "zaphfc: card %d TX [ ", hfctmp->cardno);
	for (x=0; x<count; x++) {
		printk("%#2x ",hfctmp->dtransbuf[x]);
	}
	if (hfctmp->chans[2].eoftx == 1) {
		printk("] %d bytes\n", count);
	} else {
		printk("..] %d bytes\n", count);
	}
}

static void zaphfc_debug_dchan_rx(struct hfc_card* hfctmp)
{
	int x;
	
	if (!debug)
		return;

	printk(KERN_CRIT "zaphfc: card %d RX [ ", hfctmp->cardno);
	
	if (hfctmp->chans[2].eofrx) {
		/* dont output CRC == less user confusion */
		for (x=0; x < hfctmp->chans[2].bytes2receive - 2; x++) {
			printk("%#2x ", hfctmp->drecbuf[x]);
		}

		printk("] %d bytes\n", hfctmp->chans[2].bytes2receive - 2);

	} else {

		for (x=0; x < hfctmp->chans[2].bytes2receive; x++) {
			printk("%#2x ", hfctmp->drecbuf[x]);
		}

		printk("..] %d bytes\n", hfctmp->chans[2].bytes2receive);
	}
}



static void zaphfc_drecord(struct hfc_card* hfctmp)
{
	int count=0, maxlen=0, framelen=0;
	unsigned char *f1, *f2, *crcstat;
	unsigned short *z1, *z2, oldz2, newz2;

	hfctmp->chans[2].bytes2receive=0;
	hfctmp->chans[2].eofrx = 0;

	if (!hfctmp->drec_count)
		return;
	
	
	/* put the received data into the zaptel buffer
	  we'll call zt_receive() later when the timer fires. */
	f1 = (char *)(hfctmp->fifos + hfc_FIFO_DRX_F1);
	f2 = (char *)(hfctmp->fifos + hfc_FIFO_DRX_F2);

	/* nothing received, strange eh? */
	if (*f1 == *f2) 
		return; 

	z1 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_DRX_Z1 + (*f2 * 4));
	z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_DRX_Z2 + (*f2 * 4));

	/* calculate length of frame, including 2 bytes CRC and 1 byte STAT */
	count = read_z(*z1) - read_z(*z2);

	if (count < 0) { 
		/* ring buffer wrapped */
		count += hfc_D_FIFO_SIZE; 
	}

	count++;
	framelen = count;

	crcstat = (char *)(hfctmp->fifos + hfc_FIFO_DRX_ZOFF + read_z(*z1));

	if ((framelen < 4) || (*crcstat != 0x0)) {
		/* the frame is too short for a valid HDLC frame or the CRC is borked */
		printk(KERN_CRIT "zaphfc: empty HDLC frame or bad CRC received (framelen = %d, stat = %#x, card = %d).\n", framelen, *crcstat, hfctmp->cardno);
		oldz2 = read_z(*z2);
		*f2 = ((*f2 + 1) & hfc_MAX_DFRAMES) | (hfc_MAX_DFRAMES + 1);	/* NEXT!!! */
		/* recalculate z2, because Z2 is a function of F2 Z2(F2) and we INCed F2!!! */
		z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_DRX_Z2 + (*f2 * 4));
		*z2 = write_z((oldz2 + framelen) & hfc_ZMASK);
		hfctmp->drec_count--;
		/* skip short or broken frames */
		hfctmp->chans[2].bytes2receive = 0; 
		return;
	}

	count -= 1;	/* strip STAT */
	hfctmp->chans[2].eofrx = 1;

	if (count + read_z(*z2) <= hfc_D_FIFO_SIZE) {
		maxlen = count;
	} else {
		maxlen = hfc_D_FIFO_SIZE - read_z(*z2);
	}

	/* copy first part */
	memcpy(hfctmp->drecbuf, (char *)(hfctmp->fifos + hfc_FIFO_DRX_ZOFF + read_z(*z2)), maxlen);
	hfctmp->chans[2].bytes2receive += maxlen; 

	count -= maxlen;
	if (count > 0) {
		/* ring buffer wrapped, copy rest from start of d fifo */
		memcpy(hfctmp->drecbuf + maxlen, (char *)(hfctmp->fifos + hfc_FIFO_DRX_ZOFF), count);
		hfctmp->chans[2].bytes2receive += count; 
	}

	/* frame read */
	oldz2 = read_z(*z2);
	newz2 = (oldz2 + framelen) & hfc_ZMASK;
	*f2 = ((*f2 + 1) & hfc_MAX_DFRAMES) | (hfc_MAX_DFRAMES + 1);	/* NEXT!!! */
	/* recalculate z2, because Z2 is a function of F2 Z2(F2) and we INCed F2!!! */
	z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_DRX_Z2 + (*f2 * 4));
	*z2 = write_z(newz2);
	hfctmp->drec_count--; 
	
	
	/* debug output */
	zaphfc_debug_dchan_rx(hfctmp);
}

static void zaphfc_dtransfer(struct hfc_card* hfctmp)
{
	// we are called with irqs disabled from the irq handler
	int count, maxlen, total;
	unsigned char *f1, *f2, newf1;
	unsigned short *z1, *z2, newz1;
	int frames, freebytes;

	if (hfctmp->chans[2].bytes2transmit == 0)
		return;

	if (hfctmp->state != 7 && hfctmp->state != 8 && !hfctmp->nt_mode && !timer_pending(&hfctmp->t3) && hfctmp->span.alarms == ZT_ALARM_NONE) {
	
		/* activation request */
		hfc_outb(hfctmp, hfc_STATES, hfc_STATES_DO_ACTION | hfc_STATES_ACTIVATE);
		/* start t3 timer */
		hfctmp->t3.expires = jiffies + ZAPHFC_T3_PERIOD;
		add_timer(&hfctmp->t3);
	}

	f1 = (char *)(hfctmp->fifos + hfc_FIFO_DTX_F1);
	f2 = (char *)(hfctmp->fifos + hfc_FIFO_DTX_F2);
	z1 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_DTX_Z1 + (*f1 * 4));
	z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_DTX_Z2 + (*f1 * 4));

	frames = (*f1 - *f2) & hfc_FMASK;
	if (frames < 0) {
		frames += hfc_MAX_DFRAMES + 1;
	}

	if (frames >= hfc_MAX_DFRAMES) {
		printk(KERN_CRIT "zaphfc: dchan tx fifo total number of frames exceeded!\n");
		return;
	}

	freebytes = read_z(*z2) - read_z(*z1);
	if (freebytes <= 0) {
		freebytes += hfc_D_FIFO_SIZE;
	}
	
	count = hfctmp->chans[2].bytes2transmit;

	total = count;
	if (freebytes < count) {
		printk(KERN_CRIT "zaphfc: dchan tx fifo not enough free bytes! (z1=%d, z2=%d)\n",read_z(*z1),read_z(*z2));
		return;
	}

	newz1 = (read_z(*z1) + count) & hfc_ZMASK;
	newf1 = ((*f1 + 1) & hfc_MAX_DFRAMES) | (hfc_MAX_DFRAMES + 1);	// next frame

	if (count > 0) {

		zaphfc_debug_dchan_tx(hfctmp, count);
		
		maxlen = hfc_D_FIFO_SIZE - read_z(*z1);
		if (maxlen > count) {
			maxlen = count;
		}

		memcpy((char *)(hfctmp->fifos + hfc_FIFO_DTX_ZOFF + read_z(*z1)),hfctmp->chans[2].writechunk, maxlen);
		count -= maxlen;

		if (count > 0) {
			memcpy((char *)(hfctmp->fifos + hfc_FIFO_DTX_ZOFF),(char *)(hfctmp->chans[2].writechunk + maxlen), count);
		}
	}

	*z1 = write_z(newz1);

	if (hfctmp->chans[2].eoftx == 1) {
		*f1 = newf1;
		z1 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_DTX_Z1 + (*f1 * 4));
		*z1 = write_z(newz1);
		hfctmp->chans[2].eoftx = 0;
	}
}


#ifndef ZAPHFC_HSS
static void zaphfc_btransfer(struct hfc_card* hfctmp, int whichB)
{
	/* we are called with irqs disabled from the irq handler */
	unsigned char *f1, *f2;
	unsigned short *z1, *z2, newz1;
	unsigned char* write_ptr;
	unsigned char* write_chunk;
	int freebytes;

	if (whichB == 1) {
		f1 = (char *)(hfctmp->fifos + hfc_FIFO_B1TX_F1);
		f2 = (char *)(hfctmp->fifos + hfc_FIFO_B1TX_F2);
		z1 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B1TX_Z1 + (*f1 * 4));
		z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B1TX_Z2 + (*f1 * 4));
		write_ptr = (char *)(hfctmp->fifos + hfc_FIFO_B1TX_ZOFF + read_z(*z1));
		write_chunk = hfctmp->chans[0].writechunk;
	} else {
		f1 = (char *)(hfctmp->fifos + hfc_FIFO_B2TX_F1);
		f2 = (char *)(hfctmp->fifos + hfc_FIFO_B2TX_F2);
		z1 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B2TX_Z1 + (*f1 * 4));
		z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B2TX_Z2 + (*f1 * 4));
		write_ptr = (char *)(hfctmp->fifos + hfc_FIFO_B2TX_ZOFF + read_z(*z1));
		write_chunk = hfctmp->chans[1].writechunk;
	}

	/* calculate free bytes */
	freebytes = read_z(*z2) - read_z(*z1);
	if (freebytes <= 0) {
		freebytes += hfc_B_FIFO_SIZE;
	}

	/* check if there is enough free space */
	if (freebytes < ZT_CHUNKSIZE) {
		if (hfctmp->last_msg < jiffies) {
			printk("Write buffer is full ?!?!\n");
			hfctmp->last_msg = jiffies + 2000;
		}
		return;
	}

	/* calculate new pointer where we are */
	newz1 = read_z(*z1) + ZT_CHUNKSIZE;
	if (newz1 >= (hfc_B_FIFO_SIZE + hfc_B_SUB_VAL)) {
		if (debug && newz1 > (hfc_B_FIFO_SIZE + hfc_B_SUB_VAL)) {
		printk( "Upsi, strange write pointer for card %d on channel %c. "
			"No panic, I will fix that.\n", hfctmp->cardno, (whichB==1)?'a':'b');
	}
		newz1 = hfc_B_SUB_VAL;
	}

	/* write a chunk */
	memcpy( write_ptr, write_chunk, ZT_CHUNKSIZE);

	/* set it */
	*z1 = write_z(newz1);
}


static void zaphfc_brecord(struct hfc_card *hfctmp, char whichB)
{
	/* we are called with irqs disabled from the irq handler */
	volatile unsigned char *f1, *f2;
	volatile unsigned short *z1, *z2, newz2;
	int count, bytes;
	unsigned char* read_ptr;
	static unsigned long last_j=0;
	struct zt_chan* chan;

	if (whichB == 1) {
		f1 = (char *)(hfctmp->fifos + hfc_FIFO_B1RX_F1);
		f2 = (char *)(hfctmp->fifos + hfc_FIFO_B1RX_F2);
		z1 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B1RX_Z1 + (*f1 * 4));
		z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B1RX_Z2 + (*f1 * 4));
		read_ptr = (char *)(hfctmp->fifos + hfc_FIFO_B1RX_ZOFF + read_z(*z2));
		chan = &hfctmp->chans[0];
	} else {
		f1 = (char *)(hfctmp->fifos + hfc_FIFO_B2RX_F1);
		f2 = (char *)(hfctmp->fifos + hfc_FIFO_B2RX_F2);
		z1 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B2RX_Z1 + (*f1 * 4));
		z2 = (unsigned short *)(hfctmp->fifos + hfc_FIFO_B2RX_Z2 + (*f1 * 4));
		read_ptr = (char *)(hfctmp->fifos + hfc_FIFO_B2RX_ZOFF + read_z(*z2));
		chan = &hfctmp->chans[1];
	}

	count = 0;
	newz2 = read_z(*z2);
	bytes = read_z(*z1) - read_z(*z2);

	if (bytes < 0) {
		bytes += hfc_B_FIFO_SIZE;
	}

	while(bytes > ZT_CHUNKSIZE) {

	//	memcpy(chan->readchunk, read_ptr, ZT_CHUNKSIZE);
		chan->readchunk = read_ptr;

		bytes -= ZT_CHUNKSIZE;
		newz2 += ZT_CHUNKSIZE;
		read_ptr += ZT_CHUNKSIZE;
		count ++;

		if (newz2 >= (hfc_B_FIFO_SIZE + hfc_B_SUB_VAL)) { 
			if (debug && newz2 > (hfc_B_FIFO_SIZE + hfc_B_SUB_VAL)) {
				printk( "Upsi, strange read pointer for card %d on channel %c. "
					"No panic, I will fix that.\n", hfctmp->cardno, (whichB==1)?'a':'b');
			}
			read_ptr -= newz2 - hfc_B_FIFO_SIZE - hfc_B_SUB_VAL - hfc_B_FIFO_SIZE;
			newz2     = hfc_B_SUB_VAL;
		}
	}

	if (debug && last_j < jiffies && count > 1) {
		printk("readchunkdropcounter is %d\n", count);
		last_j = jiffies + 100;
	}

	*z2 = write_z(newz2);

	if (whichB == 1) {
		zt_ec_chunk(&hfctmp->chans[0], hfctmp->chans[0].readchunk, hfctmp->chans[0].writechunk);
	} else {
		zt_ec_chunk(&hfctmp->chans[1], hfctmp->chans[1].readchunk, hfctmp->chans[1].writechunk);
	}
}


static void zaphfc_transfer(struct hfc_card* hfctmp)
{

	if (hfctmp->span.flags & ZT_FLAG_RUNNING) {

		/* clear dchan buffer */
		hfctmp->chans[2].bytes2transmit = 0;
		hfctmp->chans[2].maxbytes2transmit = hfc_D_FIFO_SIZE;

		zt_transmit(&(hfctmp->span));

		zaphfc_btransfer(hfctmp,1);
		zaphfc_btransfer(hfctmp,2);
		zaphfc_dtransfer(hfctmp);
	
	}

	zaphfc_brecord(hfctmp,1);
	zaphfc_brecord(hfctmp,2);

	if (hfctmp->drec_count > 0) {
		/* dchan data to read */
		zaphfc_drecord(hfctmp);

		if (hfctmp->chans[2].bytes2receive > 0) {
			/* debug output */
			zaphfc_debug_dchan_rx(hfctmp);
		}
	} else {
		/* hmm....ok, let zaptel receive nothing */
		hfctmp->chans[2].bytes2receive = 0;
	}

	if (hfctmp->span.flags & ZT_FLAG_RUNNING) {
		zt_receive(&(hfctmp->span));
	}

}


#else

/**
 * This function tranfers the data between hss bus and zaptel.
 */
static void zaphfc_transfer_hss(unsigned char* rx_ptr,
				unsigned char* tx_ptr,
				void* data)
{
	struct zt_chan* chan;
	struct hfc_card* hfctmp;
	int channel;
	
	chan	= (struct zt_chan*)data;;
	hfctmp	= chan->pvt;
	channel = chan->chanpos - 1;
	
	/* just for first channel */
	if (channel == 0 && (hfctmp->span.flags & ZT_FLAG_RUNNING)) {

		/* clear dchan buffer */
		hfctmp->chans[2].bytes2transmit = 0;

		zt_transmit(chan->span);
		
		zaphfc_dtransfer(hfctmp);
		zaphfc_drecord(hfctmp);
	}
	
#ifdef ZAPHFC_ZAP_LOOPBACK
	memcpy(	tx_ptr,
		rx_ptr,
		ZT_CHUNKSIZE);
#else	
	chan->readchunk = rx_ptr;
	
	memcpy(	tx_ptr,
		chan->writechunk,
		ZT_CHUNKSIZE);
#endif
	
	// !!! tess test test
	zt_ec_chunk(chan, chan->readchunk, chan->writechunk);
		
	if (channel == 1 && (hfctmp->span.flags & ZT_FLAG_RUNNING))
		zt_receive(chan->span);

}

#endif /* ZAPHFC_HSS */

static void zaphfc_change_to_f3(struct hfc_card* hfc)
{
	if (hfc->state == 7) /* should not happen */ 
		return;

	/* activation failed */
	hfc_outb(hfc, hfc_STATES, 0x40); /* set to deactivated */
	hfc->span.alarms = ZT_ALARM_RED;
	zt_alarm_notify(&hfc->span);
}


static void zaphfc_t3_expired(unsigned long data)
{
	struct hfc_card* hfc = (struct hfc_card*)data;
	
	if (!hfc) {
		printk("Invalid data on t3 timer\n");
		return;
	}

	if (debug)
		printk(KERN_INFO "T3 timer expired for card #%d\n", hfc->cardno);

	zaphfc_change_to_f3(hfc);
}


static void zaphfc_t4_expired(unsigned long data)
{
	struct hfc_card* hfc = (struct hfc_card*)data;
	
	if (!hfc) {
		printk("Invalid data on T4 timer\n");
		return;
	}

	if (debug)
		printk(KERN_INFO "T4 timer expired for card #%d\n", hfc->cardno);

	zaphfc_change_to_f3(hfc);
}


static void zaphfc_state_changed_nt(struct hfc_card* hfctmp, int state)
{
	if (debug)
		printk(KERN_CRIT "zaphfc: card %d layer 1 state = G%d\n", hfctmp->cardno, state);

	switch (state) {
		case 2:
			hfc_outb(hfctmp, hfc_STATES, hfc_STATES_ACTIVATE | hfc_STATES_DO_ACTION | hfc_STATES_NT_G2_G3);
			break;

		case 3:
	    		/* fix to G3 state (see specs) */
			hfc_outb(hfctmp, hfc_STATES, hfc_STATES_LOAD_STATE | 3);
			break;
	}
}


static void zaphfc_state_changed_te(struct hfc_card* hfctmp, int state)
{
	if (debug)
		printk(KERN_CRIT "zaphfc: card %d layer 1 state = F%d\n", hfctmp->cardno, state);

	switch (state) {

		case 3:
			/* should we here reactivate if timer t3 is still running? */
			/* set red alarm */
			if (hfctmp->state == 7 || hfctmp->state == 8) {
				/* start T4 timer */
				if (debug)
					printk(KERN_INFO "Starting T4 timer for cardno# %d\n", hfctmp->cardno);
				hfctmp->t4.expires = jiffies + ZAPHFC_T4_PERIOD;
				add_timer(&hfctmp->t4);
				return;
			}
			hfctmp->span.alarms = ZT_ALARM_RED;
			zt_alarm_notify(&hfctmp->span);
			break;

		case 7: /* activated */
			/* stop t3 timer */
			if (debug)
				printk(KERN_INFO "Stopping T3 timer for cardno #%d\n", hfctmp->cardno);
			del_timer(&hfctmp->t3);
			del_timer(&hfctmp->t4);
			/* clear alarms */
			hfctmp->span.alarms = ZT_ALARM_NONE;
			zt_alarm_notify(&hfctmp->span);
			break;

		case 8: /* lost framing */
			break;
	}

}


static void zaphfc_state_changed(struct hfc_card* hfctmp)
{
	unsigned char state = hfc_inb(hfctmp,hfc_STATES)  & hfc_STATES_STATE_MASK;

	if (debug)
	  printk("state on channel %d changed to %d\n", (hfctmp->hss_start_bit[0]/16), state);

	if (hfctmp->nt_mode)
		zaphfc_state_changed_nt(hfctmp, state);
	else {
	  /* force Layer 1 up */
	  if (state == 3)
	    hfc_outb(hfctmp, hfc_STATES,
		     hfc_STATES_DO_ACTION |
		     hfc_STATES_ACTIVATE |
		     hfc_STATES_NT_G2_G3);
	  zaphfc_state_changed_te(hfctmp, state);
	}

	 hfctmp->state = state;
}


static void zaphfc_isr_s1(struct hfc_card* hfctmp, unsigned char s1)
{
	if (!s1)
		return;

	if (s1 & hfc_INTS_DREC) {
		hfctmp->drec_count++;
	}
	
	if (s1 & hfc_INTS_L1STATE ) {
		zaphfc_state_changed(hfctmp);
	}
}


static void zaphfc_isr_s2(struct hfc_card* hfctmp, unsigned char s2)
{
	if ((s2 & 128) ) {
//		printk(	"Fatal Error: Synchronisation lost. PCI "
//			"performance too low for HFC-S PCI A(%d).\n", 
//			hfctmp->cardno);
#ifdef ZAPHFC_REBOOT_ON_SYNC_LOST
		panic("Ahhh panic... WHat to do? What to do :)\n");
#endif
	}
}


static irqreturn_t zaphfc_isr(	int irq __unused, 
			void *dev_id, 
			struct pt_regs *regs __unused) 
{
	unsigned char stat, s1, s2;
	struct hfc_card* hfctmp = dev_id;

	if (!hfctmp)
		return IRQ_HANDLED;

	stat = hfc_inb(hfctmp, hfc_STATUS);
		
	if ((stat & hfc_STATUS_ANYINT) == 0)
		return IRQ_HANDLED;

	s1 = hfc_inb(hfctmp, hfc_INT_S1);
	s2 = hfc_inb(hfctmp, hfc_INT_S2);

	zaphfc_isr_s1(hfctmp, s1);
	zaphfc_isr_s2(hfctmp, s2);
	
	return IRQ_HANDLED;
}


static void zaphfc_reset(struct hfc_card* hfctmp)
{
	/* enable memio */
	pci_write_config_word(hfctmp->pci_dev, PCI_COMMAND, PCI_COMMAND_MEMORY);

	/* disable interrupts */
	hfc_outb(hfctmp, hfc_INT_M2, 0);

	/* set as master */
	pci_set_master(hfctmp->pci_dev);

	/* softreset */
	hfc_outb(hfctmp, hfc_CIRM, hfc_CIRM_RESET);
	mdelay(30);
	hfc_outb(hfctmp, hfc_CIRM,0);
	mdelay(20);

	/* disable fifo */
	hfc_outb(hfctmp, hfc_FIFO_EN, 0);
	hfc_outb(hfctmp, hfc_TRM, 2);
	
	/* set NT or TE mode */
	if (hfctmp->nt_mode) {
		/* ST-Bit delay for NT-Mode */
		hfc_outb(hfctmp, hfc_CLKDEL, CLKDEL_NT);
    	} else {
		/* ST-Bit delay for TE-Mode */
		hfc_outb(hfctmp, hfc_CLKDEL, CLKDEL_TE);
    	}

	/* S/T Auto awake */
	hfc_outb(hfctmp, hfc_SCTRL_E, hfc_SCTRL_E_AUTO_AWAKE);

	/* transparent mode */
 	hfc_outb(hfctmp, hfc_CTMT, hfc_CTMT_TRANSB1 | hfc_CTMT_TRANSB2);

	/* set up interrupt mask */
	hfc_outb(hfctmp, hfc_INT_M1, 	hfc_INTS_DTRANS | 
					hfc_INTS_DREC | 
					hfc_INTS_L1STATE);
	
	/* Clear already pending ints */
	hfc_inb(hfctmp, hfc_INT_S1);

	if (hfctmp->master) {
		/* HFC Master Mode */
		hfc_outb(hfctmp, hfc_MST_MODE, hfc_MST_MODE_MASTER);
	} else {
     		/* HFC Slave Mode */
		hfc_outb(hfctmp, hfc_MST_MODE, hfc_MST_MODE_SLAVE);
	}

	if (hfctmp->nt_mode) {
		/* set tx_lo mode, error in datasheet ! */
    		hfc_outb(hfctmp, hfc_SCTRL, 3 | hfc_SCTRL_NONE_CAP | hfc_SCTRL_MODE_NT);
	} else {
		/* set tx_lo mode, error in datasheet ! */
    		hfc_outb(hfctmp, hfc_SCTRL, 3 | hfc_SCTRL_NONE_CAP | hfc_SCTRL_MODE_TE);
	}
    
	hfc_outb(hfctmp, hfc_SCTRL_R, 3);
#ifdef ZAPHFC_HSS
	hfc_outb(hfctmp, hfc_CONNECT, 2 | 4 | 16 | 32);
#else
	hfc_outb(hfctmp, hfc_CONNECT, 0);
#endif
	hfc_outb(hfctmp, hfc_CIRM, 0x80 | 0x40);
    
	/* Finally enable IRQ output */
	hfc_outb(hfctmp, hfc_INT_M2, hfc_M2_IRQ_ENABLE);

	/* clear pending ints */
	hfc_inb(hfctmp, hfc_INT_S1); 
	hfc_inb(hfctmp, hfc_INT_S2);
}


static int zaphfc_probe(struct pci_dev *dev, 
			const struct pci_device_id *id __unused)
{
	struct hfc_card* hfctmp;
	int ret;
	
	if (pci_enable_device(dev))
		return -1;

	ledman_cmd(LEDMAN_CMD_FLASH, (hfc_card_count)?5:4);

	pci_set_master(dev);

	/* allocate memory for hfc_card structure */
	hfctmp = kmalloc(sizeof(struct hfc_card), GFP_KERNEL);
	if (!hfctmp) {
		printk(KERN_WARNING "zaphfc: could not allocate memory\n");
		pci_disable_device(dev);
		return -ENOMEM;
	}
	memset(hfctmp, 0, sizeof(struct hfc_card));
	hfctmp->pci_dev = dev;
	hfctmp->irq	= dev->irq;
	hfctmp->pci_io  = (unsigned char*)dev->resource[1].start;
	hfctmp->master  = (1<<hfc_card_count) & master;
	hfctmp->nt_mode = (1<<hfc_card_count) & mode;

	/* check for io memory */
	if (!hfctmp->pci_io) {
		printk(KERN_WARNING "zaphfc: no iomem!\n");
		kfree(hfctmp);
		pci_disable_device(dev);
		return -ENOMEM;
	}

	/* allocating fifo memory */
	hfctmp->fifomem = pci_alloc_consistent(
					hfctmp->pci_dev, 
					65536, 
					&hfctmp->dma_handle);

	if (!hfctmp->fifomem) {
		printk(KERN_WARNING "zaphfc: could not allocate memory for fifo!\n");
		kfree(hfctmp);
		iounmap((void *) hfctmp->pci_io);
		pci_disable_device(dev);
		return -ENOMEM;
	}
	
	memset(hfctmp->fifomem, 0x0, 65536);
	pci_write_config_dword(hfctmp->pci_dev, 0x80, (u_int)hfctmp->dma_handle);
	hfctmp->pci_io = ioremap((ulong) hfctmp->pci_io, 256);
	hfctmp->fifos = (void*)(((ulong) hfctmp->fifomem));
	
	ret = request_irq(hfctmp->irq, zaphfc_isr, SA_INTERRUPT | SA_SHIRQ, "zaphfc", hfctmp);
	
	if (ret) {
		printk(KERN_WARNING "zaphfc: unable to register irq\n");
		pci_free_consistent(hfctmp->pci_dev, 65536, hfctmp->fifomem, hfctmp->dma_handle);
		kfree(hfctmp);
		iounmap((void *) hfctmp->pci_io);
		pci_disable_device(dev);
		return -EIO;
	}
	
	pci_write_config_word(hfctmp->pci_dev, PCI_COMMAND, PCI_COMMAND_MEMORY);

	/* disable interrupts */
	hfc_outb(hfctmp, hfc_INT_M1, 0x0);
	hfc_outb(hfctmp, hfc_INT_M2, 0x0);
	
	/* initializing timers */
	init_timer(&hfctmp->t3);
	hfctmp->t3.data 	= (unsigned long) hfctmp;
	hfctmp->t3.function	= zaphfc_t3_expired;

	init_timer(&hfctmp->t4);
	hfctmp->t4.data 	= (unsigned long) hfctmp;
	hfctmp->t4.function	= zaphfc_t4_expired;

	/* register with zaptel */
	zaphfc_zaptel_init(hfctmp);

	/* register to our device list */
	pci_set_drvdata(dev, hfctmp); 
	if (hfc_cards) {
		hfc_cards->next=hfctmp;
		hfctmp->prev=hfc_cards;
		hfc_cards = hfctmp;
	} else
		hfc_cards = hfctmp;
	hfctmp->cardno = hfc_card_count++;

	/* reset device */
	zaphfc_reset(hfctmp);

	printk("Registered %s as %s %s %s\n",
			pci_name(dev),
			(hfctmp->master)?"Master":"Slave",
			(hfctmp->nt_mode)?"NT":"TE",
			(loopback)?"in loopback mode":"");

	return 0;
}


static void zaphfc_remove(struct pci_dev *dev)
{
	struct hfc_card* hfctmp = pci_get_drvdata(dev);

	/* unregister from zaptel */
	zt_unregister(&hfctmp->span);
	
	/* remove from list */
	if (hfctmp->next)
		hfctmp->next->prev = hfctmp->prev;
	if (hfctmp->prev)
		hfctmp->prev->next = hfctmp->next;
	hfc_card_count--;

	/* disable interrupts and reset all pending interrupts */
	hfc_outb(hfctmp, hfc_INT_M2, 0);
	hfc_inb(hfctmp, hfc_INT_S1);
	
	mdelay(30);
	
	free_irq(hfctmp->irq, hfctmp);

	/* soft reset */
	/* XXX does not work :-( */
//	hfc_outb(hfctmp, hfc_CIRM, hfc_CIRM_RESET);
//	mdelay(30);
//	hfc_outb(hfctmp, hfc_CIRM,0);

	/* disable memio and bustmaster */
	pci_write_config_word(hfctmp->pci_dev, PCI_COMMAND, 0);

	/* free memory */
	pci_free_consistent(hfctmp->pci_dev, 65536, hfctmp->fifomem, hfctmp->dma_handle);
	iounmap((void *) hfctmp->pci_io);
	hfctmp->pci_io = NULL;

	/* free rest of memory */
	kfree(hfctmp);

	printk("HFC Card removed\n");
}

/**
 * Generic PCI Driver Registration
 */
static struct pci_device_id zaphfc_pci_ids[] = {
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_2BD0) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B000) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B006) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B007) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B008) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B009) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B00A) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B00B) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B00C) },
        {PCI_DEVICE( PCI_VENDOR_ID_CCD, 	PCI_DEVICE_ID_CCD_B100) },
        {PCI_DEVICE( PCI_VENDOR_ID_ABOCOM, 	PCI_DEVICE_ID_ABOCOM_2BD1) },
        {PCI_DEVICE( PCI_VENDOR_ID_ASUSTEK, 	PCI_DEVICE_ID_ASUSTEK_0675) },
        {PCI_DEVICE( PCI_VENDOR_ID_BERKOM, 	PCI_DEVICE_ID_BERKOM_T_CONCEPT) },
        {PCI_DEVICE( PCI_VENDOR_ID_BERKOM, 	PCI_DEVICE_ID_BERKOM_A1T) },
        {PCI_DEVICE( PCI_VENDOR_ID_ANIGMA, 	PCI_DEVICE_ID_ANIGMA_MC145575) },
        {PCI_DEVICE( PCI_VENDOR_ID_ZOLTRIX, 	PCI_DEVICE_ID_ZOLTRIX_2BD0) },
        {PCI_DEVICE( PCI_VENDOR_ID_DIGI, 	PCI_DEVICE_ID_DIGI_DF_M_IOM2_E) },
        {PCI_DEVICE( PCI_VENDOR_ID_DIGI, 	PCI_DEVICE_ID_DIGI_DF_M_E) },
        {PCI_DEVICE( PCI_VENDOR_ID_DIGI, 	PCI_DEVICE_ID_DIGI_DF_M_IOM2_A) },
        {PCI_DEVICE( PCI_VENDOR_ID_DIGI, 	PCI_DEVICE_ID_DIGI_DF_M_A) },
	{PCI_DEVICE( 0x182d, 0x3069) },
	{0 },
};

static struct pci_driver zaphfc_pci_driver = {
	.name 	  = "zaphfc",
	.id_table = zaphfc_pci_ids,
	.probe 	  = zaphfc_probe,
	.remove   = zaphfc_remove,
};

static irqreturn_t overvoltage_overcurrent_isr(	int irq __unused,
				void *data __unused,
				struct pt_regs* regs __unused)
{
	int r;

	/* The number of overcurrents/overvoltages with a certain distance */
	static int overcurrent_overvoltage_cnt = 0;
	
	/* The value of jiffies at the last occurence of the interrupt */
	static unsigned long last_event = 0;
	
	unsigned long tdiff;
	
	/* Nachschauen, ob das PwrWarn auf 0 ist
	 */
	gpio_line_get(IXP425_ISDN_PwrWarn, &r);

	/* merken, dass einen Überspannungs/Überstromsituation 
	 * aufgetreten ist.
	 * Wenn dies innerhalb eines gewissen Zeitraums nochmal passiert, so soll der GPIO
	 * deaktiviert werden.
	 */
	
	/* calculate the interval in [sec] between the last and this interrupt */
	tdiff = (jiffies - last_event) / HZ;

	if (tdiff < 3) {
		/* The Last interrupt occured less then 3 seconds ago */
		overcurrent_overvoltage_cnt++;
	} else {
		/* The Last interrupt occured less more then 3 seconds ago */
		last_event = jiffies;
	}

	/* if there are more then 10 interrupts within 3 seconds, power the device of because of safty reasons */
	if( overcurrent_overvoltage_cnt > 10 ) {
		/* power off */
		gpio_line_set(IXP425_ISDN_PwrOn, 1);	
		printk("Short circuit detected. Powering ISDN off\n");
	}
	
	#warning after an poweroff the ISDN Powersupply will not be powered on again
	
	return IRQ_HANDLED;

}

static int zaphfc_init(void)
{
	pci_register_driver(&zaphfc_pci_driver);

	ledman_cmd(LEDMAN_CMD_OFF, 5);
	ledman_cmd(LEDMAN_CMD_OFF, 4);

	if (hfc_card_count == 0) {
		printk("No cards found\n");
		return -ENODEV;
	}

	if (debug)
		printk("Found %d card(s)\n", hfc_card_count);

	/* NT configured? */
	if (mode) {
		int ret;	
		gpio_line_config(IXP425_ISDN_PwrOn, IXP425_GPIO_OUT);
		
		if(nt_protection) {
			/* Configure GPIO 0 as input in order to detect the Overvoltage / Overcurrent situation */
			gpio_line_config(0, IXP425_GPIO_IN | 
				 	    IXP425_GPIO_TRANSITIONAL);
		
			ret = request_irq(IXP425_ISDN_PwrWarn, overvoltage_overcurrent_isr,
		                               SA_TRIGGER_FALLING | SA_TRIGGER_RISING,
						"overvoltage_isr", NULL);
			if (ret) {
				/* error */
				printk("unable to register ISR for overcurrent/overvoltage detection %d\n", ret);
			}
		}
		
		/* power on */
		gpio_line_set(IXP425_ISDN_PwrOn, 0);
	}

	zaphfc_proc_register();
	
	return 0;
}

static void zaphfc_exit(void)
{ 
	pci_unregister_driver(&zaphfc_pci_driver);
	
	/* power off */
	gpio_line_set(IXP425_ISDN_PwrOn, 1);

	if(nt_protection) {
		free_irq(IXP425_ISDN_PwrWarn, NULL);
	}

	zaphfc_proc_unregister();
}

module_init(zaphfc_init);
module_exit(zaphfc_exit);

MODULE_DESCRIPTION("HFC-S PCI A Zaptel Driver");
MODULE_AUTHOR("Henry Margies <h.margies@maxina.de>");
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(pci, zaphfc_pci_ids);

MODULE_PARM(debug,"i");
MODULE_PARM(mode,"i");
MODULE_PARM(master,"i");
MODULE_PARM(nt_protection,"i");
